/* $Header: /src2/opennurbs/example_gl/example_gl.cpp 9     7/24/01 4:53p Dalelear $ */
/* $NoKeywords: $ */
//
// Copyright (c) 1993-1998 Robert McNeel & Associates. All rights reserved.
// Rhinoceros is a registered trademark of Robert McNeel & Assoicates.
//
// THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
// ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
// MERCHANTABILITY ARE HEREBY DISCLAIMED.
//				
////////////////////////////////////////////////////////////////

#include "../opennurbs.h"
#include "../opennurbs_gl.h"
#include <GL/GLaux.h>   // Open GL auxillary functions

#if defined(_WINDOWS) || defined(WIN32)
#define MY_GL_CALLBACK CALLBACK
#else
#define MY_GL_CALLBACK
#endif

// Before working on this file, be sure to study the OpenNURBS toolkit 
// file example_read.cpp and to read chapters 1 through 11 of the 
// _Open_GL_Programming_Guide_.
//
// This file contains simple example in modeled after those found in
// the _Open_GL_Programming_Guide_.  The nuts and bolts functions that
// demonstrate how to use Open GL to display OpenNURBS geometry are in
// opennurbs_gl.cpp.

////////////////////////////////////////////////////////////////////////
//
// CModel is a simple class used to hold the contents of an OpenNURBS 3dm
// file.  It contains dynamic arrays of
//    CModelRenderLight (rendering lights and their attributes)
//    CModelObject      (geometric objects and their attributes)
//    ON_Material       (rendering materials)
//    ON_Light          (rendering lights)
//
////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////

class CModel;

class CModelRenderLight
{
public:
  CModelRenderLight();
  ~CModelRenderLight();
  CModelRenderLight(const CModelRenderLight&);
  CModelRenderLight& operator=(const CModelRenderLight&);

  ON_Light m_light;
  ON_3dmObjectAttributes m_attributes;
};

CModelRenderLight::CModelRenderLight()
{
}

CModelRenderLight::~CModelRenderLight()
{
}

CModelRenderLight::CModelRenderLight(const CModelRenderLight& src) 
             : m_light(src.m_light), 
               m_attributes(src.m_attributes)
{
}

CModelRenderLight& CModelRenderLight::operator=(const CModelRenderLight& src)
{
  if ( this != &src )
  {
    m_light = src.m_light;
    m_attributes = src.m_attributes;
  }
  return *this;
}

////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////

class CModelObject
{
public:
  CModelObject();
  ~CModelObject();
  CModelObject(const CModelObject&);
  CModelObject& operator=(const CModelObject&);

  void Dump( ON_TextLog& ) const;

  // Get material to use for GL rendering
  void GetMaterial( const CModel&, ON_Material& ) const; 

  const ON_Geometry* m_geometry;
  ON_3dmObjectAttributes m_attributes;

private:
  void Destroy();
  unsigned int* m_ref_count; // reference counts used to avoid expensive object copying
};

CModelObject::CModelObject() : m_geometry(0), m_ref_count(0)
{
}

void CModelObject::Destroy()
{
  if ( m_ref_count ) {
    if ( *m_ref_count > 0 )
      (*m_ref_count)--;
    if ( *m_ref_count <= 0 ) 
    {
      delete m_ref_count;
      m_ref_count = 0;
    }
  }
  if ( !m_ref_count && m_geometry )
  {
    delete m_geometry;
    m_geometry = 0;
  }
}

CModelObject::~CModelObject()
{
  Destroy();
}

CModelObject::CModelObject(const CModelObject& src) 
             : m_geometry(0),
               m_ref_count(0)
{
  *this = src;
}

CModelObject& CModelObject::operator=(const CModelObject& src)
{
  if ( this != &src )
  {
    Destroy();
    m_geometry = src.m_geometry;
    m_attributes = src.m_attributes;
    m_ref_count = src.m_ref_count;
    if ( m_geometry ) {
      if ( m_ref_count )
        (*m_ref_count)++;
      else {
        m_ref_count = new unsigned int;
        *m_ref_count = 2; // 2 because this and src reference same m_geometry
        CModelObject* srcptr = (CModelObject*)&src; // need to defeat const
        srcptr->m_ref_count = m_ref_count;
      }
    }
  }
  return *this;
}

////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////

class CModel
{
public:
  CModel();
  ~CModel();

  BOOL Read( const char* ); // read model from file

  void Destroy(); // leaves model empty and ready to be reused

  // pedegodgical example of all the things in an OpenNURBS 3dm file.

  // start section
  int m_version;
  int m_file_length;
  ON_String m_sStartSectionComments;

  // Properties include revision history, notes, information about
  // the applicaton that created the file, and an option preview image.
  ON_3dmProperties m_properties;

  // Settings include tolerance, and unit system, and defaults used
  // for creating views and objects.
  ON_3dmSettings   m_settings;

  ON_SimpleArray<ON_Bitmap*>       m_bitmap_table;
  ON_ClassArray<ON_Material>       m_material_table;
  ON_ClassArray<ON_Layer>          m_layer_table;
  ON_ClassArray<ON_Group>          m_group_table;
  ON_ClassArray<CModelRenderLight> m_light_table;
  ON_ClassArray<CModelObject>      m_object_table;

  // gets world coordinate bounding box of objects
  BOOL GetBoundingBox( ON_BoundingBox& ) const;

  // gets a default view of the region in the box
  void DefaultView( const ON_BoundingBox& );
 
  ON_BoundingBox m_bbox;     // world bounding box of visible geometry
  ON_Viewport    m_viewport; // view projection
  ON_3dPoint     m_target;   // target point
  ON_String      m_model_name;

private:
  // prohibit copy construction and operator=
  CModel(const CModel&);
  CModel& operator=(const CModel&);
};

CModel::CModel() : m_version(0), m_file_length(0)
{
  m_sStartSectionComments.Empty();
  m_properties.Default();
  m_settings.Default();
}

CModel::~CModel()
{
  Destroy();
}

void CModelObject::GetMaterial( const CModel& model, ON_Material& material ) const
{
  BOOL bEmissiveMaterial = FALSE;
  int li = -1;
  int mi = -1;
  if ( ON_Curve::Cast(m_geometry) ) {
    // use emmissive object color for curve objects
    ON_Color c(0,0,0);
    switch ( m_attributes.ColorSource() ) 
    {
    case ON::color_from_layer:
      li = m_attributes.m_layer_index;
      if ( li >= 0 && li < model.m_layer_table.Count() )
        c = model.m_layer_table[li].Color();
      bEmissiveMaterial = TRUE;
      break;
    case ON::color_from_object:
      c = m_attributes.m_color;
      bEmissiveMaterial = TRUE;
      break;
    case ON::color_from_material:
      break;
    default:
      break;
    }
    if ( bEmissiveMaterial ) {
      ON_Color black(0,0,0);
      material.Default();
      material.SetAmbient(black);
      material.SetDiffuse(black);
      material.SetSpecular(black);
      material.SetEmission(c);
    }
  }

  if ( !bEmissiveMaterial ) {
    // use rendering matrial for meshes, surfaces, and breps
    switch ( m_attributes.MaterialSource() )
    {
    case ON::material_from_layer:
      // use material assigned to object's layer
      li = m_attributes.m_layer_index;
      if ( li >= 0 && li < model.m_layer_table.Count() )
        mi = model.m_layer_table[li].RenderMaterialIndex();
      break;
    case ON::material_from_object:
      // use material assigned to object
      mi = m_attributes.m_material_index;
      break;
    default:
      mi = -1;
    }
    if ( mi < 0 || mi >= model.m_material_table.Count() )
      material.Default();
    else
      material = model.m_material_table[mi];
  }
}


void CModel::Destroy()
{
  int i;
  m_version = 0;
  m_file_length = 0;
  m_sStartSectionComments.Empty();
  m_properties.Default();
  m_settings.Default();

  for ( i = 0; i < m_bitmap_table.Count(); i++ )
    delete m_bitmap_table[i];
  m_bitmap_table.SetCapacity(0);

  m_material_table.SetCapacity(0);

  m_layer_table.SetCapacity(0);

  m_group_table.SetCapacity(0);

  m_light_table.SetCapacity(0);

  m_object_table.SetCapacity(0);

  m_bbox.m_max.x = -1.0;
  m_bbox.m_min.x = 1.0;  
}

BOOL CModel::GetBoundingBox( ON_BoundingBox& bbox ) const
{
  int i, object_count;
  bbox.Destroy();
  object_count = m_object_table.Count();
  for ( i = 0; i < object_count; i++ ) 
  {
    m_object_table[i].m_geometry->GetBoundingBox(bbox,bbox.IsValid());
  }
  return bbox.IsValid();
}

BOOL CModel::Read( const char* filename )
{
  ON_Workspace ws;
  ON_3dmObjectAttributes object_attributes;
  BOOL rc, return_code = FALSE;

  Destroy();
  FILE* fp = ws.OpenFile( filename, "rb "); // ~ON_Workspace will close file
  if ( !fp )
    return FALSE;
  m_model_name = filename;
  ON_BinaryFile file( ON::read3dm, fp );
  for(;;)
  {
    // STEP 1: REQUIRED - Read start section
    if ( !file.Read3dmStartSection( &m_version, m_sStartSectionComments ) )
    {
      printf("FAILURE in Read3dmStartSection().\n");
      break;
    }

    // STEP 2: REQUIRED - Read properties section
    if ( !file.Read3dmProperties( m_properties ) )
    {
      printf("FAILURE in Read3dmProperties().\n");
      break;
    }

    // STEP 3: REQUIRED - Read settings section
    if ( !file.Read3dmSettings( m_settings ) )
    {
      printf("FAILURE in Read3dmSettings().\n");
      break;
    }

    // STEP 4: REQUIRED - Read embedded bitmap table
    if ( file.BeginRead3dmBitmapTable() )
    {
      // At the moment no bitmaps are embedded so this table is empty
      ON_Bitmap* pBitmap = NULL;
      for(;;) {
        pBitmap = NULL;
        rc = file.Read3dmBitmap(&pBitmap);
        if ( rc==0 )
          break; // end of bitmap table
        if ( rc < 0 ) {
          printf("WARNING: Corrupt bitmap found. Skipping rest of bitmap table.\n");
          // NOTE  You can attempt to continue reading in this situation.  If the
          //       damage to the file was isolated to the part containing this object's
          //       definition, then the remaining objects can be read.  If the damage
          //       is of a more serious nature, then you could end up in and infinte
          //       loop so you probably want to put up some sort of continue/quit
          //       dialog every time you get a negative rc.
          break;
        }
        m_bitmap_table.Append(pBitmap);
      }

      // If BeginRead3dmBitmapTable() returns TRUE, 
      // then you MUST call EndRead3dmBitmapTable().
      if ( !file.EndRead3dmBitmapTable() )
      {
        printf("FAILURE in EndRead3dmBitmapTable().\n");
        break;
      }
    }
    else     
    {
      printf("WARNING: Missing or corrupt bitmap table - attempting to continue.\n");
      //break;
    }

    // STEP 5: REQUIRED - Read render material table
    if ( file.BeginRead3dmMaterialTable() )
    {
      ON_Material* pMaterial = NULL;
      for(;;) {
        rc = file.Read3dmMaterial(&pMaterial);
        if ( rc==0 )
          break; // end of material table
        if ( rc < 0 ) {
          printf("WARNING: Corrupt render material found. Skipping to material table.\n");
          // NOTE  You can attempt to continue reading in this situation.  If the
          //       damage to the file was isolated to the part containing this object's
          //       definition, then the remaining objects can be read.  If the damage
          //       is of a more serious nature, then you could end up in and infinte
          //       loop so you probably want to put up some sort of continue/quit
          //       dialog every time you get a negative rc.
          break;
        }
        m_material_table.Append(*pMaterial);
        delete pMaterial;
        pMaterial = NULL;
      }
      
      // If BeginRead3dmMaterialTable() returns TRUE, 
      // then you MUST call EndRead3dmMaterialTable().
      if ( !file.EndRead3dmMaterialTable() )
      {
        printf("FAILURE in EndRead3dmMaterialTable().\n");
        break;
      }
    }
    else     
    {
      printf("WARNING: Missing or corrupt render material table - attempting to continue.\n");
      //break;
    }

    // STEP 6: REQUIRED - Read layer table
    if ( file.BeginRead3dmLayerTable() )
    {
      ON_Layer* pLayer = NULL;
      for(;;) {
        pLayer = NULL;
        rc = file.Read3dmLayer(&pLayer);
        if ( rc==0 )
          break; // end of layer table
        if ( rc < 0 ) {
          printf("WARNING: Corrupt layer found. Skipping rest of layer table.\n");
          // NOTE  You can attempt to continue reading in this situation.  If the
          //       damage to the file was isolated to the part containing this layers's
          //       definition, then the remaining layers can be read.  If the damage
          //       is of a more serious nature, then you could end up in and infinte
          //       loop so you probably want to put up some sort of continue/quit
          //       dialog every time you get a negative rc.
          break;
        }
        m_layer_table.Append(*pLayer);
        delete pLayer;
        pLayer = NULL;
      }
      
      // If BeginRead3dmLayerTable() returns TRUE, 
      // then you MUST call EndRead3dmLayerTable().
      if ( !file.EndRead3dmLayerTable() )
      {
        printf("FAILURE in EndRead3dmLayerTable().\n");
        break;
      }
    }
    else     
    {
      printf("WARNING: Missing or corrupt layer table - attempting to continue.\n");
      //break;
    }

    // STEP 7: REQUIRED - Read group table
    if ( file.BeginRead3dmGroupTable() )
    {
      ON_Group* pGroup = NULL;
      for(;;) {
        pGroup = NULL;
        rc = file.Read3dmGroup(&pGroup);
        if ( rc==0 )
          break; // end of group table
        if ( rc < 0 ) {
          printf("WARNING: Corrupt group found. Skipping rest of group table.\n");
          // NOTE  You can attempt to continue reading in this situation.  If the
          //       damage to the file was isolated to the part containing this group's
          //       definition, then the remaining groups can be read.  If the damage
          //       is of a more serious nature, then you could end up in and infinte
          //       loop so you probably want to put up some sort of continue/quit
          //       dialog every time you get a negative rc.
          break;
        }
        m_group_table.Append(*pGroup);
        delete pGroup;
        pGroup = NULL;
      }
      
      // If BeginRead3dmGroupTable() returns TRUE, 
      // then you MUST call EndRead3dmGroupTable().
      if ( !file.EndRead3dmGroupTable() )
      {
        printf("FAILURE in EndRead3dmGroupTable().\n");
        break;
      }
    }
    else     
    {
      printf("WARNING: Missing or corrupt group table - attempting to continue.\n");
      //break;
    }

    // STEP 8: REQUIRED - Read render lights table
    if ( file.BeginRead3dmLightTable() )
    {
      ON_Light* pLight = NULL;
      for(;;) {
        object_attributes.Default();
        rc = file.Read3dmLight(&pLight,&object_attributes);
        if ( rc==0 )
          break; // end of light table
        if ( rc < 0 ) {
          printf("WARNING: Corrupt render light found. Skipping rest of light table.\n");
          // NOTE  You can attempt to continue reading in this situation.  If the
          //       damage to the file was isolated to the part containing this light's
          //       definition, then the remaining lights can be read.  If the damage
          //       is of a more serious nature, then you could end up in and infinte
          //       loop so you probably want to put up some sort of continue/quit
          //       dialog every time you get a negative rc.
          break;
        }
        if ( pLight->IsEnabled() ) {
          switch( pLight->Style() )
          {
          case ON::view_directional_light:
          case ON::view_point_light:
          case ON::view_spot_light:
          case ON::camera_directional_light:
          case ON::camera_point_light:
          case ON::camera_spot_light:
          case ON::world_directional_light:
          case ON::world_point_light:
          case ON::world_spot_light:
          case ON::ambient_light:
            {
              CModelRenderLight& light = m_light_table.AppendNew();
              light.m_light = *pLight;
              light.m_attributes = object_attributes;
            }
            break;
          default:
            // other types of lights are not supported in OpenGL
            break;
          }
        }
        delete pLight;
        pLight = NULL;
      }
      
      // If BeginRead3dmLightTable() returns TRUE, 
      // then you MUST call EndRead3dmLightTable().
      if ( !file.EndRead3dmLightTable() )
      {
        printf("FAILURE in EndRead3dmLightTable().\n");
        break;
      }
    }
    else     
    {
      printf("WARNING: Missing or corrupt light table - attempting to continue.\n");
      //break;
    }

    // STEP 9: REQUIRED - Read object (geometry and annotation) table
    if ( file.BeginRead3dmObjectTable() )
    {
      // Just get objects that OpenGL can render and skip things like
      // annotation.
      int object_filter = ON::point_object      |
                          ON::pointset_object   |
                          ON::curve_object      |
                          ON::surface_object    |
                          ON::brep_object       |
                          ON::mesh_object;

      //ON_Object* pObject = NULL;
      for(;;) {
        CModelObject object;
        ON_Object* pObject = NULL;
        rc = file.Read3dmObject(&pObject,&object.m_attributes,object_filter);
        if ( rc == 0 )
          break; // end of object table
        if ( rc < 0 ) {
          printf("WARNING: Corrupt object found. Skipping rest of object table.\n");
          // NOTE  You can attempt to continue reading in this situation.  If the
          //       damage to the file was isolated to the part containing this object's
          //       definition, then the remaining objects can be read.  If the damage
          //       is of a more serious nature, then you could end up in and infinte
          //       loop so you probably want to put up some sort of continue/quit
          //       dialog every time you get a negative rc.
          break;
        }
        if ( pObject ) 
        {
          object.m_geometry = ON_Geometry::Cast(pObject);
          if ( object.m_geometry ) {
            m_object_table.Append(object);
          }
          else {
            delete pObject;
          }
          pObject = NULL;
        }
        else
        {
          if ( rc == 2 ) {
            // object was filtered out because it's not a point, curve, surface, brep, or mesh
            //printf("WARNING: Skipping object because it's filtered.\n");
          }
          else if ( rc == 3 )
            printf("WARNING: Skipping object because it's newer than this code.  Update your OpenNURBS toolkit.\n");
          else
            printf("WARNING: Skipping object for unknown reason.\n");
        }
      }
      
      // If BeginRead3dmObjectTable() returns TRUE, 
      // then you MUST call EndRead3dmObjectTable().
      if ( !file.EndRead3dmObjectTable() )
      {
        printf("FAILURE in EndRead3dmObjectTable().\n");
        break;
      }
    }
    else     
    {
      printf("WARNING: Missing or corrupt object table.\n");
      break;
    }

    return_code = TRUE;
    break;
  }

  return return_code;
}

void CModel::DefaultView( const ON_BoundingBox& bbox )
{
  // simple parallel projection of bounding box;
  double window_height = 1.0;
  double window_width = 1.0;
  double dx, dy, dz;
  double frus_near, frus_far;
  ON_3dPoint camLoc;
  ON_3dVector camDir, camUp;
  m_target = 0.5*(bbox.m_min + bbox.m_max);
  dx = 1.1*(bbox.m_max[0] - bbox.m_min[0]);
  dy = 1.1*(bbox.m_max[1] - bbox.m_min[1]);
  dz = 1.1*(bbox.m_max[2] - bbox.m_min[2]);
  if ( dx <= 1.0e-6 && dy <= 1.0e-6 )
    dx = dy = 2.0;
  if ( window_height*dx < window_width*dy ) {
    dx = dy*window_width/window_height;
  }
  else {
    dy = dx*window_height/window_width;
  }
  if ( dz <= 0.1*(dx+dy) )
    dz = 0.1*(dx+dy);
  dx *= 0.5;
  dy *= 0.5;
  dz *= 0.5;


  frus_near = 1.0;
  frus_far = frus_near + 2.0*dz;
  camLoc = m_target + (dz + frus_near)*ON_zaxis;
  camDir = -ON_zaxis;
  camUp = ON_yaxis;

  m_viewport.SetProjection( ON::parallel_view );
  m_viewport.SetCameraLocation( camLoc );
  m_viewport.SetCameraDirection( camDir );
  m_viewport.SetCameraUp( camUp );
  m_viewport.SetFrustum( -dx, dx, -dy, dy, frus_near, frus_far );
}

///////////////////////////////////////////////////////////////////////
//
// Globals for myDisplay() function passed to auxMainLoop()
//
//////////////////////////////////////////////////////////////////////

static GLuint glb_display_list_number = 1;         // GL display list "name"
CModel* glb_model = 0; // global pointer to active model

///////////////////////////////////////////////////////////////////////
//
// Functions used in main()
//
//////////////////////////////////////////////////////////////////////

BOOL myInitGL( const ON_Viewport&, GLUnurbsObj*& );

void myBuildDisplayList( 
      GLuint,                  // display_list_number,
      GLUnurbsObj*,            // pointer to GL nurbs render
      const CModel&            // geometry to render
      );

extern "C" {
void MY_GL_CALLBACK myNurbsErrorCallback( GLenum ); // for gluNurbsCallback()

void MY_GL_CALLBACK myDisplay( void );              // for auxMainLoop()

void MY_GL_CALLBACK myReshape( GLsizei, GLsizei );  // for auxReshapeFunc()

void MY_GL_CALLBACK myMouseLeftEvent( AUX_EVENTREC* );   // for auxMouseFunc();
void MY_GL_CALLBACK myMouseMiddleEvent( AUX_EVENTREC* ); // for auxMouseFunc();
void MY_GL_CALLBACK myMouseRightEvent( AUX_EVENTREC* );  // for auxMouseFunc();

void MY_GL_CALLBACK myKeyLeftArrowEvent( void );    // for auxKeyFunc();
void MY_GL_CALLBACK myKeyRightArrowEvent( void );   // for auxKeyFunc();
void MY_GL_CALLBACK myKeyUpArrowEvent( void );      // for auxKeyFunc();
void MY_GL_CALLBACK myKeyDownArrowEvent( void );    // for auxKeyFunc();
void MY_GL_CALLBACK myKeyViewExtents( void );       // for auxKeyFunc();

typedef void (CALLBACK* RHINO_GL_NURBS_ERROR)();
}

///////////////////////////////////////////////////////////////////////
//
// used to set projections
//
void SetGLModelViewMatrix( const ON_Viewport& );
void SetGLProjectionMatrix( ON_Viewport& );
///////////////////////////////////////////////////////////////////////

int main( int argc, const char *argv[] )
{
  BOOL bOK;
  CModel model;
  int window_width  = 500;
  int window_height = 500;
  //double port_aspect = ((double)window_width)/((double)window_height);

  glb_model = &model;

  // read the file into model
  if ( argc != 2 ) {
    printf("Syntax: %s filename.3dm\n",argv[0] );
    return 0;
  }
  const char* sFileName = argv[1];
  printf("\nFile:  %s\n", sFileName );

  // read the file
  bOK = model.Read( argv[1] );
  if ( !bOK ) {
    // read failed
    return 1;
  }

  // set model.m_bbox = world bounding box of all the objects
  bOK = model.GetBoundingBox( model.m_bbox );
  if ( !bOK ) {
    // nothing to look at in this model
    return 2;
  }

  // set model.m_view
  if ( model.m_settings.m_views.Count() > 0 ) {
    // use first viewport projection in file
    double angle;
    model.m_viewport = model.m_settings.m_views[0].m_vp;
    model.m_target = model.m_settings.m_views[0].m_target;
    model.m_viewport.GetCameraAngle( &angle );
    model.m_viewport.Extents( angle, model.m_bbox );
  }
  else {
    model.DefaultView( model.m_bbox );
  }

  // If needed, enlarge frustum so its aspect matches the window's aspect.
  // Since the Rhino file does not store the far frustum distance in the
  // file, viewports read from a Rhil file need to have the frustum's far
  // value set by inspecting the bounding box of the geometry to be
  // displayed.

  
  ///////////////////////////////////////////////////////////////////
  //
  // GL stuff starts here
  //
  for(;;) {  
    
    char sWindowTitleString[256];
    sWindowTitleString[255] = 0;
    strncpy( sWindowTitleString, argv[0], 255 );

    auxInitPosition( 0, 0, window_width, window_height );
    auxInitDisplayMode( AUX_SINGLE | AUX_RGB | AUX_DEPTH );
    auxInitWindow( sWindowTitleString );

    // register event handler functions
    auxIdleFunc( 0 );
    auxReshapeFunc( myReshape );
    auxMouseFunc( AUX_LEFTBUTTON,   AUX_MOUSEDOWN, myMouseLeftEvent );
    auxMouseFunc( AUX_LEFTBUTTON,   AUX_MOUSEUP,   myMouseLeftEvent );
    auxMouseFunc( AUX_MIDDLEBUTTON, AUX_MOUSEDOWN, myMouseMiddleEvent );
    auxMouseFunc( AUX_MIDDLEBUTTON, AUX_MOUSEUP,   myMouseMiddleEvent );
    auxMouseFunc( AUX_RIGHTBUTTON,  AUX_MOUSEDOWN, myMouseRightEvent );
    auxMouseFunc( AUX_RIGHTBUTTON,  AUX_MOUSEUP,   myMouseRightEvent );
    auxKeyFunc( AUX_LEFT,  myKeyLeftArrowEvent );
    auxKeyFunc( AUX_RIGHT, myKeyRightArrowEvent );
    auxKeyFunc( AUX_UP,    myKeyUpArrowEvent );
    auxKeyFunc( AUX_DOWN,  myKeyDownArrowEvent );
    auxKeyFunc( AUX_E,  myKeyViewExtents );
    auxKeyFunc( AUX_e,  myKeyViewExtents );
    auxKeyFunc( AUX_Z,  myKeyViewExtents );
    auxKeyFunc( AUX_z,  myKeyViewExtents );
    // setup model view matrix, GL defaults, and the GL NURBS renderer
    GLUnurbsObj* pTheGLNURBSRender = NULL; // OpenGL NURBS rendering context
    bOK = myInitGL( model.m_viewport, pTheGLNURBSRender );

    if ( bOK ) {
      // build display list
      myBuildDisplayList( glb_display_list_number,
                          pTheGLNURBSRender,
                          model );

      // look at it
      auxMainLoop( myDisplay );
    }

    gluDeleteNurbsRenderer( pTheGLNURBSRender );

    break;
  }

  //
  // GL stuff ends here
  //
  ///////////////////////////////////////////////////////////////////

  return 0;
}

///////////////////////////////////////////////////////////////////////
void SetGLModelViewMatrix( const ON_Viewport& viewport )
{
  ON_GL( viewport ); // updates GL model view matrix
}

void SetGLProjectionMatrix( ON_Viewport& viewport )
{
  int pl, pr, pb, pt;
  viewport.GetScreenPort( &pl, &pr, &pb, &pt, NULL, NULL );
  ON_GL( viewport, pl, pr, pb, pt ); // updates GL projection matrix
}

BOOL myInitGL( const ON_Viewport& viewport, GLUnurbsObj*& nobj )
{
  // set the model view transform
  SetGLModelViewMatrix( viewport );

  // this stuff works with MSVC 4.2's Open GL. Changes may be needed for other
  // GLs.
  //ON_Color background_color(0,128,128);
  ON_Color background_color(0,63,127);
  //background_color = glb_model->m_settings.m_RenderSettings.m_background_color;
  glClearColor( (float)background_color.FractionRed(), 
                (float)background_color.FractionGreen(), 
                (float)background_color.FractionBlue(), 
                1.0f
                );

  glLightModeli( GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE );
  glDisable( GL_CULL_FACE );
  
  // Rhino viewports have camera "Z" pointing at the camera in a right
  // handed coordinate system.
  glClearDepth( 0.0f );
  glEnable( GL_DEPTH_TEST );
  glDepthFunc( GL_GEQUAL );

  glEnable( GL_LIGHTING );
  glEnable( GL_DITHER );
  //glEnable( GL_AUTO_NORMAL );
  //glEnable( GL_NORMALIZE );

  // default material
  ON_GL( (ON_Material*)NULL );


  // GL rendering of NURBS objects requires a GLUnurbsObj.
  nobj = gluNewNurbsRenderer();
  if ( !nobj )
    return FALSE;
  
  gluNurbsProperty( nobj, GLU_SAMPLING_TOLERANCE,   20.0f );
  gluNurbsProperty( nobj, GLU_PARAMETRIC_TOLERANCE, 0.5f );
  gluNurbsProperty( nobj, GLU_DISPLAY_MODE,         (GLfloat)GLU_FILL );
  //gluNurbsProperty( nobj, GLU_DISPLAY_MODE,         GLU_OUTLINE_POLYGON );
  //gluNurbsProperty( nobj, GLU_DISPLAY_MODE,         GLU_OUTLINE_PATCH );
  gluNurbsProperty( nobj, GLU_SAMPLING_METHOD,      (GLfloat)GLU_PATH_LENGTH );
  //gluNurbsProperty( nobj, GLU_SAMPLING_METHOD,      GLU_PARAMETRIC_ERROR );
  //gluNurbsProperty( nobj, GLU_SAMPLING_METHOD,      GLU_DOMAIN_DISTANCE );
  gluNurbsProperty( nobj, GLU_CULLING,              (GLfloat)GL_FALSE );

  // register GL NURBS error callback
  {
    // hack to get around C vs C++ type checking trauma
    RHINO_GL_NURBS_ERROR fn;
    fn = (RHINO_GL_NURBS_ERROR)myNurbsErrorCallback;
    gluNurbsCallback( nobj, GLU_ERROR, fn );
  }

  return TRUE;
}

///////////////////////////////////////////////////////////////////////

void MY_GL_CALLBACK myReshape( GLsizei w, GLsizei h )
{
  static GLsizei w0 = 0;
  static GLsizei h0 = 0;
  if ( w != w0 || h != h0 ) {
    h0 = h;
    w0 = w;
    ON_GL( glb_model->m_viewport, 0, w-1, h-1, 0 ); // set projection transform
  }
  glViewport( 0, 0, w, h );
}

///////////////////////////////////////////////////////////////////////
static void myRotateView( ON_Viewport& viewport,
                          const ON_3dVector& axis,
                          const ON_3dPoint& center,
                          double angle )
{
  ON_Xform rot;
  ON_3dPoint camLoc;
  ON_3dVector camY, camZ;

  rot.Rotation( angle, axis, center );

  if ( !viewport.GetCameraFrame( camLoc, NULL, camY, camZ ) )
    return;

  camLoc = rot*camLoc;
  camY   = rot*camY;
  camZ   = -(rot*camZ);

  viewport.SetCameraLocation( camLoc );
  viewport.SetCameraDirection( camZ );
  viewport.SetCameraUp( camY );

  ON_GL( viewport ); // update model view
}

static void myRotateLeftRight( ON_Viewport& viewport, double angle )
{
  // ON_3dVector axis = ON_zaxis; // rotate camera about world z axis (z up feel)
  ON_3dVector axis = ON_zaxis; // rotate camera about world y axis (u up feel)
  
  ON_3dPoint center;
  if ( glb_model )
    center = glb_model->m_target;
  else
    viewport.GetFrustumCenter( center );
  myRotateView( viewport, axis, center, angle );
}

static void myRotateUpDown( ON_Viewport& viewport, double angle )
{
  // rotates camera around the screen x axis
  ON_3dVector camX;
  ON_3dPoint center;
  if ( glb_model )
    center = glb_model->m_target;
  else
    viewport.GetFrustumCenter( center );
  viewport.GetCameraFrame( NULL, camX, NULL, NULL );
  myRotateView( viewport, camX, center, angle );
}

///////////////////////////////////////////////////////////////////////

void MY_GL_CALLBACK myKeyLeftArrowEvent( void )
{
  myRotateLeftRight( glb_model->m_viewport, ON_PI/12.0 );
}

void MY_GL_CALLBACK myKeyRightArrowEvent( void )
{
  myRotateLeftRight( glb_model->m_viewport, -ON_PI/12.0 );
}

void MY_GL_CALLBACK myKeyUpArrowEvent( void )
{
  myRotateUpDown( glb_model->m_viewport, ON_PI/12.0 );
}

void MY_GL_CALLBACK myKeyDownArrowEvent( void )
{
  myRotateUpDown( glb_model->m_viewport, -ON_PI/12.0 );
}

void MY_GL_CALLBACK myKeyViewExtents( void )
{
  double half_angle = 7.5*ON_PI/180.0;
  glb_model->m_viewport.Extents( half_angle, glb_model->m_bbox );
  SetGLModelViewMatrix( glb_model->m_viewport );
  SetGLProjectionMatrix( glb_model->m_viewport );
}

///////////////////////////////////////////////////////////////////////
//
// Mouse event handling
//

static void myMouseEvent( GLint button, const AUX_EVENTREC* event )
{
  static BOOL bMouseActive = FALSE;
  static int mx0, my0;
  int mx, my;

  if ( !event ) {
    bMouseActive = FALSE;
    return;
  }

  if ( event->event == AUX_MOUSEDOWN ) {
    if ( bMouseActive ) {
      bMouseActive = FALSE;
      return;
    }
    bMouseActive = TRUE;
    mx0 = event->data[AUX_MOUSEX];
    my0 = event->data[AUX_MOUSEY];
    return;
  }

  if ( !bMouseActive || event->event != AUX_MOUSEUP )
    return;

  mx = event->data[AUX_MOUSEX];
  my = event->data[AUX_MOUSEY];

  switch (button) {
  case AUX_LEFTBUTTON:
    // zoom
    glb_model->m_viewport.ZoomToScreenRect( mx0, my0, mx, my );
    break;
  case AUX_MIDDLEBUTTON:
    break;
  case AUX_RIGHTBUTTON:
    // dolly
    {
      ON_3dVector dolly_vector;
      double d;
      ON_3dPoint camLoc;
      ON_3dVector camZ;
      glb_model->m_viewport.GetCameraFrame( camLoc, NULL, NULL, camZ );
      d = (camLoc-glb_model->m_target)*camZ;
      if ( glb_model->m_viewport.GetDollyCameraVector(mx0,my0,mx,my,d,dolly_vector) ) {
        glb_model->m_viewport.DollyCamera( dolly_vector );
      }
    }
    break;
  }

  // update GL model view and projection matrices to match viewport changes
  SetGLModelViewMatrix( glb_model->m_viewport );
  SetGLProjectionMatrix( glb_model->m_viewport );

  bMouseActive = FALSE;
}

void MY_GL_CALLBACK myMouseLeftEvent( AUX_EVENTREC* event )
{
  myMouseEvent( AUX_LEFTBUTTON, event );
}

void MY_GL_CALLBACK myMouseMiddleEvent( AUX_EVENTREC* event )
{
  myMouseEvent( AUX_MIDDLEBUTTON, event );
}

void MY_GL_CALLBACK myMouseRightEvent( AUX_EVENTREC* event )
{
  myMouseEvent( AUX_RIGHTBUTTON, event );
}

///////////////////////////////////////////////////////////////////////

void myDisplayObject( const ON_Geometry& geometry, const ON_Material& material, GLUnurbsObj* nobj )
{
  // Called from myDisplay() to show geometry.
  // Uses ON_GL() functions found in rhinoio_gl.cpp.
  const ON_Point* point;
  const ON_PointCloud* cloud;
  const ON_Brep* brep;
  const ON_Mesh* mesh;
  const ON_Curve* curve;
  const ON_Surface* surface;

  // specify rendering material
  ON_GL( material );

  brep = ON_Brep::Cast(&geometry);
  if ( brep ) {
    ON_GL(*brep, nobj);
    return;
  }

  mesh = ON_Mesh::Cast(&geometry);
  if ( mesh ) {
    ON_GL(*mesh);
    return;
  }

  curve = ON_Curve::Cast(&geometry);
  if ( curve ) {
    ON_GL( *curve, nobj );
    return;
  }

  surface = ON_Surface::Cast(&geometry);
  if ( surface ) {
    gluBeginSurface( nobj );
    ON_GL( *surface, nobj );
    gluEndSurface( nobj );
    return;
  }

  point = ON_Point::Cast(&geometry);
  if ( point ) {
    ON_GL(*point);
    return;
  }

  cloud = ON_PointCloud::Cast(&geometry);
  if ( cloud ) {
    ON_GL(*cloud);
    return;
  }

}

///////////////////////////////////////////////////////////////////////

void MY_GL_CALLBACK myDisplayLighting( const ON_Viewport& viewport,
                                       const CModel& model
                                     )
{
  int light_count = model.m_light_table.Count();
  if ( light_count > 0 ) {
    int maxlighti = light_count;
    if ( maxlighti > GL_MAX_LIGHTS )
      maxlighti = GL_MAX_LIGHTS;
    int lighti;
    for ( lighti = 0; lighti < maxlighti; lighti++ ) {
      ON_GL( model.m_light_table[lighti].m_light, lighti+GL_LIGHT0 );
    }
  }
  else {
    // use default headlight
    // use basic bright white head light with a bit of ambient
    ON_Light head_light;
    head_light.Default();
    ON_GL( head_light, GL_LIGHT0 );
  }
}

///////////////////////////////////////////////////////////////////////

#if defined(_WINDOWS) || defined(_WIN32)
static void myDrawAxesSprite( const ON_Viewport& viewport, HDC hdc )
{
  // Use simple Windows calls to draw world axes sprite in lower left corner.
  // Note that Windows has screen (0,0) in the upper left corner; i.e,
  // screen "y" increases downwards.
  if ( !hdc )
    return;
  const int axes_size = 30;

  int port_left, port_right, port_top, port_bottom;
  if ( !viewport.GetScreenPort( &port_left, &port_right, &port_bottom, &port_top, NULL, NULL ) )
    return;
  const int scr_width  = port_right - port_left; // no "+1" here
  const int scr_height = port_bottom - port_top; // no "+1" here

  if (4*axes_size >= scr_width )
    return;
  if (4*axes_size >= scr_height )
    return;

  int x0 = 3*axes_size/2;
  int y0 = port_bottom - 3*axes_size/2;
  int indx[3] = {0,1,2};
  int scr_coord[3][2];
  viewport.GetCoordinateSprite( axes_size, x0, y0, indx, scr_coord );

#define LXSIZE 3
#define LYSIZE 3
#define LOFF 3

  // draw 3 axes from back to front
  COLORREF c[3] = { RGB(255,0,0), RGB(0,255,0), RGB(0,0,255) };
  HPEN axis_pen[3];
  axis_pen[0] = CreatePen( PS_SOLID, 2, RGB(255,0,0) );
  axis_pen[1] = CreatePen( PS_SOLID, 2, RGB(0,255,0) );
  axis_pen[2] = CreatePen( PS_SOLID, 2, RGB(0,0,255) );
  HGDIOBJ saved_pen = SelectObject( hdc, axis_pen[0] );

  int i, k, x, y, lx, ly;
  for (i=0;i<3;i++) {
    k = indx[i];
    x = scr_coord[k][0];
    y = scr_coord[k][1];
    // use direction of screen vector to determine letter placement
    lx = x-x0; ly = y-y0;
    if (abs(lx) > abs(ly)) {
      // center letter to right/left of axis end
      lx = (x >= x0) ? x + LXSIZE+LOFF : x - LXSIZE-LOFF;
      ly = y;
    }
    else if (abs(ly) > abs(lx)) {
      // center letter above/below axis end
      lx = x;
      ly = (y >= y0) ? y + LYSIZE+LOFF : y - LYSIZE-LOFF;
    }
    else if (lx) {
      // diagonal axis - center letter on axis
      lx = (x >= x0) ? x + LXSIZE+LOFF : x - LXSIZE-LOFF;
      ly = (y >= y0) ? y + LYSIZE+LOFF : y - LYSIZE-LOFF;
    }
    else {
      // axis is perp to screen - center letter at axis end
      lx = x;
      ly = y;
    }
    SelectObject( hdc, axis_pen[k] );

    // draw axis
    MoveToEx( hdc, x0, y0, NULL );
    LineTo( hdc, x, y );

    // draw axis label
    switch (k) {
    case 0: // X
      MoveToEx( hdc, lx-LXSIZE, ly-LYSIZE, NULL );
      LineTo(   hdc, lx+LXSIZE, ly+LYSIZE );
      MoveToEx( hdc, lx-LXSIZE, ly+LYSIZE, NULL );
      LineTo(   hdc, lx+LXSIZE, ly-LYSIZE );
      break;
    case 1: // Y
      MoveToEx( hdc, lx-LXSIZE, ly-LYSIZE, NULL );
      LineTo(   hdc, lx,    ly    );
      LineTo(   hdc, lx+LXSIZE, ly-LYSIZE );
      MoveToEx( hdc, lx,    ly, NULL    );
      LineTo(   hdc, lx,    ly+LYSIZE );
      break;
    case 2: // Z
      MoveToEx( hdc, lx-LXSIZE, ly-LYSIZE, NULL );
      LineTo(   hdc, lx+LXSIZE, ly-LYSIZE );
      LineTo(   hdc, lx-LXSIZE, ly+LYSIZE );
      LineTo(   hdc, lx+LXSIZE, ly+LYSIZE );
      break;
    }

  }
  SelectObject( hdc, saved_pen );
  DeleteObject( axis_pen[0] );
  DeleteObject( axis_pen[1] );
  DeleteObject( axis_pen[2] );

#undef LXSIZE
#undef LYSIZE
#undef LOFF

}
#endif

///////////////////////////////////////////////////////////////////////

void myBuildDisplayList( GLuint display_list_number,
                         GLUnurbsObj* pTheGLNurbsRender,
                         const CModel& model
                         )
{
  ON_Material material;
  glNewList( display_list_number, GL_COMPILE );

  // display Rhino geometry using ON_GL() functions found in rhinoio_gl.cpp
  int i;
  const int object_count = model.m_object_table.Count();
  for ( i = 0; i < object_count; i++ ) {
    const CModelObject& object = model.m_object_table[i];
    object.GetMaterial( model, material );
    myDisplayObject( *object.m_geometry, material, pTheGLNurbsRender );
  }

  glEndList();
}

///////////////////////////////////////////////////////////////////////

void MY_GL_CALLBACK myDisplay( void )
{
  // Uses globals glb_* because the GL aux tools don't provide an
  // easy way to pass information into this callback.
  int bUseRhinoSpotlights = FALSE; // I like to use a simple headlight
                                   // for a basic preview.

  glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

  // lights
  if ( bUseRhinoSpotlights && glb_model ) {
    // Rhino spotlights (currently rotate along with geometry)
    myDisplayLighting( glb_model->m_viewport, *glb_model );
  }
  else {
    // simple bright white headlight
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glLoadIdentity();
    GLfloat pos[4]  = { (GLfloat)0.0, (GLfloat)0.0, (GLfloat)1.0, (GLfloat)0.0 };
    glLightfv( GL_LIGHT0, GL_POSITION,  pos );
    GLfloat black[4] = { (GLfloat)0.0, (GLfloat)0.0, (GLfloat)0.0, (GLfloat)1.0 };
    GLfloat white[4] = { (GLfloat)1.0, (GLfloat)1.0, (GLfloat)1.0, (GLfloat)1.0 };
    glLightfv( GL_LIGHT0, GL_AMBIENT,  black );
    glLightfv( GL_LIGHT0, GL_DIFFUSE,  white );
    glLightfv( GL_LIGHT0, GL_SPECULAR, white );
    glEnable( GL_LIGHT0 );
    glPopMatrix();
  }

  // display list built with myBuildDisplayList()
  glCallList( glb_display_list_number ); 

  glFlush();

#if defined(_WINDOWS) || defined(_WIN32)
  // Windows decorations
  myDrawAxesSprite( glb_model->m_viewport, wglGetCurrentDC() );
#endif
}

///////////////////////////////////////////////////////////////////////

void MY_GL_CALLBACK myNurbsErrorCallback( GLenum errCode )
{
  const GLubyte* s = gluErrorString( errCode );
  printf("GL NURBS ERROR: (%d) %s\n",errCode, s );
}

///////////////////////////////////////////////////////////////////////
