/* $Header: /src2/opennurbs/example_read/example_read.cpp 17    6/14/01 10:54a Dalelear $ */
/* $NoKeywords: $ */
//
// Copyright (c) 1993-1998 Robert McNeel & Associates. All rights reserved.
// Rhinoceros is a registered trademark of Robert McNeel & Assoicates.
//
// THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
// ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
// MERCHANTABILITY ARE HEREBY DISCLAIMED.
//				
////////////////////////////////////////////////////////////////
//
//  example_read.cpp  
// 
//  Example program using the Rhino file IO toolkit.  The program reads in  
//  a Rhino 3dm model file and describes its contents.  The program is a 
//  console application that takes a filename as a command line argument.
//
////////////////////////////////////////////////////////////////////////

#include "../opennurbs.h"

////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////

class CModelRenderLight
{
public:
  CModelRenderLight();
  ~CModelRenderLight();
  CModelRenderLight(const CModelRenderLight&);
  CModelRenderLight& operator=(const CModelRenderLight&);

  ON_Light m_light;
  ON_3dmObjectAttributes m_attributes;
};

CModelRenderLight::CModelRenderLight()
{
}

CModelRenderLight::~CModelRenderLight()
{
}

CModelRenderLight::CModelRenderLight(const CModelRenderLight& src) 
             : m_light(src.m_light), 
               m_attributes(src.m_attributes)
{
}

CModelRenderLight& CModelRenderLight::operator=(const CModelRenderLight& src)
{
  if ( this != &src )
  {
    m_light = src.m_light;
    m_attributes = src.m_attributes;
  }
  return *this;
}

////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////

class CModelObject
{
public:
  CModelObject();
  ~CModelObject();
  CModelObject(const CModelObject&);
  CModelObject& operator=(const CModelObject&);

  void Dump( ON_TextLog& ) const;

  const ON_Object* m_object;
  ON_3dmObjectAttributes m_attributes;

private:
  void Destroy();
  unsigned int* m_ref_count; // reference counts used to avoid expensive object copying
};

CModelObject::CModelObject() : m_object(0), m_ref_count(0)
{
}

void CModelObject::Destroy()
{
  if ( m_ref_count ) {
    if ( *m_ref_count > 0 )
      (*m_ref_count)--;
    if ( *m_ref_count <= 0 ) 
    {
      delete m_ref_count;
      m_ref_count = 0;
    }
  }
  if ( !m_ref_count && m_object )
  {
    delete m_object;
    m_object = 0;
  }
}

CModelObject::~CModelObject()
{
  Destroy();
}

CModelObject::CModelObject(const CModelObject& src) 
             : m_object(0),
               m_ref_count(0)
{
  *this = src;
}

CModelObject& CModelObject::operator=(const CModelObject& src)
{
  if ( this != &src )
  {
    Destroy();
    m_object = src.m_object;
    m_attributes = src.m_attributes;
    m_ref_count = src.m_ref_count;
    if ( m_object ) {
      if ( m_ref_count )
        (*m_ref_count)++;
      else {
        m_ref_count = new unsigned int;
        *m_ref_count = 2; // 2 because this and src reference same m_object
        CModelObject* srcptr = (CModelObject*)&src; // need to defeat const
        srcptr->m_ref_count = m_ref_count;
      }
    }
  }
  return *this;
}

////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////

class CModelUserData
{
public:
  CModelUserData();
  ~CModelUserData();
  CModelUserData(const CModelUserData&);
  CModelUserData& operator=(const CModelUserData&);

  void Dump( ON_TextLog& ) const;

  ON_UUID  m_uuid;
  ON_3dmGoo m_goo;

private:
  void Destroy();
  unsigned int* m_ref_count; // reference counts used to avoid expensive object copying
};

CModelUserData::CModelUserData() : m_uuid(ON_nil_uuid)
{
}

CModelUserData::~CModelUserData()
{
}

CModelUserData::CModelUserData(const CModelUserData& src) 
             : m_uuid(src.m_uuid),
               m_goo(m_goo)
               
{
}

CModelUserData& CModelUserData::operator=(const CModelUserData& src)
{
  if ( this != &src )
  {
    m_goo = src.m_goo;
    m_uuid = src.m_uuid;
  }
  return *this;
}

////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////

class COpenNurbsModel
{
public:
  COpenNurbsModel();
  ~COpenNurbsModel();

  void Destroy(); // leaves model empty and ready to be reused

  // pedegodgical example of all the things in an OpenNURBS 3dm file.

  // start section
  int m_version;
  int m_file_length;
  ON_String m_sStartSectionComments;

  // Properties include revision history, notes, information about
  // the applicaton that created the file, and an option preview image.
  ON_3dmProperties m_properties;

  // Settings include tolerance, and unit system, and defaults used
  // for creating views and objects.
  ON_3dmSettings   m_settings;

  ON_SimpleArray<ON_Bitmap*>       m_bitmap_table;
  ON_ClassArray<ON_Material>       m_material_table;
  ON_ClassArray<ON_Layer>          m_layer_table;
  ON_ClassArray<CModelRenderLight> m_light_table;
  ON_ClassArray<ON_Group>          m_group_table;
  ON_ClassArray<CModelObject>      m_object_table;
  ON_ClassArray<CModelUserData>    m_userdata_table;

  void Dump( ON_TextLog& ) const; // text dump of entire model

  void DumpSummary( ON_TextLog& ) const; // text dump of model properties and settings
  void DumpMaterialTable( ON_TextLog& ) const; // text dump of material table
  void DumpLayerTable( ON_TextLog& ) const; // text dump of layer table
  void DumpLightTable( ON_TextLog& ) const; // text dump of light table
  void DumpGroupTable( ON_TextLog& ) const; // text dump of group table
  void DumpObjectTable( ON_TextLog& ) const; // text dump of object table
  void DumpUserDataTable( ON_TextLog& ) const; // text dump of user data table

private:
  // prohibit copy construction and operator=
  COpenNurbsModel(const COpenNurbsModel&);
  COpenNurbsModel& operator=(const COpenNurbsModel&);
};

BOOL Read3dmFile( 
       const char*,  // name of file to read
       COpenNurbsModel&   // destination for text dump
       );

////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////

int main( int argc, const char *argv[] )
{
  ON_Workspace ws;
  COpenNurbsModel model;

  // If you want to dump the output to a file, do something like this:
  ON_TextLog dump(ws.OpenFile("textdump.txt","w"));

  // If you want to dump the output to stdout, do something like this:
  //ON_TextLog dump; // dump to stdout

  int argi;
  if ( argc < 2 ) {
    printf("Syntax: %s file1.3dm file2.3dm ...\n",argv[0] );
    return 0;
  }

  for ( argi = 1; argi < argc; argi++ ) {
    Read3dmFile( argv[argi], model );
    dump.Print("FILE: %s\n",argv[argi]);
    dump.PushIndent();
    model.Dump(dump);
    model.Destroy();
    dump.PopIndent();
  }
  
  // Uncomment the call to getchar() if you want to be able to see
  // the contents of the command window when running the
  // MS dev studio debugger.
  printf("\nFINISHED - output in textdump.txt");
  //printf("\nFINISHED - output in \"textdump.txt\".  Press any key to quit\n");
  //getchar();
 
  return 0;
}

////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////

BOOL Read3dmFile( const char* sFileName, COpenNurbsModel& model )
{
  model.Destroy(); // get rid of any residual stuff

  BOOL return_code = FALSE; // return code set to TRUE when read is completed
  ON_3dmObjectAttributes object_attributes;
  int rc; 

  printf("\nFile:  %s\n", sFileName );

  // If you are using OpenNURBS as a Windows DLL, then you MUST use
  // ON::OpenFile() to open the file.  If you are not using OpenNURBS
  // as a Windows DLL, then you may use either ON::OpenFile() or fopen()
  // to open the file.
  FILE* fp = ON::OpenFile( sFileName, "rb ");
  if ( !fp ) {
    printf("  Unable to open.\n" );
    return FALSE;
  }

  ON_BinaryFile file( ON::read3dm, fp );

  for(;;) 
  {
    // STEP 1: REQUIRED - Read start section
    if ( !file.Read3dmStartSection( &model.m_version, model.m_sStartSectionComments ) )
    {
      printf("FAILURE in Read3dmStartSection().\n");
      break;
    }

    // STEP 2: REQUIRED - Read properties section
    if ( !file.Read3dmProperties( model.m_properties ) )
    {
      printf("FAILURE in Read3dmProperties().\n");
      break;
    }

    // STEP 3: REQUIRED - Read properties section
    if ( !file.Read3dmSettings( model.m_settings ) )
    {
      printf("FAILURE in Read3dmSettings().\n");
      break;
    }

    // STEP 4: REQUIRED - Read embedded bitmap table
    if ( file.BeginRead3dmBitmapTable() )
    {
      // At the moment no bitmaps are embedded so this table is empty
      ON_Bitmap* pBitmap = NULL;
      for(;;) {
        pBitmap = NULL;
        rc = file.Read3dmBitmap(&pBitmap);
        if ( rc==0 )
          break; // end of bitmap table
        if ( rc < 0 ) {
          printf("WARNING: Corrupt bitmap found. Skipping rest of bitmap table.\n");
          // NOTE  You can attempt to continue reading in this situation.  If the
          //       damage to the file was isolated to the part containing this object's
          //       definition, then the remaining objects can be read.  If the damage
          //       is of a more serious nature, then you could end up in and infinte
          //       loop so you probably want to put up some sort of continue/quit
          //       dialog every time you get a negative rc.
          break;
        }
        model.m_bitmap_table.Append(pBitmap);
      }

      // If BeginRead3dmBitmapTable() returns TRUE, 
      // then you MUST call EndRead3dmBitmapTable().
      if ( !file.EndRead3dmBitmapTable() )
      {
        printf("FAILURE in EndRead3dmBitmapTable().\n");
        break;
      }
    }
    else     
    {
      printf("WARNING: Missing or corrupt bitmap table - attempting to continue.\n");
      //break;
    }

    // STEP 5: REQUIRED - Read render material table
    if ( file.BeginRead3dmMaterialTable() )
    {
      ON_Material* pMaterial = NULL;
      for(;;) {
        rc = file.Read3dmMaterial(&pMaterial);
        if ( rc==0 )
          break; // end of material table
        if ( rc < 0 ) {
          printf("WARNING: Corrupt render material found. Skipping to material table.\n");
          // NOTE  You can attempt to continue reading in this situation.  If the
          //       damage to the file was isolated to the part containing this object's
          //       definition, then the remaining objects can be read.  If the damage
          //       is of a more serious nature, then you could end up in and infinte
          //       loop so you probably want to put up some sort of continue/quit
          //       dialog every time you get a negative rc.
          break;
        }
        model.m_material_table.Append(*pMaterial);
        delete pMaterial;
        pMaterial = NULL;
      }
      
      // If BeginRead3dmMaterialTable() returns TRUE, 
      // then you MUST call EndRead3dmMaterialTable().
      if ( !file.EndRead3dmMaterialTable() )
      {
        printf("FAILURE in EndRead3dmMaterialTable().\n");
        break;
      }
    }
    else     
    {
      printf("WARNING: Missing or corrupt render material table - attempting to continue.\n");
      //break;
    }

    // STEP 6: REQUIRED - Read layer table
    if ( file.BeginRead3dmLayerTable() )
    {
      ON_Layer* pLayer = NULL;
      for(;;) {
        pLayer = NULL;
        rc = file.Read3dmLayer(&pLayer);
        if ( rc==0 )
          break; // end of layer table
        if ( rc < 0 ) {
          printf("WARNING: Corrupt layer found. Skipping rest of layer table.\n");
          // NOTE  You can attempt to continue reading in this situation.  If the
          //       damage to the file was isolated to the part containing this object's
          //       definition, then the remaining objects can be read.  If the damage
          //       is of a more serious nature, then you could end up in and infinte
          //       loop so you probably want to put up some sort of continue/quit
          //       dialog every time you get a negative rc.
          break;
        }
        model.m_layer_table.Append(*pLayer);
        delete pLayer;
        pLayer = NULL;
      }
      
      // If BeginRead3dmLayerTable() returns TRUE, 
      // then you MUST call EndRead3dmLayerTable().
      if ( !file.EndRead3dmLayerTable() )
      {
        printf("FAILURE in EndRead3dmLayerTable().\n");
        break;
      }
    }
    else     
    {
      printf("WARNING: Missing or corrupt layer table - attempting to continue.\n");
      //break;
    }

    // STEP 7: REQUIRED - Read group table
    if ( file.BeginRead3dmGroupTable() )
    {
      ON_Group* pGroup = NULL;
      for(;;) {
        rc = file.Read3dmGroup(&pGroup);
        if ( rc==0 )
          break; // end of group table
        if ( rc < 0 ) {
          printf("WARNING: Corrupt group found. Skipping rest of group table.\n");
          // NOTE  You can attempt to continue reading in this situation.  If the
          //       damage to the file was isolated to the part containing this group's
          //       definition, then the remaining groups can be read.  If the damage
          //       is of a more serious nature, then you could end up in and infinte
          //       loop so you probably want to put up some sort of continue/quit
          //       dialog every time you get a negative rc.
          break;
        }
        model.m_group_table.Append(*pGroup);
        delete pGroup;
        pGroup = NULL;
      }
      
      // If BeginRead3dmGroupTable() returns TRUE, 
      // then you MUST call EndRead3dmGroupTable().
      if ( !file.EndRead3dmGroupTable() )
      {
        printf("FAILURE in EndRead3dmGroupTable().\n");
        break;
      }
    }
    else     
    {
      printf("WARNING: Missing or corrupt group table - attempting to continue.\n");
      //break;
    }

    // STEP 8: REQUIRED - Read render lights table
    if ( file.BeginRead3dmLightTable() )
    {
      ON_Light* pLight = NULL;
      for(;;) {
        object_attributes.Default();
        rc = file.Read3dmLight(&pLight,&object_attributes);
        if ( rc==0 )
          break; // end of light table
        if ( rc < 0 ) {
          printf("WARNING: Corrupt render light found. Skipping rest of light table.\n");
          // NOTE  You can attempt to continue reading in this situation.  If the
          //       damage to the file was isolated to the part containing this object's
          //       definition, then the remaining objects can be read.  If the damage
          //       is of a more serious nature, then you could end up in and infinte
          //       loop so you probably want to put up some sort of continue/quit
          //       dialog every time you get a negative rc.
          break;
        }
        CModelRenderLight& light = model.m_light_table.AppendNew();
        light.m_light = *pLight;
        light.m_attributes = object_attributes;
        delete pLight;
        pLight = NULL;
      }
      
      // If BeginRead3dmLightTable() returns TRUE, 
      // then you MUST call EndRead3dmLightTable().
      if ( !file.EndRead3dmLightTable() )
      {
        printf("FAILURE in EndRead3dmLightTable().\n");
        break;
      }
    }
    else     
    {
      printf("WARNING: Missing or corrupt light table - attempting to continue.\n");
      //break;
    }

    // STEP 9: REQUIRED - Read object (geometry and annotation) table
    if ( file.BeginRead3dmObjectTable() )
    {
      // optional filter made by setting ON::object_type bits 
      // For example, if you just wanted to just read points and meshes, you would use
      // object_filter = ON::point_object | ON::mesh_object;
      int object_filter = 0; 
      int object_count = 0;

      ON_Object* pObject = NULL;
      for(;;) {
        CModelObject object;
        ON_Object* pObject = NULL;
        rc = file.Read3dmObject(&pObject,&object.m_attributes,object_filter);
        if ( rc == 0 )
          break; // end of object table
        object_count++;
        if ( rc < 0 ) {
          printf("WARNING: Corrupt object found. Skipping rest of object table.\n");
          // NOTE  You can attempt to continue reading in this situation.  If the
          //       damage to the file was isolated to the part containing this object's
          //       definition, then the remaining objects can be read.  If the damage
          //       is of a more serious nature, then you could end up in and infinte
          //       loop so you probably want to put up some sort of continue/quit
          //       dialog every time you get a negative rc.
          break;
        }
        if ( pObject ) 
        {
          object.m_object = pObject;
          pObject = 0;
          model.m_object_table.Append(object);
        }
        else
        {
          if ( rc == 2 )
            printf("WARNING: Skipping object because it's filtered.\n");
          else if ( rc == 3 )
            printf("WARNING: Skipping object because it's newer than this code.  Update your OpenNURBS toolkit.\n");
          else
            printf("WARNING: Skipping object for unknown reason.\n");
        }
      }
      
      // If BeginRead3dmObjectTable() returns TRUE, 
      // then you MUST call EndRead3dmObjectTable().
      if ( !file.EndRead3dmObjectTable() )
      {
        printf("FAILURE in EndRead3dmObjectTable().\n");
        break;
      }
    }
    else     
    {
      printf("WARNING: Missing or corrupt object table.\n");
      break;
    }

    // read user tables as anonymous tables
    // If you develop a plug-ins or application that uses OpenNURBS files,
    // you can store anything you want in a user table.
    for(;;)
    {
      ON_UUID uuid;
      if ( !file.BeginRead3dmUserTable( uuid ) )
        break; // no more user tables in the file

      CModelUserData& ud = model.m_userdata_table.AppendNew();
      ud.m_uuid = uuid;
      if ( !file.Read3dmAnonymousUserTable( ud.m_goo ) )
      {
        printf("WARNING: Corrupt user table found. Skipping rest of use table.\n");
        // NOTE  You can attempt to continue reading in this situation.  If the
        //       damage to the file was isolated to the part containing this object's
        //       definition, then the remaining objects can be read.  If the damage
        //       is of a more serious nature, then you could end up in and infinte
        //       loop so you probably want to put up some sort of continue/quit
        //       dialog every time you get a negative rc.
        break;
      }

      // If BeginRead3dmObjectTable() returns TRUE, 
      // then you MUST call EndRead3dmUserTable().
      if ( !file.EndRead3dmUserTable() )
      {
        printf("FAILURE in EndRead3dmUserTable().\n");
        break;
      }
    }

    // OPTIONAL - check for end mark
    if ( !file.Read3dmEndMark(&model.m_file_length) )
    {
      if ( file.Archive3dmVersion() != 1 ) {
        // some v1 files are missing end-of-file markers
        printf("FAILURE in Read3dmEndMark( &model.m_file_length ).\n");
      }
    }

    return_code = TRUE;
    break;
  }


  return return_code;
}

////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////

COpenNurbsModel::COpenNurbsModel() : m_version(0), m_file_length(0)
{
  m_sStartSectionComments.Empty();
  m_properties.Default();
  m_settings.Default();
}

COpenNurbsModel::~COpenNurbsModel()
{
  Destroy();
}

void COpenNurbsModel::Destroy()
{
  int i;
  m_version = 0;
  m_file_length = 0;
  m_sStartSectionComments.Empty();
  m_properties.Default();
  m_settings.Default();

  for ( i = 0; i < m_bitmap_table.Count(); i++ )
    delete m_bitmap_table[i];
  m_bitmap_table.SetCapacity(0);

  m_material_table.SetCapacity(0);

  m_layer_table.SetCapacity(0);

  m_light_table.SetCapacity(0);

  m_object_table.SetCapacity(0);
  
  m_userdata_table.SetCapacity(0);
}


void DumpView( ON_TextLog& dump, const ON_3dmView& view )
{

  const wchar_t* sViewName = view.m_name;
  if ( !sViewName )
    sViewName = L"";
  ON::view_projection proj = view.m_vp.Projection();
  

  ON_3dPoint camLoc;
  ON_3dVector camX, camY, camZ;
  BOOL bValidCamera = view.m_vp.GetCameraFrame( camLoc, camX, camY, camZ );
  double frus_left,frus_right,frus_bottom,frus_top,frus_near,frus_far;
  BOOL bValidFrustum = view.m_vp.GetFrustum(&frus_left,&frus_right,&frus_bottom,&frus_top,&frus_near,&frus_far);
  int port_left, port_right, port_bottom, port_top, port_near, port_far;
  BOOL bValidPort = view.m_vp.GetScreenPort(&port_left,&port_right,&port_bottom,&port_top,&port_near,&port_far);

  dump.Print("Viewport: name = \"%S\" projection = %s\n", 
         sViewName,
         (proj == ON::perspective_view)
         ?"perspective":"parallel" 
         );

  dump.PushIndent();

  if ( bValidCamera ) {
    dump.Print("viewport camera frame\n"
           "  location = %g, %g, %g\n"
           "  X = %g, %g, %g\n"
           "  Y = %g, %g, %g\n"
           "  Z = %g, %g, %g\n",
           camLoc.x,camLoc.y,camLoc.z,
           camX.x,camX.y,camX.z,
           camY.x,camY.y,camY.z,
           camZ.x,camZ.y,camZ.z
           );
    ON_3dPoint target_point = view.m_target;
    double target_distance = target_point.DistanceTo( camLoc );
    dump.Print("camera target\n"
               "  distance = %g\n"
               "  point = %g,%g,%g\n",
               target_distance,
               target_point.x,target_point.y,target_point.z
               );
  }


  if ( bValidFrustum ) {
    dump.Print("view frustum\n"
           "  left   = %g, right = %g\n"
           "  bottom = %g, top   = %g\n"
           "  near   = %g, far   = %g\n",
           frus_left,frus_right,
           frus_bottom,frus_top,
           frus_near,frus_far
           );
  }

  if ( bValidPort ) {
    // location of viewport window on screen
    dump.Print("viewport window screen location\n"
           "  left   = %4d, right = %4d\n"
           "  bottom = %4d, top   = %4d\n"
           "  near   = %4d, far   = %4d\n",
           port_left,port_right,
           port_bottom,port_top,
           port_near,port_far
           );
  }


  // relative position of viewport window in application main frame
  double rel_left,rel_right,rel_bottom,rel_top;
  rel_left = view.m_position.m_wnd_left;
  rel_right = view.m_position.m_wnd_right;
  rel_bottom = view.m_position.m_wnd_bottom;
  rel_top = view.m_position.m_wnd_top;
  dump.Print("relative viewport window position in application frame window\n"
             "  left   = %6.2f%%, right = %6.2f%%\n"
             "  bottom = %6.2f%%, top   = %6.2f%%\n",
             100.0*rel_left, 100.0*rel_right,
             100.0*rel_bottom, 100.0*rel_top
             );

  dump.PopIndent();

}

void COpenNurbsModel::DumpSummary( ON_TextLog& dump ) const
{
  dump.Print("File version: %d\n",m_version);
  if ( m_file_length > 0 )
    dump.Print("File length: %d bytes\n",m_file_length);
  if ( m_sStartSectionComments.Length() > 0 ) {
    dump.Print("Start section comments:\n");
    dump.PushIndent();
    dump.PrintWrappedText(m_sStartSectionComments);
    dump.PopIndent();
    dump.Print("\n");
  }
  
  m_properties.Dump(dump);

  dump.Print("\n");

  m_settings.Dump(dump);

  dump.Print("\n");

  // dump viewports
  for ( int i = 0; i < m_settings.m_views.Count(); i++ ) {
    DumpView( dump, m_settings.m_views[i] );
    dump.Print("\n");
  }

  dump.Print("Contents:\n");
  dump.PushIndent();
  dump.Print("%d embedded bitmaps\n",m_bitmap_table.Count());
  dump.Print("%d render material definitions\n",m_material_table.Count());
  dump.Print("%d layers\n",m_layer_table.Count());
  dump.Print("%d render lights\n",m_light_table.Count());
  dump.Print("%d groups\n",m_group_table.Count());
  dump.Print("%d objects\n",m_object_table.Count());
  dump.Print("%d user data objects\n",m_userdata_table.Count());
  dump.PopIndent();
}

void COpenNurbsModel::DumpMaterialTable( ON_TextLog& dump) const
{
  int i;
  for ( i = 0; i < m_material_table.Count(); i++ )
  {
    dump.Print("Material %d:\n",i);
    dump.PushIndent();
    m_material_table[i].Dump(dump);
    dump.PopIndent();
  }
}

void COpenNurbsModel::DumpLayerTable( ON_TextLog& dump) const
{
  int i;
  for ( i = 0; i < m_layer_table.Count(); i++ )
  {
    dump.Print("Layer %d:\n",i);
    dump.PushIndent();
    m_layer_table[i].Dump(dump);
    dump.PopIndent();
  }
}

void COpenNurbsModel::DumpLightTable( ON_TextLog& dump) const
{
  int i;
  for ( i = 0; i < m_light_table.Count(); i++ )
  {
    dump.Print("Light %d:\n",i);
    dump.PushIndent();
    m_light_table[i].m_attributes.Dump(dump);
    m_light_table[i].m_light.Dump(dump);
    dump.PopIndent();
  }
}

void COpenNurbsModel::DumpGroupTable( ON_TextLog& dump) const
{
  int i;
  for ( i = 0; i < m_group_table.Count(); i++ )
  {
    dump.Print("Group %d:\n",i);
    dump.PushIndent();
    m_group_table[i].Dump(dump);
    dump.PopIndent();
  }
}

void CModelObject::Dump( ON_TextLog& dump ) const
{
  m_object->Dump(dump);

  // Use Cast() if you need to get at the details.
  if ( ON_Curve::Cast(m_object) ) {
    // curve objects
    if ( ON_NurbsCurve::Cast(m_object) ) {
      const ON_NurbsCurve* pCurve = ON_NurbsCurve::Cast(m_object);
    }
    else if ( ON_ArcCurve::Cast(m_object) ) {
      const ON_ArcCurve* pCurve = ON_ArcCurve::Cast(m_object);
    }
    else if ( ON_CircleCurve::Cast(m_object) ) {
      const ON_CircleCurve* pCurve = ON_CircleCurve::Cast(m_object);
    }
    else if ( ON_CurveOnSurface::Cast(m_object) ) {
      const ON_CurveOnSurface* pCurve = ON_CurveOnSurface::Cast(m_object);
    }
    else if ( ON_BrepEdge::Cast(m_object) ) {
      const ON_BrepEdge* pCurve = ON_BrepEdge::Cast(m_object);
    }
    else if ( ON_LineCurve::Cast(m_object) ) {
      const ON_LineCurve* pCurve = ON_LineCurve::Cast(m_object);
    }
    else if ( ON_PolyCurve::Cast(m_object) ) {
      const ON_PolyCurve* pCurve = ON_PolyCurve::Cast(m_object);
    }
    else if ( ON_PolylineCurve::Cast(m_object) ) {
      const ON_PolylineCurve* pCurve = ON_PolylineCurve::Cast(m_object);
    }
    else if ( ON_CurveProxy::Cast(m_object) ) {
      const ON_CurveProxy* pCurve = ON_CurveProxy::Cast(m_object);
    }
    else {
      const ON_Curve* pCurve = ON_Curve::Cast(m_object);
    }
  }
  else if ( ON_Surface::Cast(m_object) ) {
    // surface objects
    if ( ON_NurbsSurface::Cast(m_object) ) {
      const ON_NurbsSurface* pSurface = ON_NurbsSurface::Cast(m_object);
    }
    else if ( ON_PlaneSurface::Cast(m_object) ) {
      const ON_PlaneSurface* pSurface = ON_PlaneSurface::Cast(m_object);
    }
    else if ( ON_RevSurface::Cast(m_object) ) {
      const ON_RevSurface* pSurface = ON_RevSurface::Cast(m_object);
    }
    else if ( ON_BrepFace::Cast(m_object) ) {
      const ON_BrepFace* pSurface = ON_BrepFace::Cast(m_object);
    }
    else if ( ON_SurfaceProxy::Cast(m_object) ) {
      const ON_SurfaceProxy* pSurface = ON_SurfaceProxy::Cast(m_object);
    }
    else {
      const ON_Surface* pSurface = ON_Surface::Cast(m_object);
    }
  }
  else if ( ON_Mesh::Cast(m_object) ) {
    const ON_Mesh* pMesh = ON_Mesh::Cast(m_object);
  }
}


void COpenNurbsModel::DumpObjectTable( ON_TextLog& dump) const
{
  int i;
  for ( i = 0; i < m_object_table.Count(); i++ )
  {
    dump.Print("Object %d:\n",i);
    dump.PushIndent();
    m_object_table[i].m_attributes.Dump(dump);
    m_object_table[i].Dump(dump);
    dump.PopIndent();
  }
}

void COpenNurbsModel::DumpUserDataTable( ON_TextLog& dump) const
{
  int i;
  for ( i = 0; i < m_userdata_table.Count(); i++ )
  {
    const CModelUserData& ud = m_userdata_table[i];
    dump.Print("User Data Table %d:\n",i);
    dump.PushIndent();
    dump.Print("uuid = "); dump.Print(ud.m_uuid); dump.Print("\n");
    ud.m_goo.Dump(dump);    
    dump.PopIndent();
  }
}

void COpenNurbsModel::Dump( ON_TextLog& dump ) const
{
  dump.Print("Model summary:\n");
  dump.PushIndent();
  DumpSummary(dump);
  dump.PopIndent();
  dump.Print("\n");

  dump.Print("Material table:\n");
  dump.PushIndent();
  DumpMaterialTable(dump);
  dump.PopIndent();
  dump.Print("\n");

  dump.Print("Layer table:\n");
  dump.PushIndent();
  DumpLayerTable(dump);
  dump.PopIndent();
  dump.Print("\n");

  dump.Print("Light table:\n");
  dump.PushIndent();
  DumpLightTable(dump);
  dump.PopIndent();
  dump.Print("\n");

  dump.Print("Group table:\n");
  dump.PushIndent();
  DumpGroupTable(dump);
  dump.PopIndent();
  dump.Print("\n");

  dump.Print("Object table:\n");
  dump.PushIndent();
  DumpObjectTable(dump);
  dump.PopIndent();
}

#if 0

////////////////////////////////////////////////////////////////////////
//  These Dump() function print out information about the objects //////
////////////////////////////////////////////////////////////////////////

void Dump( const CRhinoObject& object, const char* sIndent /* = NULL */ )
{
  // If you need to upcast, then use object.TypeCode() to determine
  // the specific type of Rhino object.

  if ( !sIndent )
    sIndent = "";

  // get next level of indenting for layer and material dump.
  char* sNextIndent = NextIndent( sIndent );


  if ( object.TypeCode() != TCODE_VIEWPORT && object.TypeCode() != TCODE_COMMENTBLOCK) {
    // if it's not a viewport definition or comment block, then it's an
    // object and it has layer, name, rendering, ... properties

    const CRhinoLayer* pLayer = object.Layer();
    const char* sLayerName = (pLayer) ? pLayer->Name() : 0;
    if ( !sLayerName )
      sLayerName = "current layer";

    const char* sObjectState = "unknown";
    switch( object.State() ) {
    case CRhinoObject::normal:
      sObjectState = "normal";
      break;
    case CRhinoObject::hidden:
      sObjectState = "hidden";
      break;
    case CRhinoObject::locked:
      sObjectState = "locked";
      break;
    default:
      sObjectState = "unknown";
      break;
    }
    const char* sObjectName = object.Label();
    if ( !sObjectName )
      sObjectName = "";
    const CRhinoMaterial* pMat = object.Material(); // rendering information
    printf( "%s%s object properties:\n"
              "%slayer = \"%s\"\n"
              "%sname = \"%s\"\n"
              "%sstate = %s\n",
            sIndent,  object.ClassName(), 
            sNextIndent, sLayerName, 
            sNextIndent, sObjectName,
            sNextIndent, sObjectState
            );
    if ( pMat ) {
      // dump rendering material information
      Dump( *pMat, sNextIndent );
    }
  }

  // 

  switch( object.TypeCode() ) {
    
  case TCODE_RH_POINT:    // CRhinoPointSet ( 3d point list )
    Dump( (const CRhinoPointSet&)object, sNextIndent );
    break;
    
  case TCODE_MESH_OBJECT: // CRhinoMesh ( 3d mesh of triangles and quads )
    Dump( (const CRhinoMesh&)object, sNextIndent );
    break;
    
  case TCODE_RHINO_OBJECT_NURBS_CURVE: // CRhinoNurbsCurve ( NURBS curve )
  case TCODE_AG_SPLINE: // CRhinoAgSpline ( NURBS curve from Rhino 1.0 )
    Dump( (const CRhinoNurbsCurve&)object, sNextIndent );
    break;
    
  case TCODE_RHINO_OBJECT_NURBS_SURFACE: // CRhinoNurbsSurface ( NURBS surface )
  case TCODE_AG_SURFACE: // CRhinoAgSurface ( NURBS surface from Rhino 1.0 )
    Dump( (const CRhinoNurbsSurface&)object, sNextIndent );
    break;
    
  case TCODE_AG_CURVE: // CRhinoAgCurve ( curve made from contiguous NURBS curves )
    Dump( (const CRhinoAgCurve&)object, sNextIndent );
    break;
    
  case TCODE_RHINO_OBJECT_BREP: // CRhinoBrep ( B-rep model )
  case TCODE_AG_FACE:  // CRhinoAgFace ( B-rep with one face from Rhino 1.0 );
  case TCODE_AG_SHELL: // CRhinoAgFace ( B-rep with multiple faces from Rhino 1.0 );
    Dump( (const CRhinoBrep&)object, sNextIndent );
    break;

  case TCODE_RH_SPOTLIGHT: 
    Dump( (const CRhinoSpotlight&)object, sNextIndent );
    break;

  case TCODE_VIEWPORT: 
    Dump( (const CRhinoViewport&)object, sIndent );
    break;

  case TCODE_COMMENTBLOCK: 
    // Comment blocks are supplied for people who like to read binary
    // files with text editors.  They (the blocks and perhaps the people)
    // can be safely ignored.
    printf( "%sComment block: %d bytes of ASCII text\n",
      sNextIndent, 
      object.Length() );
    break;

  default:
    // Either object that is not yet exported to public or it is something 
    // that can be ignored.  See 3dm.h for a list of type codes.  See
    // example_structure.c for more details abount chunks.
    if( object.TypeCode() & 0x80000000 ) {
      printf( "%sInteger Chunk: typecode = %08x value = %d\n",
        sNextIndent, 
        object.TypeCode(),
        object.Value() 
        );
    }
    else {
      printf( "%sData Chunk: typecode = %08x value = %d bytes\n",
        sNextIndent, 
        object.TypeCode(),
        object.Length() 
        );
    }
  }

  printf("\n");

  // for debugging:
  //if ( !object.IsValid() ) {
  //  printf("%s*** INVALID OBJECT ***\n",sIndent);
  //  exit(1);
  //}
  onfree(sNextIndent);
}

////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////
static char* NextIndent( const char* sCurrentIndent )
{
  // get next level of indenting for layer and material dump.
  const int indent_length = sCurrentIndent ? strlen(sCurrentIndent) : 0;
  char* sNextIndent = (char*)onmalloc( (indent_length+3)*sizeof(sNextIndent[0]) );
  memset( sNextIndent, ' ', indent_length+2 );
  sNextIndent[indent_length+2] = 0;
  return sNextIndent;
}


void Dump( const CRhinoPointSet& point_set, const char* sIndent /* = NULL */ )
{
  if ( !sIndent )
    sIndent = "";

  const int point_count = point_set.PointCount();

	printf( "%sPoint Set with %d points.\n", 
          sIndent,
          point_count );

  int point_index;
  for ( point_index = 0; point_index < point_count; point_index++ ) {
    double x, y, z;
    point_set.GetPoint( point_index, x, y, z );
    printf( "%s  point %3d: ( %g, %g, %g )\n",
            sIndent, 
            point_index,
            x, y, z );
  }

}

void Dump( const CRhinoMesh& mesh, const char* sIndent /* = NULL */ )
{
  if ( !sIndent )
    sIndent = "";

  const int vertex_count = mesh.NumVertices();
  const int face_count   = mesh.NumFaces();
  int triangle_count     = 0;
  int quadrangle_count   = 0;

  int face_index;
  for ( face_index = 0; face_index < face_count; face_index++ ) {
    int v0, v1, v2, v3;
    mesh.GetFace( face_index, v0, v1, v2, v3 );
    if ( v2 == v3 )
      triangle_count++;
    else
      quadrangle_count++;
  }

  // Use the GetVertex(), GetNormal(), GetTCoord(), and GetFace() 
  // member functions to get the defining parameters.

  printf( "%sMesh with %d vertices, %d faces (%d triangles and %d quads).\n"
          "%s  normals=%s, texture coordinates=%s.\n",
          sIndent, 
          vertex_count,
          face_count,
          triangle_count,
          quadrangle_count,
          sIndent,
          ( mesh.HasNormals() ) ? "yes" : "no",
          ( mesh.HasTCoords() ) ? "yes" : "no"
          );
}

void Dump( const CRhinoMaterial& mat, const char* sIndent /* = NULL */ )
{
  if ( !sIndent )
    sIndent = "";

  CRhinoColor c = mat.Ambient();  
  printf( "%sMaterial:  ambient  =  (%3d , %3d, %3d )\n",
          sIndent,
          c.Red(), c.Green(), c.Blue() );
  c = mat.Diffuse();  
  printf( "%s           diffuse  =  (%3i , %3i, %3i )\n",
          sIndent,
          c.Red(), c.Green(), c.Blue() );
  c = mat.Emission();  
  printf( "%s           emmisive =  (%3i , %3i, %3i )\n",
          sIndent,
          c.Red(), c.Green(), c.Blue() );

  c = mat.Specular();  
  printf( "%s           specular =  (%3i , %3i, %3i )\n",
          sIndent,
          c.Red(), c.Green(), c.Blue() );
  printf( "%s           shine =  %f\n",
          sIndent,
          mat.Shine() );
  printf( "%s           Transparency = %f\n",
          sIndent,
					mat.Transparency() );
  if(mat.TextureBitmapFileName())
	  printf( "%s           Texture map Bitmap FileName: %s \n",
            sIndent,
            mat.TextureBitmapFileName() );
  if(mat.BumpBitmapFileName())
	  printf( "%s           Bump map Bitmap FileName:    %s \n",
            sIndent,
            mat.BumpBitmapFileName() );
  if(mat.RenderMaterialName())
	  printf( "%s           Render Material Name:    %s \n",
            sIndent,
            mat.RenderMaterialName() );
}

void Dump( const CRhinoLayer& layer, const char* sIndent /* = NULL */ )
{
  if ( !sIndent )
    sIndent = "";

	char* state = "";
	int r,g,b;
	switch(layer.State()){
	case CRhinoLayer::normal:
		state = "normal"; 
		break;
	case CRhinoLayer::locked:
		state = "locked"; 
		break;
	case CRhinoLayer::hidden:
		state = "hidden"; 
		break;
	case CRhinoLayer::ghosted:
		state = "ghosted";
	}
	r = layer.Color().Red( );
	g = layer.Color().Green( );
	b = layer.Color().Blue( );
  printf( "%sLayer: %s  state = %s RGB color = %d %d %d\n", 
          sIndent,
          layer.Name(), state, r,g,b);
}

void Dump( const CRhinoNurbsCurve& curve, const char* sIndent /* = NULL */ )
{
  // get next level of indenting
  if ( !sIndent )
    sIndent = "";

  // See the CRhinoNurbsCurve class for member functions that will get
  // control points, knot values, and other details.
  printf( "%sNURBS curve: %s degree = %d  %d control points\n"
          "%s             closed = %s  periodic = %s\n",
          sIndent, 
          curve.IsRational() ? "rational" : "non-rational",
          curve.Degree(),
          curve.CVCount(),
          sIndent,
          curve.IsClosed()   ? "TRUE" : "FALSE",
          curve.IsPeriodic() ? "TRUE" : "FALSE"
          );
}

void Dump( const CRhinoNurbsSurface& surface, const char* sIndent /* = NULL */ )
{
  // get next level of indenting
  if ( !sIndent )
    sIndent = "";

  // See the CRhinoNurbsSurface class for member functions that will get
  // control points, knot values, and other details.
  printf( "%sNURBS Surface: %s degree = %d X %d  control net = %d X %d points\n"
          "%s               closed = %s X %s  periodic = %s X %s\n",
          sIndent, 
          surface.IsRational() ? "rational" : "non-rational",
          surface.Degree(0),
          surface.Degree(1),
          surface.CVCount(0),
          surface.CVCount(1),
          sIndent,
          surface.IsClosed(0)   ? "TRUE" : "FALSE",
          surface.IsClosed(1)   ? "TRUE" : "FALSE",
          surface.IsPeriodic(0) ? "TRUE" : "FALSE",
          surface.IsPeriodic(1) ? "TRUE" : "FALSE"
          );
}

void Dump( const CRhinoAgCurve& crv, const char* sIndent /* = NULL */ )
{
  if ( !sIndent )
    sIndent = "";

  printf( "%s3d curve has %d NURBS segments\n",
          sIndent,
          crv.NurbsCurveCount() );

  // get next level of indenting for layer and material dump.
  char* sNextIndent = NextIndent( sIndent );

  int i;
  for ( i = 0; i < crv.NurbsCurveCount(); i++ )
    Dump( *(crv[i]), sNextIndent );

  onfree(sNextIndent);
}

static int brep_i_count = 0;

void Dump( const CRhinoBrep& brep, const char* sIndent /* = NULL */ )
{
  brep_i_count++;
  if ( !sIndent )
    sIndent = "";

  printf("%sBrep:\n",sIndent);

  // get next level of indenting for layer and material dump.
  char* sNextIndent = NextIndent( sIndent );

  if ( brep.IsSurface() ) {
    printf("%s(B-rep geometry is the same as underlying surface.)\n",sNextIndent);
  }

  printf("%ssurfaces:  %d\n",sNextIndent,brep.m_S.Size());
  printf("%s3d curve:  %d\n",sNextIndent,brep.m_C3.Size());
  printf("%s2d curves: %d\n",sNextIndent,brep.m_C2.Size());
  printf("%svertices:  %d\n",sNextIndent,brep.m_V.Size());
  printf("%sedges:     %d\n",sNextIndent,brep.m_E.Size());
  printf("%strims:     %d\n",sNextIndent,brep.m_T.Size());
  printf("%sloops:     %d\n",sNextIndent,brep.m_L.Size());
  printf("%sfaces:     %d\n",sNextIndent,brep.m_F.Size());

  int vi;
  for ( vi = 0; vi < brep.m_V.Size(); vi++ ) {
    const CRhinoBrepVertex& vertex = brep.m_V[vi];
    printf("%svertex[%2d]: (%f %f %f) tolerance(%g) edges(",
           sNextIndent,vi,vertex.x,vertex.y,vertex.z,vertex.m_tolerance);
    int vei;
    for ( vei = 0; vei < vertex.m_ei.Size(); vei++ ) {
      printf( (vei)?",%d":"%d", vertex.m_ei[vei] );
    }
    printf(")\n");
  }

  int ei;
  for ( ei = 0; ei < brep.m_E.Size(); ei++ ) {
    const CRhinoBrepEdge& edge = brep.m_E[ei];
    printf("%sedge[%2d]: v0(%2d) v1(%2d) 3d_curve(%d) tolerance(%g) trims(",
           sNextIndent,ei,edge.m_vi[0],edge.m_vi[1],edge.m_c3i,edge.m_tolerance);
    int eti;
    for ( eti = 0; eti < edge.m_ti.Size(); eti++ ) {
      printf( (eti)?",%d":"%d", edge.m_ti[eti]);
    }
    printf(")\n");
  }

  int fi;
  for ( fi = 0; fi < brep.m_F.Size(); fi++ ) {
    const CRhinoBrepFace& face = brep.m_F[fi];
    printf("%sface[%2d]: surface(%d) reverse(%d) loops(",
           sNextIndent,fi,face.m_si,face.m_bRev);
    int fli;
    for ( fli = 0; fli < face.m_li.Size(); fli++ ) {
      printf( (fli)?",%d":"%d", face.m_li[fli]);
    }
    printf(")\n");

    if ( brep.FaceIsSurface(fi) ) {
      printf("%s  (Face geometry is the same as underlying surface.)\n",sNextIndent);
    }

    for ( fli = 0; fli < face.m_li.Size(); fli++ ) {
      const int li = face.m_li[fli];
      const CRhinoBrepLoop& loop = brep.m_L[li];
      const char* sLoopType = "unknown";
      switch( loop.m_type ) {
      case CRhinoBrepLoop::unknown:
        sLoopType = "unknown";
        break;
      case CRhinoBrepLoop::outer:
        sLoopType = "outer";
        break;
      case CRhinoBrepLoop::inner:
        sLoopType = "inner";
        break;
      case CRhinoBrepLoop::slit:
        sLoopType = "slit";
        break;
      }
      printf("%s  loop[%2d]: type(%s) %d trims(",
             sNextIndent, li, sLoopType, loop.m_ti.Size());
      int lti;
      for ( lti = 0; lti < loop.m_ti.Size(); lti++ ) {
        printf( (lti)?",%d":"%d", loop.m_ti[lti]);
      }
      printf(")\n");
      for ( lti = 0; lti < loop.m_ti.Size(); lti++ ) {
        const int ti = loop.m_ti[lti];
        const CRhinoBrepTrim& trim = brep.m_T[ti];
        const char* sTrimType = "?";
        const char* sTrimIso = "-?";
        switch( trim.m_type ) {
        case CRhinoBrepTrim::unknown:
          sTrimType = "unknown ";
          break;
        case CRhinoBrepTrim::boundary:
          sTrimType = "boundary";
          break;
        case CRhinoBrepTrim::mated:
          sTrimType = "mated   ";
          break;
        case CRhinoBrepTrim::seam:
          sTrimType = "seam    ";
          break;
        case CRhinoBrepTrim::singular:
          sTrimType = "singular";
          break;
        }
        switch( trim.m_iso ) {
        case CRhinoBrepTrim::not_iso:
          sTrimIso = "";
          break;
        case CRhinoBrepTrim::u_iso:
          sTrimIso = "-u iso";
          break;
        case CRhinoBrepTrim::W_iso:
          sTrimIso = "-west iso";
          break;
        case CRhinoBrepTrim::E_iso:
          sTrimIso = "-east iso";
          break;
        case CRhinoBrepTrim::v_iso:
          sTrimIso = "-v iso";
          break;
        case CRhinoBrepTrim::S_iso:
          sTrimIso = "-south iso";
          break;
        case CRhinoBrepTrim::N_iso:
          sTrimIso = "-north iso";
          break;
        }
        printf("%s    trim[%2d]: edge(%2d) v0(%2d) v1(%2d) tolerance(%g,%g)\n"
               "%s              type(%s%s) rev3d(%d) 2d_curve(%d)\n",
               sNextIndent,ti,
               trim.m_ei,trim.m_vi[0],trim.m_vi[1],
               trim.m_tolerance[0],trim.m_tolerance[1],
               sNextIndent,
               sTrimType,sTrimIso,trim.m_bRev3d,trim.m_c2i);
      }
    }
  }

  int si;
  for ( si = 0; si < brep.m_S.Size(); si++ ) {
    Dump( *brep.m_S[si], sNextIndent );
  }

  onfree(sNextIndent);
}

void Dump( const CRhinoSpotlight& spotlight, const char* sIndent /* = NULL */ )
{
  // get next level of indenting
  if ( !sIndent )
    sIndent = "";

  // See the CRhinoNurbsSurface class for member functions that will get
  // control points, knot values, and other details.
  double position[3];
  spotlight.GetPosition( position );
  printf( "%sSpotlight: location     = (%g, %g, %g)\n",
          sIndent, position[0], position[1], position[2] );

  double direction[3];
  spotlight.GetDirection( direction );
  printf( "%s           direction    = (%g, %g, %g)\n",
          sIndent, direction[0], direction[1], direction[2] );

  double cutoff_angle = spotlight.SpotCutoff();
  printf( "%s           cutoff angle = %g degrees\n",sIndent, cutoff_angle);

  double exponent = spotlight.SpotExponent();
  printf( "%s           exponent     = %g\n",sIndent, exponent);

  CRhinoColor c = spotlight.Ambient();
  printf( "%s           ambient      = (%3i , %3i, %3i )\n",
          sIndent, c.Red(), c.Green(), c.Blue() );
  
  c = spotlight.Diffuse();
  printf( "%s           diffuse      = (%3i , %3i, %3i )\n",
          sIndent, c.Red(), c.Green(), c.Blue() );

  c = spotlight.Specular();
  printf( "%s           specular     = (%3i , %3i, %3i )\n",
          sIndent, c.Red(), c.Green(), c.Blue() );

}

void Dump( const CRhinoViewport& view, const char* sIndent /* = NULL */ )
{
  // get next level of indenting
  if ( !sIndent )
    sIndent = "";

  const char* sViewName = view.ViewName();
  if ( !sViewName )
    sViewName = "";
  CRhinoViewport::eProjection proj = view.Projection();
  

  double camLoc[3], camX[3], camY[3], camZ[3];
  BOOL bValidCamera = view.GetCameraFrame( camLoc, camX, camY, camZ );
  double frus_left,frus_right,frus_bottom,frus_top,frus_near,frus_far;
  BOOL bValidFrustum = view.GetFrustum(&frus_left,&frus_right,&frus_bottom,&frus_top,&frus_near,&frus_far);
  int port_left, port_right, port_bottom, port_top, port_near, port_far;
  BOOL bValidPort = view.GetScreenPort(&port_left,&port_right,&port_bottom,&port_top,&port_near,&port_far);
  double rel_left,rel_right,rel_bottom,rel_top;
  view.GetPortPosition(&rel_left,&rel_right,&rel_bottom,&rel_top);

  printf("%sViewport: name = \"%s\" projection = %s\n", 
         sIndent,
         sViewName,
         (proj == CRhinoViewport::perspective_proj)
         ?"perspective":"parallel" 
         );

  if ( bValidCamera ) {
    printf("%s          camera frame\n"
           "%s           location = %g, %g, %g\n"
           "%s           X = %g, %g, %g\n"
           "%s           Y = %g, %g, %g\n"
           "%s           Z = %g, %g, %g\n",
           sIndent,
           sIndent,camLoc[0],camLoc[1],camLoc[2],
           sIndent,camX[0],camX[1],camX[2],
           sIndent,camY[0],camY[1],camY[2],
           sIndent,camZ[0],camZ[1],camZ[2]
           );
  }

  if ( bValidFrustum ) {
    printf("%s          camera frustum\n"
           "%s           left   = %g, right = %g\n"
           "%s           bottom = %g, top   = %g\n"
           "%s           near   = %g, far   = %g\n",
           sIndent,
           sIndent,frus_left,frus_right,
           sIndent,frus_bottom,frus_top,
           sIndent,frus_near,frus_far
           );
  }

  if ( bValidPort ) {
    printf("%s          camera port\n"
           "%s           left   = %4d, right = %4d\n"
           "%s           bottom = %4d, top   = %4d\n"
           "%s           near   = %4d, far   = %4d\n",
           sIndent,
           sIndent,port_left,port_right,
           sIndent,port_bottom,port_top,
           sIndent,port_near,port_far
           );
  }

  double target_distance = 0.0;
  if ( view.GetTargetDistance( target_distance ) ) {
    double target_point[3];
    target_point[0] = camLoc[0] - target_distance*camZ[0];
    target_point[1] = camLoc[1] - target_distance*camZ[1];
    target_point[2] = camLoc[2] - target_distance*camZ[2];
    printf("%s          camera target\n"
           "%s           distance = %g\n"
           "%s           point = %g,%g,%g\n",
           sIndent,
           sIndent,target_distance,
           sIndent,target_point[0],target_point[1],target_point[2]
           );
  }

  printf("%s          relative position in Rhino layout\n"
         "%s           left   = %6.2f%%, right = %6.2f%%\n"
         "%s           bottom = %6.2f%%, top   = %6.2f%%\n",
         sIndent,
         sIndent,100.0*rel_left,100.0*rel_right,
         sIndent,100.0*rel_bottom,100.0*rel_top
         );

}

#endif
