#include "../opennurbs.h"

// This example demonstrates how to attach customized "user data"
// to any class derived from ON_Object.  In particular, you can
// attach custom information to any piece of geometry in a 3DM
// file and have it persist in files, transform, and copy.

class MyUserData : public ON_UserData
{
  ON_OBJECT_DECLARE(MyUserData);

public:
  // Note well:
  // 1) All constructors must initialize ON_UserData::m_userdata_uuid to
  //    the UUID that identifies this kind of user data.  
  // 2) All constructors must initialize ON_UserData::m_copy_count to
  //    1 if the user data should be copied when the parent object is
  //    copied.
  // For more details, see comments in the constructor code below.
  MyUserData();
  MyUserData(const MyUserData&);

  MyUserData& operator=(const MyUserData&);
  ~MyUserData();

  // In order to get your user data to persist in files, you must
  // override ON_Object::Write() and ON_Object::Read()
  BOOL Write(
         ON_BinaryArchive&  // serialize definition to binary archive
       ) const;

  BOOL Read(
         ON_BinaryArchive&  // restore definition from binary archive
       );
  
  // You must override ON_UserData::GetDescription().
  BOOL GetDescription( ON_wString& );

  // If your user data is attached to some type of ON_Geometry and you
  // want the user data to be transformed when the parent ON_Geometry
  // is transformed, then you must override ON_UserData::Transform().
  //
  // If you don't override ON_UserData::Transform(), then the net
  // result of any transforms is stored in ON_UserData::m_userdata_xform.
  // At appropriate times, you can inspect ON_UserData::m_userdata_xform
  // and reset it to the identity after you've taken whatever actions
  // you deem to be appropriate.
  BOOL Transform( const ON_Xform& );

  // possible information you might want to attach.
  int       m_my_int;
  ON_Line   m_my_line;
  ON_String m_my_string;
};

ON_OBJECT_IMPLEMENT( MyUserData, ON_UserData, "53114529-1CD7-4872-818E-311CB19101FA" );

MyUserData::MyUserData()
{
  // Each kind of user data needs a uuid and your constructors MUST
  // initialize ON_UserData::m_userdata_uuid with the value of this
  // uuid.  The value of ON_UserData::m_userdata_uuid uuid is used to
  // identify the user data and is passed to ON_Object::GetUserData()
  // when an application wants to access user data that has been
  // attached to an object.
  //
  // In simple cases, the class UUID can be used as is shown below.
  m_userdata_uuid = MyUserData::m_MyUserData_class_id.Uuid();

  // If you want your user data to be copied when its parent ON_Object
  // is copied, then your constructor must initialize 
  // ON_UserData::m_userdata_copycount to 1.
  // By default, ON_UserData::m_userdata_copycount is zero and the
  // user data is not copied.  See the comments in the ON_UserData
  // class definition for more details.
  m_userdata_copycount = 1;

  m_my_int = 0;
}

MyUserData::MyUserData(const MyUserData& src) 
          : ON_UserData(src) // critical - be sure to call base class copy constructor
{
  // Each kind of user data needs a uuid and your constructors MUST
  // initialize ON_UserData::m_userdata_uuid with the value of this
  // uuid.  The value of ON_UserData::m_userdata_uuid uuid is used to
  // identify the user data and is passed to ON_Object::GetUserData()
  // when an application wants to access user data that has been
  // attached to an object.
  //
  // In simple cases, the class UUID can be used as is shown below.
  m_userdata_uuid = MyUserData::m_MyUserData_class_id.Uuid();

  // If you want your user data to be copied when its parent ON_Object
  // is copied, then your constructor must initialize 
  // ON_UserData::m_userdata_copycount to 1.
  // By default, ON_UserData::m_userdata_copycount is zero and the
  // user data is not copied.  See the comments in the ON_UserData
  // class definition for more details.
  m_userdata_copycount = 1;

  m_my_int = src.m_my_int;
  m_my_line = src.m_my_line;
  m_my_string = src.m_my_string;
}

MyUserData& MyUserData::operator=(const MyUserData& src)
{
  if ( this != &src ) {
    // critical - be sure to call base class operator=()
    ON_UserData::operator=(src);
    
    m_my_int = src.m_my_int;
    m_my_line = src.m_my_line;
    m_my_string = src.m_my_string;
  }
  return *this;
}

MyUserData::~MyUserData()
{
}

BOOL MyUserData::Read( ON_BinaryArchive& file )
{
  BOOL rc = TRUE;
  if ( rc ) 
    rc = file.ReadInt(&m_my_int);
  if ( rc ) 
    rc = file.ReadLine(m_my_line);
  if ( rc ) 
    rc = file.ReadString(m_my_string);
  return rc;
}

BOOL MyUserData::Write( ON_BinaryArchive& file ) const
{
  BOOL rc = TRUE;
  if ( rc ) 
    rc = file.WriteInt(m_my_int);
  if ( rc ) 
    rc = file.WriteLine(m_my_line);
  if ( rc ) 
    rc = file.WriteString(m_my_string);
  return rc;
}

BOOL MyUserData::GetDescription( ON_wString& s )
{
  s = L"my user data with point, line, and string";
  return TRUE;
}

BOOL MyUserData::Transform( const ON_Xform& xform )
{
  // Optional: call the ON_UserData::Transform() if you want the
  // ON_UserData::m_userdata_xform value to be updated.
  ON_UserData::Transform(xform);


  // Transform any geometry you have in your class.
  BOOL rc = m_my_line.Transform(xform);
  return rc;
}

static void write_file( const char* filename, const ON_Point& point )
{
  // see example_write.cpp for information about writing 3dm files
  FILE* fp = fopen( filename, "wb" );
  if ( !fp )
    return;
  ON_BinaryFile file( ON::write3dm, fp );

  file.Write3dmStartSection( 2, "example_userdata.cpp file" );
  
  ON_3dmProperties properties;
  file.Write3dmProperties( properties );

  ON_3dmSettings settings;
  file.Write3dmSettings( settings );

  file.BeginWrite3dmBitmapTable();
  file.EndWrite3dmBitmapTable();

  file.BeginWrite3dmMaterialTable();
  file.EndWrite3dmMaterialTable();

  file.BeginWrite3dmLayerTable();
  ON_Layer layer;
  layer.SetLayerIndex(0);
  layer.SetLayerName("Default");
  file.Write3dmLayer( layer );
  file.EndWrite3dmLayerTable();

  file.BeginWrite3dmGroupTable();
  file.EndWrite3dmGroupTable();

  file.BeginWrite3dmLightTable();
  file.EndWrite3dmLightTable();

  file.BeginWrite3dmObjectTable();
  ON_3dmObjectAttributes point_attributes;
  point_attributes.m_layer_index = 0;
  file.Write3dmObject( point, &point_attributes );
  file.EndWrite3dmObjectTable();

  file.Write3dmEndMark();

  fclose(fp);
}

static void read_file( const char* filename, ON_Point& point )
{
  // see example_read.cpp for information about read 3dm files
  // This code will only read the file created by write_file().
  // This code should not be used as a model for reading general 3dm files.
  FILE* fp = fopen( filename, "rb" );
  if ( !fp )
    return;
  ON_BinaryFile file( ON::read3dm, fp );

  int version;
  ON_String start_section;
  file.Read3dmStartSection( &version, start_section );

  ON_3dmProperties properties;
  file.Read3dmProperties( properties );

  ON_3dmSettings settings;
  file.Read3dmSettings( settings );

  file.BeginRead3dmBitmapTable();
  file.EndRead3dmBitmapTable();

  file.BeginRead3dmMaterialTable();
  file.EndRead3dmMaterialTable();

  file.BeginRead3dmLayerTable();
  file.EndRead3dmLayerTable();

  file.BeginRead3dmGroupTable();
  file.EndRead3dmGroupTable();

  file.BeginRead3dmLightTable();
  file.EndRead3dmLightTable();

  file.BeginRead3dmObjectTable();
  {
    ON_3dmObjectAttributes point_attributes;
    ON_Object* p = NULL;
    file.Read3dmObject( &p, &point_attributes );
    ON_Point* pPoint = ON_Point::Cast( p );
    if ( pPoint )
      point = *pPoint;

    // The next file.Read3dmObject() call is here to read the end-of-table
    // marker to prevent getting a warning about unread data when 
    // file.EndRead3dmObjectTable() is called
    p = 0;
    file.Read3dmObject( &p, NULL );
  }
  file.EndRead3dmObjectTable();


  fclose(fp);
}


int main()
{
  // uuid used to get user data via ON_Object::GetUserData()
  const ON_UUID my_user_data_uuid = MyUserData::m_MyUserData_class_id.Uuid();

  // We'll attach a MyUserData user data to a point.  In general,
  // you can attach user data to any class derived from ON_Object.
  ON_Point point(0.0,0.0,0.0);

  // User data must be created by a call to new
  MyUserData* ud = new MyUserData();
  ud->m_my_int = 1;
  ud->m_my_line.from.Set(0.0,0.0,0.0);
  ud->m_my_line.to.Set(1.0,1.0,1.0);
  ud->m_my_string = "my user data";

  // This attaches the user data to point.  When the point is destroied,
  // the user data will be destroyed.  
  //
  point.AttachUserData(ud);

  // Use ON_Object::GetUserData() to get user data.
  MyUserData* original_ud = MyUserData::Cast( point.GetUserData( my_user_data_uuid ) );

  // When the point is copied, the user data will be copied if 
  // ud->m_userdata_copycount > 0.
  //
  ON_Point copy_of_point = point;

  // Use ON_Object::GetUserData() to get user data.
  MyUserData* copy_of_ud = MyUserData::Cast( copy_of_point.GetUserData( my_user_data_uuid ) );

  // When the point is transformed, the virtual ON_UserData::Transform()
  // is called to transform the point.


  // When the point is saved to a file, the virtual ON_Object::Write() is
  // called to write the attached user data.
  write_file( "point_with_user_Data.3dm", point );

  // When the point is read from a file, the virtual ON_Object::Read() is
  // called to read the user data.
  ON_Point point_from_file;
  read_file( "point_with_user_Data.3dm", point_from_file );

  // Use ON_Object::GetUserData() to get user data.
  MyUserData* ud_from_file = MyUserData::Cast( point_from_file.GetUserData( my_user_data_uuid ) );

  // Use delete to remove user data from the object it's attached to.
  delete ud_from_file;

  return 0;
}
