/* $Header: /src2/opennurbs/opennurbs_3dm_attributes.cpp 19    1/03/01 4:59p Dalelear $ */
/* $NoKeywords: $ */
/*
//
// Copyright (c) 1993-2001 Robert McNeel & Associates. All rights reserved.
// Rhinoceros is a registered trademark of Robert McNeel & Assoicates.
//
// THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
// ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
// MERCHANTABILITY ARE HEREBY DISCLAIMED.
//				
// For complete openNURBS copyright information see <http://www.opennurbs.org>.
//
////////////////////////////////////////////////////////////////
*/

#include "opennurbs.h"

ON_3dmObjectAttributes::ON_3dmObjectAttributes()
{
  Default();
}

ON_3dmObjectAttributes::~ON_3dmObjectAttributes()
{}

ON_3dmObjectAttributes::ON_3dmObjectAttributes(const ON_3dmObjectAttributes& src)
{
  Default();
  *this = src;
}

ON_3dmObjectAttributes& ON_3dmObjectAttributes::operator=(const ON_3dmObjectAttributes& src )
{
  if ( this != &src ) {
    m_uuid = src.m_uuid;
    m_name = src.m_name;
    m_url  = src.m_url;
    m_layer_index = src.m_layer_index;
    m_material_index = src.m_material_index;
    m_color = src.m_color;
    m_line_style = src.m_line_style;
    m_wire_density = src.m_wire_density;
    m_mode = src.m_mode;
    m_color_source = src.m_color_source;
    m_line_style_source = src.m_line_style_source;
    m_material_source = src.m_material_source;
    m_group = src.m_group;
  }
  return *this;
}

void ON_3dmObjectAttributes::Default()
{
  m_uuid = ON_nil_uuid;
  m_name.Destroy();
  m_url.Destroy();
  m_layer_index=0;
  m_material_index=-1;
  m_color = ON_Color(0,0,0);
  m_line_style.Default();
  m_wire_density = 1;
  m_mode = ON::normal_object;
  m_color_source = ON::color_from_layer;
  m_line_style_source = ON::line_style_from_layer;
  m_material_source = ON::material_from_layer;
  m_group.Destroy();
}

BOOL ON_3dmObjectAttributes::Read( ON_BinaryArchive& file )
{
  Default();
  int major_version = 0;
  int minor_version = 0;
  BOOL rc = file.Read3dmChunkVersion(&major_version,&minor_version);
  if ( rc && major_version == 1 ) {
    if (rc) rc = file.ReadUuid(m_uuid);
    if (rc) rc = file.ReadInt(&m_layer_index);
    if (rc) rc = file.ReadInt(&m_material_index);
    if (rc) rc = file.ReadColor(m_color);
    if (rc) rc = file.ReadLineStyle(m_line_style);
    if (rc) rc = file.ReadInt(&m_wire_density);
    if (rc) rc = file.ReadChar(&m_mode);
    if (rc) rc = file.ReadChar(&m_color_source);
    if (rc) rc = file.ReadChar(&m_line_style_source);
    if (rc) rc = file.ReadChar(&m_material_source);
    if (rc) rc = file.ReadString(m_name);
    if (rc) rc = file.ReadString(m_url);
    if ( minor_version >= 1 ) {
      file.ReadArray( m_group );
    }
  }
  else {
    rc = FALSE;
  }
  return rc;
}


BOOL ON_3dmObjectAttributes::Write( ON_BinaryArchive& file ) const
{
  BOOL rc = file.Write3dmChunkVersion(1,1);
  // version 1.0 fields
  if (rc) rc = file.WriteUuid(m_uuid);
  if (rc) rc = file.WriteInt(m_layer_index);
  if (rc) rc = file.WriteInt(m_material_index);
  if (rc) rc = file.WriteColor(m_color);
  if (rc) rc = file.WriteLineStyle(m_line_style);
  if (rc) rc = file.WriteInt(m_wire_density);
  if (rc) rc = file.WriteChar(m_mode);
  if (rc) rc = file.WriteChar(m_color_source);
  if (rc) rc = file.WriteChar(m_line_style_source);
  if (rc) rc = file.WriteChar(m_material_source);
  if (rc) rc = file.WriteString(m_name);
  if (rc) rc = file.WriteString(m_url);

  // version 1.1 fields
  if (rc) rc = file.WriteArray(m_group);
  return rc;
}

void ON_3dmObjectAttributes::Dump( ON_TextLog& dump ) const
{
  const wchar_t* s = m_name;
  if ( !s ) s = L"";
  const char* sMode = "unknown";
  switch( Mode() )
  {
  case ON::normal_object:
    sMode = "normal";
    break;
  case ON::hidden_object:
    sMode = "hidden";
    break;
  case ON::locked_object:
    sMode = "locked";
    break;
  default:
    sMode = "unknown";
    break;
  }
  dump.Print("object name = \"%S\"\n",s);
  dump.Print("object uuid = ");
  dump.Print(m_uuid);
  dump.Print("\n");
  dump.Print("object mode = %s\n",sMode);
  dump.Print("object layer index = %d\n",m_layer_index);
  dump.Print("object material index = %d\n",m_material_index);
  const char* sMaterialSource = "unknown";
  switch(MaterialSource()) {
  case ON::material_from_layer: sMaterialSource = "layer material"; break;
  case ON::material_from_object: sMaterialSource = "object material"; break;
  }
  dump.Print("material source = %s\n",sMaterialSource);
  const int group_count = GroupCount();
  if ( group_count > 0 ) {
    const int* group = GroupList();
    dump.Print("groups: ");
    int i;
    for ( i = 0; i < group_count; i++ ) {
      if ( i )
        dump.Print(",%d",group[i]);
      else
        dump.Print("%d",group[i]);
    }
    dump.Print("\n");
  }
}

ON::object_mode ON_3dmObjectAttributes::Mode() const
{
  return ON::ObjectMode( m_mode%16 );
}

void ON_3dmObjectAttributes::SetMode( ON::object_mode m )
{
  int om = ON::ObjectMode(m);
  int dm = DisplayMode();
  m_mode = (unsigned char)(16*dm+om);
}

ON::display_mode ON_3dmObjectAttributes::DisplayMode() const
{
  return ON::DisplayMode( m_mode/16 );
}

void ON_3dmObjectAttributes::SetDisplayMode( ON::display_mode m )
{
  int om = Mode();
  int dm = ON::DisplayMode(m);
  m_mode = (unsigned char)(16*dm+om);
}

ON::object_color_source ON_3dmObjectAttributes::ColorSource() const
{
  return ON::ObjectColorSource(m_color_source);
}

void ON_3dmObjectAttributes::SetColorSource( ON::object_color_source c )
{
  m_color_source = (unsigned char)ON::ObjectColorSource(c);
}

ON::object_material_source ON_3dmObjectAttributes::MaterialSource() const
{
  return ON::ObjectMaterialSource(m_material_source);
}

void ON_3dmObjectAttributes::SetMaterialSource( ON::object_material_source c )
{
  m_material_source = (unsigned char)ON::ObjectMaterialSource(c);
}

ON::object_line_style_source ON_3dmObjectAttributes::LineStyleSource() const
{
  return ON::ObjectLineStyleSource(m_line_style_source);
}

void ON_3dmObjectAttributes::SetLineStyleSource( ON::object_line_style_source c )
{
  m_line_style_source = (unsigned char)ON::ObjectLineStyleSource(c);
}

int ON_3dmObjectAttributes::GroupCount() const
{
  return m_group.Count();
}

//////////
// returns an int array of GroupCount() zero based group indices
const int* ON_3dmObjectAttributes::GroupList() const
{
  return (m_group.Count()>0) ? m_group.Array() : 0;
}

int ON_3dmObjectAttributes::GetGroupList(ON_SimpleArray<int>& group_list) const
{
  group_list = m_group;
  return group_list.Count();
}

//////////
// returns TRUE if object is in group with specified index
BOOL ON_3dmObjectAttributes::IsInGroup(
  int group_index // zero based group index
  ) const
{
  BOOL rc = FALSE;
  const int count = m_group.Count();
  int i;
  for ( i = 0; i < count; i++ ) {
    if (m_group[i] == group_index) {
      rc = TRUE;
      break;
    }
  }
  return rc;
}

BOOL ON_3dmObjectAttributes::IsInGroups( int group_count, const int* group_list ) const
{
  // returns TRUE if object is in any of the groups in the list
  BOOL rc = FALSE;
  if ( group_count > 0 && group_list ) {
    const int obj_group_count  = GroupCount();
    const int* obj_group_list = GroupList();
    // in practice these arrays will be very short and this search will be fast
    int i, j;
    for ( i = 0; i < obj_group_count; i++ ) for ( j = 0; j < group_count; j++ ) {
      if ( obj_group_list[i] == group_list[j] )
        return TRUE;
    }
  }
  return rc;
}

BOOL ON_3dmObjectAttributes::IsInGroups( const ON_SimpleArray<int>& group_list ) const
{
  return IsInGroups( group_list.Count(), group_list.Array() );
}


//////////
// Adds object to group with specified index (If object is already in
// group, nothing is changed.)
void ON_3dmObjectAttributes::AddToGroup(
  int group_index // zero based group index
  )
{
  if ( group_index >= 0 ) {
    if ( !IsInGroup(group_index) )
      m_group.Append(group_index);
  }
}

//////////
// returns the index of the last group in the group list
// or -1 if the object is not in any groups
int ON_3dmObjectAttributes::TopGroup() const
{
  const int* top_group = m_group.Last();
  return top_group ? *top_group : -1;
}

//////////
// removes the object from the last group in the group list
void ON_3dmObjectAttributes::RemoveFromTopGroup()
{
  int c = m_group.Count();
  if ( c > 0 ) {
    c--;
    m_group.SetCount(c);
  }
}

//////////
// Removes object from group with specified index.  If object is not
// in the group, nothing is changed.
void ON_3dmObjectAttributes::RemoveFromGroup(
  int group_index // zero based group index
  )
{
  int i;
  const int count = m_group.Count();
  for ( i = 0; i < count; i++ ) {
    if (m_group[i] == group_index) {
      m_group.Remove(i);
      break;
    }
  }
}

//////////
// Removes object from all groups.
void ON_3dmObjectAttributes::RemoveFromAllGroups()
{
  m_group.Destroy();
}
