/* $Header: /src2/opennurbs/opennurbs_arc.cpp 18    7/23/01 3:11p Dalelear $ */
/* $NoKeywords: $ */
/*
//
// Copyright (c) 1993-2001 Robert McNeel & Associates. All rights reserved.
// Rhinoceros is a registered trademark of Robert McNeel & Assoicates.
//
// THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
// ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
// MERCHANTABILITY ARE HEREBY DISCLAIMED.
//				
// For complete openNURBS copyright information see <http://www.opennurbs.org>.
//
////////////////////////////////////////////////////////////////
*/

#include "opennurbs.h"

ON_Arc::ON_Arc()
{}

ON_Arc::ON_Arc( const ON_Circle& c, double angle ) : ON_Circle(c), m_angle(0.0,angle)
{}

ON_Arc::ON_Arc( const ON_Plane& p, double r, double angle ) : ON_Circle(p,r), m_angle(0.0,angle)
{}

ON_Arc::ON_Arc( const ON_3dPoint& C, double r, double angle )  : ON_Circle(C,r), m_angle(0.0,angle)
{}

ON_Arc::ON_Arc( const ON_Plane& pln, const ON_3dPoint& C, double r, double angle ) : ON_Circle(pln,C,r), m_angle(0.0,angle)
{}

ON_Arc::ON_Arc( const ON_2dPoint& P, const ON_2dPoint& Q, const ON_2dPoint& R ) : ON_Circle(P,Q,R)
{
  double angle;
  ON_Circle::ClosestPointTo( ON_3dPoint(R), &angle );
  m_angle.Set(0.0,angle);
}

ON_Arc::ON_Arc( const ON_3dPoint& P, const ON_3dPoint& Q, const ON_3dPoint& R ) : ON_Circle(P,Q,R)
{
  double angle;
  ON_Circle::ClosestPointTo( R, &angle );
  m_angle.Set(0.0,angle);
}

ON_Arc& ON_Arc::operator=( const ON_Circle& c )
{
  ON_Circle::operator=(c);
  m_angle.Set(0.0,2.0*ON_PI);
  return *this;
}


BOOL ON_Arc::Create(
  const ON_Circle& circle,
  double angle_radians          // angle in radians
  )
{
  plane = circle.plane;
  plane.UpdateEquation();
  radius = circle.radius;
  m_angle.Set(0.0,angle_radians);
  UpdatePoints();
  return IsValid();
}

BOOL ON_Arc::Create(
  const ON_Plane& pl, // circle is in this plane with center at m_origin
  double r,         // radius
  double angle_radians   // angle in radians
  )
{
  plane = pl;
  plane.UpdateEquation();
  radius = r;
  m_angle.Set(0.0,angle_radians);
  UpdatePoints();
  return IsValid();
}

BOOL ON_Arc::Create( // arc is parallel to XY plane
  const ON_3dPoint& center, // center
  double r,            // radius
  double angle_radians // angle in radians
  )
{
  plane.CreateFromNormal( center, ON_zaxis );
  radius = r;
  m_angle.Set(0.0,angle_radians);
  UpdatePoints();
  return IsValid();
}

BOOL ON_Arc::Create(        // arc parallel to a plane
  const ON_Plane& pl,       // circle will be parallel to this plane
  const ON_3dPoint& center, // center
  double r,                 // radius
  double angle_radians      // angle in radians
  )
{
  plane = pl;
  plane.origin = center;
  plane.UpdateEquation();
  radius = r;
  m_angle.Set(0.0,angle_radians);
  UpdatePoints();
  return IsValid();
}

BOOL ON_Arc::Create( // arc through 3 2d points
  const ON_2dPoint& P, // point P
  const ON_2dPoint& Q, // point Q
  const ON_2dPoint& R // point R
  )
{
  ON_Circle::Create(P,Q,R);
  double a = 0.0;
  ON_Circle::ClosestPointTo( R, &a );
  m_angle.Set(0.0,a);
  return IsValid();
}

BOOL ON_Arc::Create( // arc through 3 3d points
  const ON_3dPoint& P, // point P
  const ON_3dPoint& Q, // point Q
  const ON_3dPoint& R  // point R
  )
{
  ON_Circle::Create(P,Q,R);
  double a = 0.0;
  ON_Circle::ClosestPointTo( R, &a );
  m_angle.Set(0.0,a);
  return IsValid();
}



ON_Arc::~ON_Arc()
{}

void ON_Arc::Dump( ON_TextLog& dump ) const
{
  dump.Print("Arc: normal = ");
  dump.Print(plane.zaxis);
  dump.Print(" center = ");
  dump.Print(plane.origin);
  dump.Print(" start = ");
  dump.Print( StartPoint() );
  dump.Print(" end = ");
  dump.Print( EndPoint() );
  dump.Print(" radius = ");
  dump.Print(Radius());
  dump.Print(" angle = [");
  dump.Print(m_angle[0]);
  dump.Print(",");
  dump.Print(m_angle[1]);
  dump.Print("]\n");
}

ON_3dPoint ON_Arc::StartPoint() const
{
  return PointAt(m_angle[0]);
}

ON_3dPoint ON_Arc::MidPoint() const
{
  return PointAt(m_angle.Mid());
}

ON_3dPoint ON_Arc::EndPoint() const
{
  return PointAt(m_angle[1]);
}

BOOL ON_Arc::IsValid() const
{
  return (ON_Circle::IsValid() && AngleRadians() > ON_ZERO_TOLERANCE && AngleRadians() <= 2.0*ON_PI+ON_ZERO_TOLERANCE) ? TRUE : FALSE;
}

ON_BoundingBox ON_Arc::BoundingBox() const
{
  // TODO - compute tight arc bounding box

  // Using these knot[] and cv[] arrays makes this function
  // not use any heap memory.
  double knot[10];
  ON_4dPoint cv[9];
  ON_NurbsCurve c;
  c.m_knot = knot;
  c.m_cv = &cv[0].x;
  if ( GetNurbForm(c) )
    return c.BoundingBox();
  return ON_Circle::BoundingBox();
}


BOOL ON_Arc::UpdatePoints()
{
  m_point[0] = PointAt(m_angle.ParameterAt(0.0));
  m_point[1] = PointAt(m_angle.ParameterAt(0.5));
  m_point[2] = PointAt(m_angle.ParameterAt(1.0));
  return TRUE;
}


BOOL ON_Arc::IsCircle() const
{
  return (fabs(fabs(AngleRadians()) - 2.0*ON_PI) <= ON_ZERO_TOLERANCE) ? TRUE : FALSE;
}

double ON_Arc::AngleRadians() const
{
  return m_angle[1]-m_angle[0];
}

double ON_Arc::AngleDegrees() const
{
  return (AngleRadians()/ON_PI)*180.0;
}

ON_Interval ON_Arc::DomainRadians() const
{
  return m_angle;
}

ON_Interval ON_Arc::DomainDegrees() const
{
  const double rtd = 180.0/ON_PI;
  ON_Interval ad = m_angle;
  ad.m_t[0] *= rtd;
  ad.m_t[1] *= rtd;
  return ad;
}
BOOL ON_Arc::SetAngleRadians( double a )
{
  if ( a < 0.0 ) {
    m_angle.Set(-a,0.0);
    Reverse();
  }
  else
    m_angle.Set(0.0,a);
  return ( fabs(m_angle.Length()) <= 2.0*ON_PI ) ? TRUE : FALSE;
}

BOOL ON_Arc::SetAngleDegrees( double a )
{
  return SetAngleRadians( (a/180.0)*ON_PI );
}

BOOL ON_Arc::Reverse()
{
  m_angle.Reverse();
  plane.yaxis = -plane.yaxis;
  plane.zaxis = -plane.zaxis;

  plane.UpdateEquation();
  return TRUE;
}

double ON_Arc::Length() const
{
  return fabs(AngleRadians()*radius);
}

/* moved to opennurbs_arccurve.cpp


int ON_Arc::GetNurbForm( ON_NurbsCurve& nurbscurve ) const
{
  int rc = 0;
  if ( IsValid() ) {
    if ( IsCircle() )
      rc = ON_Circle::GetNurbForm( nurbscurve );
    else {
      double a, b, c, t, dt, angle;
      int span_count, i;
      angle = m_angle.Length();
      if (angle <= 0.5*ON_PI + ON_ZERO_TOLERANCE) {
		    span_count = 1;
        dt = 0.5;
      }
      else if (angle <= ON_PI + ON_ZERO_TOLERANCE) {
		    span_count = 2;
  		  angle *= 0.5;
        dt = 0.25;
      }
      else if (angle <= 1.5*ON_PI + ON_ZERO_TOLERANCE) {
		    span_count = 3;
  		  angle /= 3.0;
        dt = 1.0/6.0;
      }
      else {
		    span_count = 4;
  		  angle *= 0.25;
        dt = 0.125;
      }
      nurbscurve.Create( 3, TRUE, 3, 2*span_count+1 );
      ON_4dPoint* CV = (ON_4dPoint*)nurbscurve.m_cv;
      t = m_angle[0];
      for ( i = 0; i < span_count; i++ ) {
        nurbscurve.m_knot[2*i] = t;
        nurbscurve.m_knot[2*i+1] = t;
        CV[2*i] = PointAt(m_angle.ParameterAt(t));
        t += dt;
        CV[2*i+1] = PointAt(m_angle.ParameterAt(t));
        t += dt;
      }

      span_count *= 2;
      t = m_angle[1];
      CV[span_count] = PointAt(t);
      nurbscurve.m_knot[span_count] = t;
      nurbscurve.m_knot[span_count+1] = t;

	    a = cos(0.5*angle);
	    b = a - 1.0;
	    c = radius*angle;

	    for (i = 1; i < span_count; i += 2) {
		    CV[i].x +=  b * plane.origin.x;
		    CV[i].y +=  b * plane.origin.y;
		    CV[i].z +=  b * plane.origin.z;
		    CV[i].w = a;
	    }
      
      //for ( i = 1; i < span_count; i += 2 ) {
      //  t = CV[i].w;
      //  c = 1.0/t;
      //  a = CV[i].x*c; b = ArcDeFuzz(a); if ( a != b ) CV[i].x = b*t;
      //  a = CV[i].y*c; b = ArcDeFuzz(a); if ( a != b ) CV[i].y = b*t;
      //  a = CV[i].z*c; b = ArcDeFuzz(a); if ( a != b ) CV[i].z = b*t;
      //}
    }
    rc = 2;
  }
	return rc;
}
*/

// returns parameters of point on arc that is closest to given point
BOOL ON_Arc::ClosestPointTo( 
       const ON_3dPoint& pt, 
       double* t
       ) const
{
  double tt, a;
  if ( !t )
    t =&tt;
  BOOL rc = ON_Circle::ClosestPointTo(pt,t);
  if (rc) {
    if ( *t < m_angle[0] ) {
      a = 0.5*(m_angle[0] + m_angle[1] - 2.0*ON_PI);
      if ( *t < a )
        *t = m_angle[1];
      else 
        *t = m_angle[0];
    }
    else if ( *t > m_angle[1] ) {
      a = 0.5*(m_angle[0] + m_angle[1] + 2.0*ON_PI);
      if ( *t > a )
        *t = m_angle[0];
      else 
        *t = m_angle[1];
    }
  }
  return rc;
}

// returns point on circle that is arc to given point
ON_3dPoint ON_Arc::ClosestPointTo( 
       const ON_3dPoint& pt
       ) const
{
  double t = m_angle[0];
  ClosestPointTo( pt, &t );
  return PointAt(t);
}

