/* $Header: /src2/opennurbs/opennurbs_archive.cpp 76    7/09/01 11:15a Dalelear $ */
/* $NoKeywords: $ */
/*
//
// Copyright (c) 1993-2001 Robert McNeel & Associates. All rights reserved.
// Rhinoceros is a registered trademark of Robert McNeel & Assoicates.
//
// THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
// ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
// MERCHANTABILITY ARE HEREBY DISCLAIMED.
//				
// For complete openNURBS copyright information see <http://www.opennurbs.org>.
//
////////////////////////////////////////////////////////////////
*/

#include "opennurbs.h"

ON_BinaryArchive::ON_BinaryArchive( ON::archive_mode mode ) 
                 : m_3dm_version(0), 
                   m_3dm_v1_layer_index(0), m_3dm_v1_material_index(0),
                   m_3dm_v1_suppress_error_message(0),
                   m_3dm_opennurbs_version(0),
                   m_active_table(ON_BinaryArchive::no_active_table),
                   m_bDoChunkCRC(0), m_bad_CRC_count(0),
                   m_endian(ON::Endian()), 
                   m_mode(mode)
{
  // Sparc, MIPS, ... CPUs have big endian byte order
  // ON_BinaryArchives use little endian byte order

  m_bSavePreviewImage    = FALSE; // TRUE to save 200x200 preview bitmap (increases file size)
  m_bEmbedTextureBitmaps = FALSE; // TRUE to embed texture, bump, trace, and wallpaper bitmaps (increases file size)
  m_bSaveRenderMeshes    = FALSE; // TRUE to save meshes used to render B-rep objects (increases file size)
  m_bSaveAnalysisMeshes  = FALSE; // TRUE to save meshes used in surface analysis (increases file size)

  m_zlib.mode = ON::unknown_archive_mode;
  m_zlib.sizeof_buffer = sizeof(m_zlib.buffer);
  memset( &m_zlib.strm, 0, sizeof(m_zlib.strm) );
}

ON_BinaryArchive::~ON_BinaryArchive()
{
  CompressionEnd();
}


BOOL ON_BinaryArchive::ToggleByteOrder(
  int count,          // number of elements
  int sizeof_element, // size of element (2,4, or 8)
  const void* src,    // source buffer
  void* dst           // destination buffer (can be same a source buffer)
  )
{
  union {
    unsigned short x;
    unsigned char c[2];
  } u2;
  union {
    unsigned long x;
    unsigned char c[4];
  } u4;
  union {
    double x;
    unsigned char c[8];
  } u8;

  BOOL rc = (count==0 || (count>0&&src&&dst));
  if ( rc ) {
    switch(sizeof_element) {
    case 2:
      {
        const unsigned short* a2 = (const unsigned short*)src;
        unsigned char* b2 = (unsigned char*)dst;
        while(count--) {
          u2.x = *a2++;
          *b2++ = u2.c[1];
          *b2++ = u2.c[0];
        }
      }
      break;
    case 4:
      {
        const unsigned int* a4 = (const unsigned int*)src;
        unsigned char* b4 = (unsigned char*)dst;
        while(count--) {
          u4.x = *a4++;
          *b4++ = u4.c[3];
          *b4++ = u4.c[2];
          *b4++ = u4.c[1];
          *b4++ = u4.c[0];
        }
      }
      break;
    case 8:
      {
        const double* a8 = (const double*)src;
        unsigned char* b8 = (unsigned char*)dst;
        while(count--) {
          u8.x = *a8++;
          *b8++ = u8.c[7];
          *b8++ = u8.c[6];
          *b8++ = u8.c[5];
          *b8++ = u8.c[4];
          *b8++ = u8.c[3];
          *b8++ = u8.c[2];
          *b8++ = u8.c[1];
          *b8++ = u8.c[0];
        }
      }
      break;
    default:
      rc = FALSE;
      break;
    }
  }
  return rc;
}


BOOL
ON_BinaryArchive::ReadMode() const
{
  return (m_mode & ON::read) ? TRUE : FALSE;
}

BOOL
ON_BinaryArchive::WriteMode() const
{
  return (m_mode & ON::write) ? TRUE : FALSE;
}

BOOL
ON_BinaryArchive::ReadChar(    // Read an array of 8 bit chars
		size_t count,       // number of chars to read
		char*  p  
		)
{
  return ReadByte( count, p );
}

BOOL
ON_BinaryArchive::ReadChar(    // Read an array of 8 bit unsigned chars
		size_t count,       // number of unsigned chars to read
		unsigned char* p   
		)
{
  return ReadByte( count, p );
}

BOOL
ON_BinaryArchive::ReadChar(    // Read a single 8 bit char
		char* p
		)
{
  return ReadByte( 1, p );
}

BOOL
ON_BinaryArchive::ReadChar(    // Read a single 8 bit unsigned char
		unsigned char* p
		)
{
  return ReadByte( 1, p );
}

BOOL
ON_BinaryArchive::ReadShort(   // Read an array of 16 bit shorts
		size_t count,       // number of unsigned chars to read
		short* p
		)
{
  BOOL rc = ReadByte( count<<1, p );
  if ( rc && m_endian == ON::big_endian ) {
    // reverse byte order
		unsigned char* b= (unsigned char*) (p);
		unsigned char  c;
		while(count--) {
			c = b[0]; b[0] = b[1]; b[1] = c;
			b += 2;
		}
  }
  return rc;
}

BOOL
ON_BinaryArchive::ReadShort(   // Read an array of 16 bit unsigned shorts
		size_t count,       // number of unsigned chars to read
		unsigned short* p
		)
{
  return ReadShort( count, (short*)p );
}

BOOL
ON_BinaryArchive::ReadShort(   // Read a single 16 bit short
		short* p
		)
{
  return ReadShort( 1, p );
}

BOOL
ON_BinaryArchive::ReadShort(   // Read a single 16 bit unsigned short
		unsigned short* p
		)
{
  return ReadShort( 1, p );
}

BOOL
ON_BinaryArchive::ReadInt( // Read an array of 32 bit integers
		size_t count,       // number of unsigned chars to read
		int* p
		)
{
  BOOL rc = ReadByte( count<<2, p );
  if ( rc && m_endian == ON::big_endian ) {
		unsigned char* b= (unsigned char*)p;
		unsigned char  c;
		while(count--) {
			c = b[0]; b[0] = b[3]; b[3] = c;
			c = b[1]; b[1] = b[2]; b[2] = c;
			b += 4;
		}
  }
  return rc;
}

BOOL
ON_BinaryArchive::ReadInt( // Read an array of 32 bit integers
		size_t count,       // number of unsigned chars to read
		unsigned int* p
		)
{
  return ReadInt( count, (int*)p );
}

BOOL
ON_BinaryArchive::ReadInt( // Read a single 32 bit integer
		int* p
		)
{
  return ReadInt( 1, p );
}

BOOL
ON_BinaryArchive::ReadInt( // Read a single 32 bit unsigned integer
		unsigned int* p
		)
{
  return ReadInt( 1, p );
}

BOOL
ON_BinaryArchive::ReadLong( // Read an array of 32 bit integers
		size_t count,       // number of unsigned chars to read
		long* p
		)
{
  return ReadInt( count, (int*)p );
}

BOOL
ON_BinaryArchive::ReadLong( // Read an array of 32 bit integers
		size_t count,       // number of unsigned chars to read
		unsigned long* p
		)
{
  return ReadInt( count, (int*)p );
}

BOOL
ON_BinaryArchive::ReadLong( // Read a single 32 bit integer
		long* p
		)
{
  return ReadInt( 1, (int*)p );
}

BOOL
ON_BinaryArchive::ReadLong( // Read a single 32 bit unsigned integer
		unsigned long* p
		)
{
  return ReadInt( 1, (int*)p );
}

BOOL
ON_BinaryArchive::ReadFloat(   // Read an array of floats
		size_t count,       // number of unsigned chars to read
		float* p
		)
{
  // floats and ints have same size and endian issues
  return ReadInt( count, (int*)p );
}

BOOL
ON_BinaryArchive::ReadFloat(   // Read a single float
		float* p
		)
{
  return ReadFloat( 1, p );
}

BOOL
ON_BinaryArchive::ReadDouble(  // Read an array of IEEE doubles
		size_t count,       // number of unsigned chars to read
		double* p
		)
{
  BOOL rc = ReadByte( count<<3, p );
  if ( rc && m_endian == ON::big_endian ) {
		unsigned char* b=(unsigned char*)p;
		unsigned char  c;
		while(count--) {
			c = b[0]; b[0] = b[7]; b[7] = c;
			c = b[1]; b[1] = b[6]; b[6] = c;
			c = b[2]; b[2] = b[5]; b[5] = c;
			c = b[3]; b[3] = b[4]; b[4] = c;
			b += 8;
		}
  }
  return rc;
}

BOOL
ON_BinaryArchive::ReadDouble(  // Read a single double
		double* p
		)
{
  return ReadDouble( 1, p );
}

BOOL
ON_BinaryArchive::ReadColor( ON_Color& color )
{
  unsigned int colorref = 0;
  BOOL rc = ReadInt( &colorref );
  color = colorref;
  return rc;
}

BOOL ON_BinaryArchive::ReadLineStyle( ON_LineStyle& ls )
{
  ls.Default();
  short line_pattern = 0;
  short line_pattern_index = 0;
  double thickness = 0.0;
  double scale = 0.0;
  BOOL rc = ReadShort( &line_pattern );
  if (rc) rc = ReadShort( &line_pattern_index );
  if (rc) rc = ReadDouble( &thickness );
  if (rc) rc = ReadDouble( &scale );
  if (rc) {
    if ( line_pattern == ON::custom_line_pattern )
      ls.SetPatternIndex( line_pattern_index );
    else
      ls.SetPattern( ON::LinePattern(line_pattern) );
    ls.SetThickness(thickness);
    ls.SetScale(scale);
  }
  return rc;
}


BOOL
ON_BinaryArchive::ReadPoint (
  ON_2dPoint& p
  )
{
  return ReadDouble( 2, &p.x );
}

BOOL
ON_BinaryArchive::ReadPoint (
  ON_3dPoint& p
  )
{
  return ReadDouble( 3, &p.x );
}

BOOL
ON_BinaryArchive::ReadPoint (
  ON_4dPoint& p
  )
{
  return ReadDouble( 4, &p.x );
}

BOOL
ON_BinaryArchive::ReadVector (
  ON_2dVector& v
  )
{
  return ReadDouble( 2, &v.x );
}

BOOL
ON_BinaryArchive::ReadVector (
  ON_3dVector& v
  )
{
  return ReadDouble( 3, &v.x );
}

BOOL ON_BinaryArchive::WriteBoundingBox(const ON_BoundingBox& bbox)
{
  BOOL rc = WritePoint( bbox.m_min );
  if (rc) rc = WritePoint( bbox.m_max );
  return rc;
}

BOOL ON_BinaryArchive::ReadBoundingBox(ON_BoundingBox& bbox)
{
  BOOL rc = ReadPoint( bbox.m_min );
  if (rc) rc = ReadPoint( bbox.m_max );
  return rc;
}

BOOL 
ON_BinaryArchive::WriteXform( const ON_Xform& x )
{
  return WriteDouble( 16, &x.m_xform[0][0] );
}

BOOL 
ON_BinaryArchive::ReadXform( ON_Xform& x )
{
  return ReadDouble( 16, &x.m_xform[0][0] );
}

BOOL 
ON_BinaryArchive::WritePlane( const ON_Plane& plane )
{
  BOOL rc = WritePoint( plane.origin );
  if (rc) rc = WriteVector( plane.xaxis );
  if (rc) rc = WriteVector( plane.yaxis );
  if (rc) rc = WriteVector( plane.zaxis );
  if (rc) rc = WriteDouble( 4, plane.equation );
  return rc;
}

BOOL 
ON_BinaryArchive::ReadPlane( ON_Plane& plane )
{
  BOOL rc = ReadPoint( plane.origin );
  if (rc) rc = ReadVector( plane.xaxis );
  if (rc) rc = ReadVector( plane.yaxis );
  if (rc) rc = ReadVector( plane.zaxis );
  if (rc) rc = ReadDouble( 4, plane.equation );
  return rc;
}

BOOL 
ON_BinaryArchive::WriteLine( const ON_Line& line )
{
  BOOL rc = WritePoint( line.from );
  if (rc) rc = WritePoint( line.to );
  return rc;
}

BOOL 
ON_BinaryArchive::ReadLine( ON_Line& line )
{
  BOOL rc = ReadPoint( line.from );
  if (rc) rc = ReadPoint( line.to );
  return rc;
}

BOOL 
ON_BinaryArchive::WriteArc(const ON_Arc& arc )
{
  BOOL rc = WriteCircle(arc);
  if (rc)
    rc = WriteInterval(arc.m_angle);
  return rc;
}

BOOL 
ON_BinaryArchive::ReadArc( ON_Arc& arc )
{
  BOOL rc = ReadCircle(arc);
  if (rc)
    rc = ReadInterval(arc.m_angle);
  return rc;
}

BOOL 
ON_BinaryArchive::WriteCircle(const ON_Circle& circle)
{
  BOOL rc = WritePlane( circle.plane );
  if (rc)
    rc = WriteDouble( circle.radius );
  if (rc)
    rc = WritePoint( circle.m_point[0] );
  if (rc)
    rc = WritePoint( circle.m_point[1] );
  if (rc)
    rc = WritePoint( circle.m_point[2] );
  return rc;
}

BOOL 
ON_BinaryArchive::ReadCircle(ON_Circle& circle)
{
  BOOL rc = ReadPlane( circle.plane );
  if (rc)
    rc = ReadDouble( &circle.radius );
  if (rc)
    rc = ReadPoint( circle.m_point[0] );
  if (rc)
    rc = ReadPoint( circle.m_point[1] );
  if (rc)
    rc = ReadPoint( circle.m_point[2] );
  return rc;
}


BOOL 
ON_BinaryArchive::WriteInterval( const ON_Interval& t )
{
  return WriteDouble( 2, t.m_t );
}

BOOL 
ON_BinaryArchive::ReadInterval( ON_Interval& t )
{
  return ReadDouble( 2, t.m_t );
}

BOOL
ON_BinaryArchive::ReadUuid( ON_UUID& uuid )
{
  BOOL rc = ReadLong( &uuid.Data1 );
  if (rc) rc = ReadShort( &uuid.Data2 );
  if (rc) rc = ReadShort( &uuid.Data3 );
  if (rc) rc = ReadChar( 8, uuid.Data4 );
  return rc;
}

BOOL
ON_BinaryArchive::ReadTime( struct tm& utc )
{
  // utc = coordinated universal time ( a.k.a GMT, UTC )
  // (From ANSI C time() and gmtime().)
  BOOL rc = ReadInt( &utc.tm_sec );
  if ( rc )
    rc = ReadInt( &utc.tm_min );
  if ( rc )
    rc = ReadInt( &utc.tm_hour );
  if ( rc )
    rc = ReadInt( &utc.tm_mday );
  if ( rc )
    rc = ReadInt( &utc.tm_mon );
  if ( rc )
    rc = ReadInt( &utc.tm_year );
  if ( rc )
    rc = ReadInt( &utc.tm_wday );
  if ( rc )
    rc = ReadInt( &utc.tm_yday );
  if ( rc ) {
    if ( utc.tm_sec < 0 || utc.tm_sec > 60 )
      rc = FALSE;
    if ( utc.tm_min < 0 || utc.tm_min > 60 )
      rc = FALSE;
    if ( utc.tm_hour < 0 || utc.tm_hour > 24 )
      rc = FALSE;
    if ( utc.tm_mday < 0 || utc.tm_mday > 31 )
      rc = FALSE;
    if ( utc.tm_mon < 0 || utc.tm_mon > 12 )
      rc = FALSE;
    if ( utc.tm_year < 0 || utc.tm_year > 2100 )
      rc = FALSE;
    if ( utc.tm_wday < 0 || utc.tm_wday > 7 )
      rc = FALSE;
    if ( utc.tm_yday < 0 || utc.tm_yday > 366 )
      rc = FALSE;
    if ( !rc ) {
      ON_ERROR("ON_BinaryArchive::ReadTime() - bad time in archive");
    }
  }
  return rc;
}

BOOL
ON_BinaryArchive::ReadStringSize( // Read size of NULL terminated string
    size_t* sizeof_string          // (returned size includes NULL terminator)
    )
{
  return ReadInt( sizeof_string );
}

BOOL
ON_BinaryArchive::ReadString(     // Read NULL terminated string
    size_t sizeof_string,          // length = value from ReadStringSize()
    char* p            // array[length]
    )
{
  return ReadByte( sizeof_string, p );
}

BOOL
ON_BinaryArchive::ReadString(     // Read NULL terminated string
    size_t sizeof_string,          // length = value from ReadStringSize()
    unsigned char* p  // array[length]
    )
{
  return ReadByte( sizeof_string, p );
}

BOOL
ON_BinaryArchive::ReadString(    // Read NULL terminated unicode string
    size_t sizeof_string,          // length = value from ReadStringSize()
    unsigned short* p // array[length]
    )
{
  return ReadShort( sizeof_string, p );
}

BOOL
ON_BinaryArchive::ReadString( ON_String& s )
{
  s.Destroy();
  size_t sizeof_string = 0;
  BOOL rc = ReadStringSize( &sizeof_string );
  if ( rc && sizeof_string > 0 ) {
    s.ReserveArray(sizeof_string);
    ReadString( sizeof_string, s.Array() );
    s.SetLength( sizeof_string-1 );
  }
  return rc;
}

BOOL
ON_BinaryArchive::ReadString( ON_wString& s )
{
  s.Destroy();
  size_t sizeof_string = 0;
  BOOL rc = ReadStringSize( &sizeof_string );
  if ( rc && sizeof_string > 0 ) {
    s.ReserveArray(sizeof_string);
    if ( 2 == sizeof(wchar_t) ) {
      // the (unsigned short*) cast is needed to get this code to
      // compile when 2 != sizeof(wchar_t), even though the code
      // will not be executed.
      rc = ReadString( sizeof_string, (unsigned short*)s.Array() );
    }
    else {
      // UNICODE strings are stored in 3DM files as arrays of unsigned shorts
      //
      // This function is used to read strings on systems where wchar_t is
      // not 2 bytes.
      //
      // Microsoft's wchar_t is an unsigned short
      // gcc's wchar_t is an unsigned long.
      wchar_t* w = s.Array();
      size_t i;
      unsigned short c; // assumes unsigned shorts are 2 bytes
      for ( i = 0; i < sizeof_string && rc; i++ ) {
        rc = ReadShort(&c);
        w[i] = c;
      }
    }
    s.SetLength( sizeof_string-1 );
  }
  return rc;
}

BOOL
ON_BinaryArchive::ReadArray( ON_SimpleArray<char>& a )
{
  a.Empty();
  int count = 0;
  BOOL rc = ReadInt( &count );
  if ( rc && count > 0 ) {
    a.SetCapacity( count );
    rc = ReadChar( count, a.Array() );
    if ( rc )
      a.SetCount(count);
  }
  return rc;
}

BOOL
ON_BinaryArchive::ReadArray( ON_SimpleArray<short>& a )
{
  a.Empty();
  int count = 0;
  BOOL rc = ReadInt( &count );
  if ( rc && count > 0 ) {
    a.SetCapacity( count );
    rc = ReadShort( count, a.Array() );
    if ( rc )
      a.SetCount(count);
  }
  return rc;
}

BOOL
ON_BinaryArchive::ReadArray( ON_SimpleArray<int>& a )
{
  a.Empty();
  int count = 0;
  BOOL rc = ReadInt( &count );
  if ( rc && count > 0 ) {
    a.SetCapacity( count );
    rc = ReadInt( count, a.Array() );
    if ( rc )
      a.SetCount(count);
  }
  return rc;
}

BOOL
ON_BinaryArchive::ReadArray( ON_SimpleArray<float>& a )
{
  a.Empty();
  int count = 0;
  BOOL rc = ReadInt( &count );
  if ( rc && count > 0 ) {
    a.SetCapacity( count );
    rc = ReadFloat( count, a.Array() );
    if ( rc )
      a.SetCount(count);
  }
  return rc;
}

BOOL
ON_BinaryArchive::ReadArray( ON_SimpleArray<double>& a )
{
  a.Empty();
  int count = 0;
  BOOL rc = ReadInt( &count );
  if ( rc && count > 0 ) {
    a.SetCapacity( count );
    rc = ReadDouble( count, a.Array() );
    if ( rc )
      a.SetCount(count);
  }
  return rc;
}

BOOL
ON_BinaryArchive::ReadArray( ON_SimpleArray<ON_Color>& a )
{
  a.Empty();
  int count = 0;
  BOOL rc = ReadInt( &count );
  if ( rc && count > 0 ) {
    a.SetCapacity( count );
    rc = ReadInt( count, (int*)a.Array() );
    if ( rc )
      a.SetCount(count);
  }
  return rc;
}


BOOL
ON_BinaryArchive::ReadArray( ON_SimpleArray<ON_2dPoint>& a )
{
  a.Empty();
  int count = 0;
  BOOL rc = ReadInt( &count );
  if ( rc && count > 0 ) {
    a.SetCapacity( count );
    rc = ReadDouble( 2*count, &a.Array()->x );
    if ( rc )
      a.SetCount(count);
  }
  return rc;
}

BOOL
ON_BinaryArchive::ReadArray( ON_SimpleArray<ON_3dPoint>& a )
{
  a.Empty();
  int count = 0;
  BOOL rc = ReadInt( &count );
  if ( rc && count > 0 ) {
    a.SetCapacity( count );
    rc = ReadDouble( 3*count, &a.Array()->x );
    if ( rc )
      a.SetCount(count);
  }
  return rc;
}

BOOL
ON_BinaryArchive::ReadArray( ON_SimpleArray<ON_4dPoint>& a )
{
  a.Empty();
  int count = 0;
  BOOL rc = ReadInt( &count );
  if ( rc && count > 0 ) {
    a.SetCapacity( count );
    rc = ReadDouble( 4*count, &a.Array()->x );
    if ( rc )
      a.SetCount(count);
  }
  return rc;
}

BOOL
ON_BinaryArchive::ReadArray( ON_SimpleArray<ON_2dVector>& a )
{
  a.Empty();
  int count = 0;
  BOOL rc = ReadInt( &count );
  if ( rc && count > 0 ) {
    a.SetCapacity( count );
    rc = ReadDouble( 2*count, &a.Array()->x );
    if ( rc )
      a.SetCount(count);
  }
  return rc;
}

BOOL
ON_BinaryArchive::ReadArray( ON_SimpleArray<ON_3dVector>& a )
{
  a.Empty();
  int count = 0;
  BOOL rc = ReadInt( &count );
  if ( rc && count > 0 ) {
    a.SetCapacity( count );
    rc = ReadDouble( 3*count, &a.Array()->x );
    if ( rc )
      a.SetCount(count);
  }
  return rc;
}

BOOL
ON_BinaryArchive::ReadArray( ON_SimpleArray<ON_2fPoint>& a )
{
  a.Empty();
  int count = 0;
  BOOL rc = ReadInt( &count );
  if ( rc && count > 0 ) {
    a.SetCapacity( count );
    rc = ReadFloat( 2*count, &a.Array()->x );
    if ( rc )
      a.SetCount(count);
  }
  return rc;
}

BOOL
ON_BinaryArchive::ReadArray( ON_SimpleArray<ON_3fPoint>& a )
{
  a.Empty();
  int count = 0;
  BOOL rc = ReadInt( &count );
  if ( rc && count > 0 ) {
    a.SetCapacity( count );
    rc = ReadFloat( 3*count, &a.Array()->x );
    if ( rc )
      a.SetCount(count);
  }
  return rc;
}

BOOL
ON_BinaryArchive::ReadArray( ON_SimpleArray<ON_4fPoint>& a )
{
  a.Empty();
  int count = 0;
  BOOL rc = ReadInt( &count );
  if ( rc && count > 0 ) {
    a.SetCapacity( count );
    rc = ReadFloat( 4*count, &a.Array()->x );
    if ( rc )
      a.SetCount(count);
  }
  return rc;
}

BOOL
ON_BinaryArchive::ReadArray( ON_SimpleArray<ON_2fVector>& a )
{
  a.Empty();
  int count = 0;
  BOOL rc = ReadInt( &count );
  if ( rc && count > 0 ) {
    a.SetCapacity( count );
    rc = ReadFloat( 2*count, &a.Array()->x );
    if ( rc )
      a.SetCount(count);
  }
  return rc;
}

BOOL
ON_BinaryArchive::ReadArray( ON_SimpleArray<ON_3fVector>& a )
{
  a.Empty();
  int count = 0;
  BOOL rc = ReadInt( &count );
  if ( rc && count > 0 ) {
    a.SetCapacity( count );
    rc = ReadFloat( 3*count, &a.Array()->x );
    if ( rc )
      a.SetCount(count);
  }
  return rc;
}

BOOL
ON_BinaryArchive::ReadArray( ON_SimpleArray<ON_SurfaceCurvature>& a )
{
  a.Empty();
  int count = 0;
  BOOL rc = ReadInt( &count );
  if ( rc && count > 0 ) {
    a.SetCapacity( count );
    rc = ReadDouble( 2*count, &a.Array()->k1 );
    if ( rc )
      a.SetCount(count);
  }
  return rc;
}

BOOL
ON_BinaryArchive::WriteChar(    // Write an array of 8 bit chars
		size_t count,       // number of chars to write
		const char* p   
		)
{
  return WriteByte( count, p );
}

BOOL
ON_BinaryArchive::WriteChar(    // Write an array of 8 bit unsigned chars
		size_t count,       // number of unsigned chars to write
		const unsigned char* p
		)
{
  return WriteByte( count, p );
}

BOOL
ON_BinaryArchive::WriteChar(    // Write a single 8 bit char
		char c
		)
{
  return WriteByte( 1, &c );
}

BOOL
ON_BinaryArchive::WriteChar(    // Write a single 8 bit unsigned char
		unsigned char c
		)
{
  return WriteByte( 1, &c );
}

BOOL
ON_BinaryArchive::WriteShort(   // Write an array of 16 bit shorts
		size_t count,       // number of shorts to write
		const short* p
		)
{
  BOOL rc = TRUE;
  if ( m_endian == ON::big_endian ) {
    if ( count > 0 ) {
      const char* b = (const char*)p;
      while ( rc && count-- ) {
        rc = WriteByte( 1, b+1 );
        if (rc)
          rc = WriteByte( 1, b );
        b++;
        b++;
      }
    }
  }
  else {
    rc = WriteByte( count<<1, p );
  }
  return rc;
}

BOOL
ON_BinaryArchive::WriteShort(   // Write an array of 16 bit unsigned shorts
		size_t count,       // number of shorts to write
		const unsigned short* p
		)
{
  return WriteShort( count, (const short*)p );
}
BOOL
ON_BinaryArchive::WriteShort(   // Write a single 16 bit short
		short s
		)
{
  return WriteShort( 1, &s );
}

BOOL
ON_BinaryArchive::WriteShort(   // Write a single 16 bit unsigned short
		unsigned short s
		)
{
  return WriteShort( 1, &s );
}

BOOL
ON_BinaryArchive::WriteInt( // Write an array of 32 bit integers
		size_t count,	      // number of ints to write
		const int* p    
		)
{
  BOOL rc = TRUE;
  if ( m_endian == ON::big_endian ) {
    if ( count > 0 ) {
      const char* b = (const char*)p;
      while ( rc && count-- ) {
        rc = WriteByte( 1, b+3 );
        if (rc) rc = WriteByte( 1, b+2 );
        if (rc) rc = WriteByte( 1, b+1 );
        if (rc) rc = WriteByte( 1, b );
        b += 4;
      }
    }
  }
  else {
    rc = WriteByte( count<<2, p );
  }
  return rc;
}

BOOL
ON_BinaryArchive::WriteInt( // Write an array of 32 bit integers
		size_t count,	      // number of ints to write
		const unsigned int* p
		)
{
  return WriteInt( count, (const int*)p );
}

BOOL
ON_BinaryArchive::WriteInt( // Write a single 32 bit integer
		int i
		)
{
  return WriteInt( 1, &i );
}

BOOL
ON_BinaryArchive::WriteInt( // Write a single 32 bit integer
		unsigned int i
		)
{
  return WriteInt( 1, &i );
}

BOOL
ON_BinaryArchive::WriteLong( // Write an array of 32 bit integers
		size_t count,	      // number of ints to write
		const long* p    
		)
{
  return WriteInt( count, (const int*)p );
}

BOOL
ON_BinaryArchive::WriteLong( // Write an array of 32 bit longs
		size_t count,	      // number of ints to write
		const unsigned long* p
		)
{
  return WriteInt( count, (const int*)p );
}

BOOL
ON_BinaryArchive::WriteLong( // Write a single 32 bit long
		long i
		)
{
  return WriteInt( 1, (const int*)&i );
}

BOOL
ON_BinaryArchive::WriteLong( // Write a single 32 bit unsigned long
		unsigned long i
		)
{
  return WriteInt( 1, (const int*)&i );
}


BOOL
ON_BinaryArchive::WriteFloat(   // Write a number of IEEE floats
		size_t count,       // number of doubles
		const float* p
		)
{
  // floats and integers have same size and endian issues
  return WriteInt( count, (const int*)p );
}

BOOL
ON_BinaryArchive::WriteFloat(   // Write a single float
		float f
		)
{
  return WriteFloat( 1, &f );
}

BOOL
ON_BinaryArchive::WriteDouble(  // Write a single double
		size_t count,       // number of doubles
		const double* p
		)
{
  BOOL rc = TRUE;
  if ( m_endian == ON::big_endian ) {
    if ( count > 0 ) {
      const char* b = (const char*)p;
      while ( rc && count-- ) {
        rc = WriteByte( 1, b+7 );
        if (rc) rc = WriteByte( 1, b+6 );
        if (rc) rc = WriteByte( 1, b+5 );
        if (rc) rc = WriteByte( 1, b+4 );
        if (rc) rc = WriteByte( 1, b+3 );
        if (rc) rc = WriteByte( 1, b+2 );
        if (rc) rc = WriteByte( 1, b+1 );
        if (rc) rc = WriteByte( 1, b );
        b += 8;
      }
    }
  }
  else {
    rc = WriteByte( count<<3, p );
  }
  return rc;
}


BOOL
ON_BinaryArchive::WriteDouble(  // Write a single double
		const double x
		)
{
  return WriteDouble( 1, &x );
}

BOOL
ON_BinaryArchive::WriteColor( const ON_Color& color )
{
  unsigned int colorref = color;
  return WriteInt( colorref );
}

BOOL ON_BinaryArchive::WriteLineStyle( const ON_LineStyle& ls )
{
  short line_pattern = ls.Pattern();
  short line_pattern_index = ls.PatternIndex();
  BOOL rc = WriteShort( line_pattern );
  if (rc) rc = WriteShort( line_pattern_index );
  if (rc) rc = WriteDouble( ls.Thickness() );
  if (rc) rc = WriteDouble( ls.Scale() );
  return rc;
}

BOOL
ON_BinaryArchive::WritePoint (
  const ON_2dPoint& p
  )
{
  return WriteDouble( 2, &p.x );
}

BOOL
ON_BinaryArchive::WritePoint (
  const ON_3dPoint& p
  )
{
  return WriteDouble( 3, &p.x );
}

BOOL
ON_BinaryArchive::WritePoint (
  const ON_4dPoint& p
  )
{
  return WriteDouble( 4, &p.x );
}

BOOL
ON_BinaryArchive::WriteVector (
  const ON_2dVector& v
  )
{
  return WriteDouble( 2, &v.x );
}

BOOL
ON_BinaryArchive::WriteVector (
  const ON_3dVector& v
  )
{
  return WriteDouble( 3, &v.x );
}


BOOL
ON_BinaryArchive::WriteUuid( const ON_UUID& uuid )
{
  BOOL rc = WriteLong( uuid.Data1 );
  if (rc) rc = WriteShort( uuid.Data2 );
  if (rc) rc = WriteShort( uuid.Data3 );
  if (rc) rc = WriteChar( 8, uuid.Data4 );
  return rc;
}

BOOL
ON_BinaryArchive::WriteTime( const struct tm& utc )
{
  // utc = coordinated universal time ( a.k.a GMT, UTC )
  // (From ANSI C time() and gmtime().)
  // The checks are here so we can insure files don't contain
  // garbage dates and ReadTime() can treat out of bounds
  // values as file corruption errors.
  int i;
  i = (int)utc.tm_sec;  if ( i < 0 || i > 60 ) i = 0;
  BOOL rc = WriteInt( i );
  i = (int)utc.tm_min;  if ( i < 0 || i > 60 ) i = 0;
  if ( rc )
    rc = WriteInt( i );
  i = (int)utc.tm_hour;  if ( i < 0 || i > 24 ) i = 0;
  if ( rc )
    rc = WriteInt( i );
  i = (int)utc.tm_mday;  if ( i < 0 || i > 31 ) i = 0;
  if ( rc )
    rc = WriteInt( i );
  i = (int)utc.tm_mon;  if ( i < 0 || i > 12 ) i = 0;
  if ( rc )
    rc = WriteInt( i );
  i = (int)utc.tm_year;  if ( i < 1900 || i > 2100 ) i = 0;
  if ( rc )
    rc = WriteInt( i );
  i = (int)utc.tm_wday;  if ( i < 0 || i > 7 ) i = 0;
  if ( rc )
    rc = WriteInt( i );
  i = (int)utc.tm_yday;  if ( i < 0 || i > 366 ) i = 0;
  if ( rc )
    rc = WriteInt( i );
  return rc;
}

// To read strings written with WriteString(), first call
// ReadStringLength() to get the length of the string (including
// the NULL terminator, then call ReadString() with enought memory
// to load the string.
BOOL
ON_BinaryArchive::WriteString( // Write NULL terminated string
    const char* s
    )
{
  size_t sizeof_string = 0;
  if ( s ) {
    while(s[sizeof_string])
      sizeof_string++;
    if ( sizeof_string )
      sizeof_string++;
  }
  BOOL rc = WriteInt(sizeof_string);
  if ( rc && sizeof_string > 0 )
    rc = WriteByte( sizeof_string, s );
  return rc;  
}

BOOL
ON_BinaryArchive::WriteString( // Write NULL terminated string
    const unsigned char* s
    )
{
  return WriteString( (const char*)s );
}

BOOL
ON_BinaryArchive::WriteString(  // Write NULL terminated unicode string
    const unsigned short* s
    )
{
  size_t sizeof_string = 0;
  if ( s ) {
    while(s[sizeof_string])
      sizeof_string++;
    if ( sizeof_string )
      sizeof_string++;
  }
  BOOL rc = WriteInt(sizeof_string);
  if ( rc && sizeof_string > 0 )
    rc = WriteShort( sizeof_string, s );
  return rc;  
}

BOOL
ON_BinaryArchive::WriteString( const ON_String& s )
{
  size_t sizeof_string = s.Length();
  if ( sizeof_string )
    sizeof_string++;
  BOOL rc = WriteInt(sizeof_string);
  if ( rc && sizeof_string > 0 )
    rc = WriteByte( sizeof_string, s.Array() );
  return rc;  
}

BOOL
ON_BinaryArchive::WriteString( const ON_wString& s )
{
  size_t sizeof_string = s.Length();
  if ( sizeof_string )
    sizeof_string++;
  BOOL rc = WriteInt(sizeof_string);
  if ( rc && sizeof_string > 0 ) {
    if ( 2 == sizeof(wchar_t) ) {
      // the (unsigned short*) cast is neede to get this code
      // to compile on systems where wchar_t's are not 2 bytes,
      // even though the code is never executed.
      rc = WriteShort( sizeof_string, (unsigned short*)s.Array() );
    }
    else {
      // UNICODE strings are stored in 3DM files as arrays of unsigned 2 byte shorts.
      // This code is used on systems where wchar_t is not a 2 byte short.
      size_t i;
      unsigned short c; // assumes unsigned shorts are 2 bytes
      const wchar_t* w = s.Array();
      for ( i = 0; i < sizeof_string && rc; i++ ) {
        c = w[i];
        rc = WriteShort(c);
      }
    }
  }
  return rc;  
}

BOOL
ON_BinaryArchive::WriteArray( const ON_SimpleArray<char>& a )
{
  int count = a.Count();
  if ( count < 0 )
    count = 0;
  BOOL rc = WriteInt( count );
  if ( rc && count > 0 ) {
    rc = WriteChar( count, a.Array() );
  }
  return rc;
}

BOOL
ON_BinaryArchive::WriteArray( const ON_SimpleArray<short>& a )
{
  int count = a.Count();
  if ( count < 0 )
    count = 0;
  BOOL rc = WriteInt( count );
  if ( rc && count > 0 ) {
    rc = WriteShort( count, a.Array() );
  }
  return rc;
}

BOOL
ON_BinaryArchive::WriteArray( const ON_SimpleArray<int>& a )
{
  int count = a.Count();
  if ( count < 0 )
    count = 0;
  BOOL rc = WriteInt( count );
  if ( rc && count > 0 ) {
    rc = WriteInt( count, a.Array() );
  }
  return rc;
}

BOOL
ON_BinaryArchive::WriteArray( const ON_SimpleArray<float>& a )
{
  int count = a.Count();
  if ( count < 0 )
    count = 0;
  BOOL rc = WriteInt( count );
  if ( rc && count > 0 ) {
    rc = WriteFloat( count, a.Array() );
  }
  return rc;
}

BOOL
ON_BinaryArchive::WriteArray( const ON_SimpleArray<double>& a )
{
  int count = a.Count();
  if ( count < 0 )
    count = 0;
  BOOL rc = WriteInt( count );
  if ( rc && count > 0 ) {
    rc = WriteDouble( count, a.Array() );
  }
  return rc;
}

BOOL
ON_BinaryArchive::WriteArray( const ON_SimpleArray<ON_Color>& a )
{
  int count = a.Count();
  if ( count < 0 )
    count = 0;
  BOOL rc = WriteInt( count );
  if ( rc && count > 0 ) {
    rc = WriteInt( count, (int*)a.Array() );
  }
  return rc;
}

BOOL
ON_BinaryArchive::WriteArray( const ON_SimpleArray<ON_2dPoint>& a )
{
  int count = a.Count();
  if ( count < 0 )
    count = 0;
  BOOL rc = WriteInt( count );
  if ( rc && count > 0 ) {
    rc = WriteDouble( count*2, &a.Array()->x );
  }
  return rc;
}

BOOL
ON_BinaryArchive::WriteArray( const ON_SimpleArray<ON_3dPoint>& a )
{
  int count = a.Count();
  if ( count < 0 )
    count = 0;
  BOOL rc = WriteInt( count );
  if ( rc && count > 0 ) {
    rc = WriteDouble( count*3, &a.Array()->x );
  }
  return rc;
}

BOOL
ON_BinaryArchive::WriteArray( const ON_SimpleArray<ON_4dPoint>& a )
{
  int count = a.Count();
  if ( count < 0 )
    count = 0;
  BOOL rc = WriteInt( count );
  if ( rc && count > 0 ) {
    rc = WriteDouble( count*4, &a.Array()->x );
  }
  return rc;
}

BOOL
ON_BinaryArchive::WriteArray( const ON_SimpleArray<ON_2dVector>& a )
{
  int count = a.Count();
  if ( count < 0 )
    count = 0;
  BOOL rc = WriteInt( count );
  if ( rc && count > 0 ) {
    rc = WriteDouble( count*2, &a.Array()->x );
  }
  return rc;
}


BOOL
ON_BinaryArchive::WriteArray( const ON_SimpleArray<ON_3dVector>& a )
{
  int count = a.Count();
  if ( count < 0 )
    count = 0;
  BOOL rc = WriteInt( count );
  if ( rc && count > 0 ) {
    rc = WriteDouble( count*3, &a.Array()->x );
  }
  return rc;
}

BOOL
ON_BinaryArchive::WriteArray( const ON_SimpleArray<ON_2fPoint>& a )
{
  int count = a.Count();
  if ( count < 0 )
    count = 0;
  BOOL rc = WriteInt( count );
  if ( rc && count > 0 ) {
    rc = WriteFloat( count*2, &a.Array()->x );
  }
  return rc;
}

BOOL
ON_BinaryArchive::WriteArray( const ON_SimpleArray<ON_3fPoint>& a )
{
  int count = a.Count();
  if ( count < 0 )
    count = 0;
  BOOL rc = WriteInt( count );
  if ( rc && count > 0 ) {
    rc = WriteFloat( count*3, &a.Array()->x );
  }
  return rc;
}

BOOL
ON_BinaryArchive::WriteArray( const ON_SimpleArray<ON_4fPoint>& a )
{
  int count = a.Count();
  if ( count < 0 )
    count = 0;
  BOOL rc = WriteInt( count );
  if ( rc && count > 0 ) {
    rc = WriteFloat( count*4, &a.Array()->x );
  }
  return rc;
}

BOOL
ON_BinaryArchive::WriteArray( const ON_SimpleArray<ON_2fVector>& a )
{
  int count = a.Count();
  if ( count < 0 )
    count = 0;
  BOOL rc = WriteInt( count );
  if ( rc && count > 0 ) {
    rc = WriteFloat( count*2, &a.Array()->x );
  }
  return rc;
}

BOOL
ON_BinaryArchive::WriteArray( const ON_SimpleArray<ON_3fVector>& a )
{
  int count = a.Count();
  if ( count < 0 )
    count = 0;
  BOOL rc = WriteInt( count );
  if ( rc && count > 0 ) {
    rc = WriteFloat( count*3, &a.Array()->x );
  }
  return rc;
}


BOOL
ON_BinaryArchive::WriteArray( const ON_SimpleArray<ON_SurfaceCurvature>& a )
{
  int count = a.Count();
  if ( count < 0 )
    count = 0;
  BOOL rc = WriteInt( count );
  if ( rc && count > 0 ) {
    rc = WriteDouble( count*2, &a.Array()->k1 );
  }
  return rc;
}

/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////

BOOL
ON_BinaryArchive::WriteObject( const ON_Object* o )
{
  BOOL rc = FALSE;
  if ( o )
    rc = WriteObject(*o);
  else {
    // NULL object has nil UUID and no date
    rc = BeginWrite3dmChunk( TCODE_OPENNURBS_CLASS, 0 );
    if (rc) {
      rc = BeginWrite3dmChunk( TCODE_OPENNURBS_CLASS_UUID, 0 );
      if ( rc ) {
        rc = WriteUuid( ON_nil_uuid );
        if ( !EndWrite3dmChunk() ) // end of TCODE_OPENNURBS_CLASS_UUID chunk
          rc = FALSE;
      }
      if ( !EndWrite3dmChunk() )
        rc = FALSE;
    }
  }
  return rc;
}

BOOL
ON_BinaryArchive::WriteObject( const ON_Object& o )
{
  // writes polymorphic object derived from ON_Object in a way that
  // it can be recreated from ON_BinaryArchive::ReadObject
  ON_UUID uuid;
  BOOL rc = FALSE;
  const ON_ClassId* pID = o.ClassId();
  if ( !pID ) {
    ON_ERROR("ON_BinaryArchive::WriteObject() o.ClassId() returned NULL.");
    return FALSE;
  }

  rc = BeginWrite3dmChunk( TCODE_OPENNURBS_CLASS, 0 );
  if (rc) {

    // TCODE_OPENNURBS_CLASS_UUID chunk contains class's UUID 
    uuid = pID->Uuid();
    rc = BeginWrite3dmChunk( TCODE_OPENNURBS_CLASS_UUID, 0 );
    if ( rc ) {
      rc = WriteUuid( uuid );
      if ( !EndWrite3dmChunk() ) // end of TCODE_OPENNURBS_CLASS_UUID chunk
        rc = FALSE;
    }

    // TCODE_OPENNURBS_CLASS_DATA chunk contains definition of class
    if ( rc ) {
      rc = BeginWrite3dmChunk( TCODE_OPENNURBS_CLASS_DATA, 0 );
      if (rc) {
        rc = o.Write( *this );
        if ( !rc ) {
          ON_ERROR("ON_BinaryArchive::WriteObject() o.Write() failed.");
        }
        if ( !EndWrite3dmChunk() ) // end of TCODE_OPENNURBS_CLASS_DATA chunk
          rc = FALSE;
      }
    }

    // TCODE_OPENNURBS_CLASS_USERDATA chunk contains attached user data
    if (rc) {
      const ON_UserData* ud;
      ON_UUID userdata_classid;
      for (ud = o.m_userdata_list; ud && rc; ud = ud->m_userdata_next) {
        rc = BeginWrite3dmChunk( TCODE_OPENNURBS_CLASS_USERDATA, 0 );
        if (rc) {
          userdata_classid = ud->UserDataClassUuid();
          rc = Write3dmChunkVersion(2,0);
          // wrap user data header info in an anonymous chunk
          rc = BeginWrite3dmChunk( TCODE_OPENNURBS_CLASS_USERDATA_HEADER, 0 );
          if (rc) {
            if ( rc ) rc = WriteUuid( userdata_classid );
            if ( rc ) rc = WriteUuid( ud->m_userdata_uuid );
            if ( rc ) rc = WriteInt( ud->m_userdata_copycount );
            if ( rc ) rc = WriteXform( ud->m_userdata_xform );
            if ( !EndWrite3dmChunk() )
              rc = FALSE;
          }
          if (rc) {
            // wrap user data in an anonymous chunk
            rc = BeginWrite3dmChunk( TCODE_ANONYMOUS_CHUNK, 0 );
            if ( rc ) {
              rc = ud->Write(*this);
              if ( !EndWrite3dmChunk() )
                rc = FALSE;
            }
          }
          if ( !EndWrite3dmChunk() )
            rc = FALSE;
        }
      }
    }

    // TCODE_OPENNURBS_CLASS_END chunk marks end of class record
    if ( BeginWrite3dmChunk( TCODE_OPENNURBS_CLASS_END, 0 ) ) {
      if ( !EndWrite3dmChunk() )
        rc = FALSE;
    }
    else
      rc = FALSE;

    if ( !EndWrite3dmChunk() ) // end of TCODE_OPENNURBS_CLASS chunk
      rc = FALSE;
  }

  return rc;
}

BOOL
ON_BinaryArchive::ReadObject( ON_Object** ppObject )
{
  // returns 0: failure - unable to read object because of file IO problems
  //         1: success
  //         3: unable to read object because it's UUID is not registered
  //            this could happen in cases where old code is attempting to read
  //            new objects.
  unsigned int tcode;
  int length;
  ON_UUID uuid;
  const ON_ClassId* pID = 0;
  ON_Object* pObject = NULL;
  BOOL rc = FALSE;

  if ( !ppObject ) {
    ON_ERROR("ON_BinaryArchive::ReadObject() called with NULL ppObject.");
    return FALSE;
  }
  *ppObject = NULL;

  // all ON_Objects written by WriteObject are in a TCODE_OPENNURBS_CLASS chunk
  rc = BeginRead3dmChunk( &tcode, &length );
  if ( !rc )
    return FALSE;

  if ( tcode != TCODE_OPENNURBS_CLASS ) {
    ON_ERROR("ON_BinaryArchive::ReadObject() didn't find TCODE_OPENNURBS_CLASS block.");
    rc = FALSE;
  }
  else 
  {
    // we break out of this loop if something bad happens
    for (;;) {
      // read class's UUID ///////////////////////////////////////////////////////////
      rc = BeginRead3dmChunk( &tcode, &length );
      if ( !rc )
        break;
      if ( tcode != TCODE_OPENNURBS_CLASS_UUID ) {
        ON_ERROR("ON_BinaryArchive::ReadObject() didn't find TCODE_OPENNURBS_CLASS_UUID block");
        rc = FALSE;
      }
      else if ( length != 20 ) {
        ON_ERROR("ON_BinaryArchive::ReadObject() TCODE_OPENNURBS_CLASS_UUID has size != 16+4");
        rc = FALSE;
      }
      else if ( !ReadUuid( uuid ) ) {
        rc = FALSE;
      }
      if ( !EndRead3dmChunk() ) {
        rc = FALSE;
      }
      if ( !rc ) {
        break;
      }
      ///////////////////////////////////////////////////////////////////////////////

      if ( !ON_UuidCompare( &uuid, &ON_nil_uuid ) ) {
        // nil UUID written if NULL pointer is passed to WriteObject();
        rc = TRUE;
        break;
      }

      // Use UUID to get ON_ClassId for this class //////////////////////////////////
      pID = ON_ClassId::ClassId( uuid );
      if ( !pID ) {
        // Either the uuid is garbage or you are attempting to read an object with old code.
        // Visit http://www.opennurbs.org to get the latest OpenNURBS code.
        ON_WARNING("ON_BinaryArchive::ReadObject() ON_ClassId::ClassId(uuid) returned NULL.");
        rc = 3;
        break;
      }
      ///////////////////////////////////////////////////////////////////////////////

      // read class's definitions   /////////////////////////////////////////////////
      rc = BeginRead3dmChunk( &tcode, &length );
      if ( !rc )
        break;
      if ( tcode != TCODE_OPENNURBS_CLASS_DATA ) {
        ON_ERROR("ON_BinaryArchive::ReadObject() didn't find TCODE_OPENNURBS_CLASS_DATA block");
        rc = FALSE;
      }
      else {
        pObject = pID->Create();
        if ( !pObject ) {
          ON_ERROR("ON_BinaryArchive::ReadObject() pID->Create() returned NULL.");
          rc = FALSE;
        }
        else {
          rc = pObject->Read(*this);
          if ( !rc ) {
            ON_ERROR("ON_BinaryArchive::ReadObject() pObject->Read() failed.");
            delete pObject;
          }
          else {
            *ppObject = pObject;
          }
        }
      }
      if ( !EndRead3dmChunk() ) {
        rc = FALSE;
      }

      // read user data  /////////////////////////////////////////////////
      while(rc) {
        rc = BeginRead3dmChunk( &tcode, &length );
        if ( !rc )
          break;
        if ( tcode == TCODE_OPENNURBS_CLASS_USERDATA ) {
          int major_userdata_version = 0;
          int minor_userdata_version = 0;
          unsigned int t;
          int v;
          rc = Read3dmChunkVersion( &major_userdata_version, &minor_userdata_version );
          if ( rc && (major_userdata_version == 1 || major_userdata_version == 2) ) {
            ON_UUID userdata_classid;
            ON_UUID userdata_itemid;
            int userdata_copycount = 0;
            ON_Xform userdata_xform(0);
            if ( major_userdata_version == 2 ) {
              rc = BeginRead3dmChunk( &t, &v );
              if ( t != TCODE_OPENNURBS_CLASS_USERDATA_HEADER ) {
                ON_ERROR("version 2.0 TCODE_OPENNURBS_CLASS_USERDATA chunk is missing TCODE_OPENNURBS_CLASS_USERDATA_HEADER chunk.");
                rc = FALSE;
                EndRead3dmChunk();
              }
            }
            if (rc) {
              if (rc) rc = ReadUuid( userdata_classid );
              if (rc) rc = ReadUuid( userdata_itemid );
              if (rc) rc = ReadInt( &userdata_copycount );
              if (rc) rc = ReadXform( userdata_xform );
              if ( major_userdata_version == 2 ) {
                if ( !EndRead3dmChunk() ) // end of TCODE_OPENNURBS_CLASS_USERDATA_HEADER
                  rc = FALSE;
              }
            }
            if (rc) {
              rc = BeginRead3dmChunk( &t, &v );
              if (rc) {
                if ( TCODE_ANONYMOUS_CHUNK == t ) {
                  const ON_ClassId* udId = ON_ClassId::ClassId( userdata_classid );
                  if ( !udId ) // application that owns data is not loaded
                    udId = &ON_UnknownUserData::m_ON_UnknownUserData_class_id;
                  if ( !udId)
                    rc = FALSE;
                  else {
                    ON_Object* tmp = udId->Create();
                    if (tmp) {
                      ON_UserData* ud = ON_UserData::Cast(tmp);
                      if ( ud ) {
                        ud->m_userdata_copycount = userdata_copycount;
                        ud->m_userdata_uuid = userdata_itemid;
                        ud->m_userdata_xform = userdata_xform;
                        if ( ud->IsUnknownUserData() ) {
                          ON_UnknownUserData* uud = ON_UnknownUserData::Cast(ud);
                          if ( uud ) {
                            uud->m_sizeof_buffer = v;
                            uud->m_unknownclass_uuid = userdata_classid;
                          }
                        }
                        rc = ud->Read( *this );
                        if (rc) {
                          if ( !pObject->AttachUserData( ud ) )
                            delete ud;
                        }
                      }
                      else {
                        delete tmp;
                        rc = FALSE;
                      }
                    }
                    else
                      rc = FALSE;
                  }
                }
                else
                  rc = FALSE;
                if ( !EndRead3dmChunk() ) // end of TCODE_ANONYMOUS_CHUNK that's wrapped around user data
                  rc = FALSE;
              }
            }
          }
        }
        if (!EndRead3dmChunk()) {
          rc = FALSE;
          break;
        }
        if ( tcode == TCODE_OPENNURBS_CLASS_END )
          break; // done
      }
      break; // success
    }

  }
  if ( !EndRead3dmChunk() ) // TCODE_OPENNURBS_CLASS
    rc = FALSE;

  return rc;
}

BOOL ON_BinaryArchive::Write3dmChunkVersion(
  int major_version, // major // 0 to 15
  int minor_version // minor // 0 to 16
  )
{
  const unsigned char v = (unsigned char)(major_version*16+minor_version);
  return WriteChar( v );
}

BOOL ON_BinaryArchive::Read3dmChunkVersion(
  int* major_version, // major // 0 to 15
  int* minor_version // minor // 0 to 16
  )
{
  unsigned char v = 0;
  BOOL rc = ReadChar( &v );
  if ( minor_version) *minor_version = v%16;
  // The bit shift happens on the fly in the following
  // if statement.  It was happening twice which always
  // set the major version to 0
  //v >>= 4;
  if ( major_version) *major_version = (v>>4);
  return rc;
}

int ON_BinaryArchive::Archive3dmVersion() const
{
  return m_3dm_version;
}

/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////

BOOL 
ON_BinaryArchive::BeginWrite3dmChunk( unsigned int typecode, int value )
{
  m_bDoChunkCRC = FALSE; // no CRC on chunks because length is written twice.
  BOOL rc = WriteInt( typecode );
  if (rc)
    rc = WriteInt( value );
  if (rc)
    rc = PushChunk( typecode, value );
  return rc;
}

BOOL 
ON_BinaryArchive::EndWrite3dmChunk()
{
  int length = 0;
  BOOL rc = FALSE;
  ON_3DM_CHUNK* c = m_chunk.Last();
  if ( c ) {
    if ( c->m_do_length ) {
      if ( c->m_do_crc16 ) {
        // write 16 bit CRC
        unsigned char two_zero_bytes[2] = {0,0};
        unsigned short crc = ON_CRC16( c->m_crc16, 2, two_zero_bytes );
        rc = WriteShort( crc );
        if (c->m_crc16) {
          // should never happen unless ON_CRC16() code is damaged
          m_bad_CRC_count++;
          ON_ERROR("ON_BinaryArchive::EndWrite3dmChunk: CRC16 computation error.");
        }
      }
      else if ( c->m_do_crc32 ) {
        // write 32 bit CRC
        const unsigned long crc0 = c->m_crc32; 
        rc = WriteLong( crc0 );
      }
      else {
        rc = TRUE;
      }

      // write length
      m_bDoChunkCRC = 0;
      const int offset = CurrentPosition();
      length = offset - c->m_offset;
      if ( length < 0 ) {
        ON_ERROR("ON_BinaryArchive::EndWrite3dmChunk() - chunk length < 0");
        rc = FALSE;
      }
      else {
        if ( !SeekFromCurrentPosition( -length-4 ) ) {
          rc = FALSE;
        }
        else {
          if ( !WriteInt( length ) )
            rc = FALSE;
          if ( !SeekFromCurrentPosition( length ) ) {
            rc = FALSE;
          }
        }
        if ( CurrentPosition() != offset ) {
          ON_ERROR("ON_BinaryArchive::EndWrite3dmChunk() - CurrentPosition() != offset");
          rc = FALSE;
        }
      }
    }
    else {
      // "short" chunks are completely written by call to BeginWrite3dmChunk().
      rc = TRUE;
    }

    m_chunk.Remove();
    c = m_chunk.Last();
    if ( !c )
      Flush();
    m_bDoChunkCRC = c && (c->m_do_crc16 || c->m_do_crc32);
  }
  return rc;
}

BOOL ON_BinaryArchive::Write3dmGoo( const ON_3dmGoo& goo )
{
  BOOL rc = FALSE;

  if ( goo.m_typecode ) {
    const BOOL savedDoCRC = m_bDoChunkCRC;
    m_bDoChunkCRC = FALSE;
    if ( 0 != (goo.m_typecode & TCODE_SHORT) ) {
      if ( goo.m_value == 0 || (goo.m_value > 0 && goo.m_goo) ) {
        // write long chunk - do not use Begin/EndWrite3dmChunk() because
        //                    goo may contain subchunks and CRC would be
        //                    incorrectly computed.
        rc = WriteInt( goo.m_typecode );
        if (rc) rc = WriteInt( goo.m_value );
        if (rc && goo.m_value>0) rc = WriteByte( goo.m_value, goo.m_goo );
      }
    }
    else {
      // write short chunk
      rc = WriteInt( goo.m_typecode );
      if (rc) rc = WriteInt( goo.m_value );
    }
    m_bDoChunkCRC = savedDoCRC;
  }
  
  return rc;
}

BOOL ON_BinaryArchive::PeekAt3dmChunkType( unsigned int* typecode, int* value )
{
  const int pos0 = CurrentPosition();
  int t = 0, v = 0;
  BOOL rc = ReadInt( &t );
  if (rc) rc = ReadInt( &v );
  const int pos1 = CurrentPosition();
  if ( !SeekFromCurrentPosition( pos0-pos1 ) )
    rc = FALSE;
  if ( typecode )
    *typecode = t;
  if ( value )
    *value = v;
  return rc;
}

BOOL ON_BinaryArchive::Seek3dmChunkFromStart( 
      // beginning at the start of the active chunk, search portion of
      // archive included in active chunk for the start of a subchunk 
      // with the specified type.
      // if TRUE is returned, then the position is set so the next call to
      // BeginRead3dmChunk() will read a chunk with the specified typecode
      unsigned int typecode   // typecode from opennurbs_3dm.h
      )
{
  BOOL rc = FALSE;
  if ( ReadMode() ) {
    const int pos0 = CurrentPosition();
    const ON_3DM_CHUNK* c = m_chunk.Last();
    if ( c ) {
      // set archive position to the beginning of this chunk
      if ( 0 != (c->m_typecode | TCODE_SHORT) ) {
        ON_ERROR("ON_BinaryArchive::Seek3dmChunkFromStart() called with a short active chunk");
        return FALSE;
      }
      if ( pos0 < c->m_offset || pos0 > c->m_offset + c->m_value ) {
        ON_ERROR("ON_BinaryArchive::Seek3dmChunkFromStart() called with out of bounds current position");
        return FALSE;
      }
      rc = SeekFromCurrentPosition( c->m_offset - pos0 );
    }
    else {
      // set archive position to the beginning of archive chunks by skipping
      // 32 byte version info
      rc = SeekFromStart( 32 );
    }
    if (rc) {
      rc = Seek3dmChunkFromCurrentPosition( typecode );
    }
    if (!rc) {
      SeekFromStart(pos0);
    }
  }
  return rc;
}

BOOL ON_BinaryArchive::Seek3dmChunkFromCurrentPosition( 
      // beginning at the current position, search portion of archive
      // included in active chunk for the start of a subchunk with the
      // specified type.
      // if TRUE is returned, then the position is set so the next call to
      // BeginRead3dmChunk() will read a chunk with the specified typecode
      unsigned int typecode // typecode from opennurbs_3dm.h
      )
{
  BOOL rc = FALSE;
  if ( ReadMode() ) {
    const ON_3DM_CHUNK* c = m_chunk.Last();
    //const int pos0 = c ? c->m_offset : 32;
    const int pos1 = c ? c->m_offset+c->m_value : 0;
    const int pos_start = CurrentPosition();
    int pos_prev = -1;
    int pos = 0;
    unsigned int t;
    int v;
    while(pos > pos_prev) {
      pos_prev = pos;
      pos = CurrentPosition();
      if ( pos1 && pos > pos1 )
        break;
      t = !typecode;
      if ( !PeekAt3dmChunkType( &t, 0 ) )
        break;
      if ( t == typecode ) {
        rc = TRUE;
        break;
      }
      if ( !BeginRead3dmChunk( &t, &v ) )
        break;
      if ( !EndRead3dmChunk() )
        break;
    }
    if ( !rc ) {
      SeekFromStart( pos_start );
    }
  }
  return rc;
}
  
BOOL ON_BinaryArchive::BeginRead3dmChunk( unsigned int* typecode, int* value )
{
  int t = 0, v = 0;
  m_bDoChunkCRC = FALSE; // no CRC on chunk headers because length is written twice
  m_3dm_v1_suppress_error_message |= 0x0001; // disable v1 ReadByte() error message at EOF
  BOOL rc = ReadInt( &t );
  m_3dm_v1_suppress_error_message &= 0xFFFE; // enable v1 ReadByte() error message at EOF
  if (rc) {
    if ( t == TCODE_ENDOFFILE ) {
      // either this chunk is a bona fide end of file mark or it's "goo" that
      // Rhino 1.0 or the pre-February 2000 Rhino 1.1 saved and wrote.
      int sizeof_file = 0;
      if ( rc )
        rc = ReadInt( &v );
      if ( rc && v >= 4 ) {
        rc = ReadInt( &sizeof_file );
        if ( v > 4 ) {
          rc = SeekFromCurrentPosition( v - 4 ); // skip over additional end mark info added after this
                         // code was written
        }
        if ( rc ) {
          int current_pos = CurrentPosition(); 
          if ( !AtEnd() ) {
            // Rhino v1 reads chunks with unknown typecodes as a block of "goo"
            // and then saves them back into file.  This way new objects can
            // pass through old versions.  When this happens to a eof marker,
            // we get TCODE_ENDOFFILE chunks in the middle of a file.
            // This can only happen in m_3dm_version = 1 files.
            t = TCODE_ENDOFFILE_GOO;
          }
          else if ( m_3dm_version != 1 ) {
            // check that current position matches saved file size
            if ( current_pos != sizeof_file ) {
              ON_ERROR("ON_BinaryArchive::BeginRead3dmChunk() - Rogue eof marker in v2 file.\n");
            }
          }
          rc = SeekFromCurrentPosition( -v );
          if ( rc )
            rc = PushChunk( t, v );
        }
      }
      else {
        ON_ERROR( "ON_BinaryArchive::BeginRead3dmChunk() - file is damaged." );
        rc = FALSE;
        t = 0; // ?? file is trashed ??
      }
    }
    else {
      if ( rc )
        rc = ReadInt( &v );
      if ( rc )
        rc = PushChunk( t, v );
    }
  }
  if ( typecode )
    *typecode = t;
  if ( value )
    *value = v;
  return rc;
}

BOOL ON_BinaryArchive::EndRead3dmChunk()
{
  //int length = 0;
  BOOL rc = FALSE;
  ON_3DM_CHUNK* c = m_chunk.Last();
  if ( c ) {
    const int end_offset = (c->m_do_length) ? c->m_offset + c->m_value : c->m_offset;
    int file_offset = CurrentPosition();

    if ( c->m_do_length ) {
      if ( c->m_do_crc16 && file_offset+2 == end_offset) {
        // read 16 bit CRC
        unsigned char two_crc_bytes[2] = {0,0};
        rc = ReadByte( 2, two_crc_bytes );
        if (rc) {
          file_offset+=2;
          if (c->m_crc16) {
            m_bad_CRC_count++;
            ON_ERROR("ON_BinaryArchive::EndRead3dmChunk: CRC16 error.");
          }
        }
      }
      else if ( c->m_do_crc32 && file_offset+4 == end_offset) {
        // read 32 bit CRC
        unsigned long crc1 = c->m_crc32;
        unsigned long crc0;
        rc = ReadLong( &crc0 );
        if (rc) {
          file_offset+=4;
          if (crc0 != crc1) {
            m_bad_CRC_count++;
            ON_ERROR("ON_BinaryArchive::EndRead3dmChunk: CRC32 error.");
          }
        }
      }
      else {
        rc = TRUE;
      }
    }
    else {
      rc = TRUE;
    }

    // check length and seek to end of chunk if things are amiss
    if ( file_offset < c->m_offset ) {
      ON_ERROR("ON_BinaryArchive::EndRead3dmChunk: current position before start of current chunk.");
      if ( !SeekFromStart( end_offset ) )
        rc = FALSE;
    }
    else if ( file_offset > end_offset ) {
      ON_ERROR("ON_BinaryArchive::EndRead3dmChunk: current position after end of current chunk.");
      if ( !SeekFromStart( end_offset ) )
        rc = FALSE;
    }
    else if ( file_offset != end_offset ) {
      // partially read chunk - happens when chunks are skipped or old code
      // reads a new minor version of a chunk whnich has added information.
      if ( file_offset != c->m_offset ) {
        if ( m_3dm_version != 1 
             || (m_3dm_v1_suppress_error_message&0x02) == 0 ) {
          // when reading v1 files, there are some situations where
          // it is reasonable to attempt to read 4 bytes at the end
          // of a file.  The above test prevents making a call
          // to ON_WARNING() in these situations.
          ON_WARNING("ON_BinaryArchive::EndRead3dmChunk: partially read chunk - skipping bytes at end of current chunk.");
        }
      }
      if ( !SeekFromCurrentPosition( end_offset - file_offset ) )
        rc = FALSE;
    }

    m_chunk.Remove();
    c = m_chunk.Last();
    m_bDoChunkCRC = (c && (c->m_do_crc16 || c->m_do_crc32));
  }
  return rc;
}

BOOL ON_BinaryArchive::Read3dmGoo( ON_3dmGoo& goo )
{
  BOOL rc = FALSE;
  if (goo.m_goo) {
    onfree(goo.m_goo);
    goo.m_goo = 0;
  }
  goo.m_typecode = 0;
  goo.m_value = 0;
  const ON_3DM_CHUNK* c = m_chunk.Last();
  if (c) {
    goo.m_typecode = c->m_typecode;
    goo.m_value = c->m_value;
    if ( 0 == (goo.m_typecode & TCODE_SHORT) && goo.m_value > 0 ) {
      m_bDoChunkCRC = FALSE; // because goo may contain subchunks
      goo.m_goo = (unsigned char*)onmalloc( goo.m_value );
      rc = ReadByte( goo.m_value, goo.m_goo );
    }
  }
  return rc;
}

BOOL ON_BinaryArchive::PushChunk( unsigned int typecode, int value )
{
  ON_3DM_CHUNK c;
  memset(&c,0,sizeof(c));
  c.m_typecode = typecode;
  c.m_value    = value;

  // | and & are BITOPS - do NOT change to || and &&
  if ( 0 == ( TCODE_SHORT & typecode ) ) {
    if ( m_3dm_version == 1 && 0 != (TCODE_LEGACY_GEOMETRY & typecode) ) {
      // these legacy typecodes have 16 bit CRCs
      c.m_do_crc16 = 1;
      c.m_crc16 = 1;
    }
    else {
      // some other legacy typecodes that have 16 bit CRCs
      switch(typecode) {

      case TCODE_SUMMARY:
        if ( m_3dm_version == 1 ) {
          c.m_do_crc16 = 1;
          c.m_crc16 = 1;
        }
        break;

      case TCODE_OPENNURBS_OBJECT | TCODE_CRC | 0x7FFD:
        if ( m_3dm_version == 1 ) {
          // 1.1 uuid
          c.m_do_crc16 = 1;
          c.m_crc16 = 1;
        }
        else {
          // 2.0 uuid
          // 32 bit CRC
          c.m_do_crc32 = 1;
          c.m_crc32 = 0;
        }
        break;

      default:
        if ( m_3dm_version != 1 && 0 != (TCODE_CRC & typecode) ) {
          // 32 bit CRC
          c.m_do_crc32 = 1;
          c.m_crc32 = 0;
        }
        break;

      }
    }
    c.m_do_length = 1;
  }
  c.m_offset = CurrentPosition();
  m_bDoChunkCRC = c.m_do_crc16 || c.m_do_crc32;
  if ( m_chunk.Count() == 0 )
    m_chunk.Reserve(128);
  m_chunk.Append( c );
  return TRUE;
}

BOOL ON_BinaryArchive::EnableSave3dmRenderMeshes( BOOL  b /* default = TRUE */ )
{
  BOOL oldb = m_bSaveRenderMeshes;
  m_bSaveRenderMeshes = b;
  return oldb;  
}

BOOL ON_BinaryArchive::Save3dmRenderMeshes() const
{
  return m_bSaveRenderMeshes;
}

BOOL ON_BinaryArchive::EnableSave3dmAnalysisMeshes( BOOL  b /* default = TRUE */ )
{
  BOOL oldb = m_bSaveAnalysisMeshes;
  m_bSaveAnalysisMeshes = b;
  return oldb;  
}

BOOL ON_BinaryArchive::Save3dmAnalysisMeshes() const
{
  return m_bSaveAnalysisMeshes;
}

BOOL ON_BinaryArchive::Write3dmStartSection( int version, const char* sInformation )
{
  m_bad_CRC_count = 0;
  m_3dm_version = 0;
  m_3dm_opennurbs_version = ON::Version();

  char sVersion[64];
  memset( sVersion, 0, sizeof(sVersion) );
  if ( version < 1 )
    version = 2;
  // TEMPORARY 2 = X
  //if ( version == 2 ) 
  //{
  //  sprintf(sVersion,"3D Geometry File Format        X",version);
  //}
  //else 
  {
    sprintf(sVersion,"3D Geometry File Format %8d",version);
  }
  BOOL rc = WriteByte( 32, sVersion );
  if ( rc )
    rc = BeginWrite3dmChunk( TCODE_COMMENTBLOCK, 0 );
  if ( rc ) {
    if ( sInformation && sInformation[0] ) {
      rc = WriteByte( strlen(sInformation), sInformation );
    }
    if ( rc ) {
      // write information that helps determine what code wrote the 3dm file
      char s[2048];
      int s_len = 0;
      memset( s, 0, sizeof(s) );
      sprintf(s," 3DM I/O processor: OpenNURBS toolkit version %d",ON::Version());
      //strncat(s,sVERSION_FILENAME,sizeof(s) - 100 - strlen(s));
      //strcat(s," ");
      //strncat(s,sVERSION_NUMBER,sizeof(s) - 100 - strlen(s));
      strcat(s," (compiled on ");
      strcat(s,__DATE__);
      strcat(s,")\n");
      s_len = strlen(s);
      s[s_len++] = 26; // ^Z
      s[s_len++] = 0;
      rc = WriteByte( s_len, s );
    }
    if ( !EndWrite3dmChunk() ) // need to call EndWrite3dmChunk() even if a WriteByte() has failed
      rc = FALSE;
  }
  m_3dm_version = version;
  return rc;
}

BOOL ON_BinaryArchive::Read3dmStartSection( int* version, ON_String& s )
{
  // The first 24 buts of a 3dm file must be "3D Geometry File Format "
  // The next 8 bytes must be a right justified ASCII integer >= 1.  For
  // example, prior to March 2000, all 3DM files were version 1 files and
  // they began with "3D Geometry File Format        1".  At the time of
  // this writing (March 2000) there are version 1 and version 2 files.
  //
  // The next section must contain a long chunk with typecode 1.
  m_bad_CRC_count = 0;
  m_3dm_version = 0;

  // In Read3dmProperties, m_3dm_opennurbs_version will be set to 
  // the version of OpenNURBS that was used to write this archive.
  m_3dm_opennurbs_version = 0;

  unsigned int typecode = 0;
  int length = -1;
  int ver = m_3dm_version;
  if ( version )
    *version = m_3dm_version;
  s.Destroy();
  char s3d[33];
  memset( s3d, 0, sizeof(s3d) );
  BOOL rc = ReadByte( 32, s3d );
  if ( rc ) {
    if ( strncmp( s3d, "3D Geometry File Format ", 24) ) {
      // it's not a "pure" .3DM file 
      // - see if we have a .3DM file with MS OLE-goo at the start
      rc =  FALSE;
      int n, j;
      for ( n = 0; n < 4096 && !rc; n++ ) {
        for ( j = 0; j < 31; j++ )
          s3d[j] = s3d[j+1];
        if ( !ReadByte( 1, &s3d[31]) ) 
          break;
        if ( !strncmp( s3d, "3D Geometry File Format ", 24) ) {
          rc = TRUE;
          break;
        }
      }
    }
    if (rc) {
      // get version
      //char* sVersion = s3d+24;
      // skip leading spaces
      int i = 24;
      while (i < 32 && s3d[i] == ' ')
        i++;
      while (i < 32 && rc) {
        // TEMPORARY 2 = X
        if ( i == 31 && s3d[i] == 'X' ) {
          s3d[i] = '2';
        }

        if ( s3d[i] < '0' || s3d[i] > '9' ) {
          rc = FALSE; // it's not a valid .3DM file version
          break;
        }
        ver = ver*10 + ((int)(s3d[i]-'0'));
        i++;
      }
      if ( rc ) {
        m_3dm_version = ver;
        if ( version )
          *version = ver;
      }
    }
    if ( rc ) {
      rc = BeginRead3dmChunk( &typecode, &length );
      if ( rc ) {
        if ( typecode != 1 ) {
          rc = FALSE; // it's not a .3DM file
        }
        else if ( length > 0 ) {
          s.ReserveArray( length+1 );
          s.SetLength( length );
          s[length] = 0;
          ReadByte( length, s.Array() );
          while ( length > 0 && s[length-1] == 0 || s[length-1] == 26 ) {
            s[length-1] = 0;
            length--;
          }
          s.SetLength(length);
        }
      }
      if ( !EndRead3dmChunk() )
        rc = FALSE;
      if ( rc && m_3dm_version == 1 ) {
        // In March 2001, we got reports of files with V1 headers and
        // a V2 bodies.  We haven't been able to track down the application
        // that is creating these damaged files, but we can detect them
        // and read them because they all have a TCODE_PROPERTIES_TABLE
        // chunk right after the start comments chunk and no valid V1
        // file has a chunk with a TCODE_PROPERTIES_TABLE tcode.
        //
        // Rhino 1.1 version 31-May-2000 reads 2.0 files as goo.  If a user
        // saves this file for some reason (no instances have been reported)
        // the resulting file has a V1 header, V1 fluff, and a V2 body.
        // This code will cause opennurbs to read the V2 body.
        // a file that is different from those describe This code
        // detects these files.
        {

          const int pos1 = CurrentPosition();
          //int v1_fluff_chunk_count = 0;
          BOOL bCheckChunks = TRUE;

          //////////
          while(bCheckChunks) {
            if ( !PeekAt3dmChunkType(&typecode,&length) )
              break;
            switch(typecode)
            {
              case TCODE_SUMMARY:
              case TCODE_BITMAPPREVIEW:
              case TCODE_UNIT_AND_TOLERANCES:
              case TCODE_VIEWPORT:
              case TCODE_LAYER:
              case TCODE_RENDERMESHPARAMS:
              case TCODE_CURRENTLAYER:
              case TCODE_ANNOTATION_SETTINGS:
              case TCODE_NOTES:
              case TCODE_NAMED_CPLANE:
              case TCODE_NAMED_VIEW:
                // skip potential v1 fluff
                bCheckChunks = BeginRead3dmChunk( &typecode, &length );
                if ( bCheckChunks )
                  bCheckChunks = EndRead3dmChunk();
                break;

              //case TCODE_PROPERTIES_TABLE:
              //case TCODE_SETTINGS_TABLE:
              //case TCODE_OBJECT_TABLE:
              //case TCODE_BITMAP_TABLE:
              //case TCODE_LAYER_TABLE:
              //case TCODE_GROUP_TABLE:
              //case TCODE_LIGHT_TABLE:
              //case TCODE_MATERIAL_TABLE:
              //case TCODE_USER_TABLE:
              default:
                if ( TCODE_TABLE == (typecode & 0xFFFF0000) ) {
                  // Found a V2 table which has to be V1 goo
                  ON_WARNING("ON_BinaryArchive::Read3dmStartSection(): Archive has V1 header and V2 body. Continuing to read V2 body.");
                  m_3dm_version = 2;
                  if ( version )
                    *version = 2;
                }
                bCheckChunks = FALSE;
                break;
            }
          }

          if ( m_3dm_version == 1 ) {
            // move archive pointer back to 
            int pos2 = CurrentPosition();
            if ( pos2 > pos1 ) {
              SeekFromCurrentPosition( pos1 - pos2 );
            }
          }
        }
      }
    }
  }
  return rc;
}

BOOL ON_BinaryArchive::Write3dmProperties(
      const ON_3dmProperties& prop
      )
{
  BOOL rc = FALSE;
  if ( m_3dm_version == 1 ) {
    ON_String s;

    if ( rc && prop.m_RevisionHistory.IsValid() ) {
      rc = BeginWrite3dmChunk(TCODE_SUMMARY,0);
      if (rc) {
        // version 1 revision history chunk
        s = prop.m_RevisionHistory.m_sCreatedBy;
        if (rc) rc = WriteString(s);
        if (rc) rc = WriteTime( prop.m_RevisionHistory.m_create_time );
        if (rc) rc = WriteInt(0);
        s = prop.m_RevisionHistory.m_sLastEditedBy;
        if (rc) rc = WriteString(s);
        if (rc) rc = WriteTime( prop.m_RevisionHistory.m_last_edit_time );
        if (rc) rc = WriteInt(0);
        if (rc) rc = WriteInt( prop.m_RevisionHistory.m_revision_count );
        if ( !EndWrite3dmChunk() ) // writes 16 bit crc
          rc = FALSE;
      }
    }

    if ( rc && prop.m_Notes.IsValid() ) {
      rc = BeginWrite3dmChunk(TCODE_NOTES,0);
      // version 1 notes chunk
      if ( rc ) {
        if ( rc ) rc = WriteInt( prop.m_Notes.m_bVisible );
        if ( rc ) rc = WriteInt( prop.m_Notes.m_window_left );
        if ( rc ) rc = WriteInt( prop.m_Notes.m_window_top );
        if ( rc ) rc = WriteInt( prop.m_Notes.m_window_right );
        if ( rc ) rc = WriteInt( prop.m_Notes.m_window_bottom );
        s = prop.m_Notes.m_notes;
        if ( rc ) rc = WriteString( s );
        if ( !EndWrite3dmChunk() )
          rc = FALSE;
      }
    }

    if ( rc && prop.m_PreviewImage.IsValid() ) {
      rc = BeginWrite3dmChunk(TCODE_BITMAPPREVIEW,0);
      if (rc) {
        // version 1 preview image chunk
        prop.m_PreviewImage.Write(*this);
        if ( !EndWrite3dmChunk() )
          rc = FALSE;
      }
    }
  }
  else {
    // version 2+ file properties chunk
    rc = BeginWrite3dmChunk(TCODE_PROPERTIES_TABLE,0);
    if ( rc ) {
      rc = prop.Write( *this );
      if ( !EndWrite3dmChunk() )
        rc = FALSE;
    }
  }
  return rc;
}

int on_strnicmp(const char * s1, const char * s2, int n)
{
#if defined(ON_OS_WINDOWS)
  //return stricmp(s1,s2,n);
  return _strnicmp(s1,s2,n);
#else
  return strncasecmp(s1,s2,n);
#endif
}

BOOL ON_BinaryArchive::Read3dmProperties( ON_3dmProperties& prop )
{
  // In ON_3dmProperties::Read(), m_3dm_opennurbs_version will be 
  // set to the version of OpenNURBS that was used to write this archive.
  // If the file was written with by a pre 200012210 version of OpenNURBS,
  // then m_3dm_opennurbs_version will be zero.
  m_3dm_opennurbs_version = 0;

  prop.Default();

  BOOL rc = TRUE;

  // we need these when reading version 1 files
  const int pos0 = CurrentPosition();
  BOOL bHaveRevisionHistory = FALSE;
  BOOL bHaveNotes = FALSE;
  BOOL bHavePreviewImage = FALSE;
  BOOL bDone = FALSE;
  BOOL bRewindFilePointer = FALSE;

  unsigned int tcode;
  int value;
  int version = 0;

  if ( m_3dm_version != 1 ) {
    for(;;) {
      rc = BeginRead3dmChunk( &tcode, &value );
      if ( !rc ) {
        bRewindFilePointer = TRUE;
        break;
      }

      if ( tcode == TCODE_PROPERTIES_TABLE ) {
        rc = prop.Read(*this);
      }
      else {
        bRewindFilePointer = TRUE;
      }

      if ( !EndRead3dmChunk() ) {
        rc = FALSE;
        bRewindFilePointer = TRUE;
      }
      if ( tcode == TCODE_PROPERTIES_TABLE || !rc )
        break;
    }
  }
  else {
    // version 1 file
    rc = SeekFromStart(32);
    bRewindFilePointer = TRUE;
    for(;;) {
      rc = BeginRead3dmChunk( &tcode, &value );
      if ( !rc ) {
        rc = TRUE; // assume we are at the end of the file
        bRewindFilePointer = TRUE;
        break;
      }

      switch ( tcode ) {

      case 1: // comments section has application name
        if ( value > 1 ) {
          int i;
          char* name = 0;
          ON_String s;
          s.ReserveArray( value+1 );
          s.SetLength( value );
          s[value] = 0;
          ReadByte( value, s.Array() );
          while ( value > 0 && s[value-1] == 0 || s[value-1] == 26 ) {
            s[value-1] = 0;
            value--;
          }
          s.SetLength(value);
          name = s.Array();
          if ( name ) {
            while(*name) {
              if ( !on_strnicmp(name,"Interface:",10) ) {
                name += 10;
                break;
              }
              name++;
            }
            while(*name && *name <= 32 )
              name++;
            for ( i = 0; name[i] ; i++ ) {
              if ( name[i] == '(' ) {
                name[i] = 0;
                while ( i > 0 && (name[i] <= 32 || name[i] == '-') ) {
                  name[i] = 0;
                  i--;
                }
                break;
              }
            }
            if ( *name ) {
              char* details = 0;
              if ( !on_strnicmp(name,"Rhinoceros",10) ) {
                prop.m_Application.m_application_URL = "http://www.rhino3d.com";
                details = name+10;
                while ( *details && *details <= 32 )
                  details++;
                while ( (*details >= '0' && *details <= '9') || *details == '.' )
                  details++;
                if ( *details && *details <= 32 ) {
                  *details = 0;
                  details++;
                  while ( *details && (*details <= 32 ||*details == '-')) {
                    details++;
                  }
                }
              }
              if (*name)
                prop.m_Application.m_application_name = name;
              if (details && *details)
                prop.m_Application.m_application_details = details;
            }
          }
        }
        break;

      case TCODE_SUMMARY: 
        // version 1 revision history chunk (has 16 bit CRC)
        version = 1;
        bHaveRevisionHistory = TRUE;
        {
          int slength = 0;
          char* s = 0;
          if (rc) rc = ReadInt(&slength);
          if (rc && slength > 0 ) {
            s = (char*)onmalloc((slength+1)*sizeof(*s));
            memset( s, 0, (slength+1)*sizeof(*s) );
            if (rc) rc = ReadChar( slength, s );
            if ( rc )
              prop.m_RevisionHistory.m_sCreatedBy = s;
            onfree(s);
            slength = 0;
            s = 0;
          }
          if (rc) rc = ReadTime( prop.m_RevisionHistory.m_create_time );
          if (rc) rc = ReadInt(&value); // 0 in 1.x files
          if (rc) rc = ReadInt(&slength);
          if ( rc && slength > 0 ) {
            s = (char*)onmalloc((slength+1)*sizeof(*s));
            memset( s, 0, (slength+1)*sizeof(*s) );
            if (rc) rc = ReadChar( slength, s );
            if ( rc )
              prop.m_RevisionHistory.m_sLastEditedBy = s;
            onfree(s);
            slength = 0;
            s = 0;
          }
          if (rc) rc = ReadTime( prop.m_RevisionHistory.m_last_edit_time );
          if (rc) rc = ReadInt(&value); // 0 in 1.x files
          if (rc) rc = ReadInt( &prop.m_RevisionHistory.m_revision_count );
        }
        break;

      case TCODE_NOTES: 
        // version 1 notes chunk
        version = 1;
        bHaveNotes = TRUE;
        {
          int slength;
          char* s = 0;
          if ( rc ) rc = ReadInt( &prop.m_Notes.m_bVisible );
          if ( rc ) rc = ReadInt( &prop.m_Notes.m_window_left );
          if ( rc ) rc = ReadInt( &prop.m_Notes.m_window_top );
          if ( rc ) rc = ReadInt( &prop.m_Notes.m_window_right );
          if ( rc ) rc = ReadInt( &prop.m_Notes.m_window_bottom );
          if ( rc ) rc = ReadInt( &slength );
          if ( rc && slength > 0 ) {
            s = (char*)onmalloc( (slength+1)*sizeof(*s) );
            memset( s, 0, (slength+1)*sizeof(*s) );
            if ( rc ) rc = ReadChar( slength, s );
            if ( rc ) {
              prop.m_Notes.m_notes = s;
            }
            onfree(s);
          }
        }
        break;

      case TCODE_BITMAPPREVIEW: 
        // version 1 preview image chunk
        version = 1;
        rc = prop.m_PreviewImage.Read(*this);
        bHavePreviewImage = rc;
        break;

      case TCODE_CURRENTLAYER:
      case TCODE_LAYER:
        // version 1 layer and current layer chunks always came after notes/bitmap/summary
        bDone = TRUE;
        bRewindFilePointer = TRUE;
        break;

      default:
        // the call to EndRead3dmChunk() will skip over this chunk
        bRewindFilePointer = TRUE;
        break;
      }

      // this call to EndRead3dmChunk() skips any unread portions of the chunk
      if ( !EndRead3dmChunk() ) {
        rc = FALSE;
        bRewindFilePointer = TRUE;
      }

      if ( bHaveRevisionHistory && bHaveNotes && bHavePreviewImage )
        bDone = TRUE;

      if ( bDone || !rc )
        break;
    }
  }

  if ( bRewindFilePointer ) {
    // reposition file pointer
    const int pos1 = CurrentPosition();
    if ( pos0 != pos1 ) {
      SeekFromCurrentPosition( pos0-pos1 );
    }
  }

  return rc;
}

BOOL ON_BinaryArchive::Write3dmSettings(
      const ON_3dmSettings& settings
      )
{
  BOOL rc = FALSE;
  if ( m_3dm_version == 1 ) {
    // legacy v1 settings info is a bunch of unreleated top level chunks
    rc = settings.Write(*this);
  }
  else {
    // version 2+ file settings chunk
    rc = BeginWrite3dmChunk(TCODE_SETTINGS_TABLE,0);
    if ( rc ) {
      rc = settings.Write( *this );
      if ( !EndWrite3dmChunk() )
        rc = FALSE;
    }
  }
  return rc;
}

BOOL ON_BinaryArchive::Read3dmSettings( ON_3dmSettings& settings )
{
  BOOL rc = FALSE;
  unsigned int tcode;
  int value;

  if ( m_3dm_version == 1 ) {
    // read legacy v 1 info that is scattered around the file
    rc = settings.Read(*this);
  }
  else {
    rc = TRUE;
    while(rc) {
      rc = BeginRead3dmChunk( &tcode, &value );
      if ( !rc )
        break;
      if ( tcode == TCODE_SETTINGS_TABLE ) {
        // version 2 model properties
        rc = settings.Read(*this);
      }
      if ( !EndRead3dmChunk() ) {
        rc = FALSE;
        break;
      }
      if ( TCODE_SETTINGS_TABLE == tcode )
        break;
    }
  }

  return rc;
}

ON_BinaryArchive::table_type ON_BinaryArchive::TableTypeFromTypecode( unsigned int typecode )
{
  table_type tt = no_active_table;
  switch(typecode) 
  {
  case TCODE_PROPERTIES_TABLE: tt = properties_table; break;
  case TCODE_SETTINGS_TABLE:   tt = settings_table; break;
  case TCODE_BITMAP_TABLE:     tt = bitmap_table; break;
  case TCODE_MATERIAL_TABLE:   tt = material_table; break;
  case TCODE_LAYER_TABLE:      tt = layer_table; break;
  case TCODE_LIGHT_TABLE:      tt = light_table; break;
  case TCODE_OBJECT_TABLE:     tt = object_table; break;
  case TCODE_GROUP_TABLE:      tt = group_table; break;
  case TCODE_USER_TABLE:       tt = user_table; break;
  }
  return tt;
}

BOOL ON_BinaryArchive::BeginWrite3dmTable( unsigned int typecode )
{
  const table_type tt = TableTypeFromTypecode(typecode);
  if (tt == no_active_table) {
    ON_ERROR("ON_BinaryArchive::BeginWrite3dmTable() bad typecode");
    return FALSE;
  }
  if ( m_active_table != no_active_table ) {
    ON_ERROR("ON_BinaryArchive::BeginWrite3dmTable() m_active_table != no_active_table");
    return FALSE;
  }
  if ( m_chunk.Count() ) {
    ON_ERROR("ON_BinaryArchive::BeginWrite3dmTable() m_chunk.Count() > 0");
    return FALSE;
  }
  BOOL rc = BeginWrite3dmChunk(typecode,0);
  if (rc)
    m_active_table = tt;
  return rc;
}

BOOL ON_BinaryArchive::EndWrite3dmTable( unsigned int typecode )
{
  const table_type tt = TableTypeFromTypecode(typecode);
  if (tt == no_active_table) {
    ON_ERROR("ON_BinaryArchive::EndWrite3dmTable() bad typecode");
    return FALSE;
  }
  if ( m_active_table != tt ) {
    ON_ERROR("ON_BinaryArchive::EndWrite3dmTable() m_active_table != t");
    return FALSE;
  }
  if ( m_chunk.Count() != 1 ) {
    ON_ERROR("ON_BinaryArchive::EndWrite3dmTable() m_chunk.Count() != 1");
    return FALSE;
  }
  const ON_3DM_CHUNK* c = m_chunk.Last();
  if ( c->m_typecode != typecode ) {
    ON_ERROR("ON_BinaryArchive::EndWrite3dmTable() m_chunk.Last()->typecode != typecode");
    return FALSE;
  }
  BOOL rc = BeginWrite3dmChunk( TCODE_ENDOFTABLE, 0 );
  if (rc) {
    if (!EndWrite3dmChunk())
      rc = FALSE;
  }
  if (!EndWrite3dmChunk())
    rc = FALSE;
  Flush();
  m_active_table = no_active_table;
  return rc;
}

BOOL ON_BinaryArchive::BeginRead3dmTable( unsigned int typecode )
{
  BOOL rc = FALSE;
  const table_type tt = TableTypeFromTypecode(typecode);
  if (tt == no_active_table) {
    ON_ERROR("ON_BinaryArchive::BeginRead3dmTable() bad typecode");
    return FALSE;
  }
  if ( m_active_table != no_active_table ) {
    ON_ERROR("ON_BinaryArchive::BeginRead3dmTable() m_active_table != no_active_table");
    return FALSE;
  }
  if ( m_chunk.Count() ) {
    ON_ERROR("ON_BinaryArchive::BeginRead3dmTable() m_chunk.Count() > 0");
    return FALSE;
  }

  if ( m_3dm_version == 1 ) {
    // version 1 files had can have chunks in any order.  To read a "table",
    // you have to go through the entire list of chunks looking for those
    // that belong to a particular table.
    rc = SeekFromStart(32);
    m_active_table = tt;
  }
  else {
    BOOL bReadTable = TRUE;
    unsigned int tcode = !typecode;
    int value = 0;
    rc = PeekAt3dmChunkType( &tcode, &value );
    if ( rc ) {
      if ( tcode != typecode ) {
        if ( typecode == TCODE_USER_TABLE ) {
          // it's ok to not have user tables
          rc = FALSE;
          bReadTable = FALSE;
        }
        else if ( typecode == TCODE_GROUP_TABLE && m_3dm_opennurbs_version < 200012210 ) {
          // 3DM archives written before version 200012210 and before do not have group tables
          rc = TRUE;
          m_active_table = tt;
          bReadTable = FALSE;
        }
        else {
          // A required table is not at the current position in the archive
          // see if we can find it someplace else in the archive.  This can
          // happen when old code encounters a table that was added later.
          ON_ERROR("ON_BinaryArchive::BeginRead3dmTable() - current file position not at start of table - searching");
          rc = Seek3dmChunkFromStart( typecode );
        }
      }
      if ( rc && bReadTable ) {
        tcode = !typecode;
        value = 0;
        rc = BeginRead3dmChunk( &tcode, &value );
        if ( rc && tcode != typecode ) {
          ON_ERROR("ON_BinaryArchive::BeginRead3dmTable() - corrupt table - skipping");
          rc = FALSE;
          EndRead3dmChunk();
        }
        else if (rc) {
          m_active_table = tt;
        }
      }
    }
  }

  return rc;
}

BOOL ON_BinaryArchive::EndRead3dmTable( unsigned int typecode )
{
  BOOL rc = FALSE;
  const table_type tt = TableTypeFromTypecode(typecode);
  if (tt == no_active_table) {
    ON_ERROR("ON_BinaryArchive::EndRead3dmTable() bad typecode");
    return FALSE;
  }
  if ( m_active_table != tt ) {
    ON_ERROR("ON_BinaryArchive::EndRead3dmTable() m_active_table != t");
    return FALSE;
  }
  if ( m_3dm_version == 1 ) {
    if ( m_chunk.Count() != 0 ) {
      ON_ERROR("ON_BinaryArchive::EndRead3dmTable() v1 file m_chunk.Count() != 0");
      return FALSE;
    }
    rc = TRUE;
  }
  else {
    if ( m_active_table == group_table && m_3dm_opennurbs_version < 200012210 ) {
      // 3DM archives written before version 200012210 and before do not have group tables
      rc = TRUE;
    }
    else {
      if ( m_chunk.Count() != 1 ) {
        ON_ERROR("ON_BinaryArchive::EndRead3dmTable() v2 file m_chunk.Count() != 1");
        return FALSE;
      }
      const ON_3DM_CHUNK* c = m_chunk.Last();
      if ( c->m_typecode != typecode ) {
        ON_ERROR("ON_BinaryArchive::EndRead3dmTable() m_chunk.Last()->typecode != typecode");
        return FALSE;
      }
      rc = EndRead3dmChunk();
    }
  }
  m_active_table = no_active_table;
  return rc;
}

BOOL ON_BinaryArchive::BeginWrite3dmBitmapTable()
{
  return BeginWrite3dmTable( TCODE_BITMAP_TABLE );
}

BOOL ON_BinaryArchive::EndWrite3dmBitmapTable()
{
  return EndWrite3dmTable( TCODE_BITMAP_TABLE );
}

BOOL ON_BinaryArchive::Write3dmBitmap( const ON_Bitmap& bitmap )
{
  BOOL rc = FALSE;
  if ( m_3dm_version != 1 ) {
    const ON_3DM_CHUNK* c = m_chunk.Last();
    if ( c && c->m_typecode == TCODE_BITMAP_TABLE ) {
      rc = BeginWrite3dmChunk( TCODE_BITMAP_RECORD, 0 );
      if ( rc ) {
        rc = WriteObject( bitmap );
        if ( !EndWrite3dmChunk() )
          rc = FALSE;
      }
    }
    else {
      ON_ERROR("ON_BinaryArchive::Write3dmBitmap() must be called in BeginWrite3dmBitmapTable() block");
      rc = FALSE;
    }
  }
  return rc;
}

BOOL ON_BinaryArchive::BeginRead3dmBitmapTable()
{
  return BeginRead3dmTable( TCODE_BITMAP_TABLE );
}

BOOL ON_BinaryArchive::EndRead3dmBitmapTable()
{
  return EndRead3dmTable( TCODE_BITMAP_TABLE );
}

int ON_BinaryArchive::Read3dmBitmap(  // returns 0 at end of bitmap table
                                      //         1 bitmap successfully read
          ON_Bitmap** ppBitmap // bitmap returned here
          )
{
  if ( ppBitmap )
    *ppBitmap = 0;
  ON_Bitmap* bitmap = 0;
  int rc = 0;
  if ( m_3dm_version != 1 ) {
    unsigned int tcode = 0;
    int value;
    if ( BeginRead3dmChunk( &tcode, &value ) ) {
      if ( tcode == TCODE_BITMAP_RECORD ) {
        ON_Object* p = 0;
        if ( ReadObject( &p ) ) {
          bitmap = ON_Bitmap::Cast(p);
          if ( !bitmap )
            delete p;
          else
            rc = 1;
        }
        if (!bitmap) {
          ON_ERROR("ON_BinaryArchive::Read3dmBitmap() - corrupt bitmap table");
        }
        if ( ppBitmap )
          *ppBitmap = bitmap;
        else if ( bitmap )
          delete bitmap;          
      }
      else if ( tcode != TCODE_ENDOFTABLE ) {
        ON_ERROR("ON_BinaryArchive::Read3dmBitmap() - corrupt bitmap table");
      }
      EndRead3dmChunk();
    }
  }

  return rc;
}


BOOL ON_BinaryArchive::BeginWrite3dmLayerTable()
{
  BOOL rc = FALSE;
  if ( m_3dm_version != 1 ) {
    rc = BeginWrite3dmTable( TCODE_LAYER_TABLE );
  }
  else {
    if ( m_chunk.Count() ) {
      ON_ERROR("ON_BinaryArchive::BeginWrite3dmLayerTable() - chunk stack should be empty");
      return FALSE;
    }
    if ( m_active_table != no_active_table ) {
      ON_ERROR("ON_BinaryArchive::BeginWrite3dmLayerTable() - m_active_table != no_active_table");
    }
    m_active_table = layer_table;
    rc = TRUE;
  }

  return rc;
}

BOOL ON_BinaryArchive::Write3dmLayer( const ON_Layer&  layer )
{
  BOOL rc = FALSE;
  if ( m_active_table != layer_table ) {
    ON_ERROR("ON_BinaryArchive::Write3dmLayer() - m_active_table != layer_table");
  }

  if ( m_3dm_version == 1 ) {
    // legacy version 1 layer information is in a top level TCODE_LAYER chunk
    if ( m_chunk.Count() ) {
      ON_ERROR("ON_BinaryArchive::Write3dmLayer() - version 1 - chunk stack should be empty");
      return FALSE;
    }
    ON_String s = layer.LayerName();
    if ( !s.IsEmpty() ) {
      rc = BeginWrite3dmChunk( TCODE_LAYER, 0 );

      // layer name
      if (rc) {
        rc = BeginWrite3dmChunk( TCODE_LAYERNAME, 0 );
        if(rc) rc = WriteString(s);
        if (!EndWrite3dmChunk())
          rc = FALSE;
      }

      // layer color
      if (rc) {
        rc = BeginWrite3dmChunk( TCODE_RGB, layer.Color() );
        if (!EndWrite3dmChunk())
          rc = FALSE;
      }

      // layer mode normal=0/hidden=1/locked=2
      if (rc) {
        int mode = 0;
        switch( layer.Mode() ) {
        case ON::hidden_layer:
          mode = 1;
          break;
        case ON::locked_layer:
          mode = 2;
          break;
        default:
          mode = 0;
          break;
        }
        rc = BeginWrite3dmChunk( TCODE_LAYERSTATE, mode );
        if (!EndWrite3dmChunk())
          rc = FALSE;
      }

      if ( !BeginWrite3dmChunk( TCODE_ENDOFTABLE, 0 ) )
        rc = FALSE;
      if ( !EndWrite3dmChunk() )
        rc = FALSE;

      if (!EndWrite3dmChunk()) // end of TCODE_LAYER chunk
        rc = FALSE;
    }
  }
  else {
    // version 2+
    const ON_3DM_CHUNK* c = m_chunk.Last();
    if ( c && c->m_typecode == TCODE_LAYER_TABLE ) {
      rc = BeginWrite3dmChunk( TCODE_LAYER_RECORD, 0 );
      if ( rc ) {
        rc = WriteObject( layer );
        if ( !EndWrite3dmChunk() )
          rc = FALSE;
      }
    }
    else {
      ON_ERROR("ON_BinaryArchive::Write3dmLayer() must be called in BeginWrite3dmLayerTable(2) block");
      rc = FALSE;
    }
  }

  return rc;
}

BOOL ON_BinaryArchive::EndWrite3dmLayerTable()
{
  BOOL rc = FALSE;
  if ( m_3dm_version == 1 ) {
    if ( m_active_table != layer_table ) {
      ON_ERROR("ON_BinaryArchive::EndWrite3dmLayerTable() - m_active_table != layer_table");
    }
    rc = TRUE;
    m_active_table = no_active_table;
  }
  else {
    rc = EndWrite3dmTable( TCODE_LAYER_TABLE );
  }
  return rc;
}

BOOL ON_BinaryArchive::BeginRead3dmLayerTable()
{
  BOOL rc = FALSE;
  m_3dm_v1_layer_index = 0;
  rc = BeginRead3dmTable( TCODE_LAYER_TABLE );
  if ( rc && m_3dm_version == 1 ) {
    rc = Seek3dmChunkFromStart( TCODE_LAYER );
    rc = TRUE; // there are 1.0 files written by the old IO toolkit that have no layers
  }
  return rc;
}

int ON_BinaryArchive::Read3dmV1Layer( ON_Layer*& layer )
{
  int rc = 0;
  unsigned int tcode;
  int value;
  for (;;) {
    tcode = !TCODE_LAYER;
    if (!BeginRead3dmChunk(&tcode,&value))
      break; // assume we are at the end of the file
    if ( tcode == TCODE_LAYER ) {
      layer = new ON_Layer();
      layer->SetLayerIndex(m_3dm_v1_layer_index++);
      rc = 1;
      break;
    }
    if (!EndRead3dmChunk())
       break;
  }
  if ( layer ) {
    rc = FALSE;
    for (;;) {
      if (!BeginRead3dmChunk(&tcode,&value))
        break;
      switch(tcode) {
      case TCODE_LAYERNAME:
        {
          ON_String s;
          value = 0;
          ReadInt(&value);
          s.SetLength(value);
          if ( ReadByte( s.Length(), s.Array() ) ) {
            layer->SetLayerName(s);
          }
        }
        break;
      case TCODE_RGB:
        layer->SetColor( ON_Color(value) );
        break;
      case TCODE_LAYERSTATE:
        switch (value) 
        {
        case 1: // hidden
          layer->SetMode( ON::hidden_layer );
          break;
        case 2: // locked
          layer->SetMode( ON::locked_layer );
          break;
        default: // normal
          layer->SetMode( ON::normal_layer );
          break;
        }
        break;
      }
      if (!EndRead3dmChunk())
         break;
      if ( TCODE_ENDOFTABLE == tcode ) {
        rc = TRUE;
        break;
      }
    }
    if ( !EndRead3dmChunk() ) // end of TCODE_LAYER chunk
      rc = FALSE;
  }
  if ( !rc && layer ) {
    delete layer;
    layer = 0;
  }
  return rc;
}

int ON_BinaryArchive::Read3dmLayer( ON_Layer** ppLayer )
{
  if ( !ppLayer )
    return 0;
  *ppLayer = 0;
  if ( m_active_table != layer_table ) {
    ON_ERROR("ON_BinaryArchive::BeginRead3dmLayerTable() - m_active_table != no_active_table");
  }
  unsigned int tcode;
  int value;
  ON_Layer* layer = NULL;
  // returns 0 at end of layer table
  if ( m_3dm_version == 1 ) {
    Read3dmV1Layer(layer);
  }
  else {
    // version 2+
    tcode = 0;
    if ( BeginRead3dmChunk( &tcode, &value ) ) {
      if ( tcode == TCODE_LAYER_RECORD ) {
        ON_Object* p = 0;
        if ( ReadObject( &p ) ) {
          layer = ON_Layer::Cast(p);
          if ( !layer )
            delete p;
        }
        if (!layer) {
          ON_ERROR("ON_BinaryArchive::Read3dmLayer() - corrupt layer table");
        }
      }
      else if ( tcode != TCODE_ENDOFTABLE ) {
        ON_ERROR("ON_BinaryArchive::Read3dmLayer() - corrupt layer table");
      }
      EndRead3dmChunk();
    }
  }
  *ppLayer = layer;
  return (layer) ? 1 : 0;
}

BOOL ON_BinaryArchive::EndRead3dmLayerTable()
{
  BOOL rc = FALSE;
  if ( m_3dm_version == 1 ) {
    if ( m_active_table != layer_table ) {
      ON_ERROR("ON_BinaryArchive::EndRead3dmLayerTable() - m_active_table != no_active_table");
      rc = FALSE;
    }
    else if ( m_chunk.Count() ) {
      ON_ERROR("ON_BinaryArchive::EndRead3dmLayerTable() - m_chunk.Count() > 0");
      rc = FALSE;
    }
    else {
      // rewind to start of chunks
      rc = SeekFromStart(32);
    }
    m_active_table = no_active_table;
  }
  else {
    rc = EndRead3dmTable( TCODE_LAYER_TABLE );
  }
  return rc;
}


///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////

BOOL ON_BinaryArchive::BeginWrite3dmGroupTable()
{
  BOOL rc = FALSE;
  rc = BeginWrite3dmTable( TCODE_GROUP_TABLE );
  return rc;
}

BOOL ON_BinaryArchive::Write3dmGroup( const ON_Group&  group )
{
  BOOL rc = FALSE;
  if ( m_active_table != group_table ) {
    ON_ERROR("ON_BinaryArchive::Write3dmGroup() - m_active_table != group_table");
  }

  const ON_3DM_CHUNK* c = m_chunk.Last();
  if ( c && c->m_typecode == TCODE_GROUP_TABLE ) {
    rc = BeginWrite3dmChunk( TCODE_GROUP_RECORD, 0 );
    if ( rc ) {
      rc = WriteObject( group );
      if ( !EndWrite3dmChunk() )
        rc = FALSE;
    }
  }
  else {
    ON_ERROR("ON_BinaryArchive::Write3dmGroup() must be called in BeginWrite3dmGroupTable() block");
    rc = FALSE;
  }

  return rc;
}

BOOL ON_BinaryArchive::EndWrite3dmGroupTable()
{
  BOOL rc = FALSE;
  rc = EndWrite3dmTable( TCODE_GROUP_TABLE );
  return rc;
}

BOOL ON_BinaryArchive::BeginRead3dmGroupTable()
{
  if ( m_3dm_version == 1 ) {
    return TRUE;
  }
  BOOL rc = FALSE;
  rc = BeginRead3dmTable( TCODE_GROUP_TABLE );
  return rc;
}

int ON_BinaryArchive::Read3dmGroup( ON_Group** ppGroup )
{
  if ( !ppGroup )
    return 0;
  *ppGroup = 0;
  if ( m_3dm_version == 1 ) {
    return 0;
  }
  if ( m_active_table != group_table ) {
    ON_ERROR("ON_BinaryArchive::BeginRead3dmGroupTable() - m_active_table != no_active_table");
  }
  if ( m_3dm_opennurbs_version < 200012210 ) {
    // 3DM archives written before version 200012210 and before do not have group tables
    return 0;
  }

  unsigned int tcode;
  int value;
  ON_Group* group = NULL;
  tcode = 0;
  if ( BeginRead3dmChunk( &tcode, &value ) ) {
    if ( tcode == TCODE_GROUP_RECORD ) {
      ON_Object* p = 0;
      if ( ReadObject( &p ) ) {
        group = ON_Group::Cast(p);
        if ( !group )
          delete p;
      }
      if (!group) {
        ON_ERROR("ON_BinaryArchive::Read3dmGroup() - corrupt group table");
      }
    }
    else if ( tcode != TCODE_ENDOFTABLE ) {
      ON_ERROR("ON_BinaryArchive::Read3dmGroup() - corrupt group table");
    }
    EndRead3dmChunk();
  }
  *ppGroup = group;
  return (group) ? 1 : 0;
}

BOOL ON_BinaryArchive::EndRead3dmGroupTable()
{
  BOOL rc = FALSE;
  if ( m_3dm_version == 1 ) {
    return TRUE;
  }
  else {
    rc = EndRead3dmTable( TCODE_GROUP_TABLE );
  }
  return rc;
}

///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////

BOOL ON_BinaryArchive::BeginWrite3dmMaterialTable()
{
  return BeginWrite3dmTable( TCODE_MATERIAL_TABLE );
}

BOOL ON_BinaryArchive::Write3dmMaterial( const ON_Material& material )
{
  BOOL rc = FALSE;
  if ( m_active_table != material_table ) {
    ON_ERROR("ON_BinaryArchive::Write3dmMaterial() - m_active_table != material_table");
  }
  const ON_3DM_CHUNK* c = m_chunk.Last();
  if ( !c || c->m_typecode != TCODE_MATERIAL_TABLE ) {
    ON_ERROR("ON_BinaryArchive::Write3dmMaterial() - active chunk typecode != TCODE_MATERIAL_TABLE");
  }
  else {
    rc = BeginWrite3dmChunk( TCODE_MATERIAL_RECORD, 0 );
    if (rc) {
      rc = WriteObject( material );
      if ( !EndWrite3dmChunk() )
        rc = FALSE;
    }
  }
  return rc;
}

BOOL ON_BinaryArchive::EndWrite3dmMaterialTable()
{
  return EndWrite3dmTable( TCODE_MATERIAL_TABLE );
}

BOOL ON_BinaryArchive::BeginRead3dmMaterialTable()
{
  m_3dm_v1_material_index = 0;
  return BeginRead3dmTable( TCODE_MATERIAL_TABLE );
}

BOOL ON_BinaryArchive::Read3dmV1String( ON_String& s )
{
  int string_length = 0;
  s.Empty();
  BOOL rc = ReadInt( &string_length );
  if (rc) {
    s.ReserveArray(string_length+1);
    rc = ReadChar( string_length, s.Array() );
    if (rc)
      s.SetLength(string_length);
  }
  return rc;
}

BOOL ON_BinaryArchive::Read3dmV1AttributesOrMaterial( 
                         ON_3dmObjectAttributes* attributes,
                         ON_Material* material,
                         BOOL& bHaveMat,
                         unsigned int end_mark_tcode )
{
  // Check ReadV1Material() if you fix any bugs in the mateial related items

  BOOL rc = FALSE;
  unsigned int tcode, u;
  int value;
  ON_Color c;
  bHaveMat = FALSE;
  BOOL bEndRead3dmChunk_rc;

  if ( attributes ) {
    attributes->Default();
  }
  if ( material ) {
    material->Default();
    ON_Color white(255,255,255);
    ON_Color black(0,0,0);
    material->SetDiffuse( white );
    material->SetSpecular( white );
    material->SetAmbient(black);
  }

  for (;;) {
    if ( end_mark_tcode != TCODE_ENDOFTABLE ) {
      tcode = !end_mark_tcode;
      if ( !PeekAt3dmChunkType(&tcode,&value) ) {
        break; // should not happen
      }
      if ( tcode == end_mark_tcode ) {
        rc = TRUE;
        break; // done reading attributes
      }
    }
    if ( !BeginRead3dmChunk(&tcode,&value) )
      break;
    if ( tcode == end_mark_tcode ) {
      rc = EndRead3dmChunk();
      break;
    }
    
    switch( tcode ) 
    {      
    case (TCODE_OPENNURBS_OBJECT | TCODE_CRC | 0x7FFD):
      // 1.1 object 16 byte UUID + 2 byte crc
      if ( attributes )
        ReadUuid( attributes->m_uuid );
      break;
      
    case TCODE_LAYERREF:
      if ( attributes )
        attributes->m_layer_index = value;
      break;
      
    case TCODE_RGB:
      if ( value != 0xFFFFFF ) {
        if ( material ) {
          u = (unsigned int)value;
          c.SetRGB( u%256,(u>>8)%256,(u>>16)%256 );
          material->SetDiffuse(c);
          material->SetShine((u >> 24)/100.0*ON_Material::MaxShine());
        }
        bHaveMat = TRUE;
      }
      break;		
      
    case TCODE_RGBDISPLAY:
      if ( attributes ) {
        u = (unsigned int)value;
        attributes->m_color.SetRGB( u%256,(u>>8)%256,(u>>16)%256 );
      }
      break;
      
    case TCODE_TRANSPARENCY:
      if ( value > 0 && value <= 255 ) {
        if ( material )
          material->SetTransparency(value/255.0);
        bHaveMat = TRUE;
      }
      break;				
      
    case TCODE_NAME:
      if ( attributes ) {
        ON_String s;
        Read3dmV1String(s);
        if( s.Length() > 0 )
          attributes->m_name = s;
      }
      break;
   
    case TCODE_TEXTUREMAP:
      {
        ON_String s;
        Read3dmV1String(s);
        if ( s.Length() > 0 ) {
          if ( material )
            material->SetTextureBitmapFileName(s);
          bHaveMat = TRUE;
        }
      }
      break;

    case TCODE_BUMPMAP:
      if ( material ) {
        ON_String s;
        Read3dmV1String(s);
        if ( s.Length() ) {
          if ( material )
            material->SetBumpBitmapFileName(s);
          bHaveMat = TRUE;
        }
      }
      break;
      
    case TCODE_XDATA:
      // v1 "xdata"
      if ( attributes ) {
        ON_String layer_name;
        ON_String xid;
        int sizeof_xid = 0;
        int sizeof_data = 0;
        ReadInt(&sizeof_xid);
        ReadInt(&sizeof_data);
        xid.SetLength(sizeof_xid);
        ReadByte(sizeof_xid,xid.Array());
        if ( !on_stricmp("RhHidePrevLayer",xid) )
        {
          // v1 object is hidden - real layer name is in xdata
          layer_name.SetLength(sizeof_data);
          ReadByte(sizeof_data,layer_name.Array());
        }
        else if ( !on_stricmp("RhFreezePrevLayer",xid) )
        {
          // v1 object is locked - real layer name is in xdata
          layer_name.SetLength(sizeof_data);
          ReadByte(sizeof_data,layer_name.Array());
        }
        else {
          m_3dm_v1_suppress_error_message |= 0x0002; // disable v1 EndRead3dmChunk() partially read chunk warning
        }
        // call to EndRead3dmChunk() will skip unread junk
      }
      break;
      
    case TCODE_DISP_CPLINES:
      if ( value > 0 && attributes )
        attributes->m_wire_density = value;
      break;

    case TCODE_RENDER_MATERIAL_ID:
      {
        int flag;
        ON_String s;
        ReadInt(&flag);
        if ( flag == 1 ) {
          Read3dmV1String(s);
          if ( s.Length() > 0 ) {
            if ( material )
              material->SetMaterialName(s);
            bHaveMat = TRUE;
          }
        }
      }
      break;
      
    default:
      // obsolete attributes from v1
      m_3dm_v1_suppress_error_message |= 0x0002; // disable v1 EndRead3dmChunk() partially read chunk warning
      break;
    }

    bEndRead3dmChunk_rc = EndRead3dmChunk();
    m_3dm_v1_suppress_error_message &= 0xFFFD; // enable v1 EndRead3dmChunk() partially read chunk warning
    if ( !bEndRead3dmChunk_rc )
      break;
  }

  if ( bHaveMat ) {
    if ( attributes )
      attributes->m_material_index = m_3dm_v1_material_index;
    if ( material )
      material->SetMaterialIndex(m_3dm_v1_material_index);
    m_3dm_v1_material_index++;
  }

  return rc;
}



int ON_BinaryArchive::Read3dmV1Material( ON_Material** ppMaterial )
{
  int rc = 0;
  // returns -1: failure
  //          0: end of material table
  //          1: success
  //          2: skipped filtered objects

  ON_Material material;
  unsigned int tcode;
  int value;
  BOOL bHaveMat;
  BOOL bEndReadChunk_rc;
  // reads NURBS, point, and mesh objects
  for(;;)
  {
    bHaveMat = FALSE;
    rc = 0;
    tcode = 0;
    value = 0;
    if ( !BeginRead3dmChunk(&tcode,&value) ) {
      // assume we are at the end of the file
      break;
    }
    switch(tcode)
    {
    case TCODE_RH_POINT:
      // v1 3d point
      {
        ON_3dPoint pt; // need to read point to get to material defn
        ReadPoint( pt );
      }
      Read3dmV1AttributesOrMaterial( NULL, &material, bHaveMat, TCODE_ENDOFTABLE);
      break;

    case TCODE_MESH_OBJECT:
      // v1 mesh
      {
        unsigned int tc;
        int i;
        if ( !PeekAt3dmChunkType( &tc, &i ) )
          break;
        if ( tc != TCODE_COMPRESSED_MESH_GEOMETRY )
          break;
        // skip over the TCODE_COMPRESSED_MESH_GEOMETRY chunk
        if ( !BeginRead3dmChunk(&tc,&i) )
          break;
        if ( !EndRead3dmChunk() )
          break;
        // attributes and material informtion follow the TCODE_COMPRESSED_MESH_GEOMETRY chunk
        rc = Read3dmV1AttributesOrMaterial( NULL, &material, bHaveMat, TCODE_ENDOFTABLE );
      }
      break;

    case TCODE_AG_SHELL:
      // v1 polysurface
      rc = Read3dmV1AttributesOrMaterial( NULL, &material, bHaveMat, TCODE_AG_SHELLSTUFF );
      break;
    case TCODE_AG_FACE:
      // v1 trimmed surface
      rc = Read3dmV1AttributesOrMaterial( NULL, &material, bHaveMat, TCODE_AG_FACESTUFF );
      break;
    case TCODE_AG_CURVE:
      // v1 curve
      rc = Read3dmV1AttributesOrMaterial( NULL, &material, bHaveMat, TCODE_AG_CURVESTUFF );
      break;

    case TCODE_RHINOIO_OBJECT_NURBS_CURVE:
    case TCODE_RHINOIO_OBJECT_NURBS_SURFACE:
    case TCODE_RHINOIO_OBJECT_BREP:
      // old Rhino I/O toolkit nurbs curve, surface, and breps
      {
        unsigned int tc;
        int i;
        if ( !PeekAt3dmChunkType( &tc, &i ) )
          break;
        if ( tc != TCODE_RHINOIO_OBJECT_DATA )
          break;
        // skip over the TCODE_RHINOIO_OBJECT_DATA chunk
        if ( !BeginRead3dmChunk(&tc,&i) )
          break;
        if ( !EndRead3dmChunk() )
          break;
        rc = Read3dmV1AttributesOrMaterial( NULL, &material, bHaveMat, TCODE_RHINOIO_OBJECT_END );
      }
      break;
    }

    m_3dm_v1_suppress_error_message |= 0x0002; // disable v1 EndRead3dmChunk() partially read chunk warning
    bEndReadChunk_rc = EndRead3dmChunk();
    m_3dm_v1_suppress_error_message &= 0xFFFD; // enable v1 EndRead3dmChunk() partially read chunk warning
    if (!bEndReadChunk_rc ) {
      rc = -1;
      break;
    }
    if ( bHaveMat && ppMaterial) {
      *ppMaterial = new ON_Material(material);
      rc = 1;
      break;
    }
  }
  return rc;
}


int ON_BinaryArchive::Read3dmMaterial( ON_Material** ppMaterial )
{
  int rc = 0;
  if ( !ppMaterial )
    return 0;
  *ppMaterial = 0;
  ON_Material* material = NULL;
  unsigned int tcode = 0;
  int value = 0;
  if ( m_3dm_version == 1 ) {
    rc = ON_BinaryArchive::Read3dmV1Material( ppMaterial );
  }
  else {
    // version 2+
    rc = -1;
    if ( BeginRead3dmChunk( &tcode, &value ) ) {
      if ( tcode == TCODE_MATERIAL_RECORD ) {
        ON_Object* p = 0;
        if ( ReadObject( &p ) ) {
          material = ON_Material::Cast(p);
          if ( !material )
            delete p;
          else {
            if ( ppMaterial )
              *ppMaterial = material;
            rc = 1;
          }
        }
        if (!material) {
          ON_ERROR("ON_BinaryArchive::Read3dmMaterial() - corrupt material table");
        }
      }
      else if ( tcode == TCODE_ENDOFTABLE ) {
        // end of material table
        rc = 0;
      }
      else {
        ON_ERROR("ON_BinaryArchive::Read3dmMaterial() - corrupt material table");
      }
      if ( !EndRead3dmChunk() )
        rc = -1;
    }
  }
  return rc;
}

BOOL ON_BinaryArchive::EndRead3dmMaterialTable()
{
  return EndRead3dmTable( TCODE_MATERIAL_TABLE );
}


BOOL ON_BinaryArchive::BeginWrite3dmLightTable()
{
  return BeginWrite3dmTable( TCODE_LIGHT_TABLE );
}

BOOL ON_BinaryArchive::Write3dmLight( const ON_Light& light,  const ON_3dmObjectAttributes* attributes )
{
  BOOL rc = FALSE;
  if ( m_active_table != light_table ) {
    ON_ERROR("ON_BinaryArchive::Write3dmLight() - m_active_table != light_table");
  }
  const ON_3DM_CHUNK* c = m_chunk.Last();
  if ( c && c->m_typecode == TCODE_LIGHT_TABLE ) {
    rc = BeginWrite3dmChunk( TCODE_LIGHT_RECORD, 0 );
    if (rc) {
      // WriteObject writes TCODE_OPENNURBS_CLASS chunk that contains light definition
      rc = WriteObject( light );

      // optional TCODE_LIGHT_RECORD_ATTRIBUTES chunk
      if ( rc && attributes ) {
        rc = BeginWrite3dmChunk( TCODE_LIGHT_RECORD_ATTRIBUTES, 0 );
        if (rc) {
          rc = attributes->Write( *this );
          if (!EndWrite3dmChunk())
            rc = FALSE;
        }
      }

      // TCODE_LIGHT_RECORD_END chunk marks end of light record
      if ( BeginWrite3dmChunk( TCODE_LIGHT_RECORD_END, 0 ) ) {
        if (!EndWrite3dmChunk())
          rc = FALSE;
      }
      else {
        rc = FALSE;
      }

      if ( !EndWrite3dmChunk() ) // end of TCODE_LIGHT_RECORD
        rc = FALSE;
    }
  }
  else {
    ON_ERROR("ON_BinaryArchive::Write3dmMaterial() - active chunk typecode != TCODE_LIGHT_TABLE");
  }
  return rc;
}

BOOL ON_BinaryArchive::EndWrite3dmLightTable()
{
  return EndWrite3dmTable( TCODE_LIGHT_TABLE );
}

BOOL ON_BinaryArchive::BeginRead3dmLightTable()
{
  return BeginRead3dmTable( TCODE_LIGHT_TABLE );
}

int ON_BinaryArchive::Read3dmV1Light(  // returns 0 at end of light table
                    //         1 light successfully read
                    //        -1 if file is corrupt
          ON_Light** ppLight, // light returned here
          ON_3dmObjectAttributes* pAttributes// optional - if NOT NULL, object attributes are
                                  //            returned here
          )
{
  BOOL bHaveMat;
  ON_Material material;
  // TODO - read v1 lights
  if ( m_chunk.Count() != 0 ) {
    ON_ERROR("ON_BinaryArchive::Read3dmV1Light() m_chunk.Count() != 0");
    return FALSE;
  }
  BOOL rc = FALSE;
  unsigned int tcode;
  int value;

  // find TCODE_RH_SPOTLIGHT chunk
  for(;;)
  {
    if ( !BeginRead3dmChunk(&tcode,&value) )
      break; // assume we are at the end of the file
    if ( tcode == TCODE_RH_SPOTLIGHT ) {
      rc = 1;
      break;
    }
    if ( !EndRead3dmChunk() )
      break;
  }
  if (rc) {
    ON_3dPoint origin;
    ON_3dVector xaxis, yaxis;
    double radius;
    double height;
    double hotspot;

    if (rc) rc = ReadPoint( origin );
    if (rc) rc = ReadVector( xaxis );
    if (rc) rc = ReadVector( yaxis );
    if (rc) rc = ReadDouble( &radius );
    if (rc) rc = ReadDouble( &height );
    if (rc) rc = ReadDouble( &hotspot );
    if (rc && ppLight ) {
      ON_3dVector Z = ON_CrossProduct( xaxis, yaxis );
      ON_3dPoint  location = height*Z + origin;
      ON_3dVector direction = -Z;

      if( height > 0.0)
        direction *= height;
      ON_Light* light = new ON_Light;
      light->SetStyle( ON::world_spot_light );
      light->SetLocation(location);
      light->SetDirection(direction);
      light->SetSpotExponent( 64.0);
      if( radius > 0.0 && height > 0.0 )
        light->SetSpotAngleRadians( atan( radius/height));
      *ppLight = light;
    }

    if (rc && ppLight && *ppLight) {
      bHaveMat = FALSE;
      Read3dmV1AttributesOrMaterial(pAttributes,&material,bHaveMat,TCODE_ENDOFTABLE);
      if ( pAttributes )
        pAttributes->m_material_index = -1;
      if (bHaveMat)
        (*ppLight)->SetDiffuse(material.Diffuse());
    }

    if ( !EndRead3dmChunk() ) // end of TCODE_RH_SPOTLIGHT chunk
      rc = FALSE;
  }

  return rc;
}

int ON_BinaryArchive::Read3dmLight( ON_Light** ppLight, ON_3dmObjectAttributes* attributes )
{
  if ( attributes )
    attributes->Default();
  int rc = -1;
  if ( !ppLight )
    return 0;
  *ppLight = 0;
  if ( m_active_table != light_table ) {
    ON_ERROR("ON_BinaryArchive::Read3dmLight() - m_active_table != light_table");
  }
  else if ( m_3dm_version == 1 ) {
    rc = Read3dmV1Light( ppLight, attributes );
  }
  else {
    ON_Light* light = NULL;
    unsigned int tcode = 0;
    int value = 0;
    if ( BeginRead3dmChunk( &tcode, &value ) ) {
      if ( tcode == TCODE_LIGHT_RECORD ) {
        ON_Object* p = 0;
        if ( ReadObject( &p ) ) {
          light = ON_Light::Cast(p);
          if ( !light )
            delete p;
        }
        if (!light) {
          ON_ERROR("ON_BinaryArchive::Read3dmLight() - corrupt light table");
        }
        else {
          *ppLight = light;
          rc = 1;
        }
      }
      else if ( tcode != TCODE_ENDOFTABLE ) {
        ON_ERROR("ON_BinaryArchive::Read3dmLight() - corrupt light table");
      }
      else
        rc = 0;

      while(rc==1) {
        if (!BeginRead3dmChunk( &tcode, &value )) {
          rc = -1;
          break;
        }
        if ( tcode == TCODE_LIGHT_RECORD_ATTRIBUTES && attributes ) {
          if ( !attributes->Read( *this ) )
            rc = -1;
        }
        if ( !EndRead3dmChunk() ) {
          rc = -1;
          break;
        }
        if ( tcode == TCODE_LIGHT_RECORD_END )
          break;
      }

      EndRead3dmChunk();
    }
  }
  return rc;
}

BOOL ON_BinaryArchive::EndRead3dmLightTable()
{
  return EndRead3dmTable( TCODE_LIGHT_TABLE );
}

BOOL ON_BinaryArchive::BeginWrite3dmObjectTable()
{
  return BeginWrite3dmTable( TCODE_OBJECT_TABLE );
}

BOOL ON_BinaryArchive::Write3dmObject( const ON_Object& object, const ON_3dmObjectAttributes* attributes )
{
  BOOL rc = FALSE;
  if ( m_active_table != object_table ) {
    ON_ERROR("ON_BinaryArchive::Write3dmObject() - m_active_table != object_table");
  }
  const ON_3DM_CHUNK* c = m_chunk.Last();
  if ( c && c->m_typecode == TCODE_OBJECT_TABLE ) {
    Flush();
    rc = BeginWrite3dmChunk( TCODE_OBJECT_RECORD, 0 );
    if (rc) {
      // TCODE_OBJECT_RECORD_TYPE chunk integer value that can be used
      // for skipping unwanted types of objects
      rc = BeginWrite3dmChunk( TCODE_OBJECT_RECORD_TYPE, object.ObjectType() );
      if (rc) {
        if (!EndWrite3dmChunk())
          rc = FALSE;
      }

      // WriteObject writes TCODE_OPENNURBS_CLASS chunk that contains object definition
      rc = WriteObject( object );

      // optional TCODE_OBJECT_RECORD_ATTRIBUTES chunk
      if ( rc && attributes ) {
        rc = BeginWrite3dmChunk( TCODE_OBJECT_RECORD_ATTRIBUTES, 0 );
        if (rc) {
          rc = attributes->Write( *this );
          if (!EndWrite3dmChunk())
            rc = FALSE;
        }
      }

      // TCODE_OBJECT_RECORD_END chunk marks end of object record
      if ( BeginWrite3dmChunk( TCODE_OBJECT_RECORD_END, 0 ) ) {
        if (!EndWrite3dmChunk())
          rc = FALSE;
      }
      else {
        rc = FALSE;
      }

      if (!EndWrite3dmChunk()) // end of TCODE_OBJECT_RECORD
      {
        rc = FALSE;
      }
      Flush();
    }
    else {
      ON_ERROR("ON_BinaryArchive::Write3dmObject() - active chunk typecode != TCODE_OBJECT_TABLE");
    }
  }
  return rc;
}

BOOL ON_BinaryArchive::EndWrite3dmObjectTable()
{
  return EndWrite3dmTable( TCODE_OBJECT_TABLE );
}

BOOL ON_BinaryArchive::BeginRead3dmObjectTable()
{
  m_3dm_v1_material_index = 0;
  return BeginRead3dmTable( TCODE_OBJECT_TABLE );
}

BOOL ON_BinaryArchive::ReadV1_TCODE_RH_POINT(
  ON_Object** ppObject,
  ON_3dmObjectAttributes* pAttributes
  )
{
  // read v1 point
  BOOL rc = FALSE;
  BOOL bHaveMat = FALSE;
  ON_3dPoint pt;
  rc = ReadPoint(pt);
  if (rc) {
    *ppObject = new ON_Point(pt);
    Read3dmV1AttributesOrMaterial(pAttributes,NULL,bHaveMat,TCODE_ENDOFTABLE);
  }
  return rc;
}

BOOL ON_BinaryArchive::ReadV1_TCODE_MESH_OBJECT(
  ON_Object** ppObject,
  ON_3dmObjectAttributes* pAttributes
  )
{
  ON_Mesh* mesh = 0;
  BOOL rc = FALSE;
  // read v1 mesh
  unsigned int tcode;
  int i, value;
  if ( !BeginRead3dmChunk(&tcode,&value) )
    return FALSE;
  
  if ( tcode == TCODE_COMPRESSED_MESH_GEOMETRY ) for(;;) {

    int point_count = 0;
    int face_count = 0;
    BOOL bHasVertexNormals = FALSE;
    BOOL bHasTexCoords = FALSE;
    ON_BoundingBox bbox;

    if (!ReadInt(&point_count) )
      break;
    if ( point_count <= 0 )
      break;
    if (!ReadInt(&face_count) )
      break;
    if ( face_count <= 0 )
      break;
    if (!ReadInt(&bHasVertexNormals) )
      break;
    if (!ReadInt(&bHasTexCoords) )
      break;
    if ( !ReadPoint(bbox.m_min) )
      break;
    if ( !ReadPoint(bbox.m_max) )
      break;

    mesh = new ON_Mesh(face_count,point_count,bHasVertexNormals,bHasTexCoords);

    // read 3d vertex locations
    {
      ON_3dVector d = bbox.Diagonal();
      double dx = d.x / 65535.0;
      double dy = d.y / 65535.0;
      double dz = d.z / 65535.0;
      unsigned short xyz[3];
      ON_3fPoint pt;
      for ( i = 0; i < point_count; i++ ) {
        if ( !ReadShort(3,xyz) )
          break;
        pt.x = (float)(dx*xyz[0] + bbox.m_min.x);
        pt.y = (float)(dy*xyz[1] + bbox.m_min.y);
        pt.z = (float)(dz*xyz[2] + bbox.m_min.z);
        mesh->m_V.Append(pt);
      }
    }
    if ( mesh->m_V.Count() != point_count )
      break;

    // read triangle/quadrangle faces
    if ( face_count < 65535 ) {
      unsigned short abcd[4];
      for ( i = 0; i < face_count; i++ ) {
        if ( !ReadShort(4,abcd) )
          break;
        ON_MeshFace& f = mesh->m_F.AppendNew();
        f.vi[0] = abcd[0];
        f.vi[1] = abcd[1];
        f.vi[2] = abcd[2];
        f.vi[3] = abcd[3];
      }
    }
    else {
      int abcd[4];
      for ( i = 0; i < face_count; i++ ) {
        if ( !ReadInt(4,abcd) )
          break;
        ON_MeshFace& f = mesh->m_F.AppendNew();
        f.vi[0] = abcd[0];
        f.vi[1] = abcd[1];
        f.vi[2] = abcd[2];
        f.vi[3] = abcd[3];
      }
    }
    if ( mesh->m_F.Count() != face_count )
      break;

    if ( bHasVertexNormals ) {
      char xyz[3];
      ON_3fVector normal;
      for ( i = 0; i < point_count; i++ ) {
        if ( !ReadChar(3,xyz) )
          break;
        normal.x = (float)(((signed char)xyz[0])/127.0);
        normal.y = (float)(((signed char)xyz[1])/127.0);
        normal.z = (float)(((signed char)xyz[2])/127.0);
        mesh->m_N.Append(normal);
      }
      if ( mesh->m_N.Count() != mesh->m_V.Count() )
        break;
    }

    if ( bHasTexCoords ) {
      unsigned short uv[2];
      ON_2fPoint t;
      for ( i = 0; i < point_count; i++ ) {
        if ( !ReadShort(2,uv) )
          break;
        t.x = (float)(uv[0]/65535.0);
        t.y = (float)(uv[1]/65535.0);
        mesh->m_T.Append(t);
      }
      if ( mesh->m_T.Count() != mesh->m_V.Count() )
        break;
    }

    rc = TRUE;

    break;
  }
  
  if ( !EndRead3dmChunk() )
    rc = FALSE;

  if ( rc && mesh ) {
    *ppObject = mesh;
  }
  else {
    if ( mesh )
      delete mesh;
    rc = FALSE;
  }

  if ( rc && mesh ) {
    // attributes and material information follows the TCODE_COMPRESSED_MESH_GEOMETRY chunk;
    BOOL bHaveMat = FALSE;
    Read3dmV1AttributesOrMaterial(pAttributes,NULL,bHaveMat,TCODE_ENDOFTABLE);
  }

  return rc;
}

static BOOL BeginRead3dmAGSTUFF( ON_BinaryArchive& file, unsigned int stuff_tcode )
{
  // begins reading stuff chunk
  BOOL rc = FALSE;
  unsigned int tcode;
  int value;
  tcode = !stuff_tcode;
  for (;;) {
    if ( !file.BeginRead3dmChunk(&tcode,&value) )
      break;
    if ( tcode == stuff_tcode ) {
      rc = TRUE;
      break;
    }
    if ( !file.EndRead3dmChunk() )
      break;
  }
  return rc;
}

static ON_NurbsCurve* ReadV1_TCODE_AG_SPLINESTUFF( ON_BinaryArchive& file )
{
  // reads contents of a v1 TCODE_AG_SPLINESTUFF chunk
  ON_NurbsCurve* pNurbsCurve = 0;
  int i, dim, is_rat, order, cv_count, is_closed, form;
  ON_BoundingBox bbox;
  char c;
  short s;

  // read v1 agspline chunk
  if ( !file.ReadChar(1,&c) )
    return NULL;
  if ( c != 2 && c != 3 )
    return NULL;
  dim = c;
  if ( !file.ReadChar(1,&c) )
    return NULL;
  if ( c != 0 && c != 1 && c != 2 )
    return NULL;
  is_rat = c; // 0 = no, 1 = euclidean, 2 = homogeneous
  if ( !file.ReadChar(1,&c) )
    return NULL;
  if ( c < 2 )
    return NULL;
  order = c;    
  if ( !file.ReadShort(1,&s) )
    return NULL;
  if ( s < order )
    return NULL;
  cv_count = s; 

  // The "is_closed" and "form" flags are here to recording
  // the values of legacy data found in the Rhino file.  These
  // values are not used in the toolkit code.
  if ( !file.ReadByte(1,&c) )
    return NULL;  
  if (c != 0 && c != 1 && c != 2)
    return NULL;
  is_closed = c; // 0 = open, 1 = closed, 2 = periodic
  if ( !file.ReadByte(1,&c) )
    return NULL;
  form = c;

  // read bounding box
  if ( !file.ReadDouble( dim, bbox.m_min ) )
    return NULL;
  if ( !file.ReadDouble( dim, bbox.m_max ) )
    return NULL;

  pNurbsCurve = new ON_NurbsCurve( dim, is_rat?TRUE:FALSE, order, cv_count );

  BOOL rc = FALSE;
  for(;;) {

    // read legacy v1 knot vector 
    const int knot_count = order+cv_count-2;
    int       knot_index = 0;
    double    knot;

    // clamped_end_knot_flag: 0 = none, 1 = left, 2 = right, 3 = both
    char clamped_end_knot_flag = 0;
    if ( order > 2 ) 
      file.ReadChar(1,&clamped_end_knot_flag);

    // first knot(s)
    if ( !file.ReadDouble(&knot) )
      break;
    pNurbsCurve->m_knot[knot_index++] = knot;
    if (clamped_end_knot_flag % 2) {
      // clamped_start_knot
      while ( knot_index <= order-2 ) 
        pNurbsCurve->m_knot[knot_index++] = knot;
    }

    // middle knots
    while ( knot_index <= cv_count-1 ) {
      if ( !file.ReadDouble(&knot) )
        break;
      pNurbsCurve->m_knot[knot_index++] = knot;
    }
    if ( knot_index <= cv_count-1 )
      break;

    // end knot(s)
    if ( clamped_end_knot_flag >= 2 ) {
      while ( knot_index < knot_count )
        pNurbsCurve->m_knot[knot_index++] = knot;
    }
    else {
      while ( knot_index < knot_count ) {
        if ( !file.ReadDouble(&knot) )
          break;
        pNurbsCurve->m_knot[knot_index++] = knot;
      }
      if ( knot_index < knot_count )
        break;
    }

    // read legacy v1 control points
    const int cvdim = ( is_rat ) ? dim+1 : dim;
    for ( i = 0; i < cv_count; i++ ) {
      if ( !file.ReadDouble( cvdim, pNurbsCurve->CV(i) ) )
        break;
    }
    if ( i < cv_count )
      break;
    if ( is_rat ) {
      // is_rat == 1 check fails because invalid is_rat flags in old files
      // convert AG rational CVs from euclidean to homogeneous
      double w, *cv;
      int cv_index;
      for ( cv_index = 0; cv_index < cv_count; cv_index++ ) {
        cv = pNurbsCurve->CV(cv_index);
        w = cv[dim];
        for ( i = 0; i < dim; i++ )
          cv[i] *= w;
      }
    }
    if ( order == 2 && cv_count == 2 && pNurbsCurve->m_knot[0] > pNurbsCurve->m_knot[1] ) {
      // a few isolated old v1 3DM file created by Rhino 1.0 files have lines with reversed knots.
      pNurbsCurve->m_knot[0] = -pNurbsCurve->m_knot[0];
      pNurbsCurve->m_knot[1] = -pNurbsCurve->m_knot[1];
    }
    rc = TRUE;

    break;
  }
  if ( !rc && pNurbsCurve ) {
    delete pNurbsCurve;
    pNurbsCurve = 0;
  }
  return pNurbsCurve;
}

static BOOL ReadV1_TCODE_AG_SPLINE( ON_BinaryArchive& file,
  ON_NurbsCurve*& pNurbsCurve
  )
{
  // reads contents of TCODE_AG_SPLINE chunk
  pNurbsCurve = 0;
  BOOL rc = BeginRead3dmAGSTUFF(file, TCODE_AG_SPLINESTUFF );
  if ( !rc )
    return FALSE;
  pNurbsCurve = ReadV1_TCODE_AG_SPLINESTUFF(file);
  if ( !file.EndRead3dmChunk() ) // end of TCODE_AG_SPLINESTUFF chunk
    rc = FALSE;
  if ( !pNurbsCurve )
    rc = FALSE;
  return rc;
}

static ON_NurbsSurface* ReadV1_TCODE_AG_SURFACESTUFF( ON_BinaryArchive& file )
{
  // reads contents of TCODE_AG_SURFACESTUFF chunk
  ON_NurbsSurface* pNurbsSurface = 0;
  int i, j, dim=0, is_rat=0, order[2], cv_count[2], is_closed[2], is_singular[2], form;
  ON_BoundingBox bbox;
  char c;
  short s;

  // read contents of v1 TCODE_AG_SURFACESTUFF chunk
  if ( !file.ReadChar(1,&c) )
    return NULL;
  if ( c != 2 && c != 3 )
    return NULL;
  dim = c;
  if ( !file.ReadByte(1,&c) )
    return NULL;
  form = c;
  if ( !file.ReadChar(1,&c) )
    return NULL;
  if ( c < 1 )
    return NULL;
  order[0] = c+1;    
  if ( !file.ReadChar(1,&c) )
    return NULL;
  if ( c < 1 )
    return NULL;
  order[1] = c+1;    
  
  if ( !file.ReadShort(1,&s) )
    return NULL;
  if ( s < 1 )
    return NULL;
  cv_count[0] = order[0]-1+s;    
  if ( !file.ReadShort(1,&s) )
    return NULL;
  if ( s < 1 )
    return NULL;
  cv_count[1] = order[1]-1+s;    
  
  // "ratu" 0 = no, 1 = euclidean, 2 = homogeneous
  if ( !file.ReadChar(1,&c) )
    return NULL;
  if ( c == 1 ) is_rat = 1; else if ( c == 2 ) is_rat = 2;
  
  // "ratv" 0 = no, 1 = euclidean, 2 = homogeneous
  if ( !file.ReadChar(1,&c) )
    return NULL;
  if ( c == 1 ) is_rat = 1; else if ( c == 2 ) is_rat = 2;

  // The "is_closed" and "is_singular" flags are here to recording
  // the values of legacy data found in the Rhino file.  These
  // values are not used in the toolkit code.
  if ( !file.ReadByte(1,&c) )
    return NULL;  
  if (c != 0 && c != 1 && c != 2)
    return NULL;
  is_closed[0] = c; // 0 = open, 1 = closed, 2 = periodic
  if ( !file.ReadByte(1,&c) )
    return NULL;  
  if (c != 0 && c != 1 && c != 2)
    return NULL;
  is_closed[1] = c; // 0 = open, 1 = closed, 2 = periodic

  if ( !file.ReadByte(1,&c) )
    return NULL;  
  if (c != 0 && c != 1 && c != 2 && c != 3)
    return NULL;
  is_singular[0] = c;
  if ( !file.ReadByte(1,&c) )
    return NULL;  
  if (c != 0 && c != 1 && c != 2 && c != 3)
    return NULL;
  is_singular[1] = c;

  // read bounding box
  if ( !file.ReadDouble( dim, bbox.m_min ) )
    return NULL;
  if ( !file.ReadDouble( dim, bbox.m_max ) )
    return NULL;

  pNurbsSurface = new ON_NurbsSurface( dim, is_rat?TRUE:FALSE, 
                                       order[0], order[1], 
                                       cv_count[0], cv_count[1] );

  BOOL rc = FALSE;
  for (;;) {

    // read legacy v1 knot vectors 
    if ( !file.ReadDouble( order[0]+cv_count[0]-2, pNurbsSurface->m_knot[0] ) )
      break;
    if ( !file.ReadDouble( order[1]+cv_count[1]-2, pNurbsSurface->m_knot[1] ) )
      break;

    // read legacy v1 control points
    const int cvdim = ( is_rat ) ? dim+1 : dim;
    for ( i = 0; i < cv_count[0]; i++ ) {
      for ( j = 0; j < cv_count[1]; j++ ) {
        if ( !file.ReadDouble( cvdim, pNurbsSurface->CV(i,j) ) )
          break;
      }
      if ( j < cv_count[1] )
        break;
    }
    if ( i < cv_count[0] )
      break;
    if ( is_rat == 1 ) {
      double w, *cv;
      int k;
      for ( i = 0; i < cv_count[0]; i++ ) for ( j = 0; j < cv_count[1]; j++ ) {
        cv = pNurbsSurface->CV(i,j);
        w = cv[dim];
        for ( k = 0; k < dim; k++ )
          cv[k] *= w;
      }
    }
    rc = TRUE;

    break;
  }
  if ( !rc ) {
    delete pNurbsSurface;
    pNurbsSurface = 0;
  }

  return pNurbsSurface;
}

static BOOL ReadV1_TCODE_AG_SURFACE( ON_BinaryArchive& file,
  ON_NurbsSurface*& pNurbsSurface
  )
{
  pNurbsSurface = 0;
  BOOL rc = BeginRead3dmAGSTUFF( file, TCODE_AG_SURFACE );
  if ( rc ) {
    rc = BeginRead3dmAGSTUFF( file, TCODE_AG_SURFACESTUFF );
    if ( rc ) {
      pNurbsSurface = ReadV1_TCODE_AG_SURFACESTUFF( file );
      if ( !file.EndRead3dmChunk() ) // end of TCODE_AG_SURFACESTUFF chunk
        rc = FALSE;
    }
    if ( !file.EndRead3dmChunk() )
      rc = FALSE; // end of TCODE_AG_SURFACE chunk
  }
  if ( !rc && pNurbsSurface ) {
    delete pNurbsSurface;
    pNurbsSurface = 0;
  }
  return rc;
}

ON_Curve* ReadV1_TCODE_AG_CURVESTUFF( ON_BinaryArchive& file )
{
  // reads contents of a v1 TCODE_AG_CURVESTUFF chunk
  ON_Curve* curve = 0;
  ON_PolyCurve* polycurve = 0;
  ON_NurbsCurve* segment = 0;
  BOOL rc = FALSE;
  unsigned int tcode;
  int value, i;
  BOOL bIsPolyline = FALSE;
  ON_BoundingBox bbox;

  for (;;) {
    char c;
    short s;
    int segment_count = 0;
    file.ReadChar(1,&c);  
    if ( c != 2 && c != 3 )
      break;
    int dim = c;
    file.ReadChar(1,&c);
    if ( c != -1 && c != 0 && c != 1 && c != 2 )
      break;
    //int is_closed = (c) ? 1 : 0;
    file.ReadShort(&s);
    if ( s < 1 )
      break;
    file.ReadDouble( dim, bbox.m_min);
    file.ReadDouble( dim, bbox.m_max);
    segment_count = s;
    for ( i = 0; i < segment_count; i++ ) {
      segment = 0;
      if ( !file.BeginRead3dmChunk( &tcode, &value ) )
        break;
      if ( tcode == TCODE_AG_SPLINE && value > 0 ) {
        ReadV1_TCODE_AG_SPLINE(file,segment);
      }
      if ( !file.EndRead3dmChunk() ) {
        if ( segment ) {
          delete segment;
          segment = 0;
        }
        break;
      }
      if ( !segment )
        break;
      if ( i == 0 )
        polycurve = new ON_PolyCurve(segment_count);
      if ( segment->Order() != 2 || segment->IsRational() ) {
        bIsPolyline = FALSE;
        polycurve->Append(segment);
      }
      else {
        ON_LineCurve* line = new ON_LineCurve();
        line->m_t.Set( segment->m_knot[0], segment->m_knot[1] );
        segment->GetCV( 0, line->m_line.from );
        segment->GetCV( 1, line->m_line.to );
        line->m_dim = segment->m_dim;
        delete segment;
        segment = 0;
        polycurve->Append(line);
      }
    }
    if ( polycurve->Count() != segment_count )
      break;
    rc = TRUE;
    break;
  }

  if ( rc && polycurve ) {
    if ( polycurve->Count() == 1 ) {
      curve = polycurve->HarvestSegment(0);
      delete polycurve;
    }
    else if ( bIsPolyline ) {
      ON_PolylineCurve* pline = new ON_PolylineCurve();
      pline->m_dim = polycurve->Dimension();
      pline->m_t.Reserve(polycurve->Count()+1);
      pline->m_t.SetCount(polycurve->Count()+1);
      polycurve->GetSpanVector( pline->m_t.Array() );
      pline->m_pline.m_point.Reserve(polycurve->Count()+1);
      for ( i = 0; i < polycurve->Count(); i++ ) {
        pline->m_pline.m_point.Append(polycurve->SegmentCurve(i)->PointAtStart());
      }
      pline->m_pline.m_point.Append(polycurve->SegmentCurve(polycurve->Count()-1)->PointAtEnd());
      curve = pline;
      delete polycurve;
    }
    else {
      curve = polycurve;
    }
  }
  else {
    if ( polycurve )
      delete polycurve;
    rc = FALSE;
  }

  return curve;
}

BOOL ON_Brep::ReadV1_LegacyTrimStuff( ON_BinaryArchive& file, 
                            ON_BrepFace& face,
                            ON_BrepLoop& loop )
{
  // read contents of TCODE_AG_TEDGESTUFF chunk
  BOOL rc = FALSE;
  int revedge, gcon, mono;
  int curve2d_index = -1, curve3d_index = -1, trim_index = -1;
  double tol_3d, tol_2d;
  ON_Curve* curve2d = NULL;
  ON_Curve* curve3d = NULL;

  char c;
  file.ReadChar( &c );

  BOOL bHasEdge = (c % 2 ); // bit 0 = 1 if "tedge" has non NULL "->edge"
  BOOL bHasMate = (c & 6 ); // bit 1 or 2 = 1 if "tedge" has non NULL "->twin"
  BOOL bIsSeam  = (c & 2 ); // bit 1 = 1 if "tedge->twin" belongs to same face

  if ( !file.ReadInt(&revedge) )
    return FALSE;
  if ( !file.ReadInt(&gcon) )
    return FALSE;
  if ( !file.ReadInt(&mono) )
    return FALSE;
  if ( !file.ReadDouble( &tol_3d ) )
    return FALSE;
  if ( !file.ReadDouble( &tol_2d ) )
    return FALSE;

  // 2d trim curve
  if ( !BeginRead3dmAGSTUFF( file, TCODE_AG_CURVE ) )
    return FALSE;
  if ( BeginRead3dmAGSTUFF( file, TCODE_AG_CURVESTUFF ) ) {
    curve2d = ReadV1_TCODE_AG_CURVESTUFF(file);
    if ( !file.EndRead3dmChunk() ) // end of TCODE_AG_CURVESTUFF chunk
      rc = FALSE;
  }
  if ( !file.EndRead3dmChunk() ) // end of TCODE_AG_CURVE chunk
    rc = FALSE;
  if ( !curve2d )
    return FALSE;
  curve2d_index = AddTrimCurve(curve2d);
  if ( curve2d_index < 0 ) {
    delete curve2d;
    return FALSE;
  }

  // 3d curve
  if ( bHasEdge ) {
    if ( !BeginRead3dmAGSTUFF( file, TCODE_AG_CURVE ) )
      return FALSE;
    if ( BeginRead3dmAGSTUFF( file, TCODE_AG_CURVESTUFF ) ) {
      curve3d = ReadV1_TCODE_AG_CURVESTUFF(file);
      if ( !file.EndRead3dmChunk() ) // end of TCODE_AG_CURVESTUFF chunk
        rc = FALSE;
    }
    if ( !file.EndRead3dmChunk() ) // end of TCODE_AG_CURVE chunk
      rc = FALSE;
    if ( !curve3d )
      return FALSE;
    curve3d_index = AddEdgeCurve(curve3d);
    if ( curve3d_index < 0 ) {
      delete curve3d;
      return FALSE;
    }
    ON_BrepEdge& edge = NewEdge(curve3d_index);
    ON_BrepTrim& trim = NewTrim( edge, 
                                    revedge ? TRUE : FALSE,
                                    loop,
                                    curve2d_index
                                    );
    trim_index = trim.m_trim_index;
  }
  else {
    ON_BrepTrim& trim = NewTrim( revedge ? TRUE : FALSE,
                                    loop,
                                    curve2d_index
                                    );
    trim_index = trim.m_trim_index;
  }
  if ( trim_index >= 0 ) {
    ON_BrepTrim& trim = m_T[trim_index];
    trim.m__legacy_2d_tol = tol_2d;
    trim.m__legacy_3d_tol = tol_3d;
    trim.m__legacy_flags_Set(gcon,mono);
    if ( bIsSeam ) {
      trim.m_type = ON_BrepTrim::seam;
    }
    else if ( bHasMate ) {
      trim.m_type = ON_BrepTrim::mated;
    }
    else if ( bHasEdge ) {
      trim.m_type = ON_BrepTrim::boundary;
    }
    else {
      trim.m_type = ON_BrepTrim::singular;
    }
  }

  return (trim_index>=0) ? TRUE : FALSE;
}

BOOL ON_Brep::ReadV1_LegacyTrim( ON_BinaryArchive& file, 
                            ON_BrepFace& face,
                            ON_BrepLoop& loop )
{
  BOOL rc = FALSE;
  if ( !BeginRead3dmAGSTUFF( file, TCODE_AG_TEDGE ) )
    return FALSE;
  rc = BeginRead3dmAGSTUFF( file, TCODE_AG_TEDGESTUFF );
  if ( rc ) {
    rc = ReadV1_LegacyTrimStuff( file, face, loop );
    if ( !file.EndRead3dmChunk() ) // end of TCODE_AG_TEDGESTUFF
      rc = FALSE;
  }
  if ( !file.EndRead3dmChunk() ) // end of TCODE_AG_TEDGE chunk
    rc = FALSE;
  return rc;
}


BOOL ON_Brep::ReadV1_LegacyLoopStuff( ON_BinaryArchive& file, 
                               ON_BrepFace& face )
{
  // reads contents of TCODE_AG_BOUNDARYSTUFF chunk
  // read boundary
  ON_BrepLoop::TYPE loop_type = ON_BrepLoop::unknown;
  int tedge_count, btype, lti;
  double pspace_box[2][2]; // parameter space bounding box

  if ( !file.ReadInt( &tedge_count ) )
    return FALSE;
  if ( tedge_count < 1 ) {
    return FALSE;
  }
  if ( !file.ReadInt( &btype ) )
    return FALSE;
  if ( btype < -1 || btype > 1 ) {
    return FALSE;
  }
  if ( !file.ReadDouble( 4, &pspace_box[0][0] ) )
    return FALSE;
  switch( btype ) {
  case -1:
    loop_type = ON_BrepLoop::slit;
    break;
  case  0:
    loop_type = ON_BrepLoop::outer;
    break;
  case  1:
    loop_type = ON_BrepLoop::inner;
    break;
  }
  ON_BrepLoop& loop = NewLoop( loop_type, face );

  for ( lti = 0; lti < tedge_count; lti++ ) {
    if ( !ReadV1_LegacyTrim( file, face, loop ) )
      return FALSE;
  }

  return TRUE;
}

BOOL ON_Brep::ReadV1_LegacyLoop( ON_BinaryArchive& file, 
                                        ON_BrepFace& face )
{
  BOOL rc = FALSE;
  if ( !BeginRead3dmAGSTUFF( file, TCODE_AG_BOUNDARY ) )
    return FALSE;
  rc = BeginRead3dmAGSTUFF( file, TCODE_AG_BOUNDARYSTUFF );
  if ( rc ) {
    rc = ReadV1_LegacyLoopStuff( file, face );
    if ( !file.EndRead3dmChunk() ) // end of TCODE_AG_BOUNDARYSTUFF
      rc = FALSE;
  }
  if ( !file.EndRead3dmChunk() ) // end of TCODE_AG_BOUNDARY chunk
    rc = FALSE;
  return rc;
}

BOOL ON_Brep::ReadV1_LegacyFaceStuff( ON_BinaryArchive& file )
{
  // reads contents of TCODE_AG_FACESTUFF chunk
  ON_NurbsSurface* surface = 0;
  ON_Workspace ws;
  int flipnorm = 0;
  int ftype = 0;
  int bndcnt = 0;
  int twincnt = 0;
  BOOL bHasOuter = FALSE;
  ON_BoundingBox face_bbox;

  int ti0 = m_T.Count();

  BOOL rc = FALSE;

  // read flags
  if ( !file.ReadInt(&flipnorm) )
    return FALSE;
  if ( flipnorm < 0 || flipnorm > 1 )
    return FALSE;
  if ( !file.ReadInt(&ftype) )
    return FALSE;
  if ( !file.ReadInt(&bndcnt) )
    return FALSE;
  bHasOuter = (bndcnt%2); // always TRUE in v1 files
  bndcnt /= 2;

  // read bounding box
  if ( !file.ReadDouble( 3, face_bbox.m_min ) )
    return FALSE;
  if ( !file.ReadDouble( 3, face_bbox.m_max ) )
    return FALSE;

  // B-rep edge gluing info 
  if ( !file.ReadInt(&twincnt) )
    return FALSE;
  short* glue = (twincnt > 0 ) ? (short*)ws.GetMemory(twincnt*sizeof(*glue)) : NULL;
  if (twincnt > 0) {
    if ( !file.ReadShort(twincnt,glue) )
      return FALSE;
  }

  // read surface
  if ( !ReadV1_TCODE_AG_SURFACE( file, surface ) )
    return FALSE;
  if ( !surface )
    return FALSE;
  const int srf_index = AddSurface(surface);

  // create face
  ON_BrepFace& face = NewFace(srf_index);
  face.m_bRev = (flipnorm) ? TRUE : FALSE;
  face.m_li.Reserve(bndcnt);

  // read boundary loops
  int loop_index = -1;
  if ( !bHasOuter ) {
    // TODO: cook up outer boundary loop (never happes with v1 files)
    face.m_li.Append(loop_index);
  }
  int bi;
  rc = TRUE;
  for ( bi = 0; rc && bi < bndcnt; bi++ ) {
    rc = ReadV1_LegacyLoop( file, face );
  }

  if ( twincnt > 0 ) {
    // twincnt = number of seams edges in face
    // glue[] = order 2 permutation of {0,....,twincnt-1}

    // set seam_i[] = m_T[] indices of seam trims
    int si, ti;
    const int ti1 = m_T.Count();
    int* seam_i = (int*)ws.GetMemory(twincnt*sizeof(*seam_i));
    for ( ti = ti0, si = 0; ti < ti1 && si < twincnt; ti++ ) {
      if (m_T[ti].m_type != ON_BrepTrim::seam )
        continue;
      seam_i[si++] = ti;
    }

    if ( si == twincnt ) {
      // glue seams
      for ( si = 0; si < twincnt; si++ ) {
        if ( glue[si] >= 0 && glue[si] < twincnt ) {
          const int i0 = seam_i[si];
          const int i1 = seam_i[glue[si]];
          // m_T[i0] and m_T[i1] use the same edge;
          const int ei0 = m_T[i0].m_ei;
          const int ei1 = m_T[i1].m_ei;
          if ( ei0 == -1 && ei1 >= 0 ) {
            m_T[i0].m_ei = ei1;
            m_E[ei1].m_ti.Append(i0);
          }
          else if ( ei1 == -1 && ei0 >= 0 ) {
            m_T[i1].m_ei = ei0;
            m_E[ei0].m_ti.Append(i1);
          }
        }
      }
    }
  }

  return rc;
}

BOOL ON_Brep::ReadV1_LegacyShellStuff( ON_BinaryArchive& file )
{
  // read contents of TCODE_AG_SHELLSTUFF chunk
  ON_Workspace ws;
  int outer = 0;
  int facecnt = 0;
  int twincnt = 0;
  ON_BoundingBox shell_bbox;
  const int ti0 = m_T.Count();

  /* read flags */
  file.ReadInt(&outer);  
  file.ReadInt(&facecnt);  

  // read bounding box
  file.ReadPoint( shell_bbox.m_min );
  file.ReadPoint( shell_bbox.m_max );

  /* B-rep edge gluing info */
  file.ReadInt(&twincnt);
  short* glue = (twincnt > 0 ) ? (short*)ws.GetMemory(twincnt*sizeof(*glue)) : NULL;
  if (twincnt > 0)
    file.ReadShort(twincnt,glue);

  BOOL rc = TRUE;
  int fi;
  for ( fi = 0; rc && fi < facecnt; fi++ ) {
    rc = BeginRead3dmAGSTUFF( file, TCODE_AG_FACE );
    if ( rc ) {
      rc = BeginRead3dmAGSTUFF( file, TCODE_AG_FACESTUFF );
      if ( rc ) {
        rc = ReadV1_LegacyFaceStuff( file );
        if ( !file.EndRead3dmChunk() ) // end of TCODE_AG_FACESTUFF chunk
          rc = FALSE;
      }
      if ( !file.EndRead3dmChunk() ) // end of TCODE_AG_FACE chunk
        rc = FALSE;
    }
  }

  if ( twincnt > 0 ) {
    // twincnt = number of shared (inter-face) edges
    // glue[] = order 2 permutation of {0,....,twincnt-1}

    // set share_i[] = m_T[] indices of shared trims
    int si, ti;
    const int ti1 = m_T.Count();
    int* share_i = (int*)ws.GetMemory(twincnt*sizeof(*share_i));
    for ( ti = ti0, si = 0; ti < ti1 && si < twincnt; ti++ ) {
      if (m_T[ti].m_type != ON_BrepTrim::mated )
        continue;
      share_i[si++] = ti;
    }

    if ( si == twincnt ) {
      // glue seams
      for ( si = 0; si < twincnt; si++ ) {
        if ( glue[si] >= 0 && glue[si] < twincnt ) {
          const int i0 = share_i[si];
          const int i1 = share_i[glue[si]];
          // m_T[i0] and m_T[i1] use the same edge;
          const int ei0 = m_T[i0].m_ei;
          const int ei1 = m_T[i1].m_ei;
          if ( ei0 == -1 && ei1 >= 0 ) {
            m_T[i0].m_ei = ei1;
            m_E[ei1].m_ti.Append(i0);
          }
          else if ( ei1 == -1 && ei0 >= 0 ) {
            m_T[i1].m_ei = ei0;
            m_E[ei0].m_ti.Append(i1);
          }
        }
      }
    }
  }

  return rc;
}

BOOL ON_BinaryArchive::ReadV1_TCODE_AG_CURVE(
  ON_Object** ppObject,
  ON_3dmObjectAttributes* pAttributes
  )
{
  ON_Curve* curve = NULL;
  BOOL rc = FALSE;
  unsigned int tcode;
  int value;
  BOOL bHaveMat = FALSE;

  Read3dmV1AttributesOrMaterial(pAttributes,NULL,bHaveMat,TCODE_AG_CURVESTUFF);

  if ( !BeginRead3dmChunk( &tcode, &value ) )
    return FALSE;
  if ( tcode == TCODE_AG_CURVESTUFF )
    curve = ReadV1_TCODE_AG_CURVESTUFF(*this);
  rc = EndRead3dmChunk();
  if ( !curve )
    rc = FALSE;
  else
    *ppObject = curve;
  return rc;
}


BOOL ON_BinaryArchive::ReadV1_TCODE_AG_FACE(
  ON_Object** ppObject,
  ON_3dmObjectAttributes* pAttributes
  )
{
  // read V1 TCODE_AG_FACE chunk
  BOOL bHaveMat = FALSE;
  if ( !Read3dmV1AttributesOrMaterial(pAttributes,NULL,bHaveMat,TCODE_AG_FACESTUFF) )
    return FALSE;
  if ( !BeginRead3dmAGSTUFF( *this, TCODE_AG_FACESTUFF ) )
    return FALSE;
  ON_Brep* brep = new ON_Brep();
  BOOL rc = brep->ReadV1_LegacyFaceStuff( *this );
  if ( !EndRead3dmChunk() ) // end of TCODE_AG_FACESTUFF chunk
    rc = FALSE;

  if ( !rc ) {
    delete brep;
  }
  else {
    brep->SetVertices();
    brep->SetTrimIsoFlags();
    brep->SetTolsFromLegacyValues();
    *ppObject = brep;
  }

  return rc;
}

BOOL ON_BinaryArchive::ReadV1_TCODE_AG_SHELL(
  ON_Object** ppObject,
  ON_3dmObjectAttributes* pAttributes
  )
{
  // read v1 TCODE_AG_SHELL chunk
  BOOL bHaveMat = FALSE;
  if ( !Read3dmV1AttributesOrMaterial(pAttributes,NULL,bHaveMat,TCODE_AG_SHELLSTUFF) )
    return FALSE;
  if ( !BeginRead3dmAGSTUFF( *this, TCODE_AG_SHELLSTUFF ) )
    return FALSE;
  ON_Brep* brep = new ON_Brep();
  BOOL rc = brep->ReadV1_LegacyShellStuff( *this );
  if ( !EndRead3dmChunk() ) // end of TCODE_AG_SHELLSTUFF chunk
    rc = FALSE;

  if ( !rc ) {
    delete brep;
  }
  else {
    brep->SetVertices();
    brep->SetTrimIsoFlags();
    brep->SetTolsFromLegacyValues();
    *ppObject = brep;
  }

  return rc;
}


static
ON_NurbsCurve* ReadV1_RHINOIO_NURBS_CURVE_OBJECT_DATA( ON_BinaryArchive& file )
{
  // read TCODE_RHINOIO_OBJECT_DATA chunk that is contained in a
  // TCODE_RHINOIO_OBJECT_NURBS_CURVE chunk.  The TCODE_RHINOIO_OBJECT_DATA
  // chunk contains the definition of NURBS curves written by the 
  // old RhinoIO toolkit.
  ON_NurbsCurve* curve = 0;
  BOOL rc = FALSE;
  unsigned int tcode;
  int value, version, dim, is_rat, order, cv_count, flag, i;
  if ( !file.BeginRead3dmChunk( &tcode, &value ) )
    return NULL;
  if ( tcode == TCODE_RHINOIO_OBJECT_DATA ) for (;;) {
    if ( !file.ReadInt(&version) )
      break;
    // int bReverse = version & 0x100;
    version &= 0xFFFFFEFF;
    if ( version != 100 && version != 101 )
      break;
    file.ReadInt(&dim);
    if ( dim < 1 )
      break;
    file.ReadInt(&is_rat);
    if ( is_rat < 0 || is_rat > 1 )
      break;
    file.ReadInt(&order);
    if ( order < 2 )
      break;
    file.ReadInt(&cv_count);
    if ( cv_count < order )
      break;
    file.ReadInt(&flag);
    if ( flag != 0 )
      break;

    curve = new ON_NurbsCurve(dim,is_rat,order,cv_count);
    if ( !file.ReadDouble( order+cv_count-2, curve->m_knot ) )
      break;
    int cvdim = is_rat ? dim+1 : dim;
    for ( i = 0; i < cv_count; i++ ) {
      if ( !file.ReadDouble( cvdim, curve->CV(i) ) )
        break;
    }
    if ( i < cv_count )
      break;
    rc = TRUE;
    break;
  }
  if ( !file.EndRead3dmChunk() ) // end of TCODE_RHINOIO_OBJECT_DATA chunk
    rc = FALSE;
  if ( !rc && curve ) {
    delete curve;
    curve = 0;
  }

  return curve;
}

static
ON_NurbsSurface* ReadV1_RHINOIO_NURBS_SURFACE_OBJECT_DATA( ON_BinaryArchive& file )
{
  // read TCODE_RHINOIO_OBJECT_DATA chunk that is contained in a
  // TCODE_RHINOIO_OBJECT_NURBS_SURFACE chunk.  The TCODE_RHINOIO_OBJECT_DATA
  // chunk contains the definition of NURBS surfaces written by the 
  // old RhinoIO toolkit.
  BOOL rc = FALSE;
  ON_NurbsSurface* surface = 0;
  unsigned int tcode;
  int value, version, dim, is_rat, order[2], cv_count[2], flag, i, j;

  if ( !file.BeginRead3dmChunk( &tcode, &value ) )
    return FALSE;
  if ( tcode == TCODE_RHINOIO_OBJECT_DATA ) for (;;) {
    if ( !file.ReadInt(&version) )
      break;
    // int bReverse = version & 0x100;
    version &= 0xFFFFFEFF;
    if ( version != 100 && version != 101 )
      break;
    file.ReadInt(&dim);
    if ( dim < 1 )
      break;
    file.ReadInt(&is_rat);
    if ( is_rat < 0 || is_rat > 1 )
      break;
    file.ReadInt(&order[0]);
    if ( order[0] < 2 )
      break;
    file.ReadInt(&order[1]);
    if ( order[1] < 2 )
      break;
    file.ReadInt(&cv_count[0]);
    if ( cv_count[0] < order[0] )
      break;
    file.ReadInt(&cv_count[1]);
    if ( cv_count[1] < order[1] )
      break;
    file.ReadInt(&flag);
    if ( flag != 0 )
      break;

    surface = new ON_NurbsSurface(dim,is_rat,order[0],order[1],cv_count[0],cv_count[1]);
    if ( !file.ReadDouble( order[0]+cv_count[0]-2, surface->m_knot[0] ) )
      break;
    if ( !file.ReadDouble( order[1]+cv_count[1]-2, surface->m_knot[1] ) )
      break;
    int cvdim = is_rat ? dim+1 : dim;
    for ( i = 0; i < cv_count[0]; i++ ) {
      for ( j = 0; j < cv_count[1]; j++ ) {
        if ( !file.ReadDouble( cvdim, surface->CV(i,j) ) )
          break;
      }
      if ( j < cv_count[1] )
        break;
    }
    if ( i < cv_count[0] )
      break;
    rc = TRUE;
    break;
  }
  if ( !file.EndRead3dmChunk() ) // end of TCODE_RHINOIO_OBJECT_DATA
    rc = FALSE;
  if ( !rc && surface ) {
    delete surface;
    surface = 0;
  }
  return surface;
}

BOOL ON_BinaryArchive::ReadV1_TCODE_RHINOIO_OBJECT_NURBS_CURVE(
  ON_Object** ppObject,
  ON_3dmObjectAttributes* pAttributes
  )
{
  // read contents of ReadV1_TCODE_RHINOIO_OBJECT_NURBS_CURVE chunk
  // written by v1 RhinoIO toolkit
  ON_NurbsCurve* curve = 0;
  BOOL rc = FALSE;
  BOOL bHaveMat = FALSE;
  
  // reads TCODE_RHINOIO_OBJECT_DATA header and nurbs curve definition
  curve = ReadV1_RHINOIO_NURBS_CURVE_OBJECT_DATA(*this);

  if ( curve ) {
    *ppObject = curve;
    rc = TRUE;
    Read3dmV1AttributesOrMaterial(pAttributes,NULL,bHaveMat,TCODE_RHINOIO_OBJECT_END);
  }

  return rc;
}

BOOL ON_BinaryArchive::ReadV1_TCODE_RHINOIO_OBJECT_NURBS_SURFACE(
  ON_Object** ppObject,
  ON_3dmObjectAttributes* pAttributes
  )
{
  // read contents of TCODE_RHINOIO_OBJECT_NURBS_SURFACE chunk
  // written by v1 RhinoIO toolkit
  BOOL bHaveMat = FALSE;
  BOOL rc = FALSE;
  ON_NurbsSurface* surface = 0;

  surface = ReadV1_RHINOIO_NURBS_SURFACE_OBJECT_DATA( *this );

  if ( surface ) {
    *ppObject = surface;
    rc = TRUE;
    Read3dmV1AttributesOrMaterial(pAttributes,NULL,bHaveMat,TCODE_RHINOIO_OBJECT_END);
  }

  return rc;
}

static
ON_Curve* ReadV1_RHINOIO_BREP_CURVE( ON_BinaryArchive& file )
{
  ON_Curve* curve = NULL;
  ON_PolyCurve* pcurve = NULL;
  ON_NurbsCurve* nurbs_curve = NULL;
  int segment_index, segment_count = 0;
  unsigned int tcode;
  int value;
  
  if ( !file.ReadInt(&segment_count) )
    return NULL;
  if ( segment_count < 1 )
    return NULL;

  for ( segment_index = 0; segment_index < segment_count; segment_index++ ) {
    if ( !file.BeginRead3dmChunk(&tcode,&value) )
      break;
    if ( tcode == TCODE_RHINOIO_OBJECT_NURBS_CURVE ) {
      nurbs_curve = ReadV1_RHINOIO_NURBS_CURVE_OBJECT_DATA( file );
    }
    if ( !file.EndRead3dmChunk() )
      break;
    if ( !nurbs_curve )
      break;
    if ( segment_index == 0 ) {
      curve = nurbs_curve;
      nurbs_curve = 0;
    }
    else {
      if ( segment_index == 1 ) {
        pcurve = new ON_PolyCurve();
        pcurve->Append(curve);
        curve = pcurve;
      }
      pcurve->Append(nurbs_curve);
      nurbs_curve = NULL;
    }
  }

  if ( segment_index < segment_count ) {
    if ( nurbs_curve ) {
      delete nurbs_curve;
      nurbs_curve = 0;
    }
    if ( curve ) {
      delete curve;
      curve = 0;
    }
  }
  return curve;
}

BOOL ON_BinaryArchive::ReadV1_TCODE_RHINOIO_OBJECT_BREP(
  ON_Object** ppObject,
  ON_3dmObjectAttributes* pAttributes
  )
{
  ON_3dPoint m_oldTrim_mP[2];
  BOOL bHaveMat = FALSE;
  BOOL rc = FALSE;
  ON_Brep* brep = 0;
  unsigned int tcode;
  int value;
  if ( !BeginRead3dmChunk( &tcode, &value ) )
    return FALSE;
  if ( tcode == TCODE_RHINOIO_OBJECT_DATA ) for (;;) {
    int version = -1;
    int sz, i, j;
    double tol2d, tol3d;
    if ( !ReadInt( &version ) ) 
      break; // serialization version
    // version = 100 means the b-rep was written by the RhinoIO toolkit
    // version = 101 means the b-rep was written by Rhino 1.0
    if ( version != 100 && version != 101 ) {
      return FALSE;
    }

    brep = new ON_Brep();

    // 2d trimming curves
    if ( !ReadInt( &sz ) )
      break;
    if ( sz < 1 ) {
      break;
    }
    brep->m_C2.Reserve(sz);
    for ( i = 0; i < sz; i++ ) {
      ON_Curve* curve = ReadV1_RHINOIO_BREP_CURVE( *this );
      if ( !curve )
        break;
      brep->m_C2.Append(curve);
    }
    if ( i < sz )
      break;

    // 3d trimming curves
    if ( !ReadInt( &sz ) )
      break;
    if ( sz < 1 ) {
      break;
    }
    brep->m_C3.Reserve(sz);
    for ( i = 0; i < sz; i++ ) {
      ON_Curve* curve = ReadV1_RHINOIO_BREP_CURVE( *this );
      if ( !curve )
        break;
      brep->m_C3.Append(curve);
    }
    if ( i < sz )
      break;

    // 3d untrimmed surfaces
    if ( !ReadInt( &sz ) )
      break;
    if ( sz < 1 ) {
      break;
    }
    brep->m_S.Reserve(sz);
    for ( i = 0; i < sz; i++ ) {
      ON_NurbsSurface* surface = 0;
      if ( !BeginRead3dmChunk(&tcode,&value) )
        break;
      if ( tcode == TCODE_RHINOIO_OBJECT_NURBS_SURFACE ) {
        surface = ReadV1_RHINOIO_NURBS_SURFACE_OBJECT_DATA( *this );
      }
      if ( !EndRead3dmChunk() )
        break;
      if ( !surface )
        break;
      brep->m_S.Append(surface);
    }
    if ( i < sz )
      break;

    // vertices
    ReadInt( &sz );
    brep->m_V.Reserve(sz);
    for ( i = 0; i < sz; i++ ) {
      ON_BrepVertex& vertex = brep->NewVertex();
      if ( !ReadInt( &vertex.m_vertex_index ) ) break;
      if ( !ReadPoint( vertex.point ) ) break;
      if ( !ReadArray( vertex.m_ei ) ) break;
      if ( !ReadDouble( &vertex.m_tolerance ) ) break;
    }
    if ( i < sz )
      break;

    // edges
    ReadInt( &sz );
    brep->m_E.Reserve(sz);
    for ( i = 0; i < sz; i++ ) {
      ON_BrepEdge& edge = brep->NewEdge();
      if ( !ReadInt( &edge.m_edge_index ) ) break;
      if ( !ReadInt( &edge.m_c3i ) ) break;
      if ( !ReadInterval( edge.m_domain ) ) break;
      if ( !ReadInt( 2, edge.m_vi ) ) break;
      if ( !ReadArray( edge.m_ti ) ) break;
      if ( !ReadDouble( &edge.m_tolerance ) ) break;
    }
    if ( i < sz )
      break;

    // trims
    ReadInt( &sz );
    brep->m_T.Reserve(sz);
    for ( i = 0; i < sz; i++ ) {
      ON_BrepTrim& trim = brep->NewTrim();
      if ( !ReadInt( &trim.m_trim_index ) ) break;
      if ( !ReadInt( &trim.m_c2i ) ) break;
      if ( !ReadInterval( trim.m_t ) ) break;
      if ( !ReadInt( &trim.m_ei ) ) break;
      if ( !ReadInt( 2, trim.m_vi ) ) break;
      if ( !ReadInt( &trim.m_bRev3d ) ) break;
      if ( !ReadInt( &j ) ) break;
      switch(j) {
      case 1: trim.m_type = ON_BrepTrim::boundary; break;
      case 2: trim.m_type = ON_BrepTrim::mated; break;
      case 3: trim.m_type = ON_BrepTrim::seam; break;
      case 4: trim.m_type = ON_BrepTrim::singular; break;
      }
      if ( !ReadInt( &j ) ) break; // legacy iso flag - ignore and recaluate
      if ( !ReadInt( &trim.m_li ) ) break;
      if ( !ReadDouble( 2, trim.m_tolerance ) ) break;
      if ( !ReadPoint( m_oldTrim_mP[0] ) ) break;
      if ( !ReadPoint( m_oldTrim_mP[1] ) ) break;
      if ( !ReadDouble( &tol2d ) ) break;
      if ( !ReadDouble( &tol3d ) ) break;
    }
    if ( i < sz )
      break;

    // loops
    ReadInt( &sz );
    brep->m_L.Reserve(sz);
    for ( i = 0; i < sz; i++ ) {
      ON_BrepLoop& loop = brep->NewLoop(ON_BrepLoop::unknown);
      if ( !ReadInt( &loop.m_loop_index ) ) break;
      if ( !ReadArray( loop.m_ti ) ) break;
      if ( !ReadInt( &j ) ) break;
      switch (j) {
      case 1: loop.m_type = ON_BrepLoop::outer; break;
      case 2: loop.m_type = ON_BrepLoop::inner; break;
      case 3: loop.m_type = ON_BrepLoop::slit; break;
      }
      if ( !ReadInt( &loop.m_fi ) ) break;
    }
    if ( i < sz )
      break;

    // faces
    ReadInt( &sz );
    brep->m_F.Reserve(sz);
    for ( i = 0; i < sz; i++ ) {
      ON_BrepFace& face = brep->NewFace();
      if ( !ReadInt( &face.m_face_index ) ) break;
      if ( !ReadArray( face.m_li ) ) break;
      if ( !ReadInt( &face.m_si ) ) break;
      if ( !ReadInt( &face.m_bRev ) ) break;
    }
    if ( i < sz )
      break;

    // bounding box
    {
      ON_BoundingBox bbox;
      if ( !ReadPoint( bbox.m_min ) ) break;
      if ( !ReadPoint( bbox.m_max ) ) break;
    }

    rc = TRUE;
    break;
  }
  if ( !EndRead3dmChunk() )
    rc = FALSE;
  if ( rc && brep ) {
    brep->SetTrimIsoFlags();
    *ppObject = brep;
  }
  else {
    if ( brep )
      delete brep;
    rc = FALSE;
  }

  if ( rc && brep ) {
    Read3dmV1AttributesOrMaterial(pAttributes,NULL,bHaveMat,TCODE_RHINOIO_OBJECT_END);
  }

  return rc;
}

int
ON_BinaryArchive::Read3dmV1Object( 
  ON_Object** ppObject,                // object is returned here
  ON_3dmObjectAttributes* pAttributes, // optional - object attributes 
  unsigned int object_filter           // optional filter made by or-ing object_type bits
  )
{
  int rc = 0;
  // returns -1: failure
  //          0: end of geometry table
  //          1: success
  //          2: skipped filtered objects

  unsigned int tcode;
  int value;
  // reads NURBS, point, and mesh objects
  for(;;)
  {
    tcode = 0;
    value = 0;
    if ( !BeginRead3dmChunk(&tcode,&value) ) {
      rc = 0; // at the end of the file
      break;
    }
    switch(tcode)
    {
    case TCODE_RH_POINT:
      // v1 3d point
      if ( 0 != (ON::point_object & object_filter) ) {
        if (ReadV1_TCODE_RH_POINT( ppObject, pAttributes ))
          rc = 1;
        else
          rc = -1;
      }
      else {
        rc = 2;
      }
      break;
    case TCODE_MESH_OBJECT:
      // v1 mesh
      if ( 0 != (ON::mesh_object & object_filter) ) {
        if ( ReadV1_TCODE_MESH_OBJECT( ppObject, pAttributes ) )
          rc = 1;
        else
          rc = -1;
      }
      else {
        rc = 2;
      }
      break;
    case TCODE_AG_SHELL:
      // v1 polysurface
      if ( 0 != (ON::mesh_object & object_filter) ) {
        if ( ReadV1_TCODE_AG_SHELL( ppObject, pAttributes ) )
          rc = 1;
        else
          rc = -1;
      }
      else {
        rc = 2;
      }
      break;
    case TCODE_AG_FACE:
      // v1 trimmed surface
      if ( 0 != (ON::mesh_object & object_filter) ) {
        if ( ReadV1_TCODE_AG_FACE( ppObject, pAttributes ) )
          rc = 1;
        else
          rc = -1;
      }
      else {
        rc = 2;
      }
      break;
    case TCODE_AG_CURVE:
      // v1 curve
      if ( 0 != (ON::mesh_object & object_filter) ) {
        if ( ReadV1_TCODE_AG_CURVE( ppObject, pAttributes ) )
          rc = 1;
        else
          rc = -1;
      }
      else {
        rc = 2;
      }
      break;
    case TCODE_RHINOIO_OBJECT_NURBS_CURVE:
      // old Rhino I/O toolkit nurbs curve
      if ( 0 != (ON::mesh_object & object_filter) ) {
        if ( ReadV1_TCODE_RHINOIO_OBJECT_NURBS_CURVE( ppObject, pAttributes ) )
          rc = 1;
        else
          rc = -1;
      }
      else {
        rc = 2;
      }
      break;
    case TCODE_RHINOIO_OBJECT_NURBS_SURFACE:
      // old Rhino I/O toolkit nurbs surface
      if ( 0 != (ON::mesh_object & object_filter) ) {
        if ( ReadV1_TCODE_RHINOIO_OBJECT_NURBS_SURFACE( ppObject, pAttributes ) )
          rc = 1;
        else
          rc = -1;
      }
      else {
        rc = 2;
      }
      break;
    case TCODE_RHINOIO_OBJECT_BREP:
      // old Rhino I/O toolkit nurbs brep
      if ( 0 != (ON::mesh_object & object_filter) ) {
        if ( ReadV1_TCODE_RHINOIO_OBJECT_BREP( ppObject, pAttributes ) )
          rc = 1;
        else
          rc = -1;
      }
      else {
        rc = 2;
      }
      break;
    }
    if (!EndRead3dmChunk() )
      break;
    if ( rc == 1 || rc == -1 )
      break;
  }
  return rc;
}


int
ON_BinaryArchive::Read3dmObject( 
  ON_Object** ppObject,                // object is returned here
  ON_3dmObjectAttributes* pAttributes, // optional - object attributes 
  unsigned int object_filter           // optional filter made by or-ing object_type bits
  )
{
  // returns -1: failure
  //          0: end of geometry table
  //          1: success
  //          2: skipped filtered objects
  //          3: skipped new object (object's class UUID wasn't found in class list)
  BOOL rc = -1;
  if ( pAttributes )
    pAttributes->Default();
  if ( !ppObject )
    return 0;
  *ppObject = 0;
  if ( !object_filter ) // default filter (0) reads every object
    object_filter = 0xFFFFFFFF;

  if ( m_3dm_version == 1 ) {
    rc = Read3dmV1Object(ppObject,pAttributes,object_filter);
  }
  else {
    unsigned int tcode;
    int value;
    if ( BeginRead3dmChunk( &tcode, &value ) ) {
      if ( tcode == TCODE_OBJECT_RECORD ) {
        if (BeginRead3dmChunk( &tcode, &value )) {
          if ( tcode != TCODE_OBJECT_RECORD_TYPE ) {
            rc = -1;
            ON_ERROR("ON_BinaryArchive::Read3dmObject() - missing TCODE_OBJECT_RECORD_TYPE chunk.");
          }
          else if ( value && 0 == (value|object_filter) )
            rc = 2; // skip reading this object
          else
            rc = 1; // need to read this object

          if ( !EndRead3dmChunk() )
            rc = -1;

          switch(ReadObject(ppObject))
          {
          case 1:
            rc = 1; // successfully read this object
            break;
          case 3:
            rc = 3; // skipped object - assume it's just a newer object than this code reads
            break;
          default:
            rc = -1; // serious failure
            break;
          }
        }
        else
          rc = -1;
      }
      else if ( tcode != TCODE_ENDOFTABLE ) {
        ON_ERROR("ON_BinaryArchive::Read3dmLight() - corrupt light table");
        rc = -1;
      }
      else
        rc = 0;

      while(rc==1) {
        if (!BeginRead3dmChunk( &tcode, &value )) {
          rc = -1;
          break;
        }
        if ( tcode == TCODE_OBJECT_RECORD_ATTRIBUTES && pAttributes ) {
          if ( !pAttributes->Read( *this ) )
            rc = -1;
        }
        if ( !EndRead3dmChunk() ) {
          rc = -1;
        }
        if ( tcode == TCODE_OBJECT_RECORD_END )
          break;
      }

      if ( !EndRead3dmChunk() )
        rc = -1;
    }
  }

  return rc;
}

BOOL ON_BinaryArchive::EndRead3dmObjectTable()
{
  return EndRead3dmTable( TCODE_OBJECT_TABLE );
}


BOOL ON_BinaryArchive::BeginWrite3dmUserTable( const ON_UUID& usertable_uuid )
{
  if ( m_active_table != no_active_table ) {
    ON_ERROR("ON_BinaryArchive::BeginWrite3dmUserTable() - m_active_table != no_active_table");
  }
  if ( !ON_UuidCompare( &ON_nil_uuid, &usertable_uuid ) ) {
    ON_ERROR("ON_BinaryArchive::BeginWrite3dmUserTable() - nil usertable_uuid not permitted.");
    return FALSE;
  }
  BOOL rc = BeginWrite3dmTable( TCODE_USER_TABLE);
  if (rc) {
    rc = BeginWrite3dmChunk( TCODE_USER_TABLE_UUID, 0 );
    if (rc) {
      rc = WriteUuid( usertable_uuid );
      if ( !EndWrite3dmChunk() )
        rc = FALSE;
    }
    if (rc) {
      rc = BeginWrite3dmChunk( TCODE_USER_RECORD, 0 );
    }
    if ( !rc ) {
      EndWrite3dmTable( TCODE_USER_TABLE);
    }
  }
  return rc;
}

BOOL ON_BinaryArchive::Write3dmAnonymousUserTable( const ON_3dmGoo& goo )
{
  BOOL rc = FALSE;
  const ON_3DM_CHUNK* c = m_chunk.Last();
  if ( !c || c->m_typecode != TCODE_USER_RECORD ) {
    ON_ERROR("ON_BinaryArchive::Write3dmAnonymousUserTable() - active chunk not a TCODE_USER_RECORD.");
    rc = FALSE;
  }
  else if ( goo.m_typecode != TCODE_USER_RECORD ) {
    ON_ERROR("ON_BinaryArchive::Write3dmAnonymousUserTable() - goo chunk not a TCODE_USER_RECORD.");
    rc = FALSE;
  }
  else {
    rc = ( goo.m_value > 0 ) ? WriteByte( goo.m_value, goo.m_goo ) : TRUE;
  }
  return rc;
}


BOOL ON_BinaryArchive::EndWrite3dmUserTable()
{
  BOOL rc = FALSE;
  const ON_3DM_CHUNK* c = m_chunk.Last();
  if ( c && c->m_typecode == TCODE_USER_RECORD ) {
    rc = EndWrite3dmChunk();
  }
  else {
    ON_ERROR("ON_BinaryArchive::EndWrite3dmUserTable() - not in a TCODE_USER_RECORD chunk.");
    rc = FALSE;
  }
  if ( !EndWrite3dmTable(TCODE_USER_TABLE) )
    rc = FALSE;
  return rc;
}

BOOL ON_BinaryArchive::BeginRead3dmUserTable( ON_UUID& usertable_uuid )
{
  if ( m_3dm_version == 1 )
    return FALSE;
  BOOL rc = BeginRead3dmTable( TCODE_USER_TABLE );
  if ( rc ) {
    // read table id
    unsigned int tcode = !TCODE_USER_TABLE_UUID;
    int value = 0;
    if (rc) rc = BeginRead3dmChunk( &tcode, &value );
    if (rc) {
      if ( tcode != TCODE_USER_TABLE_UUID ) {
        ON_ERROR("ON_BinaryArchive::BeginRead3dmUserTable() - missing user table UUID");
        rc = FALSE;
      }
      else {
        rc = ReadUuid( usertable_uuid );
      }
      if ( !EndRead3dmChunk() )
        rc = FALSE;
    }

    tcode = !TCODE_USER_RECORD;
    value = 0;
    if (rc) rc = BeginRead3dmChunk( &tcode, &value );
    if (rc) {
      if ( tcode != TCODE_USER_RECORD ) {
        ON_ERROR("ON_BinaryArchive::BeginRead3dmUserTable() - missing user table TCODE_USER_RECORD chunk.");
        EndRead3dmChunk();
        rc = FALSE;
      }
    }

    if (!rc)
      EndRead3dmTable(TCODE_USER_TABLE);
  }
  return rc;
}

BOOL ON_BinaryArchive::Read3dmAnonymousUserTable( ON_3dmGoo& goo )
{
  BOOL rc = Read3dmGoo( goo );
  if (rc && goo.m_typecode != TCODE_USER_RECORD ) {
    ON_ERROR("ON_BinaryArchive::Read3dmAnonymousUserTable() do not read a TCODE_USER_RECORD chunk.");
    rc = FALSE;
  }
  return rc;
}

BOOL ON_BinaryArchive::EndRead3dmUserTable()
{
  if ( m_chunk.Count() != 2 ) {
    ON_ERROR("ON_BinaryArchive::EndRead3dmUserTable() m_chunk.Count() != 2");
    return FALSE;
  }
  const ON_3DM_CHUNK* c = m_chunk.Last();
  if ( c->m_typecode != TCODE_USER_RECORD ) {
    ON_ERROR("ON_BinaryArchive::EndRead3dmTable() m_chunk.Last()->typecode != TCODE_USER_RECORD");
    return FALSE;
  }
  BOOL rc = EndRead3dmChunk(); // end of TCODE_USER_RECORD chunk

  if (rc) {
    // end of table chunk
    unsigned int tcode = !TCODE_ENDOFTABLE;
    int value = 1;
    rc = BeginRead3dmChunk( &tcode, &value );
    if ( rc ) {
      if ( tcode != TCODE_ENDOFTABLE ) {
        ON_ERROR("ON_BinaryArchive::EndRead3dmTable() missing TCODE_ENDOFTABLE marker.");
      }
      if ( !EndRead3dmChunk() )
        rc = FALSE;
    }
  }

  if ( !EndRead3dmTable(TCODE_USER_TABLE) )
    rc = FALSE;
  return rc;
}

BOOL ON_BinaryArchive::Write3dmEndMark()
{
  Flush();
  if ( m_chunk.Count() != 0 ) {
    ON_ERROR( "ON_BinaryArchive::WriteEndMark() called with unfinished chunks.\n" );
    return FALSE;
  }
  int length = CurrentPosition(); // since no chunks are unfinished, everything
                                  // has been committed to disk in either
                                  // write mode.
  BOOL rc = BeginWrite3dmChunk( TCODE_ENDOFFILE, 0 );
  if ( rc ) {
    rc = WriteInt( length+12 );
    if ( !EndWrite3dmChunk() )
      rc = FALSE;
  }
  Flush();

  return rc;
}

BOOL ON_BinaryArchive::Read3dmEndMark( int* file_length )
{
  unsigned int tcode=0;
  int value=0;
  if ( file_length )
    *file_length = 0;
  m_3dm_v1_suppress_error_message |= 0x0001; // disable v1 ReadByte() error message at EOF
  BOOL rc = PeekAt3dmChunkType(&tcode,&value);
  m_3dm_v1_suppress_error_message &= 0xFFFE; // enable v1 ReadByte() error message at EOF
  if (rc)
  {
    if ( tcode == TCODE_ENDOFFILE ) {
      rc = BeginRead3dmChunk(&tcode,&value);
      if (rc) {
        int i = 0;
        rc = ReadInt(&i);
        if ( rc && file_length )
          *file_length = i;
        if ( !EndRead3dmChunk() )
          rc = FALSE;
      }
    }
  }
  return rc;
}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

ON::endian
ON_BinaryArchive::Endian() const // endian-ness of cpu
{
  return m_endian;
}

ON::archive_mode
ON_BinaryArchive::Mode() const // endian-ness of cpu
{
  return m_mode;
}

void ON_BinaryArchive::UpdateCRC( size_t count, const void* p )
{
  if ( m_bDoChunkCRC ) {
    ON_3DM_CHUNK* c = m_chunk.Last();
    if (c) {
      if ( c->m_do_crc16 )
        c->m_crc16 = ON_CRC16( c->m_crc16, count, p ); // version 1 files had 16 bit CRC
      if ( c->m_do_crc32 )
        c->m_crc32 = ON_CRC32( c->m_crc32, count, p ); // version 2 files have 32 bit CRC
    }
  }
}

int ON_BinaryArchive::BadCRCCount() const
{
  return m_bad_CRC_count;
}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

ON_BinaryFile::ON_BinaryFile( ON::archive_mode mode ) 
              : ON_BinaryArchive( mode ), 
                m_fp(0),
                m_memory_buffer_capacity(0),
                m_memory_buffer_size(0),
                m_memory_buffer_ptr(0),
                m_memory_buffer(0)
{}

ON_BinaryFile::ON_BinaryFile( ON::archive_mode mode, FILE* fp ) 
              : ON_BinaryArchive( mode ), 
                m_fp(fp),
                m_memory_buffer_capacity(0),
                m_memory_buffer_size(0),
                m_memory_buffer_ptr(0),
                m_memory_buffer(0)
{}

ON_BinaryFile::~ON_BinaryFile()
{
  EnableMemoryBuffer(0);
}

BOOL
ON_BinaryArchive::ReadByte( size_t count, void* p )
{
  BOOL rc = FALSE;
  if ( count > 0 ) {
    if ( !ReadMode() ) {
      ON_ERROR("ON_BinaryArchive::ReadByte() ReadMode() is FALSE.");
    }
    else if ( p ) {
      size_t readcount = (size_t)Read( count, p );
      if ( readcount == count ) {
        UpdateCRC( count, p );
        rc = TRUE;  
      }
      else {
        if (   readcount != 0 
            || count != 4
            || m_3dm_version != 1 
            || (m_3dm_v1_suppress_error_message&0x01) == 0 ) {
          // when reading v1 files, there are some situations where
          // it is reasonable to attempt to read 4 bytes at the end
          // of a file.  The above test prevents making a call
          // to ON_ERROR() in these situations.
          ON_ERROR("ON_BinaryArchive::ReadByte() Read() failed.");
        }
      }
    }
    else {
      ON_ERROR("ON_BinaryArchive::ReadByte() NULL file or buffer.");
    }
  }
  else {
    rc = TRUE;
  }
  return rc;
}

BOOL
ON_BinaryArchive::WriteByte( size_t count, const void* p )
{
  BOOL rc = FALSE;
  if ( count > 0 ) {
    if ( !WriteMode() ) {
      ON_ERROR("ON_BinaryArchive::WriteByte() WriteMode() is FALSE.");
    }
    else if ( p ) {
      size_t writecount = (size_t)Write( count, p );
      if ( writecount == count ) {
        UpdateCRC( count, p );
        rc = TRUE;      
      }
      else {
        ON_ERROR("ON_BinaryArchive::WriteByte() fwrite() failed.");
      }
    }
    else {
      ON_ERROR("ON_BinaryArchive::WriteByte() NULL file or buffer.");
    }
  }
  else {
    rc = TRUE;
  }
  return rc;
}

BOOL
ON_BinaryArchive::SetArchive3dmVersion(int v)
{
  BOOL rc = FALSE;
  if ( v == 1 || v == 2 ) {
    m_3dm_version = v;
    rc = TRUE;
  }
  return rc;
}

void 
ON_BinaryFile::EnableMemoryBuffer(
       int buffer_capacity
       )
{
  if ( buffer_capacity > 0 && !m_memory_buffer) {
    m_memory_buffer = (unsigned char*)onmalloc(buffer_capacity);
    if ( m_memory_buffer ) {
      m_memory_buffer_capacity = buffer_capacity;
      m_memory_buffer_size = 0;
      m_memory_buffer_ptr = 0;
    }
  }
  else {
    if ( buffer_capacity == 0 && m_memory_buffer ) {
      Flush();
      onfree(m_memory_buffer);
    }
    m_memory_buffer = 0;
    m_memory_buffer_capacity = 0;
    m_memory_buffer_size = 0;
    m_memory_buffer_ptr = 0;
  }
}


BOOL 
ON_BinaryFile::Read( size_t count, void* p )
{
  return (m_fp) ? fread( p, 1, count, m_fp ) : 0;
}

BOOL 
ON_BinaryFile::Write( size_t count, const void* p )
{
  int rc = 0;
  if ( m_fp ) {
    if ( m_memory_buffer ) {
      if ( ((int)count)+m_memory_buffer_ptr >= m_memory_buffer_capacity ) {
        if ( !Flush() ) // flush existing memory buffer to disk
          return 0;
        rc = fwrite( p, 1, count, m_fp ); // write directly to disk
      }
      else {
        // copy to the memory buffer
        memcpy( m_memory_buffer+m_memory_buffer_ptr, p, count );
        m_memory_buffer_ptr += count;
        if ( m_memory_buffer_ptr > m_memory_buffer_size )
          m_memory_buffer_size = m_memory_buffer_ptr;
        rc = count;
      }      
    }
    else
    {
      rc = fwrite( p, 1, count, m_fp );
    }
  }
  return rc;
}

BOOL 
ON_BinaryFile::Flush()
{
  BOOL rc = TRUE;
  if ( m_fp ) {
    if ( m_memory_buffer && m_memory_buffer_size > 0 ) {
      rc = fwrite( m_memory_buffer, 1, m_memory_buffer_size, m_fp );   
      rc = (rc == m_memory_buffer_size) ? TRUE : FALSE;
      if ( rc && m_memory_buffer_ptr != m_memory_buffer_size ) {
        if ( !fseek( m_fp, m_memory_buffer_size-m_memory_buffer_ptr, SEEK_CUR ) )
          rc = FALSE;
      }
      m_memory_buffer_size = 0;
      m_memory_buffer_ptr = 0;
    }
  }
  return rc;
}

int
ON_BinaryFile::CurrentPosition() const
{
  int offset = -1;
  if ( m_fp ) {
    offset = ftell(m_fp);
    if ( m_memory_buffer && m_memory_buffer_size > 0 ) {
      offset += m_memory_buffer_ptr;
    }
  }
  else {
    ON_ERROR("ON_BinaryFile::CurrentPosition() NULL file.");
  }
  return offset;
}

BOOL
ON_BinaryFile::AtEnd() const
{
  BOOL rc = TRUE;
  if ( m_fp ) {
    rc = FALSE;
    if ( ReadMode() ) {
      if ( feof( m_fp ) ) {
        rc = TRUE;
      }
      else {
        int i, buffer;
        i = fread( &buffer, 1, 1, m_fp );
        if ( feof( m_fp ) ) {
          rc = TRUE;
        }
        else {
          // back up to the byte we just read
          fseek( m_fp, -1, SEEK_CUR );
        }
      }
    }
  }
  return rc;
}

BOOL
ON_BinaryFile::SeekFromCurrentPosition( int offset )
{
  // it appears that the calls to SeekFromCurrentPosition(),
  // and consequent call to fseek(), in ON_BinaryArchive::EndWrite3DMChunk()
  // really slow down 3DM file writing on some networks.  There are
  // reports of a 100x difference in local vs network saves.
  // I could not reproduce these 100x slow saves in testing on McNeel's
  // network but we have a good quality network and a server that's
  // properly tuned.  My guess is that the slow saves are happening
  // on servers that do a poor job of cacheing because they are starved
  // for memory and/or heavily used at the time of the save.
  //
  // To speed up network saves, ON_BinaryFile can optionally use
  // it's own buffer for buffered I/O instead of relying on fwrite()
  // and the OS to handle this.
  BOOL rc = FALSE;
  if ( m_fp ) {
    if ( m_memory_buffer && 
         m_memory_buffer_ptr+offset >= 0 &&
         m_memory_buffer_ptr+offset <= m_memory_buffer_size ) {
      m_memory_buffer_ptr += offset;
      rc = TRUE;
    }
    else{
      // don't deal with memory buffer I/O if seek lands outside of
      // our current memory buffer.
      Flush();
      if ( !fseek(m_fp,offset,SEEK_CUR) )
        rc = TRUE;
      else {
        ON_ERROR("ON_BinaryFile::Seek() fseek(,SEEK_CUR) failed.");
      }
    }
  }
  return rc;
}

BOOL
ON_BinaryFile::SeekFromEnd( int offset )
{
  BOOL rc = FALSE;
  if ( m_fp ) {
    Flush(); // don't deal with memory buffer I/O in rare seek from end
    if ( !fseek(m_fp,offset,SEEK_END) )
      rc = TRUE;
    else {
      ON_ERROR("ON_BinaryFile::SeekFromEnd() fseek(,SEEK_END) failed.");
    }
  }
  return rc;
}

BOOL
ON_BinaryFile::SeekFromStart( int offset )
{
  BOOL rc = FALSE;
  if ( m_fp ) {
    Flush(); // don't deal with memory buffer I/O in rare seek from start
    if ( !fseek(m_fp,offset,SEEK_SET) )
      rc = TRUE;
    else {
      ON_ERROR("ON_BinaryFile::SeekFromStart() fseek(,SEEK_SET) failed.");
    }
  }
  return rc;
}

ON_3dmGoo::ON_3dmGoo()
        : m_typecode(0),
          m_value(0),
          m_goo(0),
          m_next_goo(0),
          m_prev_goo(0)
{}

ON_3dmGoo::~ON_3dmGoo()
{
  if ( m_prev_goo )
    m_prev_goo->m_next_goo = m_next_goo;
  if ( m_next_goo )
    m_next_goo->m_prev_goo = m_prev_goo;
  if ( m_goo ) {
    onfree(m_goo);
    m_goo = 0;
  }
}

ON_3dmGoo::ON_3dmGoo( const ON_3dmGoo& src )
        : m_typecode(0),
          m_value(0),
          m_goo(0),
          m_next_goo(0),
          m_prev_goo(0)
{
  *this = src;
}

ON_3dmGoo& ON_3dmGoo::operator=( const ON_3dmGoo& src )
{
  if ( this != &src ) {
    if ( m_goo ) {
      onfree(m_goo);
    }
    m_typecode = src.m_typecode;
    m_value = src.m_value;
    m_goo = (m_value > 0 && src.m_goo) ? (unsigned char*)onmemdup( src.m_goo, m_value ) : 0;
  }
  return *this;
}

void ON_3dmGoo::Dump(ON_TextLog& dump) const
{
  dump.Print("typecode = %08x value = %d\n",m_typecode,m_value);
}

