/* $Header: /src2/opennurbs/opennurbs_bezier.h 8     6/28/01 8:17a Dalelear $ */
/* $NoKeywords: $ */
/*
//
// Copyright (c) 1993-2001 Robert McNeel & Associates. All rights reserved.
// Rhinoceros is a registered trademark of Robert McNeel & Assoicates.
//
// THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
// ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
// MERCHANTABILITY ARE HEREBY DISCLAIMED.
//				
// For complete openNURBS copyright information see <http://www.opennurbs.org>.
//
////////////////////////////////////////////////////////////////
*/


#if !defined(OPENNURBS_BEZIER_INC_)
#define OPENNURBS_BEZIER_INC_

class ON_PolynomialCurve;
class ON_PolynomialSurface;
class ON_BezierCurve;
class ON_BezierSurface;
class ON_TextLog;
class ON_NurbsCurve;
class ON_NurbsSurface;

class ON_CLASS ON_PolynomialCurve
{
public:
  ON_PolynomialCurve();
  ON_PolynomialCurve(
    int,  // dim,
    BOOL, // TRUE if rational
    int   // order
    );
  ~ON_PolynomialCurve();
  ON_PolynomialCurve(const ON_PolynomialCurve&);
  ON_PolynomialCurve(const ON_BezierCurve&);
  ON_PolynomialCurve& operator=(const ON_PolynomialCurve&);
  ON_PolynomialCurve& operator=(const ON_BezierCurve&);

  BOOL Create(
    int,  // dim,
    BOOL, // TRUE if rational
    int   // order
    );
  void Destroy();

  BOOL Evaluate( // returns FALSE if unable to evaluate
         double,         // evaluation parameter
         int,            // number of derivatives (>=0)
         int,            // array stride (>=Dimension())
         double*         // array of length stride*(ndir+1)
         ) const;

  int m_dim;           // 1,2, or 3
  BOOL m_is_rat;       // TRUE or FALSE;
  int m_order;         // order = degree+1 >= 1;
  ON_4dPointArray m_cv; // coefficients ( m_C.Length() = order of monomial )
                       // 
  ON_Interval m_domain;
};

class ON_CLASS ON_PolynomialSurface
{
public:
  ON_PolynomialSurface();
  ON_PolynomialSurface(
    int,  // dim,
    BOOL, // TRUE if rational
    int,  // "u" order
    int   // "v" order
    );
  ~ON_PolynomialSurface();
  ON_PolynomialSurface(const ON_PolynomialSurface&);
  ON_PolynomialSurface(const ON_BezierSurface&);
  ON_PolynomialSurface& operator=(const ON_PolynomialSurface&);
  ON_PolynomialSurface& operator=(const ON_BezierSurface&);

  BOOL Create(
    int,  // dim,
    BOOL, // TRUE if rational
    int,  // "u" order
    int   // "v" order
    );
  void Destroy();

  BOOL Evaluate( // returns FALSE if unable to evaluate
         double, double, // evaluation parameter
         int,            // number of derivatives (>=0)
         int,            // array stride (>=Dimension())
         double*         // array of length stride*(ndir+1)*(ndir+2)/2
         ) const;

  int m_dim;     // 1,2, or 3
  BOOL m_is_rat; // TRUE or FALSE;
  int m_order[2];
  ON_4dPointArray m_cv; // coefficients ( m_C.Length() = m_order[0]*m_order[1]
                        // coefficient of s^m*t^n = m_cv[m_order[1]*m+n]
  ON_Interval m_domain[2];
};

class ON_CLASS ON_BezierCurve
{
public:
  ON_BezierCurve();
  ON_BezierCurve(
    int,  // dim,
    BOOL, // TRUE if rational
    int   // order
    );
  ~ON_BezierCurve();
  ON_BezierCurve(const ON_BezierCurve&);
  ON_BezierCurve(const ON_PolynomialCurve&);
  ON_BezierCurve& operator=(const ON_BezierCurve&);
  ON_BezierCurve& operator=(const ON_PolynomialCurve&);

  BOOL IsValid() const;
  void Dump( ON_TextLog& ) const; // for debugging
  int Dimension() const;

  BOOL Create(
    int,  // dim,
    BOOL, // TRUE if rational
    int   // order
    );
  void Destroy();
  void EmergencyDestroy(); // call if memory used by ON_NurbsCurve becomes invalid

  BOOL Loft( const ON_3dPointArray& );

  BOOL Loft(
    int,           // pt_dim,
    int,           // pt_count,
    int,           // pt_stride
    const double*, // pt
    int,           // t_stride
    const double*  // t parameters 
    );

  BOOL GetBBox( // returns TRUE if successful
         double*,    // minimum
         double*,    // maximum
         BOOL = FALSE  // TRUE means grow box
         ) const;

  BOOL GetBoundingBox( // returns TRUE if successful
         ON_BoundingBox&, // minimum
         BOOL = FALSE  // TRUE means grow box
         ) const;

  ON_BoundingBox BoundingBox() const;

  BOOL Transform( 
         const ON_Xform&
         );

  ON_Interval Domain() const;

  BOOL Reverse();       // reverse parameterizatrion
                        // Domain changes from [a,b] to [-b,-a]
  BOOL Evaluate( // returns FALSE if unable to evaluate
         double,         // evaluation parameter
         int,            // number of derivatives (>=0)
         int,            // array stride (>=Dimension())
         double*         // array of length stride*(ndir+1)
         ) const;

  BOOL GetNurbForm( ON_NurbsCurve& ) const;

  BOOL IsRational() const;  // TRUE if NURBS curve is rational
  
  int CVSize() const;  // number of doubles per control vertex 
               // = IsRational() ? Dim()+1 : Dim()

  int CVCount() const; // nubmer of cvs = order of bezier
  
  int Order() const;        // order = degree + 1
	
  int Degree() const;       // degree = order - 1


  double* CV(           // get pointer to control vertex
        int             // CV index ( >= 0 and < CVCount() )
        );

  const double* CV(     // get pointer to control vertex
        int             // CV index ( >= 0 and < CVCount() )
        ) const;

  double Weight(        // get value of control vertex weight
        int             // CV index ( >= 0 and < CVCount() )
        ) const;

  BOOL SetWeight(      // get value of control vertex weight
        int,            // CV index ( >= 0 and < CVCount() )
        double
        );

  BOOL SetCV(              // set a single control vertex
        int,              // CV index ( >= 0 and < CVCount() )
        ON::point_style, // style of input point
        const double*     // value of control vertex
        );

  BOOL SetCV(               // set a single control vertex
        int,               // CV index ( >= 0 and < CVCount() )
        const ON_3dPoint& // value of control vertex
                           // If NURBS is rational, weight
                           // will be set to 1.
        );

  BOOL SetCV(              // set a single control vertex
        int,              // CV index ( >= 0 and < CVCount() )
        const ON_4dPoint& // value of control vertex
                          // If NURBS is not rational, euclidean
                          // location of homogeneous point will
                          // be used.
        );

  BOOL GetCV(              // get a single control vertex
        int,              // CV index ( >= 0 and < CVCount() )
        ON::point_style, // style to use for output point
        double*           // array of length >= CVSize()
        ) const;

  BOOL GetCV(              // get a single control vertex
        int,              // CV index ( >= 0 and < CVCount() )
        ON_3dPoint&      // gets euclidean cv when NURBS is rational
        ) const;

  BOOL GetCV(              // get a single control vertex
        int,              // CV index ( >= 0 and < CVCount() )
        ON_4dPoint&      // gets homogeneous cv
        ) const;

  BOOL ZeroCVs(); // zeros control vertices and, if rational, sets weights to 1

  BOOL MakeRational();

  BOOL MakeNonRational();

  BOOL IncreaseDegree(
          int //  desired_degree
          );

  BOOL ChangeDimension(
          int //  desired_dimension
          );

  /////////////////////////////////////////////////////////////////
  // Tools for managing CV and knot memory
  BOOL ReserveCVCapacity(
    int // number of doubles to reserve
    );

  //////////
  // Trim() trins (or extends) the bezier so thatmodifies the bezier so that the result starts atremoves portions of the bezier curve outside the specified interval
  BOOL Trim( const ON_Interval& );

  //////////
  // Split() divides the Bezier curve at the specified paramter.  The parameter
  // must satisfy 0 < t < 1.  You may pass *this as one of the curves to be
  // returned.  For example, if crv were an ON_BezierCurve, then the call
  //
  //   ON_BezierCurve right_side;
  //   crv.Split( 0.5, crv, right_side );
  //
  // would split crv at the 1/2, put the left side in crv,
  // and return the right side in right_side.
  BOOL Split( 
         double, // t = splitting parameter must 0 < t < 1
         ON_BezierCurve&, // left side returned here (can pass *this)
         ON_BezierCurve&  // right side returned here (can pass *this)
         ) const;

  //////////
  // returns the length of the control polygon
  double ControlPolygonLength() const;

  /////////////////////////////////////////////////////////////////
  // Implementation
public:
  // NOTE: These members are left "public" so that expert users may efficiently
  //       create bezier curves using the default constructor and borrow the
  //       knot and CV arrays from their native NURBS representation.
  //       No technical support will be provided for users who access these
  //       members directly.  If you can't get your stuff to work, then use
  //       the constructor with the arguments and the SetKnot() and SetCV()
  //       functions to fill in the arrays.


  int     m_dim;
  BOOL    m_is_rat;
  int     m_order;
  int     m_cv_stride;
  int     m_cv_capacity;
  double* m_cv;
};

class ON_CLASS ON_BezierSurface
{
public:
  ON_BezierSurface();
  ON_BezierSurface(
    int,  // dim,
    BOOL, // TRUE if rational
    int,  // "u" order
    int   // "v" order
    );
  ~ON_BezierSurface();
  ON_BezierSurface(const ON_BezierSurface&);
  ON_BezierSurface(const ON_PolynomialSurface&);
  ON_BezierSurface& operator=(const ON_BezierSurface&);
  ON_BezierSurface& operator=(const ON_PolynomialSurface&);

  BOOL IsValid() const;
  void Dump( ON_TextLog& ) const; // for debugging
  int Dimension() const;

  BOOL Create(
    int,  // dim,
    BOOL, // TRUE if rational
    int,  // "u" order
    int   // "v" order
    );
  void Destroy();
  void EmergencyDestroy(); // call if memory used by ON_NurbsCurve becomes invalid

  BOOL Loft( const ON_ClassArray<ON_BezierCurve>& );
  BOOL Loft( const ON_SimpleArray<ON_BezierCurve*>& );

  BOOL GetBBox( // returns TRUE if successful
         double*,    // minimum
         double*,    // maximum
         BOOL = FALSE  // TRUE means grow box
         ) const;

  BOOL Transform( 
         const ON_Xform&
         );

  ON_Interval Domain(
    int // 0 = "u" domain, 1 = "v" domain
    ) const;

  BOOL Reverse( int );  // reverse parameterizatrion
                        // Domain changes from [a,b] to [-b,-a]
  
  BOOL Transpose(); // transpose surface parameterization (swap "s" and "t")

  BOOL Evaluate( // returns FALSE if unable to evaluate
         double, double, // evaluation parameter
         int,            // number of derivatives (>=0)
         int,            // array stride (>=Dimension())
         double*         // array of length stride*(ndir+1)*(ndir+2)/2
         ) const;

  BOOL GetNurbForm( ON_NurbsSurface& ) const;

  BOOL IsRational() const;  // TRUE if NURBS curve is rational
  
  int CVSize() const;       // number of doubles per control vertex 
                // = IsRational() ? Dim()+1 : Dim()
  
  int Order(        // order = degree + 1
        int // dir
        ) const;
	
  int Degree(       // degree = order - 1
        int // dir
        ) const;

  double* CV(           // get pointer to control vertex
        int,int          // CV index ( >= 0 and < CVCount() )
        );

  const double* CV(     // get pointer to control vertex
        int,int          // CV index ( >= 0 and < CVCount() )
        ) const;

  double Weight(        // get value of control vertex weight
        int,int          // CV index ( >= 0 and < CVCount() )
        ) const;

  BOOL SetWeight(      // get value of control vertex weight
        int,int,         // CV index ( >= 0 and < CVCount() )
        double
        );

  BOOL SetCV(              // set a single control vertex
        int,int,         // CV index ( >= 0 and < CVCount() )
        ON::point_style, // style of input point
        const double*     // value of control vertex
        );

  BOOL SetCV(               // set a single control vertex
        int,int,         // CV index ( >= 0 and < CVCount() )
        const ON_3dPoint& // value of control vertex
                           // If NURBS is rational, weight
                           // will be set to 1.
        );

  BOOL SetCV(              // set a single control vertex
        int,int,         // CV index ( >= 0 and < CVCount() )
        const ON_4dPoint& // value of control vertex
                          // If NURBS is not rational, euclidean
                          // location of homogeneous point will
                          // be used.
        );

  BOOL GetCV(              // get a single control vertex
        int,int,          // CV index ( >= 0 and < CVCount() )
        ON::point_style, // style to use for output point
        double*           // array of length >= CVSize()
        ) const;

  BOOL GetCV(              // get a single control vertex
        int,int,         // CV index ( >= 0 and < CVCount() )
        ON_3dPoint&      // gets euclidean cv when NURBS is rational
        ) const;

  BOOL GetCV(              // get a single control vertex
        int,int,         // CV index ( >= 0 and < CVCount() )
        ON_4dPoint&      // gets homogeneous cv
        ) const;

  BOOL ZeroCVs(); // zeros control vertices and, if rational, sets weights to 1

  BOOL MakeRational();

  BOOL MakeNonRational();

  /////////////////////////////////////////////////////////////////
  // Tools for managing CV and knot memory
  BOOL ReserveCVCapacity(
    int // number of doubles to reserve
    );

  /////////////////////////////////////////////////////////////////
  // Implementation
public:
  // NOTE: These members are left "public" so that expert users may efficiently
  //       create bezier curves using the default constructor and borrow the
  //       knot and CV arrays from their native NURBS representation.
  //       No technical support will be provided for users who access these
  //       members directly.  If you can't get your stuff to work, then use
  //       the constructor with the arguments and the SetKnot() and SetCV()
  //       functions to fill in the arrays.


  int     m_dim;
  BOOL    m_is_rat;
  int     m_order[2];
  int     m_cv_stride[2];
  int     m_cv_capacity;
  double* m_cv;
  //ON_Interval m_domain[2];
};

#if defined(ON_DLL_EXPORTS) || defined(ON_DLL_IMPORTS)
// This stuff is here because of a limitation in the way Microsoft
// handles templates and DLLs.  See Microsoft's knowledge base 
// article ID Q168958 for details.
#pragma warning( push )
#pragma warning( disable : 4231 )
ON_TEMPLATE template class ON_CLASS ON_ClassArray<ON_BezierCurve>;
ON_TEMPLATE template class ON_CLASS ON_SimpleArray<ON_BezierCurve*>;
ON_TEMPLATE template class ON_CLASS ON_ClassArray<ON_BezierSurface>;
ON_TEMPLATE template class ON_CLASS ON_SimpleArray<ON_BezierSurface*>;
#pragma warning( pop )
#endif

#endif
