/* $Header: /src2/opennurbs/opennurbs_bitmap.cpp 14    1/03/01 4:59p Dalelear $ */
/* $NoKeywords: $ */
/*
//
// Copyright (c) 1993-2001 Robert McNeel & Associates. All rights reserved.
// Rhinoceros is a registered trademark of Robert McNeel & Assoicates.
//
// THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
// ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
// MERCHANTABILITY ARE HEREBY DISCLAIMED.
//				
// For complete openNURBS copyright information see <http://www.opennurbs.org>.
//
////////////////////////////////////////////////////////////////
*/

#include "opennurbs.h"

ON_VIRTUAL_OBJECT_IMPLEMENT( ON_Bitmap, ON_Object, "390465E9-3721-11d4-800B-0010830122F0");
ON_OBJECT_IMPLEMENT( ON_OpenGLBitmap, ON_Bitmap,   "390465EA-3721-11d4-800B-0010830122F0");
ON_OBJECT_IMPLEMENT( ON_WindowsBitmap, ON_Bitmap,  "390465EB-3721-11d4-800B-0010830122F0");

ON_Bitmap::ON_Bitmap()
{}

ON_Bitmap::ON_Bitmap(const ON_Bitmap& src) : ON_Object(src)
{
  m_name = src.m_name;
}

ON_Bitmap& ON_Bitmap::operator=(const ON_Bitmap& src)
{
  if ( this != &src ) {
    ON_Object::operator=(src);
    m_name = src.m_name;
  }
  return *this;
}

ON_Bitmap::~ON_Bitmap()
{}

void ON_Bitmap::Dump( ON_TextLog& dump ) const
{
  dump.Print("width = %d pixels\n",Width());
  dump.Print("height = %d pixels\n",Height());
  dump.Print("bits per pixel = %d\n",BitsPerPixel());
  dump.Print("size of image = %d bytes\n",SizeofImage());
}


static int ON_WindowsBitmap_PaletteColorCount( int bmiHeader_biClrUsed, int bmiHeader_biBitCount )
{
  int color_count = 0;
  
  if ( bmiHeader_biClrUsed )
    color_count = bmiHeader_biClrUsed;
  else {  
    switch( bmiHeader_biBitCount )
    {
    case 1:
      color_count = 2;
    case 4:
      color_count = 16;
    case 8:
      color_count = 256;
    default:
      color_count = 0;
    }
  }  
  return color_count;
}

static int ON_WindowsBitmap_SizeofPalette( int bmiHeader_biClrUsed, int bmiHeader_biBitCount )
{
  return ( ON_WindowsBitmap_PaletteColorCount(bmiHeader_biClrUsed,bmiHeader_biBitCount) * sizeof(ON_WindowsRGBQUAD) );
}

ON_WindowsBitmap::ON_WindowsBitmap() : m_bmi(0), m_bits(0), m_bFreeBMI(0)
{}

ON_WindowsBitmap::ON_WindowsBitmap( const ON_WindowsBitmap& src ) 
                : m_bmi(0), m_bits(0), m_bFreeBMI(0)
{
  *this = src;
}

void ON_WindowsBitmap::Destroy()
{
  if (m_bmi) {
    if ( m_bFreeBMI )
      Realloc(m_bmi,0);
    m_bmi=0;
  }
  m_bits = 0;
  m_bFreeBMI = FALSE;
}

BOOL ON_WindowsBitmap::IsValid() const
{
  return ( m_bmi != NULL && m_bits != NULL && Width() > 0 && Height() > 0) ? TRUE : FALSE;
}


BOOL ON_WindowsBitmap::Create( 
       int width,
       int height,
       int bits_per_pixel // 1, 2, 4, 8, 16, 24, or 32
       )
{
  BOOL rc = FALSE;
  Destroy();
  if ( width < 1 || height < 1 )
    return FALSE;
  if (   bits_per_pixel != 1 && bits_per_pixel !=  2 && bits_per_pixel !=  4 
      && bits_per_pixel != 8 && bits_per_pixel != 16 && bits_per_pixel != 24 
      && bits_per_pixel != 32 )
    return FALSE;
  ON_WindowsBITMAPINFOHEADER bh;
  memset( &bh, 0, sizeof(bh) );
  bh.biSize = sizeof(ON_WindowsBITMAPINFO);
  bh.biWidth = width;
  bh.biHeight = height;
  bh.biPlanes = 1;
  bh.biBitCount = bits_per_pixel;
  bh.biCompression = 0; // BI_RGB
  const int sizeof_scan = (((bits_per_pixel*width) + 31) / 32) * 4;
  bh.biSizeImage = height*sizeof_scan;
  bh.biXPelsPerMeter = 0;
  bh.biYPelsPerMeter = 0;
  bh.biClrUsed = 0;
  bh.biClrImportant = 0;

  int palette_color_count = 0;
  switch ( bits_per_pixel ) {
  case 1:
    palette_color_count = 2;
    break;
  case 4:
    palette_color_count = 16;
    break;
  case 8:
    palette_color_count = 256;
    break;
  }
  const int sizeof_palette = palette_color_count*sizeof(ON_WindowsRGBQUAD);

  m_bmi = Realloc( 0, sizeof(*m_bmi)+sizeof_palette+bh.biSizeImage );
  if ( m_bmi ) {
    m_bmi->bmiHeader = bh;
    m_bits = (unsigned char*)&m_bmi->bmiColors[palette_color_count];

    // default palette is gray scale
    const int rgb_delta = 256/palette_color_count;
    int i, rgb;
    for ( i = 0, rgb = 0; i < palette_color_count; i++, rgb += rgb_delta ) {
      if ( rgb >= 256 ) rgb = 255;
      m_bmi->bmiColors[i].rgbBlue  = (unsigned char)rgb;
      m_bmi->bmiColors[i].rgbGreen = (unsigned char)rgb;
      m_bmi->bmiColors[i].rgbRed   = (unsigned char)rgb;
      m_bmi->bmiColors[i].rgbReserved = 0;
    }
    rc = TRUE;
  }
  return rc;
}


ON_WindowsBitmap::~ON_WindowsBitmap()
{
  Destroy();
}

ON_WindowsBitmap& ON_WindowsBitmap::operator=( const ON_WindowsBitmap& src )
{
  if ( this != &src ) {
    Destroy();
    ON_Bitmap::operator=(src);
    if ( src.m_bmi ) {
      const int sizeof_header  = sizeof(ON_WindowsBITMAPINFOHEADER);
      const int sizeof_palette = src.SizeofPalette();
      const int sizeof_image   = src.SizeofImage();
      const int sz = sizeof_header + sizeof_palette + sizeof_image;
      m_bmi = Realloc( 0, sz );
      if ( m_bmi ) {
        m_bFreeBMI = TRUE;
        memcpy( m_bmi, src.m_bmi, sizeof_header+sizeof_palette );
        if ( sizeof_image > 0 ) {
          m_bits = (unsigned char*)&m_bmi->bmiColors[PaletteColorCount()];
          if ( src.m_bits )
            memcpy( m_bits, src.m_bits, sizeof_image );
          else
            memset( m_bits, 0, sizeof_image );
        }
        else
          m_bits = 0;
      }
    }
  }
  return *this;
}

int ON_WindowsBitmap::Width() const
{
  return (m_bmi) ? m_bmi->bmiHeader.biWidth : 0;
}

int ON_WindowsBitmap::Height() const
{
  return (m_bmi) ? m_bmi->bmiHeader.biHeight : 0;
}

int ON_WindowsBitmap::PaletteColorCount() const
{
  return m_bmi ? ON_WindowsBitmap_PaletteColorCount(m_bmi->bmiHeader.biClrUsed, m_bmi->bmiHeader.biBitCount ) : 0;
}

int ON_WindowsBitmap::SizeofPalette() const
{
  return m_bmi ? ON_WindowsBitmap_SizeofPalette(m_bmi->bmiHeader.biClrUsed, m_bmi->bmiHeader.biBitCount ) : 0;
}

int ON_WindowsBitmap::SizeofScan() const
{
  int scan_width = 0;
  if ( m_bmi ) {
    int bitcount = m_bmi->bmiHeader.biBitCount;
    int width = Width();
    scan_width = (((bitcount*width) + 31) / 32) * 4;
  }
  return scan_width;
}

int ON_WindowsBitmap::BitsPerPixel() const
{
  return m_bmi ? m_bmi->bmiHeader.biBitCount : 0;
}

int ON_WindowsBitmap::SizeofImage() const
{
  return m_bmi ? m_bmi->bmiHeader.biSizeImage : 0;
}

unsigned char* ON_WindowsBitmap::Bits( int scan_index )
{
  const int sizeof_scan = SizeofScan();
  unsigned char* bits = m_bmi ? (unsigned char*)&m_bmi->bmiColors[PaletteColorCount()] : 0;
  if ( bits && sizeof_scan && scan_index >= 0 && scan_index < Height() ) {
    bits += (sizeof_scan*scan_index);
  }
  else {
    bits = 0;
  }
  return bits;
}

const unsigned char* ON_WindowsBitmap::Bits(int scan_index) const
{
  const int sizeof_scan = SizeofScan();
  const unsigned char* bits = m_bmi ? (const unsigned char*)&m_bmi->bmiColors[PaletteColorCount()] : 0;
  if ( bits && sizeof_scan && scan_index >= 0 && scan_index < Height() ) {
    bits += (sizeof_scan*scan_index);
  }
  else {
    bits = 0;
  }
  return bits;
}

ON_Color ON_WindowsBitmap::Pixel( int column_index, int row_index ) const
{
  return Pixel( column_index, Bits(row_index) );
}

ON_Color ON_WindowsBitmap::Pixel( int column_index, const unsigned char* scanbits ) const
{
  int r=0,g=0,b=0,a=0;

  unsigned char palindex;
  
  if ( m_bmi && column_index >= 0 && column_index <= Width() && scanbits && !m_bmi->bmiHeader.biCompression ) {
    switch( m_bmi->bmiHeader.biBitCount ) {
    case 32:
      scanbits += (column_index*4);
      b = *scanbits++;
      g = *scanbits++;
      r = *scanbits++;
      a = *scanbits;
      break;
    
    case 24:
      scanbits += (column_index*3);
      b = *scanbits++;
      g = *scanbits++;
      r = *scanbits;
      break;
    
    case 8:
      // 256 color bitmap uses palette
      palindex = scanbits[column_index];    
      b = m_bmi->bmiColors[palindex].rgbBlue;
      g = m_bmi->bmiColors[palindex].rgbGreen;
      r = m_bmi->bmiColors[palindex].rgbRed;
      a = m_bmi->bmiColors[palindex].rgbReserved;
      break;
    
    case 4:
      {
        // 16 color bitmap uses palette
        palindex = scanbits[column_index/2];
        if ( !(column_index%2) )
          palindex >>= 4;
        palindex &= 0x0F;
        b = m_bmi->bmiColors[palindex].rgbBlue;
        g = m_bmi->bmiColors[palindex].rgbGreen;
        r = m_bmi->bmiColors[palindex].rgbRed;
        a = m_bmi->bmiColors[palindex].rgbReserved;
      }
      break;
    
    case 1:
      // monochrome bitmap has 2 color palette 
      palindex = scanbits[column_index/8] >> (7-(column_index%8)) & 0x01;
      b = m_bmi->bmiColors[palindex].rgbBlue;
      g = m_bmi->bmiColors[palindex].rgbGreen;
      r = m_bmi->bmiColors[palindex].rgbRed;
      a = m_bmi->bmiColors[palindex].rgbReserved;
      break;		
    }
  }
  
  return ON_Color(r,g,b,a);
}


BOOL ON_WindowsBitmap::WriteUncompressed( ON_BinaryArchive& file ) const
{
  ON_WindowsBITMAPINFOHEADER bmiHeader;
  if ( m_bmi )
    bmiHeader = m_bmi->bmiHeader;
  else {
    memset(&bmiHeader,0,sizeof(bmiHeader));
  }
  BOOL    rc = file.WriteInt(bmiHeader.biSize);
  if (rc) rc = file.WriteInt(bmiHeader.biWidth);
  if (rc) rc = file.WriteInt(bmiHeader.biHeight);
  if (rc) rc = file.WriteShort(bmiHeader.biPlanes);
  if (rc) rc = file.WriteShort(bmiHeader.biBitCount);
  if (rc) rc = file.WriteInt(bmiHeader.biCompression);
  if (rc) rc = file.WriteInt(bmiHeader.biSizeImage);
  if (rc) rc = file.WriteInt(bmiHeader.biXPelsPerMeter);
  if (rc) rc = file.WriteInt(bmiHeader.biYPelsPerMeter);
  if (rc) rc = file.WriteInt(bmiHeader.biClrUsed);
  if (rc) rc = file.WriteInt(bmiHeader.biClrImportant);
  if ( rc ) {
    const int color_count = PaletteColorCount();
    int i;
    for (i = 0; i < color_count && rc; i++ ) {
      if (rc) rc = file.WriteChar( m_bmi->bmiColors[i].rgbBlue );
      if (rc) rc = file.WriteChar( m_bmi->bmiColors[i].rgbGreen );
      if (rc) rc = file.WriteChar( m_bmi->bmiColors[i].rgbRed );
      if (rc) rc = file.WriteChar( m_bmi->bmiColors[i].rgbReserved );
    }
    const int sizeof_image = SizeofImage();
    if ( sizeof_image > 0 && rc ) {
      if (rc) rc = file.WriteByte( sizeof_image, &m_bmi->bmiColors[color_count] );
    }
  }

  return rc;
}

BOOL ON_WindowsBitmap::ReadUncompressed( ON_BinaryArchive& file )
{
  Destroy();
  ON_WindowsBITMAPINFOHEADER bmiHeader;
  memset(&bmiHeader,0,sizeof(bmiHeader));
  BOOL    rc = file.ReadInt(&bmiHeader.biSize);
  if (rc) rc = file.ReadInt(&bmiHeader.biWidth);
  if (rc) rc = file.ReadInt(&bmiHeader.biHeight);
  if (rc) rc = file.ReadShort(&bmiHeader.biPlanes);
  if (rc) rc = file.ReadShort(&bmiHeader.biBitCount);
  if (rc) rc = file.ReadInt(&bmiHeader.biCompression);
  if (rc) rc = file.ReadInt(&bmiHeader.biSizeImage);
  if (rc) rc = file.ReadInt(&bmiHeader.biXPelsPerMeter);
  if (rc) rc = file.ReadInt(&bmiHeader.biYPelsPerMeter);
  if (rc) rc = file.ReadInt(&bmiHeader.biClrUsed);
  if (rc) rc = file.ReadInt(&bmiHeader.biClrImportant);
  if ( rc ) {
    const int sizeof_header  = sizeof(ON_WindowsBITMAPINFOHEADER);
    const int sizeof_palette = ON_WindowsBitmap_SizeofPalette(bmiHeader.biClrUsed, bmiHeader.biBitCount );
    const int sizeof_image   = bmiHeader.biSizeImage;
    const int sz = sizeof_header + sizeof_palette + sizeof_image;
    m_bmi = Realloc( 0, sz );
    if ( !m_bmi )
      rc = FALSE;
    else {
      m_bFreeBMI = TRUE;
      m_bmi->bmiHeader = bmiHeader;
      const int color_count = ON_WindowsBitmap_PaletteColorCount(bmiHeader.biClrUsed, bmiHeader.biBitCount );
      int i;
      for (i = 0; i < color_count && rc; i++ ) {
        if (rc) rc = file.ReadChar( &m_bmi->bmiColors[i].rgbBlue );
        if (rc) rc = file.ReadChar( &m_bmi->bmiColors[i].rgbGreen );
        if (rc) rc = file.ReadChar( &m_bmi->bmiColors[i].rgbRed );
        if (rc) rc = file.ReadChar( &m_bmi->bmiColors[i].rgbReserved );
      }
      if ( sizeof_image > 0 && rc ) {
        m_bits = (unsigned char*)&m_bmi->bmiColors[color_count];
        if (rc) rc = file.ReadByte( sizeof_image, &m_bmi->bmiColors[color_count] );
      }
    }
  }
  return rc;
}

BOOL ON_WindowsBitmap::Write( ON_BinaryArchive& file ) const
{
  return WriteCompressed(file);
}

BOOL ON_WindowsBitmap::Read( ON_BinaryArchive& file )
{
  BOOL rc = FALSE;
  if ( file.Archive3dmVersion() == 1 )
    rc = ReadUncompressed(file);
  else
    rc = ReadCompressed(file);
  return rc;
}

BOOL ON_WindowsBitmap::WriteCompressed( ON_BinaryArchive& file ) const
{
  //return WriteUncompressed(file);

  ON_WindowsBITMAPINFOHEADER bmiHeader;
  if ( m_bmi )
    bmiHeader = m_bmi->bmiHeader;
  else {
    memset(&bmiHeader,0,sizeof(bmiHeader));
  }
  BOOL    rc = file.WriteInt(bmiHeader.biSize);
  if (rc) rc = file.WriteInt(bmiHeader.biWidth);
  if (rc) rc = file.WriteInt(bmiHeader.biHeight);
  if (rc) rc = file.WriteShort(bmiHeader.biPlanes);
  if (rc) rc = file.WriteShort(bmiHeader.biBitCount);
  if (rc) rc = file.WriteInt(bmiHeader.biCompression);
  if (rc) rc = file.WriteInt(bmiHeader.biSizeImage);
  if (rc) rc = file.WriteInt(bmiHeader.biXPelsPerMeter);
  if (rc) rc = file.WriteInt(bmiHeader.biYPelsPerMeter);
  if (rc) rc = file.WriteInt(bmiHeader.biClrUsed);
  if (rc) rc = file.WriteInt(bmiHeader.biClrImportant);
  if ( rc ) {
    const int color_count = PaletteColorCount();
    const int sizeof_palette = color_count*sizeof(*m_bmi->bmiColors);
    const int sizeof_image = SizeofImage();
    const int sizeof_buffer = sizeof_palette + sizeof_image;
    // palette and bits are compressed in a single chunk
    rc = file.WriteCompressedBuffer( sizeof_buffer, m_bmi->bmiColors );
  }

  return rc;
}

BOOL ON_WindowsBitmap::ReadCompressed( ON_BinaryArchive& file )
{
  //return ReadUncompressed(file);

  BOOL bFailedCRC = FALSE;
  Destroy();
  ON_WindowsBITMAPINFOHEADER bmiHeader;
  memset(&bmiHeader,0,sizeof(bmiHeader));
  BOOL    rc = file.ReadInt(&bmiHeader.biSize);
  if (rc) rc = file.ReadInt(&bmiHeader.biWidth);
  if (rc) rc = file.ReadInt(&bmiHeader.biHeight);
  if (rc) rc = file.ReadShort(&bmiHeader.biPlanes);
  if (rc) rc = file.ReadShort(&bmiHeader.biBitCount);
  if (rc) rc = file.ReadInt(&bmiHeader.biCompression);
  if (rc) rc = file.ReadInt(&bmiHeader.biSizeImage);
  if (rc) rc = file.ReadInt(&bmiHeader.biXPelsPerMeter);
  if (rc) rc = file.ReadInt(&bmiHeader.biYPelsPerMeter);
  if (rc) rc = file.ReadInt(&bmiHeader.biClrUsed);
  if (rc) rc = file.ReadInt(&bmiHeader.biClrImportant);
  if ( rc ) {
    const int sizeof_header  = sizeof(ON_WindowsBITMAPINFOHEADER);
    const int sizeof_palette = ON_WindowsBitmap_SizeofPalette(bmiHeader.biClrUsed, bmiHeader.biBitCount );
    const int sizeof_image   = bmiHeader.biSizeImage;
    const int sz = sizeof_header + sizeof_palette + sizeof_image;
    m_bmi = Realloc( 0, sz );
    if ( !m_bmi )
      rc = FALSE;
    else {
      m_bFreeBMI = TRUE;
      m_bmi->bmiHeader = bmiHeader;
      const int color_count = ON_WindowsBitmap_PaletteColorCount(bmiHeader.biClrUsed, bmiHeader.biBitCount );
      int sizeof_buffer = 0;
      rc = file.ReadCompressedBufferSize( &sizeof_buffer );
      if ( rc && sizeof_buffer != color_count*((int)sizeof(*m_bmi->bmiColors)) + sizeof_image ) 
      {
        ON_ERROR("ON_WindowsBitmap::ReadCompressed() buffer size mismatch\n");
        rc = FALSE;
      }
      else if (rc) {
        m_bits = (unsigned char*)&m_bmi->bmiColors[color_count];
        // palette and bits are compressed in a single chunk
        rc = file.ReadCompressedBuffer( sizeof_buffer, m_bmi->bmiColors, &bFailedCRC );
      }
    }
  }
  return rc;
}


#if defined(ON_OS_WINDOWS) && !defined(NOGDI)

ON_WindowsBitmap::ON_WindowsBitmap( const BITMAPINFO& src )
                 : m_bmi(0), m_bits(0), m_bFreeBMI(0)
{
  *this = src;
}

static ON_WindowsBitmap_CopyHeader( const BITMAPINFOHEADER& src, ON_WindowsBITMAPINFOHEADER& dest )
{
  // src and dest memory may overlap
  ON_WindowsBITMAPINFOHEADER bmih;
  memset(&bmih,0,sizeof(bmih));
  bmih.biSize          = src.biSize;
  bmih.biWidth         = src.biWidth;
  bmih.biHeight        = src.biHeight;
  bmih.biPlanes        = src.biPlanes;
  bmih.biBitCount      = src.biBitCount;
  bmih.biCompression   = src.biCompression;
  bmih.biSizeImage     = src.biSizeImage;
  bmih.biXPelsPerMeter = src.biXPelsPerMeter;
  bmih.biYPelsPerMeter = src.biYPelsPerMeter;
  bmih.biClrUsed       = src.biClrUsed;
  bmih.biClrImportant  = src.biClrImportant;
  dest = bmih; 
}

static ON_WindowsBitmap_CopyPalette( const RGBQUAD* src, ON_WindowsRGBQUAD* dest, int color_count )
{
  // src and dest memory may overlap
  int i;
  ON_WindowsRGBQUAD q;
  for ( i = 0; i < color_count; i++ ) {
    q.rgbBlue     = src[i].rgbBlue;
    q.rgbGreen    = src[i].rgbGreen;
    q.rgbRed      = src[i].rgbRed;
    q.rgbReserved = src[i].rgbReserved;
    dest[i] = q;
  }
}

ON_WindowsBitmap& ON_WindowsBitmap::operator=( const BITMAPINFO& src )
{
  Destroy();
  if ( sizeof(BITMAPINFO) != sizeof( ON_WindowsBITMAPINFO ) ) {
    const int sizeof_header  = sizeof(ON_WindowsBITMAPINFOHEADER);
    const int sizeof_palette = ON_WindowsBitmap_SizeofPalette(src.bmiHeader.biClrUsed, src.bmiHeader.biBitCount );
    const int sizeof_image   = src.bmiHeader.biSizeImage;
    const int sz = sizeof_header + sizeof_palette + sizeof_image;
    m_bmi = Realloc( 0, sz );
    if ( m_bmi ) {
      m_bFreeBMI = TRUE;
      memset( m_bmi, 0, sz );
      ON_WindowsBitmap_CopyHeader( src.bmiHeader, m_bmi->bmiHeader );
      int color_count = ON_WindowsBitmap_PaletteColorCount(src.bmiHeader.biClrUsed, src.bmiHeader.biBitCount );
      ON_WindowsBitmap_CopyPalette( &src.bmiColors[0], &m_bmi->bmiColors[0], color_count );
      if ( sizeof_image > 0 ) {
        // copy image bits
        m_bits = (unsigned char*)&m_bmi->bmiColors[color_count];
        memcpy( m_bits, &src.bmiColors[color_count], sizeof_image );
      }
    }
  }
  else {
    // share memory
    m_bmi = (ON_WindowsBITMAPINFO*)&src;
    if ( m_bmi )
      m_bits = (unsigned char*)&m_bmi->bmiColors[PaletteColorCount()];
  }
  return *this;
}

ON_WindowsBitmap::ON_WindowsBitmap( const BITMAPINFO*& src )
                 : m_bmi(0), m_bits(0), m_bFreeBMI(0)
{
  *this = src;
}

BITMAPINFO* ON_WindowsBitmap::Convert()
{
  BITMAPINFO* bmi = 0;
  if ( m_bmi ) {
    if ( sizeof(BITMAPINFO) != sizeof( ON_WindowsBITMAPINFO ) ) {
      BITMAPINFOHEADER bmiHeader;
      memset( &bmiHeader, 0, sizeof(bmiHeader) );
      bmiHeader.biSize          = sizeof(bmiHeader);
      bmiHeader.biWidth         = m_bmi->bmiHeader.biWidth;
      bmiHeader.biHeight        = m_bmi->bmiHeader.biHeight;
      bmiHeader.biPlanes        = m_bmi->bmiHeader.biPlanes;
      bmiHeader.biBitCount      = m_bmi->bmiHeader.biBitCount;
      bmiHeader.biCompression   = m_bmi->bmiHeader.biCompression;
      bmiHeader.biSizeImage     = m_bmi->bmiHeader.biSizeImage;
      bmiHeader.biXPelsPerMeter = m_bmi->bmiHeader.biXPelsPerMeter;
      bmiHeader.biYPelsPerMeter = m_bmi->bmiHeader.biYPelsPerMeter;
      bmiHeader.biClrUsed       = m_bmi->bmiHeader.biClrUsed;
      bmiHeader.biClrImportant  = m_bmi->bmiHeader.biClrImportant;
      // copy palette and image
      const int sizeof_header  = sizeof(BITMAPINFOHEADER);
      const int sizeof_palette = ON_WindowsBitmap_SizeofPalette(bmiHeader.biClrUsed, bmiHeader.biBitCount );
      const int sizeof_image   = bmiHeader.biSizeImage;
      bmi = (BITMAPINFO*)Realloc( 0, sizeof_header+sizeof_palette+sizeof_image);
      if ( bmi ) {
        bmi->bmiHeader = bmiHeader;
        const int color_count = ON_WindowsBitmap_PaletteColorCount(bmiHeader.biClrUsed, bmiHeader.biBitCount );
        int i;
        for ( i = 0; i < color_count; i++ ) {
          bmi->bmiColors[i].rgbBlue     = m_bmi->bmiColors[i].rgbBlue;
          bmi->bmiColors[i].rgbGreen    = m_bmi->bmiColors[i].rgbGreen;
          bmi->bmiColors[i].rgbRed      = m_bmi->bmiColors[i].rgbRed;
          bmi->bmiColors[i].rgbReserved = m_bmi->bmiColors[i].rgbReserved;
        }
        if ( sizeof_image > 0 ) {
          // copy image bits
          memcpy( &bmi->bmiColors[color_count], &m_bmi->bmiColors[color_count], sizeof_image );
        }
      }
    }
    else {
      bmi = (BITMAPINFO*)m_bmi;
    }
  }
  return bmi;
}

#endif

ON_WindowsBITMAPINFO* ON_WindowsBitmap::Realloc(ON_WindowsBITMAPINFO* ptr,size_t sz)
{
  return (ON_WindowsBITMAPINFO*)onrealloc(ptr,sz);
}


////////////////////////////////////////////////////////////////////////////////
//
// OpenGL bitmap
//

ON_OpenGLBitmap::ON_OpenGLBitmap()
: m_bits(0), m_width(0), m_height(0), m_sizeof_pixel(0)
{
}

ON_OpenGLBitmap::~ON_OpenGLBitmap()
{
  Destroy();
}

BOOL ON_OpenGLBitmap::Read( ON_BinaryArchive& file )
{
  ON_ERROR("TODO - finish ON_OpenGLBitmap::Read()\n");
  return FALSE;
}

BOOL ON_OpenGLBitmap::Write( ON_BinaryArchive& file ) const
{
  ON_ERROR("TODO - finish ON_OpenGLBitmap::Write()\n");
  return FALSE;
}

unsigned char* ON_OpenGLBitmap::Realloc(unsigned char* ptr,size_t sz)
{
  return (unsigned char*)onrealloc(ptr,sz);
}

void ON_OpenGLBitmap::EmergencyDestroy()
{
  m_bits = 0;
  m_width = 0;
  m_height = 0;
  m_sizeof_pixel = 0;
}

void ON_OpenGLBitmap::Destroy()
{
  if ( m_bits ) {
    Realloc(m_bits,0);
    m_bits = 0;
  }
  m_width = 0;
  m_height = 0;
  m_sizeof_pixel = 0;
}

ON_OpenGLBitmap::ON_OpenGLBitmap( const ON_OpenGLBitmap& src )
{
  *this = src;
}

int ON_OpenGLBitmap::Width() const
{
  return m_width;
}

int ON_OpenGLBitmap::Height() const
{
  return m_height;
}

int ON_OpenGLBitmap::SizeofPixel() const
{
  return m_sizeof_pixel;
}

int ON_OpenGLBitmap::BitsPerPixel() const
{
  return m_sizeof_pixel*8;
}

int ON_OpenGLBitmap::SizeofScan() const
{
  return m_width*m_sizeof_pixel;
}

int ON_OpenGLBitmap::SizeofImage() const
{
  return SizeofImage( Width(), Height(), SizeofPixel() );
}

int ON_OpenGLBitmap::SizeofImage( int width, int height, int sizeof_pixel ) const
{
  int i = 0;
  if ( width > 0 && height > 0 && sizeof_pixel >= 3 && sizeof_pixel <= 4 ) {
    i = width*height*sizeof_pixel + 4; // extra pixel needed internally to simplify some loops
    if ( i%4 )
      i = 4*(i/4 + 1);
    i *= sizeof(unsigned char);
  }
  return i;
}


ON_OpenGLBitmap& ON_OpenGLBitmap::operator=( const ON_OpenGLBitmap& src )
{
  if ( this != &src ) {
    Destroy();
    ON_Bitmap::operator=(src);
    if ( src.m_bits && Create( src.Width(), src.Height(), src.SizeofPixel() ) ) {
      memcpy( m_bits, src.m_bits, SizeofImage() );
    }
  }
  return *this;
}

BOOL ON_OpenGLBitmap::Convert( const ON_WindowsBitmap& src, int sizeof_pixel )
{
  Destroy();
  
  if ( sizeof_pixel != 3 && sizeof_pixel != 4)
    return FALSE; // can only convert to GL RGB or RGBA type bitmaps.
  
  // copy Windows bitmap contents to GL bitmap
  
  if ( !Create( src.Width(), src.Height(), sizeof_pixel) ) {
    return FALSE;
  }
  
  if ( m_width != src.Width() || m_height != src.Height() ) {
    Destroy();
    return FALSE;
  }
  
  const unsigned char* scan;
  unsigned char *p;
  int x, y;
  ON_Color c;
  
  for ( y = 0; y < m_height; y++) {
    scan = src.Bits(y);
    p = Bits(y);
    for ( x = 0; x < m_width; x++) {
      c = src.Pixel( x, scan );
      p[0] = c.Red();
      p[1] = c.Green();
      p[2] = c.Blue();      
      if ( m_sizeof_pixel == 4)
        p[3] = 255; // alpha
      p += m_sizeof_pixel;
    }
  }
  
  return TRUE;
}


BOOL ON_OpenGLBitmap::ConvertToTextureMap()
{
  int w, h;
  BOOL rc = GetTextureSize( &w, &h );
  if ( rc )
    rc = Stretch( w, h );
  if ( !rc )
    Destroy();
  return rc;
}

BOOL ON_OpenGLBitmap::ConvertToRGB()
{
  int w, h;
  BOOL rc = IsValid();
  if ( rc && SizeofPixel() == 4 ) {
    const int width = Width();
    const int height = Height();
    const unsigned char* p0;
    unsigned char* p1;
    for ( h = 0; h < height; h++ ) {
      p1 = Bits(h);
      p0 = p1;
      for ( w = 0; w < width; w++ ) {
        *p1++ = *p0++;
        *p1++ = *p0++;
        *p1++ = *p0++;
        p0++;
      }
    }
    m_sizeof_pixel = 3;
  }
  return rc;
}

BOOL ON_OpenGLBitmap::ConvertToRGBA()
{
  BOOL rc = IsValid();
  if ( rc && SizeofPixel() == 3 ) {
    const int width = Width();
    const int height = Height();
    int w, h;
    const unsigned char* p0;
    unsigned char* p1;
    rc = FALSE;
    const int s = SizeofImage( width, height, 4 );
    if ( s > 0 ) {
      m_bits = Realloc( m_bits, s );
      if ( m_bits ) {
        for ( h = height-1; h >= 0; h-- ) {
          m_sizeof_pixel = 3;
          p0 = Bits(h) + (width-1)*m_sizeof_pixel;
          m_sizeof_pixel = 4;
          p1 = Bits(h) + (width-1)*m_sizeof_pixel;
          for ( w = 0; w < width; w++ ) {
            p1[3] = 255;
            p1[2] = p0[2];
            p1[1] = p0[1];
            p1[0] = p0[0];
            p1 -= 4;
            p0 -= 3;
          }
        }
        rc = TRUE;
      }
    }
  }
  if ( !rc )
    Destroy();
  return rc;
}

BOOL ON_OpenGLBitmap::IsValid() const
{
  return (m_width > 0 && m_height > 0 && m_sizeof_pixel >= 3 && m_sizeof_pixel <= 4 && m_bits != NULL ) ? TRUE : FALSE;
}

BOOL ON_OpenGLBitmap::Create( int width, int height, int sizeof_pixel )
{
  Destroy();
  const int s = SizeofImage( width, height, sizeof_pixel );
  if ( s > 0 ) {
    m_bits = Realloc( 0,s);
    //{
    //  static int malloc_count = 0;
    //  malloc_count++;
    //  printf("BITMAP MALLOC %2d %x (%d bytes)\n",malloc_count,m_bits,s);
    //}
    if ( m_bits ) {
      m_width = width;
      m_height = height;
      m_sizeof_pixel = sizeof_pixel;
      memset( m_bits, 255, s );
    }
  }
  return ( m_bits ) ? TRUE : FALSE;
}

BOOL ON_OpenGLBitmap::SetColor( ON_Color c )
{
  BOOL rc = FALSE;
  const unsigned char r = c.Red();
  const unsigned char g = c.Green();
  const unsigned char b = c.Blue();
  const BOOL bHasAlpha = ( SizeofPixel() == 4 );
  if ( IsValid() ) {
    int i = Width()*Height();
    unsigned char* p = m_bits;
    while ( i-- ) {
      *p++ = r;
      *p++ = g;
      *p++ = b;
      if ( bHasAlpha )
        p++;
    }
    rc = TRUE;
  }
  return rc;
}

BOOL ON_OpenGLBitmap::SetAlpha( int alpha )
{
  // 0 = transparent, 255 = opaque
  BOOL rc = IsValid();
  if ( alpha < 0 )
    alpha = 0;
  else if ( alpha > 255 )
    alpha = 255;
  if ( rc && SizeofPixel() == 3 ) {
    rc = ConvertToRGBA();
  }

  if ( IsValid() ) {
    int i = Width()*Height();
    const unsigned char a = (unsigned char)alpha;
    unsigned char* p = m_bits;
    while ( i-- ) {
      p++; // r
      p++; // g
      p++; // b
      *p++ = a;
    }
    rc = TRUE;
  }



/*
  if ( rc ) {
    const int width = Width();
    const int height = Height();
    const unsigned char a = (unsigned char)alpha;
    int w, h;
    unsigned char* p;
    for ( h = 0; h < height; h++ ) {
      p = Pixel(width-1,h);
      for ( w = 0; w < width; w++ ) {
        p[3] = a;
        p += 4;
      }
    }
  }
*/


  return rc;
}

unsigned char* ON_OpenGLBitmap::Bits( int scan_index )
{
  const int sizeof_scan = SizeofScan();
  unsigned char* bits = m_bits;
  if ( bits && sizeof_scan && scan_index >= 0 && scan_index < Height() ) {
    bits += (sizeof_scan*scan_index);
  }
  else {
    bits = 0;
  }
  return bits;
}

const unsigned char* ON_OpenGLBitmap::Bits(int scan_index) const
{
  const int sizeof_scan = SizeofScan();
  const unsigned char* bits = m_bits;
  if ( bits && sizeof_scan && scan_index >= 0 && scan_index < Height() ) {
    bits += (sizeof_scan*scan_index);
  }
  else {
    bits = 0;
  }
  return bits;
}

ON_Color ON_OpenGLBitmap::Pixel( int i, int j ) const
{
  ON_Color c(0,0,0);
  if ( 0 <= i && i < m_width && 0 <= j && j < m_height && m_bits ) {
    const unsigned char* p = m_bits + (i+j*m_width)*m_sizeof_pixel;
    if ( m_sizeof_pixel==4 ) {
      c.SetRGBA( p[0], p[1], p[2], p[3] );
    }
    else {
      c.SetRGB( p[0], p[1], p[2] );
    }
  }
  return c;
}

ON_Color ON_OpenGLBitmap::Pixel( int i, const unsigned char* scan ) const
{
  ON_Color c(0,0,0);
  if ( 0 <= i && i < m_width && scan ) {
    const unsigned char* p = scan + i*m_sizeof_pixel;
    if ( m_sizeof_pixel==4 ) {
      c.SetRGBA( p[0], p[1], p[2], p[3] );
    }
    else {
      c.SetRGB( p[0], p[1], p[2] );
    }
  }
  return c;
}

static BOOL StretchBitmapX( // maps cannot overlap
             int w0, int h0, int sizeof_pixel0, const unsigned char* map0,
             int w1, int h1, int sizeof_pixel1, unsigned char* map1
             )
{
  double r, g, b, a, s, f, fx, fy, x0, x1, y0, y1;
  int i0, j0, i1, j1, xx0, xx1, yy0, yy1;
  const unsigned char* p0;
  unsigned char* p1;
  unsigned char rr, gg, bb, aa;

  if ( w0 < 1 || h0 < 1 || w1 < 1 || h1 < 1 )
    return FALSE;
  const double dw = ((double)w0)/((double)w1);
  const double dh = ((double)h0)/((double)h1);

  aa = 255;

  for ( j1 = 0; j1 < h1; j1++ ) {
    p1 = map1 + (j1*w1*sizeof_pixel1);
    y0 = j1*dh;
    y1 = (j1 < h1 - 1) ? (j1+1)*dh : h0;
    yy0 = (int)floor(y0);
    yy1 = (int)ceil(y1);

    // these tests are "not necessary" but will prevent crashes if
    // unforseen events happen.
    if ( yy0 < 0 )
      yy0 = 0;
    if ( yy1 > h0 )
      yy1 = h0;

    yy1--;
    for ( i1 = 0; i1 < w1; i1++, p1 += sizeof_pixel1 ) {
      p1[0] = 255;
      p1[1] = 0;
      p1[2] = 0;

      r = g = b = a = s = 0.0;
      x0 = i1*dw;
      x1 = (i1 < w1 - 1) ? (i1+1)*dw : w0;
      xx0 = (int)floor(x0);
      xx1 = (int)ceil(x1);
      
      // and these tests are also "not necessary" but will prevent crashes if
      // unforseen events happen.
      if ( xx0 < 0 )
        xx0 = 0;
      if ( xx1 > w0 )
        xx1 = w0;

      xx1--;
      for ( j0 = yy0; j0 <= yy1; j0++ ) {
        p0 = map0 + ((j0*w0 + xx0)*sizeof_pixel0);
        if ( j0 == yy0 ) {
          fy = ((j0==yy1) ? y1 : floor(y0+1.0)) - y0;
          if ( fy < 0.0 )
            fy = 0.0;
        }
        else if ( j0 == yy1 ) {
          fy = y1 - ceil(y1-1.0);
          if ( fy < 0.0 ) fy = 0.0;
        }
        else
          fy = 1.0;
        for ( i0 = xx0; i0 <= xx1; i0++, p0 += sizeof_pixel0 ) {
          if ( i0 == xx0 ) {
            fx = ((i0==xx1) ? x1 : floor(x0+1.0)) - x0;
            if ( fx < 0.0 ) fx = 0.0;
          }
          else if ( i0 == xx1 ) {
            fx = x1 - ceil(x1-1.0);
            if ( fx < 0.0 ) fx = 0.0;
          }
          else
            fx = 1.0;

          f = fx*fy;

          r += f*p0[0];
          g += f*p0[1];
          b += f*p0[2];
          if ( sizeof_pixel0 == 4 )
            a += f*p0[3];
          s += f;
        }
      }
      if ( s > 0.0 ) {
        s = 1.0/s;
        r *= s;
        g *= s;
        b *= s;
        if ( sizeof_pixel0 == 4 ) a *= s; else a = 255.0;
        rr = (r > 255.0) ? 255 : (unsigned char)r;
        gg = (g > 255.0) ? 255 : (unsigned char)g;
        bb = (b > 255.0) ? 255 : (unsigned char)b;
        aa = (a > 255.0) ? 255 : (unsigned char)a;
      }
      else {
        rr = gg = bb = aa = 255;
      }
      p1[0] = rr;
      p1[1] = gg;
      p1[2] = bb;
      if ( sizeof_pixel1 == 4 )
        p1[3] = aa;
    }
  }
  return TRUE;
}

BOOL ON_OpenGLBitmap::Stretch( int width, int height )
{
  BOOL rc = FALSE;
  if ( width > 0 && height > 0 && IsValid() ) {
    if ( width == Width() && height == Height() ) {
      rc = TRUE;
    }
    else {
      ON_OpenGLBitmap map1;
      if ( map1.Create( width, height, SizeofPixel() ) ) {
        rc = StretchBitmapX( // map memory cannot overlap
               Width(), Height(), SizeofPixel(), m_bits,
               map1.Width(), map1.Height(), map1.SizeofPixel(), map1.m_bits
               );
        if ( rc ) {
          if ( m_bits ) {
            Realloc(m_bits,0);
            m_bits = NULL;
          }
          m_width        = map1.Width();
          m_height       = map1.Height();
          m_sizeof_pixel = map1.SizeofPixel();
          m_bits          = map1.m_bits;
          map1.m_bits = 0;
        }
        else {
          Destroy();
        }
      }
    }
  }
  return rc;
}

BOOL ON_OpenGLBitmap::GetTextureSize(
        int* map_width1,       // output width will be 2^m for some integer m >= 0
        int* map_height1       // output height will be 2^n for some integer n >= 0
        ) const
{
  const int max_wXh = 1024*1024;
  const int min_w = 1;
  const int min_h = 1;

  int map_width0 = m_width;
  int map_height0 = m_height;

  int w, h, i, j;
  if ( map_width1 )
    *map_width1 = 0;
  if ( map_height1 )
    *map_height1 = 0;
  if ( map_width0 < 1 || map_height0 < 1 || !map_width1 || !map_height1  )
    return FALSE;


  // get 2^m x 2^n size that best matches input size
  for ( i = 1; i < map_width0; i *= 2 )
    ; // empty for
  w = i;
  if ( w > 1 && w - map_width0 > map_width0 - w/2 )
    w /= 2;
  for ( i = 1; i < map_height0; i *= 2 )
    ; // empty for
  h = i;
  if ( h > 1 && h - map_height0 > map_height0 - h/2 )
    h /= 2;
  if (w < min_w) w = min_w;
  if (h < min_h) h = min_h;

  while ( w*h > max_wXh ) {
    i = w - map_width0;
    j = h - map_height0;
    if ( i > 0 && i > j && w > min_w )
      w /= 2;
    else if ( j > 0 && j > i && h > min_h )
      h /= 2;
    else if ( w > h && w > min_w )
      w /= 2;
    else if ( h > min_h )
      h /= 2;
    else
      return FALSE;
  }

  *map_width1  = w;
  *map_height1 = h;

  return TRUE;
}
