/* $Header: /src2/opennurbs/opennurbs_bounding_box.cpp 16    7/24/01 3:54p Dalelear $ */
/* $NoKeywords: $ */
/*
//
// Copyright (c) 1993-2001 Robert McNeel & Associates. All rights reserved.
// Rhinoceros is a registered trademark of Robert McNeel & Assoicates.
//
// THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
// ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
// MERCHANTABILITY ARE HEREBY DISCLAIMED.
//				
// For complete openNURBS copyright information see <http://www.opennurbs.org>.
//
////////////////////////////////////////////////////////////////
*/

#include "opennurbs.h"

ON_BoundingBox::ON_BoundingBox()
                : m_min(1.0,0.0,0.0), 
                  m_max(-1.0,0.0,0.0)
{}

ON_BoundingBox::ON_BoundingBox( const ON_3dPoint& min_pt, const ON_3dPoint& max_pt )
                : m_min( min_pt ), 
                  m_max( max_pt )
{}

ON_BoundingBox::~ON_BoundingBox()
{}

void ON_BoundingBox::Destroy()
{
  m_min.Zero();
  m_max.Zero();
  m_min.x = 1.0;
  m_max.x = -1.0;
}

BOOL ON_BoundingBox::IsValid() const 
{
	return m_min[0]<=m_max[0] && m_min[1]<=m_max[1] && m_min[1]<=m_max[1];
};

//////////
// ON_BoundingBox::Transform() updates the bounding box
// to be the smallest axis aligned bounding box that contains
// the transform of the eight corner points of the input
// bounding box.
BOOL ON_BoundingBox::Transform( const ON_Xform& xform )
{
  ON_3dPointArray corners;
  BOOL rc =  GetCorners( corners );
  if (rc) {
    rc = corners.Transform(xform);
    if (rc)
      rc = Set(corners);
  }
  return rc;
}

double ON_BoundingBox::Tolerance() const
{
  // rough guess at a tolerance to use for comparing
  return ON_BoundingBoxTolerance( 3, m_min, m_max );
}

ON_3dPoint& ON_BoundingBox::operator[](int i)
{
  return (i>0) ? m_max : m_min;
}

const ON_3dPoint& ON_BoundingBox::operator[](int i) const
{
  return (i>0) ? m_max : m_min;
}

ON_3dPoint ON_BoundingBox::Min() const 
{
	return m_min;
}

ON_3dPoint ON_BoundingBox::Max() const 
{
	return m_max;
}

ON_3dVector ON_BoundingBox::Diagonal() const
{
  return m_max - m_min;
}

ON_3dPoint ON_BoundingBox::Center() const
{
  return 0.5*(m_max+m_min);
}

ON_3dPoint ON_BoundingBox::Corner( int x_index, int y_index, int z_index ) const
{
  // 8 corners of box
  // x_index   0 = Min().x, 1 = Max().x
  // y_index   0 = Min().y, 1 = Max().y
  // z_index   0 = Min().z, 1 = Max().z
  ON_3dPoint corner;
  corner.x = (x_index>0) ? m_max.x : m_min.x;
  corner.y = (y_index>0) ? m_max.y : m_min.y;
  corner.z = (z_index>0) ? m_max.z : m_min.z;
  return corner;
}

BOOL
ON_BoundingBox::GetCorners( 
  ON_3dPointArray& corners// returns list of 8 corner points
  ) const
{
  double x,y,z;
  int i,j,k;
  corners.Empty();
  corners.Reserve(8);
  BOOL rc = IsValid();
  if ( rc ) {
    for( i = 0; i < 2; i++ ) {
      x = (i) ? m_max.x : m_min.x;
      for ( j = 0; j < 2; j++ ) {
        y = (j) ? m_max.y : m_min.y;
        for ( k = 0; k < 2; k++ ) {
          z = (k) ? m_max.z : m_min.z;
          corners.Append( ON_3dPoint(x,y,z) );
        }
      }
    }
  }
  return rc;
}


BOOL ON_BoundingBox::IsPointIn( const ON_3dPoint& p, BOOL bStrictlyIn ) const
{
  BOOL bIn = FALSE;
  if ( bStrictlyIn ) {
	  bIn = m_min.x<p.x && p.x<m_max.x &&
					m_min.y<p.y && p.y<m_max.y &&
					m_min.z<p.z && p.z<m_max.z;
  }
  else {
	  bIn = m_min.x<=p.x && p.x<=m_max.x &&
					m_min.y<=p.y && p.y<=m_max.y &&
					m_min.z<=p.z && p.z<=m_max.z;
  }
  return bIn;
};

ON_3dPoint ON_BoundingBox::ClosestPoint( 
  const ON_3dPoint& test_point
  ) const
{
  ON_3dPoint near_point = test_point;
  if ( IsValid() ) {
    if ( test_point.x < m_min.x )
      near_point.x = m_min.x;
    else if ( test_point.x > m_max.x )
      near_point.x = m_max.x;
    if ( test_point.y < m_min.y )
      near_point.y = m_min.y;
    else if ( test_point.y > m_max.y )
      near_point.y = m_max.y;
    if ( test_point.z < m_min.z )
      near_point.z = m_min.z;
    else if ( test_point.z > m_max.z )
      near_point.z = m_max.z;
  }
  return near_point;
}

ON_3dPoint ON_BoundingBox::FarPoint( 
  const ON_3dPoint& test_point
  ) const
{
  ON_3dPoint far_point = test_point;
  if ( IsValid() ) {
    far_point.x = ( fabs(m_min.x-test_point.x) >= fabs(m_max.x-test_point.x) )
                ? m_min.x : m_max.x;
    far_point.y = ( fabs(m_min.y-test_point.y) >= fabs(m_max.y-test_point.y) )
                ? m_min.y : m_max.y;
    far_point.z = ( fabs(m_min.z-test_point.z) >= fabs(m_max.z-test_point.z) )
                ? m_min.z : m_max.z;
  }
  return far_point;
}

//TODO:  Replace this static function with an ON_Interval member function.
//  Add to the ON_Interval class in ON_Interval.h
//
// returns true if the intersection is non-empty and sets AB to the intersection
//  BOOL   GetIntersection(ON_Interval B, ON_Interval& AB);


static BOOL Intersect( ON_Interval A, ON_Interval B, ON_Interval& AB);

BOOL Intersect( ON_Interval A, ON_Interval B, ON_Interval& AB){
	if(A.IsDecreasing()) A.Swap();
	if(B.IsDecreasing()) B.Swap();

	BOOL NotEmpty=TRUE;
	if( A.m_t[0] <= B.m_t[0] && B.m_t[0]<=A.m_t[1] &&  A.m_t[1]<= B.m_t[1]){
		AB.Set(B.m_t[0], A.m_t[1]);
	} else if( B.m_t[0] <= A.m_t[0] && A.m_t[0]<=B.m_t[1] && B.m_t[1]<=A.m_t[1]){
		AB.Set(A.m_t[0], B.m_t[1]);
	} else if( A.m_t[0] <=  B.m_t[0] && B.m_t[0] <= B.m_t[1] && B.m_t[1]<= A.m_t[1]){
		AB.Set(B.m_t[0], B.m_t[1]);
	} else if( B.m_t[0] <= A.m_t[0] && A.m_t[0] <= A.m_t[1] && A.m_t[1]<= B.m_t[1]){
		AB.Set(A.m_t[0], A.m_t[1]);
	} else if( B.m_t[0] <= A.m_t[0]  && A.m_t[0] <= A.m_t[1]  && A.m_t[1]<= B.m_t[1]){
		AB.Set(A.m_t[0], A.m_t[1]);
	} else if(A.m_t[1] < B.m_t[0] || B.m_t[1] < A.m_t[0] ){
		AB.Destroy();
		NotEmpty = FALSE;
	}
	return NotEmpty;
}

BOOL ON_BoundingBox::GetClosestPoint( 
       const ON_BoundingBox& other_box, // "other" bounding box
       ON_3dPoint& this_point, // point on "this" box that is closest to "other" box
       ON_3dPoint& other_point // point on "other" box that is closest to "this" box
       )  const
{
  ON_BoundingBox b;
	if ( !IsValid() || !other_box.IsValid() )
		return FALSE;

	for (int i=0; i<3; i++ )
  {
		ON_Interval It(m_min[i],m_max[i]);
		ON_Interval Io(other_box.m_min[i],other_box.m_max[i]);
		ON_Interval intersect;
		BOOL NotEmpty = Intersect(It,Io,intersect);
		if(NotEmpty)
    {
			this_point[i] = other_point[i] = intersect.Mid();
		} 
    else {
			if(m_max[i]< other_box.m_min[i] )
      {
				this_point[i] = m_max[i];
				other_point[i] = other_box.m_min[i];
			} 
      else {
				this_point[i] = m_min[i];
				other_point[i] = other_box.m_max[i];
			}
		}
	}
	return TRUE;
}

//////////
// Get points on bounding boxes that are farthest from each other.
BOOL ON_BoundingBox::GetFarPoint( 
       const ON_BoundingBox& other_box, // "other" bounding box
       ON_3dPoint& this_point, // point on "this" box that is farthest from "other" box point
       ON_3dPoint& other_point // point on "other" box that is farthest from "this" box point
       )  const
{
	if(!IsValid() || !other_box.IsValid())
		return FALSE;
	for(int i=0; i<3; i++){
		ON_Interval It(m_min[i], m_max[i]);
		ON_Interval Io(other_box.m_min[i], other_box.m_max[i]);
		if( It.Includes(Io) || Io.Includes(It)){
			if( m_max[i] - other_box.m_min[i] > other_box.m_max[i] - m_min[i]){
				this_point[i] = m_max[i];
				other_point[i] = other_box.m_min[i];
			} else {
				this_point[i] = m_min[i];
				other_point[i] = other_box.m_max[i];
			}
		} else {
			if( m_min[i]< other_box.m_min[i]){
				this_point[i]=m_min[i];
			} else {
				other_point[i] = other_box.m_min[i];
			}
			if( m_max[i]> other_box.m_max[i]){
				this_point[i]=m_max[i];
			} else {
				other_point[i] = other_box.m_max[i];
			}
		}
	}
	return TRUE;
}

BOOL ON_BoundingBox::IsDisjoint(const ON_BoundingBox& box) const
{	//todo wrong!
  if (!IsValid() || !box.IsValid()) return FALSE;
  int i;
  for (i=0; i<3; i++){
    if (box.m_min[i] > m_max[i] || box.m_max[i] < m_min[i]) return TRUE;
  }
  return FALSE;
}

BOOL ON_BoundingBox::SwapCoordinates( int i, int j )
{
  BOOL rc = FALSE;
  if ( IsValid() && 0 <= i && i < 3 && 0 <= j && j < 3 ) {
    rc = TRUE;
    if ( i != j ) {
      double t = m_min[i]; m_min[i] = m_min[j]; m_min[j] = t;
      t = m_max[i]; m_max[i] = m_max[j]; m_max[j] = t;
    }
  }
  return rc;
}

BOOL ON_BoundingBox::Intersection(
          const ON_BoundingBox& a
          )
{
  if ( IsValid() && a.IsValid() ) {
    if ( a.m_min.x > m_min.x )
      m_min.x = a.m_min.x;
    if ( a.m_min.y > m_min.y )
      m_min.y = a.m_min.y;
    if ( a.m_min.z > m_min.z )
      m_min.z = a.m_min.z;
    if ( a.m_max.x < m_max.x )
      m_max.x = a.m_max.x;
    if ( a.m_max.y < m_max.y )
      m_max.y = a.m_max.y;
    if ( a.m_max.z < m_max.z )
      m_max.z = a.m_max.z;
  }
  else {
    Destroy();
  }
  return IsValid();
}

BOOL ON_BoundingBox::Union(
          const ON_BoundingBox& a
          )
{
  if ( IsValid() ) {
    if ( a.IsValid() ) {
      if ( a.m_min.x < m_min.x )
        m_min.x = a.m_min.x;
      if ( a.m_min.y < m_min.y )
        m_min.y = a.m_min.y;
      if ( a.m_min.z < m_min.z )
        m_min.z = a.m_min.z;
      if ( a.m_max.x > m_max.x )
        m_max.x = a.m_max.x;
      if ( a.m_max.y > m_max.y )
        m_max.y = a.m_max.y;
      if ( a.m_max.z > m_max.z )
        m_max.z = a.m_max.z;
    }
  }
  else if ( a.IsValid() ) {
    *this = a;
  }
  else {
    Destroy();
  }
  return IsValid();
}

BOOL ON_BoundingBox::Intersection(
          const ON_BoundingBox& a,
          const ON_BoundingBox& b
          )
{
  if ( a.IsValid() && b.IsValid() ) {
    m_min.x = (a.m_min.x >= b.m_min.x) ? a.m_min.x : b.m_min.x;
    m_min.y = (a.m_min.y >= b.m_min.y) ? a.m_min.y : b.m_min.y;
    m_min.z = (a.m_min.z >= b.m_min.z) ? a.m_min.z : b.m_min.z;
    m_max.x = (a.m_max.x <= b.m_max.x) ? a.m_max.x : b.m_max.x;
    m_max.y = (a.m_max.y <= b.m_max.y) ? a.m_max.y : b.m_max.y;
    m_max.z = (a.m_max.z <= b.m_max.z) ? a.m_max.z : b.m_max.z;
  }
  else {
    Destroy();
  }
  return IsValid();
}

BOOL ON_BoundingBox::Union(
          const ON_BoundingBox& a,
          const ON_BoundingBox& b
          )
{
  if ( a.IsValid() ) {
    if ( b.IsValid() ) {
      m_min.x = (a.m_min.x <= b.m_min.x) ? a.m_min.x : b.m_min.x;
      m_min.y = (a.m_min.y <= b.m_min.y) ? a.m_min.y : b.m_min.y;
      m_min.z = (a.m_min.z <= b.m_min.z) ? a.m_min.z : b.m_min.z;
      m_max.x = (a.m_max.x >= b.m_max.x) ? a.m_max.x : b.m_max.x;
      m_max.y = (a.m_max.y >= b.m_max.y) ? a.m_max.y : b.m_max.y;
      m_max.z = (a.m_max.z >= b.m_max.z) ? a.m_max.z : b.m_max.z;
    }
    else {
      *this = a;
    }
  }
  else if ( b.IsValid() ) {
    *this = b;
  }
  else {
    Destroy();
  }
  return IsValid();
}

double ON_BoundingBox::Volume() const
{
  double dx = m_max.x - m_min.x;
  double dy = m_max.y - m_min.y;
  double dz = m_max.z - m_min.z;
  return (dx > 0.0 && dy > 0.0 && dz > 0.0) ? dx*dy*dz : 0.0;
}

double ON_BoundingBox::Area() const
{
  double dx = m_max.x - m_min.x;
  double dy = m_max.y - m_min.y;
  double dz = m_max.z - m_min.z;
  return (dx >= 0.0 && dy >= 0.0 && dz >= 0.0) ? 2.0*(dx*dy + dy*dz + dz*dx) : 0.0;
}

BOOL ON_BoundingBox::Set(     
    int dim, BOOL is_rat, int count, int stride, const double* points, 
    BOOL bGrowBox
  )
{
  double *box[2];
  if ( bGrowBox )
    bGrowBox = IsValid();
  if ( dim > 3 ) {
    box[0] = (double*)onmalloc(dim*2*sizeof(box[0][0]));
    box[1] = box[0]+dim;
    if ( bGrowBox ) {
      memset(box[0],0,dim*2*sizeof(box[0][0]));
      box[0][0] = m_min.x;
      box[0][1] = m_min.y;
      box[0][2] = m_min.z;
      box[1][0] = m_max.x;
      box[1][1] = m_max.y;
      box[1][2] = m_max.z;
    }
  }
  else {
    box[0] = &m_min.x;
    box[1] = &m_max.x;
    if ( !bGrowBox ) {
      switch(dim) {
      case 1:
        m_min.y = m_max.y = 0.0;
        // NO break
      case 2:
        m_min.z = m_max.z = 0.0;
        break;
      }
    }
  }
  BOOL rc = ON_GetPointListBoundingBox( 
                dim, is_rat, 
                count, stride, points, 
                box[0], box[1], 
                bGrowBox
                );
  if ( dim > 3 ) {
    m_min = box[0];
    m_max = box[1];
    onfree(box);
  }
  return rc;
}

BOOL ON_BoundingBox::Set ( const ON_3dPoint& pt, BOOL bGrowBox )
{
  return Set( 3, 0, 1, 3, pt, bGrowBox );
}


BOOL ON_BoundingBox::Set(     
  const ON_4dPointArray& a,
  BOOL bGrowBox
  )
{
  const int count = a.Count();
  const double* points = (count>0) ? &a.Array()[0].x :0;
  return Set( 3, TRUE, count, 4, points, bGrowBox );
}

BOOL ON_BoundingBox::Set(     
  const ON_3dPointArray& a,
  BOOL bGrowBox
  )
{
  const int count = a.Count();
  const double* points = (count>0) ? &a.Array()[0].x :0;
  return Set( 3, FALSE, count, 3, points, bGrowBox );
}

BOOL ON_BoundingBox::Set(     
  const ON_2dPointArray& a,
  BOOL bGrowBox
  )
{
  const int count = a.Count();
  const double* points = (count>0) ? &a.Array()[0].x :0;
  return Set( 2, FALSE, count, 2, points, bGrowBox );
}

ON_DEFN
ON_BoundingBox ON_PointListBoundingBox(
    int dim, BOOL is_rat, int count, int stride, const double* points, 
    BOOL bGrowBox
    )
{
  ON_BoundingBox bbox;
  BOOL rc = FALSE;
  double* bboxmin = &bbox.m_min.x;
  double* bboxmax = &bbox.m_max.x;
  if ( dim > 3 ) {
    bboxmin = (double*)alloca(2*dim*sizeof(*bboxmin));
    bboxmax = bboxmin+dim;
  }
  rc = ON_GetPointListBoundingBox( dim, is_rat, count, stride, points, bboxmin, bboxmax, bGrowBox );
  if ( rc && dim > 3 ) {
    bbox.m_min = ON_3dPoint( bboxmin );
    bbox.m_max = ON_3dPoint( bboxmax );
  }
  return bbox;
}

ON_DEFN
BOOL ON_GetPointListBoundingBox( 
    int dim, BOOL is_rat, int count, int stride, const double* points, 
    double* boxmin, double* boxmax,
    BOOL bGrowBox
    )
/*****************************************************************************
Bounding Box of a set of points
 
INPUT:
  dim           ( >= 1 ) dimension of each point
  is_rat        ( TRUE if points are rational )
  count         ( >= 1 ) number of points
  stride        ( >= (is_rat)?(dim+1):dim )
  points        array of dim*count doubles
  boxmin, boxmax      unused arrays of dim doubles
  bGrowBox       TRUE if input box should be enlarged to contain points
                 FALSE if input box should be ignored bounding box of points
                       is returned
OUTPUT:
  boxmin, boxmax      diagonal corners of bounding box
*****************************************************************************/
{
  // bounding box workhorse
  double x, w;
  int j;
  BOOL rc = FALSE;
  if ( count > 0 ) {
    if ( is_rat )
      is_rat = 1;
    if ( points && dim > 0 && (count == 1 || stride >= dim+is_rat) ) {
      // input is valid list of a least 1 point

      if ( is_rat ) {
        // bounding box of homogeneous rational points
        rc = TRUE;
        while ( count > 0 && points[dim] == 0.0 ) {
          count--;
          points += stride;
          rc = FALSE;
        }
        if ( count > 0 ) {
          if ( !bGrowBox ) {
            ON_ArrayScale( dim, 1.0/points[dim], points, boxmin );
            memcpy( boxmax, boxmin, dim*sizeof(*boxmax) );
            points += stride;
            count--;
          }
          if ( count > 0 ) {
            for ( /*empty*/; count--; points += stride ) {
              if ( points[dim] == 0.0 ) {
                rc = FALSE;
                continue;
              }
              w = 1.0/points[dim];
              for ( j = 0; j < dim; j++ ) {
                x = w*points[j];
                if (boxmin[j] > x) 
                  boxmin[j] = x; 
                else if (boxmax[j] < x) 
                  boxmax[j] = x;
              }
            }
          }
        }
      }
      else {
        // bounding box of non-rational points
        rc = TRUE;
        if ( !bGrowBox ) {
          // use first point to initialize box 
          memcpy( boxmin, points, dim*sizeof(*boxmin) );
          memcpy( boxmax, boxmin, dim*sizeof(*boxmax) );
          points += stride;
          count--;
        }
        if ( count ) {
          // grow box to contain the rest of the points
          for ( /*empty*/; count--; points += stride ) {
            for ( j = 0; j < dim; j++ ) {
              x = points[j];
              if (boxmin[j] > x) 
                boxmin[j] = x; 
              else if (boxmax[j] < x) 
                boxmax[j] = x;
            }
          }
        }
      }
    }
  }
  else if ( bGrowBox ) {
    // result is still valid if no points are added to a valid input box
    rc = TRUE;
  }
  return rc;
}

ON_DEFN
ON_BoundingBox ON_PointListBoundingBox(
    int dim, BOOL is_rat, int count, int stride, const float* points, 
    BOOL bGrowBox
    )
{
  ON_BoundingBox bbox;
  BOOL rc = FALSE;
  float* bboxmin = (float*)alloca(2*dim*sizeof(*bboxmin));
  float* bboxmax = bboxmin+dim;
  rc = ON_GetPointListBoundingBox( dim, is_rat, count, stride, points, bboxmin, bboxmax, bGrowBox );
  if ( rc ) {
    int i;
    for ( i = 0; i < dim && i < 3; i++ ) {
      bbox.m_min[i] = bboxmin[i];
      bbox.m_max[i] = bboxmax[i];
    }
  }
  return bbox;
}

ON_DEFN
BOOL ON_GetPointListBoundingBox( 
    int dim, BOOL is_rat, int count, int stride, const float* points, 
    float* boxmin, float* boxmax,
    BOOL bGrowBox
    )
/*****************************************************************************
Bounding Box of a set of points
 
INPUT:
  dim           ( >= 1 ) dimension of each point
  is_rat        ( TRUE if points are rational )
  count         ( >= 1 ) number of points
  stride        ( >= (is_rat)?(dim+1):dim )
  points        array of dim*count floats
  boxmin, boxmax      unused arrays of dim floats
  bGrowBox       TRUE if input box should be enlarged to contain points
                 FALSE if input box should be ignored bounding box of points
                       is returned
OUTPUT:
  boxmin, boxmax      diagonal corners of bounding box
*****************************************************************************/
{
  // bounding box workhorse
  float x, w;
  int j;
  BOOL rc = FALSE;
  if ( count > 0 ) {
    if ( is_rat )
      is_rat = 1;
    if ( points && dim > 0 && (count == 1 || stride >= dim+is_rat) ) {
      if ( is_rat ) {
        rc = TRUE;
        while ( count > 0 && points[dim] == 0.0 ) {
          count--;
          points += stride;
          rc = FALSE;
        }
        if ( count > 0 ) {
          if ( !bGrowBox ) {
            ON_ArrayScale( dim, 1.0f/points[dim], points, boxmin );
            memcpy( boxmax, boxmin, dim*sizeof(*boxmax) );
            points += stride;
            count--;
          }
          for ( /*empty*/; count--; points += stride ) {
            if ( points[dim] == 0.0 )
              continue;
            w = 1.0/points[dim];
            for ( j = 0; j < dim; j++ ) {
              x = w*points[j];
              if (boxmin[j] > x) 
                boxmin[j] = x; 
              else if (boxmax[j] < x) 
                boxmax[j] = x;
            }
          }
        }
      }
      else {
        rc = TRUE;
        if ( !bGrowBox ) {
          memcpy( boxmin, points, dim*sizeof(*boxmin) );
          memcpy( boxmax, boxmin, dim*sizeof(*boxmax) );
          points += stride;
          count--;
        }
        for ( /*empty*/; count--; points += stride ) {
          for ( j = 0; j < dim; j++ ) {
            x = points[j];
            if (boxmin[j] > x) 
              boxmin[j] = x; 
            else if (boxmax[j] < x) 
              boxmax[j] = x;
          }
        }
      }
    }
  }
  else if ( bGrowBox ) {
    rc = TRUE;
  }
  return rc;
}

ON_DEFN
ON_BoundingBox ON_PointGridBoundingBox(
        int dim,
        BOOL is_rat,
        int point_count0, int point_count1,
        int point_stride0, int point_stride1,
        const double* p,
        BOOL bGrowBox
    )
{
  ON_BoundingBox bbox;
  BOOL rc = FALSE;
  double* bboxmin = &bbox.m_min.x;
  double* bboxmax = &bbox.m_max.x;
  if ( dim > 3 ) {
    bboxmin = (double*)alloca(2*dim*sizeof(*bboxmin));
    bboxmax = bboxmin+dim;
  }
  rc = ON_GetPointGridBoundingBox( dim, is_rat, point_count0, point_count1, point_stride0, point_stride1, p, bboxmin, bboxmax, bGrowBox );
  if ( rc && dim > 3 ) {
    bbox.m_min = ON_3dPoint( bboxmin );
    bbox.m_max = ON_3dPoint( bboxmax );
  }
  return bbox;
}

ON_DEFN
BOOL ON_GetPointGridBoundingBox(
        int dim,
        BOOL is_rat,
        int point_count0, int point_count1,
        int point_stride0, int point_stride1,
        const double* p,
        double* boxmin, double* boxmax,
        BOOL bGrowBox
        )
{
  BOOL rc = (bGrowBox) ? TRUE : FALSE;
  int i;
  for ( i = 0; i < point_count0; i++ ) {
    if ( !ON_GetPointListBoundingBox( dim, is_rat, point_count1, point_stride1, p + i*point_stride0, boxmin, boxmax, bGrowBox ) ) {
      rc = FALSE;
      break;
    }
    else {
      bGrowBox = TRUE;
      if (!i)
        rc = TRUE;
    }
  }
  return rc;
}

double ON_BoundingBoxTolerance(
        int dim,
        const double* bboxmin,
        const double* bboxmax
        )
{
  int i;
  double x, tolerance=0.0;
  ON_ASSERT_OR_RETURN( dim > 0 && bboxmin != NULL && bboxmax != NULL,tolerance);
  for ( i = 0; i < dim; i++ ) {
    ON_ASSERT_OR_RETURN(bboxmin[i] <= bboxmax[i],tolerance);
  }
  tolerance = ON_ArrayDistance(dim,bboxmin,bboxmax)*ON_EPSILON;
  for ( i = 0; i < dim; i++ ) {
    x = (bboxmax[i] - bboxmin[i])*ON_SQRT_EPSILON;
    if ( x > tolerance )
      tolerance = x;
    x = (fabs(bboxmax[i]) - fabs(bboxmin[i]))*ON_EPSILON;
    if ( x > tolerance )
      tolerance = x;
  }
  if ( tolerance > 0.0 && tolerance < ON_ZERO_TOLERANCE )
    tolerance = ON_ZERO_TOLERANCE;
  return tolerance;
}
