/* $Header: /src2/opennurbs/opennurbs_bounding_box.h 10    7/11/01 4:53p Dalelear $ */
/* $NoKeywords: $ */
/*
//
// Copyright (c) 1993-2001 Robert McNeel & Associates. All rights reserved.
// Rhinoceros is a registered trademark of Robert McNeel & Assoicates.
//
// THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
// ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
// MERCHANTABILITY ARE HEREBY DISCLAIMED.
//				
// For complete openNURBS copyright information see <http://www.opennurbs.org>.
//
////////////////////////////////////////////////////////////////
*/

#if !defined(ON_BOUNDING_BOX_INC_)
#define ON_BOUNDING_BOX_INC_


////////////////////////////////////////////////////////////////
//
//   ON_BoundingBox  - axis aligned bounding box
//

class ON_CLASS ON_BoundingBox
{
public:
  ON_BoundingBox();
	ON_BoundingBox(
    const ON_3dPoint&, // min corner of axis aligned bounding box
    const ON_3dPoint&  // max corner of axis aligned bounding box
    );
  ~ON_BoundingBox();

  void Destroy(); // invalidates bounding box

  // operator[] returns min if index <= 0 and max if indes >= 1
  ON_3dPoint& operator[](int);
  const ON_3dPoint& operator[](int) const;

  ON_3dPoint Min() const;
  ON_3dPoint Max() const;
  ON_3dVector Diagonal() const; // max corner - min corner
  ON_3dPoint Center() const;
  ON_3dPoint Corner( // 8 corners of box
    int, // x_index   0 = Min().x, 1 = Max().x
    int, // y_index   0 = Min().y, 1 = Max().y
    int  // z_index   0 = Min().z, 1 = Max().z
    ) const;
  BOOL GetCorners( 
    ON_3dPointArray& // returns list of 8 corner points
    ) const;

  BOOL IsValid() const; // empty boxes are not valid

  //////////
  // ON_BoundingBox::Transform() updates the bounding box
  // to be the smallest axis aligned bounding box that contains
  // the transform of the eight corner points of the input
  // bounding box.
  BOOL Transform( const ON_Xform& );

  double Tolerance() const; // rough guess at a tolerance to use for comparing
                            // objects in this bounding box


  // All of these Set() functions set or expand a box to enclose the points in the arguments
  // If bGrowBox is TRUE, the existing box is expanded, otherwise it is only set to the current point list
  BOOL Set(     
    int,           // dim, 
    BOOL,          // is_rat, 
    int,           // count, 
    int,           // stride, 
    const double*, // points, 
    BOOL = FALSE    // bGrowBox
    );

  BOOL Set(
    const ON_3dPoint&,
    BOOL = FALSE // bGrowBox
    );

  BOOL Set(     
    const ON_4dPointArray&,
    BOOL = FALSE    // bGrowBox
    );

  BOOL Set(     
    const ON_3dPointArray&,
    BOOL = FALSE    // bGrowBox
    );

  BOOL Set(     
    const ON_2dPointArray&,
    BOOL = FALSE    // bGrowBox
    );

	BOOL IsPointIn(
         const ON_3dPoint&, // point to test
         BOOL = FALSE        // bStrictlyIn
                             // TRUE to test for strict ( min < point < max )
                             // FALSE to test for (min <= point <= max)
                             //       
         ) const;

  //////////
  // Point on or in the box that is closest to test_point.
  // If test_point is in or on the box, the test_point is returned.
  ON_3dPoint ClosestPoint( 
    const ON_3dPoint& // test_point
    ) const;

  //////////
  // Get points on bounding boxes that are closest to each other.
  // If the boxes intersect, then the point at the centroid of the
  // intersection is returned for both points.
  BOOL GetClosestPoint( 
         const ON_BoundingBox&, // "other" bounding box
         ON_3dPoint&, // point on "this" box that is closest to "other" box
         ON_3dPoint&  // point on "other" box that is closest to "this" box
         )  const;

  //////////
  // Point on the box that is farthest from the test_point.
  ON_3dPoint FarPoint( 
    const ON_3dPoint& // test_point
    ) const;

  //////////
  // Get points on bounding boxes that are farthest from each other.
  BOOL GetFarPoint( 
         const ON_BoundingBox&, // "other" bounding box
         ON_3dPoint&, // point on "this" box that is farthest from "other" box
         ON_3dPoint&  // point on "other" box that is farthest from "this" box
         )  const;

  // Intersection() returns TRUE if intersection is not empty
  // Invalid boxes are treated as the empty set.
  BOOL Intersection( // this = this intersect arg
         const ON_BoundingBox&
         );

  BOOL Intersection( // this = intersection of two args
         const ON_BoundingBox&, 
         const ON_BoundingBox&
         );

  double Volume() const;

  double Area() const;

  // Union() returns TRUE if union is not empty.
  // Invalid boxes are treated as the empty set.
  BOOL Union( // this = this union arg
         const ON_BoundingBox&
         );

  BOOL Union( // this = union of two args
         const ON_BoundingBox&, 
         const ON_BoundingBox&
         );
                  
  // IsDisjoint() returns TRUE if both boxes are valid and intersection is empty
	BOOL IsDisjoint(
    const ON_BoundingBox&
    ) 
    const; // TRUE if boxes are disjoint

  BOOL SwapCoordinates( int, int );

  ON_3dPoint m_min;
  ON_3dPoint m_max;
};

ON_DECL
BOOL ON_GetPointListBoundingBox( // low level workhorse function
    int,           // dimension of points
    BOOL,          // TRUE for homogeneous rational points
    int,           // number of points
    int,           // stride between points
    const double*, // point array
    double*,       // min[dim]
    double*,       // max[dim]
    BOOL           // bGrowBox
    );

ON_DECL
ON_BoundingBox ON_PointListBoundingBox(
    int,           // dimension of points
    BOOL,          // TRUE for homogeneous rational points
    int,           // number of points
    int,           // stride between points
    const double*, // point array
    BOOL           // bGrowBox
    );

ON_DECL
BOOL ON_GetPointListBoundingBox( // low level workhorse function
    int,           // dimension of points
    BOOL,          // TRUE for homogeneous rational points
    int,           // number of points
    int,           // stride between points
    const float*, // point array
    float*,       // min[dim]
    float*,       // max[dim]
    BOOL           // bGrowBox
    );

ON_DECL
ON_BoundingBox ON_PointListBoundingBox( // low level workhorse function
    int,           // dimension of points
    BOOL,          // TRUE for homogeneous rational points
    int,           // number of points
    int,           // stride between points
    const float*, // point array
    BOOL           // bGrowBox
    );

ON_DECL
BOOL ON_GetPointGridBoundingBox(
        int,           // dim
        BOOL,          // TRUE for homogeneous rational points
        int, int,      // point_count0, point_count1,
        int, int,      // point_stride0, point_stride1,
        const double*, // point grid
        double*,       // min[dim]
        double*,       // max[dim]
        BOOL           // bGrowBox
    );

ON_DECL
ON_BoundingBox ON_PointGridBoundingBox(
        int,           // dim
        BOOL,          // TRUE for homogeneous rational points
        int, int,      // point_count0, point_count1,
        int, int,      // point_stride0, point_stride1,
        const double*, // point grid
        BOOL           // bGrowBox
    );

double ON_BoundingBoxTolerance(
        int dim,
        const double* bboxmin,
        const double* bboxmax
        );

#endif
