/* $Header: /src2/opennurbs/opennurbs_brep.h 27    7/09/01 11:15a Dalelear $ */
/* $NoKeywords: $ */
/*
//
// Copyright (c) 1993-2001 Robert McNeel & Associates. All rights reserved.
// Rhinoceros is a registered trademark of Robert McNeel & Assoicates.
//
// THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
// ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
// MERCHANTABILITY ARE HEREBY DISCLAIMED.
//				
// For complete openNURBS copyright information see <http://www.opennurbs.org>.
//
////////////////////////////////////////////////////////////////
*/

////////////////////////////////////////////////////////////////
//
//   Definition of b-rep and its parts
//
////////////////////////////////////////////////////////////////

#if !defined(OPENNURBS_BREP_INC_)
#define OPENNURBS_BREP_INC_

class ON_CLASS ON_BrepVertex : public ON_Point
{
  ON_OBJECT_DECLARE(ON_BrepVertex);
public:
  int m_vertex_user_i; // integer available for application use.
                       // Constructor initializes m_user to 0.
                       // Value is not serialized and may be changed
                       // by computations.

  int m_vertex_index;  // index of vertex in ON_Brep.m_V[] array

  ON_BrepVertex();
  ON_BrepVertex(int); // vertex index
  ON_BrepVertex& operator=(const ON_BrepVertex&);


  /////////////////////////////////////////////////////////////////
  // ON_Object overrides

  BOOL IsValid() const;

  void Dump( ON_TextLog& ) const; // for debugging

  BOOL Write( ON_BinaryArchive& ) const;

  BOOL Read( ON_BinaryArchive& );

  /////////////////////////////////////////////////////////////////
  // Interface
  BOOL SetPoint( const ON_3dPoint& );
  ON_3dPoint Point() const;

  // (use ON_Brep::SetTolerance( ON_BrepVertex& ) to set tolerance
  double Tolerance() const;

  int EdgeCount() const;

  /////////////////////////////////////////////////////////////////
  // Implementation
  ON_SimpleArray<int> m_ei;   // indices of edges starting/ending at this vertex
                       //
                       // For closed edges, edge.m_vi[0] = edge.m_vi[1] and 
                       // edge.m_edge_index appears twice in the m_ei[] array.
                       // The first occurance of edge.m_edge_index in m_ei[]
                       // is for the closed edge starting the vertex.
                       // The second occurance of edge,m_edge_index in m_ei[]
                       // is for the closed edge ending at the vertex.
                       // C.f. ON_Brep::Next/PrevEdge().

  double m_tolerance;  // accuracy of vertex point (>=0.0 or ON_UNSET_VALUE)
                       //
                       // A value of ON_UNSET_VALUE indicates that the
                       // tolerance should be computed.
                       //
                       // A value of 0.0 indicates that the distance
                       // from the vertex to any applicable edge or trim
                       // end is <=  ON_ZERO_TOLERANCE
                       //
                       // If an edge begins or ends at this vertex,
                       // then the distance from the vertex's 
                       // 3d point to the appropriate end of the
                       // edge's 3d curve must be <= this tolerance.
                       //
                       // If a trim begins or ends at this vertex,
                       // then the distance from the vertex's 3d point
                       // to the 3d point on the surface obtained by
                       // evaluating the surface at the appropriate
                       // end of the trimming curve must be <= this
                       // tolerance.

private:
  ON_BrepVertex( const ON_BrepVertex& ); // no implementation

};

///////////
// An ON_BrepEdge is derived from ON_CurveProxy as a convienience 
// so that the edge can supply easy to use evaluation tools via 
// the ON_Curve virtual member functions the ON_ProxyCurve overrides.
// The ON_CurveProxy member variables of the ON_BrepEdge are always
// set as follows:
//    m_curve = m_c3[edge.m_c3i]
//    m_bReversed = FALSE,
//    m_domain = a sub-interval of m_C3[edge.m_c3i]->Domain()
class ON_CLASS ON_BrepEdge : public  ON_CurveProxy
{
  ON_OBJECT_DECLARE(ON_BrepEdge);
public:
  int m_edge_user_i;   // integer available for application use.
                       // Constructor initializes m_user to 0.
                       // Value is not serialized and may be changed
                       // by computations.

  int m_edge_index;    // index of edge in ON_Brep.m_E[] array

  ON_BrepEdge();
  ON_BrepEdge(int); // edge index
  ON_BrepEdge& operator=(const ON_BrepEdge&);

  /////////////////////////////////////////////////////////////////
  // ON_Object overrides
  //
  BOOL IsValid() const;

  void Dump( ON_TextLog& ) const; // for debugging

  BOOL Write( ON_BinaryArchive& ) const;

  BOOL Read( ON_BinaryArchive& );

  /////////////////////////////////////////////////////////////////
  // Implementation

  int m_c3i;           // index of 3d curve in m_C3[] array
                       // (edge.m_curve also points to m_C3[m_c3i])
  int m_vi[2];         // indices of starting/ending vertex
                       //
                       // For closed edges, m_vi[0] = m_vi[1] and m_edge_index
                       // appears twice in the m_V[m_vi[0]].m_ei[] array.
                       // The first occurance of m_edge_index in m_V[m_vi[0]].m_ei[]
                       // is for the closed edge starting the vertex.  The second
                       // occurance of m_edge_index in m_V[m_vi[0]].m_ei[]
                       // is for the closed edge edge ending at the vertex.
                       // C.f. ON_Brep::Next/PrevEdge().

  ON_SimpleArray<int> m_ti;   // indices of Trims that use this edge

  double m_tolerance;  // accuracy of edge curve (>=0.0 or ON_UNSET_VALUE)
                       //
                       // A value of ON_UNSET_VALUE indicates that the
                       // tolerance should be computed.
                       //
                       // The maximum distance from the edge's 3d curve
                       // to any surface of a face that has this edge as
                       // a portion of its boundary must be <= this
                       // tolerance.

  int m_selected_trim; // if edge is selected as an edge of a face, this
                       // index is set to the trim_index for that face.

private:
  friend class ON_Brep;
  ON_Brep* m_brep; // so isolated edge class edge can get at it's 3d curve
  ON_BrepEdge( const ON_BrepEdge& ); // no implementation
};

struct ON_BrepTrimPoint
{
  ON_2dPoint p; // 2d surface paramter space point
  double t;     // corresponding trim curve parameter
  double e;     // corresponding edge curve parameter (ON_UNSET_VALUE if unknown)
};

#if defined(ON_DLL_EXPORTS) || defined(ON_DLL_IMPORTS)
// This stuff is here because of a limitation in the way Microsoft
// handles templates and DLLs.  See Microsoft's knowledge base 
// article ID Q168958 for details.
#pragma warning( push )
#pragma warning( disable : 4231 )
ON_TEMPLATE template class ON_CLASS ON_SimpleArray<ON_BrepTrimPoint>;
#pragma warning( pop )
#endif


class ON_CLASS ON_BrepTrim : public  ON_Object
{
  ON_OBJECT_DECLARE(ON_BrepTrim);
public:
  int m_trim_user_i;   // integer available for application use.
                       // Constructor initializes m_trim_user_i to 0.
                       // Value is not serialized and may be changed
                       // by computations.

  int m_trim_index;  // index of trim in ON_Brep.m_T[] array

  enum TYPE {
    unknown  = 0,
    boundary = 1,       // trim has no mate and is connected to edge
    mated    = 2,       // trim shares edge with trim(s) on other face(s)
    seam     = 3,       // trim shares edge with 1 other trim on same face
    singular = 4,       // trim is along singular side of surface (no edge)
    crvonsrf = 5,       // "trim" is a curve on surface (a.k.a "snake")
    type_count = 6
  };

  ON_BrepTrim();
  ON_BrepTrim(int); // trim index
  ON_BrepTrim& operator=(const ON_BrepTrim&);

  /////////////////////////////////////////////////////////////////
  // ON_Object overrides
  //
  // (Trims are purely topologicial - geometry queries should be 
  //  directed at the trim's 2d curve or the trim's edge's 3d curve.)

  BOOL IsValid() const;

  void Dump( ON_TextLog& ) const; // for debugging

  BOOL Write( ON_BinaryArchive& ) const;

  BOOL Read( ON_BinaryArchive& );

  /////////////////////////////////////////////////////////////////
  // Interface
  
  int    m_c2i;         // index of 2d curve
  ON_Interval m_t;      // increasing (sub)domain of 2d curve used for trim geometry
  int    m_ei;          // index of 3d edge (-1 if ON_BrepTrim is singular)
  int    m_vi[2];       // indices of start/end vertices.  Trims along singular
                        // sides and trims that correspond to closed 3d edges
                        // have m_vi[0] = m_vi[1].  Note that singular trims
                        // and trims on the closed edge of a closed surface can
                        // have an open 2d trimming curve and still have 
                        // m_vi[0] = m_vi[1].
        
  BOOL   m_bRev3d;      // TRUE if the 2d trim and 3d edge have
                        //      opposite orientations.
  TYPE   m_type;
  ON_Surface::ISO  m_iso;

  int    m_li;          // index of loop that uses this trim
  double m_tolerance[2]; // each value >= 0.0 or ON_UNSET_VALUE
                         //
                         // A value of ON_UNSET_VALUE indicates that the 
                         // tolerance should be computed.
                         //
                         // m_tolerance[0] = accuracy of parameter space curve
                         //                  in first parameter
                         // m_tolerance[1] = accuracy of parameter space curve
                         //                  in second parameter
                         //
                         // If the trim is not singular, then the trim must
                         // have an edge.  If P is a 3d point on the edge's
                         // curve and surface(u,v) = Q is the point on the 
                         // surface that is closest to P, then there must
                         // be a parameter t in the interval [m_t[0], m_t[1]]
                         // such that
                         //   |u - curve2d(t)[0]| <= m_tolerance[0]
                         // and 
                         //   |v - curve2d(t)[1]| <= m_tolerance[1]
                         //
                         // If P is the 3d point for the vertex brep.m_V[m_vi[k]]
                         // and (uk,vk) is the corresponding end of the trim's
                         // parameter space curve, then there must be a surface
                         // parameter (u,v) such that:
                         // *) the distance from the 3d point surface(u,v) to P
                         //    is <= brep.m_V[m_vi[k]].m_tolerance,
                         // *) |u-uk| <= m_tolerance[0], and
                         // *) |v-vk| <= m_tolerance[1].

  //////////
  // Removed July 3, 2001 - Use ON_Brep::GetTrim3dStart() and ON_Brep::GetTrim3dStart()
  // ON_3dPoint m_P[2];      // 3d locations of trim ends.

  //////////
  // polyline approximation of trimming curve
  // runtime information - not saved
  ON_SimpleArray<ON_BrepTrimPoint> m_pline;

  //////////
  // parameter space trimming curve bounding box
  // runtime information - not saved
  ON_BoundingBox m_pbox;

public:
  // values stored in legacy file formats - ignore
  void m__legacy_flags_Set(int,int);   // used internally - ignore
  BOOL m__legacy_flags_Get(int*,int*) const; // used internally - ignore
  double m__legacy_2d_tol; // used internally - ignore
  double m__legacy_3d_tol; // used internally - ignore
  int    m__legacy_flags;  // used internally - ignore
};

class ON_CLASS ON_BrepLoop : public  ON_Object
{
  ON_OBJECT_DECLARE(ON_BrepLoop);
public:
  int m_loop_user_i;   // integer available for application use.
                       // Constructor initializes m_user to 0.
                       // Value is not serialized and may be changed
                       // by computations.

  int m_loop_index;  // index of loop in ON_Brep.m_L[] array

  enum TYPE {
    unknown  = 0,
    outer    = 1,  // 2d loop curves form a simple closed curve with a counterclockwise orientation
    inner    = 2,  // 2d loop curves form a simple closed curve with a clockwise orientation
    slit     = 3,  // always closed - used internally during splitting operations
    crvonsrf = 4,  // "loop" is a curveonsrf (a.k.a "snake") made from a single 
                   // (open or closed) trim that is has type ON_BrepTrim::crvonsrf.
    type_count = 5
  };

  ON_BrepLoop();
  ON_BrepLoop(int); // loop index
  ON_BrepLoop& operator=(const ON_BrepLoop&);

  /////////////////////////////////////////////////////////////////
  // ON_Object overrides
  //
  // (Trims are purely topologicial - geometry queries should be 
  //  directed at the trim's 2d curve or the trim's edge's 3d curve.)

  BOOL IsValid() const;

  void Dump( ON_TextLog& ) const; // for debugging

  BOOL Write( ON_BinaryArchive& ) const;

  BOOL Read( ON_BinaryArchive& );

  /////////////////////////////////////////////////////////////////
  // Interface

  //////////
  // Returns the index i such that loop.m_ti[i] = trim.m_trim_index.
  // Returns -1 if the trim is not in this loop
  int IndexOfTrim( const ON_BrepTrim& ) const;


  ON_SimpleArray<int> m_ti;   // trim indices
  TYPE         m_type;
  int          m_fi;   // index of face that uses this loop

  //////////
  // parameter space trimming loop bounding box
  // runtime information - not saved
  ON_BoundingBox m_pbox;
};

class ON_CLASS ON_BrepFace : public ON_SurfaceProxy
{
  ON_OBJECT_DECLARE(ON_BrepFace);
public:
  int m_face_user_i;   // integer available for application use.
                       // Constructor initializes m_user to 0.
                       // Value is not serialized and may be changed
                       // by computations.

  int m_face_index;  // index of face in ON_Brep.m_F[] array

  ON_BrepFace();
  ~ON_BrepFace();
  ON_BrepFace(int);
  ON_BrepFace& operator=(const ON_BrepFace&);

  const ON_Brep* Brep() const;

  /////////////////////////////////////////////////////////////////
  // ON_Object overrides
  //
  // (Faces are purely topologicial - geometry queries should be 
  //  directed at the face's 3d surface.)

  BOOL IsValid() const;

  void Dump( ON_TextLog& ) const; // for debugging

  BOOL Write( ON_BinaryArchive& ) const;

  BOOL Read( ON_BinaryArchive& );

  /////////////////////////////////////////////////////////////////
  // Rendering Interface

  int MaterialIndex() const; // if -1, use parent's material definition
  void SetMaterialIndex(int);

  ON_Mesh* Mesh( ON::mesh_type ) const;
  void DestroyMesh( ON::mesh_type );

  /////////////////////////////////////////////////////////////////
  // "Expert" Interface

  ON_SimpleArray<int> m_li; // loop indices (outer loop is m_li[0])
  int m_si;            // index of surface in b-rep m_S[] array
  BOOL m_bRev;         // TRUE if face orientation is opposite
                       //      of natural surface orientation

private:
  ON_BoundingBox m_bbox;      // 3d bounding box
  ON_Interval    m_domain[2]; // rectangular bounds of 2d curves
  ON_Mesh* m_render_mesh;
  ON_Mesh* m_analysis_mesh;
  ON_Mesh* m_preview_mesh;
  int m_material_index; // if 0 (default), ON_Brep's object attributes
                        // determine material.
private:
  friend class ON_Brep;
  ON_Brep* m_brep; // so isolated edge class edge can get at it's 3d curve
  ON_BrepFace( const ON_BrepFace& ); // no implementation
};

#if defined(ON_DLL_EXPORTS) || defined(ON_DLL_IMPORTS)
// This stuff is here because of a limitation in the way Microsoft
// handles templates and DLLs.  See Microsoft's knowledge base 
// article ID Q168958 for details.
#pragma warning( push )
#pragma warning( disable : 4231 )
ON_TEMPLATE template class ON_CLASS ON_ClassArray<ON_BrepVertex>;
ON_TEMPLATE template class ON_CLASS ON_ClassArray<ON_BrepEdge>;
ON_TEMPLATE template class ON_CLASS ON_ClassArray<ON_BrepTrim>;
ON_TEMPLATE template class ON_CLASS ON_ClassArray<ON_BrepLoop>;
ON_TEMPLATE template class ON_CLASS ON_ClassArray<ON_BrepFace>;
#pragma warning( pop )
#endif

class ON_CLASS ON_BrepVertexArray : public ON_ClassArray<ON_BrepVertex>
{
public:
  ON_BrepVertexArray();
  ~ON_BrepVertexArray();

  BOOL Read( ON_BinaryArchive& );
  BOOL Write( ON_BinaryArchive& ) const;
};

class ON_CLASS ON_BrepEdgeArray   : public  ON_ClassArray<ON_BrepEdge>
{
public:
  ON_BrepEdgeArray();
  ~ON_BrepEdgeArray();
  BOOL Read( ON_BinaryArchive& );
  BOOL Write( ON_BinaryArchive& ) const;
};

class ON_CLASS ON_BrepTrimArray   : public  ON_ClassArray<ON_BrepTrim>
{
public:
  ON_BrepTrimArray();
  ~ON_BrepTrimArray();
  BOOL Read( ON_BinaryArchive& );
  BOOL Write( ON_BinaryArchive& ) const;
};

class ON_CLASS ON_BrepLoopArray   : public  ON_ClassArray<ON_BrepLoop>
{
public:
  ON_BrepLoopArray();
  ~ON_BrepLoopArray();
  BOOL Read( ON_BinaryArchive& );
  BOOL Write( ON_BinaryArchive& ) const;
};

class ON_CLASS ON_BrepFaceArray   : public  ON_ClassArray<ON_BrepFace>
{
public:
  ON_BrepFaceArray();
  ~ON_BrepFaceArray();
  BOOL Read( ON_BinaryArchive& );
  BOOL Write( ON_BinaryArchive& ) const;
};


class ON_CLASS ON_Brep : public ON_Geometry 
{	 
  ON_OBJECT_DECLARE(ON_Brep);
public:

	// Constructors/ Destructors  - Constructors are protected 
  ON_Brep();
	~ON_Brep();		
  ON_Brep(const ON_Brep&);
  ON_Brep& ON_Brep::operator=(const ON_Brep&);

  /////////////////////////////////////////////////////////////////
  // construction/destruction helpers

  void Destroy(); // returns Brep to state it has after default construction
  void EmergencyDestroy(); // call if memory pool used by b-rep members becomes invalid
  void DestroyMesh( ON::mesh_type );

  // Create a b-rep from a NURBS surface.  If Create() returns TRUE,
  // then the surface pointer is returned NULL, the B-rep owns the surface
  // and the surface ill be destroyed by a delete call in ~ON_Brep().
  BOOL Create( ON_NurbsSurface*& );

  /////////////////////////////////////////////////////////////////
  // ON_Object overrides
  BOOL IsValid() const;

  void Dump( ON_TextLog& ) const; // for debugging

  BOOL Write( ON_BinaryArchive& ) const;

  BOOL Read( ON_BinaryArchive& );

  ON::object_type ObjectType() const;

  /////////////////////////////////////////////////////////////////
  // ON_Geometry overrides
  int Dimension() const;

  BOOL GetBBox( // returns TRUE if successful
         double*,    // minimum
         double*,    // maximum
         BOOL = FALSE  // TRUE means grow box
         ) const;

  BOOL Transform( 
         const ON_Xform&
         );

  BOOL SwapCoordinates(
        int, int        // indices of coords to swap
        );

  //////////
  // overrides ON_Geometry::HasBrepForm() and returns TRUE
  BOOL HasBrepForm() const;

  //////////
  // overrides ON_Geometry::BrepForm() and simply returns a 
  // duplicate of this brep.
  ON_Brep* BrepForm() const;

  /////////////////////////////////////////////////////////////////
  // Creation Interface

  // These add a new geometry piece to the b-rep and return the
  // index that should be used to reference the geometry.
  // -1 is returned if the input is not acceptable.
  // ~Tl2_Brep() will delete the geometry.
  int AddTrimCurve( ON_Curve* ); // 2d curve used by ON_BrepTrim
  int AddEdgeCurve( ON_Curve* ); // 3d curve used by ON_BrepEdge
  int AddSurface( ON_Surface* ); // 3d surface used by ON_BrepFace

  // These add a new topology piece of the b-rep and return a 
  // reference that is intended to be used for initialization.
  ON_BrepVertex& NewVertex();
  ON_BrepVertex& NewVertex( const ON_3dPoint& );

  ON_BrepEdge& NewEdge(
                  int = -1              // 3d curve index
                  );
  ON_BrepEdge& NewEdge( 
                  ON_BrepVertex&, // start vertex
                  ON_BrepVertex&, // end vertex
                  int = -1              // 3d curve index
                  );

  ON_BrepFace& NewFace(
                  int = -1              // surfce index
                  );

  ON_BrepLoop& NewLoop( ON_BrepLoop::TYPE );
  ON_BrepLoop& NewLoop( ON_BrepLoop::TYPE, ON_BrepFace& );

  ON_BrepTrim& NewTrim(
                  int = -1              // 2d curve index
                  );
  ON_BrepTrim& NewTrim(
                  BOOL,         // TRUE if edge and trim have opposite directions
                  ON_BrepLoop&,
                  int = -1      // 2d curve index
                  );
  ON_BrepTrim& NewTrim(
                  ON_BrepEdge&,
                  BOOL,         // TRUE if edge and trim have opposite directions
                  int = -1              // 2d curve index
                  );
  ON_BrepTrim& NewTrim(
                  ON_BrepEdge&,
                  BOOL,         // TRUE if edge and trim have opposite directions
                  ON_BrepLoop&,
                  int = -1      // 2d curve index
                  );

  void Append( const ON_Brep& ); // appends a copy of brep to this and updates
                                 // indices of appended brep parts

  // This function can be used to compute vertex information for a
  // b-rep when everything but the m_V array is properly filled in.
  // It is intended to be used when creating a ON_Brep from a 
  // definition that does not include explicit vertex information.
  void SetVertices(void);

  // This function can be used to set the ON_BrepTrim::m_iso
  // flag. It is intended to be used when creating a ON_Brep from
  // a definition that does not include compatible parameter space
  // type information.
  BOOL SetTrimIsoFlags();    // sets all trim iso flags
  BOOL SetTrimIsoFlags( ON_BrepFace& );
  BOOL SetTrimIsoFlags( ON_BrepLoop& );
  BOOL SetTrimIsoFlags( ON_BrepTrim& );

  // This function can be used to set the ON_BrepTrim::m_type
  // flag.  If the optional bLazy argument is TRUE, then only
  // trims with m_type = unknown are set.
  BOOL SetTrimTypeFlags( BOOL = FALSE );    // sets all trim iso flags
  BOOL SetTrimTypeFlags( ON_BrepFace&, BOOL = FALSE );
  BOOL SetTrimTypeFlags( ON_BrepLoop&, BOOL = FALSE );
  BOOL SetTrimTypeFlags( ON_BrepTrim&, BOOL = FALSE );

  //////////
  // GetTrim3dStart() evaluates the 3d surface at the start of the
  // parameter space (2d) trim curve.
  BOOL GetTrim3dStart( 
          int,         // index of ON_BrepTrim in m_T[] array
          ON_3dPoint& 
          ) const;

  //////////
  // GetTrim3dEnd() evaluates the 3d surface at the start of the
  // parameter space (2d) trim curve.
  BOOL GetTrim3dEnd(
          int,         // index of ON_BrepTrim in m_T[] array
          ON_3dPoint& 
          ) const;

  // This function examines the 2d parameter space curves and returns
  // the loop's type based on their orientation.  Use this function for
  // debugging loop orientation problems.
  ON_BrepLoop::TYPE ComputeLoopType( const ON_BrepLoop& ) const;

  // These set the various tolerances.  The optional BOOL argument
  // is called bLazy.  If bLazy is FALSE, the tolerance is recomputed
  // from its definition.  If bLazy is TRUE, the tolerance is computed
  // only if its current value is negative.
  BOOL SetTolerance( ON_BrepVertex&, BOOL = FALSE ) const;
  BOOL SetTolerance( ON_BrepEdge&, BOOL = FALSE ) const;

  /////////////////////////////////////////////////////////////////
  // Query Interface

  // 3d surface associated with b-rep object
  int SurfaceIndexOf( const ON_BrepTrim& ) const;
  int SurfaceIndexOf( const ON_BrepLoop& ) const;
  int SurfaceIndexOf( const ON_BrepFace& ) const; 
  ON_Surface* SurfaceOf( const ON_BrepTrim& ) const;
  ON_Surface* SurfaceOf( const ON_BrepLoop& ) const;
  ON_Surface* SurfaceOf( const ON_BrepFace& ) const; 

  // 3d edge curve associated with b-rep object
  int EdgeCurveIndexOf( const ON_BrepTrim& ) const;
  int EdgeCurveIndexOf( const ON_BrepEdge& ) const;
  ON_Curve* EdgeCurveOf( const ON_BrepTrim& ) const;
  ON_Curve* EdgeCurveOf( const ON_BrepEdge& ) const;

  // 2d trim curve associated with b-rep object
  int TrimCurveIndexOf( const ON_BrepTrim& ) const;
  ON_Curve* TrimCurveOf( const ON_BrepTrim& ) const;

  BOOL IsSolid() const; // returns TRUE if b-rep is an oriented manifold solid
  
  BOOL IsManifold( // returns TRUE if b-rep is an oriented manifold
    BOOL* = NULL,  // TRUE  is returned if b-rep is an oriented manifold
                   // FALSE is returned if b-rep is not a manifold or
                   //       is not an orientable manifold
    BOOL* = NULL   // TRUE  is returned if b-rep is a manifold with boundary
                   // FALSE is returned if b-rep is not a manifold or
                   //       is a manifold with no boundary
    ) const;

  BOOL IsSurface() const;      // returns TRUE if the b-rep has a single face
                               // and that face is geometrically the same
                               // as the underlying surface.  I.e., the face
                               // has trivial trimming.  In this case, the
                               // surface is m_S[0].
                               // The flag m_F[0].m_bRev records
                               // the correspondence between the surface's
                               // natural parametric orientation and the
                               // orientation of the b-rep.


  BOOL FaceIsSurface(          // returns TRUE if the face has a single
         int // index of face  // outer boundary and that boundary runs
         ) const;              // along the edges of the underlying surface.
                               // In this case the geometry of the surface
                               // is the same as the geometry of the face.
                               // If FaceIsSurface() is TRUE, then
                               // m_S[m_F[face_index].m_si] is the surface.
                               // The flag m_F[face_index].m_bRev records
                               // the correspondence between the surface's
                               // natural parametric orientation and the
                               // orientation of face in the b-rep.

  BOOL LoopIsSurfaceBoundary(  // returns TRUE if the loop's trims all run
         int // index of loop  // along the edge's of the underlying surface's
         ) const;              // parameter space.

  /////////////////////////////////////////////////////////////////
  // Modification Interface

  //////////
  // Clears all ON_BrepFace.m_bRev flags by calling 
  // ON_Brep::SwapFaceParameters() on each face with a TRUE m_bRev.
  BOOL FlipReversedSurfaces();

  //////////
  // Change the domain of a trim's 2d curve.  This changes only the
  // parameterization of the 2d trimming curve; the locus of the 
  // 2d trimming curve is not changed.
  BOOL SetTrimDomain(
         int, // index of trim in m_T[] array
         const ON_Interval&
         );

  //////////
  // Change the domain of an edge.  This changes only the
  // parameterization of the 3d edge curve; the locus of the 
  // 3d edge curve is not changed.
  BOOL SetEdgeDomain(
         int, // index of edge in m_E[] array
         const ON_Interval&
         );

  //////////
  // Change the domain of a face
  // This changes the parameterization of the face's surface and transforms
  // the "u" and "v" coordinates of all the face's parameter space trimming
  // curves.  The locus of the face is not changed.
  BOOL SetFaceDomain(
         int, // index of face in m_F[] array
         const ON_Interval&,
         const ON_Interval&
         );

  //////////
  // Reverse the "u" or "v" parameter of a face's surface
  // domain. This also reverses the "u" or "v" coordinates of all the 
  // face's parameter space trimming curves, reverses all the face's
  // trimming loops, and toggles the value of ON_BrepFace::m_bRev.  
  // In particular, the abstract orientation of the face is not 
  // changed, just the underlying surface geometry paremterization
  // is changed.
  BOOL ReverseFaceParameter(   // swaps "u" and "v" parameters of face's
        int, // index of face
        int // dir = 0 reverse "u", 1 reverse "v"
        );

  //////////
  // Swap (transpose) the "u" and "v" parameters of a face's surface
  // domain. This also swaps the "u" and "v" coordinates of all the 
  // face's parameter space trimming curves, reverses all the face's
  // trimming loops, and toggles the value of ON_BrepFace::m_bRev.  
  // In particular, the abstract orientation of the face is not 
  // changed, just the underlying surface geometry paremterization
  // is changed.
  BOOL SwapFaceParameters(
        int // index of face
        );

  //////////
  // Reverses entire brep orientation of all faces by toggling 
  // value of all face's ON_BrepFace::m_bRev flag.
  void Flip();

  //////////
  // reverses orientation of an edge
  void FlipEdge(ON_BrepEdge&); // reverses orientation of edge

  //////////
  // reverses orientation of a face by toggling ON_BrepFace::m_bRev
  void FlipFace(ON_BrepFace&);

  //////////
  // Reverses orientation of a trims 2d curve and associated flags.
  // This function is intended to be used by brep experts and does
  // not modify any neighboring trims or the parent loop.
  void FlipTrim(ON_BrepTrim&);

  //////////
  // Reverses orientation of trimming loop. 
  // This function is intended to be used by brep experts and does
  // does NOT modify ON_BrepLoop::m_type.  You should make sure 
  // ON_BrepLoop::m_type jibes with the loop's direction.  (Outer loops
  // should be counter-clockwise and inner loops should be clockwise.)
  // You can use ON_Brep::LoopDirection() to determine the direction of
  // a loop.
  void FlipLoop(ON_BrepLoop&); // reverses orientation of trimming loop

  //////////
  // LoopDirection() examines the 2d trimming curve geometry that defines
  // the loop  and returns
  //   +1 if the loop is a counter-clockwise loop.
  //   -1 if the loop is a clockwise loop.
  //    0 if the loop is not a continuous closed loop.
  // Since LoopDirection() calculates its result based on the 2d trimming
  // curve geometry, it can be use to set ON_BrepLoop::m_type to outer/inner
  // when translating from data definition where this distinction is murky.
  int LoopDirection( const ON_BrepLoop& ) const;


  BOOL CombineCoincidentVertices(ON_BrepVertex&, ON_BrepVertex&); // moves information to first vertex and deletes second
  BOOL CombineCoincidentEdges(ON_BrepEdge&, ON_BrepEdge&); // moves information to first edge and deletes second

  // These remove a topology piece from a b-rep but do not
  // rearrange the arrays that hold the brep objects.  The
  // deleted objects have their indices set to -1.  Deleting
  // an object that is connected to other objects will 
  // modify thos objects.
  void DeleteVertex(ON_BrepVertex&);
  void DeleteEdge(ON_BrepEdge&,BOOL); // pass TRUE to delete vertices used only by edge
  void DeleteTrim(ON_BrepTrim&,BOOL); // pass TRUE to delete edges and vertices used only by trim
  void DeleteLoop(ON_BrepLoop&,BOOL); // pass TRUE to delete edges and vertices used only by trim
  void DeleteFace(ON_BrepFace&,BOOL); // pass TRUE to delete edges and vertices used only by face
  void DeleteSurface(int);
  void Delete2dCurve(int);
  void Delete3dCurve(int);

  BOOL Compact(); // Uses the CullUnused*() members to delete any unreferenced
                  // objects from arrays, reindexes as needed, and shrinks
                  // arrays to minimum required size.

  BOOL CullUnusedFaces(); // culls faces with m_face_index == -1
  BOOL CullUnusedLoops(); // culls loops with m_loop_index == -1
  BOOL CullUnusedTrims(); // culls trims with m_trim_index == -1
  BOOL CullUnusedEdges(); // culls edges with m_edge_index == -1
  BOOL CullUnusedVertices(); // culls vertices with m_vertex_index == -1
  BOOL CullUnused3dCurves(); // culls 2d curves not referenced by a trim
  BOOL CullUnused2dCurves(); // culls 3d curves not referenced by an edge
  BOOL CullUnusedSurfaces(); // culls surfaces not referenced by a face

  /////////////////////////////////////////////////////////////////
  // Navigation Interface

  // for moving around loops - returns trim index of prev/next trim in loop
  int PrevTrim(
        int // index of current trim (m_trim_index)
        ) const;
  int NextTrim(
        int // index of current trim (m_trim_index)
        ) const;

  // for moving around vertices - returns edge index of prev/next edge in
  // using the specified vertex
  int PrevEdge(
        int,        // index of current edge (m_edge_index)
        int,        // 0 = use edge start vertex, 1 = use edge end vertex
        int* = NULL // returns 0 if next edge begins at vert
                    //         1 if next edge ends at vertex
        ) const;
  int NextEdge(
        int,        // index of current edge (m_edge_index)
        int,        // 0 = use edge start vertex, 1 = use edge end vertex
        int* = NULL // returns 0 if prev edge begins at vert
                    //         1 if prev edge ends at vertex
        ) const;

  /////////////////////////////////////////////////////////////////
  // "Expert" Interface

  // geometry 
  // (all geometry is deleted by ~ON_Brep().  Pointers can be NULL
  // or not referenced.  Use Compact() to remove unreferenced geometry.
  ON_CurveArray   m_C2;  // Pointers to parameter space trimming curves
                         // (used by trims).
  ON_CurveArray   m_C3;  // Pointers to 3d curves (used by edges).
  ON_SurfaceArray m_S;   // Pointers to parametric surfaces (used by faces)

  // topology
  // (all topology is deleted by ~ON_Brep().  Objects can be unreferenced.
  // Use Compact() to to remove unreferenced geometry.
  ON_BrepVertexArray  m_V;   // vertices
  ON_BrepEdgeArray    m_E;   // edges
  ON_BrepTrimArray    m_T;   // trims
  ON_BrepLoopArray    m_L;   // loops
  ON_BrepFaceArray    m_F;   // faces

protected:	
  ON_BoundingBox m_bbox;

private:
  // These are friends so legacy tol values stored in v1 3dm files
  // can be used to set brep edge and trimming tolerances with a call
  // to ON_Brep::SetTolsFromLegacyValues().
  friend BOOL ON_BinaryArchive::ReadV1_TCODE_AG_FACE(ON_Object**,ON_3dmObjectAttributes*);
  friend BOOL ON_BinaryArchive::ReadV1_TCODE_AG_SHELL(ON_Object**,ON_3dmObjectAttributes*);
  void Initialize();

  // helpers to set ON_BrepTrim::m_iso flag
  void SetTrimIsoFlag(int,double[6]);
  void SetTrimIsoFlag(int);

  // helpers to create and set vertices
  BOOL SetEdgeVertex(const int, const int, const int );
  BOOL HopAcrossEdge( int&, int& ) const;
  BOOL SetTrimStartVertex( const int, const int);
  void SetLoopVertices(const int);
  void ClearTrimVertices();
  void ClearEdgeVertices();

  // helpers for SwapFaceParameters()
  BOOL SwapLoopParameters(
        int // index of loop
        );
  BOOL SwapTrimParameters(
        int // index of trim
        );

  // helpers for validation checking
  BOOL
  IsValidTrim( 
        int // index of trim
        ) const;
  BOOL
  IsValidLoop( 
        int // index of loop
        ) const;
  BOOL
  IsValidFace( 
        int // index of face
        ) const;
  BOOL
  IsValidEdge( 
        int // index of edge
        ) const;
  BOOL
  IsValidVertex( 
        int // index of vertex
        ) const;

  void SetTolsFromLegacyValues();

  // read helpers to support various versions
  BOOL ReadOld100( ON_BinaryArchive& ); // reads legacy old RhinoIO toolkit b-rep
  BOOL ReadOld101( ON_BinaryArchive& ); // reads legacy Rhino 1.1 b-rep
  BOOL ReadOld200( ON_BinaryArchive&, int ); // reads legacy trimmed surface
  BOOL Read_1( ON_BinaryArchive& ); // reads openNURBS b-rep
  ON_Curve* Read100_BrepCurve( ON_BinaryArchive& ) const;
  ON_Surface* Read100_BrepSurface( ON_BinaryArchive& ) const;

  // helpers for reading legacy v1 trimmed surfaces and breps
  BOOL ReadV1_LegacyTrimStuff( ON_BinaryArchive&, ON_BrepFace&, ON_BrepLoop& );
  BOOL ReadV1_LegacyTrim( ON_BinaryArchive&, ON_BrepFace&, ON_BrepLoop& );
  BOOL ReadV1_LegacyLoopStuff( ON_BinaryArchive&, ON_BrepFace& );
  BOOL ReadV1_LegacyLoop( ON_BinaryArchive&, ON_BrepFace& );
  BOOL ReadV1_LegacyFaceStuff( ON_BinaryArchive& );
  BOOL ReadV1_LegacyShellStuff( ON_BinaryArchive& );

};

///////////////////////////////////////////////////////////////////////////////

#endif

