#include "opennurbs.h"

ON_VIRTUAL_OBJECT_IMPLEMENT(ON_Curve,ON_Geometry,"4ED7D4D7-E947-11d3-BFE5-0010830122F0");

ON_Curve::ON_Curve() : m_ctree(0)
{}

ON_Curve::ON_Curve(const ON_Curve& src) : ON_Geometry(src), m_ctree(0)
{}

ON_Curve& ON_Curve::operator=(const ON_Curve& src)
{
  if ( this != &src ) {
    DestroyCurveTree();
    ON_Geometry::operator=(src);
  }
  return *this;
}

ON_Curve::~ON_Curve()
{
  DestroyCurveTree();
}

ON::object_type ON_Curve::ObjectType() const
{
  return ON::curve_object;
}

BOOL ON_Curve::GetDomain(double* s0,double* s1) const
{
  BOOL rc = FALSE;
  ON_Interval d = Domain();
  if ( d.IsIncreasing() ) {
    if(s0) *s0 = d.Min();
    if (s1) *s1 = d.Max();
    rc = TRUE;
  }
  return rc;
}

BOOL ON_Curve::CreateCurveTree()
{
  // this virtual function is (and should be) overridden by curves that can 
  // compute curve trees
  return FALSE;
}

void ON_Curve::DestroyCurveTree()
{
  if ( m_ctree ) {
    delete m_ctree;
    m_ctree = 0;
  }
}

ON_CurveTree* ON_Curve::CurveTree() const
{
  if ( !m_ctree ) {
    ON_Curve* p = (ON_Curve*)this; // const lie
    p->CreateCurveTree();
  }
  return m_ctree;
}

BOOL ON_Curve::SetDomain( double, double )
{
  // this virtual function is overridden by curves that can change their domain.
  return FALSE;
}


BOOL ON_Curve::GetParameterTolerance( // returns tminus < tplus: paramters tminus <= s <= tplus
       double t,       // t = parameter in domain
       double* tminus, // tminus
       double* tplus   // tplus
       ) const
{
  BOOL rc = FALSE;
  ON_Interval d = Domain();
  if ( d.IsIncreasing() )
    rc = ON_GetParameterTolerance( d[0], d[1], t, tminus, tplus );
  return rc;
}


BOOL ON_Curve::IsLinear( double tolerance ) const
{
  BOOL rc = FALSE;
  if ( Dimension() == 2 || Dimension() == 3 ) {
    const int span_count = SpanCount();
    const int span_degree = Degree();
    if ( span_count > 0 ) {
      ON_SimpleArray<double> s(span_count+1);
      s.SetCount(span_count+1);
      if ( GetSpanVector( s.Array() ) ) {
        if ( tolerance == 0.0 )
          tolerance = ON_ZERO_TOLERANCE;
        ON_Line line( PointAtStart(), PointAtEnd() );
        if ( line.Length() > tolerance ) {
          double t, t0, d, delta;
          ON_Interval sp;
          int n, i, span_index;
          rc = TRUE;
          t0 = Domain()[0];
          ON_3dPoint P;

          for ( span_index = 0; span_index < span_count; span_index++ ) {
            sp.Set( s[span_index], s[span_index+1] );
            n = 2*span_degree+1;
            delta = 1.0/n;
            for ( i = (span_index)?0:1; i < n; i++ ) {
              P = PointAt( sp.ParameterAt(i*delta) );
              if ( !line.ClosestPointTo( P, &t ) )
                rc = FALSE;
              else if ( t < t0 )
                rc = FALSE;
              d = P.DistanceTo( line.PointAt(t) );
              if ( d > tolerance )
                rc = FALSE;
              t0 = t;
            }
          }
        }
      }
    }
  }
  return rc;
}

BOOL ON_Curve::IsArc( const ON_Plane* plane, ON_Arc* arc, double tolerance ) const
{
  BOOL rc = FALSE;
  double c0, c, t, delta;
  int n, i, span_index;
  ON_Plane pln;
  ON_Arc a;
  ON_3dPoint P, C;

  if ( !plane ) {
    if ( !IsPlanar(&pln,tolerance) )
      return FALSE;
    plane = &pln;
  }
  if ( !arc )
    arc = &a;
  const int span_count = SpanCount();
  const int span_degree = Degree();
  if ( span_count < 1 )
    return FALSE;
  ON_SimpleArray<double> d(span_count+1);
  d.SetCount(span_count+1);
  if ( !GetSpanVector(d.Array()) )
    return FALSE;

  const BOOL bIsClosed = IsClosed();

  ON_3dPoint P0 = PointAt( d[0] );
  t = bIsClosed ? 0.5*d[0] + 0.5*d[span_count] : d[span_count];
  ON_3dPoint P1 = PointAt( 0.5*d[0] + 0.5*t );
  ON_3dPoint P2 = PointAt( t );

  arc->Create(P0,P1,P2);
  if ( bIsClosed )
    arc->SetAngleRadians(2.0*ON_PI);


  if ( tolerance == 0.0 )
    tolerance = ON_ZERO_TOLERANCE;
  rc = TRUE;
  c0 = 0.0;
  for ( span_index = 0; rc && span_index < span_count; span_index++ ) {
    n = 2*span_degree+1;
    if ( n < 4 )
      n = 4;
    delta = 1.0/n;
    for ( i = 0; i < n; i++ ) {
      t = i*delta;
      P = PointAt( (1.0-t)*d[span_index] + t*d[span_index+1] );
      if ( !arc->ClosestPointTo(P,&c) ) {
        rc = FALSE;
        break;        
      }
      if ( c < c0 ) {
        rc = FALSE;
        break;
      }
      C = arc->PointAt(c);        
      if ( C.DistanceTo(P) > tolerance ) {
        rc = 0;
        break;
      }
      c0 = c;
    }
  }

  return rc;
}

BOOL ON_Curve::IsPlanar( ON_Plane* plane, double tolerance ) const
{
  BOOL rc = FALSE;
  const int dim = Dimension();
  if ( dim == 2 ) {
    rc = TRUE;
    if ( plane ) {
      double t0 = Domain().Min();
      ON_3dPoint pt = PointAt(t0);
      ON_3dVector x = TangentAt(t0);
      if ( !x.Unitize() )
        x = ON_xaxis;
      ON_3dVector y(-x.y,x.x,0.0);
      plane->CreateFromFrame( pt, x, y );
    }
  }
  else if ( dim == 3 ) {
    const int span_count = SpanCount();
    if ( span_count < 1 )
      return FALSE;
    const int span_degree = Degree();
    if ( span_degree < 1 )
      return FALSE;
    ON_SimpleArray<double> s(span_count+1);
    s.SetCount(span_count+1);
    if ( !GetSpanVector(s.Array()) )
      return FALSE;
    ON_Interval d = Domain();

    // use initial point, tangent, and evaluated spans to guess a plane
    ON_3dPoint pt = PointAt(d.ParameterAt(0.0));
    ON_3dVector x = TangentAt(d.ParameterAt(0.0));
    if ( x.Length() < 0.95 ) {
      return FALSE;
    }
    int n = (span_degree > 1) ? span_degree+1 : span_degree;
    double delta = 1.0/n;
    int i, span_index, hint = 0;
    ON_3dPoint q;
    ON_3dVector y;
    BOOL bNeedY = TRUE;
    for ( span_index = 0; span_index < span_count && bNeedY; span_index++ ) {
      d.Set(s[span_index],s[span_index+1]);
      for ( i = span_index ? 0 : 1; i < n && bNeedY; i++ ) {
        if ( !EvPoint( d.ParameterAt(i*delta), q, 0, &hint ) )
          return FALSE;
        y = q-pt;
        y = y - (y*x)*x;
        bNeedY = ( y.Length()  <= 1.0e-6 );
      }
    }
    if ( bNeedY )
      y.PerpendicularTo(x);
    ON_Plane pln( pt, x, y );
    if ( plane )
      *plane = pln;

    // test
    BOOL rc = TRUE;
    n = 2*span_degree + 1;
    delta = 1.0/n;
    double h = pln.EquationAt(PointAtEnd());
    if ( fabs(h) )
      rc = FALSE;
    hint = 0;
    for ( span_index = 0; rc && span_index < span_count; span_index++ ) {
      d.Set(s[span_index],s[span_index+1]);
      for ( i = 0; rc && i < n; i++ ) {
        if ( !EvPoint( d.ParameterAt(i*delta), q, 0, &hint ) )
          rc = FALSE;
        else {
          h = pln.EquationAt(q);
          if ( fabs(h) > tolerance )
            rc = FALSE;
        }
      }
    }
  }
  return rc;
}

BOOL ON_Curve::IsClosed() const
{
  BOOL rc = FALSE;
  double *a, *b, *c, *p, w[12];
  const int dim = Dimension();
  if ( dim > 1 ) {
    ON_Interval d = Domain();
    a = (dim>3) ? (double*)onmalloc(dim*4*sizeof(*a)) : w;
    b = a+dim;
    c = b+dim;
    p = c+dim;
    if ( !Evaluate( d.ParameterAt(0.0), 0, dim, a, 1 ) )
      return FALSE;
    if ( !Evaluate( d.ParameterAt(1.0/3.0), 0, dim, b, 0 ) )
      return FALSE;
    if ( !Evaluate( d.ParameterAt(2.0/3.0), 0, dim, c, 0 ) )
      return FALSE;
    if ( !Evaluate( d.ParameterAt(1.0), 0, dim, p,-1 ) )
      return FALSE;
    rc = ON_ComparePoint( dim, FALSE, a, p )?FALSE:TRUE;
    if ( rc ) {
      if ( !ON_ComparePoint( dim, FALSE, a, b ) || !ON_ComparePoint( dim, FALSE, b, c ) || !ON_ComparePoint( dim, FALSE, c, p ) )
        rc = FALSE;
    }
    if ( dim>3)
      onfree(a);
  }
  return rc;
}

BOOL ON_Curve::IsPeriodic() const
{
  return FALSE;
}

ON_3dPoint ON_Curve::PointAt( double t ) const
{
  ON_3dPoint p(0.0,0.0,0.0);
  EvPoint(t,p);
  return p;
}

ON_3dPoint ON_Curve::PointAtStart() const
{
  return PointAt(Domain().Min());
}

ON_3dPoint ON_Curve::PointAtEnd() const
{
  return PointAt(Domain().Max());
}

ON_3dVector ON_Curve::DerivativeAt( double t ) const
{
  ON_3dPoint  p(0.0,0.0,0.0);
  ON_3dVector d(0.0,0.0,0.0);
  Ev1Der(t,p,d);
  return d;
}

ON_3dVector ON_Curve::TangentAt( double t ) const
{
  ON_3dPoint p;
  ON_3dVector d1, d2, T, K;
  T = DerivativeAt(t);
  if ( !T.Unitize() ) {
    // see if lhospital's rule helps
    Ev2Der( t, p, d1, d2 );
    ON_EvCurvature( d1, d2, T, K );
  }
  return T;
}

ON_3dVector ON_Curve::CurvatureAt( double t ) const
{
  ON_3dPoint p;
  ON_3dVector d1, d2, T, K;
  Ev2Der( t, p, d1, d2 );
  ON_EvCurvature( d1, d2, T, K );
  return K;
}

BOOL ON_Curve::EvPoint( // returns FALSE if unable to evaluate
       double t,         // evaluation parameter
       ON_3dPoint& point,   // returns value of curve
       int side,        // optional - determines which side to evaluate from
                       //         0 = default
                       //      <  0 to evaluate from below, 
                       //      >  0 to evaluate from above
       int* hint       // optional - evaluation hint used to speed interative
                       //            evaluations
       ) const
{
  BOOL rc = FALSE;
  double ws[128];
  double* v;
  if ( Dimension() <= 3 ) {
    v = &point.x;
    point.x = 0.0;
    point.y = 0.0;
    point.z = 0.0;
  }
  else if ( Dimension() <= 128 ) {
    v = ws;
  }
  else {
    v = (double*)onmalloc(Dimension()*sizeof(*v));
  }
  rc = Evaluate( t, 0, Dimension(), v, side, hint );
  if ( Dimension() > 3 ) {
    point.x = v[0];
    point.y = v[1];
    point.z = v[2];
    if ( Dimension() > 128 )
      onfree(v);
  }
  return rc;
}

BOOL ON_Curve::Ev1Der( // returns FALSE if unable to evaluate
       double t,         // evaluation parameter
       ON_3dPoint& point,
       ON_3dVector& derivative,
       int side,        // optional - determines which side to evaluate from
                       //      <= 0 to evaluate from below, 
                       //      >  0 to evaluate from above
       int* hint       // optional - evaluation hint used to speed interative
                       //            evaluations
       ) const
{
  BOOL rc = FALSE;
  const int dim = Dimension();
  double ws[2*64];
  double* v;
  point.x = 0.0;
  point.y = 0.0;
  point.z = 0.0;
  derivative.x = 0.0;
  derivative.y = 0.0;
  derivative.z = 0.0;
  if ( dim <= 64 ) {
    v = ws;
  }
  else {
    v = (double*)onmalloc(2*dim*sizeof(*v));
  }
  rc = Evaluate( t, 1, dim, v, side, hint );
  point.x = v[0];
  derivative.x = v[dim];
  if ( dim > 1 ) {
    point.y = v[1];
    derivative.y = v[dim+1];
    if ( dim > 2 ) {
      point.z = v[2];
      derivative.z = v[dim+2];
      if ( dim > 64 )
        onfree(v);
    }
  }

  return rc;
}

BOOL ON_Curve::Ev2Der( // returns FALSE if unable to evaluate
       double t,         // evaluation parameter
       ON_3dPoint& point,
       ON_3dVector& firstDervative,
       ON_3dVector& secondDervative,
       int side,        // optional - determines which side to evaluate from
                       //      <= 0 to evaluate from below, 
                       //      >  0 to evaluate from above
       int* hint       // optional - evaluation hint used to speed interative
                       //            evaluations
       ) const
{
  BOOL rc = FALSE;
  const int dim = Dimension();
  double ws[3*64];
  double* v;
  point.x = 0.0;
  point.y = 0.0;
  point.z = 0.0;
  firstDervative.x = 0.0;
  firstDervative.y = 0.0;
  firstDervative.z = 0.0;
  secondDervative.x = 0.0;
  secondDervative.y = 0.0;
  secondDervative.z = 0.0;
  if ( dim <= 64 ) {
    v = ws;
  }
  else {
    v = (double*)onmalloc(3*dim*sizeof(*v));
  }
  rc = Evaluate( t, 2, dim, v, side, hint );
  point.x = v[0];
  firstDervative.x = v[dim];
  secondDervative.x = v[2*dim];
  if ( dim > 1 ) {
    point.y = v[1];
    firstDervative.y = v[dim+1];
    secondDervative.y = v[2*dim+1];
    if ( dim > 2 ) {
      point.z = v[2];
      firstDervative.z = v[dim+2];
      secondDervative.z = v[2*dim+2];
      if ( dim > 64 )
        onfree(v);
    }
  }

  return rc;
}

BOOL ON_Curve::GetClosestPoint( const ON_3dPoint& test_point,
        double* t,       // parameter of local closest point returned here
        double maximum_distance,
        const ON_Interval* sub_domain
        ) const
{
  // override this function if your curve can calculate global closest points
  return FALSE;
}

BOOL ON_Curve::GetLocalClosestPoint( const ON_3dPoint& test_point,
        double seed_parameter,
        double* t,
        const ON_Interval* sub_domain
        ) const
{
  // TODO: add simple dbrent search for base class default that will work
  //       with C1 curves that have a non-vanishing derivative.
  return FALSE;
}

BOOL ON_Curve::Length(
        double* length,               // length returned here
        double fractional_tolerance,  // default = 1.0e-8
        const ON_Interval* sub_domain // default = NULL
        ) const
{
  // override when possible
  // TODO: add simple integration routine that works with C1 curves
  if ( length )
    *length = 0.0;
  return FALSE;
}


BOOL ON_Curve::Trim( const ON_Interval& in )
{
  // TODO - make this pure virtual
  return FALSE;
}

BOOL ON_Curve::Split(
    double, // t = curve parameter to split curve at
    ON_Curve*&, // left portion returned here (can pass "this" as the pointer)
    ON_Curve*&  // right portion returned here (can pass "this" as the pointer)
  ) const
{
  // override if curve can split itself
  return FALSE;
}

int ON_Curve::GetNurbForm( // returns 0: unable to create NURBS representation
                 //            with desired accuracy.
                 //         1: success - returned NURBS parameterization
                 //            matches the curve's to wthe desired accuracy
                 //         2: success - returned NURBS point locus matches
                 //            the curve's to the desired accuracy but, on
                 //            the interior of the curve's domain, the 
                 //            curve's parameterization and the NURBS
                 //            parameterization may not match to the 
                 //            desired accuracy.
      ON_NurbsCurve&,
      double,       // OPTIONAL tolerance (default = 0 )
      const ON_Interval*  // OPTIONAL subdomain of curve (default = NULL)
      ) const
{
  // This is a virtual function
  return 0;
}

ON_CurveArray::ON_CurveArray( size_t initial_capacity ) 
              : ON_SimpleArray<ON_Curve*>(initial_capacity)
{}

ON_CurveArray::~ON_CurveArray()
{
  Destroy();
}

void ON_CurveArray::Destroy()
{
  int i = m_capacity;
  while ( i-- > 0 ) {
    if ( m_a[i] ) {
      delete m_a[i];
      m_a[i] = NULL;
    }
  }
  Empty();
}

BOOL ON_CurveArray::Duplicate( ON_CurveArray& dst ) const
{
  dst.Destroy();
  dst.SetCapacity( Capacity() );

  const int count = Count();
  int i;
  ON_Object* dup;
  ON_Curve* curve;
  for ( i = 0; i < count; i++ ) {
    curve = 0;
    if ( m_a[i] ) {
      dup = m_a[i]->Duplicate();
      curve = ON_Curve::Cast(dup);
      if ( !curve )
        delete dup;
    }
    dst.Append(curve);      
  }
  return TRUE;
}

BOOL ON_CurveArray::Write( ON_BinaryArchive& file ) const
{
  BOOL rc = file.BeginWrite3dmChunk( TCODE_ANONYMOUS_CHUNK, 0 );
  if (rc) rc = file.Write3dmChunkVersion(1,0);
  if (rc ) {
    int i;
    rc = file.WriteInt( Count() );
    for ( i = 0; rc && i < Count(); i++ ) {
      if ( m_a[i] ) {
        rc = file.WriteInt(1);
        if ( rc ) 
          rc = file.WriteObject( *m_a[i] ); // polymorphic curves
      }
      else {
        // NULL curve
        rc = file.WriteInt(0);
      }
    }
    if ( !file.EndWrite3dmChunk() )
      rc = FALSE;
  }
  return rc;
}


BOOL ON_CurveArray::Read( ON_BinaryArchive& file )
{
  int major_version = 0;
  int minor_version = 0;
  unsigned int tcode;
  int i, flag;
  Destroy();
  BOOL rc = file.BeginRead3dmChunk( &tcode, &i );
  if (rc) {
    rc = ( tcode == TCODE_ANONYMOUS_CHUNK );
    if (rc) rc = file.Read3dmChunkVersion(&major_version,&minor_version);
    if (rc && major_version == 1) {
      ON_Object* p;
      int count;
      BOOL rc = file.ReadInt( &count );
      if (rc) {
        SetCapacity(count);
        SetCount(count);
        Zero();
        int i;
        for ( i = 0; rc && i < count && rc; i++ ) {
          flag = 0;
          rc = file.ReadInt(&flag);
          if (rc && flag==1) {
            p = 0;
            rc = file.ReadObject( &p ); // polymorphic curves
            m_a[i] = ON_Curve::Cast(p);
            if ( !m_a[i] )
              delete p;
          }
        }
      }
    }
    else {
      rc = FALSE;
    }
    if ( !file.EndRead3dmChunk() )
      rc = FALSE;
  }
  return rc;
}

ON_CurveTree::ON_CurveTree() : m_root(0), m_bReparam(0), m_tolerance(0.0)
{
}


ON_CurveTree::~ON_CurveTree()
{
  if ( m_root ) {
    delete m_root;
    m_root = 0;
  }
}

ON_CurveTreeNode* ON_CurveTree::Root() const
{
  return m_root;
}

ON_CurveTreeNode* ON_CurveTree::FirstLeaf() const
{
  ON_CurveTreeNode* p = m_root;
  if ( p ) {
    while ( p->m_down[0] )
      p = p->m_down[0];
  }
  return p;
}

ON_CurveTreeNode* ON_CurveTree::LastLeaf() const
{
  ON_CurveTreeNode* p = m_root;
  if ( p ) {
    while ( p->m_down[1] )
      p = p->m_down[1];
  }
  return p;
}

ON_CurveTreeNode::ON_CurveTreeNode() 
                 : m_bez(0), 
                   m_bMonotone(0), 
                   m_radius(ON_UNSET_VALUE), 
                   m_up(0)
{
  m_down[0] = 0;
  m_down[1] = 0;
}

ON_CurveTreeNode::~ON_CurveTreeNode()
{
  if ( m_bez ) {
    delete m_bez;
    m_bez = 0;
  }
  if ( m_down[0] ) {
    delete m_down[0];
    m_down[0] = 0;
  }
  if ( m_down[1] ) {
    delete m_down[1];
    m_down[1] = 0;
  }
}


ON_CurveTreeNode* ON_CurveTreeNode::NextLeaf() const
{
  ON_CurveTreeNode* p = 0;
  if ( m_up && !m_down[1] ) {
    if ( this == m_up->m_down[0] ) {
      // we are a left hand leaf, move to the right
      p = m_up->m_down[1];

      if ( p ) {
        // then go down the left side;
        while ( p->m_down[0] )
          p = m_down[0];
      }
    }
    else if ( this == m_up->m_down[1] && m_up->m_up ) {
      // we are a right hand leaf
      // go up until we are a left hand node
      p = m_up;
      while ( p->m_up && p == p->m_up->m_down[1] )
        p = p->m_up;
      // then go up once more
      p = p->m_up;
      if ( p ) { 
        // then go down once on the right
        p = p->m_down[1];
        if (p) {
          // then go down the left side until we hit a leaf
          while (p->m_down[0])
            p = p->m_down[0];
        }
      }
    }
  }
  return p;
}


ON_CurveTreeNode* ON_CurveTreeNode::PrevLeaf() const
{
  ON_CurveTreeNode* p = 0;
  if ( m_up && !m_down[0] ) {
  }
  return p;
}


BOOL ON_CurveTreeNode::Split()
{
  BOOL rc = FALSE;
  if ( !m_down[0] && !m_down[1] ) {
    m_down[0] = new ON_CurveTreeNode;
    m_down[1] = new ON_CurveTreeNode;
    m_down[0]->m_up = this;
    m_down[0]->m_domain.Set( m_domain.m_t[0], 0.5*(m_domain.m_t[0]+m_domain.m_t[1]) );
    m_down[1]->m_up = this;
    m_down[1]->m_domain.Set( m_down[0]->m_domain.m_t[0], m_domain.m_t[1] );
    if ( m_bez && m_bez->m_order >= 2 && m_bez->m_cv ) {
      ON_Line axis;
      ON_3dPoint pt;
      double s, t, r, d;
      ON_BezierCurve* bez;
      int i,j,k,n;
      const double* a;
      const double* b;
      double* c;
      const int dim = m_bez->m_dim;
      const int is_rat = m_bez->m_is_rat;
      const int cvdim = is_rat?dim+1:dim;
      const int sizeof_cv = cvdim*sizeof(double);
      const int order = m_bez->m_order;
      const int degree = order-1;
      m_down[0]->m_bez = new ON_BezierCurve(dim,is_rat,order);
      m_down[1]->m_bez = new ON_BezierCurve(dim,is_rat,order);

      // split parent bezier
      double** cv1 = (double**)alloca( (2*order-1)*sizeof(*cv1) );
      for ( i = 0; i < order; i++ ) {
        cv1[i] = m_down[0]->m_bez->CV(i);
        cv1[i+order-1] = m_down[1]->m_bez->CV(i);
      }
      for ( i = 0; i < order; i++ ) {
        memcpy( cv1[i*2], m_bez->CV(i), sizeof_cv );
      }
      for ( i = 1; i < order; i++ ) {
        n = order-i;
        for ( j = i; j < n; j += 2 ) {
          a = cv1[i];
          b = cv1[i+2];
          c = cv1[i];
          for ( k = 0; k < cvdim; k++ )
            c[k] = 0.5*(a[k]+b[k]);
        }
      }
      memcpy( m_down[1]->m_bez->CV(0), m_down[0]->m_bez->CV(degree), sizeof_cv );

      for ( i = 0; i < 2; i++ ) {
        bez = m_down[i]->m_bez;
        // get bounding boxes
        if ( m_down[i]->m_bez->m_dim < 3 ) {
          m_down[i]->m_bbox.m_min.Zero();
          m_down[i]->m_bbox.m_max.Zero();
        }
        bez->GetBBox(m_down[i]->m_bbox.m_min,m_down[i]->m_bbox.m_max);

        // get radii
        k = TRUE; // monotone flag
        bez->GetCV(0,axis.from);
        bez->GetCV(degree,axis.to);
        t = 0.0;
        r = 0.0;
        for ( i = 1; i < degree; i++ ) {
          bez->GetCV(i,pt);
          axis.ClosestPointTo(pt,&s);
          if ( s < t )
            k = FALSE;
          d = pt.DistanceTo(axis.PointAt(s));
          if ( d > r )
            r = d;
        }
        m_down[i]->m_bMonotone = k;
        m_down[i]->m_radius = r;
      }
    }
    rc = TRUE;
  }
  return rc;
}


