/* $Header: /src2/opennurbs/opennurbs_curve.h 18    5/11/01 3:36p Dalelear $ */
/* $NoKeywords: $ */
/*
//
// Copyright (c) 1993-2001 Robert McNeel & Associates. All rights reserved.
// Rhinoceros is a registered trademark of Robert McNeel & Assoicates.
//
// THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
// ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
// MERCHANTABILITY ARE HEREBY DISCLAIMED.
//				
// For complete openNURBS copyright information see <http://www.opennurbs.org>.
//
////////////////////////////////////////////////////////////////
*/

////////////////////////////////////////////////////////////////
//
//   Definition of virtual parametric curve
//
////////////////////////////////////////////////////////////////

#if !defined(OPENNURBS_CURVE_INC_)
#define OPENNURBS_CURVE_INC_

class ON_Curve;
class ON_Plane;
class ON_Arc;
class ON_NurbsCurve;

////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////
//
// A curve cache is a simple binary tree of bezier curves
// and is used to speed up geometric calculations.

class ON_CLASS ON_CurveTreeNode
{
public:
  ON_CurveTreeNode();
  ~ON_CurveTreeNode();

  // m_domain records the portion of the curve's domain that
  // this node defines.  It is always increasing.
  ON_Interval     m_domain;

  // Bounding box of the portion of the curve with domain
  // [m_domain.Min(), m_domain.Max()]
  ON_BoundingBox  m_bbox;

  // All leaf nodes have a bezier. Non-leaf nodes may have a NULL here.
  // The domain of m_bez is alwasy [0,1] and m_domain records the portion
  // of the curve that this bezier curve corresponds to.
  ON_BezierCurve* m_bez;

  // If the bezier is monotone with respect to the chord connecting
  // its end points, then m_bMonotone is TRUE.
  BOOL m_bMonotone;

  // The maximum distance from the bezier to the bezier's chord is
  // <= m_radius.  A value of ON_UNSET_VALUE indicates the radius 
  // has not been computed.
  double  m_radius;

  // m_up moves towards the root node.  If m_up is NULL, the node
  // is the root node.
  ON_CurveTreeNode*  m_up;

  // m_down[] moves towards the leaf nodes.  Either both m_down[0]
  // and m_down[1] are NULL, or both are not NULL.  m_down[0]
  // covers the portion of the curve from m_domin.Min() to m_domain.Mid().
  // m_down[1] defines the geometry from m_domain.Mid() to m_domain.Max().
  ON_CurveTreeNode*  m_down[2];

  // Next() moves to the next leaf interval in the curve tree if this
  // node is a leaf node.
  ON_CurveTreeNode*  NextLeaf() const;

  // Prev() moves to the previous leaf interval in the curve tree if this
  // node is a leaf node.
  ON_CurveTreeNode*  PrevLeaf() const;

  // If m_down[] is {NULL,NULL}, splits this node and creates m_down[].
  BOOL Split();
};

class ON_CLASS ON_CurveTree
{
public:
  ON_CurveTree();
  ~ON_CurveTree();
  
  // root nod of the binary tree
  ON_CurveTreeNode* m_root;

  // The nodes contain bezier curves. m_bReparam is TRUE if 
  // the parametrization of the beziers (with m_domain taken
  // into account) does not match the parameterization of the curve.
  BOOL m_bReparam; 

  // Tolerance to which the locus of the beziers in the nodes
  // approximates the locus of the curve.
  double m_tolerance;

  ON_CurveTreeNode* Root() const;

  // First() returns the first leaf interval in the curve tree.
  ON_CurveTreeNode* FirstLeaf() const;

  // Last() returns the last leaf interval in the curve tree.
  ON_CurveTreeNode* LastLeaf() const;
};

////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////



class ON_CLASS ON_Curve : public ON_Geometry
{
  // pure virtual class for curve objects

  // Any object derived from ON_Curve should have a
  //   ON_OBJECT_DECLARE(ON_...);
  // as the last line of its class definition and a
  //   ON_OBJECT_IMPLEMENT( ON_..., ON_baseclass );
  // in a .cpp file.
  //
  // See the definition of ON_Object for details.
  ON_OBJECT_DECLARE(ON_Curve);

public:
  ON_Curve();
  ON_Curve(const ON_Curve&);
  ON_Curve& operator=(const ON_Curve&);
  virtual ~ON_Curve();

  //////////
  // override ON_Object::ObjectType() - returns ON::curve_object
  ON::object_type ObjectType() const;

  ////////////////////////////////////////////////////////////////////
  // curve interface

  BOOL GetDomain(double*,double*) const;

  virtual 
  ON_Interval Domain() const = 0;

  virtual
  BOOL SetDomain( double, double );

  virtual 
  int SpanCount() const = 0; // number of smooth spans in curve

  virtual
  BOOL GetSpanVector( // span "knots" 
         double* // array of length SpanCount() + 1 
         ) const = 0; // 

  virtual 
  int Degree( // returns maximum algebraic degree of any span 
                  // ( or a good estimate if curve spans are not algebraic )
    ) const = 0; 

  virtual 
  BOOL GetParameterTolerance( // returns tminus < tplus: paramters tminus <= s <= tplus
         double,  // t = parameter in domain
         double*, // tminus
         double*  // tplus
         ) const;

  virtual
  BOOL IsLinear( // TRUE if curve locus is a line segment between
                 // between specified points
        double = ON_ZERO_TOLERANCE // tolerance to use when checking linearity
        ) const;

  virtual
  BOOL IsArc( // ON_Arc.m_angle > 0 if curve locus is an arc between
              // specified points
        const ON_Plane* = NULL, // if not NULL, test is performed in this plane
        ON_Arc* = NULL, // if not NULL and TRUE is returned, then arc parameters
                         // are filled in
        double = ON_ZERO_TOLERANCE    // tolerance to use when checking
        ) const;

  virtual
  BOOL IsPlanar(
        ON_Plane* = NULL, // if not NULL and TRUE is returned, then plane parameters
                           // are filled in
        double = ON_ZERO_TOLERANCE    // tolerance to use when checking
        ) const;

  virtual
  BOOL IsInPlane(
        const ON_Plane&, // plane to test
        double = ON_ZERO_TOLERANCE    // tolerance to use when checking
        ) const = 0;

  virtual 
  BOOL IsClosed(  // TRUE if curve is closed
        ) const;

  virtual 
  BOOL IsPeriodic(  // TRUE if curve is periodic (closed and at least C2 at the start/end).
        ) const;
  
  virtual 
  BOOL Reverse(  // reverse curve direction domain changes from [a,b] to [-b,-a]
        ) =0;
  
  // simple evaluation interface - no error handling
  ON_3dPoint  PointAt( double ) const;
  ON_3dPoint  PointAtStart() const;
  ON_3dPoint  PointAtEnd() const;
  ON_3dVector DerivativeAt( double ) const;
  ON_3dVector TangentAt( double ) const;
  ON_3dVector CurvatureAt( double ) const;

  // general evaluation interface - return FALSE when errors are encountered
  BOOL EvPoint( // returns FALSE if unable to evaluate
         double,         // evaluation parameter
         ON_3dPoint&,  // returns value of curve
         int = 0,        // optional - determines which side to evaluate from
                         //         0 = default
                         //      <  0 to evaluate from below, 
                         //      >  0 to evaluate from above
         int* = 0        // optional - evaluation hint used to speed interative
                         //            evaluations
         ) const;

  BOOL Ev1Der( // returns FALSE if unable to evaluate
         double,         // evaluation parameter
         ON_3dPoint&,  // returns value of curve
         ON_3dVector&, // returns value of curve's 1st derivative
         int = 0,        // optional - determines which side to evaluate from
                         //      <= 0 to evaluate from below, 
                         //      >  0 to evaluate from above
         int* = 0        // optional - evaluation hint used to speed interative
                         //            evaluations
         ) const;

  BOOL Ev2Der( // returns FALSE if unable to evaluate
         double,         // evaluation parameter
         ON_3dPoint&,  // returns value of curve
         ON_3dVector&, // returns value of curve's 1st derivative
         ON_3dVector&, // returns value of curve's 2nd derivative
         int = 0,        // optional - determines which side to evaluate from
                         //      <= 0 to evaluate from below, 
                         //      >  0 to evaluate from above
         int* = 0        // optional - evaluation hint used to speed interative
                         //            evaluations
         ) const;

  // work horse evaluator
  virtual 
  BOOL Evaluate( // returns FALSE if unable to evaluate
         double,         // evaluation parameter
         int,            // number of derivatives (>=0)
         int,            // v[] array stride (>=Dimension())
         double*,        // v[] array of length v_stride*(ndir+1)
         int = 0,        // optional - determines which side to evaluate from
                         //         0 = default
                         //      <  0 to evaluate from below, 
                         //      >  0 to evaluate from above
         int* = 0        // optional - evaluation hint (int) used to speed
                         //            interative evaluations
         ) const = 0;

  //////////
  // Find parameter of the point on a curve that is closest to test_point.
  // If the maximum_distance parameter is > 0, then only points whose distance
  // to the given point is <= maximum_distance will be returned.  Using a 
  // positive value of maximum_distance can substantially speed up the search.
  // If the sub_domain parameter is not NULL, then the search is restricted
  // to the specified portion of the curve.
  //
  // TRUE if returned if the search is successful.  FALSE is returned if
  // the search fails.
  virtual
  BOOL GetClosestPoint( const ON_3dPoint&, // test_point
          double*,       // parameter of local closest point returned here
          double = 0.0,  // maximum_distance
          const ON_Interval* = NULL // sub_domain
          ) const;

  //////////
  // Find parameter of the point on a curve that is locally closest to 
  // the test_point.  The search for a local close point starts at 
  // seed_parameter. If the sub_domain parameter is not NULL, then
  // the search is restricted to the specified portion of the curve.
  //
  // TRUE if returned if the search is successful.  FALSE is returned if
  // the search fails.
  virtual
  BOOL GetLocalClosestPoint( const ON_3dPoint&, // test_point
          double,    // seed_parameter
          double*,   // parameter of local closest point returned here
          const ON_Interval* = NULL // sub_domain
          ) const;


  //////////
  // Length of curve.
  // TRUE if returned if the length calculation is successful.
  // FALSE is returned if the length is not calculated.
  //
  // The arc length will be computed so that
  // (returned length - real length)/(real length) <= fractional_tolerance
  // More simply, if you want N significant figures in the answer, set the
  // fractional_tolerance to 1.0e-N.  For "nice" curves, 1.0e-8 works
  // fine.  For very high degree nurbs and nurbs with bad parametrizations,
  // use larger values of fractional_tolerance.
  virtual
  BOOL Length( // returns TRUE if length successfully computed
          double*,                   // length returned here
          double = 1.0e-8,           // fractional_tolerance
          const ON_Interval* = NULL  // (optional) sub_domain
          ) const;

  //////////
  // Trim removes portions of the curve outside the specified interval
  virtual
  BOOL Trim(
    const ON_Interval& 
    );

  //////////
  // Split() divides the curve at the specified paramter.  The parameter
  // must be in the interior of the curve's domain.  The pointers passed
  // to Split must either be NULL or point to an ON_Curve object of the same
  // of the same type.  If the pointer is NULL, then a curve will be created
  // in Split().  You may pass "this" as one of the pointers to Split().
  // For example, if crv were an ON_NurbsCurve, then the call
  //
  //   ON_NurbsCurve right_side;
  //   crv.Split( crv.Domain().Mid() &crv, &right_side );
  //
  // would split crv at the parametric midpoint, put the left side in crv,
  // and return the right side in right_side.
  virtual
  BOOL Split(
      double,    // t = curve parameter to split curve at
      ON_Curve*&, // left portion returned here
      ON_Curve*&  // right portion returned here
    ) const;

  virtual
  int GetNurbForm( // returns 0: unable to create NURBS representation
                   //            with desired accuracy.
                   //         1: success - returned NURBS parameterization
                   //            matches the curve's to wthe desired accuracy
                   //         2: success - returned NURBS point locus matches
                   //            the curve's to the desired accuracy but, on
                   //            the interior of the curve's domain, the 
                   //            curve's parameterization and the NURBS
                   //            parameterization may not match to the 
                   //            desired accuracy.
        ON_NurbsCurve&,
        double = 0.0,
        const ON_Interval* = NULL     // OPTIONAL subdomain of curve
        ) const;


  virtual
  BOOL CreateCurveTree();

  void DestroyCurveTree();

  ON_CurveTree* CurveTree() const;

protected:
  // Runtime only - ignored by Read()/Write()
  ON_CurveTree* m_ctree;
};

#if defined(ON_DLL_EXPORTS) || defined(ON_DLL_IMPORTS)
// This stuff is here because of a limitation in the way Microsoft
// handles templates and DLLs.  See Microsoft's knowledge base 
// article ID Q168958 for details.
#pragma warning( push )
#pragma warning( disable : 4231 )
ON_TEMPLATE template class ON_CLASS ON_SimpleArray<ON_Curve*>;
#pragma warning( pop )
#endif

class ON_CLASS ON_CurveArray : public ON_SimpleArray<ON_Curve*>
{
public:
  ON_CurveArray( size_t = 0 );
  ~ON_CurveArray(); // deletes any non-NULL curves

  BOOL Write( ON_BinaryArchive& ) const;
  BOOL Read( ON_BinaryArchive& );

  void Destroy(); // deletes curves, sets pointers to NULL, sets count to zero

  BOOL Duplicate( ON_CurveArray& ) const; // operator= copies the pointer values
                                          // duplicate copies the curves themselves
};

#endif
