/* $Header: /src2/opennurbs/opennurbs_cylinder.cpp 5     1/03/01 4:59p Dalelear $ */
/* $NoKeywords: $ */
/*
//
// Copyright (c) 1993-2001 Robert McNeel & Associates. All rights reserved.
// Rhinoceros is a registered trademark of Robert McNeel & Assoicates.
//
// THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
// ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
// MERCHANTABILITY ARE HEREBY DISCLAIMED.
//				
// For complete openNURBS copyright information see <http://www.opennurbs.org>.
//
////////////////////////////////////////////////////////////////
*/

#include "opennurbs.h"

ON_Cylinder::ON_Cylinder() 
{
  height[0] = 0.0;
  height[1] = 0.0;
}

ON_Cylinder::ON_Cylinder(
    const ON_Circle& c
    )
{
  Create(c);
}

ON_Cylinder::ON_Cylinder(
    const ON_Circle& c,
    double h
    )
{
  Create(c,h);
}

ON_Cylinder::~ON_Cylinder()
{}

BOOL ON_Cylinder::Create(
    const ON_Circle& c
    )
{
  return Create( c, 0.0 );
}

BOOL ON_Cylinder::Create(
    const ON_Circle& c,
    double h
    )
{
  circle = c;
  if ( h > 0.0 ) {
    height[0] = 0.0;
    height[1] = h;
  }
  else {
    height[0] = h;
    height[1] = 0.0;
  }
  return IsValid();
}

BOOL ON_Cylinder::IsValid() const
{
  return circle.IsValid();
}

BOOL ON_Cylinder::IsFinite() const
{
  return height[0] != height[1];
}

const ON_3dVector& ON_Cylinder::Axis() const
{
  return circle.plane.zaxis;
}

const ON_3dPoint& ON_Cylinder::Center() const
{
  return circle.plane.origin;
}

double ON_Cylinder::Height() const
{
  return height[1] - height[0];
}

ON_Circle ON_Cylinder::CircleAt( 
      double t // linear parameter
      ) const
{
  ON_Circle c = circle;
  if ( t != 0.0 )
    c.Translate(t*circle.plane.zaxis);
  return c;
}

ON_Line ON_Cylinder::LineAt( 
      double s // angular parameter
      ) const
{
  ON_3dPoint p = circle.PointAt(s);
  ON_Line line;
  line.from = p + height[0]*circle.plane.zaxis;
  line.to   = p + height[1]*circle.plane.zaxis;
  return line;
}


ON_3dPoint ON_Cylinder::PointAt( double s, double t ) const
{
  return ( circle.PointAt(s) + t*circle.plane.zaxis );
}

ON_3dPoint ON_Cylinder::NormalAt( double s, double t ) const
{
  ON_3dVector N = ON_CrossProduct( circle.TangentAt(s), circle.plane.zaxis );
  N.Unitize();
  return N;
}

// returns parameters of point on cylinder that is closest to given point
BOOL ON_Cylinder::ClosestPointTo( 
       ON_3dPoint point, 
       double* s, // angular parameter
       double* t  // linear parameter
       ) const
{
  BOOL rc = TRUE;
  //const ON_3dVector v = point - circle.plane.origin;
  double h = (point - circle.plane.origin)*circle.plane.zaxis;
  if ( s )
    rc = circle.ClosestPointTo( point - h*circle.plane.zaxis, s );
  if ( t ) {
    if ( height[0] < height[1] ) {
      if ( h < height[0] ) h = height[0]; else if ( h > height[1] ) h = height[1];
    }
    else if ( height[0] > height[1] ) {
      if ( h > height[0] ) h = height[0]; else if ( h < height[1] ) h = height[1];
    }
    *t = h;
  }
  return rc;
}

// returns point on cylinder that is closest to given point
ON_3dPoint ON_Cylinder::ClosestPointTo( 
       ON_3dPoint point
       ) const
{
  double s, t;
  ClosestPointTo( point, &s, &t );
  return PointAt( s, t );
}

// rotate plane about its origin
BOOL ON_Cylinder::Rotate(
      double sin_angle,
      double cos_angle,
      const ON_3dVector& axis // axis of rotation
      )
{
  return Rotate( sin_angle, cos_angle, axis, circle.plane.origin );
}

BOOL ON_Cylinder::Rotate(
      double angle,               // angle in radians
      const ON_3dVector& axis // axis of rotation
      )
{
  return Rotate( sin(angle), cos(angle), axis, circle.plane.origin );
}

// rotate plane about a point and axis
BOOL ON_Cylinder::Rotate(
      double sin_angle,               // sin(angle)
      double cos_angle,               // cos(angle)
      const ON_3dVector& axis, // axis of rotation
      const ON_3dPoint& point  // center of rotation
      )
{
  return circle.Rotate( sin_angle, cos_angle, axis, point );
}

BOOL ON_Cylinder::Rotate(
      double angle,              // angle in radians
      const ON_3dVector& axis, // axis of rotation
      const ON_3dPoint&  point  // center of rotation
      )
{
  return Rotate( sin(angle), cos(angle), axis, point );
}

BOOL ON_Cylinder::Translate(
      const ON_3dVector& delta
      )
{
  return circle.Translate( delta );
}

int ON_Cylinder::GetNurbForm( ON_NurbsSurface& s ) const
{
  int rc = 0;
  if ( IsValid() && height[0] < height[1] ) {
    s.Create(3,TRUE,3,2,9,2);
    s.m_knot[0][0] = s.m_knot[0][1] = 0.0;
    s.m_knot[0][2] = s.m_knot[0][3] = 0.5*ON_PI;
    s.m_knot[0][4] = s.m_knot[0][5] = ON_PI;
    s.m_knot[0][6] = s.m_knot[0][7] = 1.5*ON_PI;
    s.m_knot[0][8] = s.m_knot[0][9] = 2.0*ON_PI;

    s.m_knot[1][0] = height[0];
    s.m_knot[1][1] = height[1]; 

    ON_Circle c0 = CircleAt(height[0]);
    ON_Circle c1 = CircleAt(height[1]);
    ON_NurbsCurve n0, n1;
    c0.GetNurbForm(n0);
    c1.GetNurbForm(n1);
    int i;
    for ( i = 0; i < 9; i++ ) {
      s.SetCV(i,0,ON::homogeneous_rational,n0.CV(i));
      s.SetCV(i,1,ON::homogeneous_rational,n1.CV(i));
    }
    rc = 2;
  }
  return rc;
}
