#include "opennurbs.h"
#include "opennurbs_layer.h"

ON_OBJECT_IMPLEMENT(ON_Layer,ON_Object,"95809813-E985-11d3-BFE5-0010830122F0");

ON_Layer::ON_Layer() : m_layer_index(0), m_iges_level(-1), m_material_index(-1), m_model_index(0), m_mode(ON::normal_layer)
{}

ON_Layer::~ON_Layer()
{
  m_name.Destroy();
}

ON_Layer::ON_Layer( const ON_Layer& src ) : m_layer_index(0), m_iges_level(-1), m_material_index(-1), m_model_index(0), m_mode(ON::normal_layer)
{
  *this = src;
}

ON_Layer& ON_Layer::operator=(const ON_Layer& src )
{
  if ( this != &src ) {
    ON_Object::operator=(src);
    SetLayerName( src.LayerName() );
    SetColor( src.Color() );
    SetMode( src.Mode() );
    SetRenderMaterialIndex( src.RenderMaterialIndex() );
    SetLayerIndex( src.LayerIndex() );
    SetModelIndex( src.ModelIndex() );
    SetIgesLevel( src.IgesLevel() );
  }
  return *this;
}

BOOL ON_Layer::operator==(const ON_Layer& src) const
{
  return m_name == src.LayerName();
}

BOOL ON_Layer::operator!=(const ON_Layer& src) const
{
  return !ON_Layer::operator==(src);
}

BOOL ON_Layer::IsValid() const
{
  const int m = m_mode;
  BOOL rc = (m>=0 && m < ON::layer_mode_count) ? TRUE : FALSE;
  if ( rc )
    rc = m_name.IsEmpty() ? FALSE : TRUE;
  return rc;
}

void ON_Layer::Dump( ON_TextLog& dump ) const
{
  const wchar_t* sName = LayerName();
  if ( !sName )
    sName = L"";
  const char* sMode = 0;
  switch( Mode() )
  {
  case ON::normal_layer:
    sMode = "normal";
    break;
  case ON::hidden_layer:
    sMode = "hidden";
    break;
  case ON::locked_layer:
    sMode = "locked";
    break;
  case ON::reference_layer:
    sMode = "reference";
    break;
  default:
    sMode = "unknown";
    break;
  }
  dump.Print("index = %d\n",LayerIndex());
  dump.Print("name = \"%S\"\n",sName);
  dump.Print("layer mode = %s\n",sMode);
  dump.Print("layer rgb = "); dump.PrintRGB(Color()); dump.Print("\n");
  dump.Print("layer material index = %d\n",RenderMaterialIndex());
}

BOOL ON_Layer::Write(
       ON_BinaryArchive& file // serialize definition to binary archive
     ) const
{
  int i;
  BOOL rc = file.Write3dmChunkVersion(1,0);
  i = Mode();
  if ( rc ) rc = file.WriteInt( i );
  if ( rc ) rc = file.WriteInt( LayerIndex() );
  if ( rc ) rc = file.WriteInt( IgesLevel() );
  if ( rc ) rc = file.WriteInt( RenderMaterialIndex() );
  if ( rc ) rc = file.WriteInt( ModelIndex() );
  if ( rc ) rc = file.WriteColor( Color() );
  if ( rc ) rc = file.WriteLineStyle( LineStyle() );
  if ( rc ) rc = file.WriteString( m_name );
  return rc;
}

BOOL ON_Layer::Read(
       ON_BinaryArchive& file // restore definition from binary archive
     )
{
  int major_version=0;
  int minor_version=0;
  int mode = ON::normal_layer;
  BOOL rc = file.Read3dmChunkVersion(&major_version,&minor_version);
  if ( rc && major_version == 1 ) {
    // common to all 1.x formats
    if ( rc ) rc = file.ReadInt( &mode );
    if ( rc ) SetMode( ON::LayerMode(mode) );
    if ( rc ) rc = file.ReadInt( &m_layer_index );
    if ( rc ) rc = file.ReadInt( &m_iges_level );
    if ( rc ) rc = file.ReadInt( &m_material_index );
    if ( rc ) rc = file.ReadInt( &m_model_index );
    if ( rc ) rc = file.ReadColor( m_color );
    if ( rc ) rc = file.ReadLineStyle( m_line_style );
    if ( rc ) rc = file.ReadString( m_name );
  }
  else {
    ON_ERROR("ON_Layer::Read() encountered a layer written by future code.");
    rc = FALSE;
  }
  return rc;
}

ON::object_type ON_Layer::ObjectType() const
{
  return ON::layer_object;
}

//////////////////////////////////////////////////////////////////////
//
// Interface

BOOL ON_Layer::SetLayerName( const char* s )
{
  m_name = s;
  return IsValid();
}

BOOL ON_Layer::SetLayerName( const wchar_t* s )
{
  m_name = s;
  return IsValid();
}

const ON_wString& ON_Layer::LayerName() const
{
  return m_name;
}

void ON_Layer::SetColor( ON_Color c)
{
  m_color = c;
}

ON_Color ON_Layer::Color() const
{
  return m_color;
}

void ON_Layer::SetLineStyle( ON_LineStyle ls )
{
  m_line_style = ls;
}

ON_LineStyle ON_Layer::LineStyle() const
{
  return m_line_style;
}

// 4 mutually exclusive layer modes: normal, hidden, locked, hidden, reference
BOOL ON_Layer::SetMode( ON::layer_mode mode )
{
  BOOL rc = FALSE;
  if (mode >= 0 && mode < ON::layer_mode_count ) {
    m_mode = mode;
    rc = TRUE;
  }
  return rc;
}

ON::layer_mode ON_Layer::Mode() const
{
  return m_mode;
}

// status determined by mode
BOOL ON_Layer::IsVisible() const
{
  BOOL rc;
  switch ( Mode() ) {
  case ON::normal_layer:
  case ON::locked_layer:
  case ON::reference_layer:
    rc = TRUE;
    break;
  case ON::hidden_layer:
    rc = FALSE;
    break;
  default:
    rc = FALSE;
    break;
  }
  return rc;
}

BOOL ON_Layer::IsChangeable() const
{
  // TRUE if objects on layer can be changed (normal)
  BOOL rc;
  switch ( Mode() ) {
  case ON::normal_layer:
    rc = TRUE;
    break;
  default:
    rc = FALSE;
    break;
  }
  return rc;
}

BOOL ON_Layer::IsSelectable() const
{
  // TRUE if objects on layer are selectable (normal and reference)
  BOOL rc;
  switch ( Mode() ) {
  case ON::normal_layer:
  case ON::reference_layer:
    rc = TRUE;
    break;
  default:
    rc = FALSE;
    break;
  }
  return rc;
}

BOOL ON_Layer::SetRenderMaterialIndex( int i )
{
  m_material_index = i;
  return TRUE;
}

int ON_Layer::RenderMaterialIndex() const
{
  return m_material_index;
}

BOOL ON_Layer::SetLayerIndex( int i )
{
  m_layer_index = i;
  return TRUE;
}

int ON_Layer::LayerIndex() const
{
  return m_layer_index;
}


BOOL ON_Layer::SetIgesLevel( int level )
{
  m_iges_level = level;
  return TRUE;
}

int ON_Layer::IgesLevel() const
{
  return m_iges_level;
}

BOOL ON_Layer::SetModelIndex( int i )
{
  m_model_index = i;
  return TRUE;
}

int ON_Layer::ModelIndex() const
{
  return m_model_index;
}

