/* $Header: /src2/opennurbs/opennurbs_mesh.h 25    5/07/01 5:00p Dalelear $ */
/* $NoKeywords: $ */
/*
//
// Copyright (c) 1993-2001 Robert McNeel & Associates. All rights reserved.
// Rhinoceros is a registered trademark of Robert McNeel & Assoicates.
//
// THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
// ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
// MERCHANTABILITY ARE HEREBY DISCLAIMED.
//				
// For complete openNURBS copyright information see <http://www.opennurbs.org>.
//
////////////////////////////////////////////////////////////////
*/

#if !defined(OPENNURBS_MESH_INC_)
#define OPENNURBS_MESH_INC_

class ON_Mesh;

///////////////////////////////////////////////////////////////////////////////
//
// Class  ON_Mesh
//
class ON_CLASS ON_MeshParameters
{
  // surface meshing perameters
public:
  ON_MeshParameters();
  ~ON_MeshParameters();
  ON_MeshParameters(const ON_MeshParameters& );
  ON_MeshParameters& operator=(const ON_MeshParameters&);
 
  BOOL operator!=(const ON_MeshParameters&) const;
  BOOL operator==(const ON_MeshParameters&) const;

  // compares with mesh's mesh parameters
  BOOL operator==(const ON_Mesh&) const;
  BOOL operator!=(const ON_Mesh&) const;
  
  void Default(); 

  void Set( // set meshing parameters from a single double
    double, // 0.0 to 1.0  0.0 = fast and jagged, 1.0 = slow and accurate
    double = 0.0001  // minimum edge length
    );


  // Compare() ignores weld and curvature settings
  int Compare( const ON_MeshParameters& ) const;

  BOOL Write( ON_BinaryArchive& ) const;
  BOOL Read( ON_BinaryArchive& );

  // Meshing happens in two stages.  The first stage creates a
  // rectangular grid.  The second stage refines the grid until
  // the mesh meets all meshing requirements.  The third stage
  // combines coincident vertices if the resulting mesh is a composite.

  BOOL m_bComputeCurvature;  // FALSE - (default) - ON_Mesh::m_K[] not computed
                             // TRUE  - ON_Mesh::m_K[] computed

  BOOL m_bSimplePlanes;      // FALSE - (default) planar surfaces are meshed
                             //          using the controls below.
                             // TRUE   - planar surfaces are meshed using
                             //          minimal number of triangles and
                             //          aspect/edge controls are ignored.

  BOOL m_bRefine;            // FALSE - skip stage 2
                             // TRUE  - (default) do stage 2

  BOOL m_bJaggedSeams;       // FALSE - (default) edges of meshes of joined 
                             //          b-rep faces match with no gaps or
                             //          "T" joints.
                             // TRUE   - faces in b-reps are meshed independently.
                             //          This is faster but results in gaps and
                             //          "T" joints along seams between faces.
    
  BOOL m_bWeld;              // FALSE - (default) skip stage 3
                             // TRUE  - do stage 3

  int m_texture_range;       // 0: surface domain
                             //      each face in a polysurface uses the
                             //      surface's domain as the texture range.
                             // 1: normalized
                             //      each face has a normalized texture range 
                             //      [0,1]x[0,1].
                             // 2: packed normalized (default)
                             //      each face in a polysurface is assigned
                             //      a texture range that is a subrectangle 
                             //      of [0,1]x[0,1].  The subrectangles are 
                             //      mutually disjoint and packed into
                             //      into [0,1]x[0,1] in a way that minimizes
                             //      distortion and maximizes the coverage
                             //      of [0,1]x[0,1].  (This texture style 
                             //      is suitable for creating texture maps 
                             //      with popular 3D painting programs.)
                           
  // These controls are used in both stages

  double m_tolerance; // maximum distance from center of edge to surface
  double m_min_edge_length; // edges shorter than m_min_edge_length will
                            // not be split even if the do not meet other
                            // meshing requirements
  double m_max_edge_length; // edges longer than m_max_edge_length will
                            // be split even when they meet all other
                            // meshing requirements

  // These controls are used during stage 1 to generate the grid
  double m_grid_aspect_ratio;  // desired aspect ratio of quads in grid
                               // 0.0 = any aspect ratio is acceptable
                               // values >0 and < sqrt(2) are treated as sqrt(2)
  int    m_grid_min_count;     // minimum number of quads in initial grid
  int    m_grid_max_count;     // desired masimum number of quads in initial grid
  double m_grid_angle;         // (in radians) maximum angle between surface
                               // normal evaluated at adjacent vertices.
                               // 0.0 is treated as pi.
  double m_grid_amplification; // The parameters above generate a grid.
                               // If you want fewer quads, set m_grid_amplification
                               // to a value < 1.  If you want more quads,
                               // set m_grid_amplification to a value > 1.
                               // default = 1 and values <= 0 are treated as 1.

  // These controls are used during stage 2 to refine the grid
  double m_refine_angle;       // (in radians) maximum angle in radians between
                               // surface normal evaluated at adjacent vertices.

  // These controls are used during stage 3
  double m_combine_angle;      // If vertices of a composite mesh are coincident
                               // and the angle between the vertex normals is
                               // less than or equal to this angle (in radians),
                               // then the vertices are combined.
  int     m_face_type;         // 0 = mixed triangle and quads
                               // 1 = all triangles
                               // 2 = all quads
};

class ON_CLASS ON_MeshCurvatureStats
{
public:
  ON_MeshCurvatureStats();
  ~ON_MeshCurvatureStats();
  ON_MeshCurvatureStats(const ON_MeshCurvatureStats& );
  ON_MeshCurvatureStats& operator=(const ON_MeshCurvatureStats&);

  void Destroy();
  void EmergencyDestroy();
  
  BOOL Set( ON::curvature_style,
            int,           // Kcount,
            const ON_SurfaceCurvature*, // K[]
            const ON_3fVector*, // N[] surface normals needed for normal sectional curvatures
            double = 0.0   // if > 0, value is used for "infinity"
            );

  BOOL Write( ON_BinaryArchive& ) const;
  BOOL Read( ON_BinaryArchive& );

  // computes curvature value from principal curvatures and style
  double Curvature( double, double ) const;

  ON::curvature_style m_style;
  
  double m_infinity; // curvature values >= this are considered infinite
                     // and not used to compute the m_average or m_adev
  int    m_count_infinite; // number of "infinte" values
  int    m_count;    // count of "finite" values
  double m_mode;     // mode of "finite" values
  double m_average;  // average of "finite" values
  double m_adev;     // average deviation of "finite" values

  ON_Interval m_range;
};

///////////////////////////////////////////////////////////////////////////////
//
// Class  ON_MeshTopology
//

struct ON_MeshTopologyVertex
{
  //////////
  // m_tope_count = number of topological edges that begin or end at this
  //                topological vertex.
  int m_tope_count;

  /////////
  // m_topei[] is an array of length m_tope_count with the indices of the
  // topological edges that begin or end at this topological vertex.
  const int* m_topei;

  /////////
  // m_v_count = number of ON_Mesh vertices that correspond to this topological
  // vertex.
  int m_v_count;

  /////////
  // m_vi[] is an array of length m_v_count with the indices of the
  // ON_Mesh vertices that correspond to this topological vertex.
  const int* m_vi;
};

struct ON_MeshTopologyEdge
{
  //////////
  // m_topvi[] = indices of the topological verteices where the edge begins
  //             and ends.
  int m_topvi[2];

  //////////
  // m_topf_count = number of topological faces tat share this topological edge
  int m_topf_count;

  /////////
  // m_topfi[] is an array of length m_topf_count with the indices of the
  // topological faces that share this topological edge.
  const int* m_topfi;
};

struct ON_MeshTopologyFace
{
  //////////
  // m_topei[] = indices of the topological edges that bound the face.
  // If m_topei[2] = m_topei[3], then the face is a triangle, otherwise
  // the face is a quad.
  int m_topei[4];

  //////////
  // If m_reve[i] is 0, then the orientation of the edge matches the
  // orientation of the face.  If m_reve[i] is 1, then the orientation
  // of the edge is opposite that of the face.
  char m_reve[4];

  BOOL IsTriangle() const;
  BOOL IsQuad() const;
  BOOL IsValid( ) const;
};

class ON_CLASS ON_MeshFace
{
public:
  int vi[4]; // vertex index - vi[2]==vi[3] for tirangles
  BOOL IsTriangle() const;
  BOOL IsQuad() const;
  BOOL IsValid( 
          int // number of vertices in mesh
        ) const;
  void Flip();
};

#if defined(ON_DLL_EXPORTS) || defined(ON_DLL_IMPORTS)
// This stuff is here because of a limitation in the way Microsoft
// handles templates and DLLs.  See Microsoft's knowledge base 
// article ID Q168958 for details.
#pragma warning( push )
#pragma warning( disable : 4231 )
ON_TEMPLATE template class ON_CLASS ON_SimpleArray<ON_MeshFace>;
ON_TEMPLATE template class ON_CLASS ON_SimpleArray<ON_MeshTopologyVertex>;
ON_TEMPLATE template class ON_CLASS ON_SimpleArray<ON_MeshTopologyEdge>;
ON_TEMPLATE template class ON_CLASS ON_SimpleArray<ON_MeshTopologyFace>;
#pragma warning( pop )
#endif

class ON_CLASS ON_MeshTopology
{
  // A mesh topology class is always associated with an ON_Mesh
  // and can be retrieved by calling ON_Mesh::Topology()
public:
  ON_MeshTopology();
  ~ON_MeshTopology();

  BOOL IsValid() const;

  void Dump( ON_TextLog& ) const;

  //////////
  // The parent ON_Mesh geometry used to compute this mesh topology.
  const ON_Mesh* m_mesh;

  //////////
  // number of topoligical vertices (<= m_mesh.VertexCount())
  int TopVertexCount() const;

  //////////
  // number of topoligical edges
  int TopEdgeCount() const;
  
  //////////
  // number of topoligical faces (same as m_mesh.FaceCount())
  int TopFaceCount() const;

  ////////
  // returns 3d location of topological vertex
  const ON_3fPoint& TopVertexPoint(
    int// ON_MeshTopology vertex topology index
    ) const;

  ////////
  // returns index of edge that connects topological vertices
  // returns -1 if no edge is found.
  int TopEdge(
    int, int // ON_MeshTopology vertex topology indices
    ) const;

  ////////
  // returns ON_MeshTopology vertex topology index of a face
  // corner.  The face is triangle iv TopFaceVertex(2) = TopFaceVertex(3)
  BOOL GetTopFaceVertices(
    int, // ON_MeshTopology face topology index (= ON_Mesh face index)
    int[4] // ON_MeshTopology vertex indices returned here
    ) const;

  //////////
  // m_topv_map[] has length m_mesh.VertexCount() and 
  // m_topv[m_topv_map[vi]] is the topological mesh vertex that is assocated
  // the with the mesh vertex m_mesh.m_V[vi].
  ON_SimpleArray<int> m_topv_map;

  ////////////
  // Array of topological mesh vertices.  See the comments in the definition
  // of ON_MeshTopologyVertex for details.
  ON_SimpleArray<ON_MeshTopologyVertex> m_topv;

  ////////////
  // Array of topological mesh edges.  See the comments in the definition
  // of ON_MeshTopologyEdge for details.
  ON_SimpleArray<ON_MeshTopologyEdge> m_tope;

  ////////////
  // Array of topological mesh faces.  The topological face
  // m_topf[fi] corresponces with the mesh face ON_Mesh.m_F[fi].
  // See the comments in the definition of ON_MeshTopologyFace for details.
  // To get the indices of the mesh topology vertices at the face corners
  // use topvi = m_topv_map[m_mesh.m_F[fi].vi[n]]
  ON_SimpleArray<ON_MeshTopologyFace> m_topf;

private:
  friend class ON_Mesh;

  BOOL Create();
  void Destroy();
  void EmergencyDestroy();

  int* GetIntArray(int);

  // efficient workspaces for
  struct memchunk
  {
    struct memchunk* next;
  } *m_memchunk;
  BOOL m_bIsValid;

private:
  // no implementation
  ON_MeshTopology(const ON_MeshTopology&);
  ON_MeshTopology& operator=(const ON_MeshTopology&);
};


class ON_CLASS ON_Mesh : public ON_Geometry
{
  ON_OBJECT_DECLARE(ON_Mesh);
public:
  ON_Mesh();
  ON_Mesh(
    int,  // initial face array capacity
    int,  // initial vertex array capacity
    BOOL, // TRUE if mesh has vertex normals
    BOOL  // TRUE if mesh has texture coordinates
    );
  ON_Mesh( const ON_Mesh& );
  ON_Mesh& operator=( const ON_Mesh& );
  ~ON_Mesh();

  void Destroy();
  void EmergencyDestroy(); // Call only when memory used by this class's
                           // members will soon become invalid for reasons 
                           // beyond your control. EmergencyDestroy() zeros
                           // anything that could possibly cause
                           // ~ON_Mesh() to crash.  Calling
                           // EmergencyDestroy() under normal conditions 
                           // will result in ~ON_Mesh() leaking
                           // memory.

  /////////////////////////////////////////////////////////////////
  // ON_Object overrides

  BOOL IsValid() const;

  void Dump( ON_TextLog& ) const; // for debugging

  BOOL Write( ON_BinaryArchive& ) const;

  BOOL Read( ON_BinaryArchive& );

  ON::object_type ObjectType() const;

  /////////////////////////////////////////////////////////////////
  // ON_Geometry overrides

  int Dimension() const;

  BOOL GetBBox( // returns TRUE if successful
         double*,    // minimum
         double*,    // maximum
         BOOL = FALSE  // TRUE means grow box
         ) const;

  BOOL Transform( 
         const ON_Xform&
         );

  BOOL SwapCoordinates(
        int, int        // indices of coords to swap
        );

  /////////////////////////////////////////////////////////////////
  // Interface
  // 

  // creation
  BOOL SetVertex(
         int,              // vertex index
         const ON_3dPoint& // vertex location
         );
  BOOL SetVertex(
         int,              // vertex index
         const ON_3fPoint& // vertex location
         );
  BOOL SetVertexNormal(
         int,               // vertex index
         const ON_3dVector& // unit normal
         );
  BOOL SetVertexNormal(
         int,               // vertex index
         const ON_3fVector& // unit normal
         );
  BOOL SetTextureDomain( 
        const ON_Interval&, // first texture parameter domain
        const ON_Interval&  // second texture parameter domain
        );
  BOOL SetTextureCoord(
         int,               // vertex index
         double, double     // texture coordinates
         );
  BOOL SetTriangle(
         int, // face index
         int,int,int // vertex indices
         );
  BOOL SetQuad(
         int, // face index
         int,int,int,int // vertex indices
         );

  // query
  int VertexCount() const;
  int FaceCount() const;
  int QuadCount(); // number of faces that are quads
  int TriangleCount(); // number of faces that are triangles
  BOOL HasVertexNormals() const; // normals at vertices
  BOOL HasFaceNormals() const;
  BOOL HasTextureCoordinates() const;
  BOOL HasPrincipalCurvatures() const;
  BOOL HasVertexColors() const;

  BOOL GetVertexNormalVectorBoundingBox( // returns TRUE if successful
         double[3],    // minimum
         double[3],    // maximum
         BOOL = FALSE  // TRUE means grow box
         ) const;

  BOOL GetTextureCoordinateBoundingBox( // returns TRUE if successful
         double[2],    // minimum
         double[2],    // maximum
         BOOL = FALSE  // TRUE means grow box
         ) const;

  BOOL GetCurvatureStats( 
         ON::curvature_style, 
         ON_MeshCurvatureStats& 
         ) const;

  void InvalidateVertexBoundingBox(); // Call if defining geometry is changed by 
                             // directly manipulating the m_V[] array.
  void InvalidateVertexNormalBoundingBox(); // Call if defining geometry is changed by 
                             // directly manipulating the m_N[] array.
  void InvalidateTextureCoordinateBoundingBox(); // Call if defining geometry is changed by 
                             // directly manipulating the m_T[] array.
  void InvalidateCurvatureStats(); // Call if defining geometry is changed by 
                             // directly manipulating the m_T[] array.
  void InvalidateBoundingBoxes(); // Invalidates all cached bounding box information.


  void Flip(); // reverses face orientations and flips vertex and face normals

  void FlipVertexNormals(); // reverses vertex normals
  void FlipFaceNormals(); // reverses face normals
  void FlipFaceOrientation(); // reverses face orientation (does nothing to normals)

  void SetMeshParameters( const ON_MeshParameters& );
  const ON_MeshParameters* MeshParameters() const;
  void DeleteMeshParameters();

  
  BOOL UnitizeVertexNormals();
  BOOL UnitizeFaceNormals();
  BOOL CountQuads();
  BOOL ConvertQuadsToTriangles();

  BOOL ComputeFaceNormals();   // compute face normals for all faces
  BOOL ComputeFaceNormal(int); // computes face normal of indexed face

  int CullDegenerateFaces(); // returns number of degenerate faces

  BOOL ComputeVertexNormals();    // uses face normals to cook up a vertex normal
  
  //////////
  // Scales textures so the texture domains are [0,1] and
  // eliminates any texture rotations.
  BOOL NormalizeTextureCoordinates();

  //////////
  // If the mesh was created by meshing a parametric surface, 
  // then GetSurfaceParameters() can compute the surface
  // parameters from texture coordinates that may have been 
  // normalized, packed, or rotated
  BOOL GetSurfaceParameters(
         double, double,   // texture coordinates
         double*, double*  // associated surface parameters
         ) const;

  BOOL EvaluateMeshGeometry( const ON_Surface& ); // evaluate surface at tcoords
                                                  // to set mesh geometry

  // finds all coincident vertices and merges them if break angle is small enough
  BOOL CombineCoincidentVertices( 
          ON_3fVector, // coordinate tols for considering vertices
                       // to be coincident
          double  // cosine normal angle tolerance in radians
                  // if vertices are coincident, then they are combined
                  // if NormalA o NormalB >= this value
          );

  void Append( const ON_Mesh& ); // appends a copy of mesh to this and updates
                                 // indices of appended mesh parts

  void SetClosed(BOOL);
  BOOL IsClosed() const;

  ///////////////////////////////////////////////////////////////////////
  //
  // mesh topology
  //
  // In order to keep the mesh facet definition simple and make the mesh
  // definition easily used in common rendering application, if two facets
  // share a vertex location but have different normals, curvatures, 
  // textures, etc., at that common vertex location, then the vertex is
  // duplicated.  When the topology of the mesh needs to be known,
  // use Topology() to get a class that provides complete topological
  // information about the mesh.
  const ON_MeshTopology& Topology() const;

  ///////////////////////////////////////////////////////////////////////
  // If you modify the mesh in any way that may change its topologyt,
  // then call DestroyTopology().  Specifically if you add or remove
  // vertices or face, change vertex locations, or change the face m_vi[]
  // values, then you must call DestroyTopology().
  void DestroyTopology();

  /////////////////////////////////////////////////////////////////
  // Implementation - mesh geometry

  ON_3fPointArray             m_V;  // vertex locations
                                    // In a case where adjacent facets share a vertex
                                    // location but have distinct normals or texture
                                    // coordinates at that location, the vertex must
                                    // be duplicated.

  ON_SimpleArray<ON_MeshFace> m_F;  // facets (triangle or quad)

  ON_3fVectorArray            m_N;  // OPTIONAL vertex unit normals
                                    // Either m_N[] has zero count or it has the same
                                    // count as m_V[], in which case m_N[j] reports
                                    // the unit normal at m_V[j].

  ON_3fVectorArray            m_FN; // OPTIONAL face unit normals
                                    // Either m_FN[] has zero count or it has the same
                                    // count as m_F[], in which case m_FN[j] reports
                                    // the unit normal for the facet m_F[j].

  /////////////////////////////////////////////////////////////////
  // Implementation - texture coordinates

  ON_2fPointArray             m_T;  // OPTIONAL texture coordinates
                                    // Either m_T[] has zero count or it has the same
                                    // count as m_V[], in which case m_T[j] reports
                                    // the texture coordinates at m_V[j].
  BOOL        m_srf_tex_rotate; // 0 = no rotate, 1 = texture coordinates rotated 90 degrees
                                //      counter clockwise from surface parameterization.
  ON_Interval m_tex_domain[2]; // Texture coordinate domain.
                               //   m_tex_domain[0].Min() <= m_T[i].x <= m_tex_domain[0].Max()
                               //   m_tex_domain[1].Min() <= m_T[i].y <= m_tex_domain[1].Max()
  ON_Interval m_srf_domain[2]; // If mesh was generated from a parametric surface,
                               // these intervals record the domain of the portion of the
                               // surface that corresponds the the m_tex_domain[]s above.
                               // See ON_Mesh::GetSurfaceParameters() for an example showing
                               // the relationship between m_tex_domain[], m_srf_domain[],
                               // and m_srf_tex_rotate.
  double m_texture_scale[2];   // If non-zero, then m_texture_scale[] reports a
                               // world coordinate texture map size that would
                               // minimize the texture distortion.

  /////////////////////////////////////////////////////////////////
  // Implementation - curvature

  ON_SimpleArray<ON_SurfaceCurvature> m_K;  // OPTIONAL surface curvatures
                                            // Either m_K[] has zero count or it has the same
                                            // count as m_V[], in which case m_K[j] reports
                                            // the surface curvatures at m_V[j].

  /////////////////////////////////////////////////////////////////
  // Implementation - false color
  ON_SimpleArray<ON_Color>    m_C;  // OPTIONAL vertex color
                                    // Either m_C[] has zero count or it has the same
                                    // count as m_V[], in which case m_C[j] reports
                                    // the color assigned to m_V[j].

  /////////////////////////////////////////////////////////////////
  // Implementation - runtime UI information
  const ON_Object* m_parent; // runtime parent geometry (use ...::Cast() to get it)
  int m_material_index;      // -1 = default if >= 0, this material overrides any other

private:
  /////////////////////////////////////////////////////////////////
  // Implementation - mesh topology
  ON_MeshTopology m_top;

  ON_MeshParameters* m_mesh_parameters; // If mesh was created from a parametric surface,
                                        // these parameters were used to create the mesh.
  int                         m_invalid_count;
  int                         m_quad_count;
  int                         m_triangle_count;

  int m_closed; // -1=unknown 0=mesh is not closed, 1 = mesh is closed, 2 = closed with duplicate vertices

  // The bounding boxes are valid if m_?box[0][0] <= m_?box[0][1];
  float m_vbox[2][3]; // 3d bounding box of all referenced vertices
  float m_nbox[2][3]; // 3d bounding box of all referenced unit normals 
                      // (for estimation of Gauss map bounds)
  float m_tbox[2][2]; // 2d bounding box of all referenced texture coordinates
  ON_MeshCurvatureStats* m_kstat[4]; // gaussian,mean,min,max,sectionx,sectiony,sectionz

private:
  BOOL Write_1( ON_BinaryArchive& ) const; // uncompressed 1.x format
  BOOL Write_2( int, ON_BinaryArchive& ) const; // compressed 2.x format
  BOOL Read_1( ON_BinaryArchive& );
  BOOL Read_2( int, ON_BinaryArchive& );
  BOOL WriteFaceArray( int, int, ON_BinaryArchive& ) const;
  BOOL ReadFaceArray( int, int, ON_BinaryArchive& );
};

#endif

