/* $Header: /src2/opennurbs/opennurbs_nurbscurve.cpp 27    7/24/01 4:04p Dalelear $ */
/* $NoKeywords: $ */
/*
//
// Copyright (c) 1993-2001 Robert McNeel & Associates. All rights reserved.
// Rhinoceros is a registered trademark of Robert McNeel & Assoicates.
//
// THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
// ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
// MERCHANTABILITY ARE HEREBY DISCLAIMED.
//				
// For complete openNURBS copyright information see <http://www.opennurbs.org>.
//
////////////////////////////////////////////////////////////////
*/

#include "opennurbs.h"

ON_OBJECT_IMPLEMENT(ON_NurbsCurve,ON_Curve,"4ED7D4DD-E947-11d3-BFE5-0010830122F0");

ON_NurbsCurve::ON_NurbsCurve()
{
  Initialize();
}

ON_NurbsCurve::ON_NurbsCurve( const ON_NurbsCurve& src )
{
  Initialize();
  *this = src;
}

ON_NurbsCurve::ON_NurbsCurve( int dim, BOOL bIsRational, int order, int cv_count )
{
  Initialize();
  Create(dim,bIsRational,order,cv_count);
}

ON_NurbsCurve::~ON_NurbsCurve()
{
  Destroy();
}

int ON_NurbsCurve::Dimension() const
{
  return m_dim;
}

BOOL ON_NurbsCurve::IsRational() const
{
  return m_is_rat;
}

int ON_NurbsCurve::CVSize() const
{
  return (m_is_rat&&m_dim>0) ? m_dim+1 : m_dim;
}

int ON_NurbsCurve::Order( void ) const
{
  return m_order;
}

int ON_NurbsCurve::CVCount( void ) const
{
  return m_cv_count;
}

int ON_NurbsCurve::KnotCount( void ) const
{
  return ON_KnotCount( m_order, m_cv_count );
}

double* ON_NurbsCurve::CV( int i )
{
  return (m_cv) ? (m_cv + i*m_cv_stride) : NULL;
}

const double* ON_NurbsCurve::CV( int i ) const
{
  return (m_cv) ? (m_cv + i*m_cv_stride) : NULL;
}

double ON_NurbsCurve::Weight( int i ) const
{
  return (m_cv && m_is_rat) ? m_cv[i*m_cv_stride+m_dim] : 1.0;
}


double ON_NurbsCurve::Knot( int i ) const
{
  return (m_knot) ? m_knot[i] : 0.0;
}

int ON_NurbsCurve::KnotMultiplicity( int i ) const
{
  return ON_KnotMultiplicity(m_order,m_cv_count,m_knot,i);
}

const double* ON_NurbsCurve::Knot() const
{
  return m_knot;
}

double ON_NurbsCurve::SuperfluousKnot( int end ) const
{
  return(m_knot) ? ON_SuperfluousKnot(m_order,m_cv_count,m_knot,end) : 0.0;
}

BOOL ON_NurbsCurve::Create(
        int dim,      // dimension (>= 1)
        BOOL is_rat,  // TRUE to make a rational NURBS
        int order,    // order (>= 2)
        int cv_count  // cv count (>= order)
        )
{
  if ( dim < 1 )
    return FALSE;
  if ( order < 2 )
    return FALSE;
  if ( cv_count < order )
    return FALSE;
  m_dim = dim;
  m_is_rat = (is_rat) ? TRUE : FALSE;
  m_order = order;
  m_cv_count = cv_count;
  m_cv_stride = (m_is_rat) ? m_dim+1 : m_dim;
  BOOL rc = ReserveKnotCapacity( KnotCount() );
  if ( !ReserveCVCapacity( CVCount()*m_cv_stride ) )
    rc = FALSE;
  return rc;
}

void ON_NurbsCurve::Destroy()
{
  double* cv = ( m_cv && m_cv_capacity ) ? m_cv : NULL;
  double* knot = ( m_knot && m_knot_capacity ) ? m_knot : NULL;
  Initialize();
  if ( cv )
    onfree(cv);
  if ( knot )
    onfree(knot);
}

void ON_NurbsCurve::EmergencyDestroy()
{
  Initialize();
}


void ON_NurbsCurve::Initialize()
{
  m_dim = 0;
  m_is_rat = 0;
  m_order = 0;
  m_cv_count = 0;
  m_knot_capacity = 0;
  m_knot = 0;
  m_cv_stride = 0;
  m_cv_capacity = 0;
  m_cv = 0;
}


ON_NurbsCurve& ON_NurbsCurve::operator=( const ON_NurbsCurve& src )
{
  if ( this != &src ) {
    ON_Curve::operator=(src);

    m_dim = src.m_dim;
    m_is_rat = src.m_is_rat;
    m_order = src.m_order;
    m_cv_count = src.m_cv_count;
    m_cv_stride = m_is_rat ? m_dim+1 : m_dim;
    if ( src.m_knot ) {
      // copy knot array
      ReserveKnotCapacity( KnotCount() );
      memcpy( m_knot, src.m_knot, KnotCount()*sizeof(*m_knot) );
    }
    if ( src.m_cv ) {
      // copy cv array
      ReserveCVCapacity( m_cv_stride*m_cv_count );
      const int dst_cv_size = CVSize()*sizeof(*m_cv);
      const int src_stride = src.m_cv_stride;
      const int dst_stride = m_cv_stride;
      const double *src_cv = src.CV(0);
      double *dst_cv = m_cv;
      if ( src_stride == dst_stride ) {
        memcpy( dst_cv, src_cv, m_cv_count*dst_stride*sizeof(*dst_cv) );
      }
      else {
        int i;
        for ( i = 0; i < m_cv_count; i++ ) {
          memcpy( dst_cv, src_cv, dst_cv_size );
          dst_cv += dst_stride;
          src_cv += src_stride;
        }
      }
    }
  }
  return *this;
}

void ON_NurbsCurve::Dump( ON_TextLog& dump ) const
{
  dump.Print( "ON_NurbsCurve dim = %d is_rat = %d\n"
               "        order = %d cv_count = %d\n",
               m_dim, m_is_rat, m_order, m_cv_count );
  dump.Print( "Knot Vector ( %d knots )\n", KnotCount() );
  dump.PrintKnotVector( m_order, m_cv_count, m_knot );
  dump.Print( "Control Points  %d %s points\n"
               "  index               value\n",
               m_cv_count, 
               (m_is_rat) ? "rational" : "non-rational" );
  if ( !m_cv ) {
    dump.Print("  NULL cv array\n");
  }
  else {
    dump.PrintPointList( m_dim, m_is_rat, 
                      m_cv_count, m_cv_stride,
                      m_cv, 
                      "  CV" );
  }
}

BOOL ON_NurbsCurve::IsValid() const
{
  BOOL rc = FALSE;
  if ( m_dim > 0 && 
       m_order >= 2 && 
       m_cv_count >= m_order && 
       m_cv_stride >= (m_is_rat)?m_dim+1 : m_dim  &&
       m_cv != NULL &&
       m_knot != NULL
       ) {
    if ( ON_IsValidKnotVector( m_order, m_cv_count, m_knot ) ) {
      rc = TRUE;
      if ( m_is_rat ) {
        // weights at fully multiple knots must be nonzero
        // partial test for weight function being zero
        double sign = 0.0;
        const double* w = &m_cv[m_dim];
        int zcount = 0;
        int i;
        for ( i = 0; i < m_cv_count && rc; i++, w += m_cv_stride ) {
          if ( *w == 0.0 )
            zcount++;
          else
            zcount = 0;
          if ( zcount >= m_order )
            rc = FALSE; // denominator is zero for entire span
          else if ( m_knot[i] == m_knot[i+m_order-2] ) {
            if ( *w == 0.0 )
              rc = FALSE;
            else if (sign == 0.0) {
              sign = (*w > 0.0) ? 1.0 : -1.0;
            }
            else if ( *w * sign <= 0.0 ) {
              rc = FALSE;
            }
          }
        }
      }
    }
  }
  return rc;
}

BOOL ON_NurbsCurve::GetBBox( // returns TRUE if successful
       double* boxmin,    // minimum
       double* boxmax,    // maximum
       BOOL bGrowBox  // TRUE means grow box
       ) const
{
  return ON_GetPointListBoundingBox( m_dim, m_is_rat, m_cv_count, m_cv_stride, m_cv, boxmin, boxmax, bGrowBox );
}

BOOL ON_NurbsCurve::Transform( const ON_Xform& xform )
{
  TransformUserData(xform);
  return ON_TransformPointList( m_dim, m_is_rat, m_cv_count, m_cv_stride, m_cv, xform );
}

BOOL ON_NurbsCurve::Write(
       ON_BinaryArchive& file // open binary file
     ) const
{
  // NOTE - check legacy I/O code if changed
  BOOL rc = file.Write3dmChunkVersion(1,0);
  if (rc) {
    if (rc) rc = file.WriteInt( m_dim );
    if (rc) rc = file.WriteInt( m_is_rat );
    if (rc) rc = file.WriteInt( m_order );
    if (rc) rc = file.WriteInt( m_cv_count );
    if (rc) rc = file.WriteInt( 0 ); // reserved - legacy flag values
    if (rc) rc = file.WriteInt( 0 ); // reserved   
    
    // write invalid bounding box - may be used in future
    if (rc) {
      ON_BoundingBox bbox;
      rc = file.WriteBoundingBox(bbox);
    }

    // write knots
    int count = m_knot ? KnotCount() : 0;
    if (rc) rc = file.WriteInt(count);
    if (rc) rc = file.WriteDouble( count, m_knot );

    // write cvs
    const int cv_size = CVSize();
    count = ( m_cv && cv_size > 0 && m_cv_count > 0  && m_cv_stride >= cv_size ) 
          ? m_cv_count
          : 0;
    if (rc) rc = file.WriteInt(count);
    if (rc && count > 0 ) {
      int i;
      for ( i = 0; i < m_cv_count && rc; i++ ) {
        rc = file.WriteDouble( cv_size, CV(i) );
      }
    }

  }
  return rc;
}

BOOL ON_NurbsCurve::Read(
       ON_BinaryArchive& file // open binary file
     )
{
  // NOTE - check legacy I/O code if changed
  Destroy();
  int major_version = 0;
  int minor_version = 0;
  BOOL rc = file.Read3dmChunkVersion(&major_version,&minor_version);
  if (rc && major_version==1) {
    // common to all 1.x versions
    int dim = 0, is_rat = 0, order = 0, cv_count = 0;
    int reserved1 = 0, reserved2 = 0;
    if (rc) rc = file.ReadInt( &dim );
    if (rc) rc = file.ReadInt( &is_rat );
    if (rc) rc = file.ReadInt( &order );
    if (rc) rc = file.ReadInt( &cv_count );
    if (rc) rc = file.ReadInt( &reserved1 ); // reserved - legacy flag values
    if (rc) rc = file.ReadInt( &reserved2 ); // reserved    

    // bounding box
    if (rc) {
      // write invalid bounding box - may be used in future
      ON_BoundingBox bbox;
      rc = file.ReadBoundingBox( bbox );
    }
    
    Create( dim, is_rat, order, cv_count );

    // read knots
    int count = 0;
    if (rc) rc = file.ReadInt(&count);
    if (rc ) rc = ReserveKnotCapacity(count);
    if (rc) rc = file.ReadDouble( count, m_knot );

    // read cvs
    count = 0;
    if (rc) rc = file.ReadInt(&count);
    const int cv_size = CVSize();
    if (rc) rc = ReserveCVCapacity( count*cv_size );
    if (count > 0 && cv_size > 0 && rc ) {
      int i;
      for ( i = 0; i < m_cv_count && rc; i++ ) {
        rc = file.ReadDouble( cv_size, CV(i) );
      }
    }
  }
  if ( !rc )
    Destroy();
  return rc;
}

ON_Interval ON_NurbsCurve::Domain() const
{
  ON_Interval d;
  if ( !ON_GetKnotVectorDomain( m_order, m_cv_count, m_knot, &d.m_t[0], &d.m_t[1] ) )
    d.Destroy();
  return d;
}

BOOL ON_NurbsCurve::SetDomain( double t0, double t1 )
{
  BOOL rc = FALSE;
  if ( m_order >= 2 && m_cv_count >= m_order && m_knot && t0 < t1 ) {
    const double k0 = m_knot[m_order-2];
    const double k1 = m_knot[m_cv_count-1];
    if ( k0 == t0 && k1 == t1 )
      rc = TRUE;
    else if ( k0 < k1 ) {
      const double d = (t1-t0)/(k1-k0);
      const double km = 0.5*(k0+k1);
      const int knot_count = KnotCount();
      int i;
      for ( i = 0; i < knot_count; i++ ) {
        if ( m_knot[i] <= km ) {
          m_knot[i] = (m_knot[i] - k0)*d + t0;
        }
        else {
          m_knot[i] = (m_knot[i] - k1)*d + t1;
        }
      }
    }
  }
  return rc;
}

int ON_NurbsCurve::SpanCount() const
{
  // number of smooth spans in curve
  return ON_KnotVectorSpanCount( m_order, m_cv_count, m_knot );
}

BOOL ON_NurbsCurve::GetSpanVector( // span "knots" 
       double* s // array of length SpanCount() + 1 
       ) const
{
  return ON_GetKnotVectorSpanVector( m_order, m_cv_count, m_knot, s );
}

int ON_NurbsCurve::Degree() const
{
  return m_order >= 2 ? m_order-1 : 0;
}


BOOL
ON_NurbsCurve::IsLinear( // TRUE if curve locus is a line segment
      double tolerance // tolerance to use when checking linearity
      ) const
{
  BOOL rc = FALSE;
  if ( m_order == 2 && m_cv_count == 2 ) {
    rc = TRUE;
  }
  else if ( IsClamped() ) {
    ON_Line L;
    ON_3dPoint P;
    double t0, t, d;
    int i;
    if ( tolerance <= 0.0 )
      tolerance = ON_ZERO_TOLERANCE;
    GetCV(0,L.from);
    GetCV(m_cv_count-1,L.to);
    t0 = 0.0;
    rc = TRUE;
    for ( i = 1; rc && i < m_cv_count-1; i++ ) {
      GetCV(i,P);
      rc = L.ClosestPointTo(P,&t);
      if (rc) {
        if ( t < t0 )
          rc = FALSE;
        d = P.DistanceTo(L.PointAt(t));
        if ( d > tolerance )
          rc = FALSE;
        t0 = t;
      }
    }
  }
  return rc;
}

BOOL
ON_NurbsCurve::IsArc( // TRUE if curve locus in an arc or circle
      const ON_Plane* plane, // if not NULL, test is performed in this plane
      ON_Arc* arc,         // if not NULL and TRUE is returned, then arc
                              // arc parameters are filled in
      double tolerance // tolerance to use when checking linearity
      ) const
{
  return FALSE;
}

BOOL
ON_NurbsCurve::IsPlanar(
      ON_Plane* plane, // if not NULL and TRUE is returned, then plane parameters
                         // are filled in
      double tolerance // tolerance to use when checking linearity
      ) const
{
  return FALSE;
}

BOOL
ON_NurbsCurve::IsInPlane(
      const ON_Plane& plane, // plane to test
      double tolerance // tolerance to use when checking linearity
      ) const
{
  return FALSE;
}

BOOL
ON_NurbsCurve::GetParameterTolerance( // returns tminus < tplus: paramters tminus <= s <= tplus
       double t,       // t = parameter in domain
       double* tminus, // tminus
       double* tplus   // tplus
       ) const
{
  BOOL rc = FALSE;
  ON_Interval d = Domain();
  if ( d.IsIncreasing() ) {
    const double* knot = Knot();
    const int order = Order();
    const int cv_count = CVCount();
    if ( t <= d[0] )
      d.m_t[0] = knot[order-1];
    else if ( t >= d[1] )
      d.m_t[1] = knot[cv_count-2];
    rc = ON_GetParameterTolerance( d.m_t[0], d.m_t[1], t, tminus, tplus );
  }
  return rc;
}

BOOL 
ON_NurbsCurve::Evaluate( // returns FALSE if unable to evaluate
       double t,       // evaluation parameter
       int der_count,  // number of derivatives (>=0)
       int v_stride,   // v[] array stride (>=Dimension())
       double* v,      // v[] array of length stride*(ndir+1)
       int side,       // optional - determines which side to evaluate from
                       //         0 = default
                       //      <  0 to evaluate from below, 
                       //      >  0 to evaluate from above
       int* hint       // optional - evaluation hint (int) used to speed
                       //            interative evaluations
       ) const
{
  BOOL rc = FALSE;
  const int span_index = ON_SpanIndex(m_order,m_cv_count,m_knot,t,side,(hint)?*hint:0);
  rc = ON_EvaluateNurbsSpan(
     m_dim, m_is_rat, m_order, 
     m_knot + span_index, 
     m_cv_stride, m_cv + (m_cv_stride*span_index),
     der_count, 
     t,
     v_stride, v 
     );
  if ( hint ) 
    *hint = span_index;
  return rc;
}


BOOL 
ON_NurbsCurve::IsClosed() const
{
  BOOL bIsClosed = FALSE;
  if ( m_dim > 0 && m_cv_count >= 4 ) {
    if ( ON_IsKnotVectorClamped( m_order, m_cv_count, m_knot ) ) {
      if ( ON_IsPointListClosed( m_dim, m_is_rat, m_cv_count, m_cv_stride, m_cv ) )
        bIsClosed = TRUE;
    }
    else if ( IsPeriodic() ) {
      bIsClosed = TRUE;
    }
    else {
      // evaluation test required
      bIsClosed = ON_Curve::IsClosed();
    }
  }
  return bIsClosed;
}

BOOL 
ON_NurbsCurve::IsPeriodic() const
{
  BOOL bIsPeriodic = ON_IsKnotVectorPeriodic( m_order, m_cv_count, m_knot );
  if ( bIsPeriodic ) {
    int i0 = m_order-2;
    int i1 = m_cv_count-1;
    const double* cv0 = m_cv + i0*m_cv_stride;
    const double* cv1 = m_cv + i1*m_cv_stride;
    for ( /*empty*/; i0 >= 0; i0--, i1-- ) {
      if ( ON_ComparePoint( m_dim, m_is_rat, cv0, cv1 ) )
        return FALSE;
      cv0 -= m_cv_stride;
      cv1 -= m_cv_stride;      
    }
  }
  return bIsPeriodic;
}

BOOL 
ON_NurbsCurve::SetWeight( int i, double w )
{
  BOOL rc = FALSE;
  if ( m_is_rat ) {
    double* cv = CV(i);
    if (cv) {
      cv[m_dim] = w;
      rc = TRUE;
    }
  }
  else if ( w == 1.0 ) {
    rc = TRUE;
  }
  return rc;
}

BOOL 
ON_NurbsCurve::SetCV( int i, ON::point_style style, const double* Point )
{
  BOOL rc = TRUE;
  int k;
  double w;

  // feeble but fast check for properly initialized class
  if ( !m_cv || i < 0 || i >= m_cv_count )
    return FALSE;

  double* cv = m_cv + i*m_cv_stride;

  switch ( style ) {

  case ON::not_rational:  // input Point is not rational
    memcpy( cv, Point, m_dim*sizeof(*cv) );
    if ( IsRational() ) {
      // NURBS curve is rational - set weight to one
      cv[m_dim] = 1.0;
    }
    break;

  case ON::homogeneous_rational:  // input Point is homogeneous rational
    if ( IsRational() ) {
      // NURBS curve is rational
      memcpy( cv, Point, (m_dim+1)*sizeof(*cv) );
    }
    else {
      // NURBS curve is not rational
      w = (Point[m_dim] != 0.0) ? 1.0/Point[m_dim] : 1.0;
      for ( k = 0; k < m_dim; k++ ) {
        cv[k] = w*Point[k];
      }
    }
    break;

  case ON::euclidean_rational:  // input Point is euclidean rational
    if ( IsRational() ) {
      // NURBS curve is rational - convert euclean point to homogeneous form
      w = Point[m_dim];
      for ( k = 0; k < m_dim; k++ )
        cv[i] = w*Point[i];
      cv[m_dim] = w;
    }
    else {
      // NURBS curve is not rational
      memcpy( cv, Point, m_dim*sizeof(*cv) );
    }
    break;

  case ON::intrinsic_point_style:
    memcpy( cv, Point, CVSize()*sizeof(*cv) );
    break;
    
  default:
    rc = FALSE;
    break;
  }
  return rc;
}

BOOL 
ON_NurbsCurve::SetCV( int i, const ON_3dPoint& point )
{
  BOOL rc = FALSE;
  double* cv = CV(i);
  if ( cv ) {
    cv[0] = point.x;
    if ( m_dim > 1 ) {
      cv[1] = point.y;
      if ( m_dim > 2 )
        cv[2] = point.z;
      if ( m_dim > 3 ) {
        memset( &cv[3], 0, (m_dim-3)*sizeof(*cv) );
      }
    }
    if ( m_is_rat ) {
      cv[m_dim] = 1.0;
    }
    rc = TRUE;
  }
  return rc;
}

BOOL 
ON_NurbsCurve::SetCV( int i, const ON_4dPoint& point )
{
  BOOL rc = FALSE;
  double* cv = CV(i);
  if ( cv ) {
    if ( m_is_rat ) {
      cv[0] = point.x;
      if ( m_dim > 1 ) {
        cv[1] = point.y;
        if ( m_dim > 2 )
          cv[2] = point.z;
        if ( m_dim > 3 ) {
          memset( &cv[3], 0, (m_dim-3)*sizeof(*cv) );
        }
      }
      cv[m_dim] = point.w;
      rc = TRUE;
    }
    else {
      double w;
      if ( point.w != 0.0 ) {
        w = 1.0/point.w;
        rc = TRUE;
      }
      else {
        w = 1.0;
      }
      cv[0] = w*point.x;
      if ( m_dim > 1 ) {
        cv[1] = w*point.y;
        if ( m_dim > 2 ) {
          cv[2] = w*point.z;
        }
        if ( m_dim > 3 ) {
          memset( &cv[3], 0, (m_dim-3)*sizeof(*cv) );
        }
      }
    }
  }
  return rc;
}

BOOL 
ON_NurbsCurve::GetCV( int i, ON::point_style style, double* Point ) const
{
  const double* cv = CV(i);
  if ( !cv )
    return FALSE;
  int dim = Dimension();
  double w = ( IsRational() ) ? cv[dim] : 1.0;
  switch(style) {
  case ON::euclidean_rational:
    Point[dim] = w;
    // no break here
  case ON::not_rational:
    if ( w == 0.0 )
      return FALSE;
    w = 1.0/w;
    while(dim--) *Point++ = *cv++ * w;
    break;
  case ON::homogeneous_rational:
    Point[dim] = w;
    memcpy( Point, cv, dim*sizeof(*Point) );
    break;
  case ON::intrinsic_point_style:
    memcpy( Point, cv, CVSize()*sizeof(*Point) );
    break;
  default:
    return FALSE;
  }
  return TRUE;
}

BOOL 
ON_NurbsCurve::GetCV( int i, ON_3dPoint& point ) const
{
  BOOL rc = FALSE;
  const double* cv = CV(i);
  if ( cv ) {
    if ( m_is_rat ) {
      if (cv[m_dim] != 0.0) {
        const double w = 1.0/cv[m_dim];
        point.x = cv[0]*w;
        point.y = (m_dim>1)? cv[1]*w : 0.0;
        point.z = (m_dim>2)? cv[2]*w : 0.0;
        rc = TRUE;
      }
    }
    else {
      point.x = cv[0];
      point.y = (m_dim>1)? cv[1] : 0.0;
      point.z = (m_dim>2)? cv[2] : 0.0;
      rc = TRUE;
    }
  }
  return rc;
}

BOOL 
ON_NurbsCurve::GetCV( int i, ON_4dPoint& point ) const
{
  BOOL rc = FALSE;
  const double* cv = CV(i);
  if ( cv ) {
    point.x = cv[0];
    point.y = (m_dim>1)? cv[1] : 0.0;
    point.z = (m_dim>2)? cv[2] : 0.0;
    point.w = (m_is_rat) ? cv[m_dim] : 1.0;
    rc = TRUE;
  }
  return rc;
}

BOOL 
ON_NurbsCurve::SetKnot( int knot_index, double k )
{
  if ( knot_index < 0 || knot_index >= KnotCount() )
    return FALSE;
  m_knot[knot_index] = k;
  return TRUE;
}

BOOL
ON_NurbsCurve::Reverse()
{
  BOOL rc0 = ON_ReverseKnotVector( m_order, m_cv_count, m_knot );
  BOOL rc1 = ON_ReversePointList( m_dim, m_is_rat, m_cv_count, m_cv_stride, m_cv );
  return rc0 && rc1;
}

BOOL
ON_NurbsCurve::SwapCoordinates( int i, int j )
{
  return  ON_SwapPointListCoordinates( m_cv_count, m_cv_stride, m_cv, i, j );
}

double ON_NurbsCurve::GrevilleAbcissa(
         int gindex  // index (0 <= index < CVCount(dir)
         ) const
{
  return ON_GrevilleAbcissa( m_order, m_knot+gindex );
}

BOOL ON_NurbsCurve::GetGrevilleAbcissae( // see ON_GetGrevilleAbcissa() for details
         double* g         // g[m_cv_count]
         ) const
{
  return ON_GetGrevilleAbcissae( m_order, m_cv_count, m_knot, FALSE, g );
}


BOOL ON_NurbsCurve::ZeroCVs()
{
  BOOL rc = FALSE;
  int i;
  if ( m_cv ) {
    if ( m_cv_capacity > 0 ) {
      memset( m_cv, 0, m_cv_capacity*sizeof(*m_cv) );
      if ( m_is_rat ) {
        for ( i = 0; i < m_cv_count; i++ ) {
          SetWeight( i, 1.0 );
        }
      }
      rc = TRUE;
    }
    else {
      double* cv;
      int s = CVSize()*sizeof(*cv);
      for ( i = 0; i < m_cv_count; i++ ) {
        cv = CV(i);
        memset(cv,0,s);
        if ( m_is_rat )
          cv[m_dim] = 1.0;
      }
      rc = (i>0) ? TRUE : FALSE;
    }
  }
  return rc;
}

BOOL ON_NurbsCurve::IsClamped( // determine if knot vector is clamped
      int end // (default= 2) end to check: 0 = start, 1 = end, 2 = start and end
      ) const
{
  return ON_IsKnotVectorClamped( m_order, m_cv_count, m_knot, end );
}
  

BOOL ON_NurbsCurve::ReserveCVCapacity(int desired_capacity)
{
  // If m_cv_capacity == 0 and m_cv != NULL, then the user
  // has hand built the ON_NurbsCurve.m_cv array and is responsible
  // for making sure it's always big enough.
  BOOL rc = TRUE;
  if ( desired_capacity > m_cv_capacity ) {
    if ( !m_cv ) {
      // no cv array - allocate one
      m_cv = (double*)onmalloc(desired_capacity*sizeof(*m_cv));
      if ( !m_cv ) {
        rc = FALSE;
      }
      else {
        m_cv_capacity = desired_capacity;
      }
    }
    else if ( m_cv_capacity > 0 ) {
      // existing m_cv[] is too small and the fact that
      // m_cv_capacity > 0 indicates that ON_NurbsCurve is
      // managing the m_cv[] memory, so we need to grow
      // the m_cv[] array.
      m_cv = (double*)onrealloc(m_cv,desired_capacity*sizeof(*m_cv));
      if ( !m_cv ) {
        rc = FALSE;
        m_cv_capacity = 0;
      }
      else {
        m_cv_capacity = desired_capacity;
      }
    }
  }
  return rc;
}

BOOL ON_NurbsCurve::ReserveKnotCapacity(int desired_capacity)
{
  // If m_knot_capacity == 0 and m_knot != NULL, then the user
  // has hand built the ON_NurbsCurve.m_knot array and is responsible
  // for making sure it's always big enough.
  BOOL rc = TRUE;
  if ( desired_capacity > m_knot_capacity ) {
    if ( !m_knot ) {
      // no knot array - allocate one
      m_knot = (double*)onmalloc(desired_capacity*sizeof(*m_knot));
      if ( !m_knot ) {
        m_knot_capacity = 0;
        rc = FALSE;
      }
      else {
        m_knot_capacity = desired_capacity;
      }
    }
    else if ( m_knot_capacity > 0 ) {
      // existing m_knot[] is too small and the fact that
      // m_knot_capacity > 0 indicates that ON_NurbsCurve is
      // managing the m_knot[] memory, so we need to grow
      // the m_knot[] array.
      m_knot = (double*)onrealloc(m_knot,desired_capacity*sizeof(*m_knot));
      if ( !m_knot ) {
        rc = FALSE;
        m_knot_capacity = 0;
      }
      else {
        m_knot_capacity = desired_capacity;
      }
    }
  }
  return rc;
}

BOOL ON_NurbsCurve::ConvertSpanToBezier( int span_index, ON_BezierCurve& bez ) const
{
  BOOL rc = FALSE;
  if ( span_index >= 0 && span_index <= m_cv_count-m_order && m_knot && m_cv ) {
    const int cvdim = CVSize();
    const int sizeof_cv = cvdim*sizeof(*bez.m_cv);
    int i;
    rc = bez.ReserveCVCapacity( cvdim*m_order );
    if ( rc ) {
      bez.m_dim = m_dim;
      bez.m_is_rat = m_is_rat;
      bez.m_order = m_order;
      bez.m_cv_stride = cvdim;
      for ( i = 0; i < m_order; i++ ) {
        memcpy( bez.CV(i), CV(span_index+i), sizeof_cv );
      }
      const double* knot = m_knot + span_index;
      ON_ConvertNurbSpanToBezier( cvdim, bez.m_order, bez.m_cv_stride, bez.m_cv,
                                  knot, knot[m_order-2], knot[m_order-1] );
    }
  }
  return rc;
}


BOOL ON_NurbsCurve::HasBezierSpans() const
{
  return ON_KnotVectorHasBezierSpans( m_order, m_cv_count, m_knot );
}

BOOL ON_NurbsCurve::MakePiecewiseBezier()
{
  BOOL rc = HasBezierSpans();
  if ( !rc && IsValid() ) {
    ON_Workspace ws;
    if ( !ClampEnd(2) )
      return FALSE;
    int span_count = SpanCount();
    //int knot_count = KnotCount();
    ReserveKnotCapacity( (span_count + 1)*(m_order-1) );
    ReserveCVCapacity( m_cv_stride*(span_count*(m_order-1) + 1) );
    double* t = ws.GetDoubleMemory( span_count+1);
    GetSpanVector( t );
    int cvdim = CVSize();
    ON_BezierCurve* bez = new ON_BezierCurve[span_count];
    int ki, spani, i;
    for ( ki = m_order-2, spani = 0; ki < m_cv_count-1 && spani < span_count; ki++ ) {
      if ( m_knot[ki] < m_knot[ki] ) {
        bez[spani].Create(m_dim,m_is_rat,m_order);
        for ( i = 0; i < m_order; i++ )
          bez[spani].SetCV(  i, ON::intrinsic_point_style, CV( i + ki - m_order + 2 ) );
        ON_ConvertNurbSpanToBezier( cvdim, bez[spani].m_order, bez[spani].m_cv_stride, bez[spani].m_cv,
                                    m_knot+ki-m_order+2, m_knot[ki], m_knot[ki+1] );
        spani++;
      }
    }
    m_cv_count = span_count*(m_order-1) + 1;
    for ( spani = 0; spani < span_count; spani++ ) {
      for ( i = 0; i < m_order; i++ ) {
        SetCV(  spani*(m_order-1) + i, ON::intrinsic_point_style, bez[spani].CV( i ) );
      }
      for ( ki = 0; ki < m_order-1; ki++ )
        m_knot[ki+spani*(m_order-1)] = t[spani];
    }
    for ( ki = 0; ki < m_order-1; ki++ )
      m_knot[ki+span_count*(m_order-1)] = t[spani];
    delete[] bez;
    rc = TRUE;
  }
  return rc;
}


double ON_NurbsCurve::ControlPolygonLength() const
{
  double length = 0.0;
  ON_GetPolylineLength( m_dim, m_is_rat, m_cv_count, m_cv_stride, m_cv, &length );
  return length;
}


BOOL ON_NurbsCurve::InsertKnot( double knot_value, int knot_multiplicity )
{
  BOOL rc = FALSE;

  if ( knot_multiplicity < 1 || knot_multiplicity > Degree() ) {
    ON_ERROR("ON_NurbsCurve::ON_InsertKnot() - illegal knot_multiplicity");
    return FALSE;
  }

  const int degree = Degree();

  double t0, t1;
  {
    ON_Interval d = Domain();
    if ( !d.IsIncreasing() )
      return FALSE;
    t0 = d[0];
    t1 = d[1];
  }
  if ( knot_value == t0 ) {
    if ( knot_multiplicity == degree ) {
      rc = ClampEnd(0);
    }
    else if ( knot_multiplicity == 1 ) {
      rc = TRUE;
    }
    else {
      ON_ERROR("ON_InsertKnot(): cannot insert knot");
      rc = FALSE;
    }
    return rc;
  }
  if ( knot_value == t1 ) {
    if ( knot_multiplicity == degree ) {
      rc = ClampEnd(1);
    }
    else if ( knot_multiplicity == 1 ) {
      rc = TRUE;
    }
    else {
      ON_ERROR("ON_InsertKnot(): cannot insert knot");
      rc = FALSE;
    }
    return rc;
  }

  BOOL bIsPeriodic = (degree>1) ? IsPeriodic() : FALSE;
  int span_index = ON_SpanIndex( m_order, m_cv_count, m_knot, knot_value, 0, 0 );

  // reserve room for new knots and cvs
  if ( !ReserveCVCapacity( m_cv_stride*(m_cv_count+knot_multiplicity) ) )
    return FALSE;
  if ( !ReserveKnotCapacity( KnotCount()+knot_multiplicity ) )
    return FALSE;
  if ( bIsPeriodic ) {
  }

  rc = TRUE;
  int span_hint = span_index;
  int new_knot_count = ON_InsertKnot( knot_value, knot_multiplicity, 
                                      CVSize(), m_order, m_cv_count, 
                                      m_cv_stride, m_cv, m_knot, &span_hint );
  if ( new_knot_count > 0 ) {
    m_cv_count += new_knot_count;
  }

  if ( bIsPeriodic && rc>0 && !IsPeriodic() ) {
    // restore periodic form
    if ( ON_MakeKnotVectorPeriodic( m_order, m_cv_count, m_knot ) ) {
      int i0, i1;
      for ( i0 = 0, i1 = m_cv_count-degree; i0 < degree; i0++, i1++ ) {
        if ( span_index < degree-1 )
          SetCV( i1, ON::intrinsic_point_style, CV(i0) ); // cv[i1] = cv[i0]
        else
          SetCV( i0, ON::intrinsic_point_style, CV(i1) ); // cv[i0] = cv[i1]
      }
    }
    else {
      ClampEnd(2);
    }
  }

  return rc;
}

BOOL ON_NurbsCurve::MakeRational()
{
  if ( !IsRational() ) {
    const int dim = Dimension();
    const int cv_count = CVCount();
    if ( cv_count > 0 && m_cv_stride >= dim && dim > 0 ) {
      const int new_stride = (m_cv_stride == dim) ? dim+1 : m_cv_stride;
      ReserveCVCapacity( cv_count*new_stride );
      const double* old_cv;
      double* new_cv;
      int cvi, j;
      for ( cvi = cv_count-1; cvi>=0; cvi-- ) {
        old_cv = CV(cvi);
        new_cv = m_cv+(cvi*new_stride);
        for ( j = dim-1; j >= 0; j-- ) {
          new_cv[j] = old_cv[j];
        }
        new_cv[dim] = 1.0;
      }
      m_cv_stride = new_stride;
      m_is_rat = 1;
    }
  }
  return IsRational();
}

BOOL ON_NurbsCurve::MakeNonRational()
{
  if ( IsRational() ) {
    const int dim = Dimension();
    const int cv_count = CVCount();
    if ( cv_count > 0 && m_cv_stride >= dim+1 && dim > 0 ) {
      double w;
      const double* old_cv;
      double* new_cv = m_cv;
      int cvi, j;
      for ( cvi = 0; cvi < cv_count; cvi++ ) {
        old_cv = CV(cvi);
        w = old_cv[dim];
        w = ( w != 0.0 ) ? 1.0/w : 1.0;
        for ( j = 0; j < dim; j++ ) {
          *new_cv++ = w*(*old_cv++);
        }
      }
      m_is_rat = 0;
      m_cv_stride = dim;
    }
  }
  return ( !IsRational() ) ? TRUE : FALSE;
}

static BOOL GetRaisedDegreeCV(int old_order, 
                              int cvdim,
                              int old_cvstride,
                              const double* oldCV, //old_cvstride*old_order
                              const double* oldkn, //2*old_degree
                              const double* newkn, //2*old_order
                              int cv_id,     //0 <= cv_id <= old_order
                              double* newCV  //cvdim
                              )

{
  int i,j,k;

  if (!oldCV || !oldkn || !newkn || !newCV || cv_id < 0 || cv_id > old_order)
    return FALSE;

  int old_degree = old_order-1;
  int new_degree = old_degree+1;

  double* t = (double*)onmalloc(old_degree*sizeof(double));
  if (!t) return FALSE;
  double* P = (double*)onmalloc(cvdim*sizeof(double));
  if (!P) {
    onfree((void*)t);
    return FALSE;
  }

  memset(newCV, 0, cvdim*sizeof(double));

  const double* kn = newkn + cv_id;

  for (i=0; i<new_degree; i++){
    k=0;
    for (j=0; j<new_degree; j++){
      if (j != i) {
        t[k] = kn[j];
        k++;
      }
    }
    if (!ON_EvaluateNurbsBlossom(cvdim, old_order, 
      old_cvstride, oldCV, oldkn, t, P)){
      onfree((void*)t);
      onfree((void*)P);
      return FALSE;
    }
    for (k=0; k<cvdim; k++) newCV[k] += P[k];
  }

  double denom = (double)new_degree;
  for (i=0; i<cvdim; i++) 
    newCV[i] /= denom;

  onfree((void*)t);
  onfree((void*)P);

  return TRUE;
}


static BOOL IncrementNurbDegree(ON_NurbsCurve& N)
//for use only in ON_NurbsCurve::IncreaseDegree().  No validation done.
//N is assumed to have sufficient knot and cv capacities, clamped ends

{
  ON_NurbsCurve M = N;
  int span_count = M.SpanCount();
  int new_kcount = M.KnotCount() + span_count + 1;
  N.m_order = M.Order()+1;
  N.m_cv_count = new_kcount - N.Order() + 2;

  //set N's knots
  int i=0;
  int k=0;
  int j;
  while (i<M.CVCount()){
    double kn = M.Knot(i);
    int mult = M.KnotMultiplicity(i);
    for (j=0; j<=mult; j++) {
      N.SetKnot(k, kn);
      k++;
    }
    i+=mult;
  }
  //zero out N's cvs
  memset(N.m_cv, 0, N.m_cv_capacity*sizeof(double));
  int cvdim = N.CVSize();
  const double* cvM;
  double* cvN;
  const double* knotN;
  const double* knotM;
  int siN = 0;
  int siM = 0;

  for (i=0; i<span_count; i++){
    knotN = &N.m_knot[siN];
    knotM = &M.m_knot[siM];
    cvM = M.CV(siM);
    cvN = N.CV(siN);
    int span_mult = N.KnotMultiplicity(siN+N.Degree()-1);
    int skip = N.Order()-span_mult;
    cvN += skip*N.m_cv_stride;
    for (j=skip; j<N.Order(); j++){//calculate each cv of the span
      GetRaisedDegreeCV(M.Order(), cvdim, M.m_cv_stride, cvM, knotM, knotN, j, cvN);
      cvN += N.m_cv_stride;
    }
    siN = ON_NextSpanIndex(N.Order(), N.CVCount(), N.m_knot, siN);
    siM = ON_NextSpanIndex(M.Order(), M.CVCount(), M.m_knot, siM);
  }

  //set first and last equal to original
  cvM = M.CV(0);
  cvN = N.CV(0);
  for (i=0; i<cvdim; i++) cvN[i] = cvM[i];

  cvM = M.CV(M.CVCount()-1);
  cvN = N.CV(N.CVCount()-1);
  for (i=0; i<cvdim; i++) cvN[i] = cvM[i];

  return TRUE;
}

BOOL ON_NurbsCurve::IncreaseDegree( int desired_degree )
{
  if ( desired_degree < 1 || desired_degree < m_order-1 ) return FALSE;
  if ( desired_degree == m_order-1 ) return TRUE;
  if (!ClampEnd(2)) return FALSE;

  int del = desired_degree - Degree();
  int new_order = Order()+del;
  int span_count = SpanCount();
  int new_kcount = KnotCount()+(span_count+1)*del;
  int new_cvcount = new_kcount - new_order + 2;

  if (!ReserveKnotCapacity(new_kcount)) return FALSE;
  if (!ReserveCVCapacity(new_cvcount*m_cv_stride)) return FALSE;

  for (int i=0; i<del; i++) {
    if (!IncrementNurbDegree(*this)) return FALSE;
  }

  return TRUE;
}

BOOL ON_NurbsCurve::ChangeDimension( int desired_dimension )
{
  BOOL rc = FALSE;
  int i, j;
  if ( desired_dimension < 1 )
    return FALSE;
  if ( desired_dimension == m_dim )
    return TRUE;
  if ( desired_dimension < m_dim ) {
    if ( m_is_rat ) {
      double* cv;
      for ( i = 0; i < m_cv_count; i++ ) {
        cv = CV(i);
        cv[desired_dimension] = cv[m_dim];
      }
    }
    m_dim = desired_dimension;
  }
  else {
    const double* cv0;
    double* cv1;
    //const int cv_size0 = CVSize();
    const int cv_size1 = m_is_rat ? m_dim + 1 : m_dim;
    const int cv_stride1 = (m_cv_stride < cv_size1) ? cv_size1 : m_cv_stride;
    if ( m_cv_stride < cv_stride1 )
      m_cv = (double*)onrealloc( m_cv, cv_stride1*m_cv_count*sizeof(*m_cv) );
    for ( i = CVCount()-1; i >= 0; i-- ) {
      cv0 = CV(i);
      cv1 = m_cv + (i*cv_stride1);
      if ( m_is_rat )
        cv1[desired_dimension] = cv0[m_dim];
      for ( j = desired_dimension-1; j >= m_dim; j-- ) {
        cv1[j] = 0.0;
      }
      for ( j = m_dim-1; j >= 0; j-- ) {
        cv1[j] = cv0[j];
      }
    }
    rc = TRUE;
  }
  return rc;
}

BOOL ON_NurbsCurve::Append( const ON_NurbsCurve& c )
{
  BOOL rc = FALSE;

  if ( CVCount() == 0 ) {
    *this = c;
    return IsValid();
  }

  if ( c.IsRational() && !IsRational() ) {
    if ( !MakeRational() )
      return FALSE;
  }
  if ( c.Degree() > Degree() ) {
    if ( !IncreaseDegree( c.Degree() ) )
      return FALSE;
  }
  if ( c.Dimension() > Dimension() ) {
    if ( !ChangeDimension( c.Dimension() ) )
      return FALSE;
  }

  if ( IsRational() && !c.IsRational() || c.Degree() < Degree() || !c.IsClamped(0) || c.Dimension() < Dimension() ) {
    ON_NurbsCurve tmp(c);
    if ( !tmp.IncreaseDegree( Degree() ) )
      return FALSE;
    if ( !tmp.ChangeDimension( Dimension() ) )
      return FALSE;
    if ( IsRational() ) {
      if ( !tmp.MakeRational() )
        return FALSE;
    }
    if ( !tmp.ClampEnd(0) )
      return FALSE;
    rc = Append(tmp);
  }

  if ( IsValid() && c.IsValid() && Degree() == c.Degree() && IsRational() == c.IsRational() && Dimension() == c.Dimension() ) {
    if ( !ClampEnd(1) )
      return FALSE;
    const double w0 = c.Weight(0);
    const double w1 = Weight(CVCount()-1);
    double w = 1.0;
    if ( IsRational() && w0 != w1 ) {
      w = w1/w0;
    }
    ReserveCVCapacity( (CVCount()+c.CVCount())*m_cv_stride );
    ReserveKnotCapacity( ON_KnotCount(Order(),CVCount()+c.CVCount()) );
    const double dk = Knot(CVCount()-1) - c.Knot(c.Order()-2);

    double* cv = CV(CVCount()-1);
    const int cv_dim = CVSize();
    const int sizeof_cv = cv_dim*sizeof(*cv);
    const int c_cv_count = c.CVCount();
    const int c_knot_count = c.KnotCount();
    int i0, i1, j;

    i0 = KnotCount();
    for ( i1 = c.Order()-1; i1 < c_knot_count; i1++ ) {
      m_knot[i0++] = c.Knot(i1) + dk;
    }

    for ( i1 = 1; i1 < c_cv_count; i1++ ) {
      cv += m_cv_stride;
      memcpy( cv, c.CV(i1), sizeof_cv );
      if ( w != 1.0 ) {
        for ( j = 0; j < cv_dim; j++ )
          cv[j] *= w;
      }
      m_cv_count++;
    }
    rc = TRUE;
  }
  return rc;
}



BOOL ON_NurbsCurve::Trim( const ON_Interval& in )
{
  if ( !in.IsIncreasing() )
    return FALSE;

  const int cv_dim = CVSize();
  const int order = Order();
  double t;
  int ki, side, i0, i1, i1_max, new_cv_count;

  // cut off right end (or extend if in.m_t[1] > Domain.Max()
  side = -1;
  t = in.m_t[1]; // trimming parameter
  ki = ON_SpanIndex( order, m_cv_count, m_knot, t, side, 0 );
  if ( !ON_EvaluateNurbsDeBoor( cv_dim, order, m_cv_stride, CV(ki),
                                m_knot + ki, side, 0.0, t ) ) {
    ON_ERROR("ON_NurbsCurve::Trim() - right end de Boor algorithm failed.");
    return FALSE;
  }
  // clamp right end knots
  m_cv_count = ki + order;
  for ( i0 = ON_KnotCount( order, m_cv_count)-1; i0 >= m_cv_count-1; i0--)
    m_knot[i0] = t;

  // cut off left end (or extend if in.m_t[0] < Domain.Max()
  side = 1;
  t = in.m_t[0]; // trimming parameter
  ki = ON_SpanIndex( order, m_cv_count, m_knot, t, side, 0 );
  if ( !ON_EvaluateNurbsDeBoor( cv_dim, order, m_cv_stride, CV(ki),
                                m_knot + ki, side, 0.0, t ) ) {
    ON_ERROR("ON_NurbsCurve::Trim() - right end de Boor algorithm failed.");
    return FALSE;
  }

  new_cv_count = m_cv_count - ki;
  if ( new_cv_count < m_cv_count ) {
    // move cvs and knots over
    i1_max = m_cv_stride*m_cv_count;
    for ( i0 = 0, i1 = ki*m_cv_stride; i1 < i1_max; i0++, i1++ )
      m_cv[i0] = m_cv[i1];
    i1_max = ON_KnotCount( order, m_cv_count );
    for ( i0 = 0, i1 = ki; i1 < i1_max; i0++, i1++ )
      m_knot[i0] = m_knot[i1];
    m_cv_count = new_cv_count;
  }
  // clamp left end knots
  for (i0 = 0; i0 <= order-2; i0++)
    m_knot[i0] = t;

  return TRUE;
}

BOOL ON_NurbsCurve::Split(
    double t,    // t = curve parameter to split curve at
    ON_Curve*& left_crv, // left portion returned here (must be an ON_NurbsCurve)
    ON_Curve*& right_crv // right portion returned here (must be an ON_NurbsCurve)
  ) const
{
  int i;
  BOOL rc = FALSE;
  if ( left_crv && !ON_NurbsCurve::Cast(left_crv) )
    return FALSE;
  if ( right_crv && !ON_NurbsCurve::Cast(right_crv) )
    return FALSE;
  if ( IsValid() && t > m_knot[m_order-1] && t < m_knot[m_cv_count-1] ) {
    ON_NurbsCurve* left = (ON_NurbsCurve*)left_crv;
    ON_NurbsCurve* right = (ON_NurbsCurve*)right_crv;
    if ( !left )
      left = new ON_NurbsCurve();
    if ( !right )
      right = new ON_NurbsCurve();
    const int span_index = ON_SpanIndex(m_order,m_cv_count,m_knot,t,1,0);
    if ( span_index >= 0 && span_index <= m_cv_count - m_order ) {
      const int cvdim = CVSize();
      const int sizeof_cv = cvdim*sizeof(double);

      // fill in left allowing for possibility that left = this
      if ( left->m_cv != m_cv ) {
        left->ReserveCVCapacity(cvdim*(m_order + span_index));
        for( i = 0; i < m_order + span_index; i++ ) {
          memcpy( left->CV(i), CV(i), sizeof_cv );
        }
      }
      if ( left->m_knot != m_knot ) {
        i = ON_KnotCount( m_order, m_order + span_index);
        left->ReserveKnotCapacity( i );
        memcpy( left->m_knot, m_knot, i*sizeof(left->m_knot[0]) );
      }
      
      // fill in right allowing for possibility that right = this
      if ( right->m_cv != m_cv || span_index > 0 )
        right->ReserveCVCapacity(cvdim*(m_cv_count - span_index));
      for( i = 0; i < m_cv_count - span_index; i++ ) {
        memmove( right->m_cv + i*cvdim, CV(i+span_index), sizeof_cv );
      }
      if ( right->m_knot != m_knot || span_index > 0 ) {
        i = ON_KnotCount(m_order, m_cv_count - span_index);
        right->ReserveKnotCapacity( i );
        memmove( right->m_knot, m_knot + span_index, i*sizeof(right->m_knot[0]) );
      }

      // set right (and possible this cv_count, etc.)
      right->m_dim = m_dim;
      right->m_is_rat = m_is_rat;
      right->m_order = m_order;
      right->m_cv_count = m_cv_count - span_index; // must happen before left->m_cv_count is set
      right->m_cv_stride = cvdim;

      // set left (and possible this cv_count, etc.)
      left->m_dim = m_dim;
      left->m_is_rat = m_is_rat;
      left->m_order = m_order;
      left->m_cv_count = m_order + span_index;
      left->m_cv_stride = cvdim;

      // trim right end of left NURBS
      i = left->m_cv_count - left->m_order;
      ON_EvaluateNurbsDeBoor( cvdim, m_order, cvdim, left->CV(i), left->m_knot + i, -1, 0.0, t );
      for ( i = left->m_cv_count-1; i < ON_KnotCount(left->m_order,left->m_cv_count); i++ )
        left->m_knot[i] = t;

      // trim left end of right NURBS
      ON_EvaluateNurbsDeBoor( cvdim, m_order, cvdim, right->m_cv, right->m_knot, +1, 0.0, t );
      for ( i = 0; i <= right->m_order-2; i++ )
        right->m_knot[i] = t;
    }
  }
  return rc;
}


int ON_NurbsCurve::GetNurbForm( ON_NurbsCurve& curve, 
                                double tolerance,
                                const ON_Interval* subdomain  // OPTIONAL subdomain of ON::ProxyCurve::Domain()
                                ) const
{
  int rc = 1;
  curve = *this;
  if ( subdomain ) {
    if ( !curve.Trim(*subdomain) )
      rc = 0;
  }
  return rc;
}

BOOL ON_NurbsCurve::ClampEnd( 
        int end // 0 = clamp start, 1 = clamp end, 2 = clamp start and end 
        )
{
  return ON_ClampKnotVector( CVSize(), m_order, 
                             m_cv_count, m_cv_stride, m_cv, 
                             m_knot, end );
}

