/* $Header: /src2/opennurbs/opennurbs_nurbssurface.h 11    4/19/01 1:26p Dalelear $ */
/* $NoKeywords: $ */
/*
//
// Copyright (c) 1993-2001 Robert McNeel & Associates. All rights reserved.
// Rhinoceros is a registered trademark of Robert McNeel & Assoicates.
//
// THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
// ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
// MERCHANTABILITY ARE HEREBY DISCLAIMED.
//				
// For complete openNURBS copyright information see <http://www.opennurbs.org>.
//
////////////////////////////////////////////////////////////////
*/

////////////////////////////////////////////////////////////////
//
//   Definition of NURBS surface
//
////////////////////////////////////////////////////////////////

#if !defined(OPENNURBS_NURBSSURFACE_INC_)
#define OPENNURBS_NURBSSURFACE_INC_

class ON_CLASS ON_TensorProduct
{
  // Pure virtual tensor passed to ON_NurbsSurface::TensorProduct()
public:

  // Evaluate() must define a function T:R^dimA X R^dimB -> R^Dimension()
  // such that
  //    T(a*A0 + (1-a)*A1, B) = a*T(A0,B) + (1-a)*T(A1,B) and
  //    T(A, b*B0 + (1-b)*B1) = b*T(A,B0) + (1-b)*T(A,B1).
  virtual
  int DimensionA() const = 0; // dimension of A space
  virtual
  int DimensionB() const = 0; // dimension of B space
  virtual
  int DimensionC() const = 0; // dimension of range space
  virtual
  BOOL Evaluate( double,        // a
                 const double*, // A
                 double,        // b
                 const double*, // B
                 double*        // C
                ) = 0;

};

class ON_Brep;
class ON_NurbsSurface;
class ON_CLASS ON_NurbsSurface : public ON_Surface
{
  ON_OBJECT_DECLARE(ON_NurbsSurface);

public:
  ON_NurbsSurface();
  ON_NurbsSurface(const ON_NurbsSurface&);
  ON_NurbsSurface(
          int,  // dimension (>= 1)
          BOOL, // TRUE to make a rational NURBS
          int,  // order0 (>= 2)
          int,  // order1 (>= 2)
          int,  // cv count0 (>= order0)
          int   // cv count1 (>= order1)
          );

  void Initialize(void);  // zeros all fields

  BOOL Create( 
          int,  // dimension (>= 1)
          BOOL, // TRUE to make a rational NURBS
          int,  // order0 (>= 2)
          int,  // order1 (>= 2)
          int,  // cv count0 (>= order0)
          int   // cv count1 (>= order1)
          );

  void Destroy();

  virtual ~ON_NurbsSurface();

  void EmergencyDestroy(); // call if memory used by this class becomes invalid

  ON_NurbsSurface& operator=(const ON_NurbsSurface&);
  
  /////////////////////////////////////////////////////////////////
  // ON_Object overrides

  BOOL IsValid() const;

  void Dump( ON_TextLog& ) const; // for debugging

  BOOL Write(
         ON_BinaryArchive&  // open binary file
       ) const;

  BOOL Read(
         ON_BinaryArchive&  // open binary file
       );

  /////////////////////////////////////////////////////////////////
  // ON_Geometry overrides

  int Dimension() const;

  BOOL GetBBox( // returns TRUE if successful
         double*,    // minimum
         double*,    // maximum
         BOOL = FALSE  // TRUE means grow box
         ) const;

  BOOL Transform( 
         const ON_Xform&
         );

  BOOL SwapCoordinates(
        int, int        // indices of coords to swap
        );

  //////////
  // overrides ON_Geometry::HasBrepForm() and returns TRUE
  BOOL HasBrepForm() const;

  //////////
  // overrides ON_Geometry::BrepForm() and simply returns a 
  // single face trivially trimmed brep.
  ON_Brep* BrepForm() const;

  /////////////////////////////////////////////////////////////////
  // ON_Surface overrides
  ON_Interval Domain(
    int // 0 gets first parameter's domain, 1 gets second parameter's domain
    ) const;


  int SpanCount(
    int // 0 gets first parameter's domain, 1 gets second parameter's domain
    ) const; // number of smooth spans in curve

  BOOL GetSpanVector( // span "knots" 
    int, // 0 gets first parameter's domain, 1 gets second parameter's domain
    double* // array of length SpanCount() + 1 
    ) const; // 

  int Degree( // returns maximum algebraic degree of any span 
                  // ( or a good estimate if curve spans are not algebraic )
    int // 0 gets first parameter's domain, 1 gets second parameter's domain
    ) const; 

  BOOL GetParameterTolerance( // returns tminus < tplus: paramters tminus <= s <= tplus
         int,     // 0 gets first parameter, 1 gets second parameter
         double,  // t = parameter in domain
         double*, // tminus
         double*  // tplus
         ) const;

  BOOL IsClosed(   // TRUE if NURBS surface is closed (either surface has
        int // dir // clamped end knots and euclidean location of start
        ) const;   // CV = euclidean location of end CV, or surface is
                   // periodic.)

  BOOL IsPeriodic( // TRUE if NURBS surface is periodic (degree > 1,
        int // dir // periodic knot vector, last degree many CVs 
        ) const;   // are duplicates of first degree many CVs.)
  
  BOOL IsSingular( // TRUE if surface side is collapsed to a point
        int        // side of parameter space to test
                   // 0 = south, 1 = east, 2 = north, 3 = west
        ) const;

  BOOL Reverse(  // reverse parameterizatrion, Domain changes from [a,b] to [-b,-a]
    int // dir  0 = "s", 1 = "t"
    );

  BOOL Transpose(); // transpose surface parameterization (swap "s" and "t")

  BOOL Evaluate( // returns FALSE if unable to evaluate
         double, double, // evaluation parameter
         int,            // number of derivatives (>=0)
         int,            // array stride (>=Dimension())
         double*,        // array of length stride*(ndir+1)*(ndir+2)/2
         int = 0,        // optional - determines which quadrant to evaluate from
                         //         0 = default
                         //         1 from NE quadrant
                         //         2 from NW quadrant
                         //         3 from SW quadrant
                         //         4 from SE quadrant
         int* = 0        // optional - evaluation hint (int[2]) used to speed
                         //            interative evaluations
         ) const;

  ON_Curve* IsoCurve(
         int,              // 0 first parameter varies and second parameter is constant
                           //   e.g., point on IsoCurve(0,c) at t is srf(t,c)
                           // 1 first parameter is constant and second parameter varies
                           //   e.g., point on IsoCurve(1,c) at t is srf(c,t)
         double            // value of constant parameter 
         ) const;

  int GetNurbForm( // returns 0: unable to create NURBS representation
                   //            with desired accuracy.
                   //         1: success - returned NURBS parameterization
                   //            matches the surface's to wthe desired accuracy
                   //         2: success - returned NURBS point locus matches
                   //            the surfaces's to the desired accuracy but, on
                   //            the interior of the surface's domain, the 
                   //            surface's parameterization and the NURBS
                   //            parameterization may not match to the 
                   //            desired accuracy.
        ON_NurbsSurface&,
        double = 0.0 // tolerance
        ) const;

  /////////////////////////////////////////////////////////////////
  // Interface

  BOOL IsRational(  // TRUE if NURBS surface is rational
        void
        ) const;
  
  int CVSize(       // number of doubles per control vertex 
        void        // = IsRational() ? Dim()+1 : Dim()
        ) const;
  
  int Order(        // order = degree + 1
        int         // dir 0 = "s", 1 = "t"
        ) const;
	
  int CVCount(      // number of control vertices
        int         // dir 0 = "s", 1 = "t"
        ) const;

  int CVCount(      // total number of control vertices
        void
        ) const;

  int KnotCount(    // total number of knots in knot vector
        int         // dir 0 = "s", 1 = "t"
        ) const;
  
  double* CV(      // get pointer to control vertex
        int, int   // CV index ( 0 <= i <= CVCount(0), 0 <= j <= CVCount(1)
        );

  const double* CV(     // get pointer to control vertex
        int, int   // CV index ( 0 <= i <= CVCount(0), 0 <= j <= CVCount(1)
        ) const;

  double Weight(        // get value of control vertex weight
        int, int   // CV index ( 0 <= i <= CVCount(0), 0 <= j <= CVCount(1)
        ) const;

  BOOL SetWeight(      // get value of control vertex weight
        int, int,   // CV index ( 0 <= i <= CVCount(0), 0 <= j <= CVCount(1)
        double
        );

  BOOL SetCV(              // set a single control vertex
        int, int,   // CV index ( 0 <= i <= CVCount(0), 0 <= j <= CVCount(1)
        ON::point_style, // style of input point
        const double*     // value of control vertex
        );

  BOOL SetCV(               // set a single control vertex
        int, int,   // CV index ( 0 <= i <= CVCount(0), 0 <= j <= CVCount(1)
        const ON_3dPoint& // value of control vertex
                           // If NURBS is rational, weight
                           // will be set to 1.
        );

  BOOL SetCV(              // set a single control vertex
        int, int,   // CV index ( 0 <= i <= CVCount(0), 0 <= j <= CVCount(1)
        const ON_4dPoint& // value of control vertex
        );

  BOOL SetCVRow(          // Sets CV( *, row_index ) 
       int,               // row_index >= 0 and < m_cv_count[1]
       const ON_3dPoint& // value of control vertex
                          // If NURBS is rational, weight
                          // will be set to 1.
       );

  BOOL SetCVRow(          // Sets CV( *, row_index ) 
       int,               // row_index >= 0 and < m_cv_count[1]
       int,               // v stride
       const double*      // v[] = values (same dim and is_rat as surface)
       );

  BOOL SetCVColumn(       // Sets CV( col_index, * ) 
       int,               // col_index >= 0 and < m_cv_count[0]
       const ON_3dPoint& // value of control vertex
                          // If NURBS is rational, weight
                          // will be set to 1.
       );

  BOOL SetCVColumn(       // Sets CV( col_index, * ) 
       int,               // col_index >= 0 and < m_cv_count[0]
       int,               // v stride
       const double*      // v[] = values (same dim and is_rat as surface)
       );

  BOOL GetCV(              // get a single control vertex
        int, int,   // CV index ( 0 <= i <= CVCount(0), 0 <= j <= CVCount(1)
        ON::point_style, // style to use for output point
        double*           // array of length >= CVSize()
        ) const;

  BOOL GetCV(              // get a single control vertex
        int, int,   // CV index ( 0 <= i <= CVCount(0), 0 <= j <= CVCount(1)
        ON_3dPoint&      // gets euclidean cv when NURBS is rational
        ) const;

  BOOL GetCV(              // get a single control vertex
        int, int,   // CV index ( 0 <= i <= CVCount(0), 0 <= j <= CVCount(1)
        ON_4dPoint&      // gets homogeneous cv
        ) const;

  int SetKnot(
        int,    // dir 0 = "s", 1 = "t"
        int,            // knot index ( 0 to KnotCount - 1 )
        double          // value for knot
        );

  double Knot(
        int,    // dir 0 = "s", 1 = "t"
        int             // knot index ( >= 0 and < Order + CV_count - 2 )
        ) const;

  int KnotMultiplicity(
        int,    // dir 0 = "s", 1 = "t"
        int             // knot index ( >= 0 and < Order + CV_count - 2 )
        ) const;

  const double* Knot(   // knot[] array
        int     // dir 0 = "s", 1 = "t"
        ) const;

  BOOL IsClamped( // determine if knot vector is clamped
        int,    // dir 0 = "s", 1 = "t"
        int = 2 // end to check: 0 = start, 1 = end, 2 = start and end
        ) const;
  
  double SuperfluousKnot(
           int,    // dir 0 = "s", 1 = "t"
           int // 0 = start, 1 = end
           ) const;

  double GrevilleAbcissa(
           int,  // dir
           int   // index (0 <= index < CVCount(dir)
           ) const;

  BOOL GetGrevilleAbcissae( // see ON_GetGrevilleAbcissa() for details
           int,      // dir
           double*   // g[cv count]
           ) const;

  BOOL SetClampedGrevilleKnotVector(
           int,          // dir
           int,          // g_stride
           const double* // g[], CVCount(dir) many Greville abcissa
           );

  BOOL SetPeriodicGrevilleKnotVector(
           int,          // dir
           int,          // g_stride
           const double* // g[], Greville abcissa
           );

  BOOL ZeroCVs(); // zeros all CVs (any weights set to 1);

  BOOL ClampEnd(
            int,         // dir 0 = "s", 1 = "t"
            int // 0 = clamp start, 1 = clamp end, 2 = clamp start and end
            );

  BOOL InsertKnot(
           int,         // dir 0 = "s", 1 = "t"
           double, // value of knot
           int=1   // multiplicity of knot ( >= 1 and <= degree )
           );

  BOOL MakeRational();

  BOOL MakeNonRational();

  BOOL IncreaseDegree(
           int,  // dir 0 = "s", 1 = "t"
           int   //  desired_degree
           );

  BOOL ChangeDimension(
           int   //  desired_dimension
           );

  // Creates a tensor product nurbs surface with srf(s,t) = T(A(s),B(t));
  BOOL TensorProduct(
        const ON_NurbsCurve&, // A
        const ON_NurbsCurve&, // B
        ON_TensorProduct&     // T
        );

  /////////////////////////////////////////////////////////////////
  // Tools for managing CV and knot memory
  BOOL ReserveKnotCapacity( // returns FALSE if allocation fails
                    // does not change m_order or m_cv_count
    int, // dir 0 = "s", 1 = "t"
    int // minimum capacity of m_knot[] array
    );
  BOOL ReserveCVCapacity(  // returns FALSE if allocation fails
                    // does not change m_order or m_cv_count
    int // minimum capacity of m_cv[] array
    );

  /////////////////////////////////////////////////////////////////
  // Implementation
public:
  // NOTE: These members are left "public" so that expert users may efficiently
  //       create NURBS curves using the default constructor and borrow the
  //       knot and CV arrays from their native NURBS representation.
  //       No technical support will be provided for users who access these
  //       members directly.  If you can't get your stuff to work, then use
  //       the constructor with the arguments and the SetKnot() and SetCV()
  //       functions to fill in the arrays.

  int     m_dim;            // (>=1)
  BOOL    m_is_rat;         // TRUE  for rational B-splines.  Control vertices
                            //       use homogeneous form.
                            // FALSE for non-rational B-splines.  Control
                            //       verticies do not have a weight coordinate.
  int     m_order[2];       // order = degree+1 (>=2)
  int     m_cv_count[2];    // number of control vertices ( >= order )

  // knot vector memory
  int     m_knot_capacity[2];
                            // If m_knot_capacity > 0, then m_knot[]
                            // is an array of at least m_knot_capacity
                            // doubles whose memory is managed by the
                            // ON_NurbsSurface class using rhmalloc(),
                            // onrealloc(), and rhfree().
                            // If m_knot_capacity is 0 and m_knot is
                            // not NULL, then  m_knot[] is assumed to
                            // be big enough for any requested operation
                            // and m_knot[] is not deleted by the
                            // destructor.
  double* m_knot[2];        // Knot vector. ( The knot vector has length
                            // m_order+m_cv_count-2. )
  
  // control vertex net memory
  int     m_cv_stride[2];   // The pointer to start of "CV[i]" is
                            //   m_cv + i*m_cv_stride.
  int     m_cv_capacity;    // If m_cv_capacity > 0, then m_cv[] is an array
                            // of at least m_cv_capacity doubles whose
                            // memory is managed by the ON_NurbsSurface
                            // class using rhmalloc(), onrealloc(), and rhfree().
                            // If m_cv_capacity is 0 and m_cv is not
                            // NULL, then m_cv[] is assumed to be big enough
                            // for any requested operation and m_cv[] is not
                            // deleted by the destructor.
  double* m_cv;             // Control points.
                            // If m_is_rat is FALSE, then control point is
                            //   ( CV(i)[0], ..., CV(i)[m_dim-1] ).
                            // If m_is_rat is TRUE, then the control point
                            // is stored in HOMOGENEOUS form and is
                            //   [ CV(i)[0], ..., CV(i)[m_dim] ].
                            // 

};


#endif
