/* $Header: /src2/opennurbs/opennurbs_object.h 15    3/15/01 5:31p Dalelear $ */
/* $NoKeywords: $ */
//
// Copyright (c) 1993-2000 Robert McNeel & Associates. All rights reserved.
// Rhinoceros is a registered trademark of Robert McNeel & Assoicates.
//
// THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
// ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
// MERCHANTABILITY ARE HEREBY DISCLAIMED.
//				
////////////////////////////////////////////////////////////////
//
//   virtual base class for all openNURBS objects
//
////////////////////////////////////////////////////////////////

#if !defined(OPENNURBS_OBJECT_INC_)
#define OPENNURBS_OBJECT_INC_

class ON_ClassId; // used for runtime class identification

////////////////////////////////////////////////////////////////
//
// Runtime class id
//
class ON_CLASS ON_ClassId
{
public:
  ON_ClassId( 
          const char*,      // sClassName
          const char*,      // sBaseClassName 
          ON_Object* (*)(), // function to create new object
          const char*       // UUID in registry format from guidgen
           );
  ~ON_ClassId();

  static const ON_ClassId* ClassId( 
          const char* // class's name
          );

  static const ON_ClassId* ClassId( 
          ON_UUID // class's uuid
          );

  //////////
  // Each class derived from ON_Object has a corresponding ON_ClassId
  // stored in a linked list and the class is marked with an integer
  // value.  ON_ClassId::IncrementMark() increments the value used to
  // mark new classes and returns the new marking value.
  static int IncrementMark();

  //////////
  // Each class derived from ON_Object has a corresponding
  // ON_ClassId stored in a linked list.  If a class definition
  // is going to disappear (which happens when the definition is
  // in a DLL that uses openNURBS as a DLL and the DLL is unloaded),
  // then the class's ON_ClassId needs to be removed from the class 
  // list.  ON_ClassId::Purge( mark ) removes all ON_ClassIds with a
  // a prescribed mark and returns the number of classes that
  // were purged.  
  // So, something like the following works:
  //
  //   int my_dll_classid_mark = ON_ClassId::IncrementMark();
  //   load DLL with classes derived from ON_Object
  //   ...
  //   ON_ClassId::Purge( my_dll_classid_mark );  // call BEFORE unloading DLL
  //   unload DLL
  static int Purge( int );

  //////////
  // Dumps the ON_ClassId list
  static void Dump( ON_TextLog& );

  // OnExitMemoryCleanup() releases all internal memory used by the
  // openNURBS classes for Cast() and serialization construction.
  // ON_ClassId::OnExitMemoryCleanup() should only be called when
  // the application is finished using the openNURBS library.
  static void OnExitMemoryCleanup();

  const char* ClassName() const;

  const char* BaseClassName() const;
  
  const ON_ClassId* BaseClass() const;

	BOOL IsDerivedFrom( const ON_ClassId* ) const;

  ON_Object* Create() const;

  ON_UUID Uuid() const;

private:
  static ON_ClassId* m_p0;     // first in the linked list of class ids
  static int m_mark0;  // current mark value
  ON_ClassId* m_pNext;         // next in the linked list of class ids
  const ON_ClassId* m_pBaseClassId;  // base class id
  char m_sClassName[80];              
  char m_sBaseClassName[80];
  ON_Object* (*m_create)();
  ON_UUID m_uuid;
  int m_mark;

private:
  ON_ClassId(); // no implementation
  ON_ClassId( const ON_ClassId&); // no implementation
  ON_ClassId& operator=( const ON_ClassId&); // no implementation
};

#define ON_CLASS_ID( cls ) ON_ClassId::ClassId( #cls )

// All classes derived from ON_Object must have ON_OBJECT_DECLARE
// in their class definition and either ON_VIRTUAL_OBJECT_IMPLEMENT
// or ON_OBJECT_IMPLEMENT in a .CPP file.
#define ON_OBJECT_DECLARE( cls ) \
  public: static const ON_ClassId m_##cls##_class_id; \
          static cls * Cast( ON_Object* ); \
          static const cls * Cast( const ON_Object* ); \
          virtual const ON_ClassId* ClassId() const; \
          virtual ON_Object* Duplicate() const;

// Objects derived from ON_Object that do not have a valid new, operator=, 
// or copy constructor must use ON_VIRTUAL_OBJECT_IMPLEMENT instead of
// ON_OBJECT_IMPLEMENT.  Objects defined with ON_VIRTUAL_OBJECT_IMPLEMENT
// cannot be serialized using ON_BinaryArchive::ReadObject()/WriteObject()
// or duplicated using ON_Object::Duplicate().
//
// The Cast() and ClassId() members work on objects defined with either
// ON_VIRTUAL_OBJECT_IMPLEMENT or ON_OBJECT_IMPLEMENT.
#define ON_VIRTUAL_OBJECT_IMPLEMENT( cls, basecls, uuid ) \
  const ON_ClassId cls::m_##cls##_class_id(#cls,#basecls,0,uuid);\
  cls * cls::Cast( ON_Object* p) {return(cls *)Cast((const ON_Object*)p);} \
  const cls * cls::Cast( const ON_Object* p) {return(p&&p->IsKindOf(&cls::m_##cls##_class_id))?(const cls *)p:0;} \
  const ON_ClassId* cls::ClassId() const {return &cls::m_##cls##_class_id;} \
  ON_Object* cls::Duplicate() const {return 0;}

// Objects derived from ON_Object that use ON_OBJECT_IMPLEMENT must
// have a valid operator= and copy constructor.  Objects defined with
// ON_OBJECT_IMPLEMENT may be serialized using 
// ON_BinaryArchive::ReadObject()/WriteObject()
// and duplicated by calling ON_Object::Duplicate().
#define ON_OBJECT_IMPLEMENT( cls, basecls, uuid ) \
  static ON_Object* CreateNew##cls() {return new cls();} \
  const ON_ClassId cls::m_##cls##_class_id(#cls,#basecls,CreateNew##cls,uuid);\
  cls * cls::Cast( ON_Object* p) {return(cls *)Cast((const ON_Object*)p);} \
  const cls * cls::Cast( const ON_Object* p) {return(p&&p->IsKindOf(&cls::m_##cls##_class_id))?(const cls *)p:0;} \
  const ON_ClassId* cls::ClassId() const {return &cls::m_##cls##_class_id;} \
  ON_Object* cls::Duplicate() const {cls* p = new cls(); if (p) *p=*this; return p;}

////////////////////////////////////////////////////////////////
//
// ON_Object: pure virtual base class for all classes that must provide
// runtime class id or support object level 3DM serialization
//
class ON_UserData;

class ON_CLASS ON_Object
{
  /////////////////////////////////////////////////////////////////
  //
  // Any object derived from ON_Object should have a
  //   ON_OBJECT_DECLARE(ON_...);
  // as the last line of its class definition and a
  //   ON_OBJECT_IMPLEMENT( ON_..., ON_baseclass );
  // in a .cpp file.
  //
  // These macros declare and implement public members
  //
  // static ON_ClassId m_ON_Object;
  // static cls * Cast( ON_Object* );
  // static const cls * Cast( const ON_Object* );
  // virtual const ON_ClassId* ClassId() const;
  ON_OBJECT_DECLARE(ON_Object);

public:
  ON_Object();
  ON_Object( const ON_Object& );
  ON_Object& operator=( const ON_Object& );
  virtual ~ON_Object();

  // IsKindOf() returns TRUE if the instantiated object is derived from the
  // class whose id is passed as the argument.  The primary reason for IsKindOf()
  // is to support the static Cast() members.  If we determine that dynamic_cast
  // is properly supported and implemented by all supported compilers, then
  // IsKindOf() may dissappear.  If an application needs to determine if a
  // pointer points to a class derived from ON_SomeClassName, then call
  // ON_SomeClassName::Cast(mystery pointer) and check for a non-null return.
  BOOL IsKindOf( 
        const ON_ClassId* // use classname::ClassId()
        ) const;

  /////////////////////////////////////////////////////////////////
  //
  // virtual functions 

  virtual
  BOOL IsValid() const = 0;

  virtual
  void Dump( ON_TextLog& ) const; // for debugging

  // Use ON_BinaryArchive::WriteObject() and ON_BinaryArchive::ReadObject()
  // for top level serialization.  These Read()/Write() members should just
  // write/read specific definitions.  In particular, they should not write/
  // read any chunk typecode or length information.  The default 
  // implementations return FALSE and do nothing.
  virtual
  BOOL Write(
         ON_BinaryArchive&  // serialize definition to binary archive
       ) const;

  virtual
  BOOL Read(
         ON_BinaryArchive&  // restore definition from binary archive
       );

  //////////
  // Useful for switch statements that need to differentiate
  // between basic object types.
  virtual
  ON::object_type ObjectType() const;

  /////////////////////////////////////////////////////////////////
  //
  // user data

  //////////
  // The ON_UserData pointer passed to AttachUserData() must be created 
  // with new.  If TRUE is returned, then ON_Object will delete the
  // user data when appropriate.  If FALSE is returned, then data could
  // not be attached and caller must delete.  AttachUserData() will
  // fail if the user data's m_userdata_uuid field is nil or not unique.
  BOOL AttachUserData( ON_UserData* );

  //////////
  // The ON_UserData pointer passed to DetachUserData() must have been
  // previously attached using AttachUserData().  If TRUE is returned,
  // then the user data was indeed attached to this object and it
  // was detached.  If FALSE is returned, then the user data was not
  // attached to this object to begin with.  In all cases, you
  // can be assured that the user data is no longer attached to "this".
  BOOL DetachUserData( ON_UserData* );

  //////////
  // The UUID is the value of the user data's m_userdata_uuid field.
  // The returned user data is still attached to the object.
  // Deleting the returned user data will automatically remove the
  // user data from the object.
  ON_UserData* GetUserData( const ON_UUID& ) const;

  //////////
  // PurgeUserData() removes all user data from object.  Use
  // delete GetUserData(...) to destroy a single piece of user data.
  void PurgeUserData();

  //////////
  // To enumerate all user data, call FirstUserData() and then use
  // ON_UserData::Next() to traverse the list.
  ON_UserData* FirstUserData() const;

  //////////
  // Objects derived from ON_Geometry must call TransformUserData()
  // in their Transform() member function.
  void TransformUserData( const ON_Xform& );

  //////////
  // Expert user tool that copies user data that has a positive 
  // m_userdata_copycount from  the source object to this.
  // Generally speaking you don't need to use CopyUserData() and
  // can simply rely on ON_Object::operator=() or the copy constructor
  // to do the right thing.
  void CopyUserData( const ON_Object& );

  //////////
  // Expert user tool Moves user data from source object to this, 
  // including user data with a nil m_userdata_copycount.  Deletes
  // any source user data with a duplicate m_userdata_uuid on this.
  void MoveUserData( ON_Object& );


private:
  friend BOOL ON_BinaryArchive::ReadObject( ON_Object** );
  friend BOOL ON_BinaryArchive::WriteObject( const ON_Object& );
  friend class ON_UserData;
  ON_UserData* m_userdata_list;
};

#endif

