/* $Header: /src2/opennurbs/opennurbs_planesurface.cpp 12    1/03/01 4:59p Dalelear $ */
/* $NoKeywords: $ */
/*
//
// Copyright (c) 1993-2001 Robert McNeel & Associates. All rights reserved.
// Rhinoceros is a registered trademark of Robert McNeel & Assoicates.
//
// THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
// ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
// MERCHANTABILITY ARE HEREBY DISCLAIMED.
//				
// For complete openNURBS copyright information see <http://www.opennurbs.org>.
//
////////////////////////////////////////////////////////////////
*/

#include "opennurbs.h"

ON_OBJECT_IMPLEMENT(ON_PlaneSurface,ON_Surface,"4ED7D4DF-E947-11d3-BFE5-0010830122F0");

ON_PlaneSurface::ON_PlaneSurface()
{}

ON_PlaneSurface::ON_PlaneSurface( const ON_PlaneSurface& src )
{
  *this = src;
}

ON_PlaneSurface& ON_PlaneSurface::operator=( const ON_PlaneSurface& src )
{
  if ( this != &src ) {
    ON_Surface::operator=(src);
    m_plane = src.m_plane;
    m_domain[0] = src.m_domain[0];
    m_domain[1] = src.m_domain[1];
  }
  return *this;
}

ON_PlaneSurface::ON_PlaneSurface( const ON_Plane& src )
{
  *this = src;
}

ON_PlaneSurface& ON_PlaneSurface::operator=( const ON_Plane& src )
{
  m_plane = src;
  m_domain[0].Set(0.0,1.0);
  m_domain[1].Set(0.0,1.0);
  return *this;
}

ON_PlaneSurface::~ON_PlaneSurface()
{}

BOOL
ON_PlaneSurface::IsValid() const
{
  return ( m_plane.IsValid() && m_domain[0].IsIncreasing() && m_domain[1].IsIncreasing() ) ? TRUE : FALSE;
}

void
ON_PlaneSurface::Dump( ON_TextLog& dump ) const
{
  dump.Print("ON_PlaneSurface\n");
}

BOOL 
ON_PlaneSurface::Write(
       ON_BinaryArchive& file  // open binary file
     ) const
{
  BOOL rc = file.Write3dmChunkVersion(1,0);
  if (rc)
    rc = file.WritePlane( m_plane );
  if (rc)
    rc = file.WriteInterval( m_domain[0] );
  if (rc)
    rc = file.WriteInterval( m_domain[1] );
  return rc;
}

BOOL 
ON_PlaneSurface::Read(
       ON_BinaryArchive& file // open binary file
     )
{
  int major_version = 0;
  int minor_version = 0;
  BOOL rc = file.Read3dmChunkVersion(&major_version,&minor_version);
  if (rc && major_version == 1) {
    // common to all 1.x formats
    if (rc)
      rc = file.ReadPlane( m_plane );
    if (rc)
      rc = file.ReadInterval( m_domain[0] );
    if (rc)
      rc = file.ReadInterval( m_domain[1] );
  }
  return rc;
}

int 
ON_PlaneSurface::Dimension() const
{
  return 3;
}

BOOL 
ON_PlaneSurface::GetBBox( // returns TRUE if successful
         double* boxmin,    // minimum
         double* boxmax,    // maximum
         BOOL bGrowBox
         ) const
{
  int i,j,k=0;
  ON_3dPoint corner[4];
  for ( i = 0; i < 2; i++ ) for ( j = 0; j < 2; j++ ) {
    corner[k++] = PointAt( m_domain[0].m_t[i], m_domain[1].m_t[j] );
  }
  return ON_GetPointListBoundingBox( 3, 0, 4, 3, corner[0], boxmin, boxmax, bGrowBox );
}

BOOL
ON_PlaneSurface::Transform( const ON_Xform& xform )
{
  TransformUserData(xform);
  return m_plane.Transform( xform );
}

ON_Interval ON_PlaneSurface::Domain( int dir ) const
{
  return dir ? m_domain[1] : m_domain[0];
}


int ON_PlaneSurface::SpanCount( int dir ) const
{
  return 1;
}

BOOL ON_PlaneSurface::GetSpanVector( int dir, double* s ) const
{
  ON_Interval d = Domain(dir);
  s[0] = d.Min();
  s[1] = d.Max();
  return d.IsIncreasing();
}

int ON_PlaneSurface::Degree( int dir ) const
{
  return 1;
}

BOOL 
ON_PlaneSurface::GetParameterTolerance(
         int dir,
         double t,  // t = parameter in domain
         double* tminus, // tminus
         double* tplus   // tplus
         ) const
{
  dir = (dir)?1:0;
  return ON_GetParameterTolerance( m_domain[dir][0], m_domain[dir][1], t, tminus, tplus );
}

BOOL 
ON_PlaneSurface::IsClosed( int dir ) const
{
  return FALSE;
}

BOOL 
ON_PlaneSurface::IsPeriodic( int dir ) const
{
  return FALSE;
}

BOOL 
ON_PlaneSurface::IsSingular( int side ) const
{
  return FALSE;
}

BOOL
ON_PlaneSurface::Reverse( int dir )
{
  m_domain[(dir)?1:0].Reverse();
  return TRUE;
}

BOOL
ON_PlaneSurface::Transpose()
{
  m_plane.Flip();
  ON_Interval i = m_domain[0];
  m_domain[0] = m_domain[1];
  m_domain[1] = i;
  return TRUE;
}

BOOL 
ON_PlaneSurface::Evaluate( // returns FALSE if unable to evaluate
       double s, double t, // evaluation parameters
       int der_count,  // number of derivatives (>=0)
       int v_stride,   // v[] array stride (>=Dimension())
       double* v,      // v[] array of length stride*(ndir+1)
       int side,       // optional - determines which side to evaluate from
                       //         0 = default
                       //      <  0 to evaluate from below, 
                       //      >  0 to evaluate from above
       int* hint       // optional - evaluation hint (int) used to speed
                       //            interative evaluations
       ) const
{
  BOOL rc = FALSE;
  ON_3dPoint P = m_plane.PointAt( s, t );
  v[0] = P.x;
  v[1] = P.y;
  v[2] = P.z;
  v += v_stride;
  if ( der_count >= 1 ) {
    double d = m_domain[0][1] - m_domain[0][0];
    if ( d > 0.0 ) {
      rc = FALSE;
    }
    else {
      d = 1.0/d;
      v[0] = d*m_plane.xaxis.x;
      v[1] = d*m_plane.xaxis.y;
      v[2] = d*m_plane.xaxis.z;
    }
    v += v_stride;

    d = m_domain[1][1] - m_domain[1][0];
    if ( d > 0.0 ) {
      rc = FALSE;
    }
    else {
      d = 1.0/d;
      v[0] = d*m_plane.yaxis.x;
      v[1] = d*m_plane.yaxis.y;
      v[2] = d*m_plane.yaxis.z;
    }
    v += v_stride;

    if ( der_count > 1 ) {
      // zero higher partials
      memset( v, 0, (((der_count+1)*(der_count+2)/2-4)*v_stride+3)*sizeof(*v) );
    }
  }
  return TRUE;
}

BOOL ON_PlaneSurface::GetClosestPoint( const ON_3dPoint& test_point,
        double* s,double* t,  // parameters of local closest point returned here
        double maximum_distance,
        const ON_Interval* sdomain, // first parameter sub_domain
        const ON_Interval* tdomain  // second parameter sub_domain
        ) const
{
  double u = 0.0, v=0.0;
  BOOL rc = m_plane.ClosestPointTo( test_point, &u, &v );
  if ( rc ) {
    if ( sdomain ) {
      if ( u < sdomain->Min() )
        u = sdomain->Min();
      else if ( u > sdomain->Max() )
        u = sdomain->Max();
    }
    if ( tdomain ) {
      if ( v < tdomain->Min() )
        v = tdomain->Min();
      else if ( v > tdomain->Max() )
        v = tdomain->Max();
    }
    if ( s )
      *s = u;
    if ( t )
      *t = v;
    if (maximum_distance > 0.0) 
    {
      ON_3dPoint pt = m_plane.PointAt(u,v);
      if ( test_point.DistanceTo(pt) > maximum_distance )
        rc = FALSE;
    }
  }
  return rc;
}

//////////
// Find parameters of the point on a surface that is locally closest to 
// the test_point.  The search for a local close point starts at 
// seed parameters. If a sub_domain parameter is not NULL, then
// the search is restricted to the specified portion of the surface.
//
// TRUE if returned if the search is successful.  FALSE is returned if
// the search fails.
BOOL ON_PlaneSurface::GetLocalClosestPoint( const ON_3dPoint& test_point, // test_point
        double s0, double t0,     // seed_parameters
        double* s,double* t,   // parameters of local closest point returned here
        const ON_Interval* sdomain, // first parameter sub_domain
        const ON_Interval* tdomain  // second parameter sub_domain
        ) const
{
  // for planes, global serach is fast and returns same answer as local search
  return GetClosestPoint(test_point,s,t,0.0,sdomain,tdomain);
}

int 
ON_PlaneSurface::GetNurbForm( // returns 0: unable to create NURBS representation
                   //            with desired accuracy.
                   //         1: success - returned NURBS parameterization
                   //            matches the surface's to wthe desired accuracy
                   //         2: success - returned NURBS point locus matches
                   //            the surfaces's to the desired accuracy but, on
                   //            the interior of the surface's domain, the 
                   //            surface's parameterization and the NURBS
                   //            parameterization may not match to the 
                   //            desired accuracy.
        ON_NurbsSurface& nurbs,
        double tolerance
        ) const
{
  BOOL rc = IsValid();
  if ( rc ) {
    nurbs.m_dim = 3;
    nurbs.m_is_rat = 0;
    nurbs.m_order[0] = nurbs.m_order[1] = 2;
    nurbs.m_cv_count[0] = nurbs.m_cv_count[1] = 2;
    nurbs.m_cv_stride[1] = nurbs.m_dim;
    nurbs.m_cv_stride[0] = nurbs.m_cv_stride[1]*nurbs.m_cv_count[1];
    nurbs.ReserveCVCapacity(12);
    nurbs.ReserveKnotCapacity(0,2);
    nurbs.ReserveKnotCapacity(1,2);
    nurbs.m_knot[0][0] = m_domain[0][0];
    nurbs.m_knot[0][1] = m_domain[0][1];
    nurbs.m_knot[1][0] = m_domain[1][0];
    nurbs.m_knot[1][1] = m_domain[1][1];
    nurbs.SetCV( 0, 0, PointAt( m_domain[0][0], m_domain[1][0] ));
    nurbs.SetCV( 0, 1, PointAt( m_domain[0][0], m_domain[1][1] ));
    nurbs.SetCV( 1, 0, PointAt( m_domain[0][1], m_domain[1][0] ));
    nurbs.SetCV( 1, 1, PointAt( m_domain[0][1], m_domain[1][1] ));
  }
  return rc;
}
