/* $Header: /src2/opennurbs/opennurbs_point.h 19    7/11/01 4:53p Dalelear $ */
/* $NoKeywords: $ */
/*
//
// Copyright (c) 1993-2001 Robert McNeel & Associates. All rights reserved.
// Rhinoceros is a registered trademark of Robert McNeel & Assoicates.
//
// THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
// ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
// MERCHANTABILITY ARE HEREBY DISCLAIMED.
//				
// For complete openNURBS copyright information see <http://www.opennurbs.org>.
//
////////////////////////////////////////////////////////////////
*/

////////////////////////////////////////////////////////////////
//
//   defines double precision point, vector, and array classes
//
////////////////////////////////////////////////////////////////
#if !defined(ON_POINT_INC_)
#define ON_POINT_INC_

class ON_BoundingBox;
class ON_Xform;
class ON_Line;
class ON_Plane;

class ON_2dPoint;
class ON_3dPoint;
class ON_4dPoint;

class ON_2dVector;
class ON_3dVector;

class ON_2fVector;
class ON_3fVector;

class ON_Interval;

////////////////////////////////////////////////////////////////
//
//   ON_Interval
//
class ON_CLASS ON_Interval
{
public:
  ////////
  // The default constructor creates the empty set interval [ON_UNSET_VALUE,ON_UNSET_VALUE]
  ON_Interval();
  ON_Interval(double,double);
  ~ON_Interval();

  BOOL operator!=(const ON_Interval&) const;
  BOOL operator==(const ON_Interval&) const;

  double m_t[2];

  void Destroy(); // sets interval to [ON_UNSET_VALUE,ON_UNSET_VALUE]
  void Set(double,double);

  double ParameterAt( // returns min*(1.0-x) + max*x
    double // normalized parameter (0 = start, 1 = end)
    ) const; 
  
  double NormalizedParameterAt( // returns x so that min*(1.0-x) + max*x = input
    double // ordinary parameter
    ) const;

  double& operator[](int); // returns (index<=0) ? m_t[0] : m_t[1]
  double operator[](int) const; // returns (index<=0) ? m_t[0] : m_t[1]

  double Min() const; // returns smaller of m_t[0] and m_t[1]
  double Max() const; // returns larger of m_t[0] and m_t[1]
  double Mid() const; // returns 0.5*(m_t[0] + m_t[1])
  double Length() const;

  BOOL IsIncreasing() const; // returns TRUE if m_t[0] < m_t[1]
  BOOL IsDecreasing() const; // returns TRUE if m_t[0] > m_t[0];
  BOOL IsSingleton() const;  // returns TRUE if m_t[0] == m_t[1] != ON_UNSET_VALUE
  BOOL IsEmptySet() const;   // returns TRUE if m_t[0] == m_t[1] == ON_UNSET_VALUE
  BOOL Compare( const ON_Interval& ) const;

  BOOL Includes(double) const; // returns TRUE if double lies in the closed interval
  BOOL Includes( const ON_Interval& ) const; // returns TRUE if argument lies in the closed interval

  void Reverse(); // changes interval to [-m_t[1],-m_t[0]]
  void Swap();    // swap m_t[0] and m_t[1]

  //////////
  // If the intersection is not empty, then 
  // intersection = [max(this.Min(),arg.Min()), min(this.Max(),arg.Max())]
  // Intersection() returns TRUE if the intersection is not empty.
  // The interval [ON_UNSET_VALUE,ON_UNSET_VALUE] is considered to be
  // the empty set interval.  The result of any intersection involving an
  // empty set interval or disjoint intervals is the empty set interval.
  BOOL Intersection( // this = this intersect arg
         const ON_Interval&
         );

  //////////
  // If the intersection is not empty, then 
  // intersection = [max(argA.Min(),argB.Min()), min(argA.Max(),argB.Max())]
  // Intersection() returns TRUE if the intersection is not empty.
  // The interval [ON_UNSET_VALUE,ON_UNSET_VALUE] is considered to be
  // the empty set interval.  The result of any intersection involving an
  // empty set interval or disjoint intervals is the empty set interval.
  BOOL Intersection( // this = intersection of two args
         const ON_Interval&, 
         const ON_Interval&
         );

  //////////
  // The union of an empty set and an increasing interval is the increasing
  // interval.  The union of two empty sets is empty. The union of an empty
  // set an a non-empty interval is the non-empty interval.
  // The union of two non-empty intervals is
  // union = [min(this.Min(),arg.Min()), max(this.Max(),arg.Max()),]
  // Union() returns TRUE if the union is not empty.
  BOOL Union( // this = this union arg
         const ON_Interval&
         );

  //////////
  // The union of an empty set and an increasing interval is the increasing
  // interval.  The union of two empty sets is empty. The union of an empty
  // set an a non-empty interval is the non-empty interval.
  // The union of two non-empty intervals is
  // union = [min(argA.Min(),argB.Min()), max(argA.Max(),argB.Max()),]
  // Union() returns TRUE if the union is not empty.
  BOOL Union( // this = union of two args
         const ON_Interval&, 
         const ON_Interval&
         );
};

////////////////////////////////////////////////////////////////
//
//   ON_2dPoint
//
class ON_CLASS ON_2dPoint
{
public:
  double x, y;

  // use implicit destructor, copy constructor
  ON_2dPoint();                         // not initialized
  ON_2dPoint(const float*);             // from array of 2 floats
  ON_2dPoint(const double*);            // from array of 2 doubles
  ON_2dPoint(double,double);
  ON_2dPoint(const ON_3dPoint& );     // from 3d point
  ON_2dPoint(const ON_4dPoint& );     // from homogeneous 4d point
  ON_2dPoint(const ON_2dVector& );    // from 2d vector

  // (double*) conversion operators
  operator double*();
  operator const double*() const;

  // use implicit operator=(const ON_2dPoint&)
  ON_2dPoint& operator=(const float*);  // point = float[2] support
  ON_2dPoint& operator=(const double*); // point = double[2] support
  ON_2dPoint& operator=(const ON_3dPoint&);
  ON_2dPoint& operator=(const ON_4dPoint&);
  ON_2dPoint& operator=(const ON_2dVector&);

  ON_2dPoint& operator*=(double);
  ON_2dPoint& operator/=(double);
  ON_2dPoint& operator+=(const ON_2dPoint&);
  ON_2dPoint& operator+=(const ON_2dVector&);
  ON_2dPoint& operator-=(const ON_2dPoint&);
  ON_2dPoint& operator-=(const ON_2dVector&);

  ON_2dPoint  operator*(double) const;
  ON_2dPoint  operator/(double) const;
  ON_2dPoint  operator+(const ON_2dPoint&) const;
  ON_2dPoint  operator+(const ON_2dVector&) const;
  ON_2dVector operator-(const ON_2dPoint&) const;
  ON_2dPoint  operator-(const ON_2dVector&) const;

  double operator*(const ON_4dPoint&) const;
  ON_2dPoint operator*(const ON_Xform&) const;

  BOOL operator==(const ON_2dPoint&) const;
  BOOL operator!=(const ON_2dPoint&) const;

  // dictionary order comparisons
  BOOL operator<=(const ON_2dPoint&) const;
  BOOL operator>=(const ON_2dPoint&) const;
  BOOL operator<(const ON_2dPoint&) const;
  BOOL operator>(const ON_2dPoint&) const;

  // index operators mimic double[2] behavior
  double& operator[](int);
  double operator[](int) const;

  // set 2d point value
  void Set(double,double);

  double DistanceTo( const ON_2dPoint& ) const;

  int MaximumCoordinateIndex() const;
  double MaximumCoordinate() const; // absolute value of maximum coordinate

  void Zero(); // set all coordinates to zero;

  // These transform the point in place. The transformation matrix acts on
  // the left of the point; i.e., result = transformation*point
  void Transform( 
        const ON_Xform&
        );

  void Rotate( // rotatation in XY plane
        double,              // angle in radians
        const ON_2dPoint&   // center of rotation
        );

  void Rotate( // rotatation in XY plane
        double,              // sin(angle)
        double,              // cos(angle)
        const ON_2dPoint&   // center of rotation
        );
};

ON_DECL
ON_2dPoint operator*(double, const ON_2dPoint&);

////////////////////////////////////////////////////////////////
//
//   ON_3dPoint
//
class ON_CLASS ON_3dPoint
{
public:
  double x, y, z;

  // use implicit destructor, copy constructor
  ON_3dPoint();                        // not initialized
  ON_3dPoint(const float*);            // from array of 3 floats
  ON_3dPoint(const double*);           // from array of 3 doubles
  ON_3dPoint(double,double,double);
  ON_3dPoint(const ON_2dPoint& );     // from 2d point
  ON_3dPoint(const ON_4dPoint& );     // from homogeneous 4d point
  ON_3dPoint(const ON_3dVector& );    // from 3d vector

  // (double*) conversion operators
  operator double*();
  operator const double*() const;

  // use implicit operator=(const ON_3dPoint&)
  ON_3dPoint& operator=(const float*);  // point = float[3] support
  ON_3dPoint& operator=(const double*); // point = double[3] support
  ON_3dPoint& operator=(const ON_2dPoint&);
  ON_3dPoint& operator=(const ON_4dPoint&);
  ON_3dPoint& operator=(const ON_3dVector&);

  ON_3dPoint& operator*=(double);
  ON_3dPoint& operator/=(double);
  ON_3dPoint& operator+=(const ON_3dPoint&);
  ON_3dPoint& operator+=(const ON_3dVector&);
  ON_3dPoint& operator-=(const ON_3dPoint&);
  ON_3dPoint& operator-=(const ON_3dVector&);

  ON_3dPoint  operator*(double) const;
  ON_3dPoint  operator/(double) const;
  ON_3dPoint  operator+(const ON_3dPoint&) const;
  ON_3dPoint  operator+(const ON_3dVector&) const;
  ON_3dVector operator-(const ON_3dPoint&) const;
  ON_3dPoint  operator-(const ON_3dVector&) const;

  double operator*(const ON_4dPoint&) const;
  ON_3dPoint operator*(const ON_Xform&) const;

  BOOL operator==(const ON_3dPoint&) const;
  BOOL operator!=(const ON_3dPoint&) const;

  // dictionary order comparisons
  BOOL operator<=(const ON_3dPoint&) const;
  BOOL operator>=(const ON_3dPoint&) const;
  BOOL operator<(const ON_3dPoint&) const;
  BOOL operator>(const ON_3dPoint&) const;

  // index operators mimic double[3] behavior
  double& operator[](int);
  double operator[](int) const;

  // set 3d point value
  void Set(double,double,double);

  double DistanceTo( const ON_3dPoint& ) const;

  int MaximumCoordinateIndex() const;
  double MaximumCoordinate() const; // absolute value of maximum coordinate
  double Fuzz( double = ON_ZERO_TOLERANCE ) const; // tolerance to use when comparing 3d points

  void Zero(); // set all coordinates to zero;

  // These transform the point in place. The transformation matrix acts on
  // the left of the point; i.e., result = transformation*point
  void Transform( 
        const ON_Xform&
        );

  void Rotate( 
        double,               // angle in radians
        const ON_3dVector&, // axis of rotation
        const ON_3dPoint&   // center of rotation
        );

  void Rotate( 
        double,               // sin(angle)
        double,               // cos(angle)
        const ON_3dVector&, // axis of rotation
        const ON_3dPoint&   // center of rotation
        );
};

ON_DECL
ON_3dPoint operator*(double, const ON_3dPoint&);

////////////////////////////////////////////////////////////////
//
//   ON_4dPoint (homogeneous coordinates)
//
class ON_CLASS ON_4dPoint
{
public:
  double x, y, z, w;

  // use implicit destructor, copy constructor
  ON_4dPoint();                         // not initialized
  ON_4dPoint(const float*);             // from array of 4 doubles
  ON_4dPoint(const double*);            // from array of 4 doubles
  ON_4dPoint(double,double,double,double);
  ON_4dPoint(const ON_2dPoint& );     // from 2d point
  ON_4dPoint(const ON_3dPoint& );     // from 3d point
  ON_4dPoint(const ON_2dVector& );    // from 2d vector
  ON_4dPoint(const ON_3dVector& );    // from 3d vector

  // (double*) conversion operators
  operator double*();
  operator const double*() const;

  // use implicit operator=(const ON_4dPoint&)
  ON_4dPoint& operator=(const float*);  // point = float[4] support
  ON_4dPoint& operator=(const double*); // point = double[4] support
  ON_4dPoint& operator=(const ON_2dPoint&);
  ON_4dPoint& operator=(const ON_3dPoint&);
  ON_4dPoint& operator=(const ON_2dVector&);
  ON_4dPoint& operator=(const ON_3dVector&);

  ON_4dPoint& operator*=(double);
  ON_4dPoint& operator/=(double);
  ON_4dPoint& operator+=(const ON_4dPoint&); // sum w = sqrt(|w1*w2|)
  ON_4dPoint& operator-=(const ON_4dPoint&); // difference w = sqrt(|w1*w2|)

  ON_4dPoint  operator*(double) const;
  ON_4dPoint  operator/(double) const;
  ON_4dPoint  operator+(const ON_4dPoint&) const; // sum w = sqrt(|w1*w2|)
  ON_4dPoint  operator-(const ON_4dPoint&) const; // difference w = sqrt(|w1*w2|)

  double operator*(const ON_4dPoint&) const;
  ON_4dPoint operator*(const ON_Xform&) const;

  // projective comparison 
  // (i.e., {x,y,z,w} == {c*x,c*y,c*z,c*w} is true for nonzero c)
  BOOL operator==(ON_4dPoint) const;
  BOOL operator!=(const ON_4dPoint&) const;

  // index operators mimic double[4] behavior
  double& operator[](int);
  double operator[](int) const;

  // set 4d point value
  void Set(double,double,double,double);

  int MaximumCoordinateIndex() const;
  double MaximumCoordinate() const; // absolute value of maximum coordinate

  void Zero();      // set all 4 coordinates to zero;
  BOOL Normalize(); // set so x^2 + y^2 + z^2 + w^2 = 1

  // These transform the point in place. The transformation matrix acts on
  // the left of the point; i.e., result = transformation*point
  void Transform( 
        const ON_Xform&
        );
};

ON_DECL
ON_4dPoint operator*(double, const ON_4dPoint&);

////////////////////////////////////////////////////////////////
//
//   ON_2dVector
//
class ON_CLASS ON_2dVector
{
public:
  double x, y;

  // use implicit destructor, copy constructor
  ON_2dVector();                     // not initialized
  ON_2dVector(const float*);         // from array of 2 floats
  ON_2dVector(const double*);        // from array of 2 doubles
  ON_2dVector(double,double);
  ON_2dVector(const ON_3dVector& ); // from 3d vector
  ON_2dVector(const ON_2dPoint& );  // from 2d point

  // (double*) conversion operators
  operator double*();
  operator const double*() const;

  // use implicit operator=(const ON_2dVector&)
  ON_2dVector& operator=(const float*);  // vector = float[2] support
  ON_2dVector& operator=(const double*); // vector = double[2] support
  ON_2dVector& operator=(const ON_3dVector&);
  ON_2dVector& operator=(const ON_2dPoint&);

  ON_2dVector  operator-() const;

  ON_2dVector& operator*=(double);
  ON_2dVector& operator/=(double);
  ON_2dVector& operator+=(const ON_2dVector&);
  ON_2dVector& operator-=(const ON_2dVector&);

  ON_2dVector  operator*(double) const;
  double operator*(const ON_2dVector&) const; // inner (dot) product
  double operator*(const ON_2fVector&) const; // inner (dot) product
  ON_2dVector  operator/(double) const;
  ON_2dVector  operator+(const ON_2dVector&) const;
  ON_2dPoint   operator+(const ON_2dPoint&) const;
  ON_2dVector  operator-(const ON_2dVector&) const;

  double operator*(const ON_4dPoint&) const;
  ON_2dVector operator*(const ON_Xform&) const;

  BOOL operator==(const ON_2dVector&) const;
  BOOL operator!=(const ON_2dVector&) const;

  // dictionary order comparisons
  BOOL operator<=(const ON_2dVector&) const;
  BOOL operator>=(const ON_2dVector&) const;
  BOOL operator<(const ON_2dVector&) const;
  BOOL operator>(const ON_2dVector&) const;

  // index operators mimic double[2] behavior
  double& operator[](int);
  double operator[](int) const;

  // set 2d vector value
  void Set(double,double);

  int MaximumCoordinateIndex() const;
  double MaximumCoordinate() const; // absolute value of maximum coordinate

  double LengthSquared() const;
  double Length() const;

  BOOL Decompose( // Computes a, b such that this vector = a*X + b*Y
         // Returns FALSE if unable to solve for a,b.  This happens
         // when X,Y is not really a basis.
         //
         // If X,Y is known to be an orthonormal frame,
         // then a = V*X, b = V*Y will compute
         // the same result more quickly.
         const ON_2dVector&, // X
         const ON_2dVector&, // Y
         double*, // a
         double*  // b
         ) const;

  int IsParallelTo( 
        // returns  1: this and other vectors are parallel
        //         -1: this and other vectors are anti-parallel
        //          0: this and other vectors are not parallel
        //             or at least one of the vectors is zero
        const ON_2dVector&,                 // other vector     
        double = ON_DEFAULT_ANGLE_TOLERANCE // optional angle tolerance (radians)
        ) const;

  BOOL IsPerpendicularTo(
        // returns TRUE:  this and other vectors are perpendicular
        //         FALSE: this and other vectors are not perpendicular
        //                or at least one of the vectors is zero
        const ON_2dVector&,                 // other vector     
        double = ON_DEFAULT_ANGLE_TOLERANCE // optional angle tolerance (radians)
        ) const;

  void Zero(); // set all coordinates to zero;
  void Reverse(); // negate all coordinates
  BOOL Unitize();  // returns FALSE if vector has zero length

  // set this vector to be perpendicular to another vector
  BOOL PerpendicularTo( // Result is not unitized. 
                        // returns FALSE if input vector is zero
        const ON_2dVector& 
        );

  // set this vector to be perpendicular to a line defined by 2 points
  BOOL PerpendicularTo( 
        const ON_2dPoint&, 
        const ON_2dPoint& 
        );

  // These transform the vector in place. The transformation matrix acts on
  // the left of the vector; i.e., result = transformation*vector
  void Transform( 
        const ON_Xform& // can use ON_Xform here
        );

  void Rotate( 
        double               // angle in radians
        );

  void Rotate( 
        double,              // sin(angle)
        double               // cos(angle)
        );
};

ON_DECL
ON_2dVector operator*(double, const ON_2dVector&);

///////////////////////////////////////////////////////////////
//
// ON_2dVector utilities
//

ON_DECL
double 
ON_DotProduct( 
    const ON_2dVector&, 
    const ON_2dVector& 
    );

ON_DECL
ON_3dVector 
ON_CrossProduct(
    const ON_2dVector&, 
    const ON_2dVector& 
    );

ON_DECL
BOOL 
ON_IsOrthogonalFrame( // TRUE if X, Y are nonzero and mutually perpindicular
    const ON_2dVector&, // X
    const ON_2dVector&  // Y
    );

ON_DECL
BOOL 
ON_IsOrthonormalFrame( // TRUE if X, Y are orthogonal and unit length
    const ON_2dVector&, // X
    const ON_2dVector&  // Y
    );

ON_DECL
BOOL 
ON_IsRightHandFrame( // TRUE if X, Y are orthonormal and right handed
    const ON_2dVector&, // X
    const ON_2dVector&  // Y
    );

////////////////////////////////////////////////////////////////
//
//   ON_3dVector
//
class ON_CLASS ON_3dVector
{
public:
  double x, y, z;

  // use implicit destructor, copy constructor
  ON_3dVector();                     // not initialized
  ON_3dVector(const float*);         // from array of 3 floats
  ON_3dVector(const double*);        // from array of 3 doubles
  ON_3dVector(double,double,double);
  ON_3dVector(const ON_2dVector& ); // from 2d vector
  ON_3dVector(const ON_3dPoint& );  // from 3d point

  // (double*) conversion operators
  operator double*();
  operator const double*() const;

  // use implicit operator=(const ON_3dVector&)
  ON_3dVector& operator=(const float*);  // vector = float[3] support
  ON_3dVector& operator=(const double*); // vector = double[3] support
  ON_3dVector& operator=(const ON_2dVector&);
  ON_3dVector& operator=(const ON_3dPoint&);
  
  ON_3dVector  operator-() const;

  ON_3dVector& operator*=(double);
  ON_3dVector& operator/=(double);
  ON_3dVector& operator+=(const ON_3dVector&);
  ON_3dVector& operator-=(const ON_3dVector&);

  ON_3dVector  operator*(double) const;
  double operator*(const ON_3dVector&) const; // inner (dot) product
  double operator*(const ON_3fVector&) const; // inner (dot) product
  ON_3dVector  operator/(double) const;
  ON_3dVector  operator+(const ON_3dVector&) const;
  ON_3dPoint   operator+(const ON_3dPoint&) const;
  ON_3dVector  operator-(const ON_3dVector&) const;

  double operator*(const ON_4dPoint&) const;
  ON_3dVector operator*(const ON_Xform&) const;

  BOOL operator==(const ON_3dVector&) const;
  BOOL operator!=(const ON_3dVector&) const;

  // dictionary order comparisons
  BOOL operator<=(const ON_3dVector&) const;
  BOOL operator>=(const ON_3dVector&) const;
  BOOL operator<(const ON_3dVector&) const;
  BOOL operator>(const ON_3dVector&) const;

  // index operators mimic double[3] behavior
  double& operator[](int);
  double operator[](int) const;

  // set 3d vector value
  void Set(double,double,double);

  int MaximumCoordinateIndex() const;
  double MaximumCoordinate() const; // absolute value of maximum coordinate

  double LengthSquared() const;
  double Length() const;

  BOOL Decompose( // Computes a, b, c such that this vector = a*X + b*Y + c*Z
         // Returns FALSE if unable to solve for a,b,c.  This happens
         // when X,Y,Z is not really a basis.
         //
         // If X,Y,Z is known to be an orthonormal frame,
         // then a = V*X, b = V*Y, c = V*Z will compute
         // the same result more quickly.
         const ON_3dVector&, // X
         const ON_3dVector&, // Y
         const ON_3dVector&, // Z
         double*, // a
         double*, // b
         double*  // c
         ) const;

  int IsParallelTo( 
        // returns  1: this and other vectors are parallel
        //         -1: this and other vectors are anti-parallel
        //          0: this and other vectors are not parallel
        //             or at least one of the vectors is zero
        const ON_3dVector&,                 // other vector     
        double = ON_DEFAULT_ANGLE_TOLERANCE // optional angle tolerance (radians)
        ) const;

  BOOL IsPerpendicularTo(
        // returns TRUE:  this and other vectors are perpendicular
        //         FALSE: this and other vectors are not perpendicular
        //                or at least one of the vectors is zero
        const ON_3dVector&,                 // other vector     
        double = ON_DEFAULT_ANGLE_TOLERANCE // optional angle tolerance (radians)
        ) const;

  double Fuzz( double = ON_ZERO_TOLERANCE ) const; // tolerance to use when comparing 3d vectors

  void Zero(); // set all coordinates to zero;
  void Reverse(); // negate all coordinates
  BOOL Unitize();  // returns FALSE if vector has zero length

  // set this vector to be perpendicular to another vector
  BOOL PerpendicularTo( // Result is not unitized. 
                        // returns FALSE if input vector is zero
        const ON_3dVector& 
        );

  // set this vector to be perpendicular to a plane defined by 3 points
  BOOL PerpendicularTo(
               // about 3 times slower than
               //    ON_3dVector N = ON_CrossProduct(P1-P0,P2-P0); 
               //    N.Unitize();
               // returns FALSE if points are coincident or colinear
         const ON_3dPoint&, const ON_3dPoint&, const ON_3dPoint& 
         );

  // These transform the vector in place. The transformation matrix acts on
  // the left of the vector; i.e., result = transformation*vector
  void Transform( 
        const ON_Xform& // can use ON_Xform here
        );

  void Rotate( 
        double,             // angle in radians
        const ON_3dVector&  // axis of rotation
        );

  void Rotate( 
        double,             // sin(angle)
        double,             // cos(angle)
        const ON_3dVector&  // axis of rotation
        );
};

ON_DECL
ON_3dVector operator*(double, const ON_3dVector&);

///////////////////////////////////////////////////////////////
//
// ON_3dVector utilities
//

ON_DECL
double 
ON_DotProduct( 
    const ON_3dVector&, 
    const ON_3dVector& 
    );


ON_DECL
ON_3dVector 
ON_CrossProduct(
    const ON_3dVector&, 
    const ON_3dVector& 
    );

ON_DECL
ON_3dVector 
ON_CrossProduct( // 3d cross product for old fashioned arrays
    const double*, // array of 3d doubles
    const double*  // array of 3d doubles
    );

ON_DECL
double 
ON_TripleProduct( 
    const ON_3dVector&,
    const ON_3dVector&,
    const ON_3dVector&
    );

ON_DECL
double 
ON_TripleProduct(  // 3d triple product for old fashioned arrays
    const double*, // array of 3d doubles
    const double*, // array of 3d doubles
    const double*  // array of 3d doubles
    );

ON_DECL
BOOL 
ON_IsOrthogonalFrame( // TRUE if X, Y, Z are nonzero and mutually perpindicular
    const ON_3dVector&, // X
    const ON_3dVector&, // Y
    const ON_3dVector&  // Z 
    );

ON_DECL
BOOL 
ON_IsOrthonormalFrame( // TRUE if X, Y, Z are orthogonal and unit length
    const ON_3dVector&, // X
    const ON_3dVector&, // Y
    const ON_3dVector&  // Z 
    );

ON_DECL
BOOL 
ON_IsRightHandFrame( // TRUE if X, Y, Z are orthonormal and right handed
    const ON_3dVector&, // X
    const ON_3dVector&, // Y
    const ON_3dVector&  // Z 
    );

///////////////////////////////////////////////////////////////
//
// common points and vectors
//
ON_EXTERN_DECL const ON_3dPoint ON_unset_point; // {ON_UNSET_VALUE,ON_UNSET_VALUE,ON_UNSET_VALUE}
ON_EXTERN_DECL const ON_3dPoint ON_origin; // {0.0, 0.0, 0.0}
ON_EXTERN_DECL const ON_3dVector ON_xaxis; // {1.0, 0.0, 0.0}
ON_EXTERN_DECL const ON_3dVector ON_yaxis; // {0.0, 1.0, 0.0}
ON_EXTERN_DECL const ON_3dVector ON_zaxis; // {0.0, 0.0, 1.0}

#include "opennurbs_fpoint.h"

////////////////////////////////////////////////////////////////
//
//   ON_SurfaceCurvature
//
class ON_CLASS ON_SurfaceCurvature
{
public:
  double k1, k2; // principal curvatures
  //ON_3dVector e1, e2; // unit principal directions

  double GaussianCurvature() const;
  double MeanCurvature() const;
  double MinimumRadius() const;
  double MaximumRadius() const;
  //double NormalCurvature(
  //  const ON_3dVector& // tangent to surface
  //  ) const;
  //double NormalSectionCurvature( 
  //  const ON_3dVector&, // section_normal
  //  const ON_3dVector&  // surface_normal 
  //  ) const;
};

#endif

