/* $Header: /src2/opennurbs/opennurbs_polycurve.cpp 16    4/27/01 10:21a Dalelear $ */
/* $NoKeywords: $ */
/*
//
// Copyright (c) 1993-2001 Robert McNeel & Associates. All rights reserved.
// Rhinoceros is a registered trademark of Robert McNeel & Assoicates.
//
// THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
// ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
// MERCHANTABILITY ARE HEREBY DISCLAIMED.
//				
// For complete openNURBS copyright information see <http://www.opennurbs.org>.
//
////////////////////////////////////////////////////////////////
*/

#include "opennurbs.h"

ON_OBJECT_IMPLEMENT(ON_PolyCurve,ON_Curve,"4ED7D4E0-E947-11d3-BFE5-0010830122F0");

ON_PolyCurve::ON_PolyCurve()
{
}

ON_PolyCurve::ON_PolyCurve( int capacity )
              : m_segment(capacity), m_t(capacity+1)
{
  m_segment.Zero();
}

ON_PolyCurve::ON_PolyCurve( const ON_PolyCurve& src )
              : m_segment(src.Count()), m_t(src.Count()+1)
{
  *this = src;
}

void ON_PolyCurve::Destroy()
{
  // release memory
  m_segment.Destroy();
  m_t.Destroy();
}

ON_PolyCurve::~ON_PolyCurve()
{
  Destroy();
}

void ON_PolyCurve::EmergencyDestroy()
{
  m_segment.EmergencyDestroy();
  m_t.EmergencyDestroy();
}

ON_PolyCurve& ON_PolyCurve::operator=( const ON_PolyCurve& src )
{
  if ( this != &src ) {
    ON_Curve::operator=(src);
    const int segment_capacity = m_segment.Capacity();
    ON_Curve** segment = m_segment.Array();
    int i;
    for ( i = 0; i < segment_capacity; i++ ) {
      if ( segment[i] ) {
        delete segment[i];
        segment[i] = 0;
      }
    }
    m_segment.SetCount(0);
    m_segment.SetCapacity(src.m_segment.Count());
    m_segment.Zero();
    m_t.SetCount(0);
    m_t.SetCapacity(src.m_t.Count());
    m_t.Zero();
    m_segment = src.m_segment;
    m_t       = src.m_t;
  }
  return *this;
}

int ON_PolyCurve::Dimension() const
{
  const ON_Curve* p = SegmentCurve(0);
  return (p) ? p->Dimension() : 0;
}

BOOL 
ON_PolyCurve::GetBBox( // returns TRUE if successful
         double* boxmin,    // minimum
         double* boxmax,    // maximum
         BOOL bGrowBox
         ) const
{
  const int count = Count();
  int segment_index;
  BOOL rc = (count>0) ? TRUE : FALSE;
  for ( segment_index = 0; segment_index < count && rc; segment_index++ ) {
    rc = m_segment[segment_index]->GetBBox( boxmin, boxmax, bGrowBox );
    bGrowBox = TRUE;
  }
  return rc;
}

BOOL
ON_PolyCurve::Transform( const ON_Xform& xform )
{
  TransformUserData(xform);
  const int count = Count();
  int segment_index;
  BOOL rc = (count>0) ? TRUE : FALSE;
  for ( segment_index = 0; segment_index < count && rc; segment_index++ ) {
    rc = m_segment[segment_index]->Transform( xform );
  }
  return rc;
}

BOOL
ON_PolyCurve::SwapCoordinates( int i, int j )
{
  const int count = Count();
  int segment_index;
  BOOL rc = (count>0) ? TRUE : FALSE;
  for ( segment_index = 0; segment_index < count && rc; segment_index++ ) {
    rc = m_segment[segment_index]->SwapCoordinates( i, j );
  }
  return rc;
}

BOOL ON_PolyCurve::IsValid() const
{
  const int count = Count();
  const int dim = Dimension();
  double s0, s1;
  ON_3dPoint p0, p1;
  int segment_index;
  BOOL rc = (count>0 && dim>0) ? TRUE : FALSE;
  if ( rc && m_t.Count() != count+1 )
    rc = FALSE;
  for ( segment_index = 0; segment_index < count && rc; segment_index++ ) {
    rc = m_segment[segment_index]->IsValid();
    if ( rc ) {
      if ( m_segment[segment_index]->Dimension() != dim )
        rc = FALSE; // all segments must have same dimension
      if ( m_t[segment_index] >= m_t[segment_index+1] )
        rc = FALSE; // segment domain must be non-empty
      if ( rc && segment_index ) {
        // check that segments are contiguous
        if ( m_segment[segment_index]->IsClosed() ) {
          rc = FALSE; // closed segments not permitted in multi segment curve
        }
        if ( rc )
          s0 =  m_segment[segment_index-1]->Domain().Max();
        if ( rc )
          rc = m_segment[segment_index-1]->EvPoint(s0,p1);
        if ( rc )
          s1 = m_segment[segment_index-1]->Domain().Min();
        if ( rc )
          rc = m_segment[segment_index]->EvPoint(s0,p0);
        if ( rc ) {
          if ( ON_ComparePoint( 3, FALSE, p0, p1 ) )
            rc = FALSE; // not contiguous
        }
      }
    }
  }
  return rc;
}

void ON_PolyCurve::Dump( ON_TextLog& dump ) const
{
  const int count = Count();
  int i;
  dump.Print( "ON_PolyCurve segment count = %d\n", count );
  dump.PushIndent();
  for ( i = 0; i < count; i++ ) {
    dump.Print( "Segment %d: (%g,%g)\n", i+1, m_t[i], m_t[i+1] );
    dump.PushIndent();
    m_segment[i]->Dump(dump);
    dump.PopIndent();
  }
  dump.PopIndent();
}

BOOL ON_PolyCurve::Write(
       ON_BinaryArchive& file // open binary file
     ) const
{
  // NOTE - if changed, check legacy I/O code
  BOOL rc = file.Write3dmChunkVersion(1,0);
  if (rc) {
    int reserved1 = 0;
    int reserved2 = 0;
    const int count = Count();
    int segment_index;

    rc = file.WriteInt( count );
    if (rc) file.WriteInt(reserved1);
    if (rc) file.WriteInt(reserved2);
    if (rc) {
      // may be set in future
      ON_BoundingBox bbox;
      rc = file.WriteBoundingBox(bbox);
    }
    if (rc) rc = file.WriteArray( m_t );

    for ( segment_index = 0; segment_index < count && rc; segment_index++ ) {
      rc = file.WriteObject( *SegmentCurve(segment_index) );
    }


  }
  return rc;
}

BOOL ON_PolyCurve::Read(
       ON_BinaryArchive& file  // open binary file
     )
{
  // NOTE - if changed, check legacy I/O code
  Destroy();
  int major_version = 0;
  int minor_version = 0;
  BOOL rc = file.Read3dmChunkVersion(&major_version,&minor_version);
  if (rc) {
    ON_Object* obj;
    ON_Curve* crv;
    int segment_index;
    int segment_count = 0;
    int reserved1 = 0;
    int reserved2 = 0;
    rc = file.ReadInt(&segment_count);
    if (rc) rc = file.ReadInt(&reserved1);
    if (rc) rc = file.ReadInt(&reserved2);
    if (rc) {
      // may be set in future
      ON_BoundingBox bbox;
      rc = file.ReadBoundingBox(bbox);
    }
    if (rc) rc = file.ReadArray( m_t );

    for ( segment_index = 0; segment_index < segment_count && rc; segment_index++ ) {
      obj = 0;
      crv = 0;
      rc = file.ReadObject( &obj );
      if (rc) {
        crv = ON_Curve::Cast(obj);
        if (crv) {
          Append(crv);
        }
        else {
          ON_ERROR("ON_PolyCurve::Read() - non ON_Curve object in segment list\n");
          delete obj;
          rc = FALSE;
        }
      }
    }
  }
  return rc;
}


ON_Interval ON_PolyCurve::Domain() const
{
  ON_Interval d;
  const int count = Count();
  if ( count > 0 && m_t[0] < m_t[count] ) {
    d.Set(m_t[0],m_t[count]);
  }
  return d;
}

int ON_PolyCurve::SpanCount() const
{
  int span_count = 0;
  const int segment_count = Count();
  int i, j;
  for ( i = 0; i < segment_count; i++  ) {
    if ( !m_segment[i] )
      return FALSE;
    j = m_segment[i]->SpanCount();
    if ( j == 0 )
      return 0;
    span_count += j;
  }
  return span_count;
}

BOOL ON_PolyCurve::GetSpanVector( // span "knots" 
       double* s // array of length SpanCount() + 1 
       ) const
{
  ON_Interval sp;
  double t, d;
  const int segment_count = Count();
  int i, j, k;
  for ( i = 0; i < segment_count; i++  ) {
    if ( !m_segment[i] )
      return FALSE;
    j = m_segment[i]->SpanCount();
    if ( j == 0 )
      return 0;
    if ( !m_segment[i]->GetSpanVector( s ) )
      return FALSE;
    sp.Set( m_t[i], m_t[i+1] );
    if ( sp.Min() != s[0] || sp.Max() != s[j] ) {
      d = s[j] - s[0];
      if ( d == 0.0 )
        return FALSE;
      d = 1.0/d;
      for ( k = 1; k < j; k++ ) {
        t = (s[j] - s[0])*d;
        s[k] = sp.ParameterAt(t);
      }
    }

    s += j;
  }
  return TRUE;
}

int ON_PolyCurve::Degree() const
{
  const int segment_count = Count();
  int span_degree = 0;
  int segment_index, d;
  for ( segment_index = 0; segment_index < segment_count; segment_index++  ) {
    if ( !m_segment[segment_index] )
      return FALSE;
    d = m_segment[segment_index]->Degree();
    if ( d <= 0 )
      return 0;
    if ( d > span_degree )
      span_degree = d;
  }
  return span_degree;
}


BOOL
ON_PolyCurve::IsLinear( // TRUE if curve locus is a line segment
      double tolerance // tolerance to use when checking linearity
      ) const
{
  BOOL rc = FALSE;
  int i, count = Count();
  if ( count > 0 ) {
    rc = TRUE;
    for ( i = 0; rc && i < count; i++ ) {
      if ( !m_segment[i] )
        rc = FALSE;
      else
        rc = m_segment[i]->IsLinear(tolerance);
    
    }
    if (rc)
      rc = ON_Curve::IsLinear(tolerance);
  }
  return rc;
}

BOOL
ON_PolyCurve::IsArc( // TRUE if curve locus in an arc or circle
      const ON_Plane* plane, // if not NULL, test is performed in this plane
      ON_Arc* arc,         // if not NULL and TRUE is returned, then arc
                              // arc parameters are filled in
      double tolerance // tolerance to use when checking linearity
      ) const
{
  return FALSE;
}

BOOL
ON_PolyCurve::IsPlanar(
      ON_Plane* plane, // if not NULL and TRUE is returned, then plane parameters
                         // are filled in
      double tolerance // tolerance to use when checking linearity
      ) const
{
  return FALSE;
}

BOOL
ON_PolyCurve::IsInPlane(
      const ON_Plane& plane, // plane to test
      double tolerance // tolerance to use when checking linearity
      ) const
{
  return FALSE;
}

BOOL 
ON_PolyCurve::IsClosed() const
{
  BOOL bIsClosed = FALSE;
  const int count = Count();
  if ( count == 1 ) {
    // evaluation test required
    const ON_Curve* c = FirstSegmentCurve();
    if ( c )
      bIsClosed = c->IsClosed();
  }
  else if ( count > 1 ) {
    bIsClosed = ON_Curve::IsClosed();
  }
  return bIsClosed;
}

BOOL 
ON_PolyCurve::IsPeriodic() const
{
  BOOL bIsPeriodic = FALSE;
  if ( Count() == 1 ) {
    const ON_Curve* c = FirstSegmentCurve();
    if ( c )
      bIsPeriodic = c->IsPeriodic();
  }
  return bIsPeriodic;
}

BOOL
ON_PolyCurve::Reverse()
{
  const int count = Count();
  int i;
  BOOL rc = (count>0) ? TRUE : FALSE;
  if ( rc ) {
    m_segment.Reverse();
    m_t.Reverse();
    for ( i = 0; i < count; i++ ) {
      m_segment[i]->Reverse();
      m_t[i] = -m_t[i];
    }
    m_t[count] = -m_t[count];
  }
  return rc;
}

BOOL ON_PolyCurve::Evaluate( // returns FALSE if unable to evaluate
       double t,       // evaluation parameter
       int der_count,  // number of derivatives (>=0)
       int v_stride,   // v[] array stride (>=Dimension())
       double* v,      // v[] array of length stride*(ndir+1)
       int side,       // optional - determines which side to evaluate from
                       //         0 = default
                       //      <  0 to evaluate from below, 
                       //      >  0 to evaluate from above
       int* hint       // optional - evaluation hint (int) used to speed
                       //            interative evaluations
       ) const
{
  BOOL rc = FALSE;
  const int count = Count();
  const int dim = Dimension();
  int segment_hint, curve_hint;
  if ( count > 0 && dim > 0 && dim <= v_stride ) {
    segment_hint = (hint) ? (*hint & 0x3FFF) : 0;
    const int segment_index = ON_SpanIndex(2,count+1,m_t,t,side,segment_hint);
    const ON_Curve* c = m_segment[segment_index];
    if ( c ) {
      double s0, s1;
      {
        ON_Interval dom = c->Domain();
        s0 = dom.Min();
        s1 = dom.Max();
      }
      if ( s0 != s1 ) {
        const double t0 = m_t[segment_index];
        const double t1 = m_t[segment_index+1];
        double s;
        if ( s0 == t0 && s1 == t1 )
          s = t;
        else {
          const double d = 1.0/(t1-t0);
          const double a = (t - t0)*d;
          const double b = (t1 - t)*d;
          s = a*s0 + b*s1;
        }
        curve_hint = ( hint && segment_hint == segment_index ) ? ((*hint)>>14) : 0;
        rc = c->Evaluate(
           s, 
           der_count, 
           v_stride, v, 
           side, 
           &curve_hint );
        if ( rc ) {
          if ( s0 > s1 && der_count > 0 ) {
            // negate odd derivatives
            int di, vi;
            for ( di = 0; di <= der_count; di++ ) {
              if ( di%2 ) {
                for ( vi = 0; vi < dim; vi++ ) {
                  v[vi] = -v[vi];
                }
              }
              v += v_stride;
            }
          }
          if ( hint )
            *hint = segment_index | (curve_hint<<14);
        }
      }
    }
  }
  return rc;
}

int
ON_PolyCurve::Count() const
{
  return m_segment.Count();
}


ON_Curve*
ON_PolyCurve::operator[](int segment_index) const
{
  return SegmentCurve(segment_index);
}

ON_Curve*
ON_PolyCurve::SegmentCurve(int segment_index) const
{
  return ( segment_index >= 0 && segment_index < Count() ) ? m_segment[segment_index] : NULL;
}

ON_Interval 
ON_PolyCurve::SegmentDomain( int segment_index ) const
{
  ON_Interval domain;
  if ( segment_index >= 0 && segment_index < Count() ) {
    domain.m_t[0] = m_t[segment_index];
    domain.m_t[1] = m_t[segment_index+1];
  }
  return domain;
}


ON_Curve*
ON_PolyCurve::FirstSegmentCurve() const
{
  return SegmentCurve(0);
}

ON_Curve*
ON_PolyCurve::LastSegmentCurve() const
{
  return SegmentCurve(Count()-1);
}

void
ON_PolyCurve::Reserve( int capacity )
{
  m_segment.Reserve(capacity);
  m_t.Reserve(capacity+1);
}

BOOL ON_PolyCurve::Prepend( ON_Curve* c )
{
  return Insert( 0, c );
}

BOOL ON_PolyCurve::Append( ON_Curve* c )
{
  return Insert( Count(), c );
}

BOOL ON_PolyCurve::Remove( )
{
  return Remove(Count()-1);
}

BOOL ON_PolyCurve::Remove( int segment_index )
{
  BOOL rc = FALSE;
  const int segment_count = Count();
  if ( segment_index >= 0 && segment_index < segment_count ) {
    delete m_segment[segment_index];
    m_segment[segment_index] = 0;
    m_segment.Remove(segment_index);
    if ( segment_index >= 1 && segment_index < segment_count-1 ) {
      double* d = m_t.Array();
      const double delta = d[segment_index] - d[segment_index+1];
      int i;
      for (i=segment_index+1; i <= segment_count; i++ ) {
        d[i] += delta;
      }
    }
    m_t.Remove(segment_index);
    rc = TRUE;
  }
  return rc;
}

BOOL ON_PolyCurve::Insert( int segment_index, ON_Curve* c )
{
  double s0, s1;
  BOOL rc = FALSE;
  const int count = Count();
  if ( segment_index >= 0 && segment_index <= count && c && c->GetDomain(&s0,&s1) ) {
    rc = TRUE;
    m_segment.Insert( segment_index, c );

    // determine polycurve parameters for this segment
    double t0, t1;
    if ( segment_index == count ) {
      // append segment
      if ( count == 0 ) {
        m_t.Append(s0);
        m_t.Append(s1);
      }
      else {
        t0 = m_t[count];
        t1 = ( s0 == t0 ) ? s1 : (s1-s0+t0);
        m_t.Append(t1);
      }
    }
    else if ( segment_index == 0 ) {
      // prepend segment
      t1 = m_t[0];
      t0 = ( s1 == t1 ) ? s0 : (s0-s1+t1);
      m_t.Insert(0,t0);
    }
    else {
      // insert segment
      t0 = m_t[segment_index];
      t1 = ( s0 == t0 ) ? s1 : (s1-s0+t0);
      const double dt = t1-t0;
      m_t.Insert(segment_index+1,t1);
      double* t = m_t.Array();
      for ( int i = segment_index+2; i <= count+1; i++ ) {
        t[i] += dt;
      }      
    }
  }
  return rc;
}

static int compar_dbl(const double* a, const double* b)
{
  if ( *a < *b )
    return -1;
  if ( *a > *b )
    return 1;
  return 0;
}

BOOL ON_PolyCurve::GetClosestPoint( const ON_3dPoint& test_point,
        double* t,       // parameter of local closest point returned here
        double maximum_distance,
        const ON_Interval* sub_domain
        ) const
{
  ON_Workspace ws;
  const int count = Count();
  BOOL rc = FALSE;
  if ( count > 0 ) {
    const ON_Curve* curve;
    ON_BoundingBox bbox;
    int i, *index;
    double d, t0, t1, s0, s1, s;
    ON_SimpleArray<double> near_dist(count);
    ON_Interval seg_domain, crv_domain;

    // get curve domain to search
    GetDomain(&t0,&t1);
    if ( sub_domain ) {
      s0 = sub_domain->Min();
      s1 = sub_domain->Max();
      if ( s0 > t0 )
        t0 = s0;
      if ( s1 < t1 )
        s1 = t1;
    }
    if ( t0 > t1 )
      return FALSE; // nothing to search
    if ( t0 == t1 ) {
      if ( maximum_distance > 0.0 ) {
        d = test_point.DistanceTo(PointAt(t0));
        if ( d <= maximum_distance )
          rc = TRUE;
      }
      else
        rc = TRUE;
      if ( rc && t )
        *t = t0;
      return rc;
    }

    // get minimum distance from test_point to each segment's bounding box
    for ( i = 0; i < count; i++ ) {
      d = 1.0e300;
      curve = m_segment[i];
      if ( m_t[i] <= t1 && m_t[i+1] >= t0 && curve ) {
        bbox = curve->BoundingBox();
        if ( bbox.IsValid() )
          d = test_point.DistanceTo(bbox.ClosestPoint(test_point));
      }
      near_dist.Append(d);
    }

    // sort segments so closest segments are tested first
    index = ws.GetIntMemory(count);
    near_dist.Sort( ON::heap_sort, index, compar_dbl );

    // test each segment
    if ( maximum_distance <= 0.0 )
      maximum_distance = -1.0;
    for ( i = 0; i < count; i++ ) {
      if ( maximum_distance > 0.0 && near_dist[index[i]] > maximum_distance )
        break; // every untested segment is too far away to matter

      if ( near_dist[index[i]] >= 1.0e300 )
        continue; // segment skipped for some reason

      curve = m_segment[index[i]];
      
      // get sub_domain for this curve
      crv_domain = curve->Domain();
      seg_domain.Set(m_t[i],m_t[i+1]);
      s0 = seg_domain.NormalizedParameterAt(t0);
      s1 = seg_domain.NormalizedParameterAt(t1);
      if ( s0 < 0.0 )
        s0 = 0.0;
      if ( s1 > 1.0)
        s1 = 1.0;
      if ( s0 > 0.0 || s1 < 1.0 ) {
        sub_domain = &seg_domain;
        s0 = crv_domain.ParameterAt(s0);
        s1 = crv_domain.ParameterAt(s1);
        crv_domain.Set(s0,s1);
        sub_domain = &crv_domain;
      }
      else {
        sub_domain = 0;
      }

      // test this curve
      if ( curve->GetClosestPoint( test_point, &s, maximum_distance, sub_domain ) ) {
        d = test_point.DistanceTo(curve->PointAt(s));
        if ( maximum_distance < 0.0 || d < maximum_distance ) {
          // this point is the best one we've got so far
          maximum_distance = d;
          if ( t )
            *t = s;
          rc = TRUE;
          if ( d == 0.0 )
            break; // can't do any better than this
        }
      }
    }
  }
  return rc;
}

BOOL ON_PolyCurve::Length(
        double* length,               // length returned here
        double fractional_tolerance,  // default = 1.0e-8
        const ON_Interval* sub_domain // default = NULL
        ) const
{
  // override when possible
  // TODO: add simple integration routine that works with C1 curves
  if ( length )
    *length = 0.0;
  if ( sub_domain && !sub_domain->IsIncreasing() )
    return FALSE;
  const int count = Count();
  ON_Interval* p = 0;
  ON_Interval s;
  double l, t0, t1;
  int i, rc;
  for ( i = 0; i < count; i++ ) {
    p = 0;
    rc = FALSE;
    if ( !m_segment[i] )
      break;
    if ( sub_domain ) {
      s.Set(m_t[0],m_t[1]);
      t0 = s.NormalizedParameterAt(sub_domain->Min());
      if ( t0 < 0.0 )
        t0 = 0.0;
      t1 = s.NormalizedParameterAt(sub_domain->Max());
      if (t1 > 1.0 )
        t1 = 1.0;
      if ( t0 >= t1 )
        continue;
      if ( t0 > 0.0 || t1 < 1.0 ) {
        s = m_segment[i]->Domain();
        t0 = s.ParameterAt(t0);
        t1 = s.ParameterAt(t1);
        s.Set(t0,t1);
        if ( !s.IsIncreasing() )
          continue;
        p = &s;
      }
    }
    rc = m_segment[i]->Length( &l, fractional_tolerance, p );
    if ( !rc )
      break;
    *length += l;
  }
  return rc;
}

BOOL ON_PolyCurve::GetLocalClosestPoint( const ON_3dPoint& test_point,
        double seed_parameter,
        double* t,
        const ON_Interval* sub_domain
        ) const
{
  const int count = Count();
  BOOL rc = FALSE;
  if ( count > 0 ) {
    const ON_Curve* curve;
    double t0, t1, s0, s1, s, crv_seed;
    ON_Interval seg_domain, crv_domain;

    // get curve domain to search
    GetDomain(&t0,&t1);
    if ( sub_domain ) {
      s0 = sub_domain->Min();
      s1 = sub_domain->Max();
      if ( s0 > t0 )
        t0 = s0;
      if ( s1 < t1 )
        s1 = t1;
    }
    if ( t0 > t1 )
      return FALSE; // nothing to search
    if ( t0 == t1 ) {
      if ( t )
        *t = t0;
      return TRUE;
    }
    if ( seed_parameter < t0 )
      seed_parameter = t0;
    else if ( seed_parameter > t1 )
      seed_parameter = t1;

    int segment_index = ON_SpanIndex(2,count+1,m_t,seed_parameter,1,0);
    curve = m_segment[segment_index];
    int next_seg = 0; // 0 means next seg can be before and after this 1

    while(curve)
    {
      // get sub_domain for this curve
      crv_domain = curve->Domain();
      seg_domain.Set(m_t[segment_index],m_t[segment_index+1]);
      crv_seed= crv_domain.ParameterAt(seg_domain.NormalizedParameterAt(seed_parameter));
      s0 = seg_domain.NormalizedParameterAt(t0);
      s1 = seg_domain.NormalizedParameterAt(t1);
      if ( s0 < 0.0 )
        s0 = 0.0;
      if ( s1 > 1.0)
        s1 = 1.0;
      if ( s0 > 0.0 || s1 < 1.0 ) {
        sub_domain = &seg_domain;
        s0 = crv_domain.ParameterAt(s0);
        s1 = crv_domain.ParameterAt(s1);
        crv_domain.Set(s0,s1);
        sub_domain = &crv_domain;
      }
      else {
        sub_domain = 0;
      }

      // test this curve
      if ( !curve->GetLocalClosestPoint( test_point, crv_seed, &s, sub_domain ) )
        break;
      s = curve->Domain().NormalizedParameterAt(s);
      seed_parameter = seg_domain.ParameterAt(s);
      if ( t )
        *t = seed_parameter;
      rc = TRUE;
      if ( s > 0.0 && s < 1.0 )
        break;
      if ( s >= 1.0 ) {
        // see if we need to continue searching on next segment
        if ( next_seg < 0 )
          break;
        next_seg = 1; // only allowed to increase from here on
        segment_index++;
        if ( segment_index >= count-1 )
          break;
        curve = m_segment[segment_index];
      }
      else {
        // see if we need to continue searching on previous segment
        if ( next_seg > 0 )
          break;
        next_seg = -1; // only allowed to decrease from here on
        segment_index--;
        if ( segment_index < 0 )
          break;
        curve = m_segment[segment_index];
      }
    }
  }
  return rc;
}


int ON_PolyCurve::GetNurbForm( 
                              ON_NurbsCurve& nurb, 
                              double tol,
                              const ON_Interval* subdomain  // OPTIONAL subdomain of ON::ProxyCurve::Domain()
                              ) const
{
  ON_Interval domain = Domain();
  if ( !domain.IsIncreasing() )
    return FALSE;
  int rc = 0;
  int si0 = 0;
  int si1 = Count();
  if ( subdomain ) {
    if ( !subdomain->IsIncreasing() )
      return 0;
    if ( !domain.Includes(subdomain->Min()) )
      return 0;
    if ( !domain.Includes(subdomain->Max()) )
      return 0;
    domain = *subdomain;
  }

  while ( si0 < si1 && m_t[si0+1] <= domain.m_t[0] )
    si0++;
  while ( si0 < si1 && m_t[si1-1] >= domain.m_t[1] )
    si1--;
  if ( si0 >= si1 )
    return 0;
  {
    ON_NurbsCurve c;
    int i, rci;
    for ( i = si0; i < si1; i++ ) {
      if ( !m_segment[i] )
        return 0;
      if ( i == si0 ) {
        rc = m_segment[i]->GetNurbForm( nurb, tol, NULL );
        if ( rc < 1 )
          return rc;
        nurb.SetDomain( m_t[i], m_t[i+1] );
      }
      else {
        rci = m_segment[i]->GetNurbForm( c, tol, NULL );
        if ( rci < 1 )
          return rci;
        else if ( rc < rci )
          rc = rci;
        c.SetDomain( m_t[i], m_t[i+1] );
        if ( !nurb.Append( c ) )
          return 0;
        c.Destroy();
      }
    }
  }

  if ( subdomain )
    nurb.Trim( *subdomain );

  return rc;
}

ON_Curve* ON_PolyCurve::HarvestSegment( int i )
{
  ON_Curve* segment_curve = 0;
  if ( i >= 0 && i < m_segment.Count() ) {
    segment_curve = m_segment[i];
    m_segment[i] = 0;
  }
  return segment_curve;
}

