/* $Header: /src2/opennurbs/opennurbs_polyline.cpp 7     1/05/01 4:11p Chuck $ */
/* $NoKeywords: $ */
/*
//
// Copyright (c) 1993-2001 Robert McNeel & Associates. All rights reserved.
// Rhinoceros is a registered trademark of Robert McNeel & Assoicates.
//
// THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
// ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
// MERCHANTABILITY ARE HEREBY DISCLAIMED.
//				
// For complete openNURBS copyright information see <http://www.opennurbs.org>.
//
////////////////////////////////////////////////////////////////
*/

#include "opennurbs.h"

ON_Polyline::ON_Polyline()
{
}

ON_Polyline::ON_Polyline(const ON_Polyline& src)
{
  *this = src;
}

void ON_Polyline::EmergencyDestroy()
{
  m_point.EmergencyDestroy();
}

ON_Polyline& ON_Polyline::operator=(const ON_Polyline& src)
{
  if ( this != &src ) {
    m_point = src.m_point;
  }
  return *this;
}

ON_Polyline::~ON_Polyline()
{
}

int ON_Polyline::PointCount() const
{
  return m_point.Count();
}

int ON_Polyline::SegmentCount() const
{
  int i = PointCount()-1;
  if (i < 0 )
    i = 0;
  return i;
}


ON_3dPoint& ON_Polyline::operator[](int point_index)
{
  return m_point[point_index];
}

const ON_3dPoint& ON_Polyline::operator[](int point_index) const
{
  return m_point[point_index];
}

double ON_Polyline::Length() const
{
  const int count = PointCount();
  double d = 0;
  int i;
  for ( i = 1; i < count; i++ ) {
    d += m_point[i].DistanceTo(m_point[i-1]);
  }
  return d;
}

ON_3dVector ON_Polyline::Direction( int segment_index ) const
{
  ON_3dVector v;
  if ( segment_index >= 0 && segment_index < PointCount()-1 ) {
    v = m_point[segment_index+1] - m_point[segment_index];
  }
  else {
    v.Zero();
  }
  return v;
}

ON_3dVector ON_Polyline::Tangent( int segment_index ) const
{
  ON_3dVector v = Direction(segment_index);
  v.Unitize();
  return v;
}

ON_3dPoint ON_Polyline::PointAt( double t ) const
{
  const int count = PointCount();
  int segment_index = 0;
  if ( count < 0 )
    return ON_origin;
  else if (count == 1 )
    return m_point[0];
  else {
    int segment_index = (int)floor(t);
    if ( segment_index < 0 ) {
      segment_index = 0;
      t = 0.0;
    }
    else if ( segment_index >= count-1 ) {
      segment_index = count-1;
      t = 1.0;
    }
    else {
      t -= ((double)segment_index);
    }
  }

  return (1-t)*m_point[segment_index] + t*m_point[segment_index+1];
}

BOOL ON_Polyline::ClosestPointTo( const ON_3dPoint& point, double *t, int segment_index0, int segment_index1 ) const
{
  BOOL rc = FALSE;
  int segment_index;
  double segment_t, segment_d, best_t, best_d;
  best_t = 0.0; // to keep lint quiet
  best_d = 0.0; // to keep lint quiet
  if ( t ) {
    if ( segment_index0 < 0 )
      segment_index0 = 0;
    if ( segment_index1 > SegmentCount() )
      segment_index1 = SegmentCount();
    for ( segment_index = segment_index0; segment_index < segment_index1; segment_index++ ) {
      double seg_length = m_point[segment_index].DistanceTo(m_point[segment_index + 1]);
      if (seg_length < ON_EPSILON)
        segment_t = 0.0;
      else {
        const ON_3dVector D = Tangent(segment_index);
        const int i = ( point.DistanceTo(m_point[segment_index]) <= point.DistanceTo(m_point[segment_index+1]) ) ? 0 : 1;
        segment_t = (point - m_point[segment_index+i])*D/seg_length;
        if ( i ) {
          segment_t = 1.0 + segment_t;
        }
        if ( segment_t < 0.0 )
          segment_t = 0.0;
        else if (segment_t > 1.0 )
          segment_t = 1.0;
      }
      segment_d = point.DistanceTo((1-segment_t)*m_point[segment_index] + segment_t*m_point[segment_index+1]);
      if ( !segment_index ) {
        best_t = segment_t;
        best_d = segment_d;
      }
      else if ( segment_d < best_d ) {
        best_t = segment_t + ((double)segment_index);
        best_d = segment_d;
      }
      rc = TRUE;
    }
  }
  if (rc)
    *t = best_t;
  return rc;
}

BOOL ON_Polyline::ClosestPointTo( const ON_3dPoint& point, double *t ) const
{
  return ClosestPointTo( point, t, 0, SegmentCount() );
}

ON_3dPoint ON_Polyline::ClosestPointTo( const ON_3dPoint& point ) const
{
  double t;
  BOOL rc = ClosestPointTo( point, &t );
  if ( !rc )
    t = 0.0;
  return PointAt(t);
}

BOOL ON_Polyline::Transform( const ON_Xform& tr )
{
  return m_point.Transform( tr );
}

BOOL ON_Polyline::SwapCoordinates( int i, int j )
{
  return m_point.SwapCoordinates(i,j);
}

// rotate line about a point and axis
BOOL ON_Polyline::Rotate(
      double sin_angle,                  // sin(angle)
      double cos_angle,                  // cos(angle)
      const ON_3dVector& axis,  // axis of rotation
      const ON_3dPoint& center  // center of rotation
      )
{
  const int count = PointCount();
  ON_Xform rot;
  rot.Rotation( sin_angle, cos_angle, axis, center );
  ON_SimpleArray<int> fix_index(128);
  int i;
  for ( i = 0; i < count; i++ ) {
    if ( m_point[i] == center )
      fix_index.Append(i);
  }
  const BOOL rc = Transform( rot );
  for ( i = 0; i < fix_index.Count(); i++ ) {
    m_point[fix_index[i]] = center;
  }
  return rc;
}

BOOL ON_Polyline::Rotate(
      double a,                 // angle in radians
      const ON_3dVector& axis, // axis of rotation
      const ON_3dPoint& center // center of rotation
      )
{
  return Rotate( sin(a), cos(a), axis, center );
}

BOOL ON_Polyline::Translate(
      const ON_3dVector& delta
      )
{
  ON_Xform tr;
  tr.Translation( delta );
  return Transform( tr );
}

