/* $Header: /src2/opennurbs/opennurbs_revsurface.cpp 11    7/11/01 4:53p Dalelear $ */
/* $NoKeywords: $ */
/*
//
// Copyright (c) 1993-2001 Robert McNeel & Associates. All rights reserved.
// Rhinoceros is a registered trademark of Robert McNeel & Assoicates.
//
// THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
// ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
// MERCHANTABILITY ARE HEREBY DISCLAIMED.
//				
// For complete openNURBS copyright information see <http://www.opennurbs.org>.
//
////////////////////////////////////////////////////////////////
*/

#include "opennurbs.h"

ON_OBJECT_IMPLEMENT( ON_RevSurface, ON_Surface, "A16220D3-163B-11d4-8000-0010830122F0");

ON_RevSurface::ON_RevSurface() : m_curve(0), 
                                 m_axis( ON_origin, ON_zaxis ), 
                                 m_angle( 0.0, 2.0*ON_PI ),
                                 m_bTransposed(0)
{}

ON_RevSurface::~ON_RevSurface()
{
  delete m_curve;
  m_curve = 0;
  m_axis.Create( ON_origin, ON_zaxis );
  m_angle.Set(0.0,2.0*ON_PI);
  m_bTransposed = FALSE;
  m_bbox.Destroy();
}

ON_RevSurface::ON_RevSurface( const ON_RevSurface& src ) : m_curve(0)
{
  *this = src;
}

ON_RevSurface& ON_RevSurface::operator=( const ON_RevSurface& src )
{
  if ( this != &src ) {
    if ( m_curve )
      delete m_curve;
    m_curve = 0;
    ON_Surface::operator=(src);
    if ( src.m_curve ) {
      ON_Object* obj = src.m_curve->Duplicate();
      m_curve = ON_Curve::Cast(obj);
      if ( !m_curve )
        delete obj;
    }
    m_axis = src.m_axis;
    m_angle = src.m_angle;
    m_bTransposed = src.m_bTransposed;
    m_bbox = src.m_bbox;
  }
  return *this;
}

BOOL ON_RevSurface::IsValid() const
{
  if ( !m_curve )
    return FALSE;
  if ( !m_curve->IsValid() )
    return FALSE;
  if ( m_curve->Dimension() != 3 )
    return FALSE;
  if ( !m_axis.IsValid() )
    return FALSE;
  if ( !m_angle.IsIncreasing() && !m_angle.IsDecreasing() )
    return FALSE;
  if ( fabs( m_angle.Length() ) > 2.0*ON_PI + ON_ZERO_TOLERANCE )
    return FALSE;
  if ( fabs( m_angle.Length() ) <= ON_ZERO_TOLERANCE )
    return FALSE;
  return TRUE;
}

void ON_RevSurface::Dump( ON_TextLog& dump ) const
{
  ON_Object::Dump(dump); // print class id
  dump.PushIndent();
  if ( m_curve ) {
    dump.Print("Revolute: ");
    dump.PushIndent();
    m_curve->Dump(dump);
    dump.PopIndent();
  }  
}

BOOL ON_RevSurface::Write( ON_BinaryArchive& file ) const
{
  BOOL rc = file.Write3dmChunkVersion(1,0);
  if (rc) {
    rc = file.WriteLine( m_axis );
    rc = file.WriteInterval( m_angle );
    rc = file.WriteBoundingBox( m_bbox );
    rc = file.WriteInt( m_bTransposed );
    if ( m_curve ) {
      rc = file.WriteChar((char)1);
      if (rc) rc = file.WriteObject(*m_curve);
    }
    else {
      rc = file.WriteChar((char)0);
    }
  }
  return rc;
}


BOOL ON_RevSurface::Read( ON_BinaryArchive& file )
{
  int major_version = 0;
  int minor_version = 0;
  BOOL rc = file.Read3dmChunkVersion(&major_version,&minor_version);
  if (rc && major_version == 1) {
    char bHaveCurve = 0;
    rc = file.ReadLine( m_axis );
    rc = file.ReadInterval( m_angle );
    rc = file.ReadBoundingBox( m_bbox );
    rc = file.ReadInt( &m_bTransposed );
    rc = file.ReadChar( &bHaveCurve );
    if ( bHaveCurve ) {
      ON_Object* obj = 0;
      rc = file.ReadObject(&obj);
      if ( obj ) {
        m_curve = ON_Curve::Cast(obj);
        if ( !m_curve )
          delete obj;
      }
    }
  }
  return rc;
}

int ON_RevSurface::Dimension() const
{
  return 3;
}

BOOL ON_RevSurface::Transform( const ON_Xform& xform )
{
  TransformUserData(xform);
  BOOL rc = (m_curve) ? m_curve->Transform(xform) : FALSE;
  if ( !m_axis.Transform(xform) )
    rc = FALSE;
  return rc;
}

BOOL ON_RevSurface::Evaluate( // returns FALSE if unable to evaluate
       double s, // angle
       double t, // curve_parameter
       // evaluation parameters
       int der_count,            // number of derivatives (>=0)
       int v_stride,            // array stride (>=Dimension())
       double* v,        // array of length stride*(ndir+1)*(ndir+2)/2
       int side,        // optional - determines which quadrant to evaluate from
                       //         0 = default
                       //         1 from NE quadrant
                       //         2 from NW quadrant
                       //         3 from SW quadrant
                       //         4 from SE quadrant
       int* hint       // optional - evaluation hint (int[2]) used to speed
                       //            interative evaluations
       ) const
{
  BOOL rc = FALSE;
  double a = cos(s);
  double b = sin(s);
  double ca[4] = {a, -b, -a,  b};
  double sa[4] = {b,  a, -b, -a};
  double x,y,z;
  int i, j, k, src_i, dst_i;
  ON_3dPoint pt;

  if ( m_bTransposed ) {
    x = s; s = t; t = x;
    if ( side == 2 ) side = 4; else if ( side == 4 ) side = 2;
  }
  const int curve_dim = m_curve ? m_curve->Dimension() : 0;
  if ( curve_dim == 2 || curve_dim == 3 ) {
    int curve_side = 0;
    switch(side) {
    case 1:
    case 2:
      curve_side = 1;
      break;
    case 3:
    case 4:
      curve_side = -1;
      break;
    }
    rc = m_curve->Evaluate( t, der_count, v_stride, v, curve_side, hint );
    if ( rc ) {
      ON_3dVector zaxis = m_axis.Tangent();
      ON_3dVector xaxis, yaxis;
      xaxis.PerpendicularTo(zaxis);
      xaxis.Unitize();
      yaxis = ON_CrossProduct(zaxis,xaxis);

      // move curve derivatives to pure t partial spaces in v[]
      if ( curve_dim == 2 ) {
        for ( i = der_count; i >= 1; i-- ) {
          // move curve derivative to proper spots
          src_i = v_stride*i;
          dst_i = v_stride*((i+1)*(i+2)/2 - 1);
          v[dst_i++] = v[src_i++];
          v[dst_i++] = 0.0;
          v[dst_i]   = v[src_i];
        }
      }
      else {
        for ( i = der_count; i >= 1; i-- ) {
          // move curve derivative to proper spots
          src_i = v_stride*i;
          dst_i = v_stride*((i+1)*(i+2)/2 - 1);
          v[dst_i++] = v[src_i++];
          v[dst_i++] = v[src_i++];
          v[dst_i]   = v[src_i];
        }
      }

      // dot curve dervatives with revolution frame's x,y,z axes
      for ( i = 0; i < der_count; i++ ) {
        dst_i = v_stride*((i+1)*(i+2)/2 - 1);
        pt = ON_3dPoint(v+dst_i)-m_axis[0]; // pt = curve derivative
        v[dst_i++] = pt*xaxis;
        v[dst_i++] = pt*yaxis;
        v[dst_i]   = pt*zaxis;
      }

      for ( i = der_count; i >= 0; i-- ) {
        for ( j = 0; j <= i; j++ ) {
          dst_i = v_stride*((i+1)*(i+2)/2 - 1);
          src_i = v_stride*((j+1)*(j+2)/2 - 1);
          k=(i-j)%4;
          a = ca[k];
          b = sa[k];
          x = a*v[src_i] - b*v[src_i+1];
          y = b*v[src_i] + a*v[src_i+1];
          z = v[src_i+2];
          v[dst_i++] = x;
          v[dst_i++] = y;
          v[dst_i]   = z;
        }
      }

      if ( m_bTransposed ) {
        for ( i = 1; i < der_count; i++ ) {
          for ( j = 0, k = i; j < k; j++, k-- ) {
            dst_i = v_stride*i*(i+1)/2;
            src_i = dst_i+k;
            dst_i += j;
            x = v[src_i]; v[src_i++] = v[dst_i]; v[dst_i++] = x;
            x = v[src_i]; v[src_i++] = v[dst_i]; v[dst_i++] = x;
            x = v[src_i]; v[src_i]   = v[dst_i]; v[dst_i]   = x;
          }
        }
      }
    }
  }
  return rc;
}


class ON_RevolutionTensor : public ON_TensorProduct
{
public:
  ON_3dPoint  O;
  ON_3dVector X;
  ON_3dVector Y;
  ON_3dVector Z;
  int DimensionA() const;
  int DimensionB() const;
  int DimensionC() const;
  BOOL Evaluate( double,        // a
                 const double*, // A
                 double,        // b
                 const double*, // B
                 double*        // C
                );
};

int ON_RevolutionTensor::DimensionA() const
{
  return 2;
}

int ON_RevolutionTensor::DimensionB() const
{
  return 3;
}

int ON_RevolutionTensor::DimensionC() const
{
  return 3;
}

BOOL ON_RevolutionTensor::Evaluate( double a, const double* ArcPoint, double b, const double* ShapePoint, double* SrfPoint )
{
	double x, y, z, c, s, rx, ry, A[2], B[3];

	if (a != 1.0) {
		A[0] = a*ArcPoint[0];
		A[1] = a*ArcPoint[1];
		ArcPoint = A;
	}

	if (b != 1.0) {
		B[0] = b*ShapePoint[0];
		B[1] = b*ShapePoint[1];
		B[2] = b*ShapePoint[2];
		ShapePoint = B;
	}
	

	x = (ShapePoint[0] - O.x)*X.x + (ShapePoint[1] - O.y)*X.y + (ShapePoint[2] - O.z)*X.z;
	y = (ShapePoint[0] - O.x)*Y.x + (ShapePoint[1] - O.y)*Y.y + (ShapePoint[2] - O.z)*Y.z;
	z = (ShapePoint[0] - O.x)*Z.x + (ShapePoint[1] - O.y)*Z.y + (ShapePoint[2] - O.z)*Z.z;

	c = ArcPoint[0];
	s = ArcPoint[1];

	rx = c*x - s*y;
	ry = s*x + c*y;

	SrfPoint[0] = O.x + rx*X.x + ry*Y.x + z*Z.x;
	SrfPoint[1] = O.y + rx*X.y + ry*Y.y + z*Z.y;
	SrfPoint[2] = O.z + rx*X.z + ry*Y.z + z*Z.z;

	return TRUE;
}

int ON_RevSurface::GetNurbForm(class ON_NurbsSurface& srf , double tolerance ) const
{
  int rc = 0;
  if ( IsValid() ) {
    ON_NurbsCurve a, c;
    ON_Arc arc;
    arc.plane.CreateFromNormal( ON_zaxis, ON_origin );
    arc.radius = 1.0;
    arc.SetAngleRadians(m_angle.Length());
    if ( arc.GetNurbForm(a) ) {
      const ON_NurbsCurve* nurbscurve = ON_NurbsCurve::Cast(m_curve);
      if ( nurbscurve )
        rc = 1;
      else {
        rc = m_curve->GetNurbForm(c,tolerance);
        if (rc)
          nurbscurve = &c;
      }
      if (nurbscurve) {
        ON_RevolutionTensor rho;
        rho.O = m_axis.from;
        rho.Z = m_axis.Direction();
        rho.Z.Unitize();
        rho.X.PerpendicularTo(rho.Z);
        rho.X.Unitize();
        rho.Y = ON_CrossProduct(rho.Z,rho.X);
        rho.Y.Unitize();
        srf.TensorProduct( a, *nurbscurve, rho );
      }
    }
  }
  return rc;
}

ON_Curve* ON_RevSurface::IsoCurve(int, double ) const
{
  return 0;
}

BOOL ON_RevSurface::Transpose()
{
  m_bTransposed = m_bTransposed ? FALSE : TRUE;
  return TRUE;
}

BOOL ON_RevSurface::Reverse( int dir )
{
  if ( m_bTransposed )
    dir = dir ? 0 : 1;
  return FALSE;
}

BOOL ON_RevSurface::IsSingular( int side ) const // 0 = south, 1 = east, 2 = north, 3 = west
{
  return FALSE;
}

BOOL ON_RevSurface::IsPeriodic( int dir ) const // dir  0 = "s", 1 = "t"
{
  if ( m_bTransposed )
    dir = dir ? 0 : 1;
  return FALSE;
}

BOOL ON_RevSurface::IsClosed( int dir ) const  // dir  0 = "s", 1 = "t"
{
  if ( m_bTransposed )
    dir = dir ? 0 : 1;
  return FALSE;
}

BOOL ON_RevSurface::GetParameterTolerance( // returns tminus < tplus: paramters tminus <= s <= tplus
         int dir,        // 0 gets first parameter, 1 gets second parameter
         double t,       // t = parameter in domain
         double* tminus, // tminus
         double* tplus   // tplus
         ) const
{
  return FALSE;
}

ON_Interval ON_RevSurface::Domain( int dir ) const
{
  ON_Interval d;
  if ( m_bTransposed )
    dir = dir ? 0 : 1;
  if ( !dir )
    d = m_angle;
  else if ( m_curve )
    d = m_curve->Domain();
  return d;
}

int ON_RevSurface::SpanCount( int dir ) const
{
  ON_Interval d;
  if ( m_bTransposed )
    dir = dir ? 0 : 1;
  return dir ? (m_curve?m_curve->SpanCount():0) : 1;
}

BOOL ON_RevSurface::GetSpanVector( int dir, double* s ) const
{
  BOOL rc = FALSE;
  if ( m_bTransposed )
    dir = dir ? 0 : 1;
  if ( !dir ) {
    s[0] = m_angle.Min();
    s[1] = m_angle.Max();
    rc = m_angle.IsIncreasing();
  }
  else if ( m_curve ) {
    rc = m_curve->GetSpanVector(s);
  }
  return rc;
}

int ON_RevSurface::Degree( int dir ) const
{
  int span_degree = 0;
  if ( m_bTransposed )
    dir = dir ? 0 : 1;
  if ( !dir )
    span_degree = 2;
  else if ( m_curve )
    span_degree = m_curve->Degree();
  return span_degree;
}

BOOL ON_RevSurface::GetBBox(    // returns TRUE if successful
       double* boxmin,  // boxmin[dim]
       double* boxmax,  // boxmax[dim]
       BOOL bGrowBox    // TRUE means grow box
       ) const
{
  return FALSE;
}


