/* $Header: /src2/opennurbs/opennurbs_surface.h 13    1/03/01 4:59p Dalelear $ */
/* $NoKeywords: $ */
/*
//
// Copyright (c) 1993-2001 Robert McNeel & Associates. All rights reserved.
// Rhinoceros is a registered trademark of Robert McNeel & Assoicates.
//
// THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
// ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
// MERCHANTABILITY ARE HEREBY DISCLAIMED.
//				
// For complete openNURBS copyright information see <http://www.opennurbs.org>.
//
////////////////////////////////////////////////////////////////
*/

////////////////////////////////////////////////////////////////
//
//   Definition of virtual parametric surface
//
////////////////////////////////////////////////////////////////

#if !defined(OPENNURBS_SURFACE_INC_)
#define OPENNURBS_SURFACE_INC_

class ON_Curve;
class ON_NurbsSurface;
class ON_CLASS ON_Surface : public ON_Geometry
{
  // Any object derived from ON_Surface should have a
  //   ON_OBJECT_DECLARE(ON_...);
  // as the last line of its class definition and a
  //   ON_OBJECT_IMPLEMENT( ON_..., ON_baseclass );
  // in a .cpp file.
  //
  // See the definition of ON_Object for details.
  ON_OBJECT_DECLARE(ON_Surface);

  // pure virtual class for surface objects
public:
  enum ISO // flags for isoparametric curves
  {
    not_iso = 0, // curve is not an isoparameteric curve
                 // note: odd values are all "x" = constant
                 //       even values > 0 are all "y" = constant
    x_iso   = 1, // curve is a "x" = constant (vertical) isoparametric
                 // curve in the interior of the surface's domain
    y_iso   = 2, // curve is a "y" = constant (horizontal) isoparametric
                 // curve in the interior of the surface's domain
    W_iso   = 3, // curve is a "x" = constant isoparametric curve
                 // along the west side of the surface's domain
    S_iso   = 4, // curve is a "y" = constant isoparametric curve
                 // along the south side of the surface's domain
    E_iso   = 5, // curve is a "x" = constant isoparametric curve
                 // along the east side of the surface's domain
    N_iso   = 6, // curve is a "y" = constant isoparametric curve
                 // along the north side of the surface's domain
    iso_count = 7
  };

public:
  ON_Surface();
  ON_Surface(const ON_Surface&);
  ON_Surface& operator=(const ON_Surface&);
  virtual ~ON_Surface();

  //////////
  // override ON_Object::ObjectType() - returns ON::surface_object
  ON::object_type ObjectType() const;

  ////////////////////////////////////////////////////////////////////
  // surface interface

  BOOL GetDomain( 
         int,              // 0 gets first parameter, 1 gets second parameter
         double*, double* 
         ) const;

  virtual
  BOOL SetDomain( 
    int, // 0 sets first parameter's domain, 1 gets second parameter's domain
    double, double 
    );

  virtual
  ON_Interval Domain(
    int // 0 gets first parameter's domain, 1 gets second parameter's domain
    ) const = 0;


  virtual 
  int SpanCount(
    int // 0 gets first parameter's domain, 1 gets second parameter's domain
    ) const = 0; // number of smooth spans in curve

  virtual
  BOOL GetSpanVector( // span "knots" 
    int, // 0 gets first parameter's domain, 1 gets second parameter's domain
    double* // array of length SpanCount() + 1 
    ) const = 0; // 

  virtual 
  int Degree( // returns maximum algebraic degree of any span 
                  // ( or a good estimate if curve spans are not algebraic )
    int // 0 gets first parameter's domain, 1 gets second parameter's domain
    ) const = 0; 

  virtual BOOL GetParameterTolerance( // returns tminus < tplus: paramters tminus <= s <= tplus
         int,     // 0 gets first parameter, 1 gets second parameter
         double,  // t = parameter in domain
         double*, // tminus
         double*  // tplus
         ) const;

  ISO IsIsoparametric( // returns isoparametric status of 2d curve
        const ON_Curve&
        ) const;

  ISO IsIsoparametric( // returns isoparametric status based on bounding box
        const ON_BoundingBox&
        ) const;

  virtual 
  BOOL IsClosed(   // TRUE if surface is closed in direction
        int        // dir  0 = "s", 1 = "t"
        ) const;

  virtual 
  BOOL IsPeriodic( // TRUE if surface is periodic in direction (default is FALSE)
        int        // dir  0 = "s", 1 = "t"
        ) const;

  virtual
  BOOL IsSingular( // TRUE if surface side is collapsed to a point
        int        // side of parameter space to test
                   // 0 = south, 1 = east, 2 = north, 3 = west
        ) const;
  
  virtual 
  BOOL Reverse(  // reverse parameterizatrion, Domain changes from [a,b] to [-b,-a]
    int // dir  0 = "s", 1 = "t"
    ) = 0;

  virtual 
  BOOL Transpose() = 0; // transpose surface parameterization (swap "s" and "t")

  // simple evaluation interface - no error handling
  ON_3dPoint  PointAt( double, double ) const;
  ON_3dVector NormalAt( double, double ) const;

  BOOL EvPoint( // returns FALSE if unable to evaluate
         double, double, // evaluation parameters
         ON_3dPoint&,  // returns value of surface
         int = 0,        // optional - determines which side to evaluate from
                         //         0 = default
                         //         1 from NE quadrant
                         //         2 from NW quadrant
                         //         3 from SW quadrant
                         //         4 from SE quadrant
         int* = 0        // optional - evaluation hint (int[2]) used to speed
                         //            interative evaluations
         ) const;

  BOOL Ev1Der( // returns FALSE if unable to evaluate
         double, double, // evaluation parameters (s,t)
         ON_3dPoint&,  // returns value of surface
         ON_3dVector&, // first partial derivatives (Ds)
         ON_3dVector&, // (Dt)
         int = 0,        // optional - determines which side to evaluate from
                         //         0 = default
                         //         1 from NE quadrant
                         //         2 from NW quadrant
                         //         3 from SW quadrant
                         //         4 from SE quadrant
         int* = 0        // optional - evaluation hint (int[2]) used to speed
                         //            interative evaluations
         ) const;

  BOOL Ev2Der( // returns FALSE if unable to evaluate
         double, double, // evaluation parameters (s,t)
         ON_3dPoint&,  // returns value of surface
         ON_3dVector&, // first partial derivatives (Ds)
         ON_3dVector&, // (Dt)
         ON_3dVector&, // second partial derivatives (Dss)
         ON_3dVector&, // (Dst)
         ON_3dVector&, // (Dtt)
         int = 0,        // optional - determines which side to evaluate from
                         //         0 = default
                         //         1 from NE quadrant
                         //         2 from NW quadrant
                         //         3 from SW quadrant
                         //         4 from SE quadrant
         int* = 0        // optional - evaluation hint (int[2]) used to speed
                         //            interative evaluations
         ) const;

  BOOL EvNormal( // returns FALSE if unable to evaluate
         double, double, // evaluation parameters (s,t)
         ON_3dPoint&,  // returns value of surface
         ON_3dVector&, // unit normal
         int = 0,        // optional - determines which side to evaluate from
                         //         0 = default
                         //         1 from NE quadrant
                         //         2 from NW quadrant
                         //         3 from SW quadrant
                         //         4 from SE quadrant
         int* = 0        // optional - evaluation hint (int[2]) used to speed
                         //            interative evaluations
         ) const;

  BOOL EvNormal( // returns FALSE if unable to evaluate
         double, double, // evaluation parameters (s,t)
         ON_3dVector&, // unit normal
         int = 0,        // optional - determines which side to evaluate from
                         //         0 = default
                         //         1 from NE quadrant
                         //         2 from NW quadrant
                         //         3 from SW quadrant
                         //         4 from SE quadrant
         int* = 0        // optional - evaluation hint (int[2]) used to speed
                         //            interative evaluations
         ) const;

  BOOL EvNormal( // returns FALSE if unable to evaluate
         double, double, // evaluation parameters (s,t)
         ON_3dPoint&,  // returns value of surface
         ON_3dVector&, // first partial derivatives (Ds)
         ON_3dVector&, // (Dt)
         ON_3dVector&, // unit normal
         int = 0,        // optional - determines which side to evaluate from
                         //         0 = default
                         //         1 from NE quadrant
                         //         2 from NW quadrant
                         //         3 from SW quadrant
                         //         4 from SE quadrant
         int* = 0        // optional - evaluation hint (int[2]) used to speed
                         //            interative evaluations
         ) const;

  // work horse evaluator
  virtual 
  BOOL Evaluate( // returns FALSE if unable to evaluate
         double, double, // evaluation parameters
         int,            // number of derivatives (>=0)
         int,            // array stride (>=Dimension())
         double*,        // array of length stride*(ndir+1)*(ndir+2)/2
         int = 0,        // optional - determines which quadrant to evaluate from
                         //         0 = default
                         //         1 from NE quadrant
                         //         2 from NW quadrant
                         //         3 from SW quadrant
                         //         4 from SE quadrant
         int* = 0        // optional - evaluation hint (int[2]) used to speed
                         //            interative evaluations
         ) const = 0;

  virtual
  ON_Curve* IsoCurve(
         int,              // 0 first parameter varies and second parameter is constant
                           //   e.g., point on IsoCurve(0,c) at t is srf(t,c)
                           // 1 first parameter is constant and second parameter varies
                           //   e.g., point on IsoCurve(1,c) at t is srf(c,t)
         double            // value of constant parameter 
         ) const;

  //////////
  // Find parameters of the point on a surface that is closest to test_point.
  // If the maximum_distance parameter is > 0, then only points whose distance
  // to the given point is <= maximum_distance will be returned.  Using a 
  // positive value of maximum_distance can substantially speed up the search.
  // If a sub_domain parameter is not NULL, then the search is restricted
  // to the specified portion of the surface.
  //
  // TRUE if returned if the search is successful.  FALSE is returned if
  // the search fails.
  virtual
  BOOL GetClosestPoint( const ON_3dPoint&, // test_point
          double*,double*,  // parameters of local closest point returned here
          double = 0.0,     // maximum_distance
          const ON_Interval* = NULL, // first parameter sub_domain
          const ON_Interval* = NULL  // second parameter sub_domain
          ) const;

  //////////
  // Find parameters of the point on a surface that is locally closest to 
  // the test_point.  The search for a local close point starts at 
  // seed parameters. If a sub_domain parameter is not NULL, then
  // the search is restricted to the specified portion of the surface.
  //
  // TRUE if returned if the search is successful.  FALSE is returned if
  // the search fails.
  virtual
  BOOL GetLocalClosestPoint( const ON_3dPoint&, // test_point
          double,double,     // seed_parameters
          double*,double*,   // parameters of local closest point returned here
          const ON_Interval* = NULL, // first parameter sub_domain
          const ON_Interval* = NULL  // second parameter sub_domain
          ) const;

  virtual
  int GetNurbForm( // returns 0: unable to create NURBS representation
                   //            with desired accuracy.
                   //         1: success - returned NURBS parameterization
                   //            matches the surface's to wthe desired accuracy
                   //         2: success - returned NURBS point locus matches
                   //            the surfaces's to the desired accuracy but, on
                   //            the interior of the surface's domain, the 
                   //            surface's parameterization and the NURBS
                   //            parameterization may not match to the 
                   //            desired accuracy.
        ON_NurbsSurface&,
        double = 0.0
        ) const;
};

#if defined(ON_DLL_EXPORTS) || defined(ON_DLL_IMPORTS)
// This stuff is here because of a limitation in the way Microsoft
// handles templates and DLLs.  See Microsoft's knowledge base 
// article ID Q168958 for details.
#pragma warning( push )
#pragma warning( disable : 4231 )
ON_TEMPLATE template class ON_CLASS ON_SimpleArray<ON_Surface*>;
#pragma warning( pop )
#endif

class ON_CLASS ON_SurfaceArray : public ON_SimpleArray<ON_Surface*>
{
public:
  ON_SurfaceArray( size_t = 0 );
  ~ON_SurfaceArray();

  BOOL Write( ON_BinaryArchive& ) const;
  BOOL Read( ON_BinaryArchive& );

  void Destroy(); // deletes surfaces in array and sets count to 0

  BOOL Duplicate( ON_SurfaceArray& ) const; // operator= copies the pointer values
                                     // duplicate copies the surfaces themselves
};

#endif
