/* $Header: /src2/opennurbs/opennurbs_viewport.h 7     1/03/01 4:59p Dalelear $ */
/* $NoKeywords: $ */
/*
//
// Copyright (c) 1993-2001 Robert McNeel & Associates. All rights reserved.
// Rhinoceros is a registered trademark of Robert McNeel & Assoicates.
//
// THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
// ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
// MERCHANTABILITY ARE HEREBY DISCLAIMED.
//				
// For complete openNURBS copyright information see <http://www.opennurbs.org>.
//
////////////////////////////////////////////////////////////////
*/

////////////////////////////////////////////////////////////////
//
//   defines ON_Viewport
//
////////////////////////////////////////////////////////////////

#if !defined(OPENNURBS_VIEWPORT_INC_)
#define OPENNURBS_VIEWPORT_INC_

///////////////////////////////////////////////////////////////////////////////
// Class  ON_Viewport
//
//	This object represents a viewing frustum
///////////////////////////////////////////////////////////////////////////////
class ON_CLASS ON_Viewport : public ON_Object 
{
public:

  // Construction
	ON_Viewport();												// create empty spot light
  ~ON_Viewport();
	ON_Viewport& operator=( const ON_Viewport& );

  // ON_Object overrides //////////////////////////////////////////////////////
  //

  BOOL IsValidCamera() const;
  BOOL IsValidFrustum() const;
  BOOL IsValid() const; // TRUE if camera, frustum and screen port are valid

  void Dump( ON_TextLog& ) const; // for debugging

  // Use ON_BinaryArchive::WriteObject() and ON_BinaryArchive::ReadObject()
  // for top level serialization.  These Read()/Write() members should just
  // write/read specific definitions.  In particular, they should not write/
  // read any chunk typecode or length information.  The default 
  // implementations return FALSE and do nothing.
  BOOL Write(
         ON_BinaryArchive&  // open binary file
       ) const;

  BOOL Read(
         ON_BinaryArchive&  // open binary file
       );


  // Interface /////////////////////////////////////////////////////////////////
  //
  void Initialize();

  BOOL SetProjection( ON::view_projection );
  ON::view_projection Projection() const;

  // These return TRUE if the current direction and up are not zero and not
  // parallel so the camera position is well defined.
  BOOL SetCameraLocation( const ON_3dPoint& );
  BOOL SetCameraDirection( const ON_3dVector& );
  BOOL SetCameraUp( const ON_3dVector& );
  //BOOL SetTargetDistance( double );

  ON_3dPoint CameraLocation() const;
  ON_3dVector CameraDirection() const;
  ON_3dVector CameraUp() const;
  //double TargetDistance() const;

  // returns TRUE if current camera orientation is valid
  BOOL GetCameraFrame(
      double*, // CameraLocation[3]
      double*, // CameraX[3]
      double*, // CameraY[3]
      double*  // CameraZ[3]
      ) const;
  // these do not check for a valid camera orientation
  ON_3dVector CameraX() const; // unit to right vector
  ON_3dVector CameraY() const; // unit up vector
  ON_3dVector CameraZ() const; // unit vector in -CameraDirection

  
  BOOL IsCameraFrameWorldPlan( 
      // Returns true if the camera direction = some world axis.
      // The indices report which axes are used.  For a "twisted"
      // plan view it is possible to have zero x and y indices.
      // This function returns true if and only if the "z" index
      // is non-zero.
      //
      // Indices are +/-1 = world +/-x, +/-2 = world +/-y, +/-3 = world +/-z,
      int*, // if true and plan is axis aligned, view x index, else 0
      int*, // if true and plan is axis aligned, view y index, else 0
      int*  // if true, view z index, else 0
      );
  BOOL GetCameraExtents( 
      // returns bounding box in camera coordinates - this is useful information
      // for setting view frustrums to include the point list
      int,           // count = number of 3d points
      int,           // stride = number of doubles to skip between points (>=3)
      const double*, // 3d points in world coordinates
      ON_BoundingBox&, // bounding box in camera coordinates
      BOOL = FALSE   // set to TRUE if you want to enlarge an existing camera coordinate box
      ) const;
  BOOL GetCameraExtents( 
      // returns bounding box in camera coordinates - this is useful information
      // for setting view frustrums to include the point list
      const ON_BoundingBox&, // world coordinate bounding box
      ON_BoundingBox&, // bounding box in camera coordinates
      BOOL = FALSE   // set to TRUE if you want to enlarge an existing camera coordinate box
      ) const;
  BOOL GetCameraExtents( 
      // returns bounding box in camera coordinates - this is useful information
      // for setting view frustrums to include the point list
      ON_3dPoint&,     // world coordinate bounding sphere center
      double,          // world coordinate bounding sphere radius
      ON_BoundingBox&, // bounding box in camera coordinates
      BOOL = FALSE     // set to TRUE if you want to enlarge an existing camera coordinate box
      ) const;

  // Rhino version 1.0 does not support skew frustums; i.e., the
  // viewing frustums used in Rhino have left = -right, bottom = -top.
  // If you specify a skew frustum, all the ON_Viewport functions
  // will work fine.  If you save it a file and read the file into Rhino 1.0,
  // then you will get a symmetric approximation.
  BOOL SetFrustum(
        double, // left
        double, // right   ( left < right )
        double, // bottom
        double, // top     ( bottom < top )
        double, // near
        double  // far     ( 0 < near < far ) // ignored by Rhino version 1.0
        );
  BOOL GetFrustum(
        double*, // left
        double*, // right  (left < right)
        double*, // bottom
        double*, // top    (bottom < top)
        double* = NULL, // near
        double* = NULL  // far  (0 < near < far)
        ) const;

  // SetFrustumAspect() changes the larger of the frustum's widht/height
  // so that the resulting value of width/height matches the requested
  // aspect.  The camera angle is not changed.  If you change the shape
  // of the view port with a call SetScreenPort(), then you generally 
  // want to call SetFrustumAspect() with the value returned by 
  // GetScreenPortAspect().
  BOOL SetFrustumAspect( double );

  // Returns frustum's width/height
  BOOL GetFrustumAspect( double& ) const;

  // Returns world coordinates of frustum's center
  BOOL GetFrustumCenter( double* ) const;

  // The near clipping plane stored in the Rhino 1.0 file is frequently very
  // small and useless for high quality z-buffer based rendering.  The far
  // clipping value is not stored in the file.  Use these functions to set
  // the frustum's near and far clipping planes to appropriate values.
  double FrustumLeft() const;
  double FrustumRight() const;
  double FrustumBottom() const;
  double FrustumTop() const;
  double FrustumNear() const;
  double FrustumFar() const;

  BOOL SetFrustumNearFar(       
         const double*,  // 3d bounding box min
         const double*   // 3d bounding box max
         );
  BOOL SetFrustumNearFar( 
         const double*,  // 3d bounding sphere center
         double          // 3d bounding sphere radius
         );
  BOOL SetFrustumNearFar( 
         double, // near ( > 0 )
         double  // far
         );

  // Location of viewport in pixels (not saved in file)
  // These are provided so you can set the port you are using and get
  // the appropriate transformation(s).  
  //
  // If you are using Windows, then port_left = 0, port_right = width-1,
  // port_bottom = height-1, port_top = 0;
  BOOL SetScreenPort(
        int,    // port_left
        int,    // port_right    ( port_left != port_right )
        int,    // port_bottom
        int,    // port_top      ( port_bottom != port_top)
        int =0, // port_near
        int =0  // port_far
        );
  BOOL GetScreenPort(
        int*,       // left
        int*,       // right         ( port_left != port_right )
        int*,       // port_bottom
        int*,       // port_top      ( port_bottom != port_top)
        int* =NULL, // port_near
        int* =NULL  // port_far
        ) const;
  BOOL GetScreenPortAspect( double& ) const; // port's |width/height|

  BOOL GetCameraAngle( 
          double*, // 1/2 of diagonal subtended angle
          double*, // 1/2 of vertical subtended angle
          double*  // 1/2 of horizontal subtended angle
          ) const;
  BOOL GetCameraAngle( 
          double*  // 1/2 of smallest subtended view angle
          ) const;
  BOOL SetCameraAngle( 
          double  // 1/2 of smallest subtended view angle
                  // 0 < angle < pi/2
          );

  // These functions assume the camera is horizontal and crop the
  // film rather than the image when the aspect of the frustum
  // is not 36/24.  (35mm film is 36mm wide and 24mm high.)
  //
  // The SetCamera35mmLenseLength() preserves camera location,
  // changes the frustum, but maintains the frsutrum's aspect.
  BOOL GetCamera35mmLenseLength( double* /*lense length*/ ) const;
  BOOL SetCamera35mmLenseLength( double /*lense length*/ );

  BOOL GetXform( 
         ON::coordinate_system, // srcCS
         ON::coordinate_system, // destCS
         ON_Xform&       // 4x4 transformation matrix (acts on the left)
         ) const;

  // display tools
  BOOL GetCoordinateSprite(
         int,        // size in pixels of coordinate sprite axes
         int, int,   // screen (x,y) for sprite origin
         int[3],     // returns depth order for axes
         int [3][2]  // screen coords for axes ends
         ) const;

  // Use Extents() as a quick way to set a viewport to so that bounding
  // volume is inside of a viewports frusmtrum.
  // The view angle is used to determine the position of the camera.
  BOOL Extents( 
         double,        // 1/2 smallest subtended view angle
                        // (0 < angle < pi/2)
         const ON_BoundingBox& // 3d world coordinate bounding box
         );
  BOOL Extents( 
         double,        // 1/2 smallest subtended view angle
                        // (0 < angle < pi/2)
         const ON_3dPoint&, // 3d world coordinate bounding sphere center
         double         // 3d sphere radius
         );

  ////////////////////////////////////////////////////////////////////////
  // View changing from screen input points.  Handy for
  // using a mouse to manipulate a view.
  //

  //////////
  // ZoomToScreenRect() may change camera and frustum settings
  BOOL ZoomToScreenRect(
         int, int,  // (x,y) screen coords of a rectangle corner
         int, int   // (x,y) screen coords of opposite rectangle cornerd
         );

  //////////
  // DollyCamera() does not update the frustum's clipping planes.
  // To update the frustum's clipping planes call DollyFrustum(d)
  // with d = dollyVector o cameraFrameZ.  To convert screen locations
  // into a dolly vector, use GetDollyCameraVector().
  BOOL DollyCamera( // Does not update frustum.  To update frustum use 
                    // DollyFrustum(d) with d = dollyVector o cameraFrameZ
          const ON_3dVector& // dolly vector in world coordinates
          );

  //////////
  // Gets a world coordinate dolly vector that can be passed to
  // DollyCamera().
  BOOL GetDollyCameraVector(
         int, int,    // (x,y) screen coords of start point
         int, int,    // (x,y) screen coords of end point
         double,      // distance of projection plane from camera.
                      // When in doubt, use 0.5*(frus_near+frus_far).
         ON_3dVector& // world coordinate dolly vector returned here
         ) const;

  //////////
  // Moves frustum's clipping planes
  BOOL DollyFrustum(
          double // distance to move in camera direction
          );

protected:

  // These boolean status flags are set to TRUE when the associated
  // fields contain valid values
  BOOL m_bValidCamera;
  BOOL m_bValidFrustum;
  BOOL m_bValidPort;

  // Camera Settings: ///////////////////////////////////////////////
  //   projection
  ON::view_projection m_projection;

  //   Camera location, direction and orientation (in world coordinates).
  //   These values are used to set the camera frame vectors CamX, CamY,
  //   CamZ.  If bValidCamera is TRUE, then the CamX, CamY and CamZ
  //   vectors are properly initialized and should be used
  //   instead of CamDir[] and CamUp[].  The frame vectors CamX, CamY, CamZ
  //   are garenteed to be a right handed orthonormal frame.  The CamDir
  //   and CamUp vectors contain the values passed into TL_SetViewCamera().
  ON_3dPoint m_CamLoc;  // camera location
  ON_3dVector m_CamDir; // from camera towards view (nonzero and not parallel to m_CamUp)
  ON_3dVector m_CamUp;  // (nonzero and not parallel to m_CamDir)
  //double m_target_distance; // if > 0, CamLoc + m_target_distance*CamDir 
  //                          // is the position of the view's "target" point.

  // The camera frame vectors are properly initialized by SetCamera()
  ON_3dVector m_CamX;
  ON_3dVector m_CamY;
  ON_3dVector m_CamZ;

  // View Frustum Settings: ///////////////////////////////////////
  //   left, right are camera X coords on near clipping plane
  //   bottom, top are camera Y coords on near clipping plane
  //   near = distance from camera to near clipping plane
  //   far = distance from camera to far clipping plane
  double m_frus_left,   m_frus_right; // frus_left < frus_right 
  double m_frus_bottom, m_frus_top;   // frus_bottom < frus_top 
  double m_frus_near,   m_frus_far;   // frus_near < frus_far 
                                      // in perspective, 0 < frus_near
  

  // Device View Port Box Settings: ( in display device coordinates ) ////
  //   The point (left,bottom,-near), in camera coordinates, of the view
  //   frustum is mapped to pixel coordinate (port_left,port_bottom,port_near).
  //   The point (right,top,-far), in camera coordinates, of the view frustum 
  //   is mapped to pixel coordinate (port_right,port_top,port_far).
  int m_port_left,   m_port_right; // port_left != port_right
  int m_port_bottom, m_port_top;   // port_bottom != port_top  
                                   // In many situations includeing Windows,
                                   //   port_left  = 0 
                                   //   port_right = viewport width-1
                                   //   port_top   = 0
                                   //   port_bottom = viewport height-1
  int m_port_near,   m_port_far;   // (If you want an 8 bit z-buffer with 
                                   // z=255 being "in front of" z=0, then
                                   // set port_near = 255 and port_far = 0.)

private:
  BOOL ON_Viewport::SetCameraFrame(); // used to set m_CamX, m_CamY, m_CamZ

	ON_OBJECT_DECLARE( ON_Viewport );
};

ON_DECL
BOOL 
ON_GetViewportRotationAngles( 
    const ON_3dVector&, // X, // X,Y,Z must be a right handed orthonormal basis
    const ON_3dVector&, // Y, 
    const ON_3dVector&, // Z,
    double*, // angle1, // returns rotation about world Z
    double*, // angle2, // returns rotation about world X ( 0 <= a2 <= pi )
    double*  // angle3  // returns rotation about world Z
    );

ON_DECL
BOOL
ON_ViewportFromRhinoView( // create ON_Viewport from legacy Rhino projection info
        ON::view_projection, // projection,
        const ON_3dPoint&, // rhvp_target, // 3d point
        double, // rhvp_angle1 in radians
        double, // rhvp_angle2 in radians
        double, // rhvp_angle3 in radians
        double, // rhvp_viewsize,     // > 0
        double, // rhvp_cameradist,   // > 0
        int, // screen_width, 
        int, // screen_height,
        ON_Viewport&
        );
#endif

