/* $Header: /src2/opennurbs/opennurbs_xform.h 6     1/03/01 4:59p Dalelear $ */
/* $NoKeywords: $ */
/*
//
// Copyright (c) 1993-2001 Robert McNeel & Associates. All rights reserved.
// Rhinoceros is a registered trademark of Robert McNeel & Assoicates.
//
// THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
// ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
// MERCHANTABILITY ARE HEREBY DISCLAIMED.
//				
// For complete openNURBS copyright information see <http://www.opennurbs.org>.
//
////////////////////////////////////////////////////////////////
*/

////////////////////////////////////////////////////////////////
//
//   defines ON_Xform (4 x 4 transformation matrix)
//
////////////////////////////////////////////////////////////////

#if !defined(ON_XFORM_INC_)
#define ON_XFORM_INC_

class ON_Matrix;

class ON_CLASS ON_Xform
{
public:
  double m_xform[4][4]; // [i][j] = row i, column j.  I.e., 
                        // [0][0] [0][1] [0][2] [0][3]
                        // [1][0] [1][1] [1][2] [1][3]
                        // [2][0] [2][1] [2][2] [2][3]
                        // [3][0] [3][1] [3][2] [3][3]

  // use implicit destructor, copy constructor
  ON_Xform();                     // zero matrix

  ON_Xform( int );                // diagonal matrix (d,d,d,1)
  ON_Xform( double );             // diagonal matrix (d,d,d,1)

#if defined(ON_COMPILER_MSC)
  // Microsoft's compiler won't pass double m[4][4] as a const double[4][4] arg.
  // Gnu's compiler handles this.
  ON_Xform( double[4][4] );       // from standard double m[4][4]
  ON_Xform( float[4][4] );        // from standard float m[4][4]
#endif
  
  ON_Xform( const double[4][4] ); // from standard double m[4][4]
  ON_Xform( const float[4][4] );  // from standard float m[4][4]
  
  ON_Xform( const double* );      // from array of 16 doubles (row0,row1,row2,row3)
  ON_Xform( const float* );       // from array of 16 floats (row0,row1,row2,row3)
  
  ON_Xform( const ON_Matrix& ); // from upper left 4x4 of an
                                    // arbitrary matrix.  Any missing
                                    // rows/columns are set to identity. 
	ON_Xform(const ON_3dPoint& P,	// as a frame. 
						const ON_3dVector& X,	
						const ON_3dVector& Y,	
						const ON_3dVector& Z); 

  // use implicit operator=(const ON_3dVector&), operator==
  
  double* operator[](int);
  const double* operator[](int) const;

  // xform = scalar results in a diagonal 3x3 with bottom row = 0,0,0,1
  ON_Xform& operator=( int );
  ON_Xform& operator=( float );
  ON_Xform& operator=( double );
  ON_Xform& operator=( const ON_Matrix& ); // from upper left 4x4 of an
                                               // arbitrary matrix.  Any missing
                                               // rows/columns are set to identity.

  // All non-commutative operations have "this" as left hand side and
  // argument as right hand side.
  ON_2dPoint operator*( const ON_2dPoint& ) const;
  ON_3dPoint operator*( const ON_3dPoint& ) const;
  ON_4dPoint operator*( const ON_4dPoint& ) const;
  
  ON_2dVector operator*( const ON_2dVector& ) const;
  ON_3dVector operator*( const ON_3dVector& ) const;
  
  ON_Xform operator*( const ON_Xform& /*rhs*/ ) const;
  ON_Xform operator+( const ON_Xform& ) const;
  ON_Xform operator-( const ON_Xform& /*rhs*/ ) const;
  
  // matrix operations
  void Transpose(); // transposes 4x4 matrix

  int 
  Rank( // returns 0 to 4
    double* = NULL // If not NULL, returns minimum pivot
  ) const;

  double
  Determinant( // returns determinant of 4x4 matrix
    double* = NULL // If not NULL, returns minimum pivot
  ) const;

  BOOL
  Invert( // If matrix is non-singular, returns TRUE,
          // otherwise returns FALSE and sets matrix to 
          // pseudo inverse.
    double* = NULL // If not NULL, returns minimum pivot
  );

  ON_Xform
  Inverse(  // If matrix is non-singular, returns inverse,
            // otherwise returns pseudo inverse.
    double* = NULL // If not NULL, returns minimum pivot
  ) const;

  // workhorses for matrix*vector calculations
  void ActOnLeft(
         double,double,double,double, // source x,y,z,w
         double[4]                    // = matrix * transpose([x,y,z,w])
         ) const;
  void ActOnRight(
         double,double,double,double, // source x,y,z,w
         double[4]                    // = [x,y,z,w] * matrix
         ) const;

  // standard transformations
  void Zero();
  void Identity();
  void Diagonal(double); // diagonal 3x3 with bottom row = 0,0,0,1
  void Scale( double, double, double );
  void Scale( const ON_3dVector& );
  void Translation( const ON_3dVector& ); // delta is right column
  void Translation( double, double, double ); // delta is right column
  void Rotation(
    double,               // sin(rotation angle)
    double,               // cos(rotation angle)
    const ON_3dVector&, // 3d unit axis
    const ON_3dPoint&   // 3d center of rotation
    );
  void Rotation(
    double,               // angle in radians
    const ON_3dVector&, // 3d unit axis
    const ON_3dPoint&   // 3d center of rotation
    );
  void Rotation( 
    // General: Fixes origin and rotates initial frame to final frame.  The
    //          frames need to be orthoganal unit vectors ordered so that
    //          Z = X x Y.
    // Details: fixes 0 and maps X0 to X1, Y0 to Y1, Z0 to Z1
    // Related Commands: ChangeBasis();
    const ON_3dVector&, // initial frame X  (X,Y,Z = orthonormal basis)
    const ON_3dVector&, // initial frame Y
    const ON_3dVector&, // initial frame Z
    const ON_3dVector&, // final frame X    (X,Y,Z = another orthonormal basis)
    const ON_3dVector&, // final frame Y
    const ON_3dVector&  // final frame Z
    );
  void Rotation( 
    // General: Like the Rotation() above, but this one first translates the
    //          initial center to the final center.
    // Details: Maps P0 to P1, P0+X0 to P1+X1, ...
    const ON_3dPoint&,  // initial frame center
    const ON_3dVector&, // initial frame X  (X,Y,Z = orthonormal basis)
    const ON_3dVector&, // initial frame Y
    const ON_3dVector&, // initial frame Z
    const ON_3dPoint&,  // final frame center
    const ON_3dVector&, // final frame X    (X,Y,Z = another orthonormal basis)
    const ON_3dVector&, // final frame Y
    const ON_3dVector&  // final frame Z
    );
  BOOL ChangeBasis( 
    // General: If you have points defined with respect to planes, this
    //          computes the transformation to change coordinates from
    //          one plane to another.  The predefined world plane
    //          ON_world_plane can be used as an argument.
    // Details: If P = plane0.Evaluate( a0,b0,c0 ) and
    //          {a1,b1,c1} = ChangeBasis(plane0,plane1)*ON_3dPoint(a0,b0,c0),
    //          then P = plane1.Evaluate( a1, b1, c1 )
    //          
    const ON_Plane&, // initial plane
    const ON_Plane&  // final plane
    );
  BOOL ChangeBasis( 
    // Details: maps (a0,b0,c0) to (a1,b1,c1) where
    //          a0*X0 + b0*Y0 + c0*Z0 = a1*X1 + b1*Y1 + c1*Z1
    //          returns FALSE if initial vectors are not a basis
    // Related Commands: Rotation();
    const ON_3dVector&, // initial X  (X,Y,Z = any basis
    const ON_3dVector&, // initial Y
    const ON_3dVector&, // initial Z
    const ON_3dVector&, // final X    (X,Y,Z = another basis)
    const ON_3dVector&, // final Y
    const ON_3dVector&  // final Z
    );
  BOOL ChangeBasis( 
    // Details: maps (a0,b0,c0) to (a1,b1,c1) where
    //          P0 + a0*X0 + b0*Y0 + c0*Z0 = P1 + a1*X1 + b1*Y1 + c1*Z1
    //          returns FALSE if initial vectors are not a basis
    // Related Commands: Rotation();
    const ON_3dPoint&,  // initial center
    const ON_3dVector&, // initial X  (X,Y,Z = any basis)
    const ON_3dVector&, // initial Y
    const ON_3dVector&, // initial Z
    const ON_3dPoint&,  // final center
    const ON_3dVector&, // final X    (X,Y,Z = any other basis)
    const ON_3dVector&, // final Y
    const ON_3dVector&  // final Z
    );

  // standard viewing transformations
  void WorldToCamera( 
         const ON_3dPoint&,  // CameraLocation
         const ON_3dVector&, // unit CameraX vector (right)
         const ON_3dVector&, // unit CameraY vector (up)
         const ON_3dVector&  // unit CameraZ vector (from screen to camera)
         );
  void CameraToWorld( 
         const ON_3dPoint&,  // CameraLocation
         const ON_3dVector&, // unit CameraX vector (right)
         const ON_3dVector&, // unit CameraY vector (up)
         const ON_3dVector&  // unit CameraZ vector (from screen to camera)
         );
  BOOL CameraToClip( // maps viewport frustum to -1 <= x,y,z <= 1 box
      BOOL, // TRUE for perspective, FALSE for orthographic
      double, double, // left != right (usually left < right )
      double, double, // bottom != top (usually bottom < top )
      double, double  // near != far (usually 0 < near < far )
      );
  BOOL ClipToCamera( // maps -1 <= x,y,z <= 1 box to viewport frustum
      int, // TRUE for perspective, FALSE for orthographic
      double, double, // left != right (usually left < right )
      double, double, // bottom != top (usually bottom < top )
      double, double  // near != far an bot are non-zero (usually 0 < near < far )
      );
  BOOL ClipToScreen( // maps -1<x<1,-1<y<1,-1<z<1 to 
                          // (left,right) x (bottom,top) x (near,far)
      double, double, // left, right,
      double, double, // bottom, top,
      double, double  // near_z, far_z
      );
  BOOL ScreenToClip( // maps (left,right) x (bottom,top) x (near,far)
                          // -1<x<1,-1<y<1,-1<z<1.
      double, double, // left, right,
      double, double, // bottom, top,
      double, double  // near_z, far_z
      );
  // Clipping transformation flags
  // returns clipflag with
  //   clipflag &  1 if transformed x is coordinate < -1
  //   clipflag &  2 if transformed x is coordinate >  1
  //   clipflag &  4 if transformed y is coordinate < -1
  //   clipflag &  8 if transformed y is coordinate >  1
  //   clipflag & 16 if transformed z is coordinate < -1
  //   clipflag & 32 if transformed z is coordinate >  1
  // illegal input returns 1|2|4|8|16|32 = 63
  int ClipFlag4d(
    const double* // 4d point
    ) const;
  int ClipFlag4d(
    int,           // number of points
    int,           // stride ( >= 4 )
    const double*, // points
    BOOL = TRUE    // FALSE to not test "z"
    ) const;
  int ClipFlag3d(
    const double* // 3d point
    ) const;
  int ClipFlag3d(
    int,           // number of points
    int,           // stride ( >= 3 )
    const double*, // points
    BOOL = TRUE    // FALSE to not test "z"
    ) const;
  int ClipFlag3dBox(
    const double*, // 3d boxmin
    const double*  // 3d boxmax
    ) const;
};


#endif
