/* $Header: /src2/opennurbs/opennurbs_zlib.cpp 10    1/03/01 4:59p Dalelear $ */
/* $NoKeywords: $ */
/*
//
// Copyright (c) 1993-2001 Robert McNeel & Associates. All rights reserved.
// Rhinoceros is a registered trademark of Robert McNeel & Assoicates.
//
// THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
// ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
// MERCHANTABILITY ARE HEREBY DISCLAIMED.
//				
// For complete openNURBS copyright information see <http://www.opennurbs.org>.
//
////////////////////////////////////////////////////////////////
*/

#include "opennurbs.h"
#include "opennurbs_zlib.h"

voidpf zcalloc (voidpf opaque, unsigned items, unsigned size)
{
    return oncalloc(items, size);
}

void  zcfree (voidpf opaque, voidpf ptr)
{
    onfree(ptr);
}

BOOL ON_BinaryArchive::WriteCompressedBuffer(
        int sizeof_inbuffer,  // sizeof uncompressed input data
        const void* inbuffer  // uncompressed input data
        )
{
  int compressed_size = 0;
  BOOL rc = FALSE;

  if ( !WriteMode() )
    return FALSE;
  if ( sizeof_inbuffer < 0 || (sizeof_inbuffer > 0 && !inbuffer) )
    return FALSE;


  // number of bytes of uncompressed data
  if (!WriteInt(sizeof_inbuffer))
    return FALSE;
  if ( sizeof_inbuffer == 0 )
    return TRUE;

  // 32 bit crc of uncompressed data
  const unsigned int buffer_crc = ON_CRC32( 0, sizeof_inbuffer, inbuffer );
  if (!WriteInt(buffer_crc))
    return FALSE;

  unsigned char method = (sizeof_inbuffer > 128) ? 1 : 0;
  if ( method ) {
    if ( !CompressionInit() ) {
      CompressionEnd();
      method = 0;
    }
  }
  if ( !WriteChar(method) )
    return FALSE;

  switch ( method )
  {
  case 0: // uncompressed
    rc = WriteByte(sizeof_inbuffer, inbuffer);
    if ( rc )
      compressed_size = sizeof_inbuffer;
    break;
  case 1: // compressed
    compressed_size = WriteDeflate( sizeof_inbuffer, inbuffer );
    rc = ( compressed_size > 0 );
    CompressionEnd();
    break;
  }


  return rc;
}

BOOL ON_BinaryArchive::ReadCompressedBufferSize( int* sizeof_outbuffer )
{
  int sz = 0;
  BOOL rc = ReadInt(&sz);
  if ( sizeof_outbuffer )
    *sizeof_outbuffer = sz;
  return rc;
}

BOOL ON_BinaryArchive::ReadCompressedBuffer( // read and uncompress
  int sizeof_outbuffer,  // sizeof of uncompressed buffer to read
  void* outbuffer,       // uncompressed output data returned here
  BOOL* bFailedCRC
  )
{
  BOOL rc = FALSE;
  unsigned int buffer_crc0 = 0;
  unsigned int buffer_crc1 = 0;
  char method = 0;

  if ( bFailedCRC)
    *bFailedCRC = FALSE;
  if ( !ReadMode() )
    return FALSE;
  if ( sizeof_outbuffer == 0 )
    return TRUE;
  if ( sizeof_outbuffer < 0 || !outbuffer )
    return FALSE;

  if ( !ReadInt(&buffer_crc0) ) // 32 bit crc of uncompressed buffer
    return FALSE;

  if ( !ReadChar(&method) )
    return FALSE;

  if ( method != 0 && method != 1 )
    return FALSE;

  switch(method)
  {
  case 0: // uncompressed
    rc = ReadByte(sizeof_outbuffer, outbuffer);
    break;
  case 1: // compressed
    rc = CompressionInit();
    if (rc)
      rc = ReadInflate( sizeof_outbuffer, outbuffer );
    CompressionEnd();
    break;
  }
  if (rc ) {
    buffer_crc1 = ON_CRC32( 0, sizeof_outbuffer, outbuffer );
    if ( buffer_crc1 != buffer_crc1 ) {
      ON_ERROR("ON_BinaryArchive::ReadCompressedBuffer() crc error");
      if ( bFailedCRC )
        *bFailedCRC = TRUE;
    }
  }

  return rc;
}

int ON_BinaryArchive::WriteDeflate( // returns number of bytes written
        int sizeof_inbuffer,  // sizeof uncompressed input data ( > 0 )
        const void* inbuffer  // uncompressed input data ( != NULL )
        )
{
  //  Based on gzwrite() and do_flush() in zlib\gzio.c
  BOOL rc = BeginWrite3dmChunk(TCODE_ANONYMOUS_CHUNK,0);
  if ( !rc )
    return FALSE;

  int out_count = 0;
  BOOL bDone = FALSE;
  int write_len = 0;
  int zrc = Z_OK;


  m_zlib.strm.next_in = (unsigned char*)inbuffer;
  m_zlib.strm.avail_in = sizeof_inbuffer;

  m_zlib.strm.next_out = m_zlib.buffer;
  m_zlib.strm.avail_out = m_zlib.sizeof_buffer;

  while(m_zlib.strm.avail_in > 0 && rc) {
    if ( m_zlib.strm.avail_out == 0 ) {
      rc = WriteChar( m_zlib.sizeof_buffer, m_zlib.buffer );
      if ( !rc )
        break;
      out_count += m_zlib.sizeof_buffer;
      m_zlib.strm.next_out = m_zlib.buffer;
      m_zlib.strm.avail_out = m_zlib.sizeof_buffer;
    }
    zrc = deflate( &m_zlib.strm, Z_NO_FLUSH );
    if ( zrc < 0 ) {
      rc = FALSE;
      break;
    }
    if ( zrc != Z_OK )
      break;
  }

  while (rc) {
    write_len = m_zlib.sizeof_buffer - m_zlib.strm.avail_out;
    if (write_len > 0) {
      rc = WriteChar( write_len, m_zlib.buffer );
      if ( !rc )
        break;
      out_count += write_len;
      m_zlib.strm.next_out = m_zlib.buffer;
      m_zlib.strm.avail_out = m_zlib.sizeof_buffer;
    }
    if (bDone)
      break;
    zrc = deflate( &m_zlib.strm, Z_FINISH );

	  /* Ignore the second of two consecutive flushes: */
	  if (write_len == 0 && zrc == Z_BUF_ERROR) 
      zrc = Z_OK;
    
    /* deflate has finished flushing only when it hasn't used up
    * all the available space in the output buffer: 
    */
    bDone = (m_zlib.strm.avail_out != 0 || zrc == Z_STREAM_END);
    
    if (zrc != Z_OK && zrc != Z_STREAM_END) {
      rc = FALSE;
      break;
    }
  }

  if ( !EndWrite3dmChunk() )
    return FALSE;

  if ( !rc ) {
    out_count = 0;
  }

  return out_count;
}


BOOL ON_BinaryArchive::ReadInflate(
        int sizeof_outbuffer,  // sizeof uncompressed data
        void* outbuffer        // buffer for uncompressed data
        )
{
  unsigned int tcode = 0;
  int value = 0;
  BOOL rc = FALSE;

  if ( !BeginRead3dmChunk(&tcode,&value) )
    return FALSE;

  if ( tcode == TCODE_ANONYMOUS_CHUNK && value > 4 ) for(;;) {
    ON_Workspace ws;
    int zrc = -1;
    int sizeof_inbuffer = value-4; // the last 4 bytes in this chunk are a 32 bit crc
    void* inbuffer;

    if ( sizeof_outbuffer <= 0 || !outbuffer)
      break;

    inbuffer = ws.GetMemory(sizeof_inbuffer);
    if ( !inbuffer )
      break;

    if ( !ReadByte( sizeof_inbuffer, inbuffer ) )
      break;

    m_zlib.strm.next_out = (unsigned char*)outbuffer;
    m_zlib.strm.avail_out = sizeof_outbuffer;

    m_zlib.strm.next_in = (unsigned char*)inbuffer;
    m_zlib.strm.avail_in = sizeof_inbuffer;

    zrc = inflate( &m_zlib.strm, Z_FINISH );

    switch(zrc)
    {
    case Z_OK:
    case Z_STREAM_END:
      rc = TRUE;
      break;
    }

    break;
  }

  if ( !EndRead3dmChunk() )
    rc = FALSE;

  return rc;
}

BOOL ON_BinaryArchive::CompressionInit()
{
  // inflateInit() and deflateInit() are in zlib 1.3.3
  BOOL rc = FALSE;
  if ( WriteMode() ) {
    rc = ( m_zlib.mode == ON::write ) ? TRUE : FALSE;
    if ( !rc ) {
      CompressionEnd();
      if ( Z_OK == deflateInit( &m_zlib.strm, Z_BEST_COMPRESSION ) ) {
        m_zlib.mode = ON::write;
        rc = TRUE;
      }
      else {
        memset(&m_zlib.strm,0,sizeof(m_zlib.strm));
      }
    }
  }
  else if ( ReadMode() ) {
    rc = ( m_zlib.mode == ON::read ) ? TRUE : FALSE;
    if ( !rc ) {
      CompressionEnd();
      if ( Z_OK == inflateInit( &m_zlib.strm ) ) {
        m_zlib.mode = ON::read;
        rc = TRUE;
      }
      else {
        memset(&m_zlib.strm,0,sizeof(m_zlib.strm));
      }
    }
  }
  else {
    CompressionEnd();
  }
  return rc;
}

void ON_BinaryArchive::CompressionEnd()
{
  // inflateEnd() and deflateEnd() are in zlib 1.3.3
  switch ( m_zlib.mode ) {
  case ON::read:
  case ON::read3dm:
    inflateEnd(&m_zlib.strm);
    break;
  case ON::write:
  case ON::write3dm:
    deflateEnd(&m_zlib.strm);
    break;
  default: // to quiet lint
    break;
  }
  memset(&m_zlib.strm,0,sizeof(m_zlib.strm));
  m_zlib.mode = ON::unknown_archive_mode;
}


