/* $Header: /Src/RhinoIO/example_brep/example_brep.cpp 8     11/11/99 10:08a Dalelear $ */
/* $NoKeywords: $ */
//
// Copyright (c) 1993-1998 Robert McNeel & Associates. All rights reserved.
// Rhinoceros is a registered trademark of Robert McNeel & Assoicates.
//
// THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
// ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
// MERCHANTABILITY ARE HEREBY DISCLAIMED.
//				
////////////////////////////////////////////////////////////////

#include "rhinoio.h"

// This example demonstrates two things:
//
// 1) How to dig through a CRhinoBrep face  and get at the surface
//    and trimming information.  See TraverseRhinoBrepFace() for
//    details.
//
// 2) How to write a Rhino B-rep.  See MakeTwistedCube() for details.

void TraverseRhinoBrepFace( 
       const CRhinoBrep& brep,
       int fi                   // brep face index
       )
{
  if ( fi < 0 || fi >= (int)brep.m_F.Size() ) {
    printf("Invalid face index\n");
    return;
  }

  const CRhinoBrepFace& face = brep.m_F[fi];

  // pSrf = underlying NURBS surface
  const CRhinoNurbsSurface* pSrf = brep.m_S[face.m_si];

  // The face is trimmed with one or more trimming loops.
  // The first trimming loop is the outer boundary.
  // All the 2d trimming curves are oriented so that the
  // active region of the trimmed surface lies to the left
  // of the 2d trimming curve.  
  //
  // If face.m_bRev is TRUE, the orientations of the face in
  // the b-rep is opposited the natural parameteric orientation
  // of the surface.

  // loop_count = number of trimming loops on this face (>=1)
  const int loop_count = face.m_li.Size(); 

  int fli; // face's loop index
  for ( fli = 0; fli < loop_count; fli++ ) {
    const int li = face.m_li[fli]; // li = brep loop index
    const CRhinoBrepLoop& loop = brep.m_L[li];

    // loop_edge_count = number of trimming edges in this loop
    const int loop_trim_count = loop.m_ti.Size();


    // count 2d and 3d NURBS curves in this loop (purely pedagogical)
    int count_2d_nurbs_curves = 0; // 2d NURBS curve counter
    int count_3d_nurbs_curves = 0; // 3d NURBS curve counter

    int lti; // loop's trim index
    for ( lti = 0; lti < loop_trim_count; lti++ ) {
      const int ti = loop.m_ti[lti]; // ti = brep trim index
      const CRhinoBrepTrim& trim = brep.m_T[ti];

      //////////////////////////////////////////////////////
      // 2d trimming information
      //
      // Each trim has a 2d parameter space curve made from
      // one or more NURBS curves.
      const int c2i = trim.m_c2i; // c2i = brep 2d curve index
      const CRhinoAgCurve* p2dPolyCrv = brep.m_C2[c2i];
      const int c2d_nurbs_count = p2dPolyCrv->NurbsCurveCount();
      int j;
      for ( j = 0; j < c2d_nurbs_count; j++ ) {
        count_2d_nurbs_curves++;
        const CRhinoNurbsCurve* p2dNurbsCrv = (*p2dPolyCrv)[j];
        // p2dNurbsCrv points to a 2d NURBS trimming curve
      }


      //////////////////////////////////////////////////////
      // topology and 3d geometry information
      //

      // Trim starts at v0 and ends at v1.  When the trim
      // is a loop or on a singular surface side, v0i and v1i
      // will be equal.
      const int v0i = trim.m_vi[0]; // v0i = brep vertex index
      const int v1i = trim.m_vi[1]; // v1i = brep vertex index
      const CRhinoBrepVertex& v0 = brep.m_V[v0i];
      const CRhinoBrepVertex& v1 = brep.m_V[v1i];
      // The vX.m_ei[] array contains the brep.m_E[] indices of
      // the edges that begin or end at vX.
      
      const int ei = trim.m_ei;
      if ( ei == -1 ) {
        // This trim lies on a portion of a singular surface side.
        // The vertex indices are still valid and will be equal.
      }
      else {
        // If trim.m_bRev3d is FALSE, the orientations of the 3d edge
        // and the 3d curve obtained by composing the surface and 2d
        // curve agree.
        //
        // If trim.m_bRev3d is TRUE, the orientations of the 3d edge
        // and the 3d curve obtained by composing the surface and 2d
        // curve are opposite.
        const CRhinoBrepEdge& edge = brep.m_E[ei];
        const int c3i = edge.m_c3i;
        const CRhinoAgCurve* p3dPolyCrv = brep.m_C3[c3i];
        const int c3d_nurbs_count = p3dPolyCrv->NurbsCurveCount();
        // (c3d_nurbs_count can be different from c2d_nurbs_count)
        for ( j = 0; j < c3d_nurbs_count; j++ ) {
          count_3d_nurbs_curves++;
          const CRhinoNurbsCurve* p3dNurbsCrv = (*p3dPolyCrv)[j];
          // p3dNurbsCrv points to a 3d NURBS curve.
        }

        // The edge.m_ti[] array contains the brep.m_T[] indices
        // for the other trims that are joined to this edge.
      }
    }
  }
}

static const int // symbolic constants to make code more readable
  A = 0,
  B = 1,
  C = 2,
  D = 3,
  E = 4,
  F = 5,
  G = 6,
  H = 7;

static const int // symbolic constants to make code more readable
  AB =  0,
  BC =  1,
  CD =  2,
  AD =  3,
  EF =  4,
  FG =  5,
  GH =  6,
  EH =  7,
  AE =  8,
  BF =  9,
  CG = 10,
  DH = 11;

static const int // symbolic constants to make code more readable
  ABCD =  0,
  BCGF =  1,
  CDHG =  2,
  ADHE =  3,
  ABFE =  4,
  EFGH =  5;


static CRhinoAgCurve* TwistedCubeTrimmingCurve(
              const CRhinoNurbsSurface& s,
              int side // 0 = SW to SE
                       // 1 = SE to NE
                       // 2 = NE to NW
                       // 3 = NW to SW
              )
{
  // A trimming curve is a 2d curve whose image lies in the surface's domain.
  // The "active" portion of the surface is to the left of the trimming curve.
  // A trimming loop consists of a simple closed curve running counter-clockwise
  // around the region it trims.

  double Start[2], End[2], u0, u1, v0, v1;

  s.Domain( 0, &u0, &u1 );
  s.Domain( 1, &v0, &v1 );

  switch ( side ) {
  case 0:  // SW to SE
    Start[0] = u0; Start[1] = v0;
    End[0]   = u1; End[1]   = v0;
    break;
  case 1: // SE to NE
    Start[0] = u1; Start[1] = v0;
    End[0]   = u1; End[1]   = v1;
    break;
  case 2: // NE to NW
    Start[0] = u1; Start[1] = v1;
    End[0]   = u0; End[1]   = v1;
    break;
  case 3: // NW to SW
    Start[0] = u0; Start[1] = v1;
    End[0]   = u0; End[1]   = v0;
    break;
  default:
    return 0;
  }

  CRhinoAgCurve* c2d = new CRhinoAgCurve();
  CRhinoNurbsCurve* pNurbsCurve = new CRhinoNurbsCurve( 
                                            2,     // dimension
                                            FALSE, // not rational
                                            1,     // degree
                                            2      // number of control vertices
                                            );
  pNurbsCurve->SetCV( 0, rh_not_rational, Start );
  pNurbsCurve->SetCV( 1, rh_not_rational, End );
  pNurbsCurve->SetKnot( 0, 0.0 );
  pNurbsCurve->SetKnot( 1, 1.0 );

  c2d->AddNurbsCurve( pNurbsCurve );

  return c2d;
}


static CRhinoAgCurve* TwistedCubeEdgeCurve( const double Start[3], const double End[3] )
{
  // creates a 3d line segment to be used as a 3d curve in a CRhinoBrep
  CRhinoAgCurve* c3d = new CRhinoAgCurve();
  CRhinoNurbsCurve* pNurbsCurve = new CRhinoNurbsCurve( 
                                            3,     // dimension
                                            FALSE, // not rational
                                            1,     // degree
                                            2      // number of control vertices
                                            );
  pNurbsCurve->SetCV( 0, rh_not_rational, Start );
  pNurbsCurve->SetCV( 1, rh_not_rational, End );
  pNurbsCurve->SetKnot( 0, 0.0 );
  pNurbsCurve->SetKnot( 1, 1.0 );

  c3d->AddNurbsCurve( pNurbsCurve );

  return c3d;
}

static CRhinoNurbsSurface* TwistedCubeSideSurface( 
                             const double SW[3], const double SE[3],
                             const double NE[3], const double NW[3]
                             )
{
  CRhinoNurbsSurface* pNurbsSurface = new CRhinoNurbsSurface(
                                        3,     // dimension
                                        FALSE, // not rational
                                        1,     // "u" degree
                                        1,     // "v" degree
                                        2,     // number of control vertices in "u" dir
                                        2      // number of control vertices in "v" dir
                                        );
  // corner CVs in counter clockwise order starting in the south west
  pNurbsSurface->SetCV( 0,0, rh_not_rational, SW );
  pNurbsSurface->SetCV( 1,0, rh_not_rational, SE );
  pNurbsSurface->SetCV( 1,1, rh_not_rational, NE );
  pNurbsSurface->SetCV( 0,1, rh_not_rational, NW );
  // "u" knots
  pNurbsSurface->SetKnot( 0,0, 0.0 );
  pNurbsSurface->SetKnot( 0,1, 1.0 );
  // "v" knots
  pNurbsSurface->SetKnot( 1,0, 0.0 );
  pNurbsSurface->SetKnot( 1,1, 1.0 );

  return pNurbsSurface;
}


static void MakeTwistedCubeEdges( CRhinoBrep* brep )
{

  // In this simple example, the edge indices exactly match the 3d
  // curve indices.  In general,the correspondence between edge and
  // curve indices can be arbitrary.  It is permitted for multiple
  // edges to use different portions of the same 3d curve.  The 
  // orientation of the edge always agrees with the natural 
  // parametric orientation of the curve.
  
  // edge that runs from A to B
  int ei = brep->m_E.Size();
  brep->m_E.Append(CRhinoBrepEdge(ei));
  CRhinoBrepEdge* edge = brep->m_E.Last();
  edge->m_c3i = AB; // index of curve from A to B;
  brep->m_C3[edge->m_c3i]->Domain( &edge->m_t[0], &edge->m_t[1] );
  edge->m_vi[0] = A; // index of vertex at A
  edge->m_vi[1] = B; // index of vertex at B
  edge->m_tolerance = 0.0; // this simple example is exact - for models with
                           // non-exact data, set tolerance as explained in
                           // definition of CRhinoBrepEdge.
  brep->m_V[edge->m_vi[0]].m_ei.Append(ei); // vertices need to know which edges use them
  brep->m_V[edge->m_vi[1]].m_ei.Append(ei);
  
  // edge that runs from B to C
  ei = brep->m_E.Size();
  brep->m_E.Append(CRhinoBrepEdge(ei));
  edge = brep->m_E.Last();
  edge->m_c3i = BC; // index of curve from B to C
  brep->m_C3[edge->m_c3i]->Domain( &edge->m_t[0], &edge->m_t[1] );
  edge->m_vi[0] = B; // index of vertex at B
  edge->m_vi[1] = C; // index of vertex at C
  edge->m_tolerance = 0.0;
  brep->m_V[edge->m_vi[0]].m_ei.Append(ei); // vertices need to know which edges use them
  brep->m_V[edge->m_vi[1]].m_ei.Append(ei);

  // edge that runs from C to D
  ei = brep->m_E.Size();
  brep->m_E.Append(CRhinoBrepEdge(ei));
  edge = brep->m_E.Last();
  edge->m_c3i = CD; // index of curve from C to D
  brep->m_C3[edge->m_c3i]->Domain( &edge->m_t[0], &edge->m_t[1] );
  edge->m_vi[0] = C; // index of vertex at C
  edge->m_vi[1] = D; // index of vertex at D
  edge->m_tolerance = 0.0;
  brep->m_V[edge->m_vi[0]].m_ei.Append(ei); // vertices need to know which edges use them
  brep->m_V[edge->m_vi[1]].m_ei.Append(ei);

  // edge that runs from A to D
  ei = brep->m_E.Size();
  brep->m_E.Append(CRhinoBrepEdge(ei));
  edge = brep->m_E.Last();
  edge->m_c3i = AD; // index of curve from A to D
  brep->m_C3[edge->m_c3i]->Domain( &edge->m_t[0], &edge->m_t[1] );
  edge->m_vi[0] = A; // index of vertex at A
  edge->m_vi[1] = D; // index of vertex at D
  edge->m_tolerance = 0.0;
  brep->m_V[edge->m_vi[0]].m_ei.Append(ei); // vertices need to know which edges use them
  brep->m_V[edge->m_vi[1]].m_ei.Append(ei);

  // edge that runs from E to F
  ei = brep->m_E.Size();
  brep->m_E.Append(CRhinoBrepEdge(ei));
  edge = brep->m_E.Last();
  edge->m_c3i = EF; // index of curve from E to F
  brep->m_C3[edge->m_c3i]->Domain( &edge->m_t[0], &edge->m_t[1] );
  edge->m_vi[0] = E; // index of vertex at E
  edge->m_vi[1] = F; // index of vertex at F
  edge->m_tolerance = 0.0;
  brep->m_V[edge->m_vi[0]].m_ei.Append(ei); // vertices need to know which edges use them
  brep->m_V[edge->m_vi[1]].m_ei.Append(ei);

  // edge that runs from F to G
  ei = brep->m_E.Size();
  brep->m_E.Append(CRhinoBrepEdge(ei));
  edge = brep->m_E.Last();
  edge->m_c3i = FG; // index of curve from F to G
  brep->m_C3[edge->m_c3i]->Domain( &edge->m_t[0], &edge->m_t[1] );
  edge->m_vi[0] = F; // index of vertex at F
  edge->m_vi[1] = G; // index of vertex at G
  edge->m_tolerance = 0.0;
  brep->m_V[edge->m_vi[0]].m_ei.Append(ei); // vertices need to know which edges use them
  brep->m_V[edge->m_vi[1]].m_ei.Append(ei);

  // edge that runs from G to H
  ei = brep->m_E.Size();
  brep->m_E.Append(CRhinoBrepEdge(ei));
  edge = brep->m_E.Last();
  edge->m_c3i = GH; // index of curve from G to H
  brep->m_C3[edge->m_c3i]->Domain( &edge->m_t[0], &edge->m_t[1] );
  edge->m_vi[0] = G; // index of vertex at G
  edge->m_vi[1] = H; // index of vertex at H
  edge->m_tolerance = 0.0;
  brep->m_V[edge->m_vi[0]].m_ei.Append(ei); // vertices need to know which edges use them
  brep->m_V[edge->m_vi[1]].m_ei.Append(ei);

  // edge that runs from E to H
  ei = brep->m_E.Size();
  brep->m_E.Append(CRhinoBrepEdge(ei));
  edge = brep->m_E.Last();
  edge->m_c3i = EH; // index of curve from E to H
  brep->m_C3[edge->m_c3i]->Domain( &edge->m_t[0], &edge->m_t[1] );
  edge->m_vi[0] = E; // index of vertex at E
  edge->m_vi[1] = H; // index of vertex at H
  edge->m_tolerance = 0.0;
  brep->m_V[edge->m_vi[0]].m_ei.Append(ei); // vertices need to know which edges use them
  brep->m_V[edge->m_vi[1]].m_ei.Append(ei);

  // edge that runs from A to E
  ei = brep->m_E.Size();
  brep->m_E.Append(CRhinoBrepEdge(ei));
  edge = brep->m_E.Last();
  edge->m_c3i = AE; // index of curve from A to E
  brep->m_C3[edge->m_c3i]->Domain( &edge->m_t[0], &edge->m_t[1] );
  edge->m_vi[0] = A; // index of vertex at A
  edge->m_vi[1] = E; // index of vertex at E
  edge->m_tolerance = 0.0;
  brep->m_V[edge->m_vi[0]].m_ei.Append(ei); // vertices need to know which edges use them
  brep->m_V[edge->m_vi[1]].m_ei.Append(ei);

  // edge that runs from B to F
  ei = brep->m_E.Size();
  brep->m_E.Append(CRhinoBrepEdge(ei));
  edge = brep->m_E.Last();
  edge->m_c3i = BF; // index of curve from B to F
  brep->m_C3[edge->m_c3i]->Domain( &edge->m_t[0], &edge->m_t[1] );
  edge->m_vi[0] = B; // index of vertex at B
  edge->m_vi[1] = F; // index of vertex at F
  edge->m_tolerance = 0.0;
  brep->m_V[edge->m_vi[0]].m_ei.Append(ei); // vertices need to know which edges use them
  brep->m_V[edge->m_vi[1]].m_ei.Append(ei);

  // edge that runs from C to G
  ei = brep->m_E.Size();
  brep->m_E.Append(CRhinoBrepEdge(ei));
  edge = brep->m_E.Last();
  edge->m_c3i = CG; // index of curve from C to G
  brep->m_C3[edge->m_c3i]->Domain( &edge->m_t[0], &edge->m_t[1] );
  edge->m_vi[0] = C; // index of vertex at C
  edge->m_vi[1] = G; // index of vertex at G
  edge->m_tolerance = 0.0;
  brep->m_V[edge->m_vi[0]].m_ei.Append(ei); // vertices need to know which edges use them
  brep->m_V[edge->m_vi[1]].m_ei.Append(ei);

  // edge that runs from D to H
  ei = brep->m_E.Size();
  brep->m_E.Append(CRhinoBrepEdge(ei));
  edge = brep->m_E.Last();
  edge->m_c3i = DH; // index of curve from D to H
  brep->m_C3[edge->m_c3i]->Domain( &edge->m_t[0], &edge->m_t[1] );
  edge->m_vi[0] = D; // index of vertex at D
  edge->m_vi[1] = H; // index of vertex at H
  edge->m_tolerance = 0.0;
  brep->m_V[edge->m_vi[0]].m_ei.Append(ei); // vertices need to know which edges use them
  brep->m_V[edge->m_vi[1]].m_ei.Append(ei);

}

static int MakeTwistedCubeTrimmingLoop(  CRhinoBrep* brep, // returns index of loop
     int fi,      // index of face
     int si,      // index of 3d surface
     int vSWi, int vSEi, int vNEi, int vNWi, // Indices of corner vertices listed in SW,SE,NW,NE order
     int eSi,     // index of edge on south side of surface
     int eS_dir,  // orientation of edge with respect to surface trim
     int eEi,     // index of edge on south side of surface
     int eE_dir,  // orientation of edge with respect to surface trim
     int eNi,     // index of edge on south side of surface
     int eN_dir,  // orientation of edge with respect to surface trim
     int eWi,     // index of edge on south side of surface
     int eW_dir   // orientation of edge with respect to surface trim
                                )
{
  const int li = brep->m_L.Size();       // index of loop we will create
  brep->m_L.Append(CRhinoBrepLoop(li));  // create empty loop with this index

  CRhinoBrepLoop& loop = brep->m_L[li];
  loop.m_type = CRhinoBrepLoop::outer;  // this loop is an outer trimming loop.
  loop.m_fi = fi;                       // index of face that uses this loop

  // Create trimming curves running counter clockwise around the surface's domain.
  // Start at the south side
  for ( int side = 0; side < 4; side++ ) {
    // side: 0=south, 1=east, 2=north, 3=west
    CRhinoAgCurve* c2 = TwistedCubeTrimmingCurve( *brep->m_S[si], side );
    const int c2i = brep->m_C2.Size();
    brep->m_C2.Append(c2);
    const int ti = brep->m_T.Size();
    brep->m_T.Append(CRhinoBrepTrim(ti));
    CRhinoBrepTrim& trim = brep->m_T[ti];
    trim.m_c2i = c2i;
    c2->Domain( &trim.m_t[0], &trim.m_t[1] ); // In this simple example, the
                                              // trim uses the entire 2d curve.
    trim.m_type = CRhinoBrepTrim::mated; // This b-rep is closed, so all trims
                                         // have mates.
    trim.m_li = li; // index of loop that uses this trim
    trim.m_tolerance[0] = 0.0; // This simple example is exact - for models with
    trim.m_tolerance[1] = 0.0; // non-exact data, set tolerance as explained in
                               // definition of CRhinoBrepTrim.
    
    switch ( side ) {
    case 0: // south
      trim.m_ei = eSi;
      trim.m_bRev3d = (eS_dir == -1);
      trim.m_vi[0] = vSWi;
      trim.m_vi[1] = vSEi;
      trim.m_iso = CRhinoBrepTrim::S_iso;
      break;
    case 1: // east
      trim.m_ei = eEi;
      trim.m_bRev3d = (eE_dir == -1);
      trim.m_vi[0] = vSEi;
      trim.m_vi[1] = vNEi;
      trim.m_iso = CRhinoBrepTrim::E_iso;
      break;
    case 2: // north
      trim.m_ei = eNi;
      trim.m_bRev3d = (eN_dir == -1);
      trim.m_vi[0] = vNEi;
      trim.m_vi[1] = vNWi;
      trim.m_iso = CRhinoBrepTrim::N_iso;
      break;
    case 3: // west
      trim.m_ei = eWi;
      trim.m_bRev3d = (eW_dir == -1);
      trim.m_vi[0] = vNWi;
      trim.m_vi[1] = vSWi;
      trim.m_iso = CRhinoBrepTrim::W_iso;
      break;
    }

    brep->m_E[trim.m_ei].m_ti.Append(ti); // edges need to know which trims
                                          // use them

    loop.m_ti.Append(ti); // loop's m_ti[] array is a list of the trims
                          // that make up the loop
  }

  return li;
}

static void MakeTwistedCubeFace( CRhinoBrep* brep,
     int si,      // index of 3d surface
     int s_dir,   // orientation of surface with respect to brep
     int vSWi, int vSEi, int vNEi, int vNWi, // Indices of corner vertices listed in SW,SE,NW,NE order
     int eSi,     // index of edge on south side of surface
     int eS_dir,  // orientation of edge with respect to surface trim
     int eEi,     // index of edge on south side of surface
     int eE_dir,  // orientation of edge with respect to surface trim
     int eNi,     // index of edge on south side of surface
     int eN_dir,  // orientation of edge with respect to surface trim
     int eWi,     // index of edge on south side of surface
     int eW_dir   // orientation of edge with respect to surface trim
                                )
{
  const int fi = brep->m_F.Size();        // index of this face
  brep->m_F.Append( CRhinoBrepFace(fi) ); // create new empty face with this index

  // fill in the face's fields
  CRhinoBrepFace& face = brep->m_F[fi];

  int fli = MakeTwistedCubeTrimmingLoop( brep, fi, si, vSWi, vSEi, vNEi, vNWi, 
                eSi, eS_dir, eEi, eE_dir, eNi, eN_dir, eWi, eW_dir );

  face.m_li.Append(fli); // fli = index of outer trimming loop
  face.m_si = si;
  face.m_bRev = (s_dir == -1);
}

static void MakeTwistedCubeFaces( CRhinoBrep* brep )
{

  MakeTwistedCubeFace( brep,
    ABCD,       // Index of surface ABCD
    +1,         // orientation of surface with respect to brep
    A, B, C, D, // Indices of vertices listed in SW,SE,NW,NE order
    AB,+1,      // South side edge and its orientation with respect to
                // to the trimming curve.  (AB)
    BC,+1,      // South side edge and its orientation with respect to
                // to the trimming curve.  (BC)
    CD,+1,      // South side edge and its orientation with respect to
                // to the trimming curve   (CD)
    AD,-1       // South side edge and its orientation with respect to
                // to the trimming curve   (AD)
    );

  MakeTwistedCubeFace( brep,
    BCGF,       // Index of surface BCGF
    -1,         // orientation of surface with respect to brep
    B, C, G, F, // Indices of vertices listed in SW,SE,NW,NE order
    BC,+1,      // South side edge and its orientation with respect to
                // to the trimming curve.  (BC)
    CG,+1,      // South side edge and its orientation with respect to
                // to the trimming curve.  (CG)
    FG,-1,      // South side edge and its orientation with respect to
                // to the trimming curve   (FG)
    BF,-1       // South side edge and its orientation with respect to
                // to the trimming curve   (BF)
    );

  MakeTwistedCubeFace( brep,
    CDHG,       // Index of surface CDHG
    -1,         // orientation of surface with respect to brep
    C, D, H, G, // Indices of vertices listed in SW,SE,NW,NE order
    CD,+1,      // South side edge and its orientation with respect to
                // to the trimming curve.  (CD)
    DH,+1,      // South side edge and its orientation with respect to
                // to the trimming curve.  (DH)
    GH,-1,      // South side edge and its orientation with respect to
                // to the trimming curve   (GH)
    CG,-1       // South side edge and its orientation with respect to
                // to the trimming curve   (CG)
    );

  MakeTwistedCubeFace( brep,
    ADHE,       // Index of surface ADHE
    +1,         // orientation of surface with respect to brep
    A, D, H, E, // Indices of vertices listed in SW,SE,NW,NE order
    AD,+1,      // South side edge and its orientation with respect to
                // to the trimming curve.  (AD)
    DH,+1,      // South side edge and its orientation with respect to
                // to the trimming curve.  (DH)
    EH,-1,      // South side edge and its orientation with respect to
                // to the trimming curve   (EH)
    AE,-1       // South side edge and its orientation with respect to
                // to the trimming curve   (AE)
    );

  MakeTwistedCubeFace( brep,
    ABFE,       // Index of surface ABFE
    -1,         // orientation of surface with respect to brep
    A, B, F, E, // Indices of vertices listed in SW,SE,NW,NE order
    AB,+1,      // South side edge and its orientation with respect to
                // to the trimming curve.  (AB)
    BF,+1,      // South side edge and its orientation with respect to
                // to the trimming curve.  (BF)
    EF,-1,      // South side edge and its orientation with respect to
                // to the trimming curve   (EF)
    AE,-1       // South side edge and its orientation with respect to
                // to the trimming curve   (AE)
    );

  MakeTwistedCubeFace( brep,
    EFGH,       // Index of surface EFGH
    -1,         // orientation of surface with respect to brep
    E, F, G, H, // Indices of vertices listed in SW,SE,NW,NE order
    EF,+1,      // South side edge and its orientation with respect to
                // to the trimming curve.  (EF)
    FG,+1,      // South side edge and its orientation with respect to
                // to the trimming curve.  (FG)
    GH,+1,      // South side edge and its orientation with respect to
                // to the trimming curve   (GH)
    EH,-1       // South side edge and its orientation with respect to
                // to the trimming curve   (EH)
    );
}


static CRhinoBrep* MakeTwistedCube()
{
  // This example demonstrates how to construct a CRhinoBrep
  // with the topology shown below.
  //
  //
  //             H-------e6-------G
  //            /                /|
  //           / |              / |
  //          /  e7            /  e5
  //         /   |            /   |
  //        /                e10  |
  //       /     |          /     |
  //      e11    E- - e4- -/- - - F
  //     /                /      /
  //    /      /         /      /
  //   D---------e2-----C      e9
  //   |     /          |     /
  //   |    e8          |    /
  //   e3  /            e1  /
  //   |                |  /
  //   | /              | /
  //   |                |/
  //   A-------e0-------B
  //
  //

  double point[8][3] = {
    {  0.0,  0.0,  0.0 },  // point A = geometry for vertex 0
    { 10.0,  0.0,  0.0 },  // point B = geometry for vertex 1
    { 10.0,  8.0, -1.0 },  // point C = geometry for vertex 2
    {  0.0,  6.0,  0.0 },  // point D = geometry for vertex 3
    {  1.0,  2.0, 11.0 },  // point E = geometry for vertex 4
    { 10.0,  0.0, 12.0 },  // point F = geometry for vertex 5
    { 10.0,  7.0, 13.0 },  // point G = geometry for vertex 6
    {  0.0,  6.0, 12.0 },  // point H = geometry for vertex 7
  };

  CRhinoBrep* brep = new CRhinoBrep();

  // create eight vertices located at the eight points
  int vi;
  for ( vi = 0; vi < 8; vi++ ) {
    CRhinoBrepVertex v(vi); // Constructor take vertex index or it can be set later.
                            // The index must be equal to the vertex's position in
                            // in the CRhinoBrep::m_V[] array                           
    v.x = point[vi][0];
    v.y = point[vi][1];
    v.z = point[vi][2];
    v.m_tolerance = 0.0; // this simple example is exact - for models with
                         // non-exact data, set tolerance as explained in
                         // definition of CRhinoBrepVertex.
    brep->m_V.Append(v);
  }

  // Create 3d curve geometry - the orientations are arbitrarily chosen
  // so that the end vertices are in alphabetical order.
  brep->m_C3.Append( TwistedCubeEdgeCurve( point[A], point[B] ) ); // line AB
  brep->m_C3.Append( TwistedCubeEdgeCurve( point[B], point[C] ) ); // line BC
  brep->m_C3.Append( TwistedCubeEdgeCurve( point[C], point[D] ) ); // line CD
  brep->m_C3.Append( TwistedCubeEdgeCurve( point[A], point[D] ) ); // line AD
  brep->m_C3.Append( TwistedCubeEdgeCurve( point[E], point[F] ) ); // line EF
  brep->m_C3.Append( TwistedCubeEdgeCurve( point[F], point[G] ) ); // line FG
  brep->m_C3.Append( TwistedCubeEdgeCurve( point[G], point[H] ) ); // line GH
  brep->m_C3.Append( TwistedCubeEdgeCurve( point[E], point[H] ) ); // line EH
  brep->m_C3.Append( TwistedCubeEdgeCurve( point[A], point[E] ) ); // line AE
  brep->m_C3.Append( TwistedCubeEdgeCurve( point[B], point[F] ) ); // line BF
  brep->m_C3.Append( TwistedCubeEdgeCurve( point[C], point[G] ) ); // line CG
  brep->m_C3.Append( TwistedCubeEdgeCurve( point[D], point[H] ) ); // line DH

  // Create the 12 edges that connect the corners of the cube.
  MakeTwistedCubeEdges( brep );

  // Create 3d surface geometry - the orientations are arbitrarily chosen so
  // that some normals point into the cube and others point out of the cube.
  brep->m_S.Append( TwistedCubeSideSurface( point[A], point[B], point[C], point[D] ) ); // ABCD
  brep->m_S.Append( TwistedCubeSideSurface( point[B], point[C], point[G], point[F] ) ); // BCGF
  brep->m_S.Append( TwistedCubeSideSurface( point[C], point[D], point[H], point[G] ) ); // CDHG
  brep->m_S.Append( TwistedCubeSideSurface( point[A], point[D], point[H], point[E] ) ); // ADHE
  brep->m_S.Append( TwistedCubeSideSurface( point[A], point[B], point[F], point[E] ) ); // ABFE
  brep->m_S.Append( TwistedCubeSideSurface( point[E], point[F], point[G], point[H] ) ); // EFGH


  // Create the CRhinoBrepFaces
  MakeTwistedCubeFaces( brep );

  return brep;
}

int main( int argc, const char* argv[] )
{
  // Before working through this example, you should understand
  // the example_write.cpp example.

  FILE* pFile = fopen( "my_brep.3dm", "wb" );

  try {
    // pFile must be a file pointer returned from fopen( , "wb" );
    CRhinoFile RF( pFile, CRhinoFile::write );

    // (optional) Specify some or all of unit system and tolerances
    RF.SetUnitSystem( CRhinoFile::inches );
    RF.SetAbsoluteTolerance( 0.01 );
    RF.SetAngleTolerance( 3.141592653589793238462643*1.5/180.0 );

    // (optional) create custom layer(s).  The CRhinoLayer::New() call
    // automatically adds the layers to the Rhino file's layer table.
    CRhinoLayerPtr LayerPtr = 
      CRhinoLayer::New( "Custom Breps",             // text name
                        CRhinoColor( 0, 255, 255 )  // wire display RGB color
                        );

    CRhinoBrep* brep = MakeTwistedCube(); // create a b-rep

    if ( brep->IsValid() ) {
      BOOL bIsSolid, bIsManifold, bIsOriented, bHasBoundary;
      // brep->IsValid(), brep->IsSolid(), and brep->IsManifold()
      // are useful debugging tools.
      bIsManifold = brep->IsManifold(&bIsOriented,&bHasBoundary);
      bIsSolid = brep->IsSolid();

      brep->SetLayer( LayerPtr ); // (optional) specify object's layer
      brep->SetLabel( "my twisted cube" ); // (optional) specify object's name
      brep->Write( RF );
    }

    delete brep;
  }

  catch( CRhinoException e ) {
    // if something goes wrong, catch the exception her
    const char* s;
    switch ( e.m_type ) {
    case CRhinoException::unable_to_write:
      s = "unable_to_write";
      break;
    case CRhinoException::unable_to_read:
      s = "unable_to_read";
      break;
    case CRhinoException::unable_to_seek:
      s = "unable_to_seek";
      break;
    case CRhinoException::unexpected_end_of_file:
      s = "unexpected_end_of_file";
      break;
    case CRhinoException::unexpected_value:
      s = "unexpected_value";
      break;
    case CRhinoException::not_supported:
      s = "not_supported";
      break;
    case CRhinoException::illegal_input:
      s = "illegal_input";
      break;
    case CRhinoException::out_of_memory:
      s = "out_of_memory";
      break;
    default:
      s = "unknow exception";
      break;
    }

    printf("Trouble writing file.  CRhinoException type: %s\n", s);
  }

  fclose( pFile );

  return 0;
}
