/* $Header: /Src/RhinoIO/example_dump/example_dump.c 12    1/07/00 9:20a Dalelear $ */
/* $NoKeywords: $ */

/*
//
// Copyright (c) 1993-1998 Robert McNeel & Associates. All rights reserved.
// Rhinoceros is a registered trademark of Robert McNeel & Assoicates.
//
// THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
// ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
// MERCHANTABILITY ARE HEREBY DISCLAIMED.
//				
////////////////////////////////////////////////////////////////
*/

/*
This example explains the low level structure of a Rhino .3dm file.
This example is included for purely pedagogical reasons.  Nothing 
in here is of any real use if all you want to do is quickly add 
Rhino file reading or writing to your program.
*/

/*
// Rhino .3DM files are binary files.  This toolkit assumes the following table
// describes the char, short, int, float and double types.
// 
//     type    size
//     char    1 byte
//     short   2 bytes     little endian byte order
//     int     4 bytes     little endian byte order
//     void*   4 bytes
//     float   4 bytes     little endian byte order IEEE float
//     double  8 bytes     little endian byte order IEEE double
// 
// (This toolkit is designed to work correctly on both little and
// big endean processors.)
// 
// A Rhino .3DM file has a 32 byte version ID followed by a list of chunks.
// There are two kinds of chunks, integer chunks and data chunks.  Integer
// chunks have the form
// 
//   int   typecode  ( the high bit 0x80000000 must be set ) 
//   int   value
// 
//   int   typecode  ( the high bit 0x80000000 must be clear ) 
//   int   length    ( can be zero )
//     length bytes of data
// 
// The first chunk after the 32 byte version idea must be a comment
// chunk.  Comments have typecode = 1.  The data in comment chunks
// is always ignored.  Usually comment chunks contain ASCII text
// but this is not required.  After the initial comment chunk, any
// type of chunk may appear.  Thus a Rhino .3DM file looks like
// 
//   32 byte version id
//     ( currently = "3D Geometry File Format        1" )
// 
//   DATA CHUNK with typecode = 1
//     
// 
//   CHUNK
// 
//   ...
// 
//   CHUNK
// 
//   EOF
// 
// Any data chunk can contain subchunks.  Thus a Rhino file can be
// read (but not understood) by the following code.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "3dm.h"

typedef int BOOL;
#define TRUE 1;
#define FALSE 0;

static const char* TypeCodeCategoryString( int /*tcode*/ );
static const char* TypeCodeString( int /*tcode*/ );

BOOL ReadRhinoFileInt( 
        FILE*,  /* pointer to opened binary file */
        BOOL,   /* TRUE if computer uses little endean byte order */
        int*    /* OUTPUT: value of integer read from file */
        );
BOOL ReadRhinoFile( 
        const char* /* full name of file */
        );

int main ( int argc, const char* argv[] )
{
  int argi;
  if ( argc < 2 ) {
    printf("Syntax: %s rhino_file_name, ...\n",argv[0]);
    return 0;
  }
  for ( argi = 1; argi < argc; argi++ ) {
    printf( "\nReading file %s\n", argv[argi] );
    ReadRhinoFile( argv[argi] );
    printf( "\n" );
  }
  /* getchar(); // uncomment when debugging */
  return 0;
}

BOOL ReadRhinoFile( const char* sFileName )
{
  /* Returns TRUE if successful, FALSE if something went wrong; */
  BOOL bFinishedFile;
  BOOL bLittleEndean;
  char sVersion[33];
  int chunk_count;
  int chunk_typecode;
  int chunk_value;
  FILE* pFile;
  const char* sFormat = NULL;

  bFinishedFile  = FALSE;
  bLittleEndean  = TRUE;
  sVersion[0]    = 0;
  chunk_count    = 0;
  chunk_typecode = 0;
  chunk_value    = 0;
  pFile          = NULL;


  /* make sure types have correct sizes */
  if ( sizeof(char) != 1 )
    return FALSE;
  if ( sizeof(short) != 2 )
    return FALSE;
  if ( sizeof(int) != 4 )
    return FALSE;
  if ( sizeof(float) != 4 )
    return FALSE;
  if ( sizeof(double) != 8 )
    return FALSE;

  /* determine endean-ness; */
  {
    union {
      int i;
      char c[4];
    } u;
    u.i    = 0;
    u.c[0] = 1;
    if      ( u.i == 1 ) {
      bLittleEndean = TRUE;
    }
    else if ( u.i == 0x01000000 ) {
      bLittleEndean = FALSE;
    }
    else
      return FALSE;
  }

  /* open file as a binary file */
  pFile = fopen( sFileName, "rb" ); 
  if ( !pFile )
    return FALSE;

  if ( 1 != fread( sVersion, 32, 1, pFile ) ) {
    fclose( pFile );
    return FALSE;
  }
  sVersion[32] = 0;
  if ( strncmp( sVersion, "3D Geometry File Format ", 24 ) ) {
    fclose( pFile );
    printf("Initial 32 bytes are incorrect.\n");
    return FALSE;
  }

  printf("Rhino file version: %s\n", sVersion );

  /* rest of file consists of chunks */
  for (;;) {
    chunk_typecode = 0;
    chunk_value = 0;
    if ( !ReadRhinoFileInt( pFile, bLittleEndean, &chunk_typecode ) ) {
      /* we should be at the end of the file */
      if ( feof( pFile ) ) {
        bFinishedFile = TRUE;
      }
      break;
    }
    if ( !ReadRhinoFileInt( pFile, bLittleEndean, &chunk_value ) )
      break;

    chunk_count++;

    if ( chunk_typecode & 0x80000000 ) {
      /* integer chunk */
      sFormat = "Chunk %6d: type =  %08x  integer value = %d\n"
                "                   =  (%s:%s)\n";
    }
    else {
      /* data chunk */
      sFormat = "Chunk %6d: type =  %08x data length = %d bytes\n"
                "                   =  (%s:%s)\n";
    }
    printf( sFormat,
            chunk_count, 
            chunk_typecode, 
            chunk_value,
            TypeCodeCategoryString(chunk_typecode),
            TypeCodeString(chunk_typecode)
            );
    /* skip to end of data chunk */
    if ( !(chunk_typecode & 0x80000000) ) {
      if ( chunk_value < 0 ) {
        printf("Data chunks must have non-negative lengths.\n");
        break;
      }
      if ( fseek( pFile, chunk_value, SEEK_CUR ) )
        break;
    }
    if ( chunk_count == 1 ) {
      if ( chunk_typecode != 1 ) {
        printf("The first chunk must have typecode = 0x00000001\n");
        break;
      }
    }
  }

  fclose( pFile );

  if ( !bFinishedFile ) {
    printf("%s is not a valid Rhino .3dm file.\n",sFileName);
  }
  else {
    printf("Read %d chunks from %s\n",chunk_count,sFileName);
  }

  return bFinishedFile;
}

BOOL ReadRhinoFileInt( FILE* pFile, BOOL bLittleEndean, int* p )
{
  union {
    int i;
    char c[4];
  } u;
  char c[4];
  BOOL rc = FALSE;
  u.i = 0;
  if ( bLittleEndean ) {
    rc = ( 1 == fread( &u.i, sizeof(u.i), 1, pFile ) );
  }
  else {
    if ( 1 == fread( c, sizeof(u.i), 1, pFile ) ){
      u.c[0] = c[3];
      u.c[1] = c[2];
      u.c[2] = c[1];
      u.c[3] = c[0];
      rc = TRUE;
    }
  }
  if ( rc )
    *p = u.i;
  return rc;
}

static const char* TypeCodeCategoryString( int tcode )
{
  const char* s;
  switch( tcode & 0x7FFF0000 ) {
  case TCODE_AG:
    s ="TCODE_AG          ";
    break;
  case TCODE_RHINO_OBJECT:
    s ="TCODE_RHINO_OBJECT";
    break;
  case TCODE_GEOMETRY:
    s ="TCODE_GEOMETRY    ";
    break;
  case TCODE_DISPLAY:
    s ="TCODE_DISPLAY     ";
    break;
  case TCODE_RENDER:
    s ="TCODE_RENDER      ";
    break;
  case TCODE_INTERFACE:
    s ="TCODE_INTERFACE   ";
    break;
  case TCODE_TOLERANCE:
    s ="TCODE_TOLERANCE   ";
    break;
  case TCODE_TABLE:
    s ="TCODE_TABLE       ";
    break;
  case TCODE_TABLEREF:
    s ="TCODE_TABLEREF    ";
    break;
  case TCODE_USER:
    s ="TCODE_USER        ";
    break;
  case TCODE_SHORT:
    s ="TCODE_SHORT       ";
    break;
  default:
    s ="unknown category";
    break;
  }
  return s;
}



static const char* TypeCodeString( int tcode )
{
  const char* s;
  switch( tcode ) {
  case TCODE_RHINO_OBJECT_DATA:
    s = "TCODE_RHINO_OBJECT_DATA"; break;
  case TCODE_RHINO_OBJECT_END:
    s = "TCODE_RHINO_OBJECT_END"; break;
  case TCODE_RHINO_OBJECT_NURBS_CURVE:
    s = "TCODE_RHINO_OBJECT_NURBS_CURVE"; break;
  case TCODE_RHINO_OBJECT_NURBS_SURFACE:
    s = "TCODE_RHINO_OBJECT_NURBS_SURFACE"; break;
  case TCODE_RHINO_OBJECT_BREP:
    s = "TCODE_RHINO_OBJECT_BREP"; break;
  case TCODE_AG_ASSEMBLY:
    s = "TCODE_AG_ASSEMBLY"; break;
  case TCODE_AG_PART:
    s = "TCODE_AG_PART"; break;
  case TCODE_AG_SHELL:
    s = "TCODE_AG_SHELL"; break;
  case TCODE_AG_FACE:
    s = "TCODE_AG_FACE"; break;
  case TCODE_AG_BOUNDARY:
    s = "TCODE_AG_BOUNDARY"; break;
  case TCODE_AG_TEDGE:
    s = "TCODE_AG_TEDGE"; break;
  case TCODE_AG_SURFACE:
    s = "TCODE_AG_SURFACE"; break;
  case TCODE_AG_CURVE:
    s = "TCODE_AG_CURVE"; break;
  case TCODE_AG_SPLINE:
    s = "TCODE_AG_SPLINE"; break;
  case TCODE_AG_POINT:
    s = "TCODE_AG_POINT"; break;
  case TCODE_STUFF:
    s = "TCODE_STUFF"; break;
  case TCODE_AG_ASSEMBLYSTUFF:
    s = "TCODE_AG_ASSEMBLYSTUFF"; break;
  case TCODE_AG_PARTSTUFF:
    s = "TCODE_AG_PARTSTUFF"; break;
  case TCODE_AG_SHELLSTUFF:
    s = "TCODE_AG_SHELLSTUFF"; break;
  case TCODE_AG_FACESTUFF:
    s = "TCODE_AG_FACESTUFF"; break;
  case TCODE_AG_BOUNDARYSTUFF:
    s = "TCODE_AG_BOUNDARYSTUFF"; break;
  case TCODE_AG_TEDGESTUFF:
    s = "TCODE_AG_TEDGESTUFF"; break;
  case TCODE_AG_SURFACESTUFF:
    s = "TCODE_AG_SURFACESTUFF"; break;
  case TCODE_AG_CURVESTUFF:
    s = "TCODE_AG_CURVESTUFF"; break;
  case TCODE_AG_SPLINESTUFF:
    s = "TCODE_AG_SPLINESTUFF"; break;
  case TCODE_AG_POINTSTUFF:
    s = "TCODE_AG_POINTSTUFF"; break;
  case TCODE_RH_POINT:
    s = "TCODE_RH_POINT"; break;
  case TCODE_RH_SPOTLIGHT:
    s = "TCODE_RH_SPOTLIGHT"; break;
  case TCODE_OLD_RH_TRIMESH:
    s = "TCODE_OLD_RH_TRIMESH"; break;
  case TCODE_OLD_MESH_VERTEX_NORMALS:
    s = "TCODE_OLD_MESH_VERTEX_NORMALS"; break;
  case TCODE_OLD_MESH_UV:
    s = "TCODE_OLD_MESH_UV"; break;
  case TCODE_OLD_FULLMESH			:
    s = "TCODE_OLD_FULLMESH			"; break;
  case TCODE_MESH_OBJECT:
    s = "TCODE_MESH_OBJECT"; break;
  case TCODE_COMPRESSED_MESH_GEOMETRY:
    s = "TCODE_COMPRESSED_MESH_GEOMETRY"; break;
  case TCODE_NAME:
    s = "TCODE_NAME"; break;
  case TCODE_VIEW:
    s = "TCODE_VIEW"; break;
  case TCODE_CPLANE:
    s = "TCODE_CPLANE"; break;
  case TCODE_NAMED_CPLANE:
    s = "TCODE_NAMED_CPLANE"; break;
  case TCODE_NAMED_VIEW:
    s = "TCODE_NAMED_VIEW"; break;
  case TCODE_VIEWPORT:
    s = "TCODE_VIEWPORT"; break;
  case TCODE_SHOWGRID:
    s = "TCODE_SHOWGRID"; break;
  case TCODE_SHOWGRIDAXES:
    s = "TCODE_SHOWGRIDAXES"; break;
  case TCODE_SHOWWORLDAXES:
    s = "TCODE_SHOWWORLDAXES"; break;
  case TCODE_VIEWPORT_POSITION:
    s = "TCODE_VIEWPORT_POSITION"; break;
  case TCODE_VIEWPORT_TRACEINFO:
    s = "TCODE_VIEWPORT_TRACEINFO"; break;
  case TCODE_SNAPSIZE			:
    s = "TCODE_SNAPSIZE			"; break;
  case TCODE_NEAR_CLIP_PLANE:
    s = "TCODE_NEAR_CLIP_PLANE"; break;
  case TCODE_HIDE_TRACE:
    s = "TCODE_HIDE_TRACE"; break;
  case TCODE_NOTES:
    s = "TCODE_NOTES"; break;
  case TCODE_LAYERTABLE:
    s = "TCODE_LAYERTABLE"; break;
  case TCODE_LAYERREF:
    s = "TCODE_LAYERREF"; break;
  case TCODE_BITMAPPREVIEW:
    s = "TCODE_BITMAPPREVIEW"; break;
  case TCODE_XDATA:
    s = "TCODE_XDATA"; break;
  case TCODE_RGB:
    s = "TCODE_RGB"; break;
  case TCODE_TEXTUREMAP:
    s = "TCODE_TEXTUREMAP"; break;
  case TCODE_BUMPMAP:
    s = "TCODE_BUMPMAP"; break;
  case TCODE_TRANSPARENCY:
    s = "TCODE_TRANSPARENCY"; break;
  case TCODE_DISP_AM_RESOLUTION:
    s = "TCODE_DISP_AM_RESOLUTION"; break;
  case TCODE_RGBDISPLAY:
    s = "TCODE_RGBDISPLAY"; break;
  case TCODE_RENDER_MATERIAL_ID:
    s = "TCODE_RENDER_MATERIAL_ID"; break;
  case TCODE_LAYER:
    s = "TCODE_LAYER"; break;
  case TCODE_LAYER_OBSELETE_1:
    s = "TCODE_LAYER_OBSELETE_1"; break;
  case TCODE_LAYER_OBSELETE_2:
    s = "TCODE_LAYER_OBSELETE_2"; break;
  case TCODE_LAYER_OBSELETE_3:
    s = "TCODE_LAYER_OBSELETE_3"; break;
  case TCODE_LAYERON:
    s = "TCODE_LAYERON"; break;
  case TCODE_LAYERTHAWED:
    s = "TCODE_LAYERTHAWED"; break;
  case TCODE_LAYERLOCKED:
    s = "TCODE_LAYERLOCKED"; break;
  case TCODE_LAYERVISIBLE:
    s = "TCODE_LAYERVISIBLE"; break;
  case TCODE_LAYERPICKABLE:
    s = "TCODE_LAYERPICKABLE"; break;
  case TCODE_LAYERSNAPABLE:
    s = "TCODE_LAYERSNAPABLE"; break;
  case TCODE_LAYERRENDERABLE:
    s = "TCODE_LAYERRENDERABLE"; break;
  case TCODE_LAYERSTATE:
    s = "TCODE_LAYERSTATE"; break;
  case TCODE_DISP_CPLINES:
    s = "TCODE_DISP_CPLINES"; break;
  case TCODE_DISP_MAXLENGTH:
    s = "TCODE_DISP_MAXLENGTH"; break;
  case TCODE_CURRENTLAYER:
    s = "TCODE_CURRENTLAYER"; break;
  case TCODE_LAYERNAME:
    s = "TCODE_LAYERNAME"; break;
  case TCODE_TOL_AGFIT:
    s = "TCODE_TOL_AGFIT"; break;
  case TCODE_TOL_AGANGLE:
    s = "TCODE_TOL_AGANGLE"; break;
  case TCODE_ENDOFTABLE:
    s = "TCODE_ENDOFTABLE"; break;
  case TCODE_COMMENTBLOCK:
    s = "TCODE_COMMENTBLOCK"; break;
  case TCODE_SUMMARY:
    s = "TCODE_SUMMARY"; break;
  case TCODE_RENDERMESHPARAMS:
    s = "TCODE_RENDERMESHPARAMS"; break;
  case TCODE_UNIT_AND_TOLERANCES:
    s = "TCODE_UNIT_AND_TOLERANCES"; break;
  case TCODE_ANALYSIS_MESH:
    s = "TCODE_ANALYSIS_MESH"; break;
  case TCODE_ANNOTATION_SETTINGS:
    s = "TCODE_ANNOTATION_SETTINGS"; break;
  case TCODE_TEXT_BLOCK:
    s = "TCODE_TEXT_BLOCK"; break;
  case TCODE_ANNOTATION_LEADER:
    s = "TCODE_ANNOTATION_LEADER"; break;
  case TCODE_LINEAR_DIMENSION:
    s = "TCODE_LINEAR_DIMENSION"; break;
  case TCODE_ANGULAR_DIMENSION:
    s = "TCODE_ANGULAR_DIMENSION"; break;
  case TCODE_RADIAL_DIMENSION:
    s = "TCODE_RADIAL_DIMENSION"; break;
  default:
    s = "unknown typecode";
    break;
  }
  return s;
}
