/* $Header: /Src/RhinoIO/example_gl/example_gl.cpp 17    11/10/99 2:59p Dalelear $ */
/* $NoKeywords: $ */
//
// Copyright (c) 1993-1998 Robert McNeel & Associates. All rights reserved.
// Rhinoceros is a registered trademark of Robert McNeel & Assoicates.
//
// THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
// ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
// MERCHANTABILITY ARE HEREBY DISCLAIMED.
//				
////////////////////////////////////////////////////////////////

#include "rhinoio.h"    // RhinoIO class definitions
#include "rhinoio_gl.h" // RhinoGL() function declarations
#include <GL/GLaux.h>   // Open GL auxillary functions

#if defined(_WINDOWS) || defined(WIN32)
#define RHINO_GL_CALLBACK CALLBACK
#else
#define RHINO_GL_CALLBACK
#endif

// Before working on this file, be sure to study the 
// Rhino IO toolkit example_read.cpp and to read chapters 1 through 11
// of the _Open_GL_Programming_Guide_.
//
// This file contains simple example in modeled after those found in
// the _Open_GL_Programming_Guide_.  The nuts and bolts functions that
// demonstrate how to use Open GL to display Rhino geometry are in
// rhinoio_gl.cpp.

///////////////////////////////////////////////////////////////////////
//
// Dynamic arrays of geometry objects, spotlights, and viewports
//
//////////////////////////////////////////////////////////////////////

typedef CRhinoArray<CRhinoObject*>    CRhinoObjectArray;
typedef CRhinoArray<CRhinoSpotlight*> CRhinoSpotlightArray;
typedef CRhinoArray<CRhinoViewport*>  CRhinoViewportArray;

///////////////////////////////////////////////////////////////////////
//
// Globals for myDisplay() function passed to auxMainLoop()
//
//////////////////////////////////////////////////////////////////////

static GLuint glb_display_list_number = 1;         // GL display list "name"
static CRhinoSpotlightArray glb_spotlight(0,TRUE); // array of lights
static CRhinoViewport glb_viewport;                // viewport used to display geometry
static double         glb_bounding_box[2][3];      // bounding box of display geometry

///////////////////////////////////////////////////////////////////////
//
// Functions used in main()
//
//////////////////////////////////////////////////////////////////////

BOOL myInitGeometry( // reads Rhino file and saves contents in arrays
       int argc, const char *argv[],
       CRhinoFile*&,          // unused pointer to a CRhinoFile
       double[2][3],          // place to store bounding box
       CRhinoObjectArray&,    // place to store pointers to geometry
       CRhinoSpotlightArray&, // place to store pointers to spotlights
       CRhinoViewportArray&   // place to store pointers to viewports
       );

BOOL myInitGL( const CRhinoViewport&, GLUnurbsObj*& );

void myBuildDisplayList( 
      GLuint,                  // display_list_number,
      GLUnurbsObj*,            // pointer to GL nurbs render
      const CRhinoObjectArray& // Rhino geometry to render
      );

extern "C" {
void RHINO_GL_CALLBACK myNurbsErrorCallback( GLenum ); // for gluNurbsCallback()

void RHINO_GL_CALLBACK myDisplay( void );              // for auxMainLoop()

void RHINO_GL_CALLBACK myReshape( GLsizei, GLsizei );  // for auxReshapeFunc()

void RHINO_GL_CALLBACK myMouseLeftEvent( AUX_EVENTREC* );   // for auxMouseFunc();
void RHINO_GL_CALLBACK myMouseMiddleEvent( AUX_EVENTREC* ); // for auxMouseFunc();
void RHINO_GL_CALLBACK myMouseRightEvent( AUX_EVENTREC* );  // for auxMouseFunc();

void RHINO_GL_CALLBACK myKeyLeftArrowEvent( void );    // for auxKeyFunc();
void RHINO_GL_CALLBACK myKeyRightArrowEvent( void );   // for auxKeyFunc();
void RHINO_GL_CALLBACK myKeyUpArrowEvent( void );      // for auxKeyFunc();
void RHINO_GL_CALLBACK myKeyDownArrowEvent( void );    // for auxKeyFunc();
void RHINO_GL_CALLBACK myKeyViewExtents( void );       // for auxKeyFunc();

typedef void (CALLBACK* RHINO_GL_NURBS_ERROR)();
}

///////////////////////////////////////////////////////////////////////
//
// used to set projections
//
void SetGLModelViewMatrix( const CRhinoViewport& );
void SetGLProjectionMatrix( CRhinoViewport& );
///////////////////////////////////////////////////////////////////////

int main( int argc, const char *argv[] )
{
  int window_width  = 500;
  int window_height = 500;
  double port_aspect = ((double)window_width)/((double)window_height);

  // read file, put geometry in the object[] array, and get bounding box

  BOOL bOK = TRUE;
  CRhinoFile* pRF = NULL;
  CRhinoViewportArray  viewport(0,TRUE);
  CRhinoObjectArray object(0,TRUE);       // array of Rhino objects to display
  bOK = myInitGeometry( argc, argv, 
                        pRF,
                        glb_bounding_box,
                        object,
                        glb_spotlight,
                        viewport
                        );
  if ( !bOK )
    return 1;

  if ( viewport.Size() > 0 && viewport[0] ) {
    // use active viewport from Rhino file
    glb_viewport = *viewport[0];
    glb_viewport.SetFrustumNearFar( glb_bounding_box[0], glb_bounding_box[1] );
  }
  else {
    // simple parallel projection of bounding box;
    double dx, dy, dz, cx, cy, cz;
    double frus_near, frus_far;
    double camLoc[3], camDir[3], camUp[3];
    cx = 0.5*(glb_bounding_box[0][0] + glb_bounding_box[1][0]);
    cy = 0.5*(glb_bounding_box[0][1] + glb_bounding_box[1][1]);
    cz = 0.5*(glb_bounding_box[0][2] + glb_bounding_box[1][2]);
    dx = 1.1*(glb_bounding_box[1][0] - glb_bounding_box[0][0]);
    dy = 1.1*(glb_bounding_box[1][1] - glb_bounding_box[0][1]);
    dz = 1.1*(glb_bounding_box[1][2] - glb_bounding_box[0][2]);
    if ( dx <= 1.0e-6 && dy <= 1.0e-6 )
      dx = dy = 2.0;
    if ( window_height*dx < window_width*dy ) {
      dx = dy*window_width/window_height;
    }
    else {
      dy = dx*window_height/window_width;
    }
    if ( dz <= 0.1*(dx+dy) )
      dz = 0.1*(dx+dy);
    dx *= 0.5;
    dy *= 0.5;
    dz *= 0.5;


    frus_near = 1.0;
    frus_far = frus_near + 2.0*dz;
    camLoc[0] = cx;
    camLoc[1] = cy;
    camLoc[2] = cz + dz + frus_near;
    camDir[0] =  0.0; 
    camDir[1] =  0.0;
    camDir[2] = -1.0;
    camUp[0] = 0.0; 
    camUp[1] = 1.0; 
    camUp[2] = 0.0;

    glb_viewport.SetProjection( CRhinoViewport::parallel_proj );
    glb_viewport.SetCamera( camLoc, camDir, camUp );
    glb_viewport.SetFrustum( -dx, dx, -dy, dy, frus_near, frus_far );

    glb_viewport.SetViewName( "XY plan view" );
  }
  // If needed, enlarge frustum so its aspect matches the window's aspect.
  // Since the Rhino file does not store the far frustum distance in the
  // file, viewports read from a Rhil file need to have the frustum's far
  // value set by inspecting the bounding box of the geometry to be
  // displayed.

  
  ///////////////////////////////////////////////////////////////////
  //
  // GL stuff starts here
  //
  for(;;) {  
    
    char sWindowTitleString[256];
    sWindowTitleString[255] = 0;
    strncpy( sWindowTitleString, argv[0], 255 );
    if ( glb_viewport.ViewName() ) {
      strncat( sWindowTitleString, " - ", 255 );
      strncat( sWindowTitleString, glb_viewport.ViewName(), 255 );
    }

    auxInitPosition( 0, 0, window_width, window_height );
    auxInitDisplayMode( AUX_SINGLE | AUX_RGB | AUX_DEPTH );
    auxInitWindow( sWindowTitleString );

    // register event handler functions
    auxIdleFunc( 0 );
    auxReshapeFunc( myReshape );
    auxMouseFunc( AUX_LEFTBUTTON,   AUX_MOUSEDOWN, myMouseLeftEvent );
    auxMouseFunc( AUX_LEFTBUTTON,   AUX_MOUSEUP,   myMouseLeftEvent );
    auxMouseFunc( AUX_MIDDLEBUTTON, AUX_MOUSEDOWN, myMouseMiddleEvent );
    auxMouseFunc( AUX_MIDDLEBUTTON, AUX_MOUSEUP,   myMouseMiddleEvent );
    auxMouseFunc( AUX_RIGHTBUTTON,  AUX_MOUSEDOWN, myMouseRightEvent );
    auxMouseFunc( AUX_RIGHTBUTTON,  AUX_MOUSEUP,   myMouseRightEvent );
    auxKeyFunc( AUX_LEFT,  myKeyLeftArrowEvent );
    auxKeyFunc( AUX_RIGHT, myKeyRightArrowEvent );
    auxKeyFunc( AUX_UP,    myKeyUpArrowEvent );
    auxKeyFunc( AUX_DOWN,  myKeyDownArrowEvent );
    auxKeyFunc( AUX_E,  myKeyViewExtents );
    auxKeyFunc( AUX_e,  myKeyViewExtents );
    auxKeyFunc( AUX_Z,  myKeyViewExtents );
    auxKeyFunc( AUX_z,  myKeyViewExtents );
    // setup model view matrix, GL defaults, and the GL NURBS renderer
    GLUnurbsObj* pTheGLNURBSRender = NULL; // OpenGL NURBS rendering context
    bOK = myInitGL( glb_viewport, pTheGLNURBSRender );

    if ( bOK ) {
      // build display list
      myBuildDisplayList( glb_display_list_number,
                          pTheGLNURBSRender,
                          object );

      // look at it
      auxMainLoop( myDisplay );
    }

    gluDeleteNurbsRenderer( pTheGLNURBSRender );

    break;
  }

  //
  // GL stuff ends here
  //
  ///////////////////////////////////////////////////////////////////

  // clean up
  int i;
  int count = object.Size();
  for ( i = 0; i < count; i++ ) {
    delete object[i];
    object[i] = NULL;
  }
  count = glb_spotlight.Size();
  for ( i = 0; i < count; i++ ) {
    delete glb_spotlight[i];
    glb_spotlight[i] = NULL;
  }
  count = viewport.Size();
  for ( i = 0; i < count; i++ ) {
    delete viewport[i];
    viewport[i] = NULL;
  }
  delete pRF; pRF = NULL;

  RhinoIOCleanup();  // see comments in rhinoio_chunk.h before using

  return 0;
}

///////////////////////////////////////////////////////////////////////

BOOL myInitGeometry( 
       int argc, const char *argv[],
       CRhinoFile*& pRF,
       double bounding_box[2][3],
       CRhinoObjectArray& object,
       CRhinoSpotlightArray& spotlight,
       CRhinoViewportArray& viewport
       ) 
{
  // opens file, 
  // reads file, 
  // puts Rhino geometry in global object[] array
  // returns TRUE if successful

  pRF = NULL;

  // get Rhino geometry from file and put in object[] array
  if ( argc != 2 ) {
    printf("Syntax: %s file1.3dm\n",argv[0] );
    return 0;
  }
  const char* sFileName = argv[1];
  printf("\nFile:  %s\n", sFileName );
  FILE* pFile = fopen( sFileName, "rb ");
  if ( !pFile ) {
    printf("Unable to open file.\n" );
    return FALSE;
  }

  BOOL bGrowBox = FALSE;
  BOOL bOK = TRUE;
  object.Reserve(200); // object[] is a global array

  // The Rhino IO tookit throw exceptions when it encounters an
  // unrecoverable error.  Use try/catch blocks as demonstrated
  // below.

  try {
		
		pRF = new CRhinoFile( pFile, CRhinoFile::read );
		
    CRhinoFile::eStatus status = CRhinoFile::good;

    for ( ;; ) {
		  CRhinoObject* pObj = pRF->Read( status );
      if ( status != CRhinoFile::good )
        break;
      switch( pObj->TypeCode() ) {

      case TCODE_RH_POINT:    // CRhinoPointSet ( 3d point list )
      case TCODE_MESH_OBJECT: // CRhinoMesh ( 3d mesh of triangles and quads )
      case TCODE_RHINO_OBJECT_NURBS_CURVE: // CRhinoNurbsCurve ( NURBS curve )
      case TCODE_RHINO_OBJECT_NURBS_SURFACE: // CRhinoNurbsSurface ( NURBS surface )
      case TCODE_RHINO_OBJECT_BREP: // CRhinoBrep ( B-rep model )
      case TCODE_AG_SPLINE:   // CRhinoAgSpline
      case TCODE_AG_SURFACE:  // CRhinoAgSurface
      case TCODE_AG_CURVE:    // CRhinoAgCurve
      case TCODE_AG_FACE:     // CRhinoBrep
      case TCODE_AG_SHELL:    // CRhinoBrep
        // add this geometry to list
        object.Append(pObj);
        // get bounding box
        if ( pObj->GetBoundingBox( bounding_box[0],bounding_box[1], bGrowBox ) )
          bGrowBox = TRUE;
        break;

      case TCODE_RH_SPOTLIGHT:    // CRhinoSpotlight
        {
          CRhinoSpotlight* pSpot = (CRhinoSpotlight*)pObj;
          spotlight.Append(pSpot);
        }
        break;

      case TCODE_VIEWPORT:    // CRhinoViewport
        {
          CRhinoViewport* pView = (CRhinoViewport*)pObj;
          viewport.Append(pView);
        }
        break;

      default:
        // throw away any other non-geometric stuff
  			delete pObj;
        break;
      }
    }
    bOK = ( status == CRhinoFile::eof );
  }
	
  catch( CRhinoException e ) {
    // if something goes wrong, catch the exception here

    bOK = FALSE;
    const char* s = NULL;
    switch ( e.m_type ) {
    case CRhinoException::unable_to_write:
      s = "unable_to_write";
      break;
    case CRhinoException::unable_to_read:
      s = "unable_to_read";
      break;
    case CRhinoException::unable_to_seek:
      s = "unable_to_seek";
      break;
    case CRhinoException::unexpected_end_of_file:
      s = "unexpected_end_of_file";
      break;
    case CRhinoException::unexpected_value:
      s = "unexpected_value";
      break;
    case CRhinoException::not_supported:
      s = "not_supported";
      break;
    case CRhinoException::illegal_input:
      s = "illegal_input";
      break;
    case CRhinoException::out_of_memory:
      s = "out_of_memory";
      break;
    default:
      s = "unknow exception";
      break;
    }

    printf("Reading threw CRhinoException: %s\n",s);

  }

  fclose( pFile );

  return bOK;
}

///////////////////////////////////////////////////////////////////////
void SetGLModelViewMatrix( const CRhinoViewport& viewport )
{
  RhinoGL( viewport ); // updates GL model view matrix
}

void SetGLProjectionMatrix( CRhinoViewport& viewport )
{
  int pl, pr, pb, pt;
  viewport.GetScreenPort( &pl, &pr, &pb, &pt, NULL, NULL );
  RhinoGL( viewport, pl, pr, pb, pt ); // updates GL projection matrix
}

BOOL myInitGL( const CRhinoViewport& viewport, GLUnurbsObj*& nobj )
{
  // set the model view transform
  SetGLModelViewMatrix( viewport );

  // this stuff works with MSVC 4.2's Open GL. Changes may be needed for other
  // GLs.

  glClearColor( 0.0f, 0.5f, 0.5f, 1.0f );

  glLightModeli( GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE );
  glDisable( GL_CULL_FACE );
  
  // Rhino viewports have camera "Z" pointing at the camera in a right
  // handed coordinate system.
  glClearDepth( 0.0f );
  glEnable( GL_DEPTH_TEST );
  glDepthFunc( GL_GEQUAL );

  glEnable( GL_LIGHTING );
  glEnable( GL_DITHER );
  //glEnable( GL_AUTO_NORMAL );
  //glEnable( GL_NORMALIZE );

  // default material
  RhinoGL( (CRhinoMaterial*)NULL );


  // GL rendering of NURBS objects requires a GLUnurbsObj.
  nobj = gluNewNurbsRenderer();
  if ( !nobj )
    return FALSE;
  
  gluNurbsProperty( nobj, GLU_SAMPLING_TOLERANCE,   20.0f );
  gluNurbsProperty( nobj, GLU_PARAMETRIC_TOLERANCE, 0.5f );
  gluNurbsProperty( nobj, GLU_DISPLAY_MODE,         (GLfloat)GLU_FILL );
  //gluNurbsProperty( nobj, GLU_DISPLAY_MODE,         GLU_OUTLINE_POLYGON );
  //gluNurbsProperty( nobj, GLU_DISPLAY_MODE,         GLU_OUTLINE_PATCH );
  gluNurbsProperty( nobj, GLU_SAMPLING_METHOD,      (GLfloat)GLU_PATH_LENGTH );
  //gluNurbsProperty( nobj, GLU_SAMPLING_METHOD,      GLU_PARAMETRIC_ERROR );
  //gluNurbsProperty( nobj, GLU_SAMPLING_METHOD,      GLU_DOMAIN_DISTANCE );
  gluNurbsProperty( nobj, GLU_CULLING,              (GLfloat)GL_FALSE );

  // register GL NURBS error callback
  {
    // hack to get around C vs C++ type checking trauma
    RHINO_GL_NURBS_ERROR fn;
    fn = (RHINO_GL_NURBS_ERROR)myNurbsErrorCallback;
    gluNurbsCallback( nobj, GLU_ERROR, fn );
  }

  return TRUE;
}

///////////////////////////////////////////////////////////////////////

void RHINO_GL_CALLBACK myReshape( GLsizei w, GLsizei h )
{
  static GLsizei w0 = 0;
  static GLsizei h0 = 0;
  if ( w != w0 || h != h0 ) {
    h0 = h;
    w0 = w;
    RhinoGL( glb_viewport, 0, w-1, h-1, 0 ); // set projection transform
  }
  glViewport( 0, 0, w, h );
}

///////////////////////////////////////////////////////////////////////
static void myRotateView( CRhinoViewport& viewport,
                          const double* axis,
                          const double* center,
                          double angle )
{
  CRhinoXform rot;
  double camLoc[3], camY[3], camZ[3];

  double ccenter[3] = {0.0,0.0,0.0};
  rot.Rotation( axis, ccenter, sin(angle), cos(angle) );

  if ( !viewport.GetCameraFrame( camLoc, NULL, camY, camZ ) )
    return;

  rot.Apply( camLoc[0], camLoc[1], camLoc[2], camLoc );
  rot.ApplyVector( camY[0], camY[1], camY[2], camY );
  rot.ApplyVector( camZ[0], camZ[1], camZ[2], camZ );

  camZ[0] = -camZ[0];
  camZ[1] = -camZ[1];
  camZ[2] = -camZ[2];
  viewport.SetCamera( camLoc, camZ, camY );

  RhinoGL( viewport ); // update model view

}

static void myRotateLeftRight( double angle )
{
  // rotates camera around the most vertical world axis
  int i;
  double axis[3] = {0.0,0.0,0.0};
  double camY[3], center[3];
  glb_viewport.GetCameraFrame( NULL, NULL, camY, NULL );
  glb_viewport.GetFrustumCenter( center );
  i = fabs(camY[1]) > fabs(camY[2]) ? 1 : 2;
  if ( fabs(camY[0]) > fabs(camY[i]) )
    i = 0;
  axis[i] = (camY[i] < 0.0) ? -1.0 : 1.0;
  myRotateView( glb_viewport, axis, center, angle );
}

static void myRotateUpDown( double angle )
{
  // rotates camera around the screen x axis
  double camX[3], center[3];
  glb_viewport.GetCameraFrame( NULL, camX, NULL, NULL );
  glb_viewport.GetFrustumCenter( center );
  myRotateView( glb_viewport, camX, center, angle );
}

///////////////////////////////////////////////////////////////////////

void RHINO_GL_CALLBACK myKeyLeftArrowEvent( void )
{
  myRotateLeftRight( RHINO_PI/12.0 );
}

void RHINO_GL_CALLBACK myKeyRightArrowEvent( void )
{
  myRotateLeftRight( -RHINO_PI/12.0 );
}

void RHINO_GL_CALLBACK myKeyUpArrowEvent( void )
{
  myRotateUpDown( RHINO_PI/12.0 );
}

void RHINO_GL_CALLBACK myKeyDownArrowEvent( void )
{
  myRotateUpDown( -RHINO_PI/12.0 );
}

void RHINO_GL_CALLBACK myKeyViewExtents( void )
{
  double half_angle = 7.5*RHINO_PI/180.0;
  glb_viewport.Extents( half_angle, glb_bounding_box[0], glb_bounding_box[1] );
  SetGLModelViewMatrix( glb_viewport );
  SetGLProjectionMatrix( glb_viewport );
}

///////////////////////////////////////////////////////////////////////
//
// Mouse event handling
//

static void myMouseEvent( GLint button, const AUX_EVENTREC* event )
{
  static bMouseActive = FALSE;
  static int mx0, my0;
  int mx, my;

  if ( !event ) {
    bMouseActive = FALSE;
    return;
  }

  if ( event->event == AUX_MOUSEDOWN ) {
    if ( bMouseActive ) {
      bMouseActive = FALSE;
      return;
    }
    bMouseActive = TRUE;
    mx0 = event->data[AUX_MOUSEX];
    my0 = event->data[AUX_MOUSEY];
    return;
  }

  if ( !bMouseActive || event->event != AUX_MOUSEUP )
    return;

  mx = event->data[AUX_MOUSEX];
  my = event->data[AUX_MOUSEY];

  switch (button) {
  case AUX_LEFTBUTTON:
    // zoom
    glb_viewport.ZoomToScreenRect( mx0, my0, mx, my );
    break;
  case AUX_MIDDLEBUTTON:
    break;
  case AUX_RIGHTBUTTON:
    // dolly
    glb_viewport.DollyAlongScreenChord( mx0, my0, mx, my );
    break;
  }

  // update GL model view and projection matrices to match viewport changes
  SetGLModelViewMatrix( glb_viewport );
  SetGLProjectionMatrix( glb_viewport );

  bMouseActive = FALSE;
}

void RHINO_GL_CALLBACK myMouseLeftEvent( AUX_EVENTREC* event )
{
  myMouseEvent( AUX_LEFTBUTTON, event );
}

void RHINO_GL_CALLBACK myMouseMiddleEvent( AUX_EVENTREC* event )
{
  myMouseEvent( AUX_MIDDLEBUTTON, event );
}

void RHINO_GL_CALLBACK myMouseRightEvent( AUX_EVENTREC* event )
{
  myMouseEvent( AUX_RIGHTBUTTON, event );
}

///////////////////////////////////////////////////////////////////////

void myDisplayObject( GLUnurbsObj* nobj, const CRhinoObject& object )
{
  // Called from myDisplay() to show geometry.
  // Uses RhinoGL() functions found in rhinoio_gl.cpp.

  RhinoGL( object.Material() );

  switch( object.TypeCode() ) {

  case TCODE_RH_POINT:    // CRhinoPointSet ( 3d point list )
    {
      const CRhinoPointSet& points = (const CRhinoPointSet&)object;
      RhinoGL( points );
    }
    break;

  case TCODE_MESH_OBJECT: // CRhinoMesh ( 3d mesh of triangles and quads )
    {
      const CRhinoMesh& mesh = (const CRhinoMesh&)object;
      RhinoGL( mesh );
    }
    break;

  case TCODE_RHINO_OBJECT_NURBS_CURVE: // CRhinoNurbsCurve ( NURBS curve )
  case TCODE_AG_SPLINE: // CRhinoAgSpline
    {
      const CRhinoNurbsCurve& nurbs_crv = (const CRhinoNurbsCurve&)object;
      gluBeginCurve( nobj );
      RhinoGL( nurbs_crv, nobj );
      gluEndCurve( nobj );
    }
    break;

  case TCODE_RHINO_OBJECT_NURBS_SURFACE: // CRhinoNurbsSurface ( NURBS surface )
  case TCODE_AG_SURFACE: // CRhinoAgSurface
    {
      const CRhinoNurbsSurface& nurbs_srf = (const CRhinoNurbsSurface&)object;
      gluBeginSurface( nobj );
      RhinoGL( nurbs_srf, nobj );
      gluEndSurface( nobj );
    }
    break;

  case TCODE_AG_CURVE: // CRhinoAgCurve ( AGLib curve )
    {
      const CRhinoAgCurve& crv = (const CRhinoAgCurve&)object;
      int i;
      for ( i = 0; i < crv.NurbsCurveCount(); i++ ) {
        myDisplayObject( nobj, *crv[i] );
      }
    }
    break;

  case TCODE_RHINO_OBJECT_BREP: // CRhinoBrep
  case TCODE_AG_FACE:
  case TCODE_AG_SHELL:
    {
      const CRhinoBrep& b = (const CRhinoBrep&)object;
      int face_index;
      const int face_count = b.m_F.Size();
      for ( face_index = 0; face_index < face_count; face_index++ ) {
        gluBeginSurface( nobj );
        RhinoGL( b, face_index, nobj );
        gluEndSurface( nobj );
      }
    }
    break;
  }

}

///////////////////////////////////////////////////////////////////////

void RHINO_GL_CALLBACK myDisplayLighting( const CRhinoViewport& viewport,
                                        const CRhinoSpotlightArray& spotlight 
                                        )
{
  // spotlight display is disabled because I want a view independent head light
  if ( spotlight.Size() > 0 ) {
    // This is not working right
    // use Rhino spotlights for illumination
    int maxlighti = spotlight.Size();
    if ( maxlighti > GL_MAX_LIGHTS )
      maxlighti = GL_MAX_LIGHTS;
    int lighti;
    for ( lighti = 0; lighti < maxlighti; lighti++ ) {
      RhinoGL( *spotlight[lighti], lighti+GL_LIGHT0 );
    }
  }
  else {
    // use default headlight
    // use basic bright white head light with a bit of ambient
    double camZ[3];
    viewport.GetCameraFrame( NULL, NULL, NULL, camZ );
    
    GLfloat pos[4]  = { (GLfloat)camZ[0], (GLfloat)camZ[1], (GLfloat)camZ[2], (GLfloat)0.0 };
    glLightfv( GL_LIGHT0, GL_POSITION,  pos );

    GLfloat black[4] = { (GLfloat)0.0, (GLfloat)0.0, (GLfloat)0.0, (GLfloat)1.0 };
    GLfloat grey2[4] = { (GLfloat)0.2, (GLfloat)0.2, (GLfloat)0.2, (GLfloat)1.0 };
    GLfloat white[4] = { (GLfloat)1.0, (GLfloat)1.0, (GLfloat)1.0, (GLfloat)1.0 };

    glLightfv( GL_LIGHT0, GL_AMBIENT,  black );
    glLightfv( GL_LIGHT0, GL_DIFFUSE,  white );
    glLightfv( GL_LIGHT0, GL_SPECULAR, white );

    glEnable( GL_LIGHT0 );
  }
}

///////////////////////////////////////////////////////////////////////

#if defined(_WINDOWS) || defined(_WIN32)
static void myDrawAxesSprite( const CRhinoViewport& viewport, HDC hdc )
{
  // Use simple Windows calls to draw world axes sprite in lower left corner.
  // Note that Windows has screen (0,0) in the upper left corner; i.e,
  // screen "y" increases downwards.
  if ( !hdc )
    return;
  const int axes_size = 30;

  int port_left, port_right, port_top, port_bottom;
  if ( !viewport.GetScreenPort( &port_left, &port_right, &port_bottom, &port_top, NULL, NULL ) )
    return;
  const int scr_width  = port_right - port_left; // no "+1" here
  const int scr_height = port_bottom - port_top; // no "+1" here

  if (4*axes_size >= scr_width )
    return;
  if (4*axes_size >= scr_height )
    return;

  int x0 = 3*axes_size/2;
  int y0 = port_bottom - 3*axes_size/2;
  int indx[3] = {0,1,2};
  int scr_coord[3][2];
  viewport.GetCoordinateSprite( axes_size, x0, y0, indx, scr_coord );

#define LXSIZE 3
#define LYSIZE 3
#define LOFF 3

  // draw 3 axes from back to front
  COLORREF c[3] = { RGB(255,0,0), RGB(0,255,0), RGB(0,0,255) };
  HPEN axis_pen[3];
  axis_pen[0] = CreatePen( PS_SOLID, 2, RGB(255,0,0) );
  axis_pen[1] = CreatePen( PS_SOLID, 2, RGB(0,255,0) );
  axis_pen[2] = CreatePen( PS_SOLID, 2, RGB(0,0,255) );
  HGDIOBJ saved_pen = SelectObject( hdc, axis_pen[0] );

  int i, k, x, y, lx, ly;
  for (i=0;i<3;i++) {
    k = indx[i];
    x = scr_coord[k][0];
    y = scr_coord[k][1];
    // use direction of screen vector to determine letter placement
    lx = x-x0; ly = y-y0;
    if (abs(lx) > abs(ly)) {
      // center letter to right/left of axis end
      lx = (x >= x0) ? x + LXSIZE+LOFF : x - LXSIZE-LOFF;
      ly = y;
    }
    else if (abs(ly) > abs(lx)) {
      // center letter above/below axis end
      lx = x;
      ly = (y >= y0) ? y + LYSIZE+LOFF : y - LYSIZE-LOFF;
    }
    else if (lx) {
      // diagonal axis - center letter on axis
      lx = (x >= x0) ? x + LXSIZE+LOFF : x - LXSIZE-LOFF;
      ly = (y >= y0) ? y + LYSIZE+LOFF : y - LYSIZE-LOFF;
    }
    else {
      // axis is perp to screen - center letter at axis end
      lx = x;
      ly = y;
    }
    SelectObject( hdc, axis_pen[k] );

    // draw axis
    MoveToEx( hdc, x0, y0, NULL );
    LineTo( hdc, x, y );

    // draw axis label
    switch (k) {
    case 0: // X
      MoveToEx( hdc, lx-LXSIZE, ly-LYSIZE, NULL );
      LineTo(   hdc, lx+LXSIZE, ly+LYSIZE );
      MoveToEx( hdc, lx-LXSIZE, ly+LYSIZE, NULL );
      LineTo(   hdc, lx+LXSIZE, ly-LYSIZE );
      break;
    case 1: // Y
      MoveToEx( hdc, lx-LXSIZE, ly-LYSIZE, NULL );
      LineTo(   hdc, lx,    ly    );
      LineTo(   hdc, lx+LXSIZE, ly-LYSIZE );
      MoveToEx( hdc, lx,    ly, NULL    );
      LineTo(   hdc, lx,    ly+LYSIZE );
      break;
    case 2: // Z
      MoveToEx( hdc, lx-LXSIZE, ly-LYSIZE, NULL );
      LineTo(   hdc, lx+LXSIZE, ly-LYSIZE );
      LineTo(   hdc, lx-LXSIZE, ly+LYSIZE );
      LineTo(   hdc, lx+LXSIZE, ly+LYSIZE );
      break;
    }

  }
  SelectObject( hdc, saved_pen );
  DeleteObject( axis_pen[0] );
  DeleteObject( axis_pen[1] );
  DeleteObject( axis_pen[2] );

#undef LXSIZE
#undef LYSIZE
#undef LOFF

}
#endif

///////////////////////////////////////////////////////////////////////

void myBuildDisplayList( GLuint display_list_number,
                         GLUnurbsObj* pTheGLNurbsRender,
                         const CRhinoObjectArray& object
                         )
{
  glNewList( display_list_number, GL_COMPILE );

  // display Rhino geometry using RhinoGL() functions found in rhinoio_gl.cpp
  int i;
  const int object_count = object.Size();
  for ( i = 0; i < object_count; i++ ) {
    myDisplayObject( pTheGLNurbsRender, *object[i] );
  }

  glEndList();
}

///////////////////////////////////////////////////////////////////////

void RHINO_GL_CALLBACK myDisplay( void )
{
  // Uses globals glb_* because the GL aux tools don't provide an
  // easy way to pass information into this callback.
  int bUseRhinoSpotlights = FALSE; // I like to use a simple headlight
                                   // for a basic preview.

  glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

  // lights
  if ( bUseRhinoSpotlights ) {
    // Rhino spotlights (currently rotate along with geometry)
    myDisplayLighting( glb_viewport, glb_spotlight );
  }
  else {
    // simple bright white headlight
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glLoadIdentity();
    GLfloat pos[4]  = { (GLfloat)0.0, (GLfloat)0.0, (GLfloat)1.0, (GLfloat)0.0 };
    glLightfv( GL_LIGHT0, GL_POSITION,  pos );
    GLfloat black[4] = { (GLfloat)0.0, (GLfloat)0.0, (GLfloat)0.0, (GLfloat)1.0 };
    GLfloat white[4] = { (GLfloat)1.0, (GLfloat)1.0, (GLfloat)1.0, (GLfloat)1.0 };
    glLightfv( GL_LIGHT0, GL_AMBIENT,  black );
    glLightfv( GL_LIGHT0, GL_DIFFUSE,  white );
    glLightfv( GL_LIGHT0, GL_SPECULAR, white );
    glEnable( GL_LIGHT0 );
    glPopMatrix();
  }

  // display list built with myBuildDisplayList()
  glCallList( glb_display_list_number ); 

  glFlush();

#if defined(_WINDOWS) || defined(_WIN32)
  // Windows decorations
  myDrawAxesSprite( glb_viewport, wglGetCurrentDC() );
#endif
}

///////////////////////////////////////////////////////////////////////

void RHINO_GL_CALLBACK myNurbsErrorCallback( GLenum errCode )
{
  const GLubyte* s = gluErrorString( errCode );
  printf("GL NURBS ERROR: (%d) %s\n",errCode, s );
}

///////////////////////////////////////////////////////////////////////
