/* $Header: /Src/RhinoIO/example_read/example_read.cpp 42    1/12/00 6:04p Dalelear $ */
/* $NoKeywords: $ */
//
// Copyright (c) 1993-1998 Robert McNeel & Associates. All rights reserved.
// Rhinoceros is a registered trademark of Robert McNeel & Assoicates.
//
// THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
// ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
// MERCHANTABILITY ARE HEREBY DISCLAIMED.
//				
////////////////////////////////////////////////////////////////
//
//  example_read.cpp  
// 
//  Example program using the Rhino file IO toolkit.  The program reads in  
//  a Rhino 3dm model file and describes its contents.  The program is a 
//  console application that takes a filename as a command line argument.
//
////////////////////////////////////////////////////////////////////////

#include "rhinoio.h"
#include <malloc.h>

////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////

BOOL ReadRhinoFile( 
       const char* // name of file to read
       );

// text dump of some objects
// char* argument is used for indenting

void Dump( const CRhinoObject&,       const char* = NULL ); 
void Dump( const CRhinoPointSet&,     const char* = NULL ); 
void Dump( const CRhinoMesh&,         const char* = NULL ); 
void Dump( const CRhinoNurbsCurve&,   const char* = NULL ); 
void Dump( const CRhinoNurbsSurface&, const char* = NULL ); 

void Dump( const CRhinoSpotlight&,    const char* = NULL );
void Dump( const CRhinoMaterial&,     const char* = NULL );
void Dump( const CRhinoLayer&,        const char* = NULL );

void Dump( const CRhinoAgCurve&,      const char* = NULL ); 

void Dump( const CRhinoBrep&,         const char* = NULL ); 

void Dump( const CRhinoViewport&,     const char* = NULL );
void Dump( const CRhinoNotes&,        const char* = NULL );

static char* NextIndent( const char* );

////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////

int main( int argc, const char *argv[] )
{
  int argi;
  if ( argc < 2 ) {
    printf("Syntax: %s file1.3dm file2.3dm ...\n",argv[0] );
    return 0;
  }
  for ( argi = 1; argi < argc; argi++ ) {
    ReadRhinoFile( argv[argi] );
  }
  
  RhinoIOCleanup();  // see comments in rhinoio_chunk.h before using

  // getchar(); // uncomment when debugging
 
  return 0;
}

////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////

BOOL ReadRhinoFile( const char* sFileName )
{
  printf("\nFile:  %s\n", sFileName );

  FILE* pFile = fopen( sFileName, "rb ");
  if ( !pFile ) {
    printf("  Unable to open.\n" );
    return FALSE;
  }

  printf("  Objects:\n" );

  BOOL bOK = TRUE;
  // The Rhino IO tookit throw exceptions when it encounters an
  // unrecoverable error.  Use try/catch blocks as demonstrated
  // below.
  try {			
		
		CRhinoFile RF( pFile, CRhinoFile::read );
		
    CRhinoFile::eStatus status = CRhinoFile::good;

    for ( ;; ) {
      // this loop reads all the objects
		  CRhinoObject* pObj = RF.Read( status );
      if ( status != CRhinoFile::good )
        break;
			Dump( *pObj, "    " );  // print some info about the object
			delete pObj;	  // delete object
    }

    // notes
    if ( RF.Notes() ) {
      Dump( *RF.Notes(), "  " );
      printf("\n");
    }

    // tolerances
    {
      double angle_tolerance, deg, absolute_tolerance;
      RF.GetAbsoluteTolerance( &absolute_tolerance );
      RF.GetAngleTolerance( &angle_tolerance );
      deg = angle_tolerance*180.0/3.141592653589793238462643;
      printf("  Absolute tolerance: %g\n", absolute_tolerance );
      printf("  Angle tolerance: %g radians ( %g degrees)\n", angle_tolerance, deg );
    }

    // unit system
    {
      const char* sUnitSystem = 0;
      CRhinoFile::eUnitSystem us;
      RF.GetUnitSystem(&us);
      switch ( us ) {
      case CRhinoFile::no_units:
        sUnitSystem = "none";
        break;
      case CRhinoFile::microns:
        sUnitSystem = "microns";
        break;
      case CRhinoFile::millimeters:
        sUnitSystem = "millimeters";
        break;
      case CRhinoFile::centimeters:
        sUnitSystem = "centimeters";
        break;
      case CRhinoFile::meters:
        sUnitSystem = "meters";
        break;
      case CRhinoFile::kilometers:
        sUnitSystem = "kilometers";
        break;
      case CRhinoFile::microinches:
        sUnitSystem = "microinches";
        break;
      case CRhinoFile::mils:
        sUnitSystem = "mils";
        break;
      case CRhinoFile::inches:
        sUnitSystem = "inches";
        break;
      case CRhinoFile::feet:
        sUnitSystem = "feet";
        break;
      case CRhinoFile::miles:
        sUnitSystem = "miles";
        break;
      default:
        sUnitSystem = "unknown";
      }
      printf("  Unit System: %s\n",sUnitSystem);
    }
    printf("\n");

		// Dump the layer table 
    //
    // Layer information is automatically handled by the CRhinoFile
    // class.  The CRhinoFile::Read() function will never return
    // a CRhinoLayer object.

    printf("  Layers:\n" );
		int layer_index;
		for( layer_index = 0; layer_index < RF.NumLayers(); layer_index++ ) {
      printf( "    Layer %3d: ", layer_index );
			Dump( *RF.Layer( layer_index ), "" );
		}

    bOK = ( status == CRhinoFile::eof );

	}
	
  catch( CRhinoException e ) {
    // if something goes wrong, catch the exception here
    bOK = FALSE;
    const char* s;
    switch ( e.m_type ) {
    case CRhinoException::unable_to_write:
      s = "unable_to_write";
      break;
    case CRhinoException::unable_to_read:
      s = "unable_to_read";
      break;
    case CRhinoException::unable_to_seek:
      s = "unable_to_seek";
      break;
    case CRhinoException::unexpected_end_of_file:
      s = "unexpected_end_of_file";
      break;
    case CRhinoException::unexpected_value:
      s = "unexpected_value";
      break;
    case CRhinoException::not_supported:
      s = "not_supported";
      break;
    case CRhinoException::illegal_input:
      s = "illegal_input";
      break;
    case CRhinoException::out_of_memory:
      s = "out_of_memory";
      break;
    default:
      s = "unknow exception";
      break;
    }

    printf("Trouble reading file.  CRhinoException type: %s\n", s);
  }

  fclose( pFile );

  return bOK;
}

////////////////////////////////////////////////////////////////////////
//  These Dump() function print out information about the objects //////
////////////////////////////////////////////////////////////////////////

void Dump( const CRhinoObject& object, const char* sIndent /* = NULL */ )
{
  // If you need to upcast, then use object.TypeCode() to determine
  // the specific type of Rhino object.

  if ( !sIndent )
    sIndent = "";

  // get next level of indenting for layer and material dump.
  char* sNextIndent = NextIndent( sIndent );


  if ( object.TypeCode() != TCODE_VIEWPORT && object.TypeCode() != TCODE_COMMENTBLOCK) {
    // if it's not a viewport definition or comment block, then it's an
    // object and it has layer, name, rendering, ... properties

    const CRhinoLayer* pLayer = object.Layer();
    const char* sLayerName = (pLayer) ? pLayer->Name() : 0;
    if ( !sLayerName )
      sLayerName = "current layer";

    const char* sObjectState = "unknown";
    switch( object.State() ) {
    case CRhinoObject::normal:
      sObjectState = "normal";
      break;
    case CRhinoObject::hidden:
      sObjectState = "hidden";
      break;
    case CRhinoObject::locked:
      sObjectState = "locked";
      break;
    default:
      sObjectState = "unknown";
      break;
    }
    const char* sObjectName = object.Label();
    if ( !sObjectName )
      sObjectName = "";
    const CRhinoMaterial* pMat = object.Material(); // rendering information
    printf( "%s%s object properties:\n"
              "%slayer = \"%s\"\n"
              "%sname = \"%s\"\n"
              "%sstate = %s\n",
            sIndent,  object.ClassName(), 
            sNextIndent, sLayerName, 
            sNextIndent, sObjectName,
            sNextIndent, sObjectState
            );
    if ( pMat ) {
      // dump rendering material information
      Dump( *pMat, sNextIndent );
    }
  }

  // 

  switch( object.TypeCode() ) {
    
  case TCODE_RH_POINT:    // CRhinoPointSet ( 3d point list )
    Dump( (const CRhinoPointSet&)object, sNextIndent );
    break;
    
  case TCODE_MESH_OBJECT: // CRhinoMesh ( 3d mesh of triangles and quads )
    Dump( (const CRhinoMesh&)object, sNextIndent );
    break;
    
  case TCODE_RHINO_OBJECT_NURBS_CURVE: // CRhinoNurbsCurve ( NURBS curve )
  case TCODE_AG_SPLINE: // CRhinoAgSpline ( NURBS curve from Rhino 1.0 )
    Dump( (const CRhinoNurbsCurve&)object, sNextIndent );
    break;
    
  case TCODE_RHINO_OBJECT_NURBS_SURFACE: // CRhinoNurbsSurface ( NURBS surface )
  case TCODE_AG_SURFACE: // CRhinoAgSurface ( NURBS surface from Rhino 1.0 )
    Dump( (const CRhinoNurbsSurface&)object, sNextIndent );
    break;
    
  case TCODE_AG_CURVE: // CRhinoAgCurve ( curve made from contiguous NURBS curves )
    Dump( (const CRhinoAgCurve&)object, sNextIndent );
    break;
    
  case TCODE_RHINO_OBJECT_BREP: // CRhinoBrep ( B-rep model )
  case TCODE_AG_FACE:  // CRhinoAgFace ( B-rep with one face from Rhino 1.0 );
  case TCODE_AG_SHELL: // CRhinoAgFace ( B-rep with multiple faces from Rhino 1.0 );
    Dump( (const CRhinoBrep&)object, sNextIndent );
    break;

  case TCODE_RH_SPOTLIGHT: 
    Dump( (const CRhinoSpotlight&)object, sNextIndent );
    break;

  case TCODE_VIEWPORT: 
    Dump( (const CRhinoViewport&)object, sIndent );
    break;

  case TCODE_COMMENTBLOCK: 
    // Comment blocks are supplied for people who like to read binary
    // files with text editors.  They (the blocks and perhaps the people)
    // can be safely ignored.
    printf( "%sComment block: %d bytes of ASCII text\n",
      sNextIndent, 
      object.Length() );
    break;

  default:
    // Either object that is not yet exported to public or it is something 
    // that can be ignored.  See 3dm.h for a list of type codes.  See
    // example_structure.c for more details abount chunks.
    if( object.TypeCode() & 0x80000000 ) {
      printf( "%sInteger Chunk: typecode = %08x value = %d\n",
        sNextIndent, 
        object.TypeCode(),
        object.Value() 
        );
    }
    else {
      printf( "%sData Chunk: typecode = %08x value = %d bytes\n",
        sNextIndent, 
        object.TypeCode(),
        object.Length() 
        );
    }
  }

  printf("\n");

  // for debugging:
  //if ( !object.IsValid() ) {
  //  printf("%s*** INVALID OBJECT ***\n",sIndent);
  //  exit(1);
  //}
  free(sNextIndent);
}

////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////
static char* NextIndent( const char* sCurrentIndent )
{
  // get next level of indenting for layer and material dump.
  const int indent_length = sCurrentIndent ? strlen(sCurrentIndent) : 0;
  char* sNextIndent = (char*)malloc( (indent_length+3)*sizeof(sNextIndent[0]) );
  memset( sNextIndent, ' ', indent_length+2 );
  sNextIndent[indent_length+2] = 0;
  return sNextIndent;
}


void Dump( const CRhinoPointSet& point_set, const char* sIndent /* = NULL */ )
{
  if ( !sIndent )
    sIndent = "";

  const int point_count = point_set.PointCount();

	printf( "%sPoint Set with %d points.\n", 
          sIndent,
          point_count );

  int point_index;
  for ( point_index = 0; point_index < point_count; point_index++ ) {
    double x, y, z;
    point_set.GetPoint( point_index, x, y, z );
    printf( "%s  point %3d: ( %g, %g, %g )\n",
            sIndent, 
            point_index,
            x, y, z );
  }

}

void Dump( const CRhinoMesh& mesh, const char* sIndent /* = NULL */ )
{
  if ( !sIndent )
    sIndent = "";

  const int vertex_count = mesh.NumVertices();
  const int face_count   = mesh.NumFaces();
  int triangle_count     = 0;
  int quadrangle_count   = 0;

  int face_index;
  for ( face_index = 0; face_index < face_count; face_index++ ) {
    int v0, v1, v2, v3;
    mesh.GetFace( face_index, v0, v1, v2, v3 );
    if ( v2 == v3 )
      triangle_count++;
    else
      quadrangle_count++;
  }

  // Use the GetVertex(), GetNormal(), GetTCoord(), and GetFace() 
  // member functions to get the defining parameters.

  printf( "%sMesh with %d vertices, %d faces (%d triangles and %d quads).\n"
          "%s  normals=%s, texture coordinates=%s.\n",
          sIndent, 
          vertex_count,
          face_count,
          triangle_count,
          quadrangle_count,
          sIndent,
          ( mesh.HasNormals() ) ? "yes" : "no",
          ( mesh.HasTCoords() ) ? "yes" : "no"
          );
}

void Dump( const CRhinoMaterial& mat, const char* sIndent /* = NULL */ )
{
  if ( !sIndent )
    sIndent = "";

  CRhinoColor c = mat.Ambient();  
  printf( "%sMaterial:  ambient  =  (%3d , %3d, %3d )\n",
          sIndent,
          c.Red(), c.Green(), c.Blue() );
  c = mat.Diffuse();  
  printf( "%s           diffuse  =  (%3i , %3i, %3i )\n",
          sIndent,
          c.Red(), c.Green(), c.Blue() );
  c = mat.Emission();  
  printf( "%s           emmisive =  (%3i , %3i, %3i )\n",
          sIndent,
          c.Red(), c.Green(), c.Blue() );

  c = mat.Specular();  
  printf( "%s           specular =  (%3i , %3i, %3i )\n",
          sIndent,
          c.Red(), c.Green(), c.Blue() );
  printf( "%s           shine =  %f\n",
          sIndent,
          mat.Shine() );
  printf( "%s           Transparency = %f\n",
          sIndent,
					mat.Transparency() );
  if(mat.TextureBitmapFileName())
	  printf( "%s           Texture map Bitmap FileName: %s \n",
            sIndent,
            mat.TextureBitmapFileName() );
  if(mat.BumpBitmapFileName())
	  printf( "%s           Bump map Bitmap FileName:    %s \n",
            sIndent,
            mat.BumpBitmapFileName() );
  if(mat.RenderMaterialName())
	  printf( "%s           Render Material Name:    %s \n",
            sIndent,
            mat.RenderMaterialName() );
}

void Dump( const CRhinoLayer& layer, const char* sIndent /* = NULL */ )
{
  if ( !sIndent )
    sIndent = "";

	char* state = "";
	int r,g,b;
	switch(layer.State()){
	case CRhinoLayer::normal:
		state = "normal"; 
		break;
	case CRhinoLayer::locked:
		state = "locked"; 
		break;
	case CRhinoLayer::hidden:
		state = "hidden"; 
		break;
	case CRhinoLayer::ghosted:
		state = "ghosted";
	}
	r = layer.Color().Red( );
	g = layer.Color().Green( );
	b = layer.Color().Blue( );
  printf( "%sLayer: %s  state = %s RGB color = %d %d %d\n", 
          sIndent,
          layer.Name(), state, r,g,b);
}

void Dump( const CRhinoNurbsCurve& curve, const char* sIndent /* = NULL */ )
{
  // get next level of indenting
  if ( !sIndent )
    sIndent = "";

  // See the CRhinoNurbsCurve class for member functions that will get
  // control points, knot values, and other details.
  printf( "%sNURBS curve: %s degree = %d  %d control points\n"
          "%s             closed = %s  periodic = %s\n",
          sIndent, 
          curve.IsRational() ? "rational" : "non-rational",
          curve.Degree(),
          curve.CVCount(),
          sIndent,
          curve.IsClosed()   ? "TRUE" : "FALSE",
          curve.IsPeriodic() ? "TRUE" : "FALSE"
          );
}

void Dump( const CRhinoNurbsSurface& surface, const char* sIndent /* = NULL */ )
{
  // get next level of indenting
  if ( !sIndent )
    sIndent = "";

  // See the CRhinoNurbsSurface class for member functions that will get
  // control points, knot values, and other details.
  printf( "%sNURBS Surface: %s degree = %d X %d  control net = %d X %d points\n"
          "%s               closed = %s X %s  periodic = %s X %s\n",
          sIndent, 
          surface.IsRational() ? "rational" : "non-rational",
          surface.Degree(0),
          surface.Degree(1),
          surface.CVCount(0),
          surface.CVCount(1),
          sIndent,
          surface.IsClosed(0)   ? "TRUE" : "FALSE",
          surface.IsClosed(1)   ? "TRUE" : "FALSE",
          surface.IsPeriodic(0) ? "TRUE" : "FALSE",
          surface.IsPeriodic(1) ? "TRUE" : "FALSE"
          );
}

void Dump( const CRhinoAgCurve& crv, const char* sIndent /* = NULL */ )
{
  if ( !sIndent )
    sIndent = "";

  printf( "%s3d curve has %d NURBS segments\n",
          sIndent,
          crv.NurbsCurveCount() );

  // get next level of indenting for layer and material dump.
  char* sNextIndent = NextIndent( sIndent );

  int i;
  for ( i = 0; i < crv.NurbsCurveCount(); i++ )
    Dump( *(crv[i]), sNextIndent );

  free(sNextIndent);
}

static int brep_i_count = 0;

void Dump( const CRhinoBrep& brep, const char* sIndent /* = NULL */ )
{
  brep_i_count++;
  if ( !sIndent )
    sIndent = "";

  printf("%sBrep:\n",sIndent);

  // get next level of indenting for layer and material dump.
  char* sNextIndent = NextIndent( sIndent );

  if ( brep.IsSurface() ) {
    printf("%s(B-rep geometry is the same as underlying surface.)\n",sNextIndent);
  }

  printf("%ssurfaces:  %d\n",sNextIndent,brep.m_S.Size());
  printf("%s3d curve:  %d\n",sNextIndent,brep.m_C3.Size());
  printf("%s2d curves: %d\n",sNextIndent,brep.m_C2.Size());
  printf("%svertices:  %d\n",sNextIndent,brep.m_V.Size());
  printf("%sedges:     %d\n",sNextIndent,brep.m_E.Size());
  printf("%strims:     %d\n",sNextIndent,brep.m_T.Size());
  printf("%sloops:     %d\n",sNextIndent,brep.m_L.Size());
  printf("%sfaces:     %d\n",sNextIndent,brep.m_F.Size());

  int vi;
  for ( vi = 0; vi < brep.m_V.Size(); vi++ ) {
    const CRhinoBrepVertex& vertex = brep.m_V[vi];
    printf("%svertex[%2d]: (%f %f %f) tolerance(%g) edges(",
           sNextIndent,vi,vertex.x,vertex.y,vertex.z,vertex.m_tolerance);
    int vei;
    for ( vei = 0; vei < vertex.m_ei.Size(); vei++ ) {
      printf( (vei)?",%d":"%d", vertex.m_ei[vei] );
    }
    printf(")\n");
  }

  int ei;
  for ( ei = 0; ei < brep.m_E.Size(); ei++ ) {
    const CRhinoBrepEdge& edge = brep.m_E[ei];
    printf("%sedge[%2d]: v0(%2d) v1(%2d) 3d_curve(%d) tolerance(%g) trims(",
           sNextIndent,ei,edge.m_vi[0],edge.m_vi[1],edge.m_c3i,edge.m_tolerance);
    int eti;
    for ( eti = 0; eti < edge.m_ti.Size(); eti++ ) {
      printf( (eti)?",%d":"%d", edge.m_ti[eti]);
    }
    printf(")\n");
  }

  int fi;
  for ( fi = 0; fi < brep.m_F.Size(); fi++ ) {
    const CRhinoBrepFace& face = brep.m_F[fi];
    printf("%sface[%2d]: surface(%d) reverse(%d) loops(",
           sNextIndent,fi,face.m_si,face.m_bRev);
    int fli;
    for ( fli = 0; fli < face.m_li.Size(); fli++ ) {
      printf( (fli)?",%d":"%d", face.m_li[fli]);
    }
    printf(")\n");

    if ( brep.FaceIsSurface(fi) ) {
      printf("%s  (Face geometry is the same as underlying surface.)\n",sNextIndent);
    }

    for ( fli = 0; fli < face.m_li.Size(); fli++ ) {
      const int li = face.m_li[fli];
      const CRhinoBrepLoop& loop = brep.m_L[li];
      const char* sLoopType = "unknown";
      switch( loop.m_type ) {
      case CRhinoBrepLoop::unknown:
        sLoopType = "unknown";
        break;
      case CRhinoBrepLoop::outer:
        sLoopType = "outer";
        break;
      case CRhinoBrepLoop::inner:
        sLoopType = "inner";
        break;
      case CRhinoBrepLoop::slit:
        sLoopType = "slit";
        break;
      }
      printf("%s  loop[%2d]: type(%s) %d trims(",
             sNextIndent, li, sLoopType, loop.m_ti.Size());
      int lti;
      for ( lti = 0; lti < loop.m_ti.Size(); lti++ ) {
        printf( (lti)?",%d":"%d", loop.m_ti[lti]);
      }
      printf(")\n");
      for ( lti = 0; lti < loop.m_ti.Size(); lti++ ) {
        const int ti = loop.m_ti[lti];
        const CRhinoBrepTrim& trim = brep.m_T[ti];
        const char* sTrimType = "?";
        const char* sTrimIso = "-?";
        switch( trim.m_type ) {
        case CRhinoBrepTrim::unknown:
          sTrimType = "unknown ";
          break;
        case CRhinoBrepTrim::boundary:
          sTrimType = "boundary";
          break;
        case CRhinoBrepTrim::mated:
          sTrimType = "mated   ";
          break;
        case CRhinoBrepTrim::seam:
          sTrimType = "seam    ";
          break;
        case CRhinoBrepTrim::singular:
          sTrimType = "singular";
          break;
        }
        switch( trim.m_iso ) {
        case CRhinoBrepTrim::not_iso:
          sTrimIso = "";
          break;
        case CRhinoBrepTrim::u_iso:
          sTrimIso = "-u iso";
          break;
        case CRhinoBrepTrim::W_iso:
          sTrimIso = "-west iso";
          break;
        case CRhinoBrepTrim::E_iso:
          sTrimIso = "-east iso";
          break;
        case CRhinoBrepTrim::v_iso:
          sTrimIso = "-v iso";
          break;
        case CRhinoBrepTrim::S_iso:
          sTrimIso = "-south iso";
          break;
        case CRhinoBrepTrim::N_iso:
          sTrimIso = "-north iso";
          break;
        }
        printf("%s    trim[%2d]: edge(%2d) v0(%2d) v1(%2d) tolerance(%g,%g)\n"
               "%s              type(%s%s) rev3d(%d) 2d_curve(%d)\n",
               sNextIndent,ti,
               trim.m_ei,trim.m_vi[0],trim.m_vi[1],
               trim.m_tolerance[0],trim.m_tolerance[1],
               sNextIndent,
               sTrimType,sTrimIso,trim.m_bRev3d,trim.m_c2i);
      }
    }
  }

  int si;
  for ( si = 0; si < brep.m_S.Size(); si++ ) {
    Dump( *brep.m_S[si], sNextIndent );
  }

  free(sNextIndent);
}

void Dump( const CRhinoSpotlight& spotlight, const char* sIndent /* = NULL */ )
{
  // get next level of indenting
  if ( !sIndent )
    sIndent = "";

  // See the CRhinoNurbsSurface class for member functions that will get
  // control points, knot values, and other details.
  double position[3];
  spotlight.GetPosition( position );
  printf( "%sSpotlight: location     = (%g, %g, %g)\n",
          sIndent, position[0], position[1], position[2] );

  double direction[3];
  spotlight.GetDirection( direction );
  printf( "%s           direction    = (%g, %g, %g)\n",
          sIndent, direction[0], direction[1], direction[2] );

  double cutoff_angle = spotlight.SpotCutoff();
  printf( "%s           cutoff angle = %g degrees\n",sIndent, cutoff_angle);

  double exponent = spotlight.SpotExponent();
  printf( "%s           exponent     = %g\n",sIndent, exponent);

  CRhinoColor c = spotlight.Ambient();
  printf( "%s           ambient      = (%3i , %3i, %3i )\n",
          sIndent, c.Red(), c.Green(), c.Blue() );
  
  c = spotlight.Diffuse();
  printf( "%s           diffuse      = (%3i , %3i, %3i )\n",
          sIndent, c.Red(), c.Green(), c.Blue() );

  c = spotlight.Specular();
  printf( "%s           specular     = (%3i , %3i, %3i )\n",
          sIndent, c.Red(), c.Green(), c.Blue() );

}

void Dump( const CRhinoViewport& view, const char* sIndent /* = NULL */ )
{
  // get next level of indenting
  if ( !sIndent )
    sIndent = "";

  const char* sViewName = view.ViewName();
  if ( !sViewName )
    sViewName = "";
  CRhinoViewport::eProjection proj = view.Projection();
  

  double camLoc[3], camX[3], camY[3], camZ[3];
  BOOL bValidCamera = view.GetCameraFrame( camLoc, camX, camY, camZ );
  double frus_left,frus_right,frus_bottom,frus_top,frus_near,frus_far;
  BOOL bValidFrustum = view.GetFrustum(&frus_left,&frus_right,&frus_bottom,&frus_top,&frus_near,&frus_far);
  int port_left, port_right, port_bottom, port_top, port_near, port_far;
  BOOL bValidPort = view.GetScreenPort(&port_left,&port_right,&port_bottom,&port_top,&port_near,&port_far);
  double rel_left,rel_right,rel_bottom,rel_top;
  view.GetPortPosition(&rel_left,&rel_right,&rel_bottom,&rel_top);

  printf("%sViewport: name = \"%s\" projection = %s\n", 
         sIndent,
         sViewName,
         (proj == CRhinoViewport::perspective_proj)
         ?"perspective":"parallel" 
         );

  if ( bValidCamera ) {
    printf("%s          camera frame\n"
           "%s           location = %g, %g, %g\n"
           "%s           X = %g, %g, %g\n"
           "%s           Y = %g, %g, %g\n"
           "%s           Z = %g, %g, %g\n",
           sIndent,
           sIndent,camLoc[0],camLoc[1],camLoc[2],
           sIndent,camX[0],camX[1],camX[2],
           sIndent,camY[0],camY[1],camY[2],
           sIndent,camZ[0],camZ[1],camZ[2]
           );
  }

  if ( bValidFrustum ) {
    printf("%s          camera frustum\n"
           "%s           left   = %g, right = %g\n"
           "%s           bottom = %g, top   = %g\n"
           "%s           near   = %g, far   = %g\n",
           sIndent,
           sIndent,frus_left,frus_right,
           sIndent,frus_bottom,frus_top,
           sIndent,frus_near,frus_far
           );
  }

  if ( bValidPort ) {
    printf("%s          camera port\n"
           "%s           left   = %4d, right = %4d\n"
           "%s           bottom = %4d, top   = %4d\n"
           "%s           near   = %4d, far   = %4d\n",
           sIndent,
           sIndent,port_left,port_right,
           sIndent,port_bottom,port_top,
           sIndent,port_near,port_far
           );
  }

  double target_distance = 0.0;
  if ( view.GetTargetDistance( target_distance ) ) {
    double target_point[3];
    target_point[0] = camLoc[0] - target_distance*camZ[0];
    target_point[1] = camLoc[1] - target_distance*camZ[1];
    target_point[2] = camLoc[2] - target_distance*camZ[2];
    printf("%s          camera target\n"
           "%s           distance = %g\n"
           "%s           point = %g,%g,%g\n",
           sIndent,
           sIndent,target_distance,
           sIndent,target_point[0],target_point[1],target_point[2]
           );
  }

  printf("%s          relative position in Rhino layout\n"
         "%s           left   = %6.2f%%, right = %6.2f%%\n"
         "%s           bottom = %6.2f%%, top   = %6.2f%%\n",
         sIndent,
         sIndent,100.0*rel_left,100.0*rel_right,
         sIndent,100.0*rel_bottom,100.0*rel_top
         );

}

void Dump( const CRhinoNotes& notes,  const char* sIndent )
{
  // NOTE WELL: CRhinoNotes is derived from CRhinoChunk and not from CRhinoObject

  char sLine[256];
  const int wrap_length = 60;
  const char* sNotes = notes.String();
  if ( !sIndent )
    sIndent = "";
  char* sNextIndent = NextIndent( sIndent );
  printf("%sNotes:\n",sIndent);
  int i  = 0;
  int i1 = 0;
  int isp = 0;
  BOOL bPrintLine = FALSE;
  while ( sNotes[i] ) {
    i1 = i;
    if ( sNotes[i] == 10 || sNotes[i] == 13 ) {
      // hard break at CR or LF
      i++;
      if ( sNotes[i] == 10 && sNotes[i-1] == 13 ) {
        // it's a CR+LF hard end of line - skip LF too
        i++;
      }      
      bPrintLine = TRUE;
    }
    else if ( i && sNotes[i] == 32 ) {
      if ( !isp ) {
        isp = i++;
      }
      if ( i < wrap_length ) {
        isp = i++;
      }
      else {
        bPrintLine = TRUE;
        if ( isp ) {
          i1 = i = isp;
          while ( sNotes[i] == 32 )
            i++;
        }
        else {
          i++;
        }
      }
    }
    else {
      i++;
    }
    if ( bPrintLine ) {
      if ( i1 >= ((int)sizeof(sLine)) )
        i1 = sizeof(sLine)-1;
      if ( i1 > 0 ) {
        strncpy( sLine, sNotes, i1 );
        sLine[i1] = 0;
        printf( "%s%s\n", sNextIndent, sLine );
      }
      else {
        printf("\n");
      }

      sNotes += i;
      i = i1 = isp = 0;
      bPrintLine = FALSE;
    }
  }
  if ( sNotes[0] ) {
    printf( "%s%s\n", sNextIndent, sNotes );
  }
  printf("\n");

  free(sNextIndent);

}