/* $Header: /Src/RhinoIO/example_write/example_write.cpp 20    1/12/00 6:04p Dalelear $ */
/* $NoKeywords: $ */
//
// Copyright (c) 1993-1998 Robert McNeel & Associates. All rights reserved.
// Rhinoceros is a registered trademark of Robert McNeel & Assoicates.
//
// THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
// ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
// MERCHANTABILITY ARE HEREBY DISCLAIMED.
//				
////////////////////////////////////////////////////////////////

#include "rhinoio.h"

typedef struct  {
  double x;
  double y;
  double z;
} Point;

static void set_units_and_tolerances_example( 
                CRhinoFile& RF,
                CRhinoFile::eUnitSystem us,
                double absolute_tolerance,
                double angle_tolerance
                )
{
  // If you want to specify a unit system or tolerance value,
  // you should set them right after the CRhinoFile object
  // is constructed.  In particular, you must set these
  // values before the first Write() call is made.
  // You do not have to set any of these values and you are
  // permitted to set some but not all. The actual writing 
  // of the values is done automatically.

  // There are 11 choices for unit system.  See the
  // CRhinoFile::eUnitSystem definition for the list.
  // The Rhino default is CRhinoFile::no_units.
  RF.SetUnitSystem( us );

  // The absolute tolerance is specified with respect to
  // world coordinates.  The Rhino default is 0.01.
  // Do not set this value unless you have a good reason
  // to do so.
  RF.SetAbsoluteTolerance( absolute_tolerance );

  // The angle tolerance is specified in radians.
  // The Rhino default is 0.05235987755983 ( 3 degrees )
  // Do not set this value unless you have a good reason
  // to do so.
  RF.SetAngleTolerance( angle_tolerance );
}

static void write_point_set_example( FILE* pFile )
{
  // example demonstrates how to write a set of points

  // Rhino IO toolkit classes throw CRhinoExceptions when
  // trouble arises.  Use try/catch as demonstrated below
  // for clean error trapping.


  try {

    // pFile must be a file pointer returned from fopen( , "wb" );
    CRhinoFile RF( pFile, CRhinoFile::write );

    // (optional) Specify some or all of unit system and tolerances
    set_units_and_tolerances_example( RF,
                              CRhinoFile::centimeters,
                              0.01,
                              3.141592653589793238462643*1.0/180.0
                              );

    // (optional) create custom layer(s).  The CRhinoLayer::New() call
    // automatically adds the layers to the Rhino file's layer table.
    CRhinoLayerPtr LayerPtr = 
      CRhinoLayer::New( "3 blue points",          // text name
                        CRhinoColor( 0, 0, 255 )  // wire display RGB color
                        );

    //////////////////////////////////////////////////////////////
    //////////////////////////////////////////////////////////////

    // create a point set to write
    CRhinoPointSet P;
    P.AddPoint( 1.0, 4.0, 5.0 );
    P.AddPoint( 2.0, 3.0, 1.0 );
    P.AddPoint( 3.0, 2.0, 4.0 );

    //////////////////////////////////////////////////////////////
    //////////////////////////////////////////////////////////////

    // (optional) Put surface on a specific layer.  By default it ends up on
    // Rhino's default layer.
    P.SetLayer( LayerPtr );

    P.Write( RF ); // save S in Rhino file
  }

  catch( CRhinoException e ) {
    // if something goes wrong, catch the exception her
    const char* s;
    switch ( e.m_type ) {
    case CRhinoException::unable_to_write:
      s = "unable_to_write";
      break;
    case CRhinoException::unable_to_read:
      s = "unable_to_read";
      break;
    case CRhinoException::unable_to_seek:
      s = "unable_to_seek";
      break;
    case CRhinoException::unexpected_end_of_file:
      s = "unexpected_end_of_file";
      break;
    case CRhinoException::unexpected_value:
      s = "unexpected_value";
      break;
    case CRhinoException::not_supported:
      s = "not_supported";
      break;
    case CRhinoException::illegal_input:
      s = "illegal_input";
      break;
    case CRhinoException::out_of_memory:
      s = "out_of_memory";
      break;
    default:
      s = "unknow exception";
      break;
    }

    printf("Trouble writing file.  CRhinoException type: %s\n", s);
  }
}


static void write_nurbs_crv_example( FILE* pFile )
{
  // example demonstrates how to write a NURBS curve

  // Rhino IO toolkit classes throw CRhinoExceptions when
  // trouble arises.  Use try/catch as demonstrated below
  // for clean error trapping.


  //////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////

  // The code between the comment bands has nothing to do with Rhino IO.
  // It is simply an easy way to get a NURBS curve to write.
  const int bIsRational = FALSE;
  const int dim = 3;
  const int degree = 3;
  const int cv_count = 6;

  // The knot vector does NOT have the 2 superfluous knots
  // at the start and end of the knot vector.  If you are
  // coming from a system that has the 2 superfluous knots,
  // just ignore them when writing a 3dm file.
  double knot[ cv_count + degree - 1 ];


  // make up a cubic knot vector with one double interior knot
  knot[0] = knot[1] = knot[2] = 0.0;
  knot[3] = knot[4] = 1.5;
  knot[5] = knot[6] = knot[7] = 4.0;

  // Rational control points can be in either homogeneous
  // or euclidean form. Non-rational control points do not
  // need to specify a weight.  
  Point CV[cv_count];

  int i;
  for ( i = 0; i < cv_count; i++ ) {
    CV[i].x = 2*i;
    CV[i].y = -i;
    CV[i].z = (i-3)*(i-3);
  }

  //////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////

  try {

    // pFile must be a file pointer returned from fopen( , "wb" );
    CRhinoFile RF( pFile, CRhinoFile::write );

    // (optional) Specify some or all of unit system and tolerances
    set_units_and_tolerances_example( RF,
                              CRhinoFile::inches,
                              0.001,
                              3.141592653589793238462643*2.0/180.0
                              );

    // (optional) create custom layer(s).  The CRhinoLayer::New() call
    // automatically adds the layers to the Rhino file's layer table.
    CRhinoLayerPtr LayerPtr = 
      CRhinoLayer::New( "Nurbs Curves",           // text name
                        CRhinoColor( 0, 255, 50 )  // wire display RGB color
                        );

    //////////////////////////////////////////////////////////////
    //////////////////////////////////////////////////////////////

    // create a curve to write
    CRhinoNurbsCurve C( dim, bIsRational, 
                        degree,
                        cv_count );

    for ( i = 0; i < C.KnotCount(); i++ )
      C.SetKnot( i, knot[i] );

    for ( i = 0; i < C.CVCount(); i++ ) {
      // If the curve is rational, use either
      // C.SetCV( i, rh_homogeneous_rational, .. ) or
      // C.SetCV( i, rh_euclidean_rational, .. ).
      C.SetCV( i, rh_not_rational, &CV[i].x );
    }

    //////////////////////////////////////////////////////////////
    //////////////////////////////////////////////////////////////

    // (optional) Put curve on a specific layer.  By default it ends up on
    // Rhino's default layer.
    C.SetLayer( LayerPtr );

    // (optional) Add a label to the object.  By default, this label is "".
    C.SetLabel( "my NURBS curve" );

    C.Write( RF ); // save S in Rhino file
  }

  catch( CRhinoException e ) {
    // if something goes wrong, catch the exception her
    const char* s;
    switch ( e.m_type ) {
    case CRhinoException::unable_to_write:
      s = "unable_to_write";
      break;
    case CRhinoException::unable_to_read:
      s = "unable_to_read";
      break;
    case CRhinoException::unable_to_seek:
      s = "unable_to_seek";
      break;
    case CRhinoException::unexpected_end_of_file:
      s = "unexpected_end_of_file";
      break;
    case CRhinoException::unexpected_value:
      s = "unexpected_value";
      break;
    case CRhinoException::not_supported:
      s = "not_supported";
      break;
    case CRhinoException::illegal_input:
      s = "illegal_input";
      break;
    case CRhinoException::out_of_memory:
      s = "out_of_memory";
      break;
    default:
      s = "unknow exception";
      break;
    }

    printf("Trouble writing file.  CRhinoException type: %s\n", s);
  }
}


static void write_nurbs_srf_example( FILE* pFile )
{
  // example demonstrates how to write a NURBS surface

  // Rhino IO toolkit classes throw CRhinoExceptions when
  // trouble arises.  Use try/catch as demonstrated below
  // for clean error trapping.


  //////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////

  // The code between the comment bands has nothing to do with Rhino IO.
  // It is simply an easy way to get a NURBS surface to write.
  const int bIsRational = FALSE;
  const int dim = 3;
  const int u_degree = 2;
  const int v_degree = 3;
  const int u_cv_count = 3;
  const int v_cv_count = 5;

  // The knot vectors do NOT have the 2 superfluous knots
  // at the start and end of the knot vector.  If you are
  // coming from a system that has the 2 superfluous knots,
  // just ignore them when writing a 3dm file.
  double u_knot[ u_cv_count + u_degree - 1 ];
  double v_knot[ v_cv_count + v_degree - 1 ];

  // make up a quadratic knot vector with no interior knots
  u_knot[0] = u_knot[1] = 0.0;
  u_knot[2] = u_knot[3] = 1.0;

  // make up a cubic knot vector with one simple interior knot
  v_knot[0] = v_knot[1] = v_knot[2] = 0.0;
  v_knot[3] = 1.5;
  v_knot[4] = v_knot[5] = v_knot[6] = 2.0;

  // Rational control points can be in either homogeneous
  // or euclidean form. Non-rational control points do not
  // need to specify a weight.  
  Point CV[u_cv_count][v_cv_count];

  int i, j;
  for ( i = 0; i < u_cv_count; i++ ) {
    for ( j = 0; j < v_cv_count; j++ ) {
      CV[i][j].x = i;
      CV[i][j].y = j;
      CV[i][j].z = i-j;
    }
  }

  //////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////

  try {

    // pFile must be a file pointer returned from fopen( , "wb" );
    CRhinoFile RF( pFile, CRhinoFile::write );

    // (optional) Specify some or all of unit system and tolerances
    set_units_and_tolerances_example( RF,
                              CRhinoFile::millimeters,
                              0.01,
                              3.141592653589793238462643*1.5/180.0
                              );

    // (optional) create custom layer(s).  The CRhinoLayer::New() call
    // automatically adds the layers to the Rhino file's layer table.
    CRhinoLayerPtr LayerPtr = 
      CRhinoLayer::New( "Nurbs Surfaces",           // text name
                        CRhinoColor( 200, 0, 255 )  // wire display RGB color
                        );

    //////////////////////////////////////////////////////////////
    //////////////////////////////////////////////////////////////

    // create a surface to write
    CRhinoNurbsSurface S( dim, bIsRational, 
                          u_degree, v_degree,
                          u_cv_count, v_cv_count );

    for ( i = 0; i < S.KnotCount(0); i++ )
      S.SetKnot( 0, i, u_knot[i] );

    for ( j = 0; j < S.KnotCount(1); j++ )
      S.SetKnot( 1, j, v_knot[j] );

    for ( i = 0; i < S.CVCount(0); i++ ) {
      for ( j = 0; j < S.CVCount(1); j++ ) {
        // If the surface is rational, use either
        // S.SetCV( i, j, rh_homogeneous_rational, .. ) or
        // S.SetCV( i, j, rh_euclidean_rational, .. ).

        S.SetCV( i, j, rh_not_rational, &CV[i][j].x );
      }
    }

    //////////////////////////////////////////////////////////////
    //////////////////////////////////////////////////////////////

    // (optional) Put surface on a specific layer.  By default it ends up on
    // Rhino's default layer.
    S.SetLayer( LayerPtr );

    // (optional) Add a label to the object.  By default, this label is "".
    S.SetLabel( "my NURBS surface" );

    {
      // (optional) create a material
      CRhinoMaterial material;
      material.SetDiffuse(  CRhinoColor( 220, 220,   0 ) );
      // The render material name is a string used to identify rendering
      // materials in RIB, POV, OBJ, ..., files.  In Rhino, the render
      // material name is set with the SetObjectMaterial command and can
      // be viewed in the Info tab of the dialog displayed by the
      // Properties command.
      material.SetRenderMaterialName( "NURBS surface rendering material" );

      S.SetMaterial( material );
    }

    S.Write( RF ); // save S in Rhino file
  }

  catch( CRhinoException e ) {
    // if something goes wrong, catch the exception her
    const char* s;
    switch ( e.m_type ) {
    case CRhinoException::unable_to_write:
      s = "unable_to_write";
      break;
    case CRhinoException::unable_to_read:
      s = "unable_to_read";
      break;
    case CRhinoException::unable_to_seek:
      s = "unable_to_seek";
      break;
    case CRhinoException::unexpected_end_of_file:
      s = "unexpected_end_of_file";
      break;
    case CRhinoException::unexpected_value:
      s = "unexpected_value";
      break;
    case CRhinoException::not_supported:
      s = "not_supported";
      break;
    case CRhinoException::illegal_input:
      s = "illegal_input";
      break;
    case CRhinoException::out_of_memory:
      s = "out_of_memory";
      break;
    default:
      s = "unknow exception";
      break;
    }

    printf("Trouble writing file.  CRhinoException type: %s\n", s);
  }
}

static void write_mesh_example( FILE* pFile )
{
  // example demonstrates how to write a mesh

  // Rhino IO toolkit classes throw CRhinoExceptions when
  // trouble arises.  Use try/catch as demonstrated below
  // for clean error trapping.


  //////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////

  // The code between the comment bands has nothing to do with Rhino IO.
  // It is simply an easy way to get a mesh.  The mesh is a pyramid with
  // a square base and 4 trianglular sides.


  //////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////

  try {

    // pFile must be a file pointer returned from fopen( , "wb" );
    CRhinoFile RF( pFile, CRhinoFile::write );

    // (optional) Specify some or all of unit system and tolerances
    set_units_and_tolerances_example( RF,
                              CRhinoFile::feet,
                              0.001,
                              3.141592653589793238462643*3/180.0
                              );

    // (optional) create custom layer(s).  The CRhinoLayer::New() call
    // automatically adds the layers to the Rhino file's layer table.
    CRhinoLayerPtr LayerPtr = 
      CRhinoLayer::New( "My Red Mesh",           // text name
                        CRhinoColor( 255, 0, 0 )  // wire display RGB color
                        );

    //////////////////////////////////////////////////////////////
    //////////////////////////////////////////////////////////////

    // create a mesh to write
    // The mesh is a pyramid with 4 triangular sides and a quadranglar 
    // base.  The mesh has 5 vertices and 5 faces.  
    // The side faces share normals at their common vertices.  The
    // quadrangular base has normals different from the side normal.
    // Coincident vertices that have distinct normals must be
    // duplicated in the vertex list.
    //
    // The apex will be at (1,1.5,4) with normal (0,0,1).
    // The base corners will be at (0,0,0), (0,2,0), (2,3,0), (0,3,0).


    // after construction of a CRhinoMesh, calls to SetNormalsExist(),
    // SetTCoordsExist(), SetNumVertices(), SetNumFaces(), and 
    // SetTCoordDomain() must precede any calls to SetVertex(), 
    // SetNormal(), SetTCoord(), or SetFace().

    CRhinoMesh M;
    M.SetNormalsExist(FALSE);  // we will specify vertex normals
    M.SetTCoordsExist(FALSE); // we will not specify texture coordinates
    M.SetVertexCount(5 + 4);  // 4 duplicates for different base normals
    M.SetFaceCount( 5 );

    // The SetVertex(), SetNormal(), SetTCoord() and SetFace() functions
    // return TRUE if successful and FALSE if input is illegal.  It is
    // a good idea to inspect this returned value.

    // apex location and normal
    M.SetVertex(0,  1.0,  1.5,  5.0 );
    M.SetNormal(0,  0.0,  0.0,  1.0 );

    // SW corner vertex for sides
    M.SetVertex(1,  0.0,  0.0,  0.0 );
    M.SetNormal(1, -1.0, -1.0,  0.0 ); // set normal will unitize if needed

    // SE corner vertex for sides
    M.SetVertex(2,  2.0,  0.0,  0.0 );
    M.SetNormal(2, +1.0, -1.0,  0.0 );

    // NE corner vertex for sides
    M.SetVertex(3,  2.0,  3.0,  0.0 );
    M.SetNormal(3, +1.0, +1.0,  0.0 );

    // NW corner vertex for sides
    M.SetVertex(4,  0.0,  3.0,  0.0 );
    M.SetNormal(4, -1.0, +1.0,  0.0 );


    // SW corner vertex for base
    M.SetVertex(5,  0.0,  0.0,  0.0 ); // == location of v1
    M.SetNormal(5,  0.0,  0.0, -1.0 );

    // SE corner vertex for base
    M.SetVertex(6,  2.0,  0.0,  0.0 ); // == location of v2
    M.SetNormal(6,  0.0,  0.0, -1.0 );

    // SW corner vertex for base
    M.SetVertex(7,  2.0,  3.0,  0.0 ); // == location of v3
    M.SetNormal(7,  0.0,  0.0, -1.0 );

    // SW corner vertex for base
    M.SetVertex(8,  0.0,  3.0,  0.0 ); // == location of v4
    M.SetNormal(8,  0.0,  0.0, -1.0 );

    // faces have vertices ordered counter-clockwise

    // South side triangle
    M.SetFace( 0,   1, 2, 0 );

    // East side triangle
    M.SetFace( 1,   2, 3, 0 );

    // North side triangle
    M.SetFace( 2,   3, 4, 0 );

    // West side triangle
    M.SetFace( 3,   4, 1, 0 );

    // last face is quadrangular base
    M.SetFace( 4,   5, 8, 7, 6 );

    //////////////////////////////////////////////////////////////
    //////////////////////////////////////////////////////////////

    // (optional) Put mesh on a specific layer.  By default it ends up on
    // Rhino's default layer.
    M.SetLayer( LayerPtr );

    // (optional) Add a label to the mesh.  By default, this label is "".
    M.SetLabel( "my pyramid" );

    M.Write( RF ); // save mesh in Rhino file
  }

  catch( CRhinoException e ) {
    // if something goes wrong, catch the exception her
    const char* s;
    switch ( e.m_type ) {
    case CRhinoException::unable_to_write:
      s = "unable_to_write";
      break;
    case CRhinoException::unable_to_read:
      s = "unable_to_read";
      break;
    case CRhinoException::unable_to_seek:
      s = "unable_to_seek";
      break;
    case CRhinoException::unexpected_end_of_file:
      s = "unexpected_end_of_file";
      break;
    case CRhinoException::unexpected_value:
      s = "unexpected_value";
      break;
    case CRhinoException::not_supported:
      s = "not_supported";
      break;
    case CRhinoException::illegal_input:
      s = "illegal_input";
      break;
    case CRhinoException::out_of_memory:
      s = "out_of_memory";
      break;
    default:
      s = "unknow exception";
      break;
    }

    printf("Trouble writing file.  CRhinoException type: %s\n", s);
  }
}

static void write_mesh_with_material_example( FILE* pFile )
{
  // example demonstrates how to write a mesh that uses
  // a rendering material.  You may want to study 
  // write_mesh_example() before examining this function.
  //
  // The key to attaching a texture is to create a mesh
  // with texture coordinates, create a material using the
  // texture as a bitmap, and to attach the material to the
  // mesh before writing the mesh.

  // Rhino IO toolkit classes throw CRhinoExceptions when
  // trouble arises.  Use try/catch as demonstrated below
  // for clean error trapping.


  //////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////

  // The code between the comment bands has nothing to do with 
  // Rhino IO.  It is simply an easy way to get a mesh with
  // texture coordinates.  The example mesh is a simple sixteen
  // facet quad mesh.  In practice,  you can add a texture to a
  // a mesh made from any combination of quads and triangles.

  //////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////

  try {

    // pFile must be a file pointer returned from fopen( , "wb" );
    CRhinoFile RF( pFile, CRhinoFile::write );


    // (optional) Specify some or all of unit system and tolerances
    set_units_and_tolerances_example( RF,
                              CRhinoFile::no_units,
                              0.001,
                              3.141592653589793238462643*2.0/180.0
                              );

    // (optional) create custom layer(s).  The CRhinoLayer::New() call
    // automatically adds the layers to the Rhino file's layer table.
    CRhinoLayerPtr LayerPtr = 
      CRhinoLayer::New( "My Textured Mesh",      // text name
                        CRhinoColor( 255, 0, 0 ) // layer's wire display RGB color
                        );

    // (optional) create a material
    //
    // The render in the current beta version of Rhino has very
    // limited support for rendering materials.  It ignores ambient,
    // emissive, and specular colors.
    //
    // This example demonstrates the code used to write a complete
    // material definition.  At some point in the future, Rhino
    // will use the complete material definition.
    CRhinoMaterial material;
    material.SetAmbient(  CRhinoColor(  40,  40,  40 ) );
    material.SetDiffuse(  CRhinoColor( 220, 220, 220 ) );
    material.SetEmission( CRhinoColor(   0,   0,   0 ) );
    material.SetSpecular( CRhinoColor( 180, 180, 180 ) );

    material.SetShine( 0.35*CRhinoMaterial::MaxShine() ); // 0 = flat
                                                          // MaxShine() = shiney

    material.SetTransparency( 0.2 );  // 0 = opaque, 1 = transparent

    // Texture and bump bitmaps can be Windows bitmap (.BMP), Targa (.TGA),
    // JPEG (.JPG), PCX or PNG files.  Version 1 of Rhino will not support
    // filenames using unicode or multibyte character sets.  As soon as
    // Rhino supports these character sets, the const char* filename will
    // changed to const _TCHAR*.

    // For Rhino to find the texture bitmap, the .3dm file and the
    // .bmp file need to be in the same directory.
    material.SetTextureBitmapFileName( "example_texture.bmp" );
    material.SetBumpBitmapFileName( NULL ); // no bump map in this example
    
    // The render material name is a string used to identify rendering
    // materials in RIB, POV, OBJ, ..., files.  In Rhino, the render
    // material name is set with the SetObjectMaterial command and can
    // be viewed in the Info tab of the dialog displayed by the
    // Properties command.
    material.SetRenderMaterialName( "my render material" );


    //////////////////////////////////////////////////////////////
    //////////////////////////////////////////////////////////////

    // Create a mesh to write.
    //
    // The mesh is a bent rectangle made from a 4 x 7 array of
    // quadrangles. The mesh has 40 vertices and 28 faces.  
    // For the sake of a simple example, this mesh has no normals.
    // In general, a Rhino mesh can have normals, a texture, and a
    // bumpmap.  See write_mesh_example() for an example with normals.


    // after construction of a CRhinoMesh, calls to SetNormalsExist(),
    // SetTCoordsExist(), SetNumVertices(), SetNumFaces(), and 
    // SetTCoordDomain() must precede any calls to SetVertex(), 
    // SetNormal(), SetTCoord(), or SetFace().

    CRhinoMesh M;
    M.SetNormalsExist(FALSE); // no normal in this example
    M.SetTCoordsExist(TRUE);  // we will specify texture coordinates
    double texMin[2] = {0.0,0.0}; // texture domain - lower left corner
    double texMax[2] = {1.0,1.0}; //   of texture bitmap corresponds to
                                  //   texMin and upper right corner of
                                  //   texture bitmap corresponds to
                                  //   texMax.
    M.SetTextureDomain( texMin, texMax );
    M.SetVertexCount( 40 );
    M.SetFaceCount( 28 );

    // The SetVertex(), SetNormal(), SetTCoord() and SetFace() functions
    // return TRUE if successful and FALSE if input is illegal.  It is
    // a good idea to inspect this returned value.

    // cook up a 5 x 8 grid of vertices
    int vertex_index = 0;
    int i, j;
    for ( i = 0; i <= 4; i++ ) {
      for ( j = 0; j <= 7; j++ ) {
        double vertex_x = i;
        double vertex_y = j;
        double vertex_z = sin( 2.0*3.14159*j/7.0 ) + cos( 3.14159*i/4.0 );
        M.SetVertex( vertex_index, vertex_x, vertex_y, vertex_z ); // 3d location

        double tcoord_u = i/4.0; // in range texMin[0] to texMax[0] (set above)
        double tcoord_v = j/7.0; // in range texMin[1] to texMax[1] (set above)
        M.SetTCoord( vertex_index, tcoord_u, tcoord_v ); // 2d texture coordinate

        vertex_index++;
      }
    }

    // faces have vertices ordered counter-clockwise

    // cook up a 4 x 7 grid of quadrangular faces
    int face_index = 0;
    for ( i = 0; i < 4; i++ ) {
      for ( j = 0; j < 7; j++ ) {
        int vi[4]; // indices of corner vertices
        vi[0] = i*8 + j;  // vertex at "lower left" corner
        vi[1] = vi[0]+8;  // vertex at "lower right" corner
        vi[2] = vi[1]+1;  // vertex at "upper left" corner
        vi[3] = vi[0]+1;  // vertex at "upper right" corner
        M.SetFace( face_index, vi[0], vi[1], vi[2], vi[3] );
        face_index++;
      }
    }

    //////////////////////////////////////////////////////////////
    //////////////////////////////////////////////////////////////

    // (optional) Put mesh on a specific layer.  By default it ends up on
    // Rhino's default layer.
    M.SetLayer( LayerPtr );

    // (optional) Add a label to the mesh.  By default, this label is "".
    M.SetLabel( "my textured flag" );

    // (optional) Specify rendering material.
    M.SetMaterial( material );

    M.Write( RF ); // save mesh in Rhino file
  }

  catch( CRhinoException e ) {
    // if something goes wrong, catch the exception her
    const char* s;
    switch ( e.m_type ) {
    case CRhinoException::unable_to_write:
      s = "unable_to_write";
      break;
    case CRhinoException::unable_to_read:
      s = "unable_to_read";
      break;
    case CRhinoException::unable_to_seek:
      s = "unable_to_seek";
      break;
    case CRhinoException::unexpected_end_of_file:
      s = "unexpected_end_of_file";
      break;
    case CRhinoException::unexpected_value:
      s = "unexpected_value";
      break;
    case CRhinoException::not_supported:
      s = "not_supported";
      break;
    case CRhinoException::illegal_input:
      s = "illegal_input";
      break;
    case CRhinoException::out_of_memory:
      s = "out_of_memory";
      break;
    default:
      s = "unknow exception";
      break;
    }

    printf("Trouble writing file.  CRhinoException type: %s\n", s);
  }
}

static void write_spot_light_example( FILE* pFile )
{
  try {

    // pFile must be a file pointer returned from fopen( , "wb" );
    CRhinoFile RF( pFile, CRhinoFile::write );

    // (optional) Specify some or all of unit system and tolerances
    set_units_and_tolerances_example( RF,
                              CRhinoFile::no_units,
                              0.01,
                              3.141592653589793238462643*3.0/180.0
                              );

    // (optional) create custom layer(s).  The CRhinoLayer::New() call
    // automatically adds the layers to the Rhino file's layer table.
    CRhinoLayerPtr LayerPtr = 
      CRhinoLayer::New( "Spot Lights",                // layer text name
                        CRhinoColor( 255, 255, 255 )  // wire display RGB color
                        );

    //////////////////////////////////////////////////////////////
    //////////////////////////////////////////////////////////////

    // create a soft blue spot light
    CRhinoSpotlight spot;

    spot.SetPosition( 2.0, 3.0, 10.0 );
    spot.SetDirection( -1.0, -1.0, -10.0 );
    spot.SetDiffuse( CRhinoColor( 0, 0, 255 ) ); // default is white
    spot.SetSpotExponent( 1.0 );    // 0.0 = hard, 1.0 = soft
    spot.SetSpotCutoff( 30.0 );     // angle in degrees (0 < angle < 90)

    spot.SetLabel( "Blue spot light" ); // object label
        
    //////////////////////////////////////////////////////////////
    //////////////////////////////////////////////////////////////

    // (optional) Put spot light on the "Spot Lights" layer.
    // By default it ends up on Rhino's default layer.
    spot.SetLayer( LayerPtr );

    // (optional) Add a label to the object.  By default, this label is "".
    spot.SetLabel( "my blue spot light" );

    spot.Write( RF ); // save S in Rhino file
  }

  catch( CRhinoException e ) {
    // if something goes wrong, catch the exception her
    const char* s;
    switch ( e.m_type ) {
    case CRhinoException::unable_to_write:
      s = "unable_to_write";
      break;
    case CRhinoException::unable_to_read:
      s = "unable_to_read";
      break;
    case CRhinoException::unable_to_seek:
      s = "unable_to_seek";
      break;
    case CRhinoException::unexpected_end_of_file:
      s = "unexpected_end_of_file";
      break;
    case CRhinoException::unexpected_value:
      s = "unexpected_value";
      break;
    case CRhinoException::not_supported:
      s = "not_supported";
      break;
    case CRhinoException::illegal_input:
      s = "illegal_input";
      break;
    case CRhinoException::out_of_memory:
      s = "out_of_memory";
      break;
    default:
      s = "unknow exception";
      break;
    }

    printf("Trouble writing file.  CRhinoException type: %s\n", s);
  }
}

static void write_viewport_example( FILE* pFile, 
                                    const double sphereCenter[3], double sphereRadius )
{
  // Writes a 7 viewport layout - 3 along the right side,
  // 3 along the left side, and 1 big on in the middle
  // that displays the space inside the sphere.

  // Viewports have a "target" point inside of the view frustum.
  // This target is the center of view rotations.

  // Viewports have a "construction plane".  This plane is
  // (optionally) displayed as a grid.
  //
  try {
    // Used to specify relative position of viewports inside of
    // main Rhino viewing window.
    const double worldO[3] = {0.0,0.0,0.0};
    const double worldpX[3] = {1.0,0.0,0.0};
    const double worldpY[3] = {0.0,1.0,0.0};
    const double worldpZ[3] = {0.0,0.0,1.0};
    const double worldmX[3] = {-1.0,0.0,0.0};
    const double worldmY[3] = {0.0,-1.0,0.0};
    const double worldmZ[3] = {0.0,0.0,-1.0};
    const double pos_x[4]  = {0.0,0.25,0.75,1.0}; 
    const double pos_y[4]  = {0.0,1.0/3.0,2.0/3.0,1.0};
    double camLoc[3], camUp[3], camDir[3];
    double fr_left, fr_right, fr_bottom, fr_top, fr_near, fr_far;
    double target_distance;

    fr_left   = -sphereRadius;
    fr_right  =  sphereRadius;
    fr_bottom = -sphereRadius;
    fr_top    =  sphereRadius;
    fr_near         = 0.001; // Rhino's default
    target_distance = 2.0*sphereRadius;
    fr_far          = 3.0*sphereRadius;  // ignored by Rhino V 1.0

    // pFile must be a file pointer returned from fopen( , "wb" );
    CRhinoFile RF( pFile, CRhinoFile::write );

    CRhinoViewport view[7]; // 7 viewports to fill in

    ////////////////////////////////////////////////////////////////
    //
    // set primary view transformation information first
    view[0].SetProjection( CRhinoViewport::parallel_proj );
    memcpy( camDir, worldmZ, 3*sizeof(camDir[0]) );
    memcpy( camUp , worldpY, 3*sizeof(camUp[0]) );
    camLoc[0] = sphereCenter[0] -target_distance*camDir[0];
    camLoc[1] = sphereCenter[1] -target_distance*camDir[1];
    camLoc[2] = sphereCenter[2] -target_distance*camDir[2];
    view[0].SetCamera( camLoc, camDir, camUp );
    view[0].SetFrustum( fr_left, fr_right, fr_bottom, fr_top, fr_near, fr_far );
    // view[0].SetPort(...); // information is not saved in Rhino file

    // secondary view "fluff"
    view[0].SetViewName("+X+Y parallel");
    //view[0].SetTargetDistance( target_distance ); // SetFrustum() positions th
    view[0].SetPortPosition( pos_x[0], pos_x[1], pos_y[2], pos_y[3] );
    view[0].SetConstructionPlane( worldO, worldpX, worldpY );

    ////////////////////////////////////////////////////////////////
    //
    // set primary view transformation information first
    view[1].SetProjection( CRhinoViewport::parallel_proj );
    memcpy( camDir, worldpY, 3*sizeof(camDir[0]) );
    memcpy( camUp , worldpZ, 3*sizeof(camUp[0]) );
    camLoc[0] = sphereCenter[0] -target_distance*camDir[0];
    camLoc[1] = sphereCenter[1] -target_distance*camDir[1];
    camLoc[2] = sphereCenter[2] -target_distance*camDir[2];
    view[1].SetCamera( camLoc, camDir, camUp );
    view[1].SetFrustum( fr_left, fr_right, fr_bottom, fr_top, fr_near, fr_far );
    // view[1].SetPort(...); // information is not saved in Rhino file

    // secondary view "fluff"
    view[1].SetViewName("+X+Z parallel");
    view[1].SetTargetDistance( target_distance );
    view[1].SetPortPosition( pos_x[0], pos_x[1], pos_y[1], pos_y[2] );
    view[1].SetConstructionPlane( worldO, worldpX, worldpZ );

    ////////////////////////////////////////////////////////////////
    //
    // set primary view transformation information first
    view[2].SetProjection( CRhinoViewport::parallel_proj );
    memcpy( camDir, worldmX, 3*sizeof(camDir[0]) );
    memcpy( camUp , worldpZ, 3*sizeof(camUp[0]) );
    camLoc[0] = sphereCenter[0] -target_distance*camDir[0];
    camLoc[1] = sphereCenter[1] -target_distance*camDir[1];
    camLoc[2] = sphereCenter[2] -target_distance*camDir[2];
    view[2].SetCamera( camLoc, camDir, camUp );
    view[2].SetFrustum( fr_left, fr_right, fr_bottom, fr_top, fr_near, fr_far );
    // view[2].SetPort(...); // information is not saved in Rhino file

    // secondary view "fluff"
    view[2].SetViewName("+Y+Z parallel");
    view[2].SetTargetDistance( target_distance );
    view[2].SetPortPosition( pos_x[0], pos_x[1], pos_y[0], pos_y[1] );
    view[2].SetConstructionPlane( worldO, worldpY, worldpZ );

    ////////////////////////////////////////////////////////////////
    //
    // set primary view transformation information first
    view[3].SetProjection( CRhinoViewport::parallel_proj );
    memcpy( camDir, worldpZ, 3*sizeof(camDir[0]) );
    memcpy( camUp , worldpY, 3*sizeof(camUp[0]) );
    camLoc[0] = sphereCenter[0] -target_distance*camDir[0];
    camLoc[1] = sphereCenter[1] -target_distance*camDir[1];
    camLoc[2] = sphereCenter[2] -target_distance*camDir[2];
    view[3].SetCamera( camLoc, camDir, camUp );
    view[3].SetFrustum( fr_left, fr_right, fr_bottom, fr_top, fr_near, fr_far );
    // view[3].SetPort(...); // information is not saved in Rhino file

    // secondary view "fluff"
    view[3].SetViewName("-X+Y parallel");
    view[3].SetTargetDistance( target_distance );
    view[3].SetPortPosition( pos_x[2], pos_x[3], pos_y[2], pos_y[3] );
    view[3].SetConstructionPlane( worldO, worldmX, worldpY );

    ////////////////////////////////////////////////////////////////
    //
    // set primary view transformation information first
    view[4].SetProjection( CRhinoViewport::parallel_proj );
    memcpy( camDir, worldmY, 3*sizeof(camDir[0]) );
    memcpy( camUp , worldpZ, 3*sizeof(camUp[0]) );
    camLoc[0] = sphereCenter[0] -target_distance*camDir[0];
    camLoc[1] = sphereCenter[1] -target_distance*camDir[1];
    camLoc[2] = sphereCenter[2] -target_distance*camDir[2];
    view[4].SetCamera( camLoc, camDir, camUp );
    view[4].SetFrustum( fr_left, fr_right, fr_bottom, fr_top, fr_near, fr_far );
    // view[4].SetPort(...); // information is not saved in Rhino file

    // secondary view "fluff"
    view[4].SetViewName("-X+Z parallel");
    view[4].SetTargetDistance( target_distance );
    view[4].SetPortPosition( pos_x[2], pos_x[3], pos_y[1], pos_y[2] );
    view[4].SetConstructionPlane( worldO, worldmX, worldpZ );

    ////////////////////////////////////////////////////////////////
    //
    // set primary view transformation information first
    view[5].SetProjection( CRhinoViewport::parallel_proj );
    memcpy( camDir, worldpX, 3*sizeof(camDir[0]) );
    memcpy( camUp , worldpZ, 3*sizeof(camUp[0]) );
    camLoc[0] = sphereCenter[0] -target_distance*camDir[0];
    camLoc[1] = sphereCenter[1] -target_distance*camDir[1];
    camLoc[2] = sphereCenter[2] -target_distance*camDir[2];
    view[5].SetCamera( camLoc, camDir, camUp );
    view[5].SetFrustum( fr_left, fr_right, fr_bottom, fr_top, fr_near, fr_far );
    // view[5].SetPort(...); // information is not saved in Rhino file

    // secondary view "fluff"
    view[5].SetViewName("-Y+Z parallel");
    view[5].SetTargetDistance( target_distance );
    view[5].SetPortPosition( pos_x[2], pos_x[3], pos_y[0], pos_y[1] );
    view[5].SetConstructionPlane( worldO, worldmY, worldpZ );

    ////////////////////////////////////////////////////////////////
    //
    // set primary view transformation information first
    target_distance = 10.0*sphereRadius;
    const double tan_half_angle = sphereRadius / target_distance;

    view[6].SetProjection( CRhinoViewport::perspective_proj );
    camDir[0] = -40.0;
    camDir[1] =  75.0;
    camDir[2] = -50.0;
    memcpy( camUp , worldpZ, 3*sizeof(camUp[0]) );
    double d = RhinoVectorLength( 3, camDir );
    camLoc[0] = sphereCenter[0] -target_distance*camDir[0]/d;
    camLoc[1] = sphereCenter[1] -target_distance*camDir[1]/d;
    camLoc[2] = sphereCenter[2] -target_distance*camDir[2]/d;
    view[6].SetCamera( camLoc, camDir, camUp );
    fr_near   = (target_distance - sphereRadius)/10.0;
    fr_far    = target_distance + 1.5*sphereRadius; // ignored by Rhino V 1.0
    d         = fr_near*tan_half_angle;
    fr_left   = -d;
    fr_right  =  d;
    fr_bottom = -d;
    fr_top    =  d;
    view[6].SetFrustum( fr_left, fr_right, fr_bottom, fr_top, fr_near, fr_far );
    // view[6].SetPort(...); // information is not saved in Rhino file

    // secondary view "fluff"
    view[6].SetViewName("skew perspective");
    view[6].SetTargetDistance( target_distance );
    view[6].SetPortPosition( pos_x[1], pos_x[2], pos_y[0], pos_y[3] );
    view[6].SetConstructionPlane( worldO, worldpX, worldpY );


    view[0].Write( RF ); // save view in Rhino file
    view[1].Write( RF ); // save view in Rhino file
    view[2].Write( RF ); // save view in Rhino file
    view[3].Write( RF ); // save view in Rhino file
    view[4].Write( RF ); // save view in Rhino file
    view[5].Write( RF ); // save view in Rhino file
    view[6].Write( RF ); // save view in Rhino file
  }

  catch( CRhinoException e ) {
    // if something goes wrong, catch the exception her
    const char* s;
    switch ( e.m_type ) {
    case CRhinoException::unable_to_write:
      s = "unable_to_write";
      break;
    case CRhinoException::unable_to_read:
      s = "unable_to_read";
      break;
    case CRhinoException::unable_to_seek:
      s = "unable_to_seek";
      break;
    case CRhinoException::unexpected_end_of_file:
      s = "unexpected_end_of_file";
      break;
    case CRhinoException::unexpected_value:
      s = "unexpected_value";
      break;
    case CRhinoException::not_supported:
      s = "not_supported";
      break;
    case CRhinoException::illegal_input:
      s = "illegal_input";
      break;
    case CRhinoException::out_of_memory:
      s = "out_of_memory";
      break;
    default:
      s = "unknow exception";
      break;
    }

    printf("Trouble writing file.  CRhinoException type: %s\n", s);
  }
}

static void make_trimming_curves( CRhinoBrep& brep, 
                                  double u0, double v0, // start point in parameter space
                                  double u1, double v1  // end point in parameter space
                                  )
{
  double P[3] = {0.0,0.0,0.0};

  CRhinoNurbsCurve* p2dCurve = new CRhinoNurbsCurve(2,FALSE,1,2);
  CRhinoNurbsCurve* p3dCurve = new CRhinoNurbsCurve(3,FALSE,1,2);

  // 2d curve's range is in surface's [0,1]x[0,1] parameter space
  P[0] = u0;
  P[1] = v0;
  p2dCurve->SetCV(0,rh_not_rational,P);
  P[0] = u1;
  P[1] = v1;
  p2dCurve->SetCV(1,rh_not_rational,P);
  p2dCurve->SetKnot( 0, 0.0 );
  p2dCurve->SetKnot( 1, 1.0 );

  // 3d curve's represent the composite map surface(2dcurve(t))
  // Since the surface is a 10x10 plane at z=3, we can just write down
  // the answer.
  P[0] = 10.0*u0;
  P[1] = 10.0*v0;
  P[2] = 3.0;
  p3dCurve->SetCV(0,rh_not_rational,P);
  P[0] = 10.0*u1;
  P[1] = 10.0*v1;
  P[2] = 3.0;
  p3dCurve->SetCV(1,rh_not_rational,P);
  p3dCurve->SetKnot( 0, 0.0 );
  p3dCurve->SetKnot( 1, 1.0 );

  CRhinoAgCurve* p2dAgCurve = new CRhinoAgCurve();
  p2dAgCurve->AddNurbsCurve(p2dCurve);
  brep.m_C2.Append(p2dAgCurve);

  CRhinoAgCurve* p3dAgCurve = new CRhinoAgCurve();
  p3dAgCurve->AddNurbsCurve(p3dCurve);
  brep.m_C3.Append(p3dAgCurve);
}

static int make_vertex( CRhinoBrep& brep, double x, double y, double z )
{
  const int vi = brep.m_V.Size(); // index of new vertex
  brep.m_V.Append(CRhinoBrepVertex(vi)); // create a new vertex

  CRhinoBrepVertex& vertex = brep.m_V[vi];

  vertex.x = x;
  vertex.y = y;
  vertex.z = z;
  vertex.m_tolerance = 0.0; // Since this simple example is exact,
                            // the tolerance is 0.0.  See the comments
                            // in the definition of CRhinoBrepVertex
                            // for more details.

  return vi;
}

static int make_edge( CRhinoBrep& brep, int v0i, int v1i, int c3i )
{
  const int ei = brep.m_E.Size(); // index of new edge
  brep.m_E.Append( CRhinoBrepEdge(ei) ); // create new edge

  CRhinoBrepEdge& edge = brep.m_E[ei];

  edge.m_c3i = c3i;

  // in this simple example, the edge uses the entire 3d curve
  brep.m_C3[edge.m_c3i]->Domain( &edge.m_t[0], &edge.m_t[1] );

  edge.m_vi[0] = v0i;
  edge.m_vi[1] = v1i;

  // vertices need to know which edges use them
  brep.m_V[edge.m_vi[0]].m_ei.Append(ei);
  brep.m_V[edge.m_vi[1]].m_ei.Append(ei);

  edge.m_tolerance = 0.0; // Since this simple example is exact,
                          // the tolerance is 0.0.  See the comments
                          // in the definition of CRhinoBrepVertex
                          // for more details.

  return ei;
}

static int make_trim( CRhinoBrep& brep, int li, int v0i, int v1i, int c2i, int ei )
{
  const int ti = brep.m_T.Size();
  brep.m_T.Append(CRhinoBrepTrim(ti)); // create new trim;

  CRhinoBrepTrim& trim = brep.m_T[ti];

  trim.m_c2i = c2i;

  // in this simple example, the trim uses the entire 2d curve
  brep.m_C2[trim.m_c2i]->Domain( &trim.m_t[0], &trim.m_t[1] );
  
  trim.m_ei = ei; // use -1 if the trim lies along a singular surface side
  if ( ei >= 0 ) {
    brep.m_E[ei].m_ti.Append(ti);
  }

  trim.m_vi[0] = v0i;
  trim.m_vi[1] = v1i;
  trim.m_bRev3d = FALSE; // 3d edge orientation agrees with surface(2dcurve(t))

  // In this simple example, all trims are boundaries.  See example_brep.cpp
  // for an example with mated trims.
  trim.m_type = CRhinoBrepTrim::boundary;

  trim.m_iso = CRhinoBrepTrim::not_iso;

  trim.m_li = li;

  // Since this simple example is exact, the tolerances are 0.0.  See the comments
  // in the definition of CRhinoBrepTrim for more details.
  trim.m_tolerance[0] = 0.0;
  trim.m_tolerance[1] = 1.0;

  return ti;
}


static void write_trimmed_surface_example( FILE* pFile )
{
  // trimmed surfaces are written as a CRhinoBrep that has
  // a single surface and a single CRhinoBrepFace.
  //
  // Trimming loops are simple closed curves and are oriented
  // so that the active portion of the trimmed surface's
  // domain lies to the left of the trimming curves.

  CRhinoBrep brep;

  // Create a 10x10 plane surface at z=3 with domain [0,1]x[0,1]
  CRhinoNurbsSurface* pSurface = new CRhinoNurbsSurface(3,FALSE,1,1,2,2);
  {
    double P[4][3] = { {0.0,0.0,3.0}, {10.0,0.0,3.0}, {10.0,10.0,3.0}, {0.0,10.0,3.0} };
    pSurface->SetCV(0,0,rh_not_rational,P[0]);
    pSurface->SetCV(1,0,rh_not_rational,P[1]);
    pSurface->SetCV(1,1,rh_not_rational,P[2]);
    pSurface->SetCV(0,1,rh_not_rational,P[3]);
    // domain is [0,1]x[0,1]
    pSurface->SetKnot(0,0,0.0);
    pSurface->SetKnot(0,1,1.0);
    pSurface->SetKnot(1,0,0.0);
    pSurface->SetKnot(1,1,1.0);
  }
  const int si = brep.m_S.Size(); // index of surface
  brep.m_S.Append(pSurface);

  // create simple trimming triangle
  make_trimming_curves( brep,  0.1, 0.2,   0.9, 0.15 );
  make_trimming_curves( brep,  0.9, 0.15,  0.7, 0.8  );
  make_trimming_curves( brep,  0.7, 0.8,   0.1, 0.2  );

  // there are vertices at the 3 corners
  make_vertex( brep, 1.0, 2.0, 3.0  );
  make_vertex( brep, 9.0, 1.5, 3.0 );
  make_vertex( brep, 7.0, 8.0, 3.0 );

  // there are 3 edges along the sides of the triangle
  make_edge( brep, 0, 1, 0 ); // start vertex, end vertex, 3d curve index
  make_edge( brep, 1, 2, 1 ); // start vertex, end vertex, 3d curve index
  make_edge( brep, 2, 0, 2 ); // start vertex, end vertex, 3d curve index

  // outer boundary trimming loops
  const int li = brep.m_L.Size(); // index of outer boundary loop
  brep.m_L.Append( CRhinoBrepLoop( li ) ); // create new loop
  CRhinoBrepLoop& loop = brep.m_L[li];

  loop.m_type = CRhinoBrepLoop::outer;

  // geometrically, loops are made from a contiguous list of 2d parameter space
  // curves that form a simple closed curve.
  {
    int ti;
    ti = make_trim( brep, li, 0, 1, 0, 0 ); // loop index, start vertex, end vertex, 2d curve indes, edge index
    loop.m_ti.Append(ti);
    ti = make_trim( brep, li, 1, 2, 1, 1 ); // loop index, start vertex, end vertex, 2d curve indes, edge index
    loop.m_ti.Append(ti);
    ti = make_trim( brep, li, 2, 0, 2, 2 ); // loop index, start vertex, end vertex, 2d curve indes, edge index
    loop.m_ti.Append(ti);
  }

  const int fi = brep.m_F.Size(); // index of face
  brep.m_F.Append( CRhinoBrepFace( fi ) ); // create new loop
  CRhinoBrepFace& face = brep.m_F[fi];
  face.m_li.Append(li);
  brep.m_L[li].m_fi = fi;
  face.m_si = si;
  face.m_bRev = FALSE; // surface normal agrees with face normal

  // when debugging your code, IsValid(), IsSolid(), IsManifold() are useful
  // to check.
  if ( !brep.IsValid() ) {
    printf("brep is not valid\n");
    return;
  }
  BOOL bIsManifold, bIsOriented, bHasBoundary, bIsSolid;
  bIsManifold = brep.IsManifold( &bIsOriented, &bHasBoundary );
  bIsSolid = brep.IsSolid();
  
  {
    // (optional) create a material
    //
    // The render in the current beta version of Rhino has very
    // limited support for rendering materials.  It ignores ambient,
    // emissive, and specular colors.
    //
    // This example demonstrates the code used to write a complete
    // material definition.  At some point in the future, Rhino
    // will use the complete material definition.
    CRhinoMaterial material;
    material.SetAmbient(  CRhinoColor(  40,  40,  40 ) );
    material.SetDiffuse(  CRhinoColor( 220, 220, 220 ) );
    material.SetEmission( CRhinoColor(   0,   0,   0 ) );
    material.SetSpecular( CRhinoColor( 180, 180, 180 ) );
    material.SetShine( 0.35*CRhinoMaterial::MaxShine() ); // 0 = flat
                                                          // MaxShine() = shiney
    material.SetTransparency( 0.2 );  // 0 = opaque, 1 = transparent

    // The render material name is a string used to identify rendering
    // materials in RIB, POV, OBJ, ..., files.  In Rhino, the render
    // material name is set with the SetObjectMaterial command and can
    // be viewed in the Info tab of the dialog displayed by the
    // Properties command.
    material.SetRenderMaterialName( "some trimmed surface rendering material" );

    brep.SetMaterial( material );
  }

  try {
    CRhinoFile RF( pFile, CRhinoFile::write );

    // (optional) assign rendering material properties
    brep.Write( RF );
  }

  catch( CRhinoException e ) {
    // if something goes wrong, catch the exception her
    const char* s;
    switch ( e.m_type ) {
    case CRhinoException::unable_to_write:
      s = "unable_to_write";
      break;
    case CRhinoException::unable_to_read:
      s = "unable_to_read";
      break;
    case CRhinoException::unable_to_seek:
      s = "unable_to_seek";
      break;
    case CRhinoException::unexpected_end_of_file:
      s = "unexpected_end_of_file";
      break;
    case CRhinoException::unexpected_value:
      s = "unexpected_value";
      break;
    case CRhinoException::not_supported:
      s = "not_supported";
      break;
    case CRhinoException::illegal_input:
      s = "illegal_input";
      break;
    case CRhinoException::out_of_memory:
      s = "out_of_memory";
      break;
    default:
      s = "unknow exception";
      break;
    }

    printf("Trouble writing file.  CRhinoException type: %s\n", s);
  }
}

int main ( int argc, const char* argv[] )
{

  // If you want to learn to write b-rep models, first work through
  // this example paying close attention to write_trimmed_surface_example(),
  // then examime example_brep.cpp.

  FILE* pFile;

  pFile = fopen( "my_points.3dm", "wb" );
  write_point_set_example( pFile );
  fclose( pFile );

  pFile = fopen( "my_nurbs_curve.3dm", "wb" );
  write_nurbs_crv_example( pFile );
  fclose( pFile );

  pFile = fopen( "my_nurbs_surface.3dm", "wb" );
  write_nurbs_srf_example( pFile );
  fclose( pFile );

  pFile = fopen( "my_mesh.3dm", "wb" );
  write_mesh_example( pFile );
  fclose( pFile );
  
  pFile = fopen( "my_mesh_with_material.3dm", "wb" );
  write_mesh_with_material_example( pFile );
  fclose( pFile );
  
  pFile = fopen( "my_spot_light.3dm", "wb" );
  write_spot_light_example( pFile );
  fclose( pFile );
  
  pFile = fopen( "my_viewports.3dm", "wb" );
  double sphereCenter[3] = {0.0,0.0,0.0};
  double sphereRadius = 10.0;
  write_viewport_example( pFile, sphereCenter, sphereRadius );
  fclose( pFile );
  
  pFile = fopen( "my_trimmed_surface.3dm", "wb" );
  write_trimmed_surface_example( pFile );
  fclose( pFile );

  RhinoIOCleanup();  // see comments in rhinoio_chunk.h before using

  return 0;
}
