/* $Header: /Src/RhinoIO/rhinoio_agcurve.cpp 25    11/10/99 2:56p Dalelear $ */
/* $NoKeywords: $ */
//
// Copyright (c) 1993-1998 Robert McNeel & Associates. All rights reserved.
// Rhinoceros is a registered trademark of Robert McNeel & Assoicates.
//
// THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
// ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
// MERCHANTABILITY ARE HEREBY DISCLAIMED.
//				
////////////////////////////////////////////////////////////////

#include "rhinoio_object.h"
#include "rhinoio_nurbscrv.h"
#include "rhinoio_agspline.h"
#include "rhinoio_agcurve.h"

RHINO_CHUNK_IMPLEMENT( CRhinoAgCurve, TCODE_AG_CURVE )

void CRhinoAgCurve::Initialize(void)
{
  m_dim = 0;
  m_is_closed = 0;
}

CRhinoAgCurve::CRhinoAgCurve() : CRhinoObject( TCODE_AG_CURVE ) 
{
  Initialize();
}

CRhinoAgCurve::~CRhinoAgCurve( )
{
}

BOOL 
CRhinoAgCurve::IsClosed() const
{
  BOOL bIsClosed = FALSE;
  if ( NurbsCurveCount() == 1 )
    bIsClosed = m_nurbs_curve[0]->IsClosed();
  else {
    double d[6];
    const int dim = Dimension();
    double* P0 = (dim <= 3) ? d : (double*)rhmalloc( dim*2*sizeof(*P0) );
    double* P1 = P0 + dim;
    EndPoint( 0, P0 );
    EndPoint( 1, P1 );
    bIsClosed = RhinoIsEqual( dim, P0, P1 );
    if ( dim > 3 )
      rhfree(P0);
  }
  return bIsClosed;
}

BOOL 
CRhinoAgCurve::IsPeriodic() const
{
  BOOL bIsPeriodic = FALSE;
  if ( NurbsCurveCount() == 1 ) {
    bIsPeriodic = m_nurbs_curve[0]->IsPeriodic();
  }
  return bIsPeriodic;
}

CRhinoFile::eStatus
CRhinoAgCurve::Read( CRhinoFile& file)
{
  // read attributes
  CRhinoFile::eStatus status = ReadAttributes( file, TCODE_AG_CURVESTUFF );
  if ( status != CRhinoFile::good )
    return status;

  unsigned data_typecode = 0;
  int data_length = 0;
  do {
    status = file.BeginReadChunk( data_typecode, data_length );
    if ( status != CRhinoFile::good )
      return status;
    status = CRhinoFile::crc_error;
    if ( data_typecode != TCODE_AG_CURVESTUFF || data_length <= 0 )
      break;

    char c;
    short s;
    int nbs = 0;

    /* read AG_CURVE flags */
    file.ReadByte(1,&c);  
    if ( c != 2 && c != 3 )
      return CRhinoFile::crc_error;
    m_dim = c;
    file.ReadByte(1,&c);
    if ( c != 0 && c != 1 && c != 2 )
      return CRhinoFile::crc_error;
    m_is_closed = (c) ? 1 : 0;
    file.ReadShort(1,&s);
    if ( s < 1 )
      return CRhinoFile::crc_error;
    nbs = s;
    AddNurbsCurveCapacity( nbs );

    // read AG_CURVE bounding box
    file.ReadDouble( m_dim, m_bounding_box[0] );
    file.ReadDouble( m_dim, m_bounding_box[1] );

    // read AG_SPLINES
    BOOL bIsPolyline = TRUE;
    status = CRhinoFile::good;
    int i;
    for ( i = 0; i < nbs && status == CRhinoFile::good; i++ ) {
      unsigned int typecode;
      int length;
      status = file.BeginReadChunk( typecode, length );
      if ( status != CRhinoFile::good )
        break;
      if ( typecode != TCODE_AG_SPLINE || length < 1 )
        break;
      //m_nurbs_curve.assign(i,new CRhinoAgSpline());
      CRhinoAgSpline* pBs = new CRhinoAgSpline();
      if ( !pBs ) {
        throw CRhinoException( CRhinoException::out_of_memory );
      }
      status = pBs->Read( file );
      CRhinoFile::eStatus end_status = file.EndChunk();
      if ( status == CRhinoFile::good )
        status = end_status;
      m_nurbs_curve.Append(pBs);
      if ( pBs->IsRational() || pBs->Degree() != 1 )
        bIsPolyline = FALSE;
    }
    if ( status == CRhinoFile::good ) {
      if ( i != nbs )
        status = CRhinoFile::crc_error;
      else
        status = IsValid() ? CRhinoFile::good : CRhinoFile::crc_error;
      if ( status == CRhinoFile::good && m_nurbs_curve.Size() > 1 && bIsPolyline ) {
        // merge all line segments into a single polyline
        nbs = m_nurbs_curve.Size();
        int point_count = 1-nbs;
        for ( i = 0; i < nbs; i++ ) {
          point_count += m_nurbs_curve[i]->CVCount();
        }
        if ( point_count >= nbs+1 ) {
          CRhinoAgSpline* pBs = new CRhinoAgSpline( Dimension(), FALSE, 1, point_count );
          pBs->SetCV( 0, rh_not_rational, m_nurbs_curve[0]->CV(0) );
          pBs->SetKnot( 0, m_nurbs_curve[0]->Knot(0) );
          point_count = 1;
          int j, jmax;
          double k;
          for ( i = 0; i < nbs; i++ ) {
            jmax = m_nurbs_curve[i]->CVCount();
            for ( j = 1; j < jmax; j++ ) {
              k = m_nurbs_curve[i]->Knot(j);
              if ( k <= pBs->Knot(point_count-1) )
                continue;
              pBs->SetCV( point_count, rh_not_rational, m_nurbs_curve[i]->CV(j) );
              pBs->SetKnot( point_count, k );
              point_count++;
            }
          }
          if ( point_count < 2 ) {
            delete pBs;
          }
          else {
            // remove bag-o-line segments
            for ( i = nbs-1; i >= 0; i-- ) {
              delete m_nurbs_curve[i];
              m_nurbs_curve[i] = NULL;
            }
            m_nurbs_curve.Empty();

            // add new polyline replacement
            m_nurbs_curve.Append(pBs);
            i = IsValid();
          }
        }
      }
    }
  } while ( FALSE );

  CRhinoFile::eStatus end_status = file.EndChunk();
  if ( status == CRhinoFile::good )
    status = end_status;

  return status;
}

void
CRhinoAgCurve::Write(CRhinoFile& file) const
{
  if ( !IsValid() ) {
    throw CRhinoException( CRhinoException::illegal_input );
  }
  // write CRhinoAgCurve as CRhinoNurbsCurve's to avoid getting 
  // AG_CURVE structures that do not comply with AG restrictions.
  int i;
  for ( i = 0; i < m_nurbs_curve.Size(); i++ ) {
    m_nurbs_curve[i]->Write(file);
  }
}

BOOL
CRhinoAgCurve::IsValid(void) const
{
  if ( m_nurbs_curve.Size() < 1 )
    return FALSE;
  if ( m_is_closed < 0 || m_is_closed > 1 )
    return FALSE;
  int i;
  for ( i = 0; i < m_nurbs_curve.Size(); i++ ) {
    if( !m_nurbs_curve[i] )
      return FALSE;
    if ( m_dim != m_nurbs_curve[i]->Dimension() )
      return FALSE;
    if ( !m_nurbs_curve[i]->IsValid() )
      return FALSE;
  }
  return TRUE;
}

BOOL
CRhinoAgCurve::EndPoint( int end_index, double* P ) const
{
  // Quick hack to get edge ends - doesn't work for 
  // periodic NURBS (which are not used in AG_TEDGEs).
  // When time permits, I'll add an evaluator to the
  // CRhinoNurbsCurve class.
  BOOL rc = FALSE;
  CRhinoNurbsCurve* pNurbsCurve = NULL;
  switch ( end_index ) {
  case 0:
    pNurbsCurve = m_nurbs_curve[0];
    if ( pNurbsCurve )
      rc = pNurbsCurve->GetCV(0,rh_not_rational,P);
    break;
  case 1:
    pNurbsCurve = *m_nurbs_curve.Last();
    if ( pNurbsCurve )
      rc = pNurbsCurve->GetCV(pNurbsCurve->CVCount()-1,rh_not_rational,P);
    break;
  }
  return rc;
}

BOOL
CRhinoAgCurve::Domain( double* t0, double* t1 ) const
{
  BOOL rc = FALSE;
  CRhinoNurbsCurve* pNurbsCurve = NULL;
  if ( t0 ) {
    pNurbsCurve = m_nurbs_curve[0];
    if ( pNurbsCurve )
      rc = pNurbsCurve->Domain(t0,NULL);
  }
  if ( t1 ) {
    pNurbsCurve = *m_nurbs_curve.Last();
    if ( pNurbsCurve )
      rc = pNurbsCurve->Domain( NULL, t1 );
  }
  return rc;
}

BOOL
CRhinoAgCurve::GetBoundingBox(
          double box_min[3],
          double box_max[3],
          BOOL bGrowBox     // = FALSE
          ) const
{
  BOOL rc = TRUE;
  int bsi;
  for ( bsi = 0; bsi < m_nurbs_curve.Size(); bsi++ ) {
    if ( !m_nurbs_curve[bsi] )
      continue;
    if ( !m_nurbs_curve[bsi]->GetBoundingBox( box_min, box_max, bGrowBox ) ) {
      rc = FALSE;
      continue;
    }
    bGrowBox = TRUE;
  }
  return rc;
}

BOOL
CRhinoAgCurve::Reverse()
{
  BOOL rc = FALSE;
  const int bs_count = m_nurbs_curve.Size();
  int bsi;
  for ( bsi = 0; bsi < bs_count; bsi++ ) {
    if ( !m_nurbs_curve[bsi] ) {
      rc = FALSE;
    }
    else {
      rc = m_nurbs_curve[bsi]->Reverse();
    }
    if ( !rc ) {
      // undo any changes
      while ( --bsi >= 0 ) {
        m_nurbs_curve[bsi]->Reverse();
      }
      return FALSE;
    }
  }
  m_nurbs_curve.Reverse();
  return rc;
}

BOOL
CRhinoAgCurve::SwapCoordinates(int i, int j )
{
  BOOL rc = FALSE;
  const int bs_count = m_nurbs_curve.Size();
  int bsi;
  for ( bsi = 0; bsi < bs_count; bsi++ ) {
    if ( !m_nurbs_curve[bsi] ) {
      rc = FALSE;
    }
    else {
      rc = m_nurbs_curve[bsi]->SwapCoordinates(i,j);
    }
    if ( !rc ) {
      // undo any changes
      while ( --bsi >= 0 )
        m_nurbs_curve[bsi]->SwapCoordinates(i,j);
    }
  }
  return rc;
}

//int CRhinoAgCurve::AgSplineCount() const
//{ // OBSOLETE - use NurbsCurveCount()
//  return m_nurbs_curve.Size(); 
//}

int CRhinoAgCurve::NurbsCurveCount() const
{ 
  return m_nurbs_curve.Size(); 
}

CRhinoNurbsCurve* CRhinoAgCurve::operator[] (int index) const
{ 
  return (index < m_nurbs_curve.Size()) ? m_nurbs_curve[index] : NULL; 
}

void CRhinoAgCurve::AddNurbsCurveCapacity( int capacity )
{
  m_nurbs_curve.Reserve(capacity);
}

BOOL CRhinoAgCurve::AddNurbsCurve(CRhinoNurbsCurve* p)
{
  BOOL rc = FALSE;
  if (p) {
    const int spline_dim = p->Dimension();
    if ( spline_dim == 2 || spline_dim == 3 ) {
      if ( m_nurbs_curve.Size() == 0 ) {
        m_dim = spline_dim;
        p->GetBoundingBox( m_bounding_box[0], m_bounding_box[1], FALSE );
      }
      if ( m_dim == spline_dim ) {
        m_nurbs_curve.Append(p);
        p->GetBoundingBox( m_bounding_box[0], m_bounding_box[1], TRUE );
        rc = TRUE;
      }
    }
  }
  return rc;
}

int CRhinoAgCurve::Dimension(void) const
{
  return m_dim; 
}
