/* $Header: /Src/RhinoIO/rhinoio_agface.cpp 14    11/10/99 2:56p Dalelear $ */
/* $NoKeywords: $ */
//
// Copyright (c) 1993-1998 Robert McNeel & Associates. All rights reserved.
// Rhinoceros is a registered trademark of Robert McNeel & Assoicates.
//
// THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
// ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
// MERCHANTABILITY ARE HEREBY DISCLAIMED.
//				
////////////////////////////////////////////////////////////////
 
#include "rhinoio_brep.h"
#include "rhinoio_agface.h"

RHINO_CHUNK_IMPLEMENT( CRhinoAgFace, TCODE_AG_FACE )

CRhinoFile::eStatus
CRhinoAgFace::Read( CRhinoFile& file)
{
  CRhinoFile::eStatus status = ReadAgFace( file );
  if ( status == CRhinoFile::good ) {
    SetVertices();
    SetTolsFromFileValues();
  }
  return status;
}

CRhinoFile::eStatus
CRhinoBrep::ReadAgFace( CRhinoFile& file, BOOL bSkipAttributes /* = FALSE */ )
{
  CRhinoWorkSpace ws;

  const int ti0 = m_T.Size(); // needed to glue seams together

  CRhinoFile::eStatus status;
  if ( bSkipAttributes ) {
    status = SkipToAG_STUFF( file, TCODE_AG_FACESTUFF );
  }
  else {
    // read attributes
    status = ReadAttributes( file, TCODE_AG_FACESTUFF );
  }
  if ( status != CRhinoFile::good )
    return status;

  unsigned int data_typecode = 0;
  int data_length = 0;

  do {
    status = file.BeginReadChunk( data_typecode, data_length );
    if ( status != CRhinoFile::good )
      return status;
    status = CRhinoFile::crc_error;
    if ( data_typecode != TCODE_AG_FACESTUFF || data_length <= 0 )
      break;

    int flipnorm = 0;
    int ftype = 0;
    int bndcnt = 0;
    int twincnt = 0;
    BOOL bHasOuter = FALSE;

    /* read AG_FACE flags */
    file.ReadInteger(1,&flipnorm);  
    if ( flipnorm < 0 || flipnorm > 1 )
      return CRhinoFile::crc_error;
    file.ReadInteger(1,&ftype);  
    file.ReadInteger(1,&bndcnt);  
    bHasOuter = (bndcnt%2);
    bndcnt /= 2;

    // read AG_FACE bounding box
    file.ReadDouble( 3, m_bounding_box[0] );
    file.ReadDouble( 3, m_bounding_box[1] );

    /* B-rep edge gluing info */
    file.ReadInteger(1,&twincnt);
    short* glue = (twincnt > 0 ) ? (short*)ws.GetMemory(twincnt*sizeof(*glue)) : NULL;
    if (twincnt > 0)
      file.ReadShort(twincnt,glue);

    // read AG_SURFACE
    int srf_index = -1;
    status = Read_AG_SURFACE( file, &srf_index );
    if ( status != CRhinoFile::good )
      break;
    
    // create face
    const int fi = m_F.Size();
    m_F.Append( CRhinoBrepFace(fi) );
    CRhinoBrepFace& face = m_F[fi];
    face.m_bRev = (flipnorm) ? TRUE : FALSE;
    face.m_si = srf_index;
    face.m_li.Reserve(bndcnt);

    // read AG_BOUNDARIES
    int loop_index = -1;
    if ( !bHasOuter ) {
      // TODO: cook up outer boundary loop 
      face.m_li.Append(loop_index);
    }
    int bi;
    for ( bi = 0; bi < bndcnt; bi++ ) {
      loop_index = -1;
      status = Read_AG_BOUNDARY( file, &loop_index );
      if ( status != CRhinoFile::good )
        break;
      face.m_li.Append( loop_index );
      m_L[loop_index].m_fi = m_F.Size()-1;
    }

    if ( status != CRhinoFile::good )
      break;

    if ( twincnt > 0 ) {
      // twincnt = number of seams edges in face
      // glue[] = order 2 permutation of {0,....,twincnt-1}

      // set seam_i[] = m_T[] indices of seam trims
      int si, ti;
      const int ti1 = m_T.Size();
      int* seam_i = (int*)ws.GetMemory(twincnt*sizeof(*seam_i));
      for ( ti = ti0, si = 0; ti < ti1 && si < twincnt; ti++ ) {
        if (m_T[ti].m_type != CRhinoBrepTrim::seam )
          continue;
        seam_i[si++] = ti;
      }

      if ( si == twincnt ) {
        // glue seams
        for ( si = 0; si < twincnt; si++ ) {
          const int i0 = seam_i[si];
          const int i1 = seam_i[glue[si]];
          // m_T[i0] and m_t[i1] use the same edge;
          const int ei0 = m_T[i0].m_ei;
          const int ei1 = m_T[i1].m_ei;
          if ( ei0 == -1 && ei1 >= 0 ) {
            m_T[i0].m_ei = ei1;
            m_E[ei1].m_ti.Append(i0);
          }
          else if ( ei1 == -1 && ei0 >= 0 ) {
            m_T[i1].m_ei = ei0;
            m_E[ei0].m_ti.Append(i1);
          }
        }
      }
    }

    // set face's trim's CRhinoBrepTrim::m_iso flag
    SetTrimIsoFlag( fi );

  } while ( FALSE );

  CRhinoFile::eStatus end_status = file.EndChunk();
  if ( status == CRhinoFile::good )
    status = end_status;

  return status;
}

void
CRhinoAgFace::Write(CRhinoFile& file) const
{
  CRhinoBrep::Write( file );
}

BOOL
CRhinoAgFace::IsValid(void) const
{
  if ( m_S.Size() != 1 )
    return FALSE;
  if ( m_F.Size() != 1 )
    return FALSE;
  const CRhinoBrepFace& face = m_F[0];
  if ( face.m_si != 0 )
    return FALSE;
  return CRhinoBrep::IsValid();
}
