/* $Header: /Src/RhinoIO/rhinoio_agspline.cpp 11    10/20/99 1:27p Dalelear $ */
/* $NoKeywords: $ */
//
// Copyright (c) 1993-1998 Robert McNeel & Associates. All rights reserved.
// Rhinoceros is a registered trademark of Robert McNeel & Assoicates.
//
// THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
// ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
// MERCHANTABILITY ARE HEREBY DISCLAIMED.
//				
////////////////////////////////////////////////////////////////

#include "rhinoio_object.h"
#include "rhinoio_nurbscrv.h"
#include "rhinoio_agspline.h"

RHINO_CHUNK_IMPLEMENT( CRhinoAgSpline, TCODE_AG_SPLINE )

void CRhinoAgSpline::Initialize( void )
{
  // ALERT: SetTypeCode() is used only by CRhinoAgSpline and CRhinoAgSurface
  // All other geometry constructors should be derived from CRhinoObject
  // and call : CRhinoObject( TCODE_* ) during construction.
  SetTypeCode( TCODE_AG_SPLINE );
  m_is_closed = 0;
  m_form = 0;
}

// Initialize an empty CRhinoAgSpline
CRhinoAgSpline::CRhinoAgSpline()
{
  Initialize();
}

// Initialize an empty CRhinoNurbsCurve
CRhinoAgSpline::CRhinoAgSpline(
    int   dimension,    // dimension ( >= 1 and usually = 3 )
    BOOL  bIsRational,  // TRUE if curve is rational
    int   degree,       // degree ( >= 1 )
    int   cv_count      // number of CVs ( >= degree+1 )
  ) 
  : CRhinoNurbsCurve( dimension, bIsRational, degree, cv_count )
{
  Initialize();
}

CRhinoFile::eStatus
CRhinoAgSpline::Read( CRhinoFile& file)
{
  // read attributes
  CRhinoFile::eStatus status = ReadAttributes( file, TCODE_AG_SPLINESTUFF );
  if ( status != CRhinoFile::good )
    return status;

  unsigned int data_typecode = 0;
  int data_length = 0;
  do {
    status = file.BeginReadChunk( data_typecode, data_length );
    if ( status != CRhinoFile::good )
      return status;
    status = CRhinoFile::crc_error;
    if ( data_typecode != TCODE_AG_SPLINESTUFF || data_length <= 0 )
      break;

    // read AG_SPLINE flags 
    int is_rat;    // 0 = no, 1 = euclidean, 2 = homogeneous
    {
      char c;
      short s;
      file.ReadByte(1,&c);
      if ( c != 2 && c != 3 )
        break;
      m_dim = c;
      file.ReadByte(1,&c);  
      if ( c != 0 && c != 1 && c != 2 )
        break;
      is_rat = c; // 0 = no, 1 = euclidean, 2 = homogeneous
      m_bIsRat = (is_rat != 0) ? TRUE : FALSE;
      file.ReadByte(1,&c);  
      if ( c < 2 )
        break;
      m_order = c;    
      file.ReadShort(1,&s); 
      if ( s < m_order )
        break;
      m_cv_count = s; 

      // The "m_is_closed" and "m_form" flags are here to recording
      // the values of legacy data found in the Rhino file.  These
      // values are not used in the toolkit code.
      file.ReadByte(1,&c);  
      if (c != 0 && c != 1 && c != 2)
        break;
      m_is_closed = c; // 0 = open, 1 = closed, 2 = periodic
      file.ReadByte(1,&c);  
      m_form = c;
    }

    // read AG_SPLINE bounding box
    file.ReadDouble( m_dim, m_bounding_box[0] );
    file.ReadDouble( m_dim, m_bounding_box[1] );

    // read AG_SPLINE knot vector 
    {
      m_knot_capacity = m_cv_count+m_order-2;
      m_knot = (double*)rhmalloc( m_knot_capacity*sizeof(*m_knot) );
      if ( !m_knot ) {
        throw CRhinoException( CRhinoException::out_of_memory );
      }

      const int knot_count = KnotCount();
      int       knot_index = 0;
      double    knot;

      // clamped_end_knot_flag: 0 = none, 1 = left, 2 = right, 3 = both
      char clamped_end_knot_flag = 0;
      if ( m_order > 2 ) 
        file.ReadByte(1,&clamped_end_knot_flag);
    
      // first knot(s)
      file.ReadDouble(1,&knot);
      m_knot[knot_index++] = knot;
      if (clamped_end_knot_flag % 2) {
        // clamped_start_knot
        while ( knot_index <= m_order-2 ) 
          m_knot[knot_index++] = knot;
      }
    
      // middle knots
      while ( knot_index <= m_cv_count-1 ) {
        file.ReadDouble(1,&knot);
        m_knot[knot_index++] = knot;
      }
    
      // end knot(s)
      if ( clamped_end_knot_flag >= 2 ) {
        while ( knot_index < knot_count )
          m_knot[knot_index++] = knot;
      }
      else {
        while ( knot_index < knot_count ) {
          file.ReadDouble(1,&knot);
          m_knot[knot_index++] = knot;
        }
      }
    }

    // read AG_SPLINE control points
    const int cvdim = ( m_bIsRat ) ? m_dim+1 : m_dim;
    m_cv_stride = cvdim;
    m_cv_capacity = cvdim*m_cv_count;
    m_cv = (double*)rhmalloc( m_cv_capacity*sizeof(*m_cv) );
    if ( !m_cv ) {
      throw CRhinoException( CRhinoException::out_of_memory );
    }
    file.ReadDouble( m_cv_capacity, m_cv );
    if ( m_bIsRat ) {
      // is_rat == 1 check fails because invalid is_rat flags in old files
      // convert AG rational CVs from euclidean to homogeneous
      double w, *cv;
      int cv_index, i;
      for ( cv_index = 0; cv_index < m_cv_count; cv_index++ ) {
        cv = m_cv + cv_index*m_cv_stride;
        w = cv[m_dim];
        for ( i = 0; i < m_dim; i++ )
          cv[i] *= w;
      }
    }
    status = IsValid() ? CRhinoFile::good : CRhinoFile::crc_error;
  } while (FALSE);
  CRhinoFile::eStatus end_status = file.EndChunk();
  if ( status == CRhinoFile::good )
    status = end_status;

  return status;
}

void
CRhinoAgSpline::Write(CRhinoFile& file) const
{
  if ( !IsValid() ) {
    throw CRhinoException( CRhinoException::illegal_input );
  }
  // write CRhinoAgSpline as a CRhinoNurbsCurve's to avoid getting 
  // AG_SPLINE structures that do not comply with AG restrictions.
  CRhinoNurbsCurve::Write(file);
}

BOOL
CRhinoAgSpline::IsValid(void) const
{
  if ( m_is_closed < 0 || m_is_closed > 2 )
    return FALSE;
  if ( Dimension() < 2 || Dimension() > 3 )
    return FALSE;
  return CRhinoNurbsCurve::IsValid();
}
