/* $Header: /Src/RhinoIO/rhinoio_agsurface.cpp 8     10/20/99 1:27p Dalelear $ */
/* $NoKeywords: $ */
//
// Copyright (c) 1993-1998 Robert McNeel & Associates. All rights reserved.
// Rhinoceros is a registered trademark of Robert McNeel & Assoicates.
//
// THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
// ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
// MERCHANTABILITY ARE HEREBY DISCLAIMED.
//				
////////////////////////////////////////////////////////////////

#include "rhinoio_object.h"
#include "rhinoio_nurbssrf.h"
#include "rhinoio_agsurface.h"

RHINO_CHUNK_IMPLEMENT( CRhinoAgSurface, TCODE_AG_SURFACE )

void CRhinoAgSurface::Initialize( void )
{
  // ALERT: SetTypeCode() is used only by CRhinoAgSurface and CRhinoAgSurface
  // All other geometry constructors should be derived from CRhinoObject
  // and call : CRhinoObject( TCODE_* ) during construction.
  SetTypeCode( TCODE_AG_SURFACE );
  m_is_closed[0] = m_is_closed[1] = 0;
  m_is_singular[0] = m_is_singular[1] = 0;
  m_form = 0;
}

// Initialize an empty CRhinoAgSurface
CRhinoAgSurface::CRhinoAgSurface()
{
  Initialize();
}

// Initialize an empty CRhinoAgSurface
CRhinoAgSurface::CRhinoAgSurface(
    int   dimension,    // dimension ( >= 1 and usually = 3 )
    BOOL  bIsRational,  // TRUE if surface is rational
    int   degree0,      // degree for first parameter ( >= 1 )
    int   degree1,      // degree for second parameter ( >= 1 )
    int   cv_count0,    // number of CVs for first parameter ( >= degree0+1 )
    int   cv_count1     // number of CVs for second parameter ( >= degree1+1 )
  ) : CRhinoNurbsSurface( dimension, bIsRational, degree0, degree1, cv_count0, cv_count1 )
{
  Initialize();
}

CRhinoFile::eStatus
CRhinoAgSurface::Read( CRhinoFile& file)
{
  // read attributes
  CRhinoFile::eStatus status = ReadAttributes( file, TCODE_AG_SURFACESTUFF );
  if ( status != CRhinoFile::good )
    return status;

  unsigned int data_typecode = 0;
  int data_length = 0;
  do {
    status = file.BeginReadChunk( data_typecode, data_length );
    if ( status != CRhinoFile::good )
      return status;
    status = CRhinoFile::crc_error;
    if ( data_typecode != TCODE_AG_SURFACESTUFF || data_length <= 0 )
      break;

    // read AG_SURFACE flags 
    char c;
    short s;
    int is_rat = 0; // 0 = no, 1 = eucidean, 2 = homogeneous

    // read AG_SURFACE flags
    file.ReadByte(1,&c);  
    if ( c != 3 )
      break;
    m_dim = 3;
    file.ReadByte(1,&c);
    m_form = c; // AGLib's "stype" flag
    file.ReadByte(1,&c);
    if ( c < 1 )
      break;
    m_order[0] = c+1;
    file.ReadByte(1,&c);
    if ( c < 1 )
      break;
    m_order[1] = c+1;
    file.ReadShort(1,&s); 
    if ( s < 1 )
      break;
    m_cv_count[0] = m_order[0]-1+s;
    file.ReadShort(1,&s);
    if ( s < 1 )
      break;
    m_cv_count[1] = m_order[1]-1+s;
    file.ReadByte(1,&c); // AGLib's ratu flag
    if ( c == 1 )
      is_rat = 1;
    else if ( c )
      is_rat = 2;
    file.ReadByte(1,&c);  // AGLib's ratu flag
    if ( c == 1 )
      is_rat = 1;
    else if ( c )
      is_rat = 2;
    m_bIsRat = (is_rat) ? TRUE : FALSE;
    file.ReadByte(1,&c);
    if ( c != 0 && c != 1 && c != 2 )
      break;
    m_is_closed[0] = c;
    file.ReadByte(1,&c);
    if ( c != 0 && c != 1 && c != 2 )
      break;
    m_is_closed[1] = c;
    file.ReadByte(1,&c);
    if ( c != 0 && c != 1 && c != 2 && c != 3 )
      break;
    m_is_singular[0] = c;
    file.ReadByte(1,&c);
    if ( c != 0 && c != 1 && c != 2 && c != 3 )
      break;
    m_is_singular[1] = c;


    // read AG_SURFACE bounding box
    file.ReadDouble( m_dim, m_bounding_box[0] );
    file.ReadDouble( m_dim, m_bounding_box[1] );

    // read AG_SURFACE knot vectors 
    int dir;
    for ( dir = 0; dir < 2; dir++) {
      m_knot_capacity[dir] = m_cv_count[dir]+m_order[dir]-2;
      m_knot[dir] = (double*)rhmalloc( m_knot_capacity[dir]*sizeof(*m_knot[dir]) );
      if ( !m_knot[dir] ) {
        throw CRhinoException( CRhinoException::out_of_memory );
      }
      file.ReadDouble( m_knot_capacity[dir], m_knot[dir] );
    }

    // read AG_SURFACE control points
    const int cvdim = ( m_bIsRat ) ? m_dim+1 : m_dim;
    m_cv_stride[1] = cvdim;
    m_cv_stride[0] = cvdim*m_cv_count[1];
    m_cv_capacity = cvdim*m_cv_count[0]*m_cv_count[1];
    m_cv = (double*)rhmalloc( m_cv_capacity*sizeof(*m_cv) );
    if ( !m_cv ) {
      throw CRhinoException( CRhinoException::out_of_memory );
    }
    file.ReadDouble( m_cv_capacity, m_cv );
    if ( is_rat == 1 ) {
      // convert AG rational CVs from euclidean to homogeneous
      double w, *cv;
      int cv_index, i;
      for ( cv_index = 0; cv_index < m_cv_count[0]*m_cv_count[1]; cv_index++ ) {
        cv = m_cv + cv_index*cvdim;
        w = cv[m_dim];
        for ( i = 0; i < m_dim; i++ )
          cv[i] *= w;
      }
    }
    status = IsValid() ? CRhinoFile::good : CRhinoFile::crc_error;
  } while (FALSE);
  CRhinoFile::eStatus end_status = file.EndChunk();
  if ( status == CRhinoFile::good )
    status = end_status;

  return status;
}

void
CRhinoAgSurface::Write(CRhinoFile& file) const
{
  if ( !IsValid() ) {
    throw CRhinoException( CRhinoException::illegal_input );
  }
  // write CRhinoAgSurface as a CRhinoNurbsSurface's to avoid getting 
  // AG_SPLINE structures that do not comply with AG restrictions.
  CRhinoNurbsSurface::Write(file);
}

BOOL
CRhinoAgSurface::IsValid(void) const
{
  if ( m_is_closed[0] < 0 || m_is_closed[0] > 2 )
    return FALSE;
  if ( m_is_closed[1] < 0 || m_is_closed[1] > 2 )
    return FALSE;
  if ( Dimension() < 2 || Dimension() > 3 )
    return FALSE;
  return CRhinoNurbsSurface::IsValid();
}
