/* $Header: /Src/RhinoIO/rhinoio_array.h 15    10/20/99 1:27p Dalelear $ */
/* $NoKeywords: $ */
//
// Copyright (c) 1993-1998 Robert McNeel & Associates. All rights reserved.
// Rhinoceros is a registered trademark of Robert McNeel & Assoicates.
//
// THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
// ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
// MERCHANTABILITY ARE HEREBY DISCLAIMED.
//				
////////////////////////////////////////////////////////////////
//
//   Defines RhinoIO dynamic array type
//
////////////////////////////////////////////////////////////////

#if !defined(RHINO_ARRAY_INC_)
#define RHINO_ARRAY_INC_

//  use CRhinoArray when "T" is a simple type or ~T is the default constructor
template <class T> class CRhinoArray
{
public:
  // construction ////////////////////////////////////////////////////////
  CRhinoArray( 
      int = 0,  // initial capacity
      int = 0   // simple type (TRUE or FALSE)
                // If "T" is not a class or ~T does not rhfree/delete
                // memory, then you may use TRUE to get more efficient
                // memory managment using rhmalloc()/rhrealloc()/rhfree()
      );

  virtual ~CRhinoArray();

	CRhinoArray<T>& operator=(const CRhinoArray<T>&);

  // query ///////////////////////////////////////////////////////////////
	int Size() const;      // number of elements in array

  const T& operator[]( int ) const;  // The operator[] functions do
                                     // not check for a valid index.
  T& operator[]( int );              // Caller must be sure that
                                     // 0 <= index < Size().
  const T* Last() const;             // can return NULL if size = 0
  T* Last();
  
  // array operations ////////////////////////////////////////////////////
  void Append( const T& );  // append copy of element using operator=()
  T& Append();              // append a default element
  void RemoveLast();        // remove last element (nothing deleted)
  void Empty();             // remove any elements from array
  void Reverse();           // reverse order
  
  // memory managment ////////////////////////////////////////////////////
  void Reserve( int );     // increase capacity
  void Shrink();           // remove unused capacity

protected:
  // implimentation //////////////////////////////////////////////////////
  T* Array() const;                   // pointer to array
  void SetSize( int );        // increase/decrease size
  void SetCapacity( int );    // shrink/grows capacity

  T* m_p;
  int    m_memory_type;  // 0 = use new[]/delete[],
                         // 1 = use rhmalloc()/rhrealloc()/rhfree(),
	int m_size;
	int m_capacity;  
  
private:
  // prohibit copy construction
  CRhinoArray(const CRhinoArray<T>&); // no implementation
};

/////////////////////////////////////////////////////////////////////////////////////
//  Class CRhinoArray<>
/////////////////////////////////////////////////////////////////////////////////////

// construction ////////////////////////////////////////////////////////
template <class T>
inline CRhinoArray<T>::CRhinoArray( 
                         int initial_capacity,  // = 0
                         int bSimpleType        // = FALSE
                         )
                      : m_p( 0 ),
                        m_memory_type( bSimpleType?1:0 ),
                        m_size( 0 ),
                        m_capacity( 0 )
{
  if ( initial_capacity > 0 ) 
    Reserve( initial_capacity );
}

template <class T>
inline CRhinoArray<T>::~CRhinoArray()
{ 
  SetCapacity(0);
}

template <class T>
inline CRhinoArray<T>& CRhinoArray<T>::operator=( const CRhinoArray<T>& src )
{
  if( &src == this )
    return *this;
  if( src.m_capacity > 0 )
    Reserve( src.m_capacity );
  if ( m_memory_type == 1 && src.m_memory_type == 1 ) {
    m_size = src.m_size;
    if ( m_size > 0 )
      memcpy( m_p, src.m_p, m_size*sizeof(T) );
  }
  else {
    for( m_size = 0; m_size < src.m_size; m_size++ )
      m_p[m_size] = src.m_p[m_size];
  }
  return *this;
}

// query ///////////////////////////////////////////////////////////////

template <class T>
inline int CRhinoArray<T>::Size() const
{
  return m_size;
}

template <class T>
inline T& CRhinoArray<T>::operator[]( int i )
{ 
  return m_p[i]; 
}

template <class T>
inline const T& CRhinoArray<T>::operator[](int i) const
{
  return m_p[i];
}

template <class T>
inline T* CRhinoArray<T>::Last()
{ 
  return (m_size > 0) ? m_p+(m_size-1) : 0;
}

template <class T>
inline const T* CRhinoArray<T>::Last() const
{
  return (m_size > 0) ? m_p+(m_size-1) : 0;
}

// array operations ////////////////////////////////////////////////////

template <class T>
inline void CRhinoArray<T>::Append( const T& x ) 
{
  Reserve(m_size+1);
	m_p[m_size++] = x;
}

template <class T>
inline T& CRhinoArray<T>::Append() 
{
  Reserve(m_size+1);
	return m_p[m_size++];
}

template <class T>
inline void CRhinoArray<T>::RemoveLast() 
{
  if ( m_size > 0 ) {
    m_size--;
  }
}

template <class T>
inline void CRhinoArray<T>::Empty()
{
  SetSize(0);
}

template <class T>
inline void CRhinoArray<T>::Reverse()
{
  // NOTE:
  // If anything in "T" depends on the value of this's address,
  // then don't call Reverse().
  char scratch[512];
  const int ele_size = sizeof(T);
  int i = 0;  
  int j = m_size;
  void* t = (ele_size > 512) ? rhmalloc(ele_size) : scratch;
  if ( t ) {
    j--; // make j and int to avoid signed/unsigned trauma
    for ( /*empty*/; i < j; i++, j-- ) {
      memcpy( t,     m_p+i, ele_size);
      memcpy( m_p+i, m_p+j, ele_size);
      memcpy( m_p+j, t,     ele_size);
    }
  }
  if ( ele_size > 512 )
    rhfree(t);
}


// memory managment ////////////////////////////////////////////////////

template <class T>
inline void CRhinoArray<T>::Reserve( int capacity ) 
{
  if( m_capacity < capacity )
    SetCapacity( capacity );
}

template <class T>
inline void CRhinoArray<T>::Shrink()
{
  SetCapacity( m_size );
}

// implimentation //////////////////////////////////////////

template <class T>
inline T* CRhinoArray<T>::Array() const
{
  return m_p;
}

template <class T>
inline void CRhinoArray<T>::SetSize( int size ) 
{
  if ( size >= 0 && size <= m_capacity ) {
    m_size = size;
  }
}

template <class T>
inline void CRhinoArray<T>::SetCapacity( int capacity ) 
{
  if ( capacity != m_capacity ) {
    if ( m_size > capacity )
      SetSize(capacity);
    if( capacity > 0 ) {
      switch ( m_memory_type ) {

      case 0: // new[] / delete[]
        { 
          m_capacity = capacity;
          T* newdata = new T[m_capacity];
          int i;
          for( i = 0; i < m_size; i++ ) 
            newdata[i] = m_p[i];   // copy data to new array
          delete[] m_p;
          m_p = newdata;
        }
        break;

      case 1: // rhmalloc()/rhrealloc()/rhfree()
        if ( m_p ) {
          m_p = (T*)rhrealloc( (void*)m_p, capacity*sizeof(T) );
        }
        else {
          m_p = (T*)rhmalloc( capacity*sizeof(T) );
        }
        if ( capacity > m_capacity )
          memset( &m_p[m_capacity], 0, (capacity-m_capacity)*sizeof(T) );
        m_capacity = capacity;
        break;
      }
    }
    else if ( m_p ) {
      switch ( m_memory_type ) {

      case 0: // new[] / delete[]
        delete[] m_p;
        break;

      case 1: // rhmalloc()/rhrealloc()/rhfree()
        rhfree( m_p );
        break;
      }

      m_p = 0;
      m_size = m_capacity = 0;
    }
  }
}



#endif

