/* $Header: /Src/RhinoIO/rhinoio_brep.h 30    11/10/99 2:56p Dalelear $ */
/* $NoKeywords: $ */
//
// Copyright (c) 1993-1998 Robert McNeel & Associates. All rights reserved.
// Rhinoceros is a registered trademark of Robert McNeel & Assoicates.
//
// THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
// ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
// MERCHANTABILITY ARE HEREBY DISCLAIMED.
//				
////////////////////////////////////////////////////////////////
//
//   defines CRhinoBrepObject
//
////////////////////////////////////////////////////////////////

#if !defined(RHINO_BREP_INC_)
#define RHINO_BREP_INC_

#include "rhinoio_object.h"

#include "rhinoio_nurbscrv.h"
#include "rhinoio_agspline.h"
#include "rhinoio_agcurve.h"


#include "rhinoio_nurbssrf.h"
#include "rhinoio_agsurface.h"

///////////////////////////////////////////////////////////////////////////////
// Class  CRhinoBrep
//
//  This is an abstract base class for B-rep objects stored in a Rhino 3d Model
//  files.  CRhinoAgFace and CRhinoAgShell classes are derived from CRhinoBrep
//
///////////////////////////////////////////////////////////////////////////////


// NOTE: The Vertex, Edge, Trim, Loop, Face classes should be defined
//       inside of the CRhinoBrep class.  This approach worked with
//       version 4 of Microsoft C++ compiler and STL.  Due to changes
//       in version 5 Microsoft C++ STL, the classes had to be moved
//       out of the CRhinoBrep class definition.

// CRhinoBrepVertex, CRhinoBrepEdge, CRhinoBrepTrim, CRhinoBrepLoop,
// and CRhinoBrepFace are derived from CRhinoBrepEntity.

class  CRhinoBrepEntity
{
public:
  CRhinoBrepEntity(int = -1);
  CRhinoBrepEntity& operator=(const CRhinoBrepEntity&);

  int m_user;   // integer available for application use.
                // constructor initializes m_user to 0.
  int m_index;  // index of item in CRhinoBrep array

private:
  // prohibit copy construction of Brep atoms.
  CRhinoBrepEntity( const CRhinoBrepEntity& ); // no implementation
};

inline CRhinoBrepEntity::CRhinoBrepEntity(int i /*= -1*/)
                        : m_user(0), m_index(i) 
  {}

class CRhinoBrepVertex : public  CRhinoBrepEntity
{
public:
  CRhinoBrepVertex(int = -1);
  CRhinoBrepVertex& operator=(const CRhinoBrepVertex&);
  double x, y, z;
  CRhinoIntArray m_ei; // indices of edges using this vertex
  double m_tolerance;  // accuracy of vertex point (>=0.0 or -1.0)
                       //
                       // A value of -1.0 indicates that the tolerance
                       // should be computed.
                       //
                       // If an edge begins or ends at this vertex
                       // then the distance from the vertex's 
                       // 3d point to the appropriate end of the
                       // edge's 3d curve must be <= this tolerance.
                       //
                       // If a trim begins or ends at this vertex,
                       // then the distance from the vertex's 3d point
                       // to the 3d point on the surface obtained by
                       // evaluating the surface at the appropriate
                       // end of the trimming curve must be <= this
                       // tolerance.

  void Write( CRhinoFile& ) const;
  void Read( CRhinoFile& );
};

class CRhinoBrepEdge : public  CRhinoBrepEntity
{
public:
  CRhinoBrepEdge(int = -1);
  CRhinoBrepEdge& operator=(const CRhinoBrepEdge&);
  int m_c3i;           // index of 3d curve
  double m_t[2];       // (sub)domain of 3d curve used for edge geometry
  int m_vi[2];         // indices of starting/ending vertex
  CRhinoIntArray m_ti; // indices of Trims that use this edge
  double m_tolerance;  // accuracy of edge curve (>=0.0 or -1.0)
                       //
                       // A value of -1.0 indicates that the tolerance
                       // should be computed.
                       //
                       // The maximum distance from the edge's 3d curve
                       // to a surface of a face that has this edge as
                       // a portion of its boundary must be <= this
                       // tolerance.

  void Write( CRhinoFile& ) const;
  void Read( CRhinoFile& );
};

class CRhinoBrepTrim : public  CRhinoBrepEntity
{
public:
  CRhinoBrepTrim(int = -1);
  CRhinoBrepTrim& operator=(const CRhinoBrepTrim&);

  int    m_c2i;         // index of 2d curve
  double m_t[2];        // (sub)domain of 2d curve used for trim geometry
  int    m_ei;          // index of 3d edge (-1 if CRhinoBrepTrim is singular)
  int    m_vi[2];       // indices of end vertices
  BOOL   m_bRev3d;      // TRUE if the 2d trim and 3d edge have
                        //      opposite orientations.
  enum {
    unknown  = 0,
    boundary = 1,       // trim has no mate and is connected to edge
    mated    = 2,       // trim shares edge with trim(s) on other face(s)
    seam     = 3,       // trim shares edge with 1 other trim on same face
    singular = 4        // trim is along singular side of surface (no edge)
  } m_type;

  enum {
    not_iso  = 0,       // trim is not an isoparameteric curve
                        // note: odd values are all "u" = constant
                        //       even values > 0 are all "v" = constant
    u_iso    = 1,       // trim is a "u" = constant isoparametric curve
                        // in the interior of the surface's domain
    v_iso    = 2,       // trim is a "v" = constant isoparametric curve
                        // in the interior of the surface's domain
    W_iso    = 3,       // trim is a "u" = constant isoparametric curve
                        // along the west side of the surface's domain
    S_iso    = 4,       // trim is a "v" = constant isoparametric curve
                        // along the south side of the surface's domain
    E_iso    = 5,       // trim is a "u" = constant isoparametric curve
                        // along the east side of the surface's domain
    N_iso    = 6        // trim is a "v" = constant isoparametric curve
                        // along the north side of the surface's domain
  } m_iso;

  int    m_li;          // index of loop that uses this trim
  double m_tolerance[2]; // each value >= 0.0 or -1.0
                         //
                         // A value of -1.0 indicates that the tolerance
                         // should be computed.
                         //
                         // m_tolerance[0] = accuracy of parameter space curve
                         //                  in first parameter
                         // m_tolerance[1] = accuracy of parameter space curve
                         //                  in second parameter
                         //
                         // If the trim is not singular, then the trim must
                         // have an edge.  If P is a 3d point on the edge's
                         // curve and surface(u,v) = Q is the point on the 
                         // surface that is closest to P, then there must
                         // be a parameter t in the interval [m_t[0], m_t[1]]
                         // such that
                         //   |u - curve2d(t)[0]| <= m_tolerance[0]
                         // and 
                         //   |v - curve2d(t)[1]| <= m_tolerance[1]
                         //
                         // If P is the 3d point for the vertex brep.m_V[m_vi[k]]
                         // and (uk,vk) is the corresponding end of the trim's
                         // parameter space curve, then there must be a surface
                         // parameter (u,v) such that:
                         // *) the distance from the 3d point surface(u,v) to P
                         //    is <= brep.m_V[m_vi[k]].m_tolerance,
                         // *) |u-uk| <= m_tolerance[0], and
                         // *) |v-vk| <= m_tolerance[1].

  double m_P[2][3];      // Place to store 3d locations of trim ends.  You
                         // must have acces to a surface evaluator in order
                         // to set these fields.  They are useful for translating
                         // Rhino geometry into formats that require these end
                         // locations.

  void Write( CRhinoFile& ) const;
  void Read( CRhinoFile& );
private:
  // values stored in file
  friend class CRhinoBrep;
  double m_2d_tol;      // used internally - ignore
  double m_3d_tol;      // used internally - ignore
};

class CRhinoBrepLoop : public  CRhinoBrepEntity
{
public:
  CRhinoBrepLoop(int = -1);
  CRhinoBrepLoop& operator=(const CRhinoBrepLoop&);

  CRhinoIntArray m_ti;  // trim indices
  enum {
    unknown  = 0,
    outer    = 1,
    inner    = 2,
    slit     = 3
  } m_type;
  int m_fi;             // index of face that uses this loop

  void Write( CRhinoFile& ) const;
  void Read( CRhinoFile& );
};

class CRhinoBrepFace : public  CRhinoBrepEntity
{
public:
  CRhinoBrepFace(int = -1);
  CRhinoBrepFace& operator=(const CRhinoBrepFace&);

  CRhinoIntArray m_li; // loop indices (outer loop is m_li[0])
  int m_si;            // index of surface in b-rep m_S[] array
  BOOL m_bRev;         // TRUE if face orientation is opposite
                       //      of natural surface orientation

  void Write( CRhinoFile& ) const;
  void Read( CRhinoFile& );
};

typedef CRhinoArray<CRhinoBrepVertex>    CRhinoVertexArray;
typedef CRhinoArray<CRhinoBrepEdge>      CRhinoEdgeArray;
typedef CRhinoArray<CRhinoBrepTrim>      CRhinoTrimArray;
typedef CRhinoArray<CRhinoBrepLoop>      CRhinoLoopArray;
typedef CRhinoArray<CRhinoBrepFace>      CRhinoFaceArray;

class CRhinoCurveArray : public CRhinoArray<CRhinoAgCurve*>
{
public:
  CRhinoCurveArray( size_t = 0 );
  ~CRhinoCurveArray();
};

inline 
CRhinoCurveArray::CRhinoCurveArray( size_t initial_capacity )
                   : CRhinoArray<CRhinoAgCurve*>(initial_capacity,TRUE)
{}

inline 
CRhinoCurveArray::~CRhinoCurveArray()
{
  int i = m_size;
  while ( i-- > 0 ) {
    if ( m_p[i] ) {
      delete m_p[i];
      m_p[i] = NULL;
    }
  }
}

class CRhinoSurfaceArray : public CRhinoArray<CRhinoNurbsSurface*>
{
public:
  CRhinoSurfaceArray( size_t = 0 );
  ~CRhinoSurfaceArray();
};

inline 
CRhinoSurfaceArray::CRhinoSurfaceArray( size_t initial_capacity )
                   : CRhinoArray<CRhinoNurbsSurface*>(initial_capacity,TRUE)
{}

inline 
CRhinoSurfaceArray::~CRhinoSurfaceArray()
{
  int i = m_size;
  while ( i-- > 0 ) {
    if ( m_p[i] ) {
      delete m_p[i];
      m_p[i] = NULL;
    }
  }
}


class CRhinoBrep : public CRhinoObject 
{	 
	RHINO_CHUNK_DECLARE( CRhinoBrep )
public:

	// Constructors/ Destructors  - Constructors are protected 
  CRhinoBrep();
	~CRhinoBrep();		

  // CRhinoChunk overrides
  BOOL IsValid( void ) const;

  // CRhinoObject overrides
	CRhinoFile::eStatus Read(   	// read chunk data from file 
    CRhinoFile&
    );
	
  void Write(                   // Writes NURBS surface to file
                                // Throws a CRhinoException if IsValid()
                                // returns FALSE or fwrite() fails.
    CRhinoFile&
    ) const;

  BOOL GetBoundingBox(
          double[3], // min
          double[3], // max
          BOOL = FALSE // set to TRUE if box should be grown
          ) const;

  BOOL SwapCoordinates(
        int, int        // indices of coords to swap
        );

  // CRhinoBrepFace query
  BOOL IsSolid() const; // returns TRUE if b-rep is an oriented manifold solid
  
  BOOL IsManifold( // returns TRUE if b-rep is an oriented manifold
    BOOL* = NULL,  // TRUE  is returned if b-rep is an oriented manifold
                   // FALSE is returned if b-rep is not a manifold or
                   //       is not an orientable manifold
    BOOL* = NULL   // TRUE  is returned if b-rep is a manifold with boundary
                   // FALSE is returned if b-rep is not a manifold or
                   //       is a manifold with no boundary
    ) const;

  // CRhinoBrepFace query
  BOOL IsSurface() const;      // returns TRUE if the b-rep has a single face
                               // and that face is geometrically the same
                               // as the underlying surface.  I.e., the face
                               // has trivial trimming.  In this case, the
                               // surface is m_S[0].


  BOOL FaceIsSurface(          // returns TRUE if the face has a single
         int // index of face  // outer boundary and that boundary runs
         ) const;              // along the edges of the underlying surface.
                               // In this case the geometry of the surface
                               // is the same as the geometry of the face.
                               // If FaceIsSurface() is TRUE, then
                               // m_S[m_F[face_index].m_si] is the surface.

  BOOL LoopIsSurfaceBoundary(  // returns TRUE if the loop's trims all run
         int // index of loop  // along the edge's of the underlying surface's
         ) const;              // parameter space.

  // CRhinoBrepFace orientation
  BOOL FlipReversedSurfaces(); // clears all CRhinoBrepFace.m_bRev flags
                               // by calling SwapFaceParameters() on each
                               // face with a TRUE m_bRev.

  BOOL SwapFaceParameters(     // swaps "u" and "v" parameters of face's
        int // index of face   // surface,  swaps "u" and "v" coordinates
        );                     // of parameter space trimming curves,
                               // reverses trimming loops, and toggles
                               // value of CRhinoBrepFace::m_bRev.

  // for moving around loops - returns trim index of prev/next trim in loop
  int PrevTrim(
        int // index of current trim (trim.m_index)
        ) const;
  int NextTrim(
        int // index of current trim (trim.m_index)
        ) const;

  // for moving around vertices - returns edge index of prev/next edge in
  // using the specified vertex
  int PrevEdge(
        int,        // index of current edge (edge.m_index)
        int,        // 0 = use edge start vertex, 1 = use edge end vertex
        int* = NULL // returns 0 if next edge begins at vert
                    //         1 if next edge ends at vertex
        ) const;
  int NextEdge(
        int,        // index of current edge (edge.m_index)
        int,        // 0 = use edge start vertex, 1 = use edge end vertex
        int* = NULL // returns 0 if prev edge begins at vert
                    //         1 if prev edge ends at vertex
        ) const;

  // This function can be used to compute vertex information for a
  // b-rep when everything but the m_V array is properly filled in.
  // It is intended to be used when creating a CRhinoBrep from a 
  // definition that does not include explicit vertex information.
  void SetVertices(void);

  // This function can be used to set the CRhinoBrepTrim::m_iso
  // flags for all trims in the b-rep.
  // It is intended to be used when creating a CRhinoBrep from a 
  // definition that does not include compatible parameter space
  // type information.
  void SetTrimIsoFlags();


  // geometry
  CRhinoCurveArray   m_C2;  // parameter space trimming curves (used by trims)
  CRhinoCurveArray   m_C3;  // 3d curves (used by edges)
  CRhinoSurfaceArray m_S;   // parametric surfaces (used by faces)

  // topology
  CRhinoVertexArray  m_V;   // vertices
  CRhinoEdgeArray    m_E;   // edges
  CRhinoTrimArray    m_T;   // trims
  CRhinoLoopArray    m_L;   // loops
  CRhinoFaceArray    m_F;   // faces

  CRhinoFile::eStatus ReadBrep( CRhinoFile& ); // Low level tool that reads the 
                                               // TCODE_RHINO_OBJECT_DATA chunk
                                               // that contains the parameters
                                               // that define the b-rep.
protected:	
	CRhinoBrep(
    int,     // typecode
    int = 0  // integer value or length of data
    );

  // helpers to read AG_SHELL and AG_FACE sub-chunks 
  CRhinoFile::eStatus SkipToAG_STUFF( CRhinoFile&, const int );
  CRhinoFile::eStatus Read_AG_SURFACE( CRhinoFile&, int* );
  CRhinoFile::eStatus Read_AG_CURVE( CRhinoFile&, int* );
  CRhinoFile::eStatus Read_AG_BOUNDARY( CRhinoFile&, int* );
  CRhinoFile::eStatus Read_AG_TEDGE( CRhinoFile&, int* );
  CRhinoFile::eStatus ReadAgFace( CRhinoFile&, BOOL = FALSE );
  double m_bounding_box[2][3];

  // helpers for setting tols on file read
  void SetTolsFromFileValues();

private:
  void Initialize();
  void WriteBrepCurve( CRhinoFile&, const CRhinoAgCurve& ) const;
  CRhinoAgCurve* ReadBrepCurve( CRhinoFile& ) const;

  // helpers to set CRhinoBrepTrim::m_iso flag
  void SetTrimIsoFlag(int,double[6]);
  void SetTrimIsoFlag(int);

  // helpers to create and set vertices
  BOOL SetEdgeVertex(const int, const int, const int );
  BOOL HopAcrossEdge( int&, int& ) const;
  BOOL SetTrimStartVertex( const int, const int);
  void SetLoopVertices(const int);
  void ClearTrimVertices();
  void ClearEdgeVertices();

  // helpers for SwapFaceParameters()
  BOOL SwapLoopParameters(
        int // index of loop
        );
  BOOL SwapTrimParameters(
        int // index of trim
        );

  // helpers for validation checking
  BOOL
  IsValidTrim( 
        int // index of trim
        ) const;
  BOOL
  IsValidLoop( 
        int // index of loop
        ) const;
  BOOL
  IsValidFace( 
        int // index of face
        ) const;
  BOOL
  IsValidEdge( 
        int // index of edge
        ) const;
  BOOL
  IsValidVertex( 
        int // index of edge
        ) const;
};

///////////////////////////////////////////////////////////////////////////////

#endif
