/* $Header: /Src/RhinoIO/rhinoio_chunk.h 9     10/20/99 1:27p Dalelear $ */
/* $NoKeywords: $ */
//
// Copyright (c) 1993-1998 Robert McNeel & Associates. All rights reserved.
// Rhinoceros is a registered trademark of Robert McNeel & Assoicates.
//
// THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
// ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
// MERCHANTABILITY ARE HEREBY DISCLAIMED.
//				
////////////////////////////////////////////////////////////////
//
//   defines CRhinoChunk, CRhinoEndOfTable, 
//           RHINO_DECLARE_CHUNK, and RHINO_IMPLEMENT_CHUNK
//
////////////////////////////////////////////////////////////////

#include "rhinoio_def.h"
#include "rhinoio_exception.h"
#include "rhinoio_layerptr.h"
#include "rhinoio_file.h"

#if !defined(RHINO_CHUNK_INC_)
#define RHINO_CHUNK_INC_

// Defines for declaring and defining classes derived from CRhinoChunk.
// All classes that represent an actual item in a Rhino .3dm file
// are derived from CRhinoChunk (or CRhinoObject).  These classes must
// have a RHINO_CHUNK_DECLARE() in their declaration and an
// RHINO_CHUNK_IMPLEMENT() in their implementation.

#define RHINO_CHUNK_DECLARE(classname) \
  public: \
   static const int m_##classname##_typecode;\
   static int TypeCode( void );\
   const char* ClassName( void ) const;\
  private: \
   static CRhinoChunk* CreateNew( int, int );

#define RHINO_CHUNK_IMPLEMENT(classname,tcode) \
  CRhinoChunk* classname::CreateNew(int,int){return new classname();}\
  const char* classname::ClassName(void) const {return #classname;}\
	const int classname::m_##classname##_typecode = CRhinoChunk::RegisterAdd(tcode,& classname::CreateNew);\
	int classname::TypeCode( void ) {return m_##classname##_typecode;}


///////////////////////////////////////////////////////////////////////////////
//
void RhinoIOCleanup(void);
//
// RhinoIOCleanup() deletes registry entries that are created by
// the calls to CRhinoChunk::RegisterAdd().  CRhinoChunk::RegisterAdd()
// is used to initialize the 
//
//   static int CRhinoBlahBlah::m_CRhinoBlahBlah_typecode
//
// members that are declared in the invocations of the
// RHINO_CHUNK_DECLARE macro.  Each invocation of the
// RHINO_CHUNK_IMPLEMENT macro makes a single call to
// CRhinoChunk::RegisterAdd() and allocates 8 bytes.
//
// Practically speaking, you do not want to call RhinoIOCleanup().
// 1) The Rhino IO toolkit will not be able to read any files after
//    calling RhinoIOCleanup().
// 2) Calling RhinoIOCleanup() will rhfree a most a few hundred bytes.
//
// If you are using leak detecting software and you get reports of
// CRhinoChunk::RegistryEntry classes being created in
// CRhinoChunk::RegisterAdd() and never deleted, then call
// RhinoIOCleanup() at the very end of your program.
//
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// Class  CRhinoChunk
//
//  This is an abstract base class for chunks stored in a Rhino 3d Model files.
//	Each chunk has a CRhinoHeader.  The TypeCode in the header identifies its type.
///////////////////////////////////////////////////////////////////////////////
class CRhinoChunk
{
public:
  // types and enums /////////////////////////////////////////////////////////
	enum eChunkStyle {integer, data};	// two types of chunks

  // Static Functions /////////////////////////////////////////////////////////
	static BOOL IsShortTypeCode( // TRUE if typecode is for a short chunk
                               // ( typecode, integer value )
         int // typecode
                  
         );
  static BOOL IsLongTypeCode(  // TRUE if typecode is for a long chunk
                               // ( typecode, length, length bytes )
         int // typecode 
         );
  static BOOL IsCRCTypeCode(   // true if typecode is for a chunk that has a
                               // 16 bit CRC
                               // ( typecode, length, length-2 bytes, 2 byte CRC )
         int // typecode 
         );
    
  // Constructors/ Destructors  - Constructors are protected //////////////////
  static CRhinoChunk* CreateNew(  // creates a CRhinoChunk of the correct 
                                  // derived type if the typecode is registered
                                  // using the RHINO_CHUNK_DECLARE and
                                  // RHINO_CHUNK_IMPLEMENT macros.
        int,     // typecode, 
        int = 0  // integer value for integer chunks
                 // length for data chunks
        );
	virtual ~CRhinoChunk();

  // Chunk typecode and information interface /////////////////////////////////
  int TypeCode( void ) const;	// see 3dm.h for interpretation of typecodes
  int Length( void ) const;   // number of bytes after chunk header
                              // = 0 for integer chunks, 
                              // = length of data for data chunks

	int Value( void ) const;		// = 0 for data chunks
				                      // = integer value for integer chunks
  void SetValue ( int );
	
	virtual BOOL HasCRC( void ) const;

	eChunkStyle ChunkStyle( void ) const;

  // Chunk type query /////////////////////////////////////////////////////////

  BOOL IsAttribute( void ) const;	// High level chunks are either attributes or
									                // or objects. Layer is the only attribute 
                                  // for now 
  virtual BOOL IsObject( void )	const;	// Is this derived from a CRhinoObject

  virtual BOOL IsValid( void ) const;    // Is this a valid CRhinoObject


  virtual const char* ClassName( void ) const;

  // Reading and writing.  Registered long chunks must override
  // Read() and Write().
	virtual CRhinoFile::eStatus Read(CRhinoFile&);							
	virtual void Write(CRhinoFile &) const;

  friend void RhinoIOCleanup(void);

protected:	
	CRhinoChunk(
    int = 0, // typecode
    int = 0  // integer value or length of data
    );

	// We maintain a register of all classes derived from CRhinoChunk listed
	// with there corresponding type code and CreateNew(...) member function.
	// CRhinoFile::ReadChunk uses this to create and initialize a chunk of
	// the correct type.

  int SetTypeCode( int ); // use in constructors only

	static int RegisterAdd( 
              int,                       // typecode
              CRhinoChunk* (*)(int,int)  // Classes CreateNew() function
                                         // from DECLARE/IMPLEMENT_RHINO_OBJECT
                                         // macros.
              );
private:	
	int m_typecode;
	int m_info;
  int m_data_offset;  // set when reading an unregistered chunk

  CRhinoChunk(const CRhinoChunk&);	// prohibit with no implementation
	CRhinoChunk& operator=(const CRhinoChunk&); // prohibit with no implementation 

  // CRhinoChunk registry
  class RegistryEntry{
	public:
		int m_typecode;
		CRhinoChunk*  (*m_CreateNew)( int /*typecode*/, int /*info*/ );  
		RegistryEntry *m_next;
	};
	static RegistryEntry* m_registry;	// This is a registry of all derived types

  friend class CRhinoFile;
};

///////////////////////////////////////////////////////////////////////////////

inline int CRhinoChunk::SetTypeCode( int typecode )
  { int rc = m_typecode; m_typecode = typecode; return rc; }

inline CRhinoChunk::CRhinoChunk(int typecode, int i) 
  : m_typecode(typecode), m_info(i) 
  { m_data_offset = 0; }

inline CRhinoChunk::~CRhinoChunk()
  { }

inline int CRhinoChunk::TypeCode( void ) const
  {return m_typecode;}

inline int CRhinoChunk::Length( void ) const
  {return (IsLongTypeCode(m_typecode)) ? m_info : 0;}

inline int CRhinoChunk::Value( void ) const
  {return (IsShortTypeCode(m_typecode)) ? m_info : 0;}

inline void CRhinoChunk::SetValue( int i )
  {if (IsShortTypeCode(m_typecode)) m_info = i;}

inline BOOL CRhinoChunk::HasCRC( void ) const
  { return IsCRCTypeCode( m_typecode ); }

inline BOOL CRhinoChunk::IsAttribute( void ) const
  {return m_typecode == TCODE_LAYER;}

inline CRhinoChunk::eChunkStyle CRhinoChunk::ChunkStyle() const 
  {return  (IsShortTypeCode( m_typecode )) ? integer : data; }

inline BOOL CRhinoChunk::IsObject( void ) const
	{return FALSE;}

inline BOOL CRhinoChunk::IsValid( void ) const
  { return FALSE; }

inline const char* CRhinoChunk::ClassName( void ) const
  { return "CRhinoChunk"; }

///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// Class CRhinoEndOfTable
//		Marks the end of a sequence of subchunks
///////////////////////////////////////////////////////////////////////////////
class CRhinoEndOfTable: public CRhinoChunk 
{
	RHINO_CHUNK_DECLARE( CRhinoEndOfTable )
public:

	CRhinoEndOfTable( int = 0 );
  ~CRhinoEndOfTable();
};

inline CRhinoEndOfTable::CRhinoEndOfTable(int i) 
                        : CRhinoChunk(TCODE_ENDOFTABLE,i)
  {}

inline CRhinoEndOfTable::~CRhinoEndOfTable()
  {}

#endif
