/* $Header: /Src/RhinoIO/rhinoio_file.cpp 34    1/07/00 9:34a Dalelear $ */
/* $NoKeywords: $ */
//
// Copyright (c) 1993-1998 Robert McNeel & Associates. All rights reserved.
// Rhinoceros is a registered trademark of Robert McNeel & Assoicates.
//
// THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
// ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
// MERCHANTABILITY ARE HEREBY DISCLAIMED.
//				
////////////////////////////////////////////////////////////////

#include "rhinoio_chunk.h"
#include "rhinoio_color.h"
#include "rhinoio_layer.h"
#include "rhinoio_tolerances.h"

// TODO remove this include when link hack is fixed.  See
// CRhinoFile::CRhinoFile() for details.
#include "rhinoio.h"

/* 16 bit cyclic redundancy check uses CCITT generator polynomial */
static unsigned short CRC_CCITT_TABLE[256] =
 {0x0000,0x1021,0x2042,0x3063,0x4084,0x50A5,0x60C6,0x70E7,0x8108,0x9129,0xA14A,0xB16B,0xC18C,0xD1AD,0xE1CE,0xF1EF,
  0x1231,0x0210,0x3273,0x2252,0x52B5,0x4294,0x72F7,0x62D6,0x9339,0x8318,0xB37B,0xA35A,0xD3BD,0xC39C,0xF3FF,0xE3DE,
  0x2462,0x3443,0x0420,0x1401,0x64E6,0x74C7,0x44A4,0x5485,0xA56A,0xB54B,0x8528,0x9509,0xE5EE,0xF5CF,0xC5AC,0xD58D,
  0x3653,0x2672,0x1611,0x0630,0x76D7,0x66F6,0x5695,0x46B4,0xB75B,0xA77A,0x9719,0x8738,0xF7DF,0xE7FE,0xD79D,0xC7BC,
  0x48C4,0x58E5,0x6886,0x78A7,0x0840,0x1861,0x2802,0x3823,0xC9CC,0xD9ED,0xE98E,0xF9AF,0x8948,0x9969,0xA90A,0xB92B,
  0x5AF5,0x4AD4,0x7AB7,0x6A96,0x1A71,0x0A50,0x3A33,0x2A12,0xDBFD,0xCBDC,0xFBBF,0xEB9E,0x9B79,0x8B58,0xBB3B,0xAB1A,
  0x6CA6,0x7C87,0x4CE4,0x5CC5,0x2C22,0x3C03,0x0C60,0x1C41,0xEDAE,0xFD8F,0xCDEC,0xDDCD,0xAD2A,0xBD0B,0x8D68,0x9D49,
  0x7E97,0x6EB6,0x5ED5,0x4EF4,0x3E13,0x2E32,0x1E51,0x0E70,0xFF9F,0xEFBE,0xDFDD,0xCFFC,0xBF1B,0xAF3A,0x9F59,0x8F78,
  0x9188,0x81A9,0xB1CA,0xA1EB,0xD10C,0xC12D,0xF14E,0xE16F,0x1080,0x00A1,0x30C2,0x20E3,0x5004,0x4025,0x7046,0x6067,
  0x83B9,0x9398,0xA3FB,0xB3DA,0xC33D,0xD31C,0xE37F,0xF35E,0x02B1,0x1290,0x22F3,0x32D2,0x4235,0x5214,0x6277,0x7256,
  0xB5EA,0xA5CB,0x95A8,0x8589,0xF56E,0xE54F,0xD52C,0xC50D,0x34E2,0x24C3,0x14A0,0x0481,0x7466,0x6447,0x5424,0x4405,
  0xA7DB,0xB7FA,0x8799,0x97B8,0xE75F,0xF77E,0xC71D,0xD73C,0x26D3,0x36F2,0x0691,0x16B0,0x6657,0x7676,0x4615,0x5634,
  0xD94C,0xC96D,0xF90E,0xE92F,0x99C8,0x89E9,0xB98A,0xA9AB,0x5844,0x4865,0x7806,0x6827,0x18C0,0x08E1,0x3882,0x28A3,
  0xCB7D,0xDB5C,0xEB3F,0xFB1E,0x8BF9,0x9BD8,0xABBB,0xBB9A,0x4A75,0x5A54,0x6A37,0x7A16,0x0AF1,0x1AD0,0x2AB3,0x3A92,
  0xFD2E,0xED0F,0xDD6C,0xCD4D,0xBDAA,0xAD8B,0x9DE8,0x8DC9,0x7C26,0x6C07,0x5C64,0x4C45,0x3CA2,0x2C83,0x1CE0,0x0CC1,
  0xEF1F,0xFF3E,0xCF5D,0xDF7C,0xAF9B,0xBFBA,0x8FD9,0x9FF8,0x6E17,0x7E36,0x4E55,0x5E74,0x2E93,0x3EB2,0x0ED1,0x1EF0};


static const char IOVERSION[] = "3D Geometry File Format        1";
const int Supported_Version=1;


////////////////////////////////////////////////////////////////////////////////
// Class CRhinoFile
//    High level functions.
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// CRhinoFile::CRhinoFile(FILE* fp, eMode m)
//	 The first parameter is a non-null FILE* to an opened binary file with the 
//   correct file mode.  The second paremeter specifies the mode as either read
//   write.  
//
//	 The first 32 bytes which identify the file as a Rhino model file are read
//	 or written, 
////////////////////////////////////////////////////////////////////////////////
CRhinoFile::CRhinoFile(FILE* fp, CRhinoFile::eMode m, BOOL bSkipVersionCheck ) 
           : m_mode(m), 
             m_pFile(fp), 
             m_beof(FALSE), 
             m_bBigEndian(FALSE), 
             m_version(0),
             m_bWriteTols(FALSE),
             m_bHaveAbsTol(FALSE),
             m_abs_tol(0.01),
             m_bHaveRelTol(FALSE),
             m_rel_tol(0.01),
             m_bHaveAngTol(FALSE),
             m_ang_tol(0.05235987755983),
             m_bHaveUnits(FALSE),
             m_units(CRhinoFile::no_units),
             m_notes(0),
             m_pLayer(NULL),
             m_Hdr(64,TRUE)
{
	if(fp==NULL)
		throw CRhinoException( CRhinoException::illegal_input);

	PushChunkRecord(); // initialize the stack of ChunkRecords	

	// determine endianness of executing processor.
	// 3dm files are stored in little endian (like Intel).
	m_bBigEndian = ComputeBigEndian();  

	if(m_mode==read) {  // Read the first 32 bytes including the version number
  
    // TODO: Remove this hack when VC linking bug is fixed
    {
      int typecode;

      typecode = CRhinoNotes::TypeCode();
      if ( !typecode )
  		  throw CRhinoException( CRhinoException::unexpected_value);
      typecode = CRhinoXdata::TypeCode();
      if ( !typecode )
  		  throw CRhinoException( CRhinoException::unexpected_value);
      typecode = CRhinoLayerIndex::TypeCode();
      if ( !typecode )
  		  throw CRhinoException( CRhinoException::unexpected_value);
      typecode = CRhinoNurbsCurve::TypeCode();
      if ( !typecode )
  		  throw CRhinoException( CRhinoException::unexpected_value);
      typecode = CRhinoNurbsSurface::TypeCode();
      if ( !typecode )
  		  throw CRhinoException( CRhinoException::unexpected_value);
      typecode = CRhinoBrep::TypeCode();
      if ( !typecode )
  		  throw CRhinoException( CRhinoException::unexpected_value);
      typecode = CRhinoBumpMap::TypeCode();
      if ( !typecode )
  		  throw CRhinoException( CRhinoException::unexpected_value);
      typecode = CRhinoEndOfTable::TypeCode();
      if ( !typecode )
  		  throw CRhinoException( CRhinoException::unexpected_value);
      typecode = CRhinoLayer::TypeCode();
      if ( !typecode )
  		  throw CRhinoException( CRhinoException::unexpected_value);
      typecode = CRhinoMesh::TypeCode();
      if ( !typecode )
  		  throw CRhinoException( CRhinoException::unexpected_value);
      typecode = CRhinoMeshGeometry::TypeCode();
      if ( !typecode )
  		  throw CRhinoException( CRhinoException::unexpected_value);
      typecode = CRhinoName::TypeCode();
      if ( !typecode )
  		  throw CRhinoException( CRhinoException::unexpected_value);
      typecode = CRhinoPointSet::TypeCode();
      if ( !typecode )
  		  throw CRhinoException( CRhinoException::unexpected_value);
      typecode = CRhinoRGB::TypeCode();
      if ( !typecode )
  		  throw CRhinoException( CRhinoException::unexpected_value);
      typecode = CRhinoTextureMap::TypeCode();
      if ( !typecode )
  		  throw CRhinoException( CRhinoException::unexpected_value);
      typecode = CRhinoRenderMaterialName::TypeCode();
      if ( !typecode )
  		  throw CRhinoException( CRhinoException::unexpected_value);
      typecode = CRhinoTransparency::TypeCode();
      if ( !typecode )
  		  throw CRhinoException( CRhinoException::unexpected_value);
      typecode = CRhinoAgSpline::TypeCode();
      if ( !typecode )
  		  throw CRhinoException( CRhinoException::unexpected_value);
      typecode = CRhinoAgCurve::TypeCode();
      if ( !typecode )
  		  throw CRhinoException( CRhinoException::unexpected_value);
      typecode = CRhinoAgSurface::TypeCode();
      if ( !typecode )
  		  throw CRhinoException( CRhinoException::unexpected_value);
      typecode = CRhinoAgFace::TypeCode();
      if ( !typecode )
  		  throw CRhinoException( CRhinoException::unexpected_value);
      typecode = CRhinoAgShell::TypeCode();
      if ( !typecode )
  		  throw CRhinoException( CRhinoException::unexpected_value);
      typecode = CRhinoSpotlight::TypeCode();
      if ( !typecode )
  		  throw CRhinoException( CRhinoException::unexpected_value);
      typecode = CRhinoViewport::TypeCode();
      if ( !typecode )
  		  throw CRhinoException( CRhinoException::unexpected_value);
      typecode = CRhinoUnitsAndTolerances::TypeCode();
      if ( !typecode )
  		  throw CRhinoException( CRhinoException::unexpected_value);
    }


    if ( !bSkipVersionCheck ) {
		  char s[128];
		  ReadByte(32,s);
		  s[32] = 0;
      if (strncmp(IOVERSION,s,24)) {
			  throw CRhinoException(CRhinoException::unexpected_value);
      }
		  m_version = 0;
      if (sscanf(s+24,"%d",&m_version) != 1) {
			  throw CRhinoException(CRhinoException::unexpected_value);
      }

		  // Check that version of file is supported by this reader
      if (m_version <= 0) {
			  throw CRhinoException(CRhinoException::unexpected_value);
      }
		  
      if(m_version> Supported_Version) {
			  throw CRhinoException(CRhinoException::not_supported);
      }
    }

	}
	else {	// mode==write.  write 32 bytes including the version number
		m_version=Supported_Version;
	  WriteByte(32,IOVERSION);

    // Next chunk must be a comment section.  It can be (and is) empty.
    {
      int i = 1;
      WriteInteger(1,&i); // comment chunk typecode
      i = 0;
      WriteInteger(1,&i); // length of comment chunk
    }
    m_bWriteTols = TRUE;
	}

  m_pLayer = new CRhinoLayerTable;
	
  return;
}

void CRhinoFile::PushChunkRecord()
{
  RhinoChunkRecord cr;
  memset(&cr,0,sizeof(cr));
  cr.index = m_Hdr.Size();
  m_Hdr.Append(cr); 
} 

void CRhinoFile::PopChunkRecord() 
{
	if( m_Hdr.Size() > 0 ) {
		m_Hdr.RemoveLast();
  }
	else {
		throw CRhinoException(CRhinoException::illegal_input);
  }
}   

RhinoChunkRecord* CRhinoFile::CurChunk()
{
  return m_Hdr.Last(); 
}


CRhinoFile::~CRhinoFile()
{
	delete m_pLayer;
  delete m_notes;
}

////////////////////////////////////////////////////////////////////////////////
// CRhinoFile::Read(eStatus status)
//	 Read next object out of file.  Can return status of good, eof,or crc_error.
//   Returns a NULL pointer only when status = eof.  The user should delete
//	 the returned object when finished with it.  
//
//	 If the next chunk in the file is a global attribute, like a layer definition,
//	 it is read in and recorded, then the next chunk is read.  When an object chunk
//   is read the correct derived type of CRhinoObject is made and initialized from
//   the information in the chunk.  If there is no specific type corresponding
//   to the typecode in the header an anonyous CRhinoChunk is created.
////////////////////////////////////////////////////////////////////////////////
CRhinoObject*
CRhinoFile::Read(eStatus & status)
{
  CRhinoChunk* pChunk = NULL;
  for (;;) {
	  pChunk = ReadChunk(status);
    if ( status != good ) {
      // CRC error or read failure
      if ( pChunk && !pChunk->IsObject() ) {
        delete pChunk;
        pChunk = NULL;
      }
      break;
    }

    // If we've just read a geometric object, hand it to caller
    if ( pChunk->IsObject() )
      break;

    // Rhino files contain several non-geometric objects at the top
    // level.  These include layer definitions, construction plane definitions,
    // viewport definitions, system tolerances, unit system, etc.
    // The IO toolkit ignores everything except layer, tolerance, notes and
    // unit system information.
		switch( pChunk->TypeCode() ) {

		case TCODE_LAYER:
      {
			  CRhinoLayer* pLayer = (CRhinoLayer*)pChunk;
			  AddLayer( pLayer);	// add to layer table
        pChunk = NULL;
			}
			break;

		case TCODE_UNIT_AND_TOLERANCES:
      {
        CRhinoUnitsAndTolerances* pUT = (CRhinoUnitsAndTolerances*)pChunk;
        m_bHaveRelTol = TRUE;
        m_rel_tol = pUT->m_rel_tol;
        SetAbsoluteTolerance( pUT->m_abs_tol );
        SetAngleTolerance( pUT->m_ang_tol );
        SetUnitSystem( pUT->m_units );
        delete pUT;
        pChunk = NULL;
      }
      break;

    case TCODE_NOTES:
      {
        if ( m_notes ) {
          delete m_notes;
          m_notes = 0;
        }
        m_notes = (CRhinoNotes*)(pChunk);
        pChunk = NULL;
      }

    default:
			delete pChunk; // something we don't care about
      pChunk = NULL;
      break;
		}
  }
	
	return (CRhinoObject*)(pChunk);
}

// returns pointer to the ith layer in the layer table.  Returns a NULL if
// i is out of range.
CRhinoLayer* 
CRhinoFile::Layer(int i) const {			
	if(i>=0 && i<NumLayers())
		return (*m_pLayer)[i];
	return NULL;
}

////////////////////////////////////////////////////////////////////////////////
// Returns index of layer or <0 if *pLay is not found in the layer table..
////////////////////////////////////////////////////////////////////////////////
int 
CRhinoFile::LookupLayer(const CRhinoLayer* pLay) const {
	// Simple linear search through the table
	int i=0;
	for(i=0; i<NumLayers(); i++)
		if( *Layer(i) == *pLay ) return i;
	return -1;
}

////////////////////////////////////////////////////////////////////////////////
// Add a layer to the table. retuns layer index.  Caller should use 
// LookupLayer first and gaurentee that the layer isn't already in the 
// table.
////////////////////////////////////////////////////////////////////////////////
int 
CRhinoFile::AddLayer( const CRhinoLayer* pLay){
	m_pLayer->Append( CRhinoLayerPtr( (CRhinoLayer*)pLay) );
	return m_pLayer->Size()-1;
}

////////////////////////////////////////////////////////////////////////////////
// Returns the layer index of a layer if it's in the table, otherwise
// in write mode, write layer to the file and update layer table.
// Returns -1 if passed a NULL pointer.
////////////////////////////////////////////////////////////////////////////////
int 
CRhinoFile::LayerIndex(const CRhinoLayer* play){
	int ind;
	if(play){
		ind = LookupLayer( play );
		if(ind<0 && Mode()==write){
			play->Write(*this);		// write layer definition
			ind = AddLayer( play );
		}
	}
	else ind =-1;
	return ind;
}


//
//  Low Level Read/Write Functions
//

////////////////////////////////////////////////////////////////////////////////
// Reads a chunk header.  Can return eof.  File mode must be read.
// Every BeginChunk() must have a matching EndChunk().  Chunks can be nested.
////////////////////////////////////////////////////////////////////////////////
CRhinoFile::eStatus
CRhinoFile::BeginReadChunk( unsigned int& typecode, int& info )
{
  int i0,i1;

	if ( m_mode != read ) {
	  throw CRhinoException(CRhinoException::illegal_input);
  }

	PushChunkRecord();   //New ChunkRecord
	
	// test for end of file
	int c= fgetc(m_pFile);
	if( c == EOF ){
		m_beof = TRUE;
		return eof;
	}
	c = ungetc(c,m_pFile);
	if( c == EOF ) {
		throw CRhinoException(CRhinoException::unable_to_seek);
  }
	
  ReadInteger( 1, &i0 );  // read the type code
  ReadInteger( 1, &i1 );  // read the second header int
  typecode = i0;
  info     = i1;

	// Initialize the ChunkRecord
  RhinoChunkRecord* record = CurChunk();

	record->typecode = i0;
	record->offset   = ftell(m_pFile);
  record->length   = 0;
  record->do_crc   = FALSE;
  record->crc      = 0;
  if( CRhinoChunk::IsLongTypeCode( i0 ) ) {
    record->length = i1;
    if ( info > 0 )
      record->offset += i1;
    if ( CRhinoChunk::IsCRCTypeCode( i0 ) ) {
		  record->do_crc = TRUE;
		  record->crc    = 1;      // CRC seed
    }
	}

	return good;
}

////////////////////////////////////////////////////////////////////////////////
// Writes a chunk header.  File mode must be write. 
// Always returns good.  When writing a chunk with ChunkStyle()==data, 
// the info field of the header is ignorred.  Every BeginChunk() must have a matching 
// EndChunk().  Chunks can be nested.
////////////////////////////////////////////////////////////////////////////////
CRhinoFile::eStatus
CRhinoFile::BeginWriteChunk( const CRhinoChunk& chunk )
{
  return BeginWriteChunk( chunk.TypeCode(), chunk.Value() );
}

CRhinoFile::eStatus
CRhinoFile::BeginWriteChunk( int typecode, int value /* = 0 */ )
{
	if( m_mode!= write ) {
		throw CRhinoException(CRhinoException::illegal_input);
	}

  if ( m_bWriteTols ) {
    m_bWriteTols = FALSE;
    CRhinoUnitsAndTolerances UT;
    UT.m_abs_tol = m_abs_tol;
    UT.m_rel_tol = m_rel_tol;
    UT.m_ang_tol = m_ang_tol;
    UT.m_units   = m_units;
    UT.Write( *this );
  }

  if ( CRhinoChunk::IsLongTypeCode( typecode ) )
    value = 0;

  PushChunkRecord();   //New ChunkRecord

	WriteInteger( 1, &typecode );		// write the type code

	// Initialize the ChunkRecord
  RhinoChunkRecord* record = CurChunk();
  record->length   = value;
	record->typecode = typecode;
	record->offset   = ftell(m_pFile);
	if(record->offset<0) {
		throw CRhinoException(CRhinoException::unable_to_seek);
  }
	WriteInteger( 1, &value );
  if( CRhinoChunk::IsCRCTypeCode( typecode ) ) {
		record->do_crc=TRUE;
		record->crc=1;      //seed the CRC
	}
	return good;
}

static void PadCRC(unsigned short& crc);

////////////////////////////////////////////////////////////////////////////////
// When writing, write CRC if appropriate and write the chunk length in header. 
// When reading, read and check the crc if appropriate. Can return crc_error on
// reading.
////////////////////////////////////////////////////////////////////////////////
CRhinoFile::eStatus
CRhinoFile::EndChunk()
{
	eStatus status = good;				// status to return
	RhinoChunkRecord* record = CurChunk();	

	switch(m_mode) {

	case read:
		if(record->do_crc) {
			/* read crc */
			unsigned short crc;
			ReadByte(2,&crc);
 			if (record->crc) 
				status = crc_error;
		}
    if ( ftell( m_pFile ) != record->offset )
      fseek( m_pFile, record->offset, SEEK_SET );
		break;
	
	case write:
		if (record->do_crc) {	        
      // write out the crc	
			PadCRC(record->crc);
			WriteShort(1,(short*) &record->crc);
		}
    if ( CRhinoChunk::IsLongTypeCode(record->typecode) ) {  
      // ChunkStyle()==data - compute length of data and crc
			int length = ftell(m_pFile) - record->offset;
			if (length < (int)sizeof(length)) 
				throw CRhinoException(CRhinoException::unable_to_seek);

			/* skip back to place where length is stored */
			if ( fseek(m_pFile,-length,1) != 0 ) 
				throw CRhinoException(CRhinoException::unable_to_seek);
 
			/* write number of bytes in this chunk */
			length -= sizeof(length);
			WriteInteger(1,&length);
			
			/* jump back to where we were */
  		if ( fseek(m_pFile,length,1) != 0 ) 
				throw CRhinoException(CRhinoException::unable_to_seek);
		} 
	
	}

	PopChunkRecord(); // done with that chunk
	return status;
}


// If reading, skips past the data field of the current chunk.  SkipChunk() is 
// ignored on writting. EndChunk() must still be called after SkipChunk(),
// although the crc will not be checked. 
void
CRhinoFile::SkipChunk()
{
	if( m_mode == read ) {
    RhinoChunkRecord* record = CurChunk();
		int end_of_chunk_offset = record->offset + record->length;	
		record->do_crc = FALSE;
		record->crc    = 0;
		if( fseek(m_pFile, end_of_chunk_offset, SEEK_SET) ) {
			throw CRhinoException( CRhinoException::unable_to_seek );
    }
	}
}

//////////////////////////////////////////////////////////////////////////////////
// Low level Read functions
//////////////////////////////////////////////////////////////////////////////////


// Reads Bytes.  If any read error throw exception.  
// update crc of current chunk if enabled.
void CRhinoFile::ReadByte(size_t count, void* pntr)
{

	if (!count) 
		return;

	/* read bytes */
	if (fread(pntr,count,1,m_pFile)!=1) 
		throw CRhinoException( CRhinoException::unable_to_read);
	
	if (CurChunk()->do_crc) {
		/* update crc remainder */ 
		unsigned char* b = (unsigned char*)(pntr);
		unsigned short 	remainder = CurChunk()->crc;
		while (count--) {
			unsigned short r1 = CRC_CCITT_TABLE[(remainder & ((unsigned short)0xff00))>>8];
			remainder = (remainder << 8) ^ (*b++);
			remainder ^= r1;  
		}
		CurChunk()->crc = remainder;
	} 
 	return;
  }

// Read a number of 16 bit shorts.  If executing on a BigEndian machine swap bytes.   
void CRhinoFile::ReadShort(size_t count, short* pntr){
	ReadByte((count)<<1, pntr);
	
	if(IsBigEndian()){   //Swap bytes
		unsigned char* b=(unsigned char*) (pntr);
		unsigned char  c;
		while(count--) {
			c = b[0]; b[0] = b[1]; b[1] = c;
			b += 2;
		}
	}
}
	
// Read a number of 32 bit integers.  If executing on a BigEndian machine swap bytes.   
void CRhinoFile::ReadInteger(size_t count, int* pntr){
	ReadByte((count)<<2, pntr);
	
	if(IsBigEndian()){   //Swap bytes
		unsigned char* b= (unsigned char*) (pntr);
		unsigned char  c;
		while(count--) {
			c = b[0]; b[0] = b[3]; b[3] = c;
			c = b[1]; b[1] = b[2]; b[2] = c;
			b += 4;
		}
	}
}

// Read a number of 64 bit doubles.  If executing on a BigEndian machine swap bytes.   
void CRhinoFile::ReadDouble(size_t count, double* pntr){
	ReadByte((count)<<3, pntr);
	
	if(IsBigEndian()) {   //Swap bytes
		unsigned char* b=(unsigned char*)(pntr);
		unsigned char  c;
		while(count--) {
			c = b[0]; b[0] = b[7]; b[7] = c;
			c = b[1]; b[1] = b[6]; b[6] = c;
			c = b[2]; b[2] = b[5]; b[5] = c;
			c = b[3]; b[3] = b[4]; b[4] = c;
			b += 8;
		}
	} 
	return;
}

// Returns length of string to be read in next
int CRhinoFile::StringLength()
{
	int len;
	ReadInteger(1,&len);
	return len;
}

// Read a string (not necc. null terminated).  
void CRhinoFile::ReadString( size_t length, char* pntr)
{
	ReadByte(length, pntr);
	return;
}

void CRhinoFile::ReadInteger( int* pi )
{
  ReadInteger( 1, pi );
}

void CRhinoFile::ReadIntArray( CRhinoIntArray& a )
{
  int sz;
  ReadInteger( &sz );
  if ( sz > a.m_capacity )
    a.SetCapacity( sz );
  a.SetSize(sz);
  if ( a.m_size > 0 )
    ReadInteger( a.m_size, a.m_p );
}

void CRhinoFile::ReadDouble( double* pd )
{
  ReadDouble( 1, pd );
}

// ReadString allocates and initializes a null terminated string.
// User must delete[] when done.
void CRhinoFile::ReadString(char*& str)
{   

	int length = StringLength();
	str = new char[length+1];

	ReadByte(length,str);
	str[length] = 0;
	return;
}

//////////////////////////////////////////////////////////////////////////////////
// Low level Write functions
//////////////////////////////////////////////////////////////////////////////////
// Write a number of bytes 
void CRhinoFile::WriteByte( size_t count, const char* pntr)
{
  
  unsigned short r1;

  if (!count) 
    return;

  /* write bytes */
  if (fwrite(pntr,count,1,m_pFile)!=1) 
	  throw CRhinoException(CRhinoException::unable_to_write);
  
  if (CurChunk()->do_crc) {
    /* update crc remainder */ 
    const unsigned char* b = (const unsigned char*) (pntr);
    unsigned short remainder = CurChunk()->crc;
    while (count--) {
      r1 = CRC_CCITT_TABLE[(remainder & ((unsigned short)0xff00))>>8];
      remainder = (remainder << 8) ^ (*b++);
      remainder ^= r1;  
    }
    CurChunk()->crc = remainder;
  }

  return;
}

// Write a number of 16 bit shorts
void CRhinoFile::WriteShort( size_t count, const short* pntr){
  if(IsBigEndian()){
	  const char* b = (const char*) pntr;
	  while (count) {
		WriteByte(1,b+1);
		WriteByte(1,b);
		b += 2;
		count--;
	  }
  }
  else
	  WriteByte(count<<1, (const char*) pntr);
  return;
}

// Write a number of 32 bit integers
void CRhinoFile::WriteInteger( size_t count, const int* pntr){
  if(IsBigEndian()){
	  const char* b = (const char*) pntr;
	  while (count) {
		WriteByte(1,b+3);
		WriteByte(1,b+2);
		WriteByte(1,b+1);
		WriteByte(1,b);
		b += 4;
		count--;
	  }
  }
  else
	  WriteByte(count<<2, (const char*) pntr);
  return;
}

// Write a number of IEEE doubles
void CRhinoFile::WriteDouble( size_t count, const double* pntr){
 if(IsBigEndian()){
	  const char* b = (const char*) pntr;
	  while (count) {
		WriteByte(1,b+7);
		WriteByte(1,b+6);
		WriteByte(1,b+5);
		WriteByte(1,b+4);
		WriteByte(1,b+3);
		WriteByte(1,b+2);
		WriteByte(1,b+1);
		WriteByte(1,b);
		b += 8;
		count--;
	  }
  }
  else
	  WriteByte(count<<3, (const char *) pntr);
  return;
}
	
// Write a null terminated string.  The length is written as an integer then
// followed by the string characters excluding the terminating null character.
void CRhinoFile::WriteString(const char* pntr){
  int len = (pntr) ? strlen(pntr) : 0;
  WriteInteger(1,&len);
  if (len > 0)
    WriteByte(len,pntr); 
}

void CRhinoFile::WriteInteger( int i )
{
  WriteInteger( 1, &i );
}

void CRhinoFile::WriteIntArray( const CRhinoIntArray& a )
{
  WriteInteger( a.Size() );
  if ( a.Size() > 0 )
    WriteInteger( a.Size(), a.Array() );
}

void CRhinoFile::WriteDouble( double d )
{
  WriteDouble( 1, &d );
}

static void PadCRC(unsigned short& crc)
{ /* used to add two extra 0 bytes to remainder */
  int count=2;
  unsigned short remainder, r1;
  unsigned char zero=0;
  remainder = crc;
  while (count--) {
    r1 = CRC_CCITT_TABLE[(remainder & ((unsigned short)0xff00))>>8];
    remainder = (remainder << 8) ^ zero;
    remainder ^= r1;  
  }
  crc = remainder;
  return;
}


//Returns true if executing machine is big endian.
BOOL CRhinoFile::ComputeBigEndian(){
  /* Sparc, MIPS, etc. CPUs have big endian byte order */
  union {
    int i;
    char c[4];
  } u;
  u.i = 1;
  return ( (u.c[3] == 1) ? TRUE : FALSE );
}

////////////////////////////////////////////////////////////////////////////////
// CRhinoFile::ReadChunk(eStatus status)
//	 Read next chunk from the file.  Can return status of good, eof or crc_error.
//   Returns a non-NULL pointer only when status == good.  The user should delete
//	 the returned chunk when finished with it.  
//
//	 When a chunk
//   is read the correct derived type of CRhinoChunk is made and initialized from
//   the information in the chunk.  If there is no specific type corresponding
//   to the typecode in the header an anonymous CRhinoChunk is created.
////////////////////////////////////////////////////////////////////////////////
CRhinoChunk*
CRhinoFile::ReadChunk( eStatus& status )
{
	CRhinoChunk* pChunk = NULL;
  eStatus read_status = not_supported;
  eStatus end_status  = not_supported;
  unsigned int typecode = 0;
  int info = 0;

	status = BeginReadChunk( typecode, info );

	if ( status == good ) {
    // if the typecode is in the CRhinoChunk registry, then the
    // call to CreateNew() will create the appropriate CRhinoChunk
    // child and read its defining data.
	  pChunk = CRhinoChunk::CreateNew( typecode, info );
 	  read_status = pChunk->Read(*this);  // read in chunk information but not crc
	  end_status = EndChunk();
    if ( status == good ) {
      status = read_status;
      if ( status == good ) {
        status = end_status;
      }
    }
  }
	return pChunk;
}

// Computes value for superfluous knot used in systems like OpenGL and 3dsMax
double RhinoSuperfluousKnot( int order, int cv_count, const double* knot,
                             int end )
{
  double k;
  const int knot_count = order+cv_count-2;
  // gets superfluous knot for translation to other formats
  k = knot[(end) ? knot_count-1 : 0];
  if (order > 2 && cv_count >= 2*order - 2 && cv_count >= 6 ) {
    // check for non-clamped knots
    if (end) {
      if ( knot[cv_count-1] < knot[knot_count-1] )
        k += (knot[order+1] - knot[order]);
    }
    else {
      if ( knot[0] < knot[order-2] )
        k -= (knot[cv_count-order+1] - knot[cv_count-order]);
    }
  }
  return k;
}

// Used in bounding box computations
BOOL RhinoBoundingBox(
       int dim,
       BOOL bIsRational,
       int count,
       int stride,
       const double* P,
       double box_min[3],
       double box_max[3],
       BOOL bGrowBox  // = FALSE
       )
{
  BOOL rc = FALSE;
  int i, j;
  double x, w;

  if ( dim < 1 || dim > 3 || stride < dim || count < 1 || !P )
    return rc;

  rc = TRUE;

  if ( !bGrowBox ) {
    box_min[0] = box_min[1] = box_min[2] = 0.0;
    box_max[0] = box_max[1] = box_max[2] = 0.0;
    if ( bIsRational ) {
      if ( P[dim] == 0.0 ) {
        rc = RhinoBoundingBox( dim, bIsRational, count-1, stride, P+stride, box_min, box_max, bGrowBox );
        return FALSE;
      }
      w = 1.0/P[dim];
    }
    else {
      w = 1.0;
    }
    for ( j = 0; j < dim; j++ ) {
      box_min[j] = box_max[j] = w*P[j];
    }
    P += stride;
    count--;
  }

  if ( bIsRational ) {
    for ( i = 0; i < count; i++ ) {
      if ( P[dim] != 0.0 ) {
        w = 1.0/P[dim];
        for ( j = 0; j < dim; j++ ) {
          x = w*P[j];
          if ( x < box_min[j] )
            box_min[j] = x;
          else if ( x > box_max[j] )
            box_max[j] = x;
        }
      }
      else
        rc = FALSE;
      P += stride;
    }
  }
  else {
    for ( i = 0; i < count; i++ ) {
      for ( j = 0; j < dim; j++ ) {
        if ( P[j] < box_min[j] )
          box_min[j] = P[j];
        else if ( P[j] > box_max[j] )
          box_max[j] = P[j];
      }
      P += stride;
    }
  }

  return rc;
}

BOOL RhinoIsEqual(
       int dim,
       const double* pointA,
       const double* pointB
       )
{
  const double eps = RHINO_SQRT_EPSILON;
  double d, m, x, x0, x1;
  m = eps;
  d = 0.0;
  while(dim--) {
    x0 = *pointA++;
    x1 = *pointB++;
    x = fabs(x0);
    if (x > m) m = x;
    x = fabs(x1);
    if (x > m) m = x;
    x = fabs(x0-x1);
    if (x > d) d = x;
  }
  return (d <= m*eps) ? 1 : 0;
}

// Used to determine when a knot vector is periodic
BOOL RhinoIsKnotVectorPeriodic(
       int order,
       int cv_count,
       const double* knot
       )

{
  const double eps = RHINO_SQRT_EPSILON;
  double tol;
  const double* k1;
  int i;

  if ( order <= 1 || cv_count < order || !knot )
    return FALSE;
  if (order < 3 || cv_count < order+2) 
    return FALSE;
  tol = fabs(knot[order-1] - knot[order-3])*eps;
  if (tol < fabs(knot[cv_count-1] - knot[order-2])*eps) 
    tol = fabs(knot[cv_count-1] - knot[order-2])*eps;
  k1 = knot+cv_count-order+1;
  i = 2*(order-2);
  while(i--) {
    if (fabs(knot[1] - knot[0] + k1[0] - k1[1]) > tol)
      return FALSE;
    knot++; k1++;
  }
  return TRUE;
}

// Used to determine when a knot vector is clamped
BOOL RhinoIsKnotVectorClamped(
       int order,
       int cv_count,
       const double* knot
       )

{
  if ( order <= 1 || cv_count < order || !knot )
    return FALSE;
  return ( knot[0] == knot[order-2] && knot[cv_count-1] == knot[order+cv_count-3] ) ? TRUE : FALSE;
}

BOOL RhinoGetKnotVectorDomain(
       int order,
       int cv_count,
       const double* knot,
       double* k0, double* k1
       )
{
  if ( order < 2 || cv_count < order || knot == NULL )
    return FALSE;
  if ( k0 )
    *k0 = knot[order-2];
  if ( k1 )
    *k1 = knot[cv_count-1];
  return TRUE;
}

RHINO_KNOT_STYLE RhinoKnotVectorStyle( 
       int order,
       int cv_count,
       const double* knot
       )
{
  RHINO_KNOT_STYLE s = rh_unknown_knots;
  if ( order >= 2 && cv_count >= order) {
    const int knot_count = order+cv_count-2;
    const double delta = 0.5*((knot[order-1] - knot[order-2]) + (knot[cv_count-1] - knot[cv_count-2]));
    const double ktol = delta*1.0e-6;
    int i;
    if ( RhinoIsKnotVectorClamped( order, cv_count, knot ) ) {
      s = rh_bezier_ends;
      for ( i = order-1; i < cv_count-1; i++ ) {
        if ( fabs(knot[i] - knot[i-1] - delta) > ktol ) {
          break;
        }
      }
      if ( i >= cv_count-1 ) {
        s = rh_quasi_uniform_knots;
      }
      else {
        const int degree = order-1;
        for ( i = 0; i < knot_count; i += degree ) {
          if ( knot[i] != knot[i+degree-1] )
            break; 
        }
        if ( i >= knot_count )
          s = rh_piecewise_bezier_knots;
      }
    }
    else {
      // check for uniform knots
      s = rh_non_uniform_knots;
      for ( i = 1; i < knot_count; i++ ) {
        if ( fabs(knot[i] - knot[i-1] - delta) > ktol ) {
          break;
        }
      }
      if ( i >= knot_count )
        s = rh_uniform_knots; 
    }
  }
  return s;
};

// Used to reverse knot vectors
BOOL RhinoReverseKnotVector(
       int order,
       int cv_count,
       double* knot
       )
{
  if ( order < 2 || cv_count < order || knot == NULL )
    return FALSE;
  const int knot_count = (order+cv_count-2);
  double t;
  int i, j;
  for ( i = 0, j = knot_count-1; i <= j; i++, j-- ) {
    t = knot[i];
    knot[i] = -knot[j];
    knot[j] = -t;
  }
  return TRUE;
}

// Used to reverse lists of points
BOOL RhinoReversePointList(
       int dim,
       int count,
       int stride,
       double* p
       )
{
  if ( count < 1 || dim < 1 || dim > stride || stride < 1 || p == NULL )
    return FALSE;
  if ( count == 1 )
    return TRUE;
  const size_t ele_size = dim*sizeof(*p);
  void* t = rhmalloc(ele_size);
  int i, j;
  for ( i = 0, j = (count-1)*stride; i < j; i += stride, j -= stride ) {
    memcpy( t,   p+i, ele_size );
    memcpy( p+i, p+j, ele_size );
    memcpy( p+j, t,   ele_size );
  }
  rhfree(t);
  return TRUE;
}

BOOL RhinoGetPerpendicular(
       int dim, 
       const double* V, 
       double* Vperp
       )
{
  double M,m,x,a,b;
  int i,j,k,tmp;
	if (dim == 3) {
    k = 2;
    i = (fabs(V[1]) > fabs(V[0])) ? 1 : 0;
    if (fabs(V[2]) > fabs(V[i]))
      i = 2;
    j = (i+1)%3; k = (i+2)%3;
    if (fabs(V[k]) > fabs(V[j])) 
      {tmp=j;j=k;k=tmp;}
    a = V[i];
    b = -V[j];
    Vperp[i] = b;
    Vperp[j] = a;
    Vperp[k] = 0.0;
    if (a == 0.0)
      return FALSE;
    b /= a;
    a = 1.0/(fabs(a)*sqrt(1.0+b*b));
    Vperp[i] *= a; Vperp[j] *= a;
  }
  else if (dim == 2) {
    if ( V[0] == 0.0 && V[1] == 0.0 )
      return FALSE;
    Vperp[1] =  V[0];
    Vperp[0] = -V[1];
  }
  else if (dim < 2) {
    *Vperp = *V;
    return FALSE;
  }
  else {
    memset(Vperp,0,dim*sizeof(*Vperp));
    if (dim < 2) return -1;
    if (fabs(V[0]) < fabs(V[1]))
      {i=1;j=0;}
    else 
      {i=0;j=1;}
    M = fabs(V[i]); m = fabs(V[j]);
    for (k = 2; k < dim; k++) {
      x = fabs(V[k]);
      if (x > M) 
        {j=i;m=M;i=k;M=x;} 
      else if (x > m)
        {j=k;m=x;}
    }
    if (M == 0.0) 
      return FALSE;
    Vperp[j] = V[i];
    Vperp[i] = -V[j];
  }
  return TRUE;
}

double RhinoDistance( int dim, const double* A, const double* B )
{
  double ws[6];
  double* v;
  int i;
  double d = 0.0;
  if ( dim > 0 && A && B ) {
    v = (dim > 6) ? (double*)rhmalloc(dim*sizeof(*v)) : ws;
    i = dim;
    while (i--)
      *v++ = *B++ - *A++;
    d = RhinoVectorLength(dim,v-dim);
    if ( dim > 6)
      rhfree(v);
  }
  return d;
}

void RhinoCrossProduct( const double* A, const double* B, double* AxB )
{
  AxB[0]=A[1]*B[2]-A[2]*B[1];
  AxB[1]=A[2]*B[0]-A[0]*B[2];
  AxB[2]=A[0]*B[1]-A[1]*B[0];
}

double RhinoVectorLength(
         int dim, 
         const double* v
         )
{
  double a, b, c, len;
  switch(dim) {
  case 1:
    len = fabs(*v); 
    break;
  case 2:
    a = fabs(*v++); b = fabs(*v);
    if (a > b) 
      {b /= a; len = a*sqrt(1.0+b*b);}
    else if (b > a) 
      {a /= b; len = b*sqrt(1.0+a*a);}
    else
      len = a*1.414213562373095048801689;
    break;
  case 3:
    a = fabs(*v++); b = fabs(*v++); c = fabs(*v);
    if (a >= b) {
      if (a >= c) {
        if (a == b && a == c) 
          len = a*1.732050807568877293527446;
        else
          {b /= a; c /= a; len = a*sqrt(1.0 + (b*b + c*c));}
      }
      else 
        {a /= c; b /= c; len = c*sqrt(1.0 + (a*a + b*b));}
    }
    else if (b >= c) 
      {a /= b; c /= b; len = b*sqrt(1.0 + (a*a + c*c));}
    else 
      {b /= c; a /= c; len = c*sqrt(1.0 + (a*a + b*b));}
    break;
  default:
    len = *v++;
    len = len*len;
    while (--dim) {a = *v++; len += a*a;}
    len = sqrt(len);
    break;
  }
  return len;
}


const CRhinoNotes* CRhinoFile::Notes( ) const
{
  return m_notes;
}


BOOL CRhinoFile::GetAbsoluteTolerance( double* abs_tol ) const
{
  if ( abs_tol )
    *abs_tol = m_abs_tol;
  return m_bHaveAbsTol;
}

BOOL CRhinoFile::SetAbsoluteTolerance( double abs_tol )
{
  BOOL rc = FALSE;
  if ( abs_tol > 0.0 ) {
    m_abs_tol = abs_tol;
    m_bHaveAbsTol = TRUE;
    rc = TRUE;
  }
  return rc;
}

BOOL CRhinoFile::GetAngleTolerance( double* ang_tol ) const
{
  if ( ang_tol )
    *ang_tol = m_ang_tol;
  return m_bHaveAngTol;
}

BOOL CRhinoFile::SetAngleTolerance( double ang_tol )
{
  BOOL rc = FALSE;
  if ( ang_tol > 0.0 ) {
    m_ang_tol = ang_tol;
    m_bHaveAngTol = TRUE;
    rc = TRUE;
  }
  return rc;
}

BOOL CRhinoFile::GetUnitSystem( eUnitSystem* us )  const
{
  if ( us )
    *us = m_units;
  return m_bHaveUnits;
}

BOOL CRhinoFile::SetUnitSystem( eUnitSystem us )
{
  BOOL rc = FALSE;
  int i = us;
  if ( i >= no_units && i <= miles ) {
    m_units = us;
    m_bHaveUnits = TRUE;
    rc = TRUE;
  }
  return rc;
}

int Rhino_stricmp( const char* s1, const char* s2 )
{
  // choose the one that works for your compiler
#if defined(WIN32)
  //return stricmp(s1,s2);
  return _stricmp(s1,s2);
#else
  return strcasecmp(s1,s2);
#endif
}

