/* $Header: /Src/RhinoIO/rhinoio_file.h 12    11/10/99 2:56p Dalelear $ */
/* $NoKeywords: $ */
//
// Copyright (c) 1993-1998 Robert McNeel & Associates. All rights reserved.
// Rhinoceros is a registered trademark of Robert McNeel & Assoicates.
//
// THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
// ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
// MERCHANTABILITY ARE HEREBY DISCLAIMED.
//				
////////////////////////////////////////////////////////////////
//
//   Defines CRhinoFile
//
////////////////////////////////////////////////////////////////

#if !defined(RHINO_FILE_INC_)
#define RHINO_FILE_INC_


///////////////////////////////////////////////////////////////////////////////
// Class  CRhinoFile
//
//  Class for reading and writing Rhino files.  A Rhino 3dm model file
//  contains object models and attributes.  This class has both a high level
//  and a low level interface for reading and writing Rhino files.
//
//  The high level interface to the CRhinoFile consists of reading and writing
//  CRhinoObjects to the file.  Attributes can be written to the file explicitly
//  or will be written as needed while writing objects.  Attributes read from the 
//  file are accessible from the CRhinoFile
//
//	The low level member functions provide access to the chunks that make up
//  a Rhino 3dm model file.  The file is a binary file consisting of a 32 byte 
//	ascii header identifying it as a Rhino 3dm file, followed by a sequence of chunks.
//  Each chunk has a 2 int (8 byte) header and an optional, variable length data field.
//  followed optionally by a 2 byte CRC.  Sub-chunks can be nested indside the data 
//  field.
///////////////////////////////////////////////////////////////////////////////

typedef struct 
{
  // utility struct used internally
  int typecode; 
	int length;    // chunk length on read including crc
	int offset;    // When writing, offset into file to chunk length field. 
                 // When reading, offset into file of byte after chunk end

	BOOL do_crc;   
	unsigned short crc;      // CRC accumulator.  Only data is CRC checked not headers or enclosed chunks
	unsigned short padding;  // keep struct size a multiple of 4

  int index; // index in m_Hdr[] stack
} RhinoChunkRecord;

class CRhinoNotes;

class CRhinoFile 
{
public:

	enum eMode {      // file access mode
    read, 
    write
  };

	enum eStatus {    // Status from reads and writes 
    good          = 0, // 
    crc_error     = 1, // read detected a CRC error or an unexpected value
    eof           = 2, // read found end of file
    not_supported = 3, // read or write not supported and won't be in the future
    not_finished  = 4  // read or write failed because code isn't finished
  };

  enum eUnitSystem {
    no_units    =  0,  // (default setting)
    microns     =  1,  // 1.0e-6 meters
    millimeters =  2,  // 1.0e-3 meters
    centimeters =  3,  // 1.0e-2 meters
    meters      =  4,
    kilometers  =  5,  // 1.0e+3 meters
    microinches =  6,  // 1.0e-6 inches
    mils        =  7,  // 1.0e-3 inches
    inches      =  8,  // 0.0254 meters
    feet        =  9,  // 12 inches
    miles       = 10   // 63360 inches
  };

  
  ////////////////////////////////////////////////////////////////////////
	CRhinoFile(FILE*, eMode, BOOL=FALSE ); // A pointer to an open file is required.
                             // The mode must be read or write.
                             // See example_read.cpp and example_write.cpp
                             // for details.
                             // The last BOOL defaults to FALSE and is
                             // for internal use only.
	~CRhinoFile();

  // basic file status queries
	BOOL  EndOfFile() const;	// End of file encountered. Next read will fail.  
  eMode Mode() const;			  // file mode (read or write)
  int Version() const;			// Version of 3dm file

	CRhinoObject* Read(eStatus & );	// Read next object in file.  See
                                  // example_read.cpp for usage.

  ////////////////////////////////////////////////////////////////////////
  // user comments (set/view with Rhino "Notes" command
  const CRhinoNotes* Notes() const; // returns NULL if file has no notes.

  ////////////////////////////////////////////////////////////////////////
  // System tolerance and unit system settings.  
  //
  // Angle tolerance is in radians.  Absolute tolerance is the largest
  // deviation permitted in fitting operations and the largest gap 
  // permitted in joining operations.
  //
  // When reading files, the tolerances and unit system settings are set
  // as the information is found in the file.  The file example_read.cpp
  // demonstrates when you can read tolerances.
  //
  // When writing files, the tolerances and unit system settings must be set
  // before the first geometric object is written.  The file example_write.cpp
  // demonstrates how and when to set tolerances and/or unit systems.
  //
  // Tolerances are with respect to the world coordinates in the
  // geometry definitions.  The unit system and tolerance settings
  // are treated independently.
  //
  // Rhino's default values are
  //   absolute tolerance = 0.01
  //   angle tolerance    = 3.141592653589793238462643*3.0/180.0 (3 degrees)
  //   unit system        = no units

  BOOL GetAbsoluteTolerance( double* ) const; // Returns FALSE if tolerance has
                                              // not been read or set.  In this
                                              // case the default tolerance is
                                              // returned.
  BOOL SetAbsoluteTolerance( double );        // Sets tolerance that will be
                                              // saved when file is written.
  BOOL GetAngleTolerance( double* ) const;    // Returns FALSE if angle tolerance
                                              // has not been read or set.  In 
                                              // this case the default angle
                                              // tolerance is returned.
                                              // The angle tolerance is in radians.
  BOOL SetAngleTolerance( double );           // Sets angle tolerance (radians)
                                              // that will be saved when file 
                                              // is written.
  BOOL GetUnitSystem( eUnitSystem* )  const;  // Returns FALSE if a unit system
                                              // has not been read or set.  In
                                              // this case "no_units" is returned.
  BOOL SetUnitSystem( eUnitSystem );          // Sets the unit system.

  ////////////////////////////////////////////////////////////////////////////
  // Layers
  //
  // A CRhinoFile maintains a table of Layer references.  As layers are read
	// or written references are added to the table.  When writing an object 
	// with a layer, the layer is automatically written to the file if it 
  // hasn't been written.  See example_write.cpp for details.
	int NumLayers() const;						// Number of layer in table
	
	CRhinoLayer* Layer(int i) const;					// Pointer to the ith layer

	int LookupLayer(const CRhinoLayer*) const;// Returns index of layer or <0.
	int LayerIndex(const CRhinoLayer*);				// if not already there write layer to
																						// the file and update table

  ////////////////////////////////////////////////////////////////////////////
	// Low Level Functions
  //
	//	These functions are concerened with reading and writing chunks to
	//  the file.  There is an abstract base class CRhinoChunk to represent 
	//  chunks in general.  Derived from this are various specific types of chunks.
	//  The ReadChunk(...) function and CRhinoChunk::Write(...) function read and
	//  write chunks.  

	CRhinoChunk* ReadChunk(eStatus & );					// read next chunk in file   

	//	These functions are concerened with reading and writing parts of chunks.
	//  Chunks are initiated and terminated by BeginChunk and EndChunk.  These
	//  functions handle the header and crc, as well as an internal stack which
	//  keeps track of the chunk nesting.  
	//
	//  The other functions handle reading and writing data to the chunk. 

	
	eStatus BeginReadChunk(		// Reads a chunk's typecode and info header.
                            // Can return eof. Every BeginReadChunk() 
                            // must have a matching EndChunk().  Chunks
                            // can be nested.
    unsigned int&,  // chunk typecode
    int&            // chunk info integer
    );

	eStatus BeginWriteChunk(	// Writes a chunk header.  Always returns good.
		                        // When writing a chunk with ChunkStyle()==data, the 						
		      									// info field of the header is ignorred.          
														// Every BeginWriteChunk() must have a matching
														// EndChunk().  Chunks can be nested.
    const CRhinoChunk& 
    );

	eStatus BeginWriteChunk(	// Writes a chunk header.  Always returns good.
		                        // When writing a chunk with ChunkStyle()==data, the 						
		      									// info field of the header is ignorred.          
														// Every BeginWriteChunk() must have a matching
														// EndChunk().  Chunks can be nested.
    int,    // typecode
    int = 0 // value for integer chunks
    );

	eStatus  EndChunk();	// When writing, write CRC if appropriate and chunk
						  // length in header.  When reading, read and check 
						 	// the crc if appropriate. Can return crc_error on
							// reading.

	void SkipChunk();		// If reading skip past the data field of the current 
							// chunk.  SkipChunk() is ignorred on writting; 
							// EndChunk() must still be called after SkipChunk(),
							// although the crc will not be checked. 

	void ReadByte(         // Read a number of bytes
			size_t,           // number of bytes to read
			void*   
			);
	void ReadShort(       // Read a number of 16 bit shorts
			size_t,          // number of shorts to read
			short*    
			);  
	void ReadInteger(     // Read a number of 32 bit integers
			size_t,	      // number of ints to read
			int*      
			); 
  void ReadIntArray( CRhinoIntArray& );
	void ReadDouble(     // Read a number of IEEE doubles
			size_t,         // number of doubles
			double*
			);
  int StringLength();  // Returns length of string to be read in next

	void ReadString(     // Read a string (not null terminated)
			size_t,          // Length returned from StringLength()
			char*
			); 
	void ReadString(char*& str);	// ReadString allocates and initializes a 
																// null terminated string.
																// User must delete[] when done.
	void ReadInteger( int* );     // Read a single 32 bit integer
	void ReadDouble( double* );  // Read a single IEEE double
		
	void WriteByte(      // Write a number of bytes
			size_t, 
			const char*
			);
	void WriteShort(     // Write a number of 16 bit shorts
			size_t, 
			const short*
			);  
	void WriteInteger(   // Write a number of 32 bit integers
			size_t, 
			const int*
			); 
  void WriteIntArray( const CRhinoIntArray& );
	void WriteDouble(    // Write a number of IEEE doubles
			size_t, 
			const double* 
			); 
	void WriteString(	// Write a null terminated string.  Terminating
			const char*		// null character is not written
			);
	void WriteInteger( int );   // Write a single 32 bit integer
	void WriteDouble( double ); // Write a single IEEE double
	
	FILE* GetFile();    // Don't use this.


private:
	eMode m_mode;								// file mode, read or write
	FILE* m_pFile;							// file pointer
	BOOL m_beof;								// end of file encountered
	BOOL m_bBigEndian;					// TRUE if Big Endian 
	int m_version;							// version of file being read or written

  BOOL m_bWriteTols;          // TRUE if tols and units have not been written
  BOOL m_bHaveAbsTol;
  double m_abs_tol;
  BOOL m_bHaveRelTol;
  double m_rel_tol;
  BOOL m_bHaveAngTol;
  double m_ang_tol;
  BOOL m_bHaveUnits;
  eUnitSystem m_units;

  CRhinoNotes* m_notes;

	BOOL IsBigEndian();					// TRUE if Big Endian 

  BOOL ComputeBigEndian();		// computes if executing processor is big endian

	CRhinoLayerTable* m_pLayer; // Layer table is a dynamic array of CRhinoLayerPtr

	int AddLayer( const CRhinoLayer* );		// Add a layer to the table. rtns index

	// Stack of chunk records. Store one record for each nested chunk
	//  currently open (between BeginChunk() and EndChunk()).  
	CRhinoArray<RhinoChunkRecord> m_Hdr;
	RhinoChunkRecord* CurChunk();         // current chunk
	void PushChunkRecord();
	void PopChunkRecord();

	CRhinoFile();										// prohibit
	CRhinoFile(const CRhinoFile&);	// prohibit
	CRhinoFile& operator=(const CRhinoFile&); // prohibit
};

inline
BOOL CRhinoFile::EndOfFile() const
  {return m_beof; }

inline
CRhinoFile::eMode CRhinoFile::Mode() const
  {return m_mode; }	

inline
int CRhinoFile::Version() const
  {return m_version; }         

inline
int CRhinoFile::NumLayers() const
  { return m_pLayer->Size(); }

inline
FILE* CRhinoFile::GetFile()
	{ return m_pFile; }

inline
BOOL CRhinoFile::IsBigEndian()
  { return m_bBigEndian; }

#endif
