/* $Header: /Src/RhinoIO/rhinoio_gl.cpp 17    11/10/99 2:56p Dalelear $ */
/* $NoKeywords: $ */
//
// Copyright (c) 1993-1998 Robert McNeel & Associates. All rights reserved.
// Rhinoceros is a registered trademark of Robert McNeel & Assoicates.
//
// THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
// ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
// MERCHANTABILITY ARE HEREBY DISCLAIMED.
//				
////////////////////////////////////////////////////////////////

#include "rhinoio.h"    // RhinoIO class definitions
#include "rhinoio_gl.h" // RhinoGL() function declarations

void RhinoGL( const int order,     // CRhinoNurbs order
              const int cv_count,  // CRhinoNurbs cv count
              const double* knot,  // CRhinoNurbs knot vector
              GLfloat* glknot,     // GL knot vector
              int bPermitScaling,  // TRUE if knot scaling is allowed
              double* scale        // If not NULL and knot scaling is
                                   // allowed, then the scaling
                                   // parameters are returned here.
                                   // glknot = (knot - scale[0])*scale[1]
            )
{
  // Because GL uses floats instead of doubles for knot vectors and
  // because some GLs are intolerant of closely spaced knots,
  // the returned glknots[] may be re-scaled when bPermitScaling
  // is TRUE.  When the knots belong to a trimmed surface, any rescaling
  // done to the surface's knots must be applied to the trimming geometry.

  const int knot_count = order + cv_count - 2;
  const int nknots = knot_count+2;
  
  // GL knot vectors have old-fashioned extra knot at start and end
  const double k0 = RhinoSuperfluousKnot( order, cv_count, knot,0);
  const double k1 = RhinoSuperfluousKnot( order, cv_count, knot,1);

  if ( scale ) {
    scale[0] = 0.0;
    scale[1] = 1.0;
  }
  
  int i, j;
  if ( bPermitScaling ) {
    double d0 = knot[order-2];
    double dk = 1.0;
    if ( bPermitScaling ) {
      double dmin = 1.0;
      double dmax = 1.0;
      double d;
      for ( i = 1; i < knot_count; i++ ) {
        d = knot[i] - knot[i-1];
        if ( d <= 0.0 )
          continue; // multiple knot
        if ( d < dmin )
          dmin = d;
        else if ( d > dmax )
          dmax = d;
      }
      if ( dmin > 0.0 && dmax >= dmin ) {
        if ( dmin < 1.0e-2 )
          dk = 1.0e-2/dmin;
        else if ( dmax > 1.0e4 ) {
          if ( 1.0e4*dmin >= 1.0e-2*dmax )
            dk = 1.0e4/dmax;        
        }
      }
    }
    if ( scale ) {
      scale[0] = d0;
      scale[1] = dk;
    }
    glknot[0] = (GLfloat)((k0-d0)*dk);
    for( i = 1, j = 0; j < knot_count; i++, j++ )
      glknot[i] = (GLfloat)((knot[j]-d0)*dk);
    glknot[nknots-1] = (GLfloat)((k1-d0)*dk);
  }
  else {
    glknot[0] = (GLfloat)k0;
    for( i = 1, j = 0; j < knot_count; i++, j++ )
      glknot[i] = (GLfloat)knot[j];
    glknot[nknots-1] = (GLfloat)k1;
  }
}

static void GetGLCV( const int dim, const int is_rat, const double* cv,
                     double xform[4][4], 
                     GLfloat* glcv )
{
  if ( xform ) {
    const double x = cv[0];
    const double y = cv[1];
    const double z = (dim == 3) ? cv[2] : 0.0;
    const double w = (is_rat) ? cv[dim] : 1.0;
    glcv[0] = (GLfloat)(xform[0][0]*x + xform[0][1]*y + xform[0][2]*z + xform[0][3]*w);
    glcv[1] = (GLfloat)(xform[1][0]*x + xform[1][1]*y + xform[1][2]*z + xform[1][3]*w);
    if ( dim == 3 )
      glcv[2] = (GLfloat)(xform[2][0]*x + xform[2][1]*y + xform[2][2]*z + xform[2][3]*w);
    if ( is_rat )
      glcv[dim] = (GLfloat)(xform[3][0]*x + xform[3][1]*y + xform[3][2]*z + xform[3][3]*w);
  }
  else {
    glcv[0] = (GLfloat)cv[0];
    glcv[1] = (GLfloat)cv[1];
    if ( dim == 3)
      glcv[2] = (GLfloat)cv[2];
    if ( is_rat == 3)
      glcv[dim] = (GLfloat)cv[dim];
  }
}

void RhinoGL( const CRhinoNurbsCurve& c,
              GLUnurbsObj* nobj, // created with gluNewNurbsRenderer )
              GLenum type, // = 0 (and type is automatically set)
              int bPermitKnotScaling,
              double* knot_scale,
              double xform[][4]
            )
{
}

void RhinoGL( int dim, int is_rat, int nurb_order, int cv_count,
              const double* knot_vector, 
              int cv_stride, const double* cv,
              GLUnurbsObj* nobj, // created with gluNewNurbsRenderer )
              GLenum type, // = 0 (and type is automatically set)
              int bPermitKnotScaling,
              double* knot_scale,
              double xform[][4]
            )
{
  int i;

  double dbg_ks[2];

  GLint nknots = nurb_order + cv_count; // GL knot count = TL knot count + 2
  GLfloat* knot = (GLfloat*)rhmalloc( nknots*sizeof(*knot) );
  RhinoGL( nurb_order, cv_count, knot_vector, knot, bPermitKnotScaling, dbg_ks);//knot_scale );

  // control vertices
  const int cv_size = (is_rat) ? dim+1: dim;
  GLint stride = cv_stride;
  GLfloat* ctlarray = (GLfloat*)rhmalloc( stride*cv_count*sizeof(*ctlarray) );
  for ( i = 0; i < cv_count; i++ ) {
    GetGLCV( dim, is_rat, cv + i*cv_stride, xform, ctlarray + stride*i );
  }

  GLint order = nurb_order;

  if ( type == 0 ) {
    switch ( dim ) {
    case 2: // must be a GLU_MAP1_TRIM_2/3
      type = ( is_rat ) 
           ? GLU_MAP1_TRIM_3   // rational 2d trim uses homogeneous coords
           : GLU_MAP1_TRIM_2;  // non-rational 2d trim uses euclidean coords
      break;
    case 3: // must be a GLU_MAP1_VERTEX_3/4
      type = ( is_rat ) 
           ? GL_MAP1_VERTEX_4  // rational 3d curve uses homogeneous coords
           : GL_MAP1_VERTEX_3; // non-rational 3d curve used euclidean coords
      break;
    }
  }

  gluNurbsCurve(
    nobj,
    nknots,
    knot,
    stride, 	
    ctlarray, 	
    order, 	
    type	
  );	

  rhfree( ctlarray );
  rhfree( knot );
}



// See comments in rhinoio_gl.h for calling instructions.

void RhinoGL( const CRhinoNurbsSurface& s,
              GLUnurbsObj* nobj, // created with gluNewNurbsRenderer )
              GLenum type,       // = 0 (and type is automatically set)
              int bPermitKnotScaling,
              double* knot_scale0,
              double* knot_scale1
             )
{
  int i, j, k;

  // The "bPermitScaling" parameters to the RhinoGL() call that
  // fills in the knot vectors is set to FALSE because any
  // rescaling that is applied to a surface domain must also
  // be applied to paramter space trimming curve geometry.

  // GL "s" knots
  GLint sknot_count = s.KnotCount(0) + 2;
  GLfloat* sknot = (GLfloat*)rhmalloc( sknot_count*sizeof(*sknot) );
  RhinoGL( s.Order(0), s.CVCount(0), s.Knot(0), sknot, 
           bPermitKnotScaling, knot_scale0 );

  // GL "t" knots
  GLint tknot_count = s.KnotCount(1) + 2;
  GLfloat* tknot = (GLfloat*)rhmalloc( tknot_count*sizeof(*tknot) );
  RhinoGL( s.Order(1), s.CVCount(1), s.Knot(1), tknot,
           bPermitKnotScaling, knot_scale1 );

  // control vertices
  const int cv_size= s.CVSize();
  const int cv_count[2] = {s.CVCount(0), s.CVCount(1)};
  GLint s_stride = cv_size*cv_count[1];
  GLint t_stride = cv_size;
  GLfloat* ctlarray = (GLfloat*)rhmalloc( s_stride*cv_count[0]*sizeof(*ctlarray) );
  for ( i = 0; i < cv_count[0]; i++ ) {
    for ( j = 0; j < cv_count[1]; j++ ) {
      const double*  cv = s.CV(i,j);
      GLfloat* gl_cv = ctlarray + s_stride*i + t_stride*j;
      for ( k = 0; k < cv_size; k++ ) {
        gl_cv[k] = (GLfloat)cv[k];
      }
    }
  }
  
  GLint sorder = s.Order(0);
  GLint torder = s.Order(1);

  if ( type == 0 ) {
    // set GL surface type for 3d CVs in homogeneous/euclidean form.
    type = ( s.IsRational() ) ? GL_MAP2_VERTEX_4 : GL_MAP2_VERTEX_3;
  }

  gluNurbsSurface (
    nobj,
    sknot_count,
    sknot,
    tknot_count,
    tknot,
    s_stride, 	
    t_stride, 	
    ctlarray, 	
    sorder, 	
    torder, 	
    type	
  );	

  rhfree( ctlarray );
  rhfree( tknot );
  rhfree( sknot );
}

// See comments in rhinoio_gl.h for calling instructions.

void RhinoGL( const CRhinoBrep& b,
              int face_index,       // index of face in B-rep
              GLUnurbsObj* nobj     // created with gluNewNurbsRenderer )
            )
{
  double knot_scale[2][2] = {{0.0,1.0},{0.0,1.0}};

  if ( face_index < 0 || face_index >= (int)b.m_F.Size() )
    return; // bogus face index

  const CRhinoBrepFace& face = b.m_F[face_index];

  // untrimmed surface
  const CRhinoNurbsSurface* srf = b.m_S[face.m_si];
  RhinoGL( *srf,
           nobj, 
           (srf->IsRational()) ? GL_MAP2_VERTEX_4 : GL_MAP2_VERTEX_3,
           TRUE, knot_scale[0], knot_scale[1]
          );

  int fli, li, lti, ti, cni;
  if ( face.m_li.Size() == 1 ) {
    // there is only one trimming loop - if it's trivial, we are done.
    double u[2], v[2], box[2][3];
    u[0] = srf->Knot(0,srf->Order(0)-2);
    u[1] = srf->Knot(0,srf->CVCount(0)-1);
    v[0] = srf->Knot(1,srf->Order(1)-2);
    v[1] = srf->Knot(1,srf->CVCount(1)-1);
    double utol = (u[1]-u[0])*1.0e-6;
    double vtol = (v[1]-v[0])*1.0e-6;
    u[0] += utol;
    u[1] -= utol;
    v[0] += vtol;
    v[1] -= vtol;
    BOOL bTrimmed = FALSE;
    const int face_loop_count = face.m_li.Size();
    for ( fli = 0; !bTrimmed && fli < face_loop_count; fli++ ) {
      li = face.m_li[fli];
      const CRhinoBrepLoop& loop = b.m_L[li];
      const int loop_trim_count = loop.m_ti.Size();
      for ( lti = 0; !bTrimmed && lti < loop_trim_count; lti++ ) {
        ti = loop.m_ti[lti];
        const CRhinoBrepTrim& trim = b.m_T[ti];
        const CRhinoAgCurve* crv2d = b.m_C2[trim.m_c2i];
        for ( cni = 0; !bTrimmed && cni < crv2d->NurbsCurveCount(); cni++ ) {
          CRhinoNurbsCurve* nurbs2d = (*crv2d)[cni];
          nurbs2d->GetBoundingBox( box[0], box[1] );
          if ( box[0][0] >= u[1] )
            continue;
          if ( box[0][1] >= v[1] )
            continue;
          if ( box[1][0] <= u[0] )
            continue;
          if ( box[1][1] <= v[0] )
            continue;
          bTrimmed = TRUE;
        }
      }
    }
    if ( !bTrimmed )
      return;
  }

  // any knot scaling applied to the surface must also be applied to
  // the paramter space trimming geometry
  double xform[4][4] 
    = {{knot_scale[0][1], 0.0, 0.0, -knot_scale[0][0]*knot_scale[0][1] },
       {0.0, knot_scale[1][1], 0.0, -knot_scale[1][0]*knot_scale[1][1] },
       {0.0, 0.0, 1.0, 0.0},
       {0.0, 0.0, 0.0, 1.0}};

  // Add face's 2d trimming loop(s)
  const int face_loop_count = face.m_li.Size();
  for ( fli = 0; fli < face_loop_count; fli++ ) {
    gluBeginTrim( nobj );

    li = face.m_li[fli];
    const CRhinoBrepLoop& loop = b.m_L[li];
    const int loop_trim_count = loop.m_ti.Size();
    for ( lti = 0; lti < loop_trim_count; lti++ ) {
      ti = loop.m_ti[lti];
      const CRhinoBrepTrim& trim = b.m_T[ti];
      const CRhinoAgCurve* crv2d = b.m_C2[trim.m_c2i];
      for ( cni = 0; cni < crv2d->NurbsCurveCount(); cni++ ) {
        CRhinoNurbsCurve* nurbs2d = (*crv2d)[cni];
        RhinoGL( *nurbs2d,
                 nobj, 
                 (nurbs2d->IsRational()) ? GLU_MAP1_TRIM_3 : GLU_MAP1_TRIM_2,
                 TRUE, NULL, xform
                 );
      }
    }

    gluEndTrim( nobj );
  }

}



static double dist_sq( const double* a, const double* b )
{
  double dx = a[0]-b[0];
  double dy = a[1]-b[1];
  double dz = a[2]-b[2];
  return (dx*dx+dy*dy+dz*dz);
}

void RhinoGL( const CRhinoMesh& mesh )
{
  int i0, i1, i2, j0, j1, j2;
  int fi, vi[4];
  double v[4][3], n[4][3], t[4][2];

  const int face_count = mesh.FaceCount();
  const BOOL bHasNormals = mesh.HasNormals();
  const BOOL bHasTCoords = mesh.HasTCoords();

  glBegin(GL_TRIANGLES);
  RhinoGL( mesh.Material() );
  for ( fi = 0; fi < face_count; fi++ ) {

    mesh.GetFace( fi, vi[0], vi[1], vi[2], vi[3] );
    mesh.GetVertex( vi[0], v[0][0], v[0][1], v[0][2] );
    mesh.GetVertex( vi[1], v[1][0], v[1][1], v[1][2] );
    mesh.GetVertex( vi[2], v[2][0], v[2][1], v[2][2] );

    if ( bHasNormals ) {
      mesh.GetNormal( vi[0], n[0][0], n[0][1], n[0][2] );
      mesh.GetNormal( vi[1], n[1][0], n[1][1], n[1][2] );
      mesh.GetNormal( vi[2], n[2][0], n[2][1], n[2][2] );
    }

    if ( bHasTCoords ) {
      mesh.GetTCoord( vi[0], t[0][0], t[0][1] );
      mesh.GetTCoord( vi[1], t[1][0], t[1][1] );
      mesh.GetTCoord( vi[2], t[2][0], t[2][1] );
    }

    if ( vi[2] != vi[3] ) {
      // quadrangle - render as two triangles
      mesh.GetVertex( vi[3], v[3][0], v[3][1], v[3][2] );
      if ( bHasNormals )
        mesh.GetNormal( vi[3], n[3][0], n[3][1], n[3][2] );
      if ( bHasTCoords )
        mesh.GetTCoord( vi[3], t[3][0], t[3][1] );
      if ( dist_sq( v[0], v[2] ) <= dist_sq( v[1], v[3] ) ) {
        i0 = 0; i1 = 1; i2 = 2;
        j0 = 0; j1 = 2; j2 = 3;
      }
      else {
        i0 = 1; i1 = 2; i2 = 3;
        j0 = 1; j1 = 3; j2 = 0;
      }
    }
    else {
      // single triangle
      i0 = 0; i1 = 1; i2 = 2;
      j0 = j1 = j2 = 0;
    }

    // first triangle
    if ( bHasNormals )
      glNormal3d( n[i0][0], n[i0][1], n[i0][2] );
    if ( bHasTCoords )
      glTexCoord2d( t[i0][0], t[i0][1] );
    glVertex3d( v[i0][0], v[i0][1], v[i0][2] );

    if ( bHasNormals )
      glNormal3d( n[i1][0], n[i1][1], n[i1][2] );
    if ( bHasTCoords )
      glTexCoord2d( t[i1][0], t[i1][1] );
    glVertex3d( v[i1][0], v[i1][1], v[i1][2] );

    if ( bHasNormals )
      glNormal3d( n[i2][0], n[i2][1], n[i2][2] );
    if ( bHasTCoords )
      glTexCoord2d( t[i2][0], t[i2][1] );
    glVertex3d( v[i2][0], v[i2][1], v[i2][2] );

    if ( j0 != j1 ) {
      // if we have a quad, second triagle
      if ( bHasNormals )
        glNormal3d( n[j0][0], n[j0][1], n[j0][2] );
      if ( bHasTCoords )
        glTexCoord2d( t[j0][0], t[j0][1] );
      glVertex3d( v[j0][0], v[j0][1], v[j0][2] );

      if ( bHasNormals )
        glNormal3d( n[j1][0], n[j1][1], n[j1][2] );
      if ( bHasTCoords )
        glTexCoord2d( t[j1][0], t[j1][1] );
      glVertex3d( v[j1][0], v[j1][1], v[j1][2] );

      if ( bHasNormals )
        glNormal3d( n[j2][0], n[j2][1], n[j2][2] );
      if ( bHasTCoords )
        glTexCoord2d( t[j2][0], t[j2][1] );
      glVertex3d( v[j2][0], v[j2][1], v[j2][2] );
    }

  }
  glEnd();
}

void RhinoGL( const CRhinoPointSet& set )
{
  int i;
  const double* P;
  glBegin(GL_POINTS);
  for ( i = 0; i < set.PointCount(); i++ ) {
    P = set.Point(i);
    glVertex3d( P[0], P[1], P[2] );
  }
  glEnd();
}

void RhinoGL( const CRhinoMaterial& m )
{
  RhinoGL( &m );
}

void RhinoGL( const CRhinoColor& rc, double alpha, GLfloat c[4] )
{
  c[0] = (GLfloat)rc.FractionRed();
  c[1] = (GLfloat)rc.FractionGreen();
  c[2] = (GLfloat)rc.FractionBlue();
  c[3] = (GLfloat)alpha;
}

void RhinoGL( const CRhinoColor& rc, GLfloat c[4] )
{
  c[0] = (GLfloat)rc.FractionRed();
  c[1] = (GLfloat)rc.FractionGreen();
  c[2] = (GLfloat)rc.FractionBlue();
  c[3] = (GLfloat)1.0;
}


void RhinoGL( const CRhinoMaterial* pMat )
{
  // set GL material to match Rhino material
  if ( !pMat ) {
    CRhinoMaterial default_mat;
    RhinoGL( &default_mat );
  }
  else {
    GLfloat ambient[4], diffuse[4], specular[4], emission[4];
    GLfloat alpha = (GLfloat)(1.0 - pMat->Transparency());
    RhinoGL( pMat->Ambient(), alpha, ambient );
    RhinoGL( pMat->Diffuse(), alpha, diffuse );
    RhinoGL( pMat->Specular(), alpha, specular );
    RhinoGL( pMat->Emission(), alpha, emission );
    GLint shine = (GLint)(128.0*(pMat->Shine() / CRhinoMaterial::MaxShine()));
    if ( shine == 0 ) {
      specular[0]=specular[1]=specular[2]=(GLfloat)0.0;
    }
    glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT,   ambient  );
    glMaterialfv( GL_FRONT_AND_BACK, GL_DIFFUSE,   diffuse  );
    glMaterialfv( GL_FRONT_AND_BACK, GL_SPECULAR,  specular );
    glMaterialfv( GL_FRONT_AND_BACK, GL_EMISSION,  emission );
    glMateriali(  GL_FRONT_AND_BACK, GL_SHININESS, shine );
  }
}

void RhinoGL( const CRhinoSpotlight* spot, GLenum light_index )
{
  if ( spot )
    RhinoGL( *spot, light_index );
  else {
    GLfloat ambient[4]   = { (GLfloat)0.7, (GLfloat)0.7, (GLfloat)0.7, (GLfloat)1.0};
    GLfloat diffuse[4]   = { (GLfloat)0.8, (GLfloat)0.8, (GLfloat)0.8, (GLfloat)1.0 };
    GLfloat specular[4]  = { (GLfloat)1.0, (GLfloat)1.0, (GLfloat)1.0, (GLfloat)1.0 };
    glLightfv( GL_LIGHT0, GL_AMBIENT, ambient );
    glLightfv( GL_LIGHT0, GL_DIFFUSE, diffuse );
    glLightfv( GL_LIGHT0, GL_SPECULAR, specular );
    glEnable( light_index );
  }
}

void RhinoGL( const CRhinoSpotlight& spot, GLenum light_index )
{
  GLfloat ambient[4], diffuse[4], specular[4];
  RhinoGL( spot.Ambient(), ambient );
  RhinoGL( spot.Diffuse(), diffuse );
  RhinoGL( spot.Specular(), specular );
  glLightfv( light_index, GL_AMBIENT, ambient );
  glLightfv( light_index, GL_DIFFUSE, diffuse );
  glLightfv( light_index, GL_SPECULAR, specular );
  
  double d[4];
  spot.GetPosition( d );
  GLfloat position[4] = {(GLfloat)d[0],(GLfloat)d[1],(GLfloat)d[2],(GLfloat)1.0};
  glLightfv( light_index, GL_POSITION, position );

  spot.GetDirection( d );
  GLfloat direction[3];
  direction[0] = (GLfloat)d[0]; 
  direction[1] = (GLfloat)d[1]; 
  direction[2] = (GLfloat)d[2];
  glLightfv( light_index, GL_SPOT_DIRECTION, direction );

  glLightf( light_index, GL_SPOT_EXPONENT, (GLfloat)(spot.SpotExponent()*128.0)  );
  glLightf( light_index, GL_SPOT_CUTOFF, (GLfloat)spot.SpotCutoff() );

  glLightf( light_index, GL_CONSTANT_ATTENUATION,  (GLfloat)1.0  );
  glLightf( light_index, GL_LINEAR_ATTENUATION,    (GLfloat)0.0  );
  glLightf( light_index, GL_QUADRATIC_ATTENUATION, (GLfloat)0.0  );

  glEnable( light_index );
}

void RhinoGL( CRhinoViewport& viewport,
              int port_left, int port_right,
              int port_bottom, int port_top,
              CRhinoXform& projectionMatrix,
              CRhinoXform& modelviewMatrix
              )
{
  double frus_aspect;
  double frus_left, frus_right, frus_bottom, frus_top, frus_near, frus_far;
  double frus_width, frus_height, d;

  const int window_width  = abs(port_right - port_left);
  const int window_height = abs(port_top - port_bottom);

  viewport.GetFrustumAspect( frus_aspect );

  if ( fabs( window_height*frus_aspect - window_width ) > 1 ) {
    // If the window aspect is different from the frustum's aspect, adjust
    // the frustum to get things back to 1:1.
    viewport.GetFrustum( &frus_left, &frus_right, 
                          &frus_bottom, &frus_top, 
                          &frus_near, &frus_far
                          );
    frus_width  = frus_right - frus_left;
    frus_height = frus_top - frus_bottom;
    if ( frus_width < frus_height )
      frus_height = frus_width;
    else
      frus_width = frus_height;
    if ( window_height*frus_width < window_width*frus_height ) {
      d = window_width*frus_height/window_height;
      d = 0.5*(d - (frus_right - frus_left));
      frus_left  -= d;
      frus_right +=  d;
    }
    else {
      d = window_height*frus_width/window_width;
      d = 0.5*(d - (frus_top - frus_bottom));
      frus_bottom -= d;
      frus_top    += d;
    }
    viewport.SetFrustum( frus_left, frus_right, 
                          frus_bottom, frus_top, 
                          frus_near, frus_far 
                          );
  }

  viewport.SetScreenPort( port_left, port_right, port_bottom, port_top,
                          0, 0xff );


  BOOL bHaveCameraToClip = viewport.GetXform( 
                                       CRhinoViewport::camera_cs,  
                                       CRhinoViewport::clip_cs,
                                       projectionMatrix 
                                       );

  if ( bHaveCameraToClip ) {
    projectionMatrix.Transpose();
    glMatrixMode(GL_PROJECTION);
    glLoadMatrixd( &projectionMatrix.m_xform[0][0] );
  }
}

void RhinoGL( CRhinoViewport& viewport,
              int port_left, int port_right,
              int port_bottom, int port_top
              )
{
  // sets model view and projection matrix
  CRhinoXform projectionMatrix;
  CRhinoXform modelviewMatrix;
  RhinoGL( viewport,
           port_left, port_right,
           port_bottom, port_top,
           projectionMatrix, modelviewMatrix 
         );
}

void RhinoGL( const CRhinoViewport& viewport,
              CRhinoXform& modelviewMatrix
              )
{
  // sets model view matrix
  BOOL bHaveWorldToCamera = viewport.GetXform( 
                                       CRhinoViewport::world_cs,  
                                       CRhinoViewport::camera_cs,
                                       modelviewMatrix 
                                       );
  if ( bHaveWorldToCamera ) {
    modelviewMatrix.Transpose();
    glMatrixMode(GL_MODELVIEW);
    glLoadMatrixd( &modelviewMatrix.m_xform[0][0] );
  }
}

void RhinoGL( const CRhinoViewport& viewport )
{
  // sets model view matrix
  CRhinoXform modelviewMatrix;
  RhinoGL( viewport, modelviewMatrix );
}
