/* $Header: /Src/RhinoIO/rhinoio_layer.cpp 7     9/27/99 5:10p Dalelear $ */
/* $NoKeywords: $ */
//
// Copyright (c) 1993-1998 Robert McNeel & Associates. All rights reserved.
// Rhinoceros is a registered trademark of Robert McNeel & Assoicates.
//
// THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
// ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
// MERCHANTABILITY ARE HEREBY DISCLAIMED.
//				
////////////////////////////////////////////////////////////////

#include "rhinoio_def.h"
#include "rhinoio_exception.h"
#include "rhinoio_layerptr.h"
#include "rhinoio_file.h"
#include "rhinoio_chunk.h"
#include "rhinoio_color.h"
#include "rhinoio_layer.h"

////////////////////////////////////////////////////////////////
//   Class CRhinoLayer
////////////////////////////////////////////////////////////////
RHINO_CHUNK_IMPLEMENT(CRhinoLayer, TCODE_LAYER)

CRhinoLayerPtr
CRhinoLayer::New(const char* name, CRhinoColor color, eLayerState state)
{

	CRhinoLayer* pNew = new CRhinoLayer(name, color, state);
	return CRhinoLayerPtr(pNew);
}

CRhinoLayer::CRhinoLayer(const char* name, CRhinoColor color, eLayerState state)
            : CRhinoChunk(TCODE_LAYER), 
              m_Color(color), 
              m_state(state)
{

	if(name){
		size_t len = strlen(name);
		m_pName = new char[len+1];
		strcpy(m_pName, name);
	} else
		m_pName=NULL;

	m_RefCount=0;
}

/*
CRhinoLayer::CRhinoLayer(const CRhinoHeader& head) : CRhinoChunk(head) {
	m_pName=NULL;
	m_Color = CRhinoColor();
	m_State = normal;
	m_RefCount=0;
  }
*/
	
CRhinoLayer::CRhinoLayer() : CRhinoChunk(TCODE_LAYER) {
	m_pName    = NULL;
	m_Color    = CRhinoColor();
	m_state    = normal;
	m_RefCount = 0;
}


CRhinoLayer::~CRhinoLayer(){
	//assert(m_RefCount==0);			// Should never delete if there are references
	delete[] m_pName;
}

void
CRhinoLayer::SetName(const char* str)
{

	delete[] m_pName;
	if(str){
		size_t len = strlen(str);
		m_pName = new char[len+1];
		strcpy(m_pName, str);
	} else
		m_pName=NULL;
	return;
}

CRhinoFile::eStatus  
CRhinoLayer::Read(CRhinoFile& file)
{
	CRhinoFile::eStatus status = CRhinoFile::good;

	SetState( normal );

	for ( ;; ) {

    unsigned int typecode = 0;
    int info = 0;

	  if( file.BeginReadChunk(typecode, info) != CRhinoFile::good ) {
		  throw CRhinoException(CRhinoException::unexpected_end_of_file);
    }

		switch( typecode ) {

		case TCODE_LAYERNAME:
			if(m_pName) delete[] m_pName;
			m_pName=NULL;
			file.ReadString(m_pName);
			break;

		case TCODE_RGB:
			m_Color = info;
			break;

		case TCODE_LAYERSTATE:
      switch ( info ) {
      case 0: // on
        SetState( normal );
        break;
      case 1: // off
        SetState( hidden );
        break;
      case 2: // locked
        SetState( locked );
        break;
      }
			break;		

		default: // Skip any unknown sub-chunks
			file.SkipChunk();
      break;
		}
		
		if ( file.EndChunk() == CRhinoFile::crc_error )
			status = CRhinoFile::crc_error;

    if ( typecode == TCODE_ENDOFTABLE ) {
      break;
    }

	}


	return status;
}

void 
CRhinoLayer::Write(CRhinoFile& file) const
{
  file.BeginWriteChunk( TCODE_LAYER );
  {
    file.BeginWriteChunk( TCODE_LAYERNAME );
    file.WriteString(Name());
    file.EndChunk();

    file.BeginWriteChunk( TCODE_RGB, Color());
    file.EndChunk();

    int state = 0;  // 0 = on, 1 = off, 2 = locked
    switch ( State() ) {
    case normal:
      state = 0;
      break;
    case hidden:
      state = 1;
      break;
    case locked:
      state = 2;
      break;
    case ghosted:
      state = 0; // currently Rhino does not supported a ghosted layer state
      break;
    }
    file.BeginWriteChunk( TCODE_LAYERSTATE, state );
    file.EndChunk();

    // These 4 values are obsolete.
    // file.BeginWriteChunk( TCODE_LAYERVISIBLE, cntl.m_Visible );
    // file.EndChunk();
    // file.BeginWriteChunk( TCODE_LAYERPICKABLE, cntl.m_Pickable );
    // file.EndChunk();
    // file.BeginWriteChunk( TCODE_LAYERSNAPABLE, cntl.m_Snapable );
    // file.EndChunk();
    // file.BeginWriteChunk( TCODE_LAYERRENDERABLE, cntl.m_Renderable );
    // file.EndChunk();

    file.BeginWriteChunk( TCODE_ENDOFTABLE );
    file.EndChunk();
  }
  file.EndChunk();	// end TCODE_LAYER

  return;
}


const char* CRhinoLayer::Name( void )  const	
{
  return (const char*) m_pName;
}

CRhinoColor CRhinoLayer::Color( void ) const	
{
  return m_Color;
}

CRhinoLayer::eLayerState CRhinoLayer::State( void ) const
{
  return m_state;
}

void CRhinoLayer::SetColor( CRhinoColor c )
{
  m_Color = c;
} 

void CRhinoLayer::SetState( CRhinoLayer::eLayerState s )
{
  m_state = s;
} 



//
//  Two layers are == if there name is the same
//
BOOL
CRhinoLayer::operator==(const CRhinoLayer& rhs) const {
	if( strcmp( Name(), rhs.Name()) ) return FALSE;
	return TRUE;
}


///////////////////////////////////////////////////////////////////////////////
// Class CRhinoLayerPtr
///////////////////////////////////////////////////////////////////////////////

// Copy Constuctor 
CRhinoLayerPtr::CRhinoLayerPtr( const CRhinoLayerPtr& pnt){
	m_pPointer = pnt.m_pPointer;	
	if(m_pPointer) m_pPointer->m_RefCount++;
}


// Construct from a CRhinoLayer*
CRhinoLayerPtr::CRhinoLayerPtr(CRhinoLayer* play){
	m_pPointer = play;
	if(play) play->m_RefCount++;
	}


// assignment operator
CRhinoLayerPtr& CRhinoLayerPtr::operator=(const CRhinoLayerPtr& rhs){
	if(rhs.m_pPointer) rhs.m_pPointer->m_RefCount++;
	if(m_pPointer)
		if(--m_pPointer->m_RefCount ==0) delete m_pPointer;
	m_pPointer = rhs.m_pPointer;
	return *this;
}


CRhinoLayerPtr::~CRhinoLayerPtr(){
	if(m_pPointer) 
		if(--m_pPointer->m_RefCount == 0) delete m_pPointer;
}
