/* $Header: /Src/RhinoIO/rhinoio_material.cpp 9     1/12/00 5:57p Dalelear $ */
/* $NoKeywords: $ */
//
// Copyright (c) 1993-1998 Robert McNeel & Associates. All rights reserved.
// Rhinoceros is a registered trademark of Robert McNeel & Assoicates.
//
// THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
// ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
// MERCHANTABILITY ARE HEREBY DISCLAIMED.
//				
////////////////////////////////////////////////////////////////

#include "rhinoio_def.h"
#include "rhinoio_exception.h"
#include "rhinoio_layerptr.h"
#include "rhinoio_file.h"
#include "rhinoio_chunk.h"
#include "rhinoio_color.h"
#include "rhinoio_material.h"

////////////////////////////////////////////////////////////////
//   Class CRhinoMaterial
////////////////////////////////////////////////////////////////

double CRhinoMaterial::m_max_shine = 255.0f;

double CRhinoMaterial::MaxShine()
{
  return m_max_shine;
}

// Default constructor
CRhinoMaterial::CRhinoMaterial() 
               : m_ambient( 128, 128, 128 ),
                 m_diffuse( 128, 128, 128 ),
                 m_emission( 0, 0, 0 ),
                 m_specular( 255, 255, 255 ),
                 m_shine(0.0),
                 m_transparency(0.0),
                 m_sTextureBitmapFileName(0),
                 m_sBumpBitmapFileName(0),
                 m_sRenderMaterialName(0)
{}

// Copy Constructor
CRhinoMaterial::CRhinoMaterial(const CRhinoMaterial& src)
               : m_ambient( 128, 128, 128 ),
                 m_diffuse( 128, 128, 128 ),
                 m_emission( 0, 0, 0 ),
                 m_specular( 255, 255, 255 ),
                 m_shine(0.0),
                 m_transparency(0.0),
                 m_sTextureBitmapFileName(0),
                 m_sBumpBitmapFileName(0),
                 m_sRenderMaterialName(0)
{
  *this = src;
}

// operator =
CRhinoMaterial& CRhinoMaterial::operator=(const CRhinoMaterial& src)
{
  if ( this != &src ) {
	  SetAmbient ( src.Ambient() );
	  SetDiffuse ( src.Diffuse() );
	  SetEmission ( src.Emission() );
	  SetSpecular ( src.Specular() );
	  SetShine ( src.Shine() );
	  SetTransparency ( src.Transparency() );

	  SetTextureBitmapFileName(  src.TextureBitmapFileName()  );
	  SetBumpBitmapFileName(  src.BumpBitmapFileName()  );
	  SetRenderMaterialName(  src.RenderMaterialName()  );
  }
	return *this;
}


CRhinoMaterial::~CRhinoMaterial()
{
  delete [] m_sTextureBitmapFileName;
  m_sTextureBitmapFileName = 0;
  delete [] m_sBumpBitmapFileName;
  m_sBumpBitmapFileName = 0;
  delete [] m_sRenderMaterialName;
  m_sRenderMaterialName = 0;
}


// Shine values are in range 0.0 to CRhinoMaterial::GetMaxShine()
double CRhinoMaterial::Shine() const
{
  return m_shine;
}

void CRhinoMaterial::SetShine( double shine )
{
  if ( shine < 0.0 )
    m_shine = 0.0;
  else if ( shine > m_max_shine)
    m_shine = m_max_shine;
  else
    m_shine = (float)shine;
}

  // Transparency values are in range 0.0 = opaque to 1.0 = transparent
double CRhinoMaterial::Transparency( ) const
{
  return  m_transparency;
}

void CRhinoMaterial::SetTransparency( double transparency )
{
  if ( transparency < 0.0 )
    m_transparency = 0.0f;
  else if ( transparency > 1.0)
    m_transparency = 1.0f;
  else
    m_transparency = (float)transparency;
}


CRhinoColor CRhinoMaterial::Ambient() const
{
  return m_ambient;
}

CRhinoColor CRhinoMaterial::Diffuse( ) const
{
  return m_diffuse;
}

CRhinoColor CRhinoMaterial::Emission( ) const
{
  return m_emission;
}

CRhinoColor CRhinoMaterial::Specular() const
{
  return m_specular;
}

void CRhinoMaterial::SetAmbient( CRhinoColor  c )
{
  m_ambient = c;
}

void CRhinoMaterial::SetDiffuse(  CRhinoColor c )
{
  m_diffuse = c ;
}

void CRhinoMaterial::SetEmission( CRhinoColor c )
{
  m_emission = c ;
}

void CRhinoMaterial::SetSpecular( CRhinoColor c )
{
  m_specular = c;
}


const char* CRhinoMaterial::TextureBitmapFileName( ) const
{
	return (const char*) m_sTextureBitmapFileName;
}

void CRhinoMaterial::SetTextureBitmapFileName( const char* sFileName )
{
	delete[] m_sTextureBitmapFileName;
  m_sTextureBitmapFileName = 0;
  int len = sFileName ? strlen(sFileName) : 0;
	if(len > 0){
		m_sTextureBitmapFileName = new char[len+1];
		strcpy(m_sTextureBitmapFileName, sFileName);
	}
	return;
}


const char* CRhinoMaterial::BumpBitmapFileName( ) const
{
	return (const char*) m_sBumpBitmapFileName;
}

void CRhinoMaterial::SetBumpBitmapFileName( const char* sFileName )
{
	delete[] m_sBumpBitmapFileName;
  m_sBumpBitmapFileName = 0;
  int len = sFileName ? strlen(sFileName) : 0;
	if(len > 0){
		m_sBumpBitmapFileName = new char[len+1];
		strcpy(m_sBumpBitmapFileName, sFileName);
	}
	return;
}

const char* CRhinoMaterial::RenderMaterialName( ) const
{
	return (const char*) m_sRenderMaterialName;
}

void CRhinoMaterial::SetRenderMaterialName( const char* name )
{
	delete[] m_sRenderMaterialName;
  m_sRenderMaterialName = 0;
  int len = (name) ? strlen(name) : 0;
	if (len > 0) {
		m_sRenderMaterialName = new char[len+1];
		strcpy(m_sRenderMaterialName, name);
	}
	return;
}


//////////////////////////////////////////////////////////////////////////////////
// Class CRhinoRGB
//////////////////////////////////////////////////////////////////////////////////
RHINO_CHUNK_IMPLEMENT(CRhinoRGB, TCODE_RGB)

void CRhinoRGB::SetColor( CRhinoColor c )
{
	SetValue( (Value() & 0xFF000000) | (c & 0x00FFFFFF) );
}

void CRhinoRGB::SetShine( int shine )
{
	if ( shine < 0 ) shine = 0; else if (shine > 255) shine = 255;
	SetValue(  (shine<<24) | (Value() & 0x00FFFFFF) );
}

//////////////////////////////////////////////////////////////////////////////////
// Class CRhinoTransparency
//////////////////////////////////////////////////////////////////////////////////
RHINO_CHUNK_IMPLEMENT(CRhinoTransparency, TCODE_TRANSPARENCY)

CRhinoTransparency::CRhinoTransparency(int transparency)
                   : CRhinoChunk(TCODE_TRANSPARENCY)
{
  if ( transparency < 0 ) 
    transparency = 0;
  else if ( transparency > 255 )
    transparency = 255;
	SetValue(transparency);
}

CRhinoTransparency::CRhinoTransparency(double transparency)
                   : CRhinoChunk(TCODE_TRANSPARENCY)
{
  if ( transparency < 0.0 ) 
    transparency = 0.0;
  else if ( transparency > 1.0 )
    transparency = 1.0;
	SetValue((int)(transparency*255));
}

//////////////////////////////////////////////////////////////////////////////////
// Class CRhinoRenderMaterialName
//////////////////////////////////////////////////////////////////////////////////
RHINO_CHUNK_IMPLEMENT(CRhinoRenderMaterialName, TCODE_RENDER_MATERIAL_ID)

CRhinoRenderMaterialName::CRhinoRenderMaterialName(int): CRhinoChunk(TCODE_RENDER_MATERIAL_ID) 
{
	m_Name=NULL;
}

CRhinoRenderMaterialName::~CRhinoRenderMaterialName() 
{
	delete[] m_Name;
}
	
void CRhinoRenderMaterialName::SetName(const char* name)
{
	delete[] m_Name;
  m_Name = 0;

  int len = (name) ? strlen(name) : 0;
  if ( len > 0 ) {
	  m_Name = new char[len+1];
    strcpy(m_Name, name);
  }
	return;
}

CRhinoFile::eStatus 
CRhinoRenderMaterialName::Read(CRhinoFile& file)
{
  int type = -1;
  file.ReadInteger( &type );
  switch ( type ) {
  case 0: // no render material present
    break;
  case 1: // ASCII render material name
    {
      char* name = NULL;
      file.ReadString(name);
      SetName(name);
      delete[] name;
    }
    break;
  default:
    file.SkipChunk(); // skip unknown information
    break;
  }
  return CRhinoFile::good;
}

void 
CRhinoRenderMaterialName::Write(CRhinoFile &file) const 
{
	file.BeginWriteChunk( *this );
  const char* s = Name();
  if ( s && s[0] ) {
    file.WriteInteger( 1 ); // 1 = ASCII render material name
	  file.WriteString( Name() );
  }
  else {
    file.WriteInteger( 0 ); // 0 = no render material name present
  }
	file.EndChunk();
	return;
}

const char*  CRhinoRenderMaterialName::Name( void ) const
{
  return (const char*) m_Name; 
}

//////////////////////////////////////////////////////////////////////////////////
// Class CRhinoTextureMap
//////////////////////////////////////////////////////////////////////////////////
RHINO_CHUNK_IMPLEMENT(CRhinoTextureMap, TCODE_TEXTUREMAP)

CRhinoTextureMap::CRhinoTextureMap(int): CRhinoChunk(TCODE_TEXTUREMAP) 
{
	m_Name=NULL;
}

CRhinoTextureMap::~CRhinoTextureMap() 
{
	delete[] m_Name;
}
	
void CRhinoTextureMap::SetFileName(const char* name)
{
	delete[] m_Name;
  m_Name = 0;
  
  int len = name ? strlen(name) : 0;
  if ( len > 0 ) {
	  m_Name = new char[len+1];
	  strcpy(m_Name, name);
  }
	return;
}

CRhinoFile::eStatus 
CRhinoTextureMap::Read(CRhinoFile& file){
	char* name = NULL;
	file.ReadString(name);
	SetFileName(name);
	delete[] name;
	return CRhinoFile::good;
}


void 
CRhinoTextureMap::Write(CRhinoFile &file) const 
{
	file.BeginWriteChunk( *this );
	file.WriteString( FileName() );
	file.EndChunk();
	return;
}

const char*  CRhinoTextureMap::FileName( void ) const
{
  return (const char*) m_Name; 
}

//////////////////////////////////////////////////////////////////////////////////
// Class CRhinoBumpMap
//////////////////////////////////////////////////////////////////////////////////
RHINO_CHUNK_IMPLEMENT(CRhinoBumpMap, TCODE_BUMPMAP)

CRhinoBumpMap::CRhinoBumpMap(): CRhinoChunk(TCODE_BUMPMAP)
{
	m_Name=NULL;
}

CRhinoBumpMap::~CRhinoBumpMap() 
{
	delete[] m_Name;
}
	
void CRhinoBumpMap::SetFileName( const char* sFileName )
{
	delete[] m_Name;
  m_Name = 0;

  int len = (sFileName) ? strlen(sFileName) : 0;
  if ( len > 0 ) {
	  m_Name = new char[len+1];
    m_Name[len] = 0;
    if ( len > 0 )
	    strcpy(m_Name, sFileName);
  }
}

CRhinoFile::eStatus 
CRhinoBumpMap::Read(CRhinoFile& file)
{
	char* name = NULL;
	file.ReadString(name);
	SetFileName(name);
	delete[] name;
	return CRhinoFile::good;
}


void 
CRhinoBumpMap::Write(CRhinoFile &file) const 
{
	file.BeginWriteChunk( *this );
	file.WriteString( FileName() );
	file.EndChunk();
}

const char* CRhinoBumpMap::FileName( void ) const
{
  return (const char*) m_Name; 
}
