/* $Header: /Src/RhinoIO/rhinoio_material.h 7     1/12/00 5:57p Dalelear $ */
/* $NoKeywords: $ */
//
// Copyright (c) 1993-1998 Robert McNeel & Associates. All rights reserved.
// Rhinoceros is a registered trademark of Robert McNeel & Assoicates.
//
// THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
// ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
// MERCHANTABILITY ARE HEREBY DISCLAIMED.
//				
////////////////////////////////////////////////////////////////
//
//   defines CRhinoMaterial, 
//           CRhinoRGB, CRhinoTransparency, CRhinoTextureMap, 
//           CRhinoBumpMap, CRhinoRenderMaterialName
//
////////////////////////////////////////////////////////////////

#if !defined(RHINO_MATERIAL_INC_)
#define RHINO_MATERIAL_INC_

///////////////////////////////////////////////////////////////////////////////
// Class CRhinoMaterial
//  This class defines a material in Rhino.  Warning: The texture and bumpmap
//  interface may change in the future.
// 
///////////////////////////////////////////////////////////////////////////////
class CRhinoMaterial
{
public:
  static double MaxShine();			// maximum value of shine exponent

  CRhinoMaterial();							// Default grey color
  CRhinoMaterial(const CRhinoMaterial&);    // copy constructor
  ~CRhinoMaterial();						// destructor

  CRhinoMaterial& operator=(const CRhinoMaterial&);	// assignment operator

  CRhinoColor Ambient() const;
  CRhinoColor Diffuse() const;
  CRhinoColor Emission() const;
  CRhinoColor Specular() const;

  // Shine values are in range 0.0 to CRhinoMaterial::GetMaxShine()
  double Shine() const;

  // Transparency values are in range 0.0 = opaque to 1.0 = transparent
  double Transparency() const;
  
  // Texture bitmap file names
  const char* TextureBitmapFileName() const;

  // Bump bitmap file names
  const char* BumpBitmapFileName() const;

  // Render Material Name
  const char* RenderMaterialName() const;
  // Texture and bump bitmaps can be Windows bitmap (.BMP), Targa (.TGA),
  // JPEG (.JPG), PCX or PNG files.  Version 1 of Rhino will not support
  // filenames using unicode or multibyte character sets.  As soon as
  // Rhino supports these character sets, the const char* filename will
  // changed to const _TCHAR*.

  void SetAmbient(  CRhinoColor );
  void SetDiffuse(  CRhinoColor );
  void SetEmission( CRhinoColor );
  void SetSpecular( CRhinoColor );
  void SetShine( double );         // 0 to CRhinoMaterial::MaxShine()
  void SetTransparency( double );  // 0 = opaque, 1 = transparent
  void SetTextureBitmapFileName( const char* ); // ASCII bitmap filename
  void SetBumpBitmapFileName( const char* );    // ASCII bitmap filename
  void SetRenderMaterialName( const char* );    // render material used in RIB, POV, OBJ, ...

private:
  static double m_max_shine;

  CRhinoColor   m_ambient;
  CRhinoColor   m_diffuse;
  CRhinoColor   m_emission;
  CRhinoColor   m_specular;
  double        m_shine;        // 0.0 = none to GetMaxShine()=maximum
  double        m_transparency; // 0.0 = opaque to 1.0 = transparent
  char*         m_sTextureBitmapFileName;
  char*         m_sBumpBitmapFileName;
  char*         m_sRenderMaterialName;
};

///////////////////////////////////////////////////////////////////////////////
// Class CRhinoRGB
//   This chunk stores a chunk stores CRhinoColor and a shine value.
//   It is used for materials and layers. 
///////////////////////////////////////////////////////////////////////////////
class CRhinoRGB: public CRhinoChunk 
{
	RHINO_CHUNK_DECLARE( CRhinoRGB )
public:
  // construction
	CRhinoRGB( 
    CRhinoColor, 
    int = 0          // shine 0 to 255
    );
	CRhinoRGB( 
    int = 0 // ((shine&255)<<24) | ((r&255)<<16) | ((g&255) << 8) | (b&255)
    );
  ~CRhinoRGB();

	CRhinoColor Color( void ) const;
	int Shine( void ) const; // 0 to 255

	void SetColor( CRhinoColor );
	void SetShine( 
    int // 0 to 255
    );
};

///////////////////////////////////////////////////////////////////////////////

inline CRhinoColor CRhinoRGB::Color( void ) const
  {return CRhinoColor( Value() );}

inline int CRhinoRGB::Shine( void ) const
  {return (Value()>>24);}

inline CRhinoRGB::CRhinoRGB( CRhinoColor c, int shine )
                 : CRhinoChunk(TCODE_RGB)
  {SetColor(c);	SetShine(shine);}

inline CRhinoRGB::CRhinoRGB( int color_ref )
                 : CRhinoChunk(TCODE_RGB, color_ref)
  {}

inline CRhinoRGB::~CRhinoRGB()
  {}

///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// Class CRhinoTransparency ( Value() 0 = opaque, 255 = 
//   This chunk stores a transparency value.  Used in materials.
///////////////////////////////////////////////////////////////////////////////
class CRhinoTransparency: public CRhinoChunk 
{
	RHINO_CHUNK_DECLARE( CRhinoTransparency )
public:
	CRhinoTransparency( int = 0 ); // 0 = opaque, 255 = transparent
	CRhinoTransparency( double );  // 0.0 = opaque, 1.0 = transparent
	~CRhinoTransparency();
};

inline CRhinoTransparency::~CRhinoTransparency()
  {}

///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// Class CRhinoRenderMaterialName
//   This chunk stores the render material name used with RIB, POV, OBJ, ...
///////////////////////////////////////////////////////////////////////////////
class CRhinoRenderMaterialName: public CRhinoChunk 
{
	RHINO_CHUNK_DECLARE( CRhinoRenderMaterialName )
public:
	CRhinoRenderMaterialName( int = 0 );
	~CRhinoRenderMaterialName();
		
	const char* Name( void ) const;	// Filename of bitmap 
	void SetName( const char* );	

	// Overeride Read and write defualts
	void Write(CRhinoFile &) const;		// write chunk to file. 
	CRhinoFile::eStatus Read(CRhinoFile&);							

private:
	char* m_Name;
};

///////////////////////////////////////////////////////////////////////////////
// Class CRhinoTextureMap
//   This chunk stores a texture bitmap filename.  Used in materials.
///////////////////////////////////////////////////////////////////////////////
class CRhinoTextureMap: public CRhinoChunk 
{
	RHINO_CHUNK_DECLARE( CRhinoTextureMap )
public:
	CRhinoTextureMap( int = 0 );
	~CRhinoTextureMap();
		
	const char*  FileName( void ) const;	// Filename of bitmap 
	void SetFileName( const char* );	

	// Overeride Read and write defualts
	void Write(CRhinoFile &) const;		// write chunk to file. 
	CRhinoFile::eStatus Read(CRhinoFile&);							

private:
	char* m_Name;
};

///////////////////////////////////////////////////////////////////////////////
// Class CRhinoBumpMap
//   This chunk stores a bump bitmap filename.  Used in materials.
///////////////////////////////////////////////////////////////////////////////

class CRhinoBumpMap: public CRhinoChunk 
{
	RHINO_CHUNK_DECLARE( CRhinoBumpMap )
public:

  // construction
	CRhinoBumpMap();
	~CRhinoBumpMap();
		
	// virtual overrides
	void Write(CRhinoFile &) const;		// write chunk to file. 
	CRhinoFile::eStatus Read(CRhinoFile&);							

  // interface
  const char* FileName( void ) const;	// Filename of bitmap 
	void SetFileName( const char* );	

private:
	char* m_Name;
};

#endif
