/* $Header: /Src/RhinoIO/rhinoio_mesh.cpp 7     10/28/98 12:18p Dalelear $ */
/* $NoKeywords: $ */
//
// Copyright (c) 1993-1998 Robert McNeel & Associates. All rights reserved.
// Rhinoceros is a registered trademark of Robert McNeel & Assoicates.
//
// THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
// ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
// MERCHANTABILITY ARE HEREBY DISCLAIMED.
//				
////////////////////////////////////////////////////////////////

#include "rhinoio_object.h"
#include "rhinoio_mesh.h"


RHINO_CHUNK_IMPLEMENT(CRhinoMeshGeometry, TCODE_COMPRESSED_MESH_GEOMETRY)
RHINO_CHUNK_IMPLEMENT(CRhinoMesh, TCODE_MESH_OBJECT)

// Create an empty mesh with no normal or texture coordinates.
CRhinoMesh::CRhinoMesh(): CRhinoObject(TCODE_MESH_OBJECT), 
													m_V(m_pV), 
                          m_F(m_pF), 
                          m_N(m_pN), 
                          m_TC(m_pTC) 
{
	Initialize();
}

// Initialize CRhinoMesh data members fo an empty mesh
void
CRhinoMesh::Initialize(){
	m_pGeo = new CRhinoMeshGeometry;         // empty mesh geometry
	m_TDomain[0][0]= m_TDomain[0][1] = 0;
	m_TDomain[1][0]= m_TDomain[1][1] = 0;
	m_HasTCoords = FALSE;
	m_ScaledTDomain = FALSE;

	SynchPointers();
}

void 
CRhinoMesh::SynchPointers(){			// sync up these copies with those in *m_pGeo
	m_pV = m_pGeo->m_V;
	m_pF = m_pGeo->m_F;
	m_pN = m_pGeo->m_N;
	if(!HasTCoords() || !m_ScaledTDomain)
		m_pTC = m_pGeo->m_TC;
	return;
}

CRhinoMesh::~CRhinoMesh(){
	delete m_pGeo;
	if(HasTCoords() && m_ScaledTDomain)
		delete[] m_pTC;
}


// Computes the bounding box of the mesh from the vertices.  
// If no vertices return low = high =(0,0,0).
void 
CRhinoMesh::GetBoundingBox(CRhino3DPoint& low, CRhino3DPoint& high ) const{
	if(VertexCount()==0){
		low[0]=low[1]=low[2]=0.0;
		high[0]=high[1]=high[2]=0.0;
		return;
	}
	
	int i,j;
	for(j=0; j<3; j++)
		high[j] = low[j] = m_V[0][j];

	for(i=1; i<VertexCount(); i++)
		for(j=0; j<3; j++) {
			if( m_V[i][j] < low[j]) low[j] = m_V[i][j];
			else if( m_V[i][j] > high[j]) high[j] = m_V[i][j];
	}
	return;
}

//////////////////////////////////////////////////////////////////////////// 
//       CRhinoMesh::Read(....)
//
//  Read TCODE_MESH_OBJECT chunk data and initialize a CRhinoMesh object.
//  This chunk contains a sequence of subchunks.  The mesh geometry is 
//  stored in a TCODE_COMPRESSED_MESH_GEOMETRY sub-chunk.  Other chunks
//  define the layer and material properties.
//
//  Function can return 'crc_error' if a sub-chunk has a CRC error, otherwise
//  it normally will return 'good'.
//////////////////////////////////////////////////////////////////////////// 
CRhinoFile::eStatus
CRhinoMesh::Read( CRhinoFile& file)
{
  CRhinoFile::eStatus status = CRhinoFile::good;

  {
    CRhinoChunk* pChunk = file.ReadChunk(status);
    if ( status == CRhinoFile::good && pChunk->TypeCode() != TCODE_COMPRESSED_MESH_GEOMETRY ) {
      status = CRhinoFile::crc_error;
      delete pChunk;
    }
    if ( status != CRhinoFile::good )
      return status;

    if ( m_pGeo )
      delete m_pGeo;
		m_pGeo = (CRhinoMeshGeometry*)pChunk;
		pChunk = NULL;

		// The Texture domain is the unit square.
		if(m_pGeo->TCoord){
			m_HasTCoords=TRUE;
			m_ScaledTDomain=FALSE;
			m_TDomain[0][0]=m_TDomain[0][1]=0;
			m_TDomain[1][0]=m_TDomain[1][1]=1;
		}
		else{
			m_HasTCoords=FALSE;
			m_ScaledTDomain=FALSE;;
			m_TDomain[0][0]=m_TDomain[0][1]=0;
			m_TDomain[1][0]=m_TDomain[1][1]=0;
		}
    SynchPointers();
  }
  
  CRhinoFile::eStatus att_status = ReadAttributes( file, TCODE_ENDOFTABLE );
  if ( status == CRhinoFile::good )
    status = att_status;

	return status;
}

// normalize to a unit vector. If 0 then leave alone.
static void Normalize(CRhino3DPoint& V)
{
	double mag=0;
	int j=0;
	for(j=0; j<3; j++)
		mag += V[j] * V[j];
	mag = sqrt(mag);
	if(mag)
		for(j=0;j<3;j++)
			V[j] /= mag;
		return;
	
}



//////////////////////////////////////////////////////////////////////////
//		CRhinoMeshGeometry::Read(...) 
//
//	Read the data field of the TCODE_COMPRESSED_MESH_GEOMETRY chunk.
//	This chunk contains all the mesh geometry.  It has no sub-chunks.
//	The data for vertices, faces, normals and texture coordinates 
//  are compressed for space efficiency.  Each section of
//  code below shows the specific compression scheme used.
//  Always returns status of good.
////////////////////////////////////////////////////////////////////////// 
CRhinoFile::eStatus
CRhinoMeshGeometry::Read( CRhinoFile& file )
{ 
	
	file.ReadInteger(1,&nVerts);		// read number of vertices
	file.ReadInteger(1,&nFaces);		// read number of faces
	file.ReadInteger(1,&Normals);		// read BOOL  Normals indicator
	file.ReadInteger(1,&TCoord);		// read BOOL  TCoord indicator

	// allocate the arrays
	m_V = new CRhino3DPoint[nVerts];
	m_F = new int[nFaces][4];
	if(Normals)
		m_N = new CRhino3DPoint[nVerts];
	if(TCoord)
		m_TC = new CRhino2DPoint[nVerts];

	// Read the corners of the mesh bounding box. 
	CRhino3DPoint BoundBox[2];					// read coordinates of low
	file.ReadDouble(6,(double*)(BoundBox));  	//    and high corners 

	
	// use CRhinoWorkSpace for temporary memory
	CRhinoWorkSpace work;	
	unsigned short (*V)[3];
	int size = nVerts*sizeof(V[0]);
	V = (unsigned short (*)[3]) work.GetMemory(size);

	
	// Read compressed vertex coordinates stored as 3 shorts per vertex.
	// The vertex positions are stored in a coordinate system oriented
	// along the bounding box.  The lower bounding box corner has coordinate
	// (0,0,0) and the higher corner has coordinate (65535, 65535, 65535).
	
	file.ReadShort(3*nVerts, (short*) V); 
	double invscale[3];		// scale factor in each coordinate direction
	int i,j;
	for( j=0; j<3;j++){
		double diff=(BoundBox[1][j] - BoundBox[0][j]);
		invscale[j]= diff/65535;				;
	};
	for(i=0;i<nVerts;i++){
		for( j=0; j<3;j++)
			m_V[i][j] = V[i][j]*invscale[j]+BoundBox[0][j]; 
	}

	// Read in face definitions.  Each face is a quadruple of vertex indices. 
	// If there are less than 65536 vertices the indicies are stored as
	// unsigned shorts otherwise as unsigned ints.
	if(nVerts<65536){
		unsigned short (*F)[4];
		size = nFaces*sizeof(F[0]);
		F = (unsigned short (*)[4]) work.GetMemory(size);
		file.ReadShort( 4*nFaces, (short*) F);
		for ( i = 0; i < nFaces; i++ ) {
			for ( j = 0; j < 4; j++ ) {
				m_F[i][j] = F[i][j];		// convert to ints
			}
		}
	}
	else
		file.ReadInteger(4*nFaces, (int*) m_F);
	

	// Read in normals.  Unit normals are scaled by 127 and stored as 
	// signed chars. 
	if(Normals){
		signed char (*N)[3];
		size = nVerts*sizeof(N[0]);
		N = (signed char (*)[3]) work.GetMemory(size);
		file.ReadByte(size,N);	
		for ( i = 0; i < nVerts; i++ ) {
			for(j=0; j<3; j++)					// decompress
				m_N[i][j] = N[i][j]/127.0;

			Normalize(m_N[i]);

		}						
	}

	// Read texture coordinates.  The texture coordinates are scaled by
	// 65535 and stored as unsigned shorts (2 bytes).
	if(TCoord){
		unsigned short (*T)[2];
		size = nVerts * sizeof(T[0]);
		T = (unsigned short (*)[2]) work.GetMemory(size);
		file.ReadShort(2*nVerts, (short*) T );
		// uncompress texture coordinates
		for ( i = 0; i < nVerts; i++ ) {
			m_TC[i][0] = T[i][0]/65535.0;
			m_TC[i][1] = T[i][1]/65535.0;
		}						
	}
	return CRhinoFile::good;
}


//////////////////////////////////////////////////////////////////////////// 
//       CRhinoMesh::Write(....)
//
//  Write a CRhinoMesh object out to a Rhino 3dm file as a
//  TCODE_MESH_OBJECT chunk.
// 
//  This chunk contains a sequence of subchunks.  The mesh geometry is 
//  stored in a TCODE_COMPRESSED_MESH_GEOMETRY sub-chunk.  Other chunks
//  define the layer and material properties.
//////////////////////////////////////////////////////////////////////////// 
void
CRhinoMesh::Write(CRhinoFile& file) const 
{
	// Make sure layer definition is in the file
	file.LayerIndex( Layer() );	// Get layer index or -1
  
  file.BeginWriteChunk( *this );
	
	if(HasTCoords() && m_ScaledTDomain)
  {		// scale TCoords and store in m_pGeo
		m_pGeo->TCoord = TRUE;
		m_pGeo->m_TC = new CRhino2DPoint[VertexCount()];
	
		double scale[2];
		scale[0]= m_TDomain[1][0] - m_TDomain[0][0];	// extent to Texture Domain
		scale[1]= m_TDomain[1][1] - m_TDomain[0][1];

		for (int i = 0; i < VertexCount(); i++ ) {
			// clamp coordinates into the domain
			double u = m_TC[i][0];
			double v = m_TC[i][1];
			if(u<m_TDomain[0][0]) u = m_TDomain[0][0];
			if(u>m_TDomain[1][0]) u = m_TDomain[1][0];
			if(v<m_TDomain[0][1]) v = m_TDomain[0][1];
			if(v>m_TDomain[1][1]) v = m_TDomain[1][1];
			// Transform to the unit square
			if(scale[0]) u = ( u - m_TDomain[0][0])/ scale[0];
			if(scale[1]) v = ( v - m_TDomain[0][1])/ scale[1];

			m_pGeo->m_TC[i][0] = u;
			m_pGeo->m_TC[i][1] = v;
		}						
	}

	m_pGeo->Write(file);

	if(HasTCoords() && m_ScaledTDomain) {
    // scale TCoords and store in m_pGeo
		delete[] m_pGeo->m_TC;
		m_pGeo->m_TC = NULL;
		m_pGeo->TCoord=FALSE;
	}


  WriteAttributes( file, TCODE_ENDOFTABLE );

	file.EndChunk();			// end chunk 
	return;
}


//////////////////////////////////////////////////////////////////////////
//		CRhinoMeshGeometry::Write(...) 
//
//	Write a TCODE_COMPRESSED_MESH_GEOMETRY chunk.
//	This chunk contains all the mesh geometry.  It has no sub-chunks.
//	The data for vertices, faces, normals and texture coordinates 
//  are compressed for space efficiency.  Each section of
//  code below shows the specific compression scheme used.
////////////////////////////////////////////////////////////////////////// 
void 
CRhinoMeshGeometry::Write(CRhinoFile& file) const 
{

	CRhino3DPoint BoundBox[2];
	GetBoundingBox(BoundBox[0], BoundBox[1]);  // computes the bounding box


	file.BeginWriteChunk( *this );

	file.WriteInteger(1,&nVerts);	// Write number of vertices
	file.WriteInteger(1,&nFaces);		// Write number of faces
	file.WriteInteger(1,&Normals);		// Write BOOL indicating normals
	file.WriteInteger(1,&TCoord);		// Write BOOL indicating Tcoords
	file.WriteDouble(6,(double*)BoundBox);	// Write coordinates of low and high
											//      corners of bounding box
	
	// Transform vertex coordinates to a bounding box alligned coordinate system.
	// In this system the lower bounding box corner is at the origin and
	// the higher corner has coordinate (65535, 65535, 65535).  The vertex 
	// positions in this coordinate system are rounded to integers and stored
	// as 2 byte unsigned integers.
	// TODO: will this work if shorts are more than 2 bytes???

	CRhinoWorkSpace work;		//Allocate temporary work space
	unsigned short (*V)[3];
	int size=nVerts*sizeof(V[0]);
	V = (unsigned short (*)[3]) work.GetMemory(size);
	
	// Transform coordinates
	double scale[3];
	int i,j;
	for( i=0; i<3;i++){
		double diff=(BoundBox[1][i] - BoundBox[0][i]);
		scale[i]= (diff)? 65535/diff : 0.0;				;
	};

	for(i=0;i<nVerts;i++){
		for( j=0; j<3;j++)
			V[i][j] = (unsigned short) floor( 
				   scale[j]*(m_V[i][j] - BoundBox[0][j]) +.5); 
	}
	
	file.WriteShort(3*nVerts, (short*) V);	//write vertex positions


	// Write face definitions.  Each face is a quadruple of vertex indices. 
	// If there are less than 65536 vertices the indicies are stored as
	// unsigned shorts otherwise as unsigned ints.
	if(nVerts<65536){
		unsigned short (*F)[4];
		size = nFaces * sizeof(F[0]);
		F = (unsigned short (*)[4]) work.GetMemory(size);
		for ( i = 0; i < nFaces; i++ ) {
			for (j = 0; j < 4; j++ ) {
				F[i][j] = m_F[i][j];
			}
		}
	
		file.WriteShort(4*nFaces, (short*) F);

	} 
	else 
		file.WriteInteger(4*nFaces, (int*) m_F);

	// Write out normals.  Normalized normals are scaled by 127 rounded to the 
	// nearest integer and stored as signed chars. 
	if(Normals){
		signed char (*N)[3];
		size = nVerts*sizeof(N[0]);
		N = (signed char (*)[3]) work.GetMemory(size);		
		for (i=0; i<nVerts; i++) {
			Normalize( m_N[i] );
			for(j=0; j<3; j++)
				N[i][j] = (signed char) floor(127*m_N[i][j] + .5);
		}

		file.WriteByte(size, (char*) N);	// Write normals

	}

	// Write texture coordinates.  The texture domain is the unit square.
	// They are scaled 
	// by 65535 rounded to the nearest integer and stored as unsigned shorts
	// (2 bytes).
	

	if ( TCoord ) {
		unsigned short (*T)[2];
		size = nVerts * sizeof(T[0]);
		T = (unsigned short (*)[2]) work.GetMemory(size);
		for ( i = 0; i < nVerts; i++ ) {
			// clamp coordinates into the domain
			double u = m_TC[i][0];
			double v = m_TC[i][1];
			if(u<0) u = 0;
			if(u>1) u = 1;
			if(v<0) v = 0;
			if(v>1) v = 1;

			// scale by 65535 and round to a int
			T[i][0] = (unsigned short) floor(u*65535.0 + .5);
			T[i][1] = (unsigned short) floor(v*65535.0 + .5);
		}						

		file.WriteShort(2*nVerts, (short*) T);

	}
	file.EndChunk();			//TCODE_Compressed_Mesh_Geometry
	return;
}

BOOL
CRhinoMeshGeometry::GetBoundingBox( // returns FALSE if object has no box
               double boxmin[3], // min
               double boxmax[3], // max
               BOOL bGrowBox     // = FALSE
               ) const
{
  if ( !m_V ) {
    if ( !bGrowBox ) {
      if ( boxmin) boxmin[0] = boxmin[1] = boxmin[2] = 0.0;
      if ( boxmax) boxmax[0] = boxmax[1] = boxmax[2] = 0.0;
    }
    return (bGrowBox && nVerts==0);
  }
  return RhinoBoundingBox( 3, FALSE, 
                           nVerts, 3, &m_V[0][0],
                           boxmin, boxmax, bGrowBox 
                           );
}

//////////////////////////////////////////////////////////////////////////
//       CRhinoMesh::SetVertexCount(...)
//	Delete all previous vertex information (vertices, normals and texture 
//  coordinates).  Allocate new arrays for this information and leave  it
//  unintialized.  If n<0 treat it like 0.
//////////////////////////////////////////////////////////////////////////
void
CRhinoMesh::SetVertexCount( int n )
{
	delete[] m_pGeo->m_V;
	delete[] m_pGeo->m_N;

	if(HasTCoords() && m_ScaledTDomain)
		delete[] m_pTC;
	else if(HasTCoords() && !m_ScaledTDomain)
		delete[] m_pGeo->m_TC;

	if(n<=0){
		m_pGeo->m_V = NULL;
		m_pGeo->m_N = NULL;
		m_pGeo->m_TC = NULL;
		m_pTC = NULL;
		m_pGeo->nVerts = 0;
	} else {
		m_pGeo->m_V = new CRhino3DPoint[n];
		m_pGeo->nVerts = n;
		if(HasNormals())
			m_pGeo->m_N = new CRhino3DPoint[n];
		else 
			m_pGeo->m_N = NULL;
		if(HasTCoords() && m_ScaledTDomain)
			m_pTC = new CRhino2DPoint[n];
		else if(HasTCoords() && !m_ScaledTDomain)
			m_pGeo->m_TC = new CRhino2DPoint[n];
		else 
			m_pGeo->m_TC = NULL;
	}

	SynchPointers();
	return;
}

//////////////////////////////////////////////////////////////////////////
//	Delete all face data.  Allocate a new array and leave it unintialized.
//  If n<0 treat it like 0.
void
CRhinoMesh::SetFaceCount(int n)
{

	delete[] m_F;

	if(n<=0){
		m_pGeo->m_F = NULL;
		m_pGeo->nFaces = 0;
	} else {
		m_pGeo->m_F =  new int[n][4];
    memset( m_pGeo->m_F, 0xFF, n*4*sizeof(m_F[0][0]) );
		m_pGeo->nFaces = n;
	}

	SynchPointers();
	return;	
}

//////////////////////////////////////////////////////////////////////////
//	Set or reset HasNormals indicator.  If no change to indicator do nothing.
//  If indicator changes either allocate or delete array of normals.
//	A new array is left uninitialized.
//////////////////////////////////////////////////////////////////////////
void
CRhinoMesh::SetNormalsExist(BOOL b)
{

	if(b&&m_pGeo->Normals  || !(b||m_pGeo->Normals)) // If no change
		return;

	if(b) 	// allocate array for normals
		m_pGeo->m_N =  new CRhino3DPoint[VertexCount()];

	else{	// delete array for normals
		delete[] m_pGeo->m_N;
		m_pGeo->m_N = NULL;
	}

	m_pGeo->Normals = (b)? TRUE: FALSE;			// m_pGeo->Normals = b;

	SynchPointers();
	return;
}

//////////////////////////////////////////////////////////////////////////
//	Set or reset HasTCoords indicator.  If no change to indicator do nothing.
//  If indicator changes either allocate or delete array of texture coordinates.
//  When the array is allocated, the texture domain is set to the unit square.
//	The new array is left uninitialized.
//////////////////////////////////////////////////////////////////////////
void
CRhinoMesh::SetTCoordsExist(BOOL b)
{

	if(b&&HasTCoords() || !(b||HasTCoords()) )	// If no change
		return;

	if(b){ 	// allocate array for Tcoords
		m_HasTCoords=TRUE;
		m_ScaledTDomain=FALSE;
		m_pGeo->TCoord =TRUE;

		m_pGeo->m_TC =  new CRhino2DPoint[VertexCount()];
		CRhino2DPoint UnitSquare[2] = {{0.0,0.0}, {1.0,1.0}};
		int i,j;								
		for(i=0;i<2;i++)
				for(j=0;j<2;j++) m_TDomain[i][j] = UnitSquare[i][j];
	}
	else{	// delete array for Tcoords
		if(m_ScaledTDomain){
			delete[] m_TC;
			m_pTC = NULL;
		} else {
			delete[] m_pGeo->m_TC;
			m_pGeo->TCoord = FALSE;
			m_pGeo->m_TC=NULL;
			m_pTC=NULL;
		}

		m_HasTCoords = FALSE;
		m_ScaledTDomain=FALSE;
		CRhino2DPoint Origin[2] = {{0.0,0.0}, {0.0,0.0}};
		int i,j;								
		for(i=0;i<2;i++)
				for(j=0;j<2;j++) m_TDomain[i][j] = Origin[i][j];	
	}

	SynchPointers();
	return;
}

void 
CRhinoMesh::GetTextureDomain(CRhino2DPoint& low, CRhino2DPoint& high) const
{
	low[0] = m_TDomain[0][0];
	low[1] = m_TDomain[0][1];
	high[0] = m_TDomain[1][0];
	high[1] = m_TDomain[1][1];
	return;
}

////////////////////////////////////////////////////////////////////////////
//   Set the texture domain.  No effect unless HasTCoord() is true.
//  
//
void 
CRhinoMesh::SetTextureDomain(CRhino2DPoint low, CRhino2DPoint high)
{
	if(low[0]>high[0] || low[1]>high[1])
		throw CRhinoException(CRhinoException::illegal_input);

	if(!HasTCoords() || m_TDomain[0][0]==low[0]  &&	 m_TDomain[0][1] == low[1] && 
										 m_TDomain[1][0]==high[0] &&  m_TDomain[1][1] == high[1]) 
			return;					// do nothing;
	
	if(!m_ScaledTDomain){   // changing from default domain
		m_pTC = m_pGeo->m_TC;   // move the array 
		m_pGeo->TCoord=FALSE;
		m_pGeo->m_TC = NULL;
		m_ScaledTDomain = TRUE;
		}
	else if(m_TDomain[0][0]==0  &&	 m_TDomain[0][1] == 0 && 
					m_TDomain[1][0]==1.0 &&  m_TDomain[1][1] == 1.0){	// changing to default domain
			m_pGeo->TCoord=TRUE;
			m_pGeo->m_TC = m_pTC;
			m_ScaledTDomain=FALSE;
		}
	
	m_TDomain[0][0] = low[0];
	m_TDomain[0][1] = low[1]; 
	m_TDomain[1][0] = high[0]; 
	m_TDomain[1][1] = high[1];
	
	SynchPointers();
	return;
}

BOOL
CRhinoMesh::SetVertex( int v_index, 
                       double x, double y, double z )
{
  if ( v_index < 0 || v_index >= VertexCount() )
    return FALSE;

  m_V[v_index][0] = x;
  m_V[v_index][1] = y;
  m_V[v_index][2] = z;
  return TRUE;
}


BOOL
CRhinoMesh::SetNormal( int v_index, 
                       double x, double y, double z )
{
  if ( !m_pGeo->Normals || v_index < 0 || v_index >= VertexCount() )
    return FALSE;

  double l = x*x + y*y + z*z;
  if ( l <= 0.0 )
    return FALSE;
  if ( fabs(l-1.0) > 2.0e-4 ) {
    l = 1.0/sqrt(l);
    x *= l;
    y *= l;
    z *= l;
  }

  m_N[v_index][0] = x;
  m_N[v_index][1] = y;
  m_N[v_index][2] = z;

  return TRUE;
}

BOOL
CRhinoMesh::SetTCoord( int v_index, 
                       double u, double v )
{
  if ( !m_HasTCoords || v_index < 0 || v_index >= VertexCount() )
    return FALSE;

  if ( u < m_TDomain[0][0] )
    u = m_TDomain[0][0];
  else if (u > m_TDomain[1][0])
    u = m_TDomain[1][0];
  if ( v < m_TDomain[0][1] )
    v = m_TDomain[0][1];
  else if ( v > m_TDomain[1][1] )
    v = m_TDomain[1][1];

  m_TC[v_index][0] = u;
  m_TC[v_index][1] = v;

  return TRUE;
}

BOOL
CRhinoMesh::SetFace( int f_index, 
                     int v0_index, int v1_index, int v2_index )
{
  if ( f_index < 0 || f_index >= FaceCount() )
    return FALSE;
  if ( v0_index < 0 || v0_index >= VertexCount() )
    return FALSE;
  if ( v1_index < 0 || v1_index >= VertexCount() )
    return FALSE;
  if ( v2_index < 0 || v2_index >= VertexCount() )
    return FALSE;
  m_F[f_index][0] = v0_index;
  m_F[f_index][1] = v1_index;
  m_F[f_index][2] = v2_index;
  m_F[f_index][3] = v2_index;
  return TRUE;
}


BOOL
CRhinoMesh::SetFace( int f_index, 
                     int v0_index, int v1_index, int v2_index, int v3_index )
{
  if ( f_index < 0 || f_index >= FaceCount() )
    return FALSE;
  if ( v0_index < 0 || v0_index >= VertexCount() )
    return FALSE;
  if ( v1_index < 0 || v1_index >= VertexCount() )
    return FALSE;
  if ( v2_index < 0 || v2_index >= VertexCount() )
    return FALSE;
  if ( v3_index < 0 || v3_index >= VertexCount() )
    return FALSE;
  m_F[f_index][0] = v0_index;
  m_F[f_index][1] = v1_index;
  m_F[f_index][2] = v2_index;
  m_F[f_index][3] = v3_index;
  return TRUE;
}

BOOL
CRhinoMesh::GetFace( 
            int f_index, 
            int& v0_index, int& v1_index, 
            int& v2_index, int& v3_index ) const
{
  BOOL rc = FALSE;
  // if the face is a triangle, then v2_index = v3_index.
  if ( f_index >= 0 && f_index < FaceCount() ) {
    v0_index = m_F[f_index][0];
    v1_index = m_F[f_index][1];
    v2_index = m_F[f_index][2];
    v3_index = m_F[f_index][3];
    rc = TRUE;
  }
  else {
    v0_index = 0;
    v1_index = 0;
    v2_index = 0;
    v3_index = 0;
  }
  return rc;
}

BOOL
CRhinoMesh::GetVertex( 
            int v_index, 
            double& x, double& y, double& z ) const
{
  BOOL rc = FALSE;
  // if the face is a triangle, then v2_index = v3_index.
  if ( v_index >= 0 && v_index < VertexCount() ) {
    x = m_V[v_index][0];
    y = m_V[v_index][1];
    z = m_V[v_index][2];
    rc = TRUE;
  }
  return rc;
}

BOOL
CRhinoMesh::GetNormal( 
            int v_index, 
            double& x, double& y, double& z ) const
{
  BOOL rc = FALSE;
  // if the face is a triangle, then v2_index = v3_index.
  if ( v_index >= 0 && v_index < VertexCount() && HasNormals() ) {
    x = m_N[v_index][0];
    y = m_N[v_index][1];
    z = m_N[v_index][2];
    rc = TRUE;
  }
  else {
    x = y = z = 0.0;
  }
  return rc;
}

BOOL
CRhinoMesh::GetTCoord( 
            int v_index, 
            double& u, double& v ) const
{
  BOOL rc = FALSE;
  // if the face is a triangle, then v2_index = v3_index.
  if ( v_index >= 0 && v_index < VertexCount() && HasTCoords() ) {
    u = m_TC[v_index][0];
    v = m_TC[v_index][1];
    rc = TRUE;
  }
  else {
    u = v = 0.0;
  }
  return rc;
}

// Count the number of triangles in the mesh
int CRhinoMesh::TriangleCount( void ) const
{
  const int face_count = FaceCount();
  int triangle_count = 0;
  int face_index;
  for ( face_index = 0; face_index < face_count; face_index++ ) {
    if ( IsTriangle( face_index ) )
      triangle_count++;
  }
  return triangle_count;
}

BOOL
CRhinoMesh::GetBoundingBox(
          double box_min[3],
          double box_max[3],
          BOOL bGrowBox     // = FALSE
          ) const
{
  BOOL rc = RhinoBoundingBox( 3, 0, VertexCount(), 3, &m_pGeo->m_V[0][0],
                              box_min, box_max, bGrowBox );
  return rc;
}

BOOL
CRhinoMesh::IsValid( void ) const
{
  // TODO: add more mesh validation
  if ( VertexCount() < 1 )
    return FALSE;
  return TRUE;
}
