/* $Header: /Src/RhinoIO/rhinoio_mesh.h 6     10/28/98 12:11p Dalelear $ */
/* $NoKeywords: $ */
//
// Copyright (c) 1993-1998 Robert McNeel & Associates. All rights reserved.
// Rhinoceros is a registered trademark of Robert McNeel & Assoicates.
//
// THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
// ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
// MERCHANTABILITY ARE HEREBY DISCLAIMED.
//				
////////////////////////////////////////////////////////////////
//
//   defines CRhinoMeshGeometry, CRhinoMesh
//
////////////////////////////////////////////////////////////////

#if !defined(RHINO_MESH_INC_)
#define RHINO_MESH_INC_

///////////////////////////////////////////////////////////////////////////////
//  Class CRhinoMeshGeometry
// This chunk contains all the geometry for a CRhinoMesh object.  It has vertex
// locations, face definitions, normals and texture coordinates.  The data is 
// stored in a compressed format. 
///////////////////////////////////////////////////////////////////////////////
//class CRhinoMeshGeometry: public CRhinoChunk 
class CRhinoMeshGeometry: public CRhinoObject
{
	RHINO_CHUNK_DECLARE( CRhinoMeshGeometry )
public:
	CRhinoMeshGeometry();
	~CRhinoMeshGeometry();

	int nVerts;
	int nFaces;
	BOOL Normals;
	BOOL TCoord;

	CRhino3DPoint* m_V;		// Array of vertex positions
	int (*m_F)[4];				// Array of vertex indices for each face.
	CRhino3DPoint * m_N;	// Array of unit normal for each vertex
	CRhino2DPoint * m_TC;	// Texture coordinates for each vertex
												// Should be in rectangular domain given
												// by GetTextureDomain

	// Override Read, Write, and GetBoundingBox defualts
	void Write(CRhinoFile &) const;	
	CRhinoFile::eStatus Read(CRhinoFile&);							
  BOOL GetBoundingBox( // returns FALSE if object has no box
                 double[3], // min
                 double[3], // max
                 BOOL = FALSE // set to TRUE if box should be grown
                 ) const;
};

//inline CRhinoMeshGeometry::CRhinoMeshGeometry()
//:CRhinoChunk(TCODE_COMPRESSED_MESH_GEOMETRY) 
//  {nVerts =0;nFaces =0;Normals=0;TCoord=0;m_V=NULL;m_F=NULL;m_N=NULL;m_TC=NULL;}
inline CRhinoMeshGeometry::CRhinoMeshGeometry()
:CRhinoObject(TCODE_COMPRESSED_MESH_GEOMETRY) 
  {nVerts =0;nFaces =0;Normals=0;TCoord=0;m_V=NULL;m_F=NULL;m_N=NULL;m_TC=NULL;}

inline CRhinoMeshGeometry::~CRhinoMeshGeometry()
  {delete[] m_V;delete[] m_F;delete[] m_N;delete[] m_TC;}


//////////////////////////////////////////////////////////////////////////////
//   Class CRhinoMesh
// Mesh object composed of 3 and 4 sided faces.  Unit normals and texture
// coordinates may be associated to each vertex.  Each face is represented
// as a quadruple of vertex indices (vertices are indexed 0 to NumVertex()-1).
// For triangular faces the last 2 indices are the same.
//////////////////////////////////////////////////////////////////////////////
class CRhinoMesh: public CRhinoObject 
{
	RHINO_CHUNK_DECLARE( CRhinoMesh )
public:
	// Note: Use these as arrays or pointers, but the pointer is const. 
	//       All these arrays are indexed from 0 to VertexCount()-1.
	CRhino3DPoint *const& m_V;		// Array of vertex positions 
	int (*const& m_F)[4];					// Array of vertex indices for each face.
                                // The vertices are listed in counter-clockwise
                                // order.
	CRhino3DPoint *const& m_N;		// Array of unit normal for each vertex
	CRhino2DPoint *const& m_TC;		// Texture coordinates for each vertex
																//    Should be in rectangular domain given
																//    by GetTextureDomain

	CRhinoMesh();			// Create an empty mesh

	virtual ~CRhinoMesh();

  // CRhinoObject overrides
  BOOL IsValid( void ) const;

  CRhinoFile::eStatus Read(CRhinoFile&);		// reads chunk data from file
	void Write(CRhinoFile&) const;				    // write a chunk to the file
  BOOL GetBoundingBox(
          double[3], // min
          double[3], // max
          BOOL = FALSE // set to TRUE if box should be grown
          ) const;

  int VertexCount( void ) const;       // Number of vertices
  int FaceCount( void ) const;         // Total number of faces
	int QuadrangleCount( void ) const;   // Number of faces which are 4 sided
	int TriangleCount( void ) const;     // Number of faces which are 3 sided
  BOOL IsQuadrangle(
        int // face_index
        ) const;
  BOOL IsTriangle(
        int // face_index
        ) const;

	BOOL HasNormals(void) const;
	BOOL HasTCoords(void) const;
		

  // You must call SetNumVertices(), SetNumFaces(), SetNormalsExist(),
  // SetTCoordsExist(), and SetTextureDomain() before using any of 
  // SetVertex(), SetFace(), SetNormal(), or SetTCoord().

	void SetVertexCount(int);		// Delete vertices and allocate new array
	void SetFaceCount(int);			// Delete face array and allocate new array
	void SetNormalsExist(BOOL = TRUE);		// Allocate arrays for normals if true
	void SetTCoordsExist(BOOL = TRUE);		// Allocate TCoord array if true

	// Texture Domain is a rectangular bounding box containing all the texture
	// coordinates.  The low left and high right corners are given by these functions
	// The default texture domain is the unit square.
	void GetTextureDomain(CRhino2DPoint& low, CRhino2DPoint& high) const;
	void SetTextureDomain(CRhino2DPoint low, CRhino2DPoint high);
	
  BOOL SetVertex(
         int,                   // index of vertex = 0 to VertexCount()-1
         double, double, double
         );

  BOOL SetNormal(
         int,                   // index of vertex = 0 to VertexCount()-1
         double, double, double
         );

  BOOL SetTCoord(
         int,                   // index of vertex = 0 to VertexCount()-1
         double, double
         );

  BOOL SetFace(
         int,                   // index of face = 0 to FaceCount()-1
         int, int, int          // vertex indices of triangle corners
                                // in counter-clockwise order
         );

  BOOL SetFace(
         int,                   // index of face = 0 to FaceCount()-1
         int, int, int, int     // vertex indices of quad corners
                                // in counter-clockwise order
         );

  BOOL GetVertex(
         int,                   // index of vertex = 0 to VertexCount()-1
         double&, double&, double& // OUTPUT: coordinates
         ) const;

  BOOL GetNormal(
         int,                   // index of vertex = 0 to VertexCount()-1
         double&, double&, double& // OUTPUT: coordinates
         ) const;

  BOOL GetTCoord(
         int,                   // index of vertex = 0 to VertexCount()-1
         double&, double&       // OUTPUT: texture coordinates
         ) const;

  BOOL GetFace(
         int,                   // index of face = 0 to FaceCount()-1
         int&, int&, int&, int& // OUTPUT: vertex indices of face corners
                                //         listed in counter-clockwise order.
                                //         If the last two indices are
                                //         equal, then the face is a
                                //         triangle.
         ) const;


	// corners of box bounding all vertices with least and greatest coordinates 
	void GetBoundingBox(CRhino3DPoint& low, CRhino3DPoint& high) const;   
	
  // obsolete
  int NumVertices(void) const;
	int NumFaces(void) const;
	
private:
	void Initialize();				// Initialization for constructors

	CRhinoMeshGeometry * m_pGeo; //All the geometry

														// Copies of the pointers in m_pGeo
	CRhino3DPoint* m_pV;			// Pointer to vertex positions
	int (*m_pF)[4];						// Pointer to face definitions
	CRhino3DPoint * m_pN;			// Pointer to unit normals

														// Sometimes a copy from m_pGeo 
	CRhino2DPoint * m_pTC;		// Pointer to texture coordinates

	void SynchPointers();			// sync up these copies with those in *m_pGeo

	BOOL m_HasTCoords;						// Texture coordinates present
	CRhino2DPoint m_TDomain[2];		// Texture coordinate domain
	BOOL m_ScaledTDomain;					// True if HasTCoord() and TDomain is not the unit square
						// If TRUE, m_pTC is the array of texture coordintes and *m_pGeo doesn't have TC's
						// If FALSE and m_HasTCoords=TRUE, m_pTC is copy of m_pGeo->m_TC
				
															// otherwise m_pTC is just a copy.
};		


inline int CRhinoMesh::VertexCount(void) const 
            {return m_pGeo->nVerts;}
inline int CRhinoMesh::FaceCount(void) const 
            {return m_pGeo->nFaces;}
inline int CRhinoMesh::QuadrangleCount(void) const 
            {return FaceCount() - TriangleCount();}
inline BOOL CRhinoMesh::HasNormals(void) const 
            {return m_pGeo->Normals;}
inline BOOL CRhinoMesh::HasTCoords(void) const 
            {return m_HasTCoords;}
inline BOOL CRhinoMesh::IsTriangle(int face_index) const
            {return m_F[face_index][2] == m_F[face_index][3];}
inline BOOL CRhinoMesh::IsQuadrangle(int face_index) const
            {return m_F[face_index][2] != m_F[face_index][3];}

// obsolete
inline int CRhinoMesh::NumVertices(void) const {return m_pGeo->nVerts;}
inline int CRhinoMesh::NumFaces(void)    const {return m_pGeo->nFaces;}

#endif
