/* $Header: /Src/RhinoIO/rhinoio_nurbscrv.cpp 18    10/20/99 1:28p Dalelear $ */
/* $NoKeywords: $ */
//
// Copyright (c) 1993-1998 Robert McNeel & Associates. All rights reserved.
// Rhinoceros is a registered trademark of Robert McNeel & Assoicates.
//
// THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
// ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
// MERCHANTABILITY ARE HEREBY DISCLAIMED.
//				
////////////////////////////////////////////////////////////////
 
#include "rhinoio_object.h"
#include "rhinoio_nurbscrv.h"

RHINO_CHUNK_IMPLEMENT(CRhinoNurbsCurve, TCODE_RHINO_OBJECT_NURBS_CURVE)

// Initialize an empty CRhinoNurbsCurve
CRhinoNurbsCurve::CRhinoNurbsCurve() 
: CRhinoObject( TCODE_RHINO_OBJECT_NURBS_CURVE ) 
{
  Initialize();
}


// Initialize an empty CRhinoNurbsCurve
CRhinoNurbsCurve::CRhinoNurbsCurve(
    int   dimension,    // dimension ( >= 1 and usually = 3 )
    BOOL  bIsRational,  // TRUE if curve is rational
    int   degree,       // degree ( >= 1 )
    int   cv_count      // number of CVs ( >= degree+1 )
  ) 
: CRhinoObject( TCODE_RHINO_OBJECT_NURBS_CURVE ) 
{
  Initialize();
  if ( dimension < 1 ||
       degree < 1 ||
       cv_count <= degree ) {
		throw CRhinoException( CRhinoException::illegal_input);
  }
  m_dim           = dimension;
  m_bIsRat        = (bIsRational) ? TRUE : FALSE;
  m_order         = degree+1;
  m_cv_count      = cv_count;
  m_knot_capacity = m_order + m_cv_count - 2;
  const int cvdim = (m_bIsRat) ? m_dim+1 : m_dim;
  m_cv_capacity   = m_cv_count*cvdim;
  m_cv_stride     = cvdim;

  // rhmalloc() is used instead of new[] so that the arrays can be
  // efficiently resized using rhrealloc().

  m_knot = (double*)rhmalloc( m_knot_capacity*sizeof(*m_knot) );
  if ( !m_knot ) {
		throw CRhinoException( CRhinoException::out_of_memory);
  }
  m_knot[0] = DBL_MAX;
  m_knot[m_knot_capacity-1] = -DBL_MAX;
  m_cv = (double*)rhmalloc( m_cv_capacity*sizeof(*m_cv) );
  if ( !m_cv ) {
		throw CRhinoException( CRhinoException::out_of_memory);
  }
  m_cv[0] = DBL_MAX;
  m_cv[m_cv_capacity-1] = -DBL_MAX;
}


CRhinoNurbsCurve::~CRhinoNurbsCurve()
{
  if ( m_cv_capacity > 0 && m_cv ) {
    rhfree( m_cv );
  }
	if ( m_knot_capacity > 0 && m_knot ) {
    rhfree( m_knot );
  }
}

BOOL 
CRhinoNurbsCurve::IsClosed() const
{
  double ws[16];
  BOOL bIsClosed = FALSE;
  const int cvsize = CVSize();
  if ( m_dim > 0 ) {
    if ( RhinoIsKnotVectorClamped( m_order, m_cv_count, m_knot ) ) {
      double* P0 = (cvsize > 8) ? (double*)rhmalloc( cvsize*2*sizeof(*P0) ) : ws;
      double* P1 = P0 + m_dim;
      GetCV( 0, rh_not_rational, P0 );
      GetCV( CVCount()-1, rh_not_rational, P1 );
      bIsClosed = RhinoIsEqual( m_dim, P0, P1 );
      if ( cvsize > 8 )
        rhfree(P0);
    }
    else if ( IsPeriodic() ) {
      bIsClosed = TRUE;
    }
  }
  return bIsClosed;
}

BOOL 
CRhinoNurbsCurve::IsPeriodic() const
{
  BOOL bIsPeriodic = RhinoIsKnotVectorPeriodic( m_order, m_cv_count, m_knot );
  if ( bIsPeriodic ) {
    int i0 = m_order-2;
    int i1 = m_cv_count-1;
    const int cvdim = (m_bIsRat) ? m_dim+1 : m_dim;
    const double* cv0 = m_cv + i0*m_cv_stride;
    const double* cv1 = m_cv + i1*m_cv_stride;
    for ( /*empty*/; i0 >= 0; i0--, i1-- ) {
      if ( !RhinoIsEqual( cvdim, cv0, cv1 ) )
        return FALSE;
      cv0 -= m_cv_stride;
      cv1 -= m_cv_stride;      
    }
  }
  return bIsPeriodic;
}


BOOL 
CRhinoNurbsCurve::SetCV( int i, RHINO_POINT_STYLE style, const double* Point )
{
  int k;
  double w;

  // feeble but fast check for properly initialized class
  if ( !m_cv || i < 0 || i >= m_cv_count )
    return FALSE;

  double* cv = m_cv + i*m_cv_stride;

  switch ( style ) {

  case rh_not_rational:  // input Point is not rational
    memcpy( cv, Point, m_dim*sizeof(*cv) );
    if ( IsRational() ) {
      // NURBS curve is rational - set weight to one
      cv[m_dim] = 1.0;
    }
    break;

  case rh_homogeneous_rational:  // input Point is homogeneous rational
    if ( IsRational() ) {
      // NURBS curve is rational
      memcpy( cv, Point, (m_dim+1)*sizeof(*cv) );
    }
    else {
      // NURBS curve is not rational
      w = (Point[m_dim] != 0.0) ? 1.0/Point[m_dim] : 1.0;
      for ( k = 0; k < m_dim; k++ ) {
        cv[k] = w*Point[k];
      }
    }
    break;

  case rh_euclidean_rational:  // input Point is euclidean rational
    if ( IsRational() ) {
      // NURBS curve is rational - convert euclean point to homogeneous form
      w = Point[m_dim];
      for ( k = 0; k < m_dim; k++ )
        cv[k] = w*Point[k];
      cv[m_dim] = w;
    }
    else {
      // NURBS curve is not rational
      memcpy( cv, Point, m_dim*sizeof(*cv) );
    }
    break;
  }
  return TRUE;
}

BOOL 
CRhinoNurbsCurve::GetCV( int i, RHINO_POINT_STYLE style, double* Point ) const
{
  const double* cv = CV(i);
  if ( !cv )
    return FALSE;
  int dim = Dimension();
  double w = ( IsRational() ) ? cv[dim] : 1.0;
  switch(style) {
  case rh_euclidean_rational:
    Point[dim] = w;
    // no break here
  case rh_not_rational:
    if ( w == 0.0 )
      return FALSE;
    w = 1.0/w;
    while(dim--) *Point++ = *cv++ * w;
    break;
  case rh_homogeneous_rational:
    Point[dim] = w;
    memcpy( Point, cv, dim*sizeof(*Point) );
    break;
  default:
    return FALSE;
  }
  return TRUE;
}

BOOL 
CRhinoNurbsCurve::SetKnot( int knot_index, double k )
{
  if ( knot_index < 0 || knot_index >= KnotCount() )
    return FALSE;
  m_knot[knot_index] = k;
  return TRUE;
}

BOOL CRhinoNurbsCurve::IsValid() const
{
  // check fields for illegal values
  if ( m_dim < 1 )
    return FALSE;
  if ( m_order < 2 )
    return FALSE;
  if ( m_cv_count < m_order )
    return FALSE;
  if ( !m_knot )
    return FALSE;
  if ( !m_cv )
    return FALSE;

  const int d = m_order-1;
  const int k0 = m_order-2;
  const int k1 = m_cv_count-1;
  const int kcnt = m_order+m_cv_count-2;
  int i, j;
  // knots must be monotone increasing
  for ( i = 1; i < kcnt; i++ ) {
    if ( m_knot[i-1] > m_knot[i] )
      return FALSE;
  }
  // first span must be non-empty
  if ( m_knot[k0] >= m_knot[k0+1] )
    return FALSE;
  // last span must be non-empty
  if ( m_knot[k1-1] >= m_knot[k1] )
    return FALSE;
  // max knot multiplicity <= degree
  for ( i = k0, j = k0+d; j <= k1; i++,j++ ) {
    if ( m_knot[i] >= m_knot[j] )
      return FALSE;
  }

  const int cvdim = (m_bIsRat) ? m_dim+1 : m_dim;

  // make sure cv strides have reasonable values
  if ( m_cv_stride < cvdim )
    return FALSE;

  // make sure cvs have been initialized.
  if ( m_cv[0] == DBL_MAX || m_cv[ m_cv_count-1 ] == -DBL_MAX )
    return FALSE;

  return TRUE;
}

CRhinoFile::eStatus
CRhinoNurbsCurve::Read( CRhinoFile& file)
{
  CRhinoFile::eStatus status = CRhinoFile::not_supported;
 
  int data_typecode = 0;
  int data_length = 0;

  file.ReadInteger( 1, &data_typecode );
  if ( data_typecode != TCODE_RHINO_OBJECT_DATA )
    return CRhinoFile::crc_error;
  file.ReadInteger( 1, &data_length );
  if ( data_length <= 0 )
    return CRhinoFile::crc_error;
  
  int version = -1;

  // read NURBS curve definition
  file.ReadInteger( 1, &version );
  
  // int bReverse = version & 0x100;
  version &= 0xFFFFFEFF;

  switch ( version ) {

  case 100: // written by RhinoIO toolkit
  case 101: // written by Rhino
    {
      int i5[5];
      file.ReadInteger( 5, i5 );

      m_dim         = i5[0];
      m_bIsRat      = (i5[1]==1);
      m_order       = i5[2];
      m_cv_count    = i5[3];

      if ( m_dim < 1  ||
           i5[1] < 0 || i5[1] > 1 ||
           m_order < 2 ||
           m_cv_count < m_order ||
           i5[4] != 0 ) {
        status = CRhinoFile::crc_error;
      }
      else {
        m_knot_capacity = m_cv_count+m_order-2;
        m_knot = (double*)rhmalloc( m_knot_capacity*sizeof(*m_knot) );
        if ( !m_knot ) {
          throw CRhinoException( CRhinoException::out_of_memory );
        }
        file.ReadDouble( m_knot_capacity, m_knot );
        const int cvdim = ( m_bIsRat ) ? m_dim+1 : m_dim;
        m_cv_stride = cvdim;
        m_cv_capacity = cvdim*m_cv_count;
        m_cv = (double*)rhmalloc( m_cv_capacity*sizeof(*m_cv) );
        if ( !m_cv ) {
          throw CRhinoException( CRhinoException::out_of_memory );
        }
        file.ReadDouble( m_cv_capacity, m_cv );
        status = IsValid() ? CRhinoFile::good : CRhinoFile::crc_error;
      }
    }
    break;

  default:
    status = CRhinoFile::not_supported;
    break;
  }

  // read attributes
  CRhinoFile::eStatus att_status = ReadAttributes( file, TCODE_RHINO_OBJECT_END );
  if ( status == CRhinoFile::good )
    status = att_status;

  /*
  if ( bReverse ) {
    // add code to reverse crv
  }
  */

  return status;
}

void
CRhinoNurbsCurve::Write(CRhinoFile& file) const
{
  if ( !IsValid() ) {
    throw CRhinoException( CRhinoException::illegal_input );
  }

	// Make sure layer definition is in the file
	file.LayerIndex( Layer() );	// Get layer index or -1

  // for each point in the pointset write out a chunk
	//file.BeginWriteChunk( *this );
  file.BeginWriteChunk( CRhinoNurbsCurve::TypeCode() ); // so CRhinoAgSpline::Write() will work
  {

    file.BeginWriteChunk( TCODE_RHINO_OBJECT_DATA );
    {
      // write NURBS curve data
      const int version = 100;
      int i6[6] = { version,
                    Dimension(),
                    (IsRational() ? 1 : 0),
                    Order(),
                    CVCount(),
                    0
      };

      file.WriteInteger( 6, i6 );

      file.WriteDouble( KnotCount(), m_knot );

      const int imax = CVCount();
      const int cvdim = CVSize();
      int i;
      for ( i = 0; i < imax; i++ ) {
        file.WriteDouble( cvdim, CV(i) );
      }

    }
    file.EndChunk();

    WriteAttributes( file, TCODE_RHINO_OBJECT_END );
  }
  file.EndChunk();
}

void
CRhinoNurbsCurve::Initialize( )
{
  m_dim           = 0;
  m_bIsRat        = FALSE;
  m_order         = 0;
  m_cv_count      = 0;

  m_knot_capacity = 0;
  m_knot          = NULL;

  m_cv_stride     = 0;
  m_cv_capacity   = 0;
  m_cv            = NULL;
  m_flag          = 0;
  m_type          = rh_unknown_crv;
}

BOOL
CRhinoNurbsCurve::Domain( double* t0, double* t1 ) const
{
  return RhinoGetKnotVectorDomain( m_order, m_cv_count, m_knot, t0, t1 );
}

BOOL
CRhinoNurbsCurve::Reverse()
{
  BOOL rc0 = RhinoReverseKnotVector( m_order, m_cv_count, m_knot );
  BOOL rc1 = RhinoReversePointList( CVSize(), m_cv_count, m_cv_stride, m_cv );
  return rc0 && rc1;
}

BOOL
CRhinoNurbsCurve::SwapCoordinates( int i, int j )
{
  if ( i < 0 || j < 0 || i >= m_dim || j >= m_dim )
    return FALSE;
  const int cvcount = CVCount();
  if ( cvcount < 2 || Dimension() < 2 || CV(0) == NULL )
    return FALSE;
  if ( i == j )
    return TRUE;
  double t;
  int cvi;
  double* cv;
  for ( cvi = 0; cvi < cvcount; cvi++ ) {
    cv = CV(cvi);
    t = cv[i]; 
    cv[i] = cv[j];
    cv[j] = t;
  }
  return TRUE;
}

BOOL
CRhinoNurbsCurve::GetBoundingBox(
          double box_min[3],
          double box_max[3],
          BOOL bGrowBox     // = FALSE
          ) const
{
  BOOL rc = RhinoBoundingBox( m_dim, m_bIsRat, 
                         m_cv_count, 
                         m_cv_stride,
                         m_cv,
                         box_min, box_max, bGrowBox );
  return rc;
}
