/* $Header: /Src/RhinoIO/rhinoio_nurbscrv.h 18    10/20/99 1:28p Dalelear $ */
/* $NoKeywords: $ */
//
// Copyright (c) 1993-1998 Robert McNeel & Associates. All rights reserved.
// Rhinoceros is a registered trademark of Robert McNeel & Assoicates.
//
// THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
// ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
// MERCHANTABILITY ARE HEREBY DISCLAIMED.
//				
////////////////////////////////////////////////////////////////
//
//   defines CRhinoNurbsCurve
//
////////////////////////////////////////////////////////////////


#if !defined(RHINO_NURBSCRV_INC_)
#define RHINO_NURBSCRV_INC_

///////////////////////////////////////////////////////////////////////////////
// Class  CRhinoNurbsCurve
//
//	This object represents a tensor product NURBS curve.
///////////////////////////////////////////////////////////////////////////////

// If (X,Y,Z) is the 3D location of a NURBS control point, it may be
// represented in at least three different ways.
//
// Not rational:         ( X, Y, Z )
// Homogeneous rational: ( w*X, w*Y, w*Z, w )  w != 0
// Euclidean rational:   ( X, Y, Z, w )
//
// Internally, non-rational CRhinoNurbsCurve classes use the not rational form
// and rational CRhinoNurbsCurve classes use the homogeneous rational form.
// The SetCV() and GetCV() functions have a RHINO_POINT_STYLE argument that
// allows you to specify which form you want to use.

class CRhinoNurbsCurve: public CRhinoObject
{
	RHINO_CHUNK_DECLARE( CRhinoNurbsCurve )
public:

  // Construction
	CRhinoNurbsCurve();
	CRhinoNurbsCurve(     // Constructs a NURBS curve that can be filled in
                        // with calls to SetCV() and SetKnot().
                        // Throws a CRhinoException if input is illegal
                        // or not enough memory.
    int,  // dimension ( >= 1 and usually = 3 )
    BOOL, // TRUE if curve is rational
    int,  // degree for first parameter ( >= 1 )
    int   // cv_count = number of CVs
  );
  ~CRhinoNurbsCurve();

  // CRhinoObject overrides
	CRhinoFile::eStatus Read(   	// read chunk data from file 
    CRhinoFile&
    );
	void Write(                   // Writes NURBS curve to file
                                // Throws a CRhinoException if IsValid()
                                // returns FALSE or fwrite() fails.
    CRhinoFile&
    ) const;
  BOOL IsValid(                 // Call after class is fully initialized to
                                // perform a simple and quick validity check.
                                // Will detect illegal knot vectors and
                                // incorrect settings of integer m_* members.
                                // Does NOT check control vertices for
                                // grossly degenerate conditions.
    void
    ) const;
  BOOL GetBoundingBox(
          double[3], // min
          double[3], // max
          BOOL = FALSE // set to TRUE if box should be grown
          ) const;

  // Interface
  int Dimension(    // dimension of NURBS curve (usually 3)
        void
        ) const;

  BOOL IsRational(  // TRUE if NURBS curve is rational
        void
        ) const;
  
  BOOL IsClosed(  // TRUE if NURBS curve is closed (either curve has
        void      // clamped end knots and euclidean location of start
        ) const;  // CV = euclidean location of end CV, or curve is
                  // periodic.)

  BOOL IsPeriodic(  // TRUE if NURBS curve is periodic (degree > 1,
        void        // periodic knot vector, last degree many CVs 
        ) const;    // are duplicates of first degree many CVs.)
  
  int CVSize(       // number of doubles per control vertex 
        void        // = IsRational() ? Dim()+1 : Dim()
        ) const;
  
  int Order(        // order = degree + 1
        void
        ) const;
	
  int Degree(       // degree = order - 1
        void
        ) const;

  int CVCount(      // number of control vertices
        void 
        ) const;

  int KnotCount(    // total number of knots in knot vector
        void
        ) const;
  
  const double* CV(     // get pointer to control vertex
        int             // CV index ( >= 0 and < CVCount() )
        ) const;

  double* CV(           // get pointer to control vertex
        int             // CV index ( >= 0 and < CVCount() )
        );

  int SetCV(            // set a single control vertex
        int,            // CV index ( >= 0 and < CVCount() )
        RHINO_POINT_STYLE, // style of input point
        const double*   // value of control vertex
        );

  int GetCV(            // get a single control vertex
        int,            // CV index ( >= 0 and < CVCount() )
        RHINO_POINT_STYLE, // style to use for output point
        double*         // array of length >= CVSize()
        ) const;

  int SetKnot(
        int,            // knot index ( 0 to KnotCount - 1 )
        double          // value for knot
        );

  double Knot(
        int             // knot index ( >= 0 and < Order + CV_count - 2 )
        ) const;

  const double* Knot(   // knot[] array
        ) const;

  double SuperfluousKnot(
           int // 0 = start, 1 = end
           ) const;

  RHINO_KNOT_STYLE KnotStyle() const;

  BOOL Domain(
        double*,  // if not NULL, returns domain start
        double*   // if not NULL, returns domain end
        ) const;

  BOOL Reverse();       // reverse parameterizatrion
                        // Domain changes from [a,b] to [-b,-a]

  BOOL SwapCoordinates(
        int, int        // indices of coords to swap
        );

// Implementation
// protected:
public:

  // NOTE: These fields are left "public" so that expert users may efficiently
  //       create NURBS curves using the default constructor and borrow the
  //       knot and CV arrays from their native NURBS representation.
  //       No technical support will be provided for users who access these
  //       fields directly.  If you can't get your stuff to work, then use
  //       the constructor with the arguments and the SetKnot() and SetCV()
  //       functions to fill in the arrays.

  int     m_dim;            // Must be 3 for exposed Rhino IO uses.
                            // (Can be anything when a 2 variable B-spline
                            //  function is needed internally.)
  BOOL    m_bIsRat;         // TRUE  for rational B-splines.  Control vertices
                            //       use homogeneous form.
                            // FALSE for non-rational B-splines.  Control
                            //       verticies do not have a weight coordinate.
  int     m_order;          // order = degree+1
  int     m_cv_count;       // number of control vertices ( >= order )

  // knot vector memory
  int     m_knot_capacity;
                            // If m_knot_capacity > 0, then m_knot[]
                            // is an array of at least m_knot_capacity
                            // doubles whose memory is managed by the
                            // CRhinoNurbsCurve class using rhmalloc(),
                            // rhrealloc(), and rhfree().
                            // If m_knot_capacity is 0 and m_knot is
                            // not NULL, then  m_knot[] is assumed to
                            // be big enough for any requested operation
                            // and m_knot[] is not deleted by the
                            // destructor.
  double* m_knot;           // Knot vector. ( The knot vector has length
                            // m_order+m_cv_count-2. )
  
  // control vertex net memory
  int     m_cv_stride;      // The pointer to start of "CV[i]" is
                            //   m_cv + i*m_cv_stride.
  int     m_cv_capacity;    // If m_cv_capacity > 0, then m_cv[] is an array
                            // of at least m_cv_capacity doubles whose
                            // memory is managed by the CRhinoNurbsCurve
                            // class using rhmalloc(), rhrealloc(), and rhfree().
                            // If m_cv_capacity is 0 and m_cv is not
                            // NULL, then m_cv[] is assumed to be big enough
                            // for any requested operation and m_cv[] is not
                            // deleted by the destructor.
  double* m_cv;             // Control points.
                            // If m_bIsRat is FALSE, then control point is
                            //   ( CV(i)[0], ..., CV(i)[m_dim-1] ).
                            // If m_bIsRat is TRUE, then the control point
                            // is stored in HOMOGENEOUS form and is
                            //   [ CV(i)[0], ..., CV(i)[m_dim] ].

  int m_flag;               // A place to store user defined flags - ignored
                            // by IO toolkit read/write functions but useful
                            // for setting flags used in translation.

  RHINO_CURVE_TYPE m_type;  // A place to store curve type information
                            // Ignored by IO toolkit read/write functions.

private:
  void Initialize(void);  // zeros all fields
  CRhinoNurbsCurve(const CRhinoNurbsCurve&); // no implementation
	CRhinoNurbsCurve& operator=(const CRhinoNurbsCurve&); // no implementation
};

inline int CRhinoNurbsCurve::Dimension() const
{
  return m_dim;
}

inline BOOL CRhinoNurbsCurve::IsRational() const
{
  return m_bIsRat;
}

inline int CRhinoNurbsCurve::CVSize() const
{
  return (m_bIsRat) ? m_dim+1 : m_dim;
}

inline int CRhinoNurbsCurve::Order( void ) const
{
  return m_order;
}

inline int CRhinoNurbsCurve::Degree( void ) const
{
  return (m_order>=2) ? m_order-1 : 0;
}

inline int CRhinoNurbsCurve::CVCount( void ) const
{
  return m_cv_count;
}

inline int CRhinoNurbsCurve::KnotCount( void ) const
{
  return (m_cv_count>= m_order&&m_order>=2) ? m_cv_count+m_order-2 : 0;
}

inline const double* CRhinoNurbsCurve::CV( int i ) const
{
  return (m_cv) ? (m_cv + i*m_cv_stride) : NULL;
}

inline double* CRhinoNurbsCurve::CV( int i )
{
  return (m_cv) ? (m_cv + i*m_cv_stride) : NULL;
}

inline double CRhinoNurbsCurve::Knot( int i ) const
{
  return (m_knot) ? m_knot[i] : 0.0;
}

inline const double* CRhinoNurbsCurve::Knot() const
{
  return m_knot;
}

inline double CRhinoNurbsCurve::SuperfluousKnot( int end ) const
  { return(m_knot) ? RhinoSuperfluousKnot(m_order,m_cv_count,m_knot,end) : 0.0;}

inline RHINO_KNOT_STYLE CRhinoNurbsCurve::KnotStyle() const
{
  return RhinoKnotVectorStyle( m_order, m_cv_count, m_knot );
}

#endif
