/* $Header: /Src/RhinoIO/rhinoio_nurbssrf.cpp 18    11/10/99 2:56p Dalelear $ */
/* $NoKeywords: $ */
//
// Copyright (c) 1993-1998 Robert McNeel & Associates. All rights reserved.
// Rhinoceros is a registered trademark of Robert McNeel & Assoicates.
//
// THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
// ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
// MERCHANTABILITY ARE HEREBY DISCLAIMED.
//				
////////////////////////////////////////////////////////////////
 
#include "rhinoio_object.h"
#include "rhinoio_nurbssrf.h"

RHINO_CHUNK_IMPLEMENT(CRhinoNurbsSurface, TCODE_RHINO_OBJECT_NURBS_SURFACE)

// Initialize an empty CRhinoNurbsSurface
CRhinoNurbsSurface::CRhinoNurbsSurface() 
: CRhinoObject( TCODE_RHINO_OBJECT_NURBS_SURFACE ) 
{
  Initialize();
}

// Initialize an empty CRhinoNurbsSurface
CRhinoNurbsSurface::CRhinoNurbsSurface(
    int   dimension,    // dimension ( >= 1 and usually = 3 )
    BOOL  bIsRational,  // TRUE if surface is rational
    int   degree0,      // degree for first parameter ( >= 1 )
    int   degree1,      // degree for second parameter ( >= 1 )
    int   cv_count0,    // number of CVs for first parameter ( >= degree0+1 )
    int   cv_count1     // number of CVs for second parameter ( >= degree1+1 )
  ) : CRhinoObject( TCODE_RHINO_OBJECT_NURBS_SURFACE ) 
{
  Initialize();
  if ( dimension < 1 ||
       degree0 < 1 ||
       degree1 < 1 ||
       cv_count0 <= degree0 ||
       cv_count1 <= degree1 ) {
		throw CRhinoException( CRhinoException::illegal_input);
  }
  m_dim         = dimension;
  m_bIsRat      = (bIsRational) ? TRUE : FALSE;
  m_order[0]    = degree0+1;
  m_order[1]    = degree1+1;
  m_cv_count[0] = cv_count0;
  m_cv_count[1] = cv_count1;
  m_knot_capacity[0] = m_order[0] + m_cv_count[0] - 2;
  m_knot_capacity[1] = m_order[1] + m_cv_count[1] - 2;
  const int cvdim = (m_bIsRat) ? m_dim+1 : m_dim;
  m_cv_capacity = m_cv_count[0]*m_cv_count[1]*cvdim;
  m_cv_stride[1] = cvdim;
  m_cv_stride[0] = m_cv_count[1]*cvdim;

  // rhmalloc() is used instead of new[] so that the arrays can be
  // efficiently resized using rhrealloc().

  m_knot[0] = (double*)rhmalloc( m_knot_capacity[0]*sizeof(*m_knot[0]) );
  if ( !m_knot[0] ) {
		throw CRhinoException( CRhinoException::out_of_memory);
  }
  m_knot[0][0] = DBL_MAX;
  m_knot[0][m_knot_capacity[0]-1] = -DBL_MAX;
  m_knot[1] = (double*)rhmalloc( m_knot_capacity[1]*sizeof(*m_knot[1]) );
  if ( !m_knot[1] ) {
		throw CRhinoException( CRhinoException::out_of_memory);
  }
  m_knot[1][0] = DBL_MAX;
  m_knot[1][m_knot_capacity[1]-1] = -DBL_MAX;
  m_cv = (double*)rhmalloc( m_cv_capacity*sizeof(*m_cv) );
  if ( !m_cv ) {
		throw CRhinoException( CRhinoException::out_of_memory);
  }
  m_cv[0] = DBL_MAX;
  m_cv[m_cv_capacity-1] = -DBL_MAX;
}


CRhinoNurbsSurface::~CRhinoNurbsSurface()
{
  if ( m_cv_capacity > 0 && m_cv ) {
    rhfree( m_cv );
  }
	if ( m_knot_capacity[1] > 0 && m_knot[1] ) {
    rhfree( m_knot[1] );
  }
	if ( m_knot_capacity[0] > 0 && m_knot[0] ) {
    rhfree( m_knot[0] );
  }
}


BOOL
CRhinoNurbsSurface::IsClosed( int dir ) const // dir = 0 for "u", 1 for "v"
{
  BOOL bIsClosed = FALSE;
  if (dir >= 0 && dir <= 1 && m_dim > 0) {
    if ( RhinoIsKnotVectorClamped( m_order[dir], m_cv_count[dir], m_knot[dir] ) ) {
      // test CVs
      const int i0 = 0;
      const int i1 = m_cv_count[dir]-1;
      const int j0 = 0;
      const int j1 = m_cv_count[1-dir];
      int ii0, ii1, jj0, jj1, j;
      const double *cv0, *cv1;
      const int cvsize = CVSize();
      bIsClosed = TRUE;
      for ( j = j0; j < j1 && bIsClosed; j++ ) {
        if ( dir ) {
          jj0 = i0;
          jj1 = i1;
          ii0 = ii1 = j;
        }
        else {
          ii0 = i0;
          ii1 = i1;
          jj0 = jj1 = j;
        }
        cv0 = CV(ii0,jj0);
        cv1 = CV(ii1,jj1);
        bIsClosed = RhinoIsEqual( cvsize, cv0, cv1 );
      }
    }
    else {
      bIsClosed = IsPeriodic( dir );
    }
  }
  return bIsClosed;
}

BOOL
CRhinoNurbsSurface::IsPeriodic( int dir ) const // dir = 0 for "u", 1 for "v"
{
  BOOL bIsPeriodic = FALSE;
  if (dir >= 0 && dir <= 1 && m_dim > 0 ) {
    if ( RhinoIsKnotVectorPeriodic( m_order[dir], m_cv_count[dir], m_knot[dir] ) ) {
      // test CVs
      // test CVs
      const int degree = m_order[dir]-1;
      const int i0 = 0;
      const int i1 = m_cv_count[dir]-degree;
      const int j0 = 0;
      const int j1 = m_cv_count[1-dir];
      int ii0, ii1, jj0, jj1, i, j;
      const double *cv0, *cv1;
      const int cvsize = CVSize();
      bIsPeriodic = TRUE;
      for ( j = j0; j < j1 && bIsPeriodic; j++ ) {
        for ( i = 0; i < degree && bIsPeriodic; i++ ) {
          if ( dir ) {
            jj0 = i0+i;
            jj1 = i1+i;
            ii0 = ii1 = j;
          }
          else {
            ii0 = i0+i;
            ii1 = i1+i;
            jj0 = jj1 = j;
          }
          cv0 = CV(ii0,jj0);
          cv1 = CV(ii1,jj1);
          bIsPeriodic = RhinoIsEqual( cvsize, cv0, cv1 );
        }
      }
    }
  }
  return bIsPeriodic;
}

BOOL
CRhinoNurbsSurface::IsSingular( CRhinoNurbsSurface::eSurfaceSide side ) const
{
  BOOL bIsSingular = FALSE;
  if ( m_dim > 0 ) {
    switch( side ) {
    case south_side:
    case north_side:
      break;
    case east_side:
    case west_side:
      break;
    default:
      break;
    }    
  }
  return bIsSingular;
}

BOOL 
CRhinoNurbsSurface::SetCV( int i, int j, RHINO_POINT_STYLE style, const double* Point )
{
  int k;
  double w;

  // feeble but fast check for properly initialized class
  if ( !m_cv || i < 0 || j < 0 || i >= m_cv_count[0] || j >= m_cv_count[1] )
    return FALSE;

  double* cv = m_cv + i*m_cv_stride[0] + j*m_cv_stride[1];

  switch ( style ) {

  case rh_not_rational:  // input Point is not rational
    memcpy( cv, Point, m_dim*sizeof(*cv) );
    if ( IsRational() ) {
      // NURBS surface is rational - set weight to one
      cv[m_dim] = 1.0;
    }
    break;

  case rh_homogeneous_rational:  // input Point is homogeneous rational
    if ( IsRational() ) {
      // NURBS surface is rational
      memcpy( cv, Point, (m_dim+1)*sizeof(*cv) );
    }
    else {
      // NURBS surface is not rational
      w = (Point[m_dim] != 0.0) ? 1.0/Point[m_dim] : 1.0;
      for ( k = 0; k < m_dim; k++ ) {
        cv[k] = w*Point[k];
      }
    }
    break;

  case rh_euclidean_rational:  // input Point is euclidean rational
    if ( IsRational() ) {
      // NURBS surface is rational - convert euclean point to homogeneous form
      w = Point[m_dim];
      for ( k = 0; k < m_dim; k++ )
        cv[k] = w*Point[k];
      cv[m_dim] = w;
    }
    else {
      // NURBS surface is not rational
      memcpy( cv, Point, m_dim*sizeof(*cv) );
    }
    break;
  }
  return TRUE;
}

BOOL 
CRhinoNurbsSurface::GetCV( int i, int j, RHINO_POINT_STYLE style, double* Point ) const
{
  const double* cv = CV(i,j);
  if ( !cv )
    return FALSE;
  int dim = Dimension();
  double w = ( IsRational() ) ? cv[dim] : 1.0;
  switch(style) {
  case rh_euclidean_rational:
    Point[dim] = w;
    // no break here
  case rh_not_rational:
    if ( w == 0.0 )
      return FALSE;
    w = 1.0/w;
    while(dim--) *Point++ = *cv++ * w;
    break;
  case rh_homogeneous_rational:
    Point[dim] = w;
    memcpy( Point, cv, dim*sizeof(*Point) );
    break;
  default:
    return FALSE;
  }
  return TRUE;
}

BOOL 
CRhinoNurbsSurface::SetKnot( int parameter_index, int knot_index, double k )
{
  if ( parameter_index < 0 || parameter_index > 1 )
    return FALSE;
  if ( knot_index < 0 || knot_index >= KnotCount(parameter_index) )
    return FALSE;
  m_knot[parameter_index][knot_index] = k;
  return TRUE;
}

BOOL CRhinoNurbsSurface::IsValid() const
{
  // check fields for illegal values
  if ( m_dim < 1 )
    return FALSE;
  if ( m_order[0] < 2 )
    return FALSE;
  if ( m_order[1] < 2 )
    return FALSE;
  if ( m_cv_count[0] < m_order[0] )
    return FALSE;
  if ( m_cv_count[1] < m_order[1] )
    return FALSE;
  if ( !m_knot[0] )
    return FALSE;
  if ( !m_knot[1] )
    return FALSE;
  if ( !m_cv )
    return FALSE;

  int pi;
  for ( pi = 0; pi < 2; pi++ ) {
    const int d = m_order[pi]-1;
    const int k0 = m_order[pi]-2;
    const int k1 = m_cv_count[pi]-1;
    const int kcnt = m_order[pi]+m_cv_count[pi]-2;
    int i, j;
    // knots must be monotone increasing
    for ( i = 1; i < kcnt; i++ ) {
      if ( m_knot[pi][i-1] > m_knot[pi][i] )
        return FALSE;
    }
    // first span must be non-empty
    if ( m_knot[pi][k0] >= m_knot[pi][k0+1] )
      return FALSE;
    // last span must be non-empty
    if ( m_knot[pi][k1-1] >= m_knot[pi][k1] )
      return FALSE;
    // max knot multiplicity <= degree
    for ( i = k0, j = k0+d; j <= k1; i++,j++ ) {
      if ( m_knot[pi][i] >= m_knot[pi][j] )
        return FALSE;
    }
  }

  const int cvdim = (m_bIsRat) ? m_dim+1 : m_dim;

  // make sure cv strides have reasonable values
  if ( m_cv_stride[0] < cvdim )
    return FALSE;
  if ( m_cv_stride[1] < cvdim )
    return FALSE;
  if ( m_cv_stride[0] == m_cv_stride[1] )
    return FALSE;


  if ( m_cv_stride[1] < m_cv_stride[0] ) {
    if ( m_cv_stride[1]*m_cv_count[1] > m_cv_stride[0] )
      return FALSE;
  }
  else { // m_cv_stride[1] < m_cv_stride[0]
    if ( m_cv_stride[0]*m_cv_count[0] > m_cv_stride[1] )
      return FALSE;
  }

  // make sure cvs have been initialized.
  if ( m_cv[0] == DBL_MAX || 
       m_cv[ m_cv_count[0]*m_cv_count[1]*cvdim-1 ] == -DBL_MAX )
    return FALSE;

  return TRUE;
}


CRhinoFile::eStatus
CRhinoNurbsSurface::Read( CRhinoFile& file)
{
  CRhinoFile::eStatus status = CRhinoFile::not_supported;
  
  int data_typecode = 0;
  int data_length = 0;

  file.ReadInteger( 1, &data_typecode );
  if ( data_typecode != TCODE_RHINO_OBJECT_DATA )
    return CRhinoFile::crc_error;
  file.ReadInteger( 1, &data_length );
  if ( data_length <= 0 )
    return CRhinoFile::crc_error;

  int version = -1;

  file.ReadInteger( 1, &version );

  // int bReverse = version & 0x100;
  version &= 0xFFFFFEFF;
  
  switch ( version ) {

  case 100: // written by RhinoIO toolkit
  case 101: // written by Rhino
    {
      int i7[7];
      file.ReadInteger( 7, i7 );

      m_dim         = i7[0];
      m_bIsRat      = (i7[1]==1);
      m_order[0]    = i7[2];
      m_order[1]    = i7[3];
      m_cv_count[0] = i7[4];
      m_cv_count[1] = i7[5];

      if ( m_dim < 1  ||
           i7[1] < 0 || i7[1] > 1 ||
           m_order[0] < 2 ||
           m_order[1] < 2 ||
           m_cv_count[0] < m_order[0] ||
           m_cv_count[1] < m_order[1] ||
           i7[6] != 0 ) {
        status = CRhinoFile::crc_error;
      }
      else {
        int pi;
        for ( pi = 0; pi < 2; pi++ ) {
          m_knot_capacity[pi] = m_cv_count[pi]+m_order[pi]-2;
          m_knot[pi] = (double*)rhmalloc( m_knot_capacity[pi]*sizeof(*m_knot[pi]) );
          if ( !m_knot[pi] ) {
            throw CRhinoException( CRhinoException::out_of_memory );
          }
          file.ReadDouble( m_knot_capacity[pi], m_knot[pi] );
        }
        const int cvdim = ( m_bIsRat ) ? m_dim+1 : m_dim;
        m_cv_stride[0] = cvdim*m_cv_count[1];
        m_cv_stride[1] = cvdim;
        m_cv_capacity = cvdim*m_cv_count[0]*m_cv_count[1];
        m_cv = (double*)rhmalloc( m_cv_capacity*sizeof(*m_cv) );
        if ( !m_cv ) {
          throw CRhinoException( CRhinoException::out_of_memory );
        }
        file.ReadDouble( m_cv_capacity, m_cv );
        status = IsValid() ? CRhinoFile::good : CRhinoFile::crc_error;
      }
    }
    break;

  default:
    status = CRhinoFile::not_supported;
    file.SkipChunk();
    break;
  }

  // read attributes
  CRhinoFile::eStatus att_status = ReadAttributes( file, TCODE_RHINO_OBJECT_END );
  if ( status == CRhinoFile::good )
    status = att_status;

  return status;
}

void
CRhinoNurbsSurface::Write(CRhinoFile& file) const
{
  if ( !IsValid() ) {
    throw CRhinoException( CRhinoException::illegal_input );
  }

	// Make sure layer definition is in the file
	file.LayerIndex( Layer() );	// Get layer index or -1

	file.BeginWriteChunk( *this );
  {

    file.BeginWriteChunk( TCODE_RHINO_OBJECT_DATA );
    {
      // write NURBS surface data
      const int version = 100;
      int i8[8] = { version,
                   Dimension(),
                   (IsRational() ? 1 : 0),
                   Order(0),
                   Order(1),
                   CVCount(0),
                   CVCount(1),
                   0
      };

      file.WriteInteger( 8, i8 );

      file.WriteDouble( KnotCount(0), m_knot[0] );
      file.WriteDouble( KnotCount(1), m_knot[1] );

      const int imax = CVCount(0);
      const int jmax = CVCount(1);
      const int cvdim = CVSize();
      int i, j;
      for ( i = 0; i < imax; i++ ) {
        for ( j = 0; j < jmax; j++ ) {
          file.WriteDouble( cvdim, CV(i,j) );
        }
      }

    }
    file.EndChunk();

    WriteAttributes( file, TCODE_RHINO_OBJECT_END );
  }
  file.EndChunk();
}

void
CRhinoNurbsSurface::Initialize( )
{
  m_dim = 0;
  m_bIsRat = FALSE;
  m_order[0] = m_order[1] = 0;
  m_cv_count[0] = m_cv_count[1] = 0;

  m_knot_capacity[0] = m_knot_capacity[1] = 0;
  m_knot[0] = m_knot[1] = NULL;

  m_cv_stride[0] = m_cv_stride[1] = 0;
  m_cv_capacity = 0;
  m_cv = NULL;
  m_flag[0] = 0;
  m_flag[1] = 0;
  m_type = rh_unknown_srf;
}

BOOL
CRhinoNurbsSurface::GetBoundingBox(
          double box_min[3],
          double box_max[3],
          BOOL bGrowBox     // = FALSE
          ) const
{
  BOOL rc = RhinoBoundingBox( m_dim, m_bIsRat, 
                         m_cv_count[0]*m_cv_count[1], 
                         m_cv_stride[m_cv_stride[0] < m_cv_stride[1] ? 0 : 1],
                         m_cv,
                         box_min, box_max, bGrowBox );
  return rc;
}

BOOL
CRhinoNurbsSurface::Domain(
        int pi,      // parameter index 0 = "u", 1 = "v"
        double* t0,  // if not NULL, returns domain start
        double* t1   // if not NULL, returns domain end
        ) const
{
  if ( pi != 0 && pi != 1 )
    return FALSE;
  return RhinoGetKnotVectorDomain( m_order[pi], m_cv_count[pi], m_knot[pi], t0, t1 );
}

BOOL
CRhinoNurbsSurface::Reverse(   // reverse parameterizatrion
        int pi // parameter index 0 = "u", 1 = "v"
    )
{
  if ( pi != 0 && pi != 1 )
    return FALSE;
  BOOL rc = RhinoReverseKnotVector( m_order[pi], m_cv_count[pi], m_knot[pi] );
  const int cv_size = CVSize();
  int i;
  if ( pi ) {
    // reverse "v"
    for ( i = 0; i < m_cv_count[0]; i++ ) {
      rc &= RhinoReversePointList( cv_size, m_cv_count[1], m_cv_stride[1], CV(i,0) );
    }
  }
  else {
    // reverse "u"
    for ( i = 0; i < m_cv_count[1]; i++ ) {
      rc &= RhinoReversePointList( cv_size, m_cv_count[0], m_cv_stride[0], CV(0,i) );
    }
  }
  return rc;
}

BOOL
CRhinoNurbsSurface::SwapCoordinates(
        int i, int j        // indices of coords to swap
        )
{
  if ( i < 0 || j < 0 || i >= m_dim || j >= m_dim )
    return FALSE;
  const int cvcount = CVCount();
  if ( cvcount < 4 || Dimension() < 2 || CV(0,0) == NULL )
    return FALSE;
  if ( i == j )
    return TRUE;
  double t;
  int cvi, cvj;
  double* cv;
  const int cvi_count = CVCount(0);
  const int cvj_count = CVCount(1);
  for ( cvi = 0; cvi < cvi_count; cvi++ ) for (cvj = 0; cvj < cvj_count; cvj++ ) {
    cv = CV(cvi,cvj);
    t = cv[i]; 
    cv[i] = cv[j];
    cv[j] = t;
  }

  return TRUE;
}

BOOL
CRhinoNurbsSurface::SwapParameters() // swaps "u" and "v"
{
  int i;
  double* p;
  i = m_flag[0]; m_flag[0] = m_flag[1]; m_flag[1] = i;
  i = m_order[0]; m_order[0] = m_order[1]; m_order[1] = i;
  i = m_cv_count[0]; m_cv_count[0] = m_cv_count[1]; m_cv_count[1] = i;
  i = m_cv_stride[0]; m_cv_stride[0] = m_cv_stride[1]; m_cv_stride[1] = i;
  i = m_knot_capacity[0]; m_knot_capacity[0] = m_knot_capacity[1]; m_knot_capacity[1] = i;
  p = m_knot[0]; m_knot[0] = m_knot[1]; m_knot[1] = p;
  return TRUE;
}

