/* $Header: /Src/RhinoIO/rhinoio_nurbssrf.h 16    11/10/99 2:56p Dalelear $ */
/* $NoKeywords: $ */
//
// Copyright (c) 1993-1998 Robert McNeel & Associates. All rights reserved.
// Rhinoceros is a registered trademark of Robert McNeel & Assoicates.
//
// THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
// ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
// MERCHANTABILITY ARE HEREBY DISCLAIMED.
//				
////////////////////////////////////////////////////////////////
//
//   defines CRhinoNurbsSurface
//
////////////////////////////////////////////////////////////////

#if !defined(RHINO_NURBSSRF_INC_)
#define RHINO_NURBSSRF_INC_

///////////////////////////////////////////////////////////////////////////////
// Class  CRhinoNurbsSurface
//
//	This object represents a tensor product NURBS surface.
///////////////////////////////////////////////////////////////////////////////

// If (X,Y,Z) is the 3D location of a NURBS control point, it may be
// represented in at least three different ways.
//
// Not rational:         ( X, Y, Z )
// Homogeneous rational: ( w*X, w*Y, w*Z, w )  w != 0
// Euclidean rational:   ( X, Y, Z, w )
//
// Internally, non-rational CRhinoNurbsSurface classes use the not rational form
// and rational CRhinoNurbsSurface classes use the homogeneous rational form.
// The SetCV() and GetCV() functions have a RHINO_POINT_STYLE argument that
// allows you to specify which form you want to use.

class CRhinoNurbsSurface: public CRhinoObject
{
	RHINO_CHUNK_DECLARE( CRhinoNurbsSurface )
public:
  enum eSurfaceSide {
    south_side = 1,
    east_side  = 2,
    north_side = 4,
    west_side  = 8
  };

  // Construction
	CRhinoNurbsSurface();
	CRhinoNurbsSurface(   // Constructs a NURBS surface that can be filled in
                        // with calls to SetCV() and SetKnot().
                        // Throws a CRhinoException if input is illegal
                        // or not enough memory.
    int,  // dimension ( >= 1 and usually = 3 )
    BOOL, // TRUE if surface is rational
    int,  // degree0 = degree for first parameter ( >= 1 )
    int,  // degree1 = degree for second parameter ( >= 1 )
    int,  // cv_count0 = number of CVs for first parameter ( >= degree0+1 )
    int   // cv_count1 = number of CVs for second parameter ( >= degree1+1 )

  );
  ~CRhinoNurbsSurface();

  // CRhinoObject overrides
	CRhinoFile::eStatus Read(   	// read chunk data from file 
    CRhinoFile&
    );
	void Write(                   // Writes NURBS surface to file
                                // Throws a CRhinoException if IsValid()
                                // returns FALSE or fwrite() fails.
    CRhinoFile&
    ) const;
  BOOL IsValid(                 // Call after class is fully initialized to
                                // perform a simple and quick validity check.
                                // Will detect illegal knot vectors and
                                // incorrect settings of integer m_* members.
                                // Does NOT check control vertices for
                                // grossly degenerate conditions.
    void
    ) const;
  BOOL GetBoundingBox(
          double[3], // min
          double[3], // max
          BOOL = FALSE // set to TRUE if box should be grown
          ) const;


  // Interface
  int Dimension(    // dimension of tensor product NURBS surface (usually 3)
        void
        ) const;

  BOOL IsRational(  // TRUE if tensor product NURBS is rational
        void
        ) const;
  
  BOOL IsClosed(  // TRUE if NURBS surface is closed (either curve has
                  // clamped end knots and euclidean location of start
                  // CV = euclidean location of end CV, or surface is
                  // periodic.)
        int       // parameter index 0 = "u", 1 = "v"
        ) const;

  BOOL IsPeriodic(  // TRUE if NURBS surface is periodic (degree > 1,
                    // periodic knot vector, last degree many CVs 
                    // are duplicates of first degree many CVs.)
        int         // parameter index 0 = "u", 1 = "v"
        ) const;

  BOOL IsSingular(    // TRUE if NURBS is singular along specified side
         eSurfaceSide
         ) const;
  
  int CVSize(       // number of doubles per control vertex 
        void        // = IsRational() ? Dim()+1 : Dim()
        ) const;
  
  int Order(        // order = degree + 1
        int  // parameter index 0 = "u", 1 = "v"
        ) const;
	
  int Degree( 
        int  // parameter index 0 = "u", 1 = "v"
        ) const;

  int CVCount(      // number of control vertices in parameter direction
        int  // parameter index 0 = "u", 1 = "v"
        ) const;

  int CVCount(      // total number of control vertices
        void 
        ) const;

  int KnotCount(    // total number of knots in knot vector (Order+CVCount-2)
        int  // parameter index 0 = "u", 1 = "v"
        ) 
        const;
  
  const double* CV(     // get pointer to control vertex
        int,            // first parameter index ( >= 0 and < Count(0) )
        int             // second parameter index ( >= 0 and < Count(1) )
        ) const;

  double* CV(           // get pointer to control vertex
        int,            // first parameter index ( >= 0 and < Count(0) )
        int             // second parameter index ( >= 0 and < Count(1) )
        );

  int SetCV(            // set a single control vertex
        int,            // first parameter index ( >= 0 and < Count(0) )
        int,            // second parameter index ( >= 0 and < Count(1) )
        RHINO_POINT_STYLE, // style of input point
        const double*   // value of control vertex
        );

  int GetCV(            // get a single control vertex
        int,            // first parameter index ( >= 0 and < Count(0) )
        int,            // second parameter index ( >= 0 and < Count(1) )
        RHINO_POINT_STYLE, // style to use for output point
        double*         // array of length >= CVSize()
        ) const;

  int SetKnot(
        int,            // parameter index 0 = "u", 1 = "v"
        int,            // knot index ( 0 to KnotCount - 1 )
        double          // value for knot
        );

  double Knot(
        int,            // parameter index 0 = "u", 1 = "v"
        int             // knot index ( >= 0 and < Order + CV_count - 2 )
        ) const;

  const double* Knot(   // knot[] array
        int             // parameter index 0 = "u", 1 = "v"
        ) const;

  double SuperfluousKnot(
           int, // parameter index 0 = "u", 1 = "v"
           int  // 0 = start, 1 = end
           ) const;

  RHINO_KNOT_STYLE KnotStyle(
           int // parameter index 0 = "u", 1 = "v"
           ) const;

  BOOL Domain(
        int,      // parameter index 0 = "u", 1 = "v"
        double*,  // if not NULL, returns domain start
        double*   // if not NULL, returns domain end
        ) const;

  BOOL Reverse(   // reverse parameterizatrion
        int       // parameter index 0 = "u", 1 = "v"
    );    

  BOOL SwapCoordinates(
        int, int        // indices of coords to swap
        );

  BOOL SwapParameters(); // swaps "u" and "v"

// Implementation
// protected:
public:

  // NOTE: These fields are left "public" so that expert users may efficiently
  //       create NURBS surfaces using the default constructor and borrow the
  //       knot and CV arrays from their native NURBS representation.
  //       No technical support will be provided for users who access these
  //       fields directly.  If you can't get your stuff to work, then use
  //       the constructor with the arguments and the SetKnot() and SetCV()
  //       functions to fill in the arrays.

  int     m_dim;            // Must be 3 for exposed Rhino IO uses.
                            // (Can be anything when a 2 variable B-spline
                            //  function is needed internally.)
  BOOL    m_bIsRat;         // TRUE  for rational B-splines.  Control vertices
                            //       use homogeneous form.
                            // FALSE for non-rational B-splines.  Control
                            //       verticies do not have a weight coordinate.
  int     m_order[2];       // order = degree+1
  int     m_cv_count[2];    // size of control vertex net

  // knot vector memory
  int     m_knot_capacity[2];
                            // If m_knot_capacity[i] > 0, then m_knot[i][]
                            // is an array of at least m_knot_capacity[i]
                            // doubles whose memory is managed by the
                            // CRhinoNurbsSurface class using rhmalloc(),
                            // rhrealloc(), and rhfree().
                            // If m_knot_capacity[i] is 0 and m_knot[i] is
                            // not NULL, then  m_knot[i][] is assumed to
                            // be big enough for any requested operation
                            // and m_knot[i][] is not deleted by the
                            // destructor.
  double* m_knot[2];        // Knot vectors. ( The knot vectors have length
                            // m_order[i]+m_cv_count[i]-2. )
  
  // control vertex net memory
  int     m_cv_stride[2];   // The pointer to start of "CV[i][j]" is
                            // m_cv + i*m_cv_stride[0] + j*m_cv_stride[1].
  int     m_cv_capacity;    // If m_cv_capacity > 0, then m_cv[] is an array
                            // of at least m_cv_capacity doubles whose
                            // memory is managed by the CRhinoNurbsSurface
                            // class using rhmalloc(), rhrealloc(), and rhfree().
                            // If m_cv_capacity is 0 and m_cv is not
                            // NULL, then m_cv[] is assumed to be big enough
                            // for any requested operation and m_cv[] is not
                            // deleted by the destructor.
  double* m_cv;             // Control points.
                            // If m_bIsRat is FALSE, then control point is
                            //   ( CV(i)(j)[0], ..., CV(i)(j)[m_dim-1] ).
                            // If m_bIsRat is TRUE, then the control point
                            // is stored in HOMOGENEOUS form and is
                            //   [ CV(i)(j)[0], ..., CV(i)(j)[m_dim] ].
                            // 
  
  int m_flag[2];            // A place to store user defined flags - ignored
                            // by IO toolkit read/write functions but useful
                            // for setting flags used in translation.

  RHINO_SURFACE_TYPE m_type; // A place to store surface type information
                             // Ignored by IO toolkit read/write functions.

private:
  void Initialize();         // initializes all fields
  CRhinoNurbsSurface(const CRhinoNurbsSurface&); // no implementation
	CRhinoNurbsSurface& operator=(const CRhinoNurbsSurface&); // no implementation

};

inline int CRhinoNurbsSurface::Dimension() const
{
  return m_dim;
}

inline BOOL CRhinoNurbsSurface::IsRational() const
{
  return m_bIsRat;
}

inline int CRhinoNurbsSurface::CVSize() const
{
  return (m_bIsRat) ? m_dim+1 : m_dim;
}

inline int CRhinoNurbsSurface::Order( int i ) const
{
  return (i) ? ( (i==1) ? m_order[1] : 0 ) : m_order[0];
}

inline int CRhinoNurbsSurface::Degree( int i ) const
{
  int order = Order(i);
  return ( order >= 2 ) ? order-1 : 0;
}

inline int CRhinoNurbsSurface::CVCount( int i ) const
{
  return (i) ? ( (i==1) ? m_cv_count[1] : 0 ) : m_cv_count[0];
}

inline int CRhinoNurbsSurface::CVCount( ) const
{
  return m_cv_count[0]*m_cv_count[1];
}

inline int CRhinoNurbsSurface::KnotCount( int i ) const
{
  return (i) ? ( (i==1) ? (m_order[1]+m_cv_count[1]-2) : 0 ) : (m_order[0]+m_cv_count[0]-2);
}

inline const double* CRhinoNurbsSurface::CV( int i, int j ) const
{
  return m_cv + i*m_cv_stride[0] + j*m_cv_stride[1];
}

inline double* CRhinoNurbsSurface::CV( int i, int j )
{
  return m_cv + i*m_cv_stride[0] + j*m_cv_stride[1];
}

inline double CRhinoNurbsSurface::Knot( int i, int k ) const
{
  return m_knot[i][k];
}

inline const double* CRhinoNurbsSurface::Knot( int i ) const
{
  return m_knot[i];
}

inline double CRhinoNurbsSurface::SuperfluousKnot( int i, int end ) const
  { return(m_knot[i]) ? RhinoSuperfluousKnot(m_order[i],m_cv_count[i],m_knot[i],end) : 0.0;}

inline RHINO_KNOT_STYLE CRhinoNurbsSurface::KnotStyle( int i ) const
{
  return RhinoKnotVectorStyle( m_order[i], m_cv_count[i], m_knot[i] );
}

#endif
