/* $Header: /Src/RhinoIO/rhinoio_object.cpp 13    1/07/00 9:34a Dalelear $ */
/* $NoKeywords: $ */
//
// Copyright (c) 1993-1998 Robert McNeel & Associates. All rights reserved.
// Rhinoceros is a registered trademark of Robert McNeel & Assoicates.
//
// THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
// ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
// MERCHANTABILITY ARE HEREBY DISCLAIMED.
//				
////////////////////////////////////////////////////////////////

#include "rhinoio_object.h"


////////////////////////////////////////////////////////////////
//   Class CRhinoObject
////////////////////////////////////////////////////////////////

CRhinoObject& CRhinoObject::operator=( const CRhinoObject& src )
{
  if ( &src != this ) {
    if ( m_pMaterial ) {
	    delete m_pMaterial;
	    m_pMaterial = NULL;
    }
    if ( m_pName ) {
      delete m_pName;
      m_pName = NULL;
    }
    m_LayPnt = src.m_LayPnt;
    m_wire_color = src.m_wire_color;
    {
      const CRhinoMaterial* pMat = src.Material();
      if ( pMat )
        SetMaterial( *pMat );
    }
    if ( src.m_pName ) {
      m_pName = new CRhinoName( src.m_pName->Name() );
    }
    m_state = src.m_state;
    if ( m_sXdataLayerName ) {
      rhfree( m_sXdataLayerName );
      m_sXdataLayerName = 0;
    }
    if ( src.m_sXdataLayerName ) {
      m_sXdataLayerName = rhstrdup( src.m_sXdataLayerName );
    }
  }
  return *this;
}

CRhinoColor CRhinoObject::WireColor( void ) const
{
  CRhinoColor c( m_wire_color );
  if ( m_wire_color == 0xCC000000 && Layer() ) {
    c = Layer()->Color();
  }
  return c;
}

void CRhinoObject::SetWireColor( const CRhinoColor& color )
{
  m_wire_color = color;
}

void CRhinoObject::SetWireColorByLayer(void)
{
  m_wire_color = 0xCC000000;
}

void CRhinoObject::SetLabel( const char* sLabel )
{
  if ( !m_pName ) {
    m_pName = new CRhinoName( sLabel );
  }
  else 
    m_pName->SetName( sLabel );
}

void
CRhinoObject::SetLayer( int layer_index, const CRhinoFile& file){
	if( layer_index >= 0 && layer_index < file.NumLayers() ) {
		CRhinoLayer* pLayer =  file.Layer(layer_index) ;
    if ( pLayer ) {
		  SetLayer(pLayer);
      if ( pLayer->Name() ) {
        if ( !Rhino_stricmp( pLayer->Name(), "Hidden" ) ) {
          m_state = hidden;
        }
        else if ( !Rhino_stricmp( pLayer->Name(), "Locked" ) ) {
          m_state = locked;
        }
      }
    }
	}
}

void CRhinoObject::WriteAttributes( 
                     CRhinoFile& file, 
                     const int end_mark_typecode ) const
{
  int layer_index = file.LayerIndex( Layer() );
  WriteLayerIndex( file, layer_index );

  if ( m_pName ) {
    m_pName->Write( file );
  }

  if ( m_wire_color != 0xCC000000 ) {
    file.BeginWriteChunk( TCODE_RGBDISPLAY, m_wire_color );
    file.EndChunk();
  }
  
  WriteMaterial( file, Material() );

  if ( end_mark_typecode != 0 ) {
    file.BeginWriteChunk( end_mark_typecode );
    file.EndChunk();
  }
}

void CRhinoObject::SetLayerAndState( CRhinoFile& file, const CRhinoXdata* px )
{
  if ( px && px->TypeCode() == TCODE_XDATA ) {
    const char* sTrueLayerName = 0;
    int nTrueLayerNameLength = 0;
    const char* sXdataId = px->Id();
    if ( !Rhino_stricmp ("RhHidePrevLayer", sXdataId ) ) {
      nTrueLayerNameLength = px->BufferSize();
      sTrueLayerName = (const char*)px->BufferData();
      m_state = hidden;
    }
    else if ( !Rhino_stricmp ("RhFreezePrevLayer", sXdataId ) ) {
      nTrueLayerNameLength = px->BufferSize();
      sTrueLayerName = (const char*)px->BufferData();
      m_state = locked;
    }
    if ( nTrueLayerNameLength > 0 && sTrueLayerName && sTrueLayerName[0] ) {
      int layer_index = 0;
      const int layer_count = file.NumLayers();
      m_sXdataLayerName = (char*)rhmalloc(nTrueLayerNameLength+1);
      memcpy( m_sXdataLayerName, sTrueLayerName, nTrueLayerNameLength );
      m_sXdataLayerName[nTrueLayerNameLength] = 0;
      for ( layer_index = 0; layer_index < layer_count; layer_index++ ) {
        const CRhinoLayer* pLayer = file.Layer( layer_index );
        if ( pLayer && pLayer->Name() && !Rhino_stricmp( pLayer->Name(), m_sXdataLayerName ) ) {
          SetLayer( layer_index, file );
          break;
        }
      }
    }
  }
}

CRhinoFile::eStatus
CRhinoObject::ReadAttributes( CRhinoFile& file, const int end_mark_typecode )
{
  CRhinoFile::eStatus status = CRhinoFile::good;
	CRhinoChunk* pChunk = NULL;
  BOOL bHasMaterial = FALSE;
  CRhinoMaterial material;

  // AG objects are written in an old way that requires this hack
  BOOL bTypeCodePeek = ( (end_mark_typecode & (TCODE_AG | TCODE_STUFF)) == 
                         (TCODE_AG | TCODE_STUFF) );

  for (;;) {

    if ( bTypeCodePeek ) {
      // terminating chunk is long and must be read carefully
      FILE* pFile = file.GetFile();
	    int c = fgetc( pFile );
	    if( c == EOF ){
     		throw CRhinoException(CRhinoException::unexpected_end_of_file);
	    }
	    c = ungetc( c, pFile );
	    if( c == EOF ) {
		    throw CRhinoException(CRhinoException::unable_to_seek);
      }
      int typecode = 0;
	    file.ReadInteger( 1, &typecode );			// read the type code
      if ( fseek( pFile, -((int)sizeof(typecode)), SEEK_CUR ) ) {
		    throw CRhinoException(CRhinoException::unable_to_seek);
      }
      if ( typecode == end_mark_typecode )
        break;
    }

  	pChunk = file.ReadChunk( status );

    if ( status != CRhinoFile::good )
      break;
	  if ( pChunk->TypeCode() == end_mark_typecode )
      break;

		switch( pChunk->TypeCode() ) {

		case TCODE_LAYERREF:
			SetLayer(pChunk->Value(), file);
			break;

		case TCODE_XDATA:
      // look for "real" layer name on hidden or locked objects
      SetLayerAndState( file, (CRhinoXdata*)(pChunk) );
			break;

    case TCODE_RGBDISPLAY: // wire frame color
      m_wire_color = pChunk->Value();
      break;

    case TCODE_NAME:       // label
      {
			  CRhinoName* pName = (CRhinoName*)(pChunk);
        SetLabel( pName->Name() );
      }
      break;

		case TCODE_RGB:
      {
			  bHasMaterial = TRUE;
			  CRhinoRGB* prgb = (CRhinoRGB*) (pChunk);
			  material.SetAmbient(prgb->Color());
			  material.SetDiffuse(prgb->Color());
			  material.SetShine(prgb->Shine());
			}
			break;

		case TCODE_TRANSPARENCY:
      {
			  bHasMaterial = TRUE;
			  int transparency = pChunk->Value(); // ( 0 to 255 )
			  if( transparency < 0 )
          transparency = 0; 
        else if ( transparency > 255 ) 
          transparency = 255;
			  material.SetTransparency( transparency/255.0 );
			}
			break;

		case TCODE_TEXTUREMAP:
      {
			  bHasMaterial = TRUE;
			  CRhinoTextureMap* pTex = (CRhinoTextureMap*)(pChunk);
			  material.SetTextureBitmapFileName( pTex->FileName() );
			}
			break;

		case TCODE_BUMPMAP: 
      {
			  bHasMaterial = TRUE;
			  CRhinoBumpMap* pBump = (CRhinoBumpMap*)(pChunk);
			  material.SetBumpBitmapFileName( pBump->FileName() );
			}
			break;

		case TCODE_RENDER_MATERIAL_ID: 
      {
			  bHasMaterial = TRUE;
			  CRhinoRenderMaterialName* pRMname = (CRhinoRenderMaterialName*)(pChunk);
			  material.SetRenderMaterialName( pRMname->Name() );
			}
			break;

		}

		delete pChunk;
    pChunk = NULL;
	}

  if ( pChunk ) 
	  delete pChunk;

  if ( bHasMaterial )
    SetMaterial( material );

	if ( status == CRhinoFile::eof)
		throw CRhinoException(CRhinoException::unexpected_end_of_file);
  
  return status;
}

// Write a LayerRef if index is in table. 
void 
CRhinoObject::WriteLayerIndex(CRhinoFile& file,int index) const
{
	if(index>=0 && index<file.NumLayers() ){
		file.BeginWriteChunk( TCODE_LAYERREF, index );
		file.EndChunk();
	}
	return;
}


// Make a copy of the CRhinoMaterial 
void CRhinoObject::SetMaterial(const CRhinoMaterial& mat)
{
	delete m_pMaterial;
	m_pMaterial = new CRhinoMaterial(mat);
}

// Write material if not NULL.  Writes RGB,and Transparency chunks and if required
// writes texture map and bump map chunks.
void 
CRhinoObject::WriteMaterial(CRhinoFile& file, const CRhinoMaterial* pMat) const
{
	if(pMat==NULL)
		return;

	// Take the diffuse color as the color to put in the 3dm file
	CRhinoRGB rgb(pMat->Diffuse(), (int) pMat->Shine() );
	rgb.Write(file);

	CRhinoTransparency tran( int(255 * pMat->Transparency()));
	tran.Write(file);

	if( pMat->TextureBitmapFileName() ){
		CRhinoTextureMap Tmap;
		Tmap.SetFileName ( pMat->TextureBitmapFileName() );
		Tmap.Write(file);
	}

	if( pMat->BumpBitmapFileName() ){
		CRhinoBumpMap Bmap;
		Bmap.SetFileName ( pMat->BumpBitmapFileName() );
		Bmap.Write(file);
	}

	if( pMat->RenderMaterialName() ){
		CRhinoRenderMaterialName RMname;
		RMname.SetName( pMat->RenderMaterialName() );
		RMname.Write(file);
	}
}	

BOOL 
CRhinoObject::GetBoundingBox( double[3], double[3], BOOL /* = FALSE */ ) const
{
  return FALSE;
}

CRhinoObject::CRhinoObject( int typecode, int i )
: CRhinoChunk( typecode, i )
{
  m_pMaterial = NULL;
  m_pName = NULL;
  m_wire_color = 0xCC000000;
  m_state = CRhinoObject::normal;
  m_sXdataLayerName = 0;
}

CRhinoObject::~CRhinoObject()
{
  delete m_pName;
  delete m_pMaterial;
  if ( m_sXdataLayerName ) {
    rhfree( m_sXdataLayerName );
  }
}

BOOL CRhinoObject::IsObject( void ) const
{
  return TRUE;
}

const char* CRhinoObject::ClassName( void ) const
{
  return "CRhinoObject";
}

const char* CRhinoObject::Label( void ) const
{
  return (m_pName) ? m_pName->Name() : NULL;
}

CRhinoLayer* CRhinoObject::Layer() const
{
  return m_LayPnt;
}  

void CRhinoObject::SetLayer( CRhinoLayer* pLayer )			
{
  m_LayPnt = CRhinoLayerPtr( pLayer);
}

CRhinoObject::eObjectState CRhinoObject::State() const
{
  return m_state;
}  

CRhinoMaterial* CRhinoObject::Material() const
{
  return m_pMaterial;
}  
