/* $Header: /Src/RhinoIO/rhinoio_pointset.cpp 5     10/20/99 1:28p Dalelear $ */
/* $NoKeywords: $ */
//
// Copyright (c) 1993-1998 Robert McNeel & Associates. All rights reserved.
// Rhinoceros is a registered trademark of Robert McNeel & Assoicates.
//
// THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
// ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
// MERCHANTABILITY ARE HEREBY DISCLAIMED.
//				
////////////////////////////////////////////////////////////////
 
#include "rhinoio_object.h"
#include "rhinoio_pointset.h"

RHINO_CHUNK_IMPLEMENT(CRhinoPointSet, TCODE_RH_POINT)

// Initialize an empty CRhinoPointSet
CRhinoPointSet::CRhinoPointSet() : CRhinoObject( TCODE_RH_POINT ) {
  m_count = 0;
  m_stride = 3;
  m_capacity = 0;
  m_P = NULL;
}

CRhinoPointSet::~CRhinoPointSet()
{
  if ( m_P && m_capacity > 0 ) {
    rhfree( m_P );
  }
}

void CRhinoPointSet::SetCapacity( int capacity )
{
  if ( capacity < 0 ) {
    throw CRhinoException( CRhinoException::illegal_input );
  }

  if ( capacity == 0 ) {
    if ( m_P && m_capacity > 0 ) 
      rhfree( m_P );
    m_count = 0;
    m_capacity = 0;
    m_P = NULL;
  }
  else if ( m_P && m_capacity > 0 ) {
    // grow/shrink m_P
    if ( capacity != m_capacity ) {
      m_P = (double*)rhrealloc( m_P, capacity*m_stride*sizeof(*m_P) );
      if ( !m_P ) {
        throw CRhinoException( CRhinoException::out_of_memory );
      }
    }
    else {
      rhfree( m_P );
      m_P = NULL;
    }
    m_capacity = capacity;
    if ( m_count > capacity )
      m_count = capacity;
  }
  else if ( m_P ) {
    // copy values from borrowed array to m_P
    double* pBorrowed = m_P;
    m_P = (double*)rhmalloc( capacity*m_stride*sizeof(*m_P) );
    if ( !m_P ) {
      throw CRhinoException( CRhinoException::out_of_memory );
    }
    if ( m_count > 0 ) {
      if ( m_count > capacity )
        m_count = capacity;
      memcpy( (void*)m_P, (void*)pBorrowed, m_count*m_stride*sizeof(*m_P) );
    }
    m_capacity = capacity;
  }
  else {
    // allocate m_P
    m_P = (double*)rhmalloc( capacity*m_stride*sizeof(*m_P) );
    m_capacity = capacity;
    m_count = 0;
  }
}

void CRhinoPointSet::SetPointCount( int count )
{
  if ( count != m_count ) {
    if ( !m_P || (m_capacity > 0 && count > m_capacity) ) {
      SetCapacity( count );
    }
    m_count = count;
  }
}

BOOL CRhinoPointSet::SetPoint( int point_index, double x, double y, double z )
{
  if ( point_index < 0 )
    return FALSE;

  if ( point_index >= m_capacity ) {
    int new_capacity;
    if ( point_index == 0 ) {
      new_capacity = 1;
    }
    else if ( point_index < 4 ) {
      new_capacity = 8;
    }
    else {
      new_capacity = 2*point_index;
    }
    SetCapacity( new_capacity );
  }
  double* P = m_P + point_index*m_stride;
  P[0] = x;
  P[1] = y;
  P[2] = z;
  if ( point_index <=  m_count )
    m_count = point_index+1;

  return TRUE;
}


BOOL CRhinoPointSet::GetPoint ( int point_index, double& x, double& y, double& z ) const
{
  const double* P = Point(point_index);
  if ( !P )
    return FALSE;
  x = P[0];
  y = P[1];
  z = P[2];
  return TRUE;
}

BOOL CRhinoPointSet::IsValid( void ) const
{
  return ( m_P != NULL && m_count > 0 && m_stride >= 3 );
}

CRhinoFile::eStatus
CRhinoPointSet::Read( CRhinoFile& file)
{
	SetPointCount(1); // 3dm file currently only has single points
	file.ReadDouble(3, m_P );
	
	// read attributes
  CRhinoFile::eStatus status = ReadAttributes( file, TCODE_ENDOFTABLE );

	return status;
}



void
CRhinoPointSet::Write(CRhinoFile& file) const
{
	// Make sure layer definition is in the file
	file.LayerIndex( Layer() );	// Get layer index or -1

	// for each point in the pointset write out a chunk
  const int point_count = PointCount();
	int point_index;
	for( point_index = 0; point_index < point_count; point_index++ ) {
		file.BeginWriteChunk( *this );
    {
			file.WriteDouble(3, Point( point_index ) );
      WriteAttributes( file, TCODE_ENDOFTABLE );
    }
		file.EndChunk();
	}
	return;
}

BOOL
CRhinoPointSet::GetBoundingBox(
          double box_min[3],
          double box_max[3],
          BOOL bGrowBox     // = FALSE
          ) const
{
  return RhinoBoundingBox( 3, FALSE, m_count, m_stride, m_P, box_min, box_max, bGrowBox );
}
