/* $Header: /Src/RhinoIO/rhinoio_spotlight.cpp 4     10/15/98 1:20p Dalelear $ */
/* $NoKeywords: $ */
//
// Copyright (c) 1993-1998 Robert McNeel & Associates. All rights reserved.
// Rhinoceros is a registered trademark of Robert McNeel & Assoicates.
//
// THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
// ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
// MERCHANTABILITY ARE HEREBY DISCLAIMED.
//				
////////////////////////////////////////////////////////////////
 
#include "rhinoio_object.h"
#include "rhinoio_spotlight.h"

RHINO_CHUNK_IMPLEMENT(CRhinoSpotlight, TCODE_RH_SPOTLIGHT)

#define DEGREES_TO_RADIANS RHINO_PI/180.0

// Initialize an empty CRhinoSpotlight
CRhinoSpotlight::CRhinoSpotlight() : CRhinoObject( TCODE_RH_SPOTLIGHT ),
                                     m_spot_exponent(0.0),
                                     m_spot_cutoff(45.0),
                                     m_ambient(0,0,0),
                                     m_diffuse(255,255,255), 
                                     m_specular(255,255,255),
                                     m_r(1.0),
                                     m_h(1.0)
{
  m_Position[0] =  0.0;
  m_Position[1] =  0.0;
  m_Position[2] =  0.0;
  m_SpotDirection[0] =  0.0;
  m_SpotDirection[1] =  0.0;
  m_SpotDirection[2] = -1.0;
  CRhinoMaterial mat;
  mat.SetAmbient( m_ambient );
  mat.SetDiffuse( m_diffuse );
  mat.SetSpecular( m_specular );
  SetMaterial(mat);
}

CRhinoSpotlight::~CRhinoSpotlight()
{
}

BOOL CRhinoSpotlight::IsValid( void ) const
{
  if ( m_SpotDirection[0] == 0.0 && m_SpotDirection[1] == 0.0 && m_SpotDirection[2] == 0.0 )
    return FALSE;
  if ( m_spot_exponent < 0.0 || m_spot_exponent > 1.0 )
    return FALSE;
  if ( m_spot_cutoff < 0.0 || m_spot_cutoff > 90 ) {
    if ( m_spot_cutoff != 180.0 )
      return FALSE;
  }
  return TRUE;
}

CRhinoFile::eStatus
CRhinoSpotlight::Read( CRhinoFile& file)
{
  double d[12], a;
  file.ReadDouble( 12, d );

  m_Position[0] = d[0];
  m_Position[1] = d[1];
  m_Position[2] = d[2];

  RhinoCrossProduct( d+3, d+6, m_SpotDirection );
  a = RhinoVectorLength( 3, m_SpotDirection );
  if ( a > 0.0 ) {
    a = -1.0/a;
    m_SpotDirection[0] *= a;
    m_SpotDirection[1] *= a;
    m_SpotDirection[2] *= a;
  }
  else {
    m_SpotDirection[0] =  0.0;
    m_SpotDirection[1] =  0.0;
    m_SpotDirection[2] = -1.0;
  }

  m_r = ( d[9]  > 0.0 ) ? d[9]  : 1.0;
  m_h = ( d[10] > 0.0 ) ? d[10] : 1.0;

  m_Position[0] -= m_SpotDirection[0]*m_h;
  m_Position[1] -= m_SpotDirection[1]*m_h;
  m_Position[2] -= m_SpotDirection[2]*m_h;

  if ( d[9] > 0.0 && d[10] > 0.0 ) {
    m_spot_cutoff = atan2( d[9], d[10] );
    m_spot_cutoff /= DEGREES_TO_RADIANS;
  }
  else {
    m_spot_cutoff = 180.0;
  }
  m_spot_exponent = (d[11] >= 0.0 && d[11] <= 1.0) ? 1.0-d[11] : 0.0;

  CRhinoFile::eStatus status = ReadAttributes( file, TCODE_ENDOFTABLE );

  m_ambient.SetRGB(0,0,0);
  m_diffuse.SetRGB(255,255,255);
  m_specular.SetRGB(255,255,255);
  CRhinoMaterial* pMat = Material();
  if ( !pMat ) {
    CRhinoMaterial mat;
    mat.SetDiffuse(  CRhinoColor(255,255,255) );
    SetMaterial( mat );
    pMat = Material();
  }
  if ( pMat ) {
    pMat->SetAmbient( m_ambient );
    pMat->SetSpecular( m_specular );
    m_diffuse = pMat->Diffuse();
  }

	return status;
}

void
CRhinoSpotlight::Write(CRhinoFile& file) const
{
  double d[12];
  d[0]  =  m_Position[0] + m_h*m_SpotDirection[0];
  d[1]  =  m_Position[1] + m_h*m_SpotDirection[1];
  d[2]  =  m_Position[2] + m_h*m_SpotDirection[2];
  d[3]  =  1.0;
  d[4]  =  0.0;
  d[5]  =  0.0;
  d[6]  =  0.0;
  d[7]  =  1.0;
  d[8]  =  0.0;
  d[9]  =  m_r;
  d[10] =  m_h;
  d[11] =  1.0-m_spot_exponent;
  if ( d[11] < 0.0 ) d[11] = 0.0; else if ( d[11] > 1.0 ) d[11] = 1.0;

  {
    double X[3], Y[3], Z[3];
    Z[0] = -m_SpotDirection[0];
    Z[1] = -m_SpotDirection[1];
    Z[2] = -m_SpotDirection[2];
    RhinoGetPerpendicular( 3, Z, X );
    RhinoCrossProduct( Z, X, Y );
    double a = RhinoVectorLength( 3, Y );
    if ( a > 0.0 ) {
      a = 1.0/a;
      Y[0] *= a;
      Y[1] *= a;
      Y[2] *= a;
      d[3] = X[0];
      d[4] = X[1];
      d[5] = X[2];
      d[6] = Y[0];
      d[7] = Y[1];
      d[8] = Y[2];
    }
  }

  if ( m_spot_cutoff > 0.0 && m_spot_cutoff < 90.0 ) {
    d[9]  = m_h*tan(m_spot_cutoff * DEGREES_TO_RADIANS);
  }

  // Make sure layer definition is in the file
	file.LayerIndex( Layer() );	// Get layer index or -1

	file.BeginWriteChunk( *this );
  file.WriteDouble( 12, d );
  WriteAttributes( file, TCODE_ENDOFTABLE );
  file.EndChunk();

  return;
}

CRhinoColor CRhinoSpotlight::Ambient() const
{
  return m_ambient;
}

CRhinoColor CRhinoSpotlight::Diffuse() const
{
  return m_diffuse;
}

CRhinoColor CRhinoSpotlight::Specular() const
{
  return m_specular;
}

double CRhinoSpotlight::SpotExponent() const
{
  return m_spot_exponent;
}

double CRhinoSpotlight::SpotCutoff() const
{
  return m_spot_cutoff;
}

void CRhinoSpotlight::GetDirection( double direction[3] ) const
{
  if ( direction ) {
    direction[0] = m_SpotDirection[0];
    direction[1] = m_SpotDirection[1];
    direction[2] = m_SpotDirection[2];
  }
}

void CRhinoSpotlight::GetDirection( double& x, double& y, double& z ) const
{
  x = m_SpotDirection[0];
  y = m_SpotDirection[1];
  z = m_SpotDirection[2];
}

void CRhinoSpotlight::GetPosition( double position[3] ) const
{
  if ( position ) {
    position[0] = m_Position[0];
    position[1] = m_Position[1];
    position[2] = m_Position[2];
  }
}

void CRhinoSpotlight::GetPosition( double& x, double& y, double& z ) const
{
  x = m_Position[0];
  y = m_Position[1];
  z = m_Position[2];
}

void CRhinoSpotlight::SetAmbient( CRhinoColor c )
{
   m_ambient = c;
   CRhinoMaterial* pMat = Material();
   if (pMat)
     pMat->SetAmbient(c);
}

void CRhinoSpotlight::SetDiffuse( CRhinoColor c )
{
   m_diffuse = c;
   CRhinoMaterial* pMat = Material();
   if (pMat)
     pMat->SetDiffuse(c);
}

void CRhinoSpotlight::SetSpecular( CRhinoColor c )
{
   m_specular = c;
   CRhinoMaterial* pMat = Material();
   if (pMat)
     pMat->SetSpecular(c);
}

void CRhinoSpotlight::SetSpotExponent( double e )
{
  if ( e >= 0.0 )
    m_spot_exponent = (e > 1.0) ? 1.0 : e;
}

void CRhinoSpotlight::SetSpotCutoff( double a )
{
  if ( a == 180.0 || (a > 0.0 && a < 90.0) ) {
    m_spot_cutoff = a;
    m_r = m_h;
    if ( a > 0.0 && a < 90.0 ) {
      m_r = m_h*tan(m_spot_cutoff*DEGREES_TO_RADIANS);
    }
  }
}

void CRhinoSpotlight::SetDirection( const double* direction )
{
  double d[3] = {0.0,0.0,-1.0};
  if ( direction ) {
    d[0] = direction[0];
    d[1] = direction[1];
    d[2] = direction[2];
  }
  double len = RhinoVectorLength( 3, d );
  if ( len > 0.0 ) {
    m_h = len;
    m_r = (m_spot_cutoff > 0.0 && m_spot_cutoff < 90.0) 
        ? m_h*tan(m_spot_cutoff*DEGREES_TO_RADIANS)
        : m_h;
    len = 1.0/len;
    m_SpotDirection[0] = len*d[0];
    m_SpotDirection[1] = len*d[1];
    m_SpotDirection[2] = len*d[2];
  }
}

void CRhinoSpotlight::SetDirection( double x, double y, double z )
{
  double d[3];
  d[0] = x;
  d[1] = y;
  d[2] = z;
  SetDirection( d );
}

void CRhinoSpotlight::SetPosition( const double* position )
{
  double d[3] = {0.0,0.0,0.0};
  if ( position ) {
    d[0] = position[0];
    d[1] = position[1];
    d[2] = position[2];
  }
  m_Position[0] = d[0];
  m_Position[1] = d[1];
  m_Position[2] = d[2];
}

void CRhinoSpotlight::SetPosition( double x, double y, double z )
{
  double d[3];
  d[0] = x;
  d[1] = y;
  d[2] = z;
  SetPosition( d );
}

