/* $Header: /Src/RhinoIO/rhinoio_viewport.h 20    1/29/99 2:14p Dalelear $ */
/* $NoKeywords: $ */
//
// Copyright (c) 1993-1998 Robert McNeel & Associates. All rights reserved.
// Rhinoceros is a registered trademark of Robert McNeel & Assoicates.
//
// THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
// ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
// MERCHANTABILITY ARE HEREBY DISCLAIMED.
//				
////////////////////////////////////////////////////////////////
//
//   defines CRhinoViewport
//
////////////////////////////////////////////////////////////////

#if !defined(RHINO_VIEWPORT_INC_)
#define RHINO_VIEWPORT_INC_

///////////////////////////////////////////////////////////////////////////////
// Class  CRhinoViewport
//
//	This object represents a viewing frustum
///////////////////////////////////////////////////////////////////////////////
class CRhinoViewport: public CRhinoObject 
{
	RHINO_CHUNK_DECLARE( CRhinoViewport )
public:
  enum eProjection 
  { 
    parallel_proj    = 1,
    perspective_proj = 2 
  };
  enum eCoordinateSystem 
  { 
    world_cs  = 0, 
    camera_cs = 1, 
    clip_cs   = 2, 
    screen_cs = 3 
  };

  // Construction
	CRhinoViewport();												// create empty spot light
  ~CRhinoViewport();
	CRhinoViewport& operator=( const CRhinoViewport& );

  // CRhinoObject overrides
	CRhinoFile::eStatus Read(CRhinoFile&);	// read chunk data from file
	void Write(CRhinoFile&) const;					// write spot light chunks to file
  BOOL IsValid() const;

  // Interface
  BOOL SetViewName( const char* );
  const char* ViewName() const;

  BOOL SetProjection( eProjection );
  eProjection Projection() const;

  BOOL SetCamera( // Direction and Up must be nonzero and not parallel.
      const double*, // CameraLocation[3]  (NULL for origin)
      const double*, // CameraDirection[3] (NULL for -Z)
      const double*  // CameraUp[3]        (NULL for +Y)
      );
  BOOL GetCamera( // returns values used to set camera position
                  // Use GetCameraFrame() if you want an orthonormal frame.
      double*, // CameraLocation[3]
      double*, // CameraDirection[3]
      double*  // CameraUp[3]
      ) const;
  BOOL GetCameraFrame(
      double*, // CameraLocation[3]
      double*, // CameraX[3]
      double*, // CameraY[3]
      double*  // CameraZ[3]
      ) const;
  BOOL IsCameraFrameWorldPlan( 
      // Returns true if the camera direction = some world axis.
      // The indices report which axes are used.  For a "twisted"
      // plan view it is possible to have zero x and y indices.
      // This function returns true if and only if the "z" index
      // is non-zero.
      //
      // Indices are +/-1 = world +/-x, +/-2 = world +/-y, +/-3 = world +/-z,
      int*, // if true and plan is axis aligned, view x index, else 0
      int*, // if true and plan is axis aligned, view y index, else 0
      int*  // if true, view z index, else 0
      );
  BOOL GetCameraExtents( 
      // returns bounding box in camera coordinates - this is useful information
      // for setting view frustrums to include the point list
      int,           // count = number of 3d points
      int,           // stride = number of doubles to skip between points (>=3)
      const double*, // 3d points in world coordinates
      double[3],     // minimum corner of bounding box in camera coordinates
      double[3],     // maximum corner of bounding box in camera coordinates
      BOOL = FALSE   // set to TRUE if you want to enlarge an existing box
      ) const;
  BOOL GetCameraExtents( 
      // returns bounding box in camera coordinates - this is useful information
      // for setting view frustrums to include the point list
      const double*, // world coordinate bounding box minimum
      const double*, // world coordinate bounding box maximum
      double[3],     // minimum corner of bounding box in camera coordinates
      double[3],     // maximum corner of bounding box in camera coordinates
      BOOL = FALSE   // set to TRUE if you want to enlarge an existing box
      ) const;
  BOOL GetCameraExtents( 
      // returns bounding box in camera coordinates - this is useful information
      // for setting view frustrums to include the point list
      const double*, // world coordinate bounding sphere center
      double,        // world coordinate bounding sphere radius
      double[3],     // minimum corner of bounding box in camera coordinates
      double[3],     // maximum corner of bounding box in camera coordinates
      BOOL = FALSE   // set to TRUE if you want to enlarge an existing box
      ) const;

  // Rhino version 1.0 does not support skew frustums; i.e., the
  // viewing frustums used in Rhino have left = -right, bottom = -top.
  // If you specify a skew frustum, all the CRhinoViewport functions
  // will work fine.  If you save it a file and read the file into Rhino,
  // then you will get a symmetric approximation.
  BOOL SetFrustum(
        double, // left
        double, // right   ( left < right )
        double, // bottom
        double, // top     ( bottom < top )
        double, // near
        double  // far     ( 0 < near < far ) // ignored by Rhino version 1.0
        );
  BOOL GetFrustum(
        double*, // left
        double*, // right  (left < right)
        double*, // bottom
        double*, // top    (bottom < top)
        double* = NULL, // near
        double* = NULL  // far  (0 < near < far)
        ) const;
  BOOL GetFrustumAspect( double& ) const; // frustum's width/height
  BOOL GetFrustumCenter( double* ) const; // world coordinates of frustum's center

  // The near clipping plane stored in the Rhino file is frequently very
  // small and useless for high quality z-buffer based rendering.  The far
  // clipping value is not stored in the file.  Use these functions to set
  // the frustum's near and far clipping planes to appropriate values.
  BOOL SetFrustumNearFar( 
         const double*,  // 3d bounding box min
         const double*   // 3d bounding box max
         );
  BOOL SetFrustumNearFar( 
         const double*,  // 3d bounding sphere center
         double          // 3d bounding sphere radius
         );
  BOOL SetFrustumNearFar( 
         double, // near ( > 0 )
         double  // far
         );

  // Location of viewport in pixels (not saved in file)
  // These are provided so you can set the port you are using and get
  // the appropriate transformation(s).  
  //
  // If you are using Windows, then port_left = 0, port_right = width-1,
  // port_bottom = height-1, port_top = 0;
  BOOL SetScreenPort(
        int,    // port_left
        int,    // port_right    ( port_left != port_right )
        int,    // port_bottom
        int,    // port_top      ( port_bottom != port_top)
        int =0, // port_near
        int =0  // port_far
        );
  BOOL GetScreenPort(
        int*,       // left
        int*,       // right         ( port_left != port_right )
        int*,       // port_bottom
        int*,       // port_top      ( port_bottom != port_top)
        int* =NULL, // port_near
        int* =NULL  // port_far
        ) const;
  BOOL GetScreenPortAspect( double& ) const; // port's |width/height|

  // Relative position of the viewport in Rhino's parent window.
  // For example, to put a view port in the upper left quadrant
  // of the parent window, use left=0,right=0.5,bottom=0.5, top=1.0.
  // This information is saved in the Rhino file.  It does not come 
  // into play in in viewing transformations.  When the Rhino file
  // is loaded, the information is used to position the view window.
  // If the resulting window size has a different aspect than the frustum,
  // the frustum is expanded along the appropriate axis.
  BOOL SetPortPosition( 
        double, // relative_left
        double, // relative_right   ( 0 <= relative_left < relative_right <= 1 )
        double, // relative_bottom
        double  // relative_top     ( 0 <= relative_bottom < relative_top <= 1 )
        );
  BOOL GetPortPosition( 
        double*, // relative_left
        double*, // relative_right   ( 0 <= left < right <= 1 )
        double*, // relative_bottom
        double*  // relative_top     ( 0 <= bottom < top <= 1 )
        ) const;

  // Rhino rotates viewports around a "target" point.. The point lies on
  // center axis fo the frustum and must be between the clipping planes.
  // If necessary, setting this value will modify frus_near and/or frus_far
  // so that 0 < frus_near < target_distance < frus_far.
  BOOL SetTargetDistance(
         double // ratget_distance ( > 0 )
         );
  BOOL GetTargetDistance(double&) const;
  BOOL GetTargetPoint(double*) const;

  // Rhino viewports have a "construction plane" that is displayed as a grid
  BOOL SetConstructionPlane( 
         const double*, // origin
         const double*, // X axis
         const double*  // Y axis
         );
  BOOL GetConstructionPlane( 
         double*, // origin
         double*, // X axis
         double*, // Y axis
         double*  // Z axis
         ) const;
  
  BOOL GetCameraAngle( 
          double*, // 1/2 of diagonal subtended angle
          double*, // 1/2 of vertical subtended angle
          double*  // 1/2 of horizontal subtended angle
          ) const;
  BOOL GetCameraAngle( 
          double*  // 1/2 of smallest subtended view angle
          ) const;
  BOOL SetCameraAngle( 
          double  // 1/2 of smallest subtended view angle
                  // 0 < angle < pi/2
          );

  // These functions assume the camera is horizontal and crop the
  // film rather than the image when the aspect of the frustum
  // is not 36/24.  (35mm film is 36mm wide and 24mm high.)
  //
  // The SetCamera35mmLenseLength() preserves camera location,
  // changes the frustum, but maintains the frsutrum's aspect.
  BOOL GetCamera35mmLenseLength( double* /*lense length*/ ) const;
  BOOL SetCamera35mmLenseLength( double /*lense length*/ );

  BOOL GetXform( 
         eCoordinateSystem, // srcCS
         eCoordinateSystem, // destCS
         CRhinoXform&       // 4x4 transformation matrix (acts on the left)
         ) const;

  // display tools
  BOOL GetCoordinateSprite(
         int,        // size in pixels of coordinate sprite axes
         int, int,   // screen (x,y) for sprite origin
         int[3],     // returns depth order for axes
         int [3][2]  // screen coords for axes ends
         ) const;

  // Use Extents() as a quick way to set a viewport to so that bounding
  // volume is inside of a viewports frusmtrum.
  // The view angle is used to determine the position of the camera.
  BOOL Extents( 
         double,        // 1/2 smallest subtended view angle
                        // (0 < angle < pi/2)
         const double*, // 3d world coordinate bounding box min
         const double*  // 3d world coordinate bounding box max
         );
  BOOL Extents( 
         double,        // 1/2 smallest subtended view angle
                        // (0 < angle < pi/2)
         const double*, // 3d world coordinate bounding sphere center
         double         // 3d sphere radius
         );

  // View changing from screen input points.  Handy for
  // using a mouse to manipulate a view.
  //
  // ZoomToScreenRect() may change camera and frustum settings
  // DollyAlongScreenChord() may change camera settings
  BOOL ZoomToScreenRect(
         int, int,  // (x,y) screen coords of a rectangle corner
         int, int   // (x,y) screen coords of opposite rectangle cornerd
         );
  BOOL DollyAlongScreenChord(
         int, int, // (x,y) screen coords of source point
         int, int  // (x,y) screen coords of destination point
         );

  // Rhino uses a custom legacy format for view projection information
  BOOL FromRhino( // sets standard view parameters from file info
          CRhinoViewport::eProjection, // projection,
          const double*,               // rhvp_target, // 3d point
          double, double, double,      // angles in radians
          double,                      // rhvp_viewsize,     // > 0
          double,                      // rhvp_cameradist    // > 0
          int=0, int=0                 // width and height of screen port
                                       // (use 0,0) if unknown
                                       // Must still call SetScreenPort()
                                       // to complete initialization
          );

  BOOL DollyCamera( // Does not update target and frustum.
                    // To update target and/or frustum use 
                    // DollyTarget(d) and/or DollyFrustum(d) with
                    // d = dollyVector o cameraFrameZ
          const double* // 3d dolly vector
          );
  BOOL DollyTarget(
          double // distance to move in camera direction
          );
  BOOL DollyFrustum(
          double // distance to move in camera direction
          );

protected:

  // These boolean status flags are set to TRUE when the associated
  // fields contain valid values
  BOOL m_bValidCamera;
  BOOL m_bValidFrustum;
  BOOL m_bValidPort;

  // Camera Settings: ///////////////////////////////////////////////
  //   projection
  eProjection m_projection;

  //   Camera location, direction and orientation (in world coordinates).
  //   These values are used to set the camera frame vectors CamX, CamY,
  //   CamZ.  If bValidCamera is TRUE, then the CamX, CamY and CamZ
  //   vectors are properly initialized and should be used
  //   instead of CamDir[] and CamUp[].  The frame vectors CamX, CamY, CamZ
  //   are garenteed to be a right handed orthonormal frame.  The CamDir
  //   and CamUp vectors contain the values passed into TL_SetViewCamera().
  double m_CamLoc[4]; // CamLoc[3]: 1 = finite, 0 = infinite
  double m_CamDir[3]; // nonzero vector from camera towards view 
  double m_CamUp[3];  // nonzero vector not parallel to CamDir

  // The camera frame vectors are properly initialized by the 
  // TL_SetViewCamera() and  TL_CreateView*() functions.
  double m_CamX[3], m_CamY[3], m_CamZ[3]; // Unit camera axes in world coords 

  // View Frustum Settings: ///////////////////////////////////////
  //   left, right are camera X coords on near clipping plane
  //   bottom, top are camera Y coords on near clipping plane
  //   near = distance from camera to near clipping plane
  //   far = distance from camera to far clipping plane
  double m_frus_left,   m_frus_right; // frus_left < frus_right 
  double m_frus_bottom, m_frus_top;   // frus_bottom < frus_top 
  double m_frus_near,   m_frus_far;   // frus_near < frus_far 
                                  // in perspective, 0 < frus_near

  // Device View Port Box Settings: ( in display device coordinates ) ////
  //   The point (left,bottom,-near), in camera coordinates, of the view
  //   frustum is mapped to pixel coordinate (port_left,port_bottom,port_near).
  //   The point (right,top,-far), in camera coordinates, of the view frustum 
  //   is mapped to pixel coordinate (port_right,port_top,port_far).
  int m_port_left,   m_port_right; // port_left != port_right
  int m_port_bottom, m_port_top;   // port_bottom != port_top  
                                   // In many situations includeing Windows,
                                   //   port_left  = 0 
                                   //   port_right = viewport width-1
                                   //   port_top   = 0
                                   //   port_bottom = viewport height-1
  int m_port_near,   m_port_far;   // (If you want an 8 bit z-buffer with 
                                   // z=255 being "in front of" z=0, then
                                   // set port_near = 255 and port_far = 0.)

private:
  double m_target_distance; // ( > 0 ) distance from camer to target point
  double ViewSize() const;  // "radius" of view at target point
  BOOL GetAngles( double*, double*, double* ) const;

  // construction plane
  double m_cp_origin[3];
  double m_cp_X[3];
  double m_cp_Y[3];
  double m_cp_Z[3];

  // Rhino client area
  double m_client_x[2]; // 0 to 1 
  double m_client_y[2];

  // Grid display flags
  BOOL m_bShowGrid;
  BOOL m_bShowGridAxes;
  BOOL m_bShowWorldAxes;


  char* m_sViewName;
};

#endif

