/* $Header: /Src/RhinoIO/rhinoio_xform.h 14    11/10/98 12:35p Dalelear $ */
/* $NoKeywords: $ */
//
// Copyright (c) 1993-1998 Robert McNeel & Associates. All rights reserved.
// Rhinoceros is a registered trademark of Robert McNeel & Assoicates.
//
// THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
// ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
// MERCHANTABILITY ARE HEREBY DISCLAIMED.
//				
////////////////////////////////////////////////////////////////
//
//   defines CRhinoXform (4 x 4 transformation matrix)
//
////////////////////////////////////////////////////////////////

#if !defined(RHINO_XFORM_INC_)
#define RHINO_XFORM_INC_

class CRhinoXform
{
public:
  CRhinoXform();                     // zero matrix
  CRhinoXform( const CRhinoXform& ); // copy constructor
  CRhinoXform( double );             // diagonal matrix
  CRhinoXform( double[4][4] );       // from old standard 4x4 matrix
  CRhinoXform( const double* );      // from array of 16 doubles

  ~CRhinoXform();
  
  CRhinoXform& operator=( const CRhinoXform& );
  CRhinoXform& operator=( int );
  CRhinoXform& operator=( float );
  CRhinoXform& operator=( double );
  
  // All non-commutative operations have "this" as left hand side and
  // argument as right hand side.
  CRhinoXform operator*( double ) const;
  CRhinoXform operator*( const CRhinoXform& /*rhs*/ ) const;
  CRhinoXform operator+( double ) const;
  CRhinoXform operator+( const CRhinoXform& ) const;
  CRhinoXform operator-( double /*rhs*/ ) const;
  CRhinoXform operator-( const CRhinoXform& /*rhs*/ ) const;

  CRhinoXform& operator*=( double );
  CRhinoXform& operator*=( const CRhinoXform& /*rhs*/ );
  CRhinoXform& operator+=( double );
  CRhinoXform& operator+=( const CRhinoXform& );
  CRhinoXform& operator-=( double /*rhs*/ );
  CRhinoXform& operator-=( const CRhinoXform& /*rhs*/ );
  
  double* operator[](int);
  const double* operator[](int) const;

  CRhinoXform& Identity();
  CRhinoXform& Zero();

  CRhinoXform& Translation( double, double, double );
  CRhinoXform& Translation( const double* );

  CRhinoXform& Scale( double, double, double );
  CRhinoXform& Scale( const double* );

  CRhinoXform& Rotation(
    const double*,  // 3d unit axis (NULL for Z axis)
    const double*,  // 3d center of rotation (NULL for origin)
    double,         // sin(rotation angle)
    double          // cos(rotation angle)
    );

  CRhinoXform& Rotation(
    double, double, double, // 3d unit axis (NULL for Z axis)
    double, double, double, // 3d center of rotation (NULL for origin)
    double,         // sin(rotation angle)
    double          // cos(rotation angle)
    );

  CRhinoXform& WorldToFrame(
    const double*, // 3d center
    const double*, // unitX
    const double*, // unitY
    const double*  // unitZ   (CameraX-Z = orthonormal frame)
    );

  CRhinoXform& FrameToWorld(
    const double*, // 3d center
    const double*, // unitX
    const double*, // unitY
    const double*  // unitZ   (CameraX-Z = orthonormal frame)
    );

  CRhinoXform& CameraToClip( // maps viewport frustum to -1 <= x,y,z <= 1 box
      BOOL bPerspective,
      double, double, // left != right (usually left < right )
      double, double, // bottom != top (usually bottom < top )
      double, double  // near != far (usually 0 < near < far )
      );

  CRhinoXform& ClipToCamera( // maps -1 <= x,y,z <= 1 box to viewport frustum
      BOOL bPerspective,
      double, double, // left != right (usually left < right )
      double, double, // bottom != top (usually bottom < top )
      double, double  // near != far an bot are non-zero (usually 0 < near < far )
      );

  CRhinoXform& ClipToScreen(
      double, double, // left, right,
      double, double, // bottom, top,
      double, double  // near_z, far_z
      );

  CRhinoXform& ScreenToClip(
      double, double, // left, right,
      double, double, // bottom, top,
      double, double  // near_z, far_z
      );

  CRhinoXform& Transpose();

  double Determinant() const;
  CRhinoXform Inverse() const;
  BOOL Invert();

  BOOL Apply(  // apply transformation on the left of the 2d point
    double,    // euclidean point x
    double,    // euclidean point y
    double[2]  // transformed 2d point in euclidean coordinates
    ) const;

  BOOL Apply(  // apply transformation on the left of the 3d point
    double,    // euclidean point x
    double,    // euclidean point y
    double,    // euclidean point z
    double[3]  // transformed 3d point in euclidean coordinates
    ) const;

  BOOL Apply(  // apply transformation on the left of the homogeneous point
    double,    // homogeneous point x
    double,    // homogeneous point y
    double,    // homogeneous point z
    double,    // homogeneous point w
    double[4]  // transformed homogeneous point in homogeneous coordinates
    ) const;

  BOOL ApplyVector( // apply transformation on the left of the 3d vector
    double,    // euclidean vector delta x
    double,    // euclidean vector delta y
    double,    // euclidean vector delta z
    double[3]  // transformed 3d vector in euclidean coordinates
    ) const;

  // returns clipflag with
  //   clipflag &  1 if transformed x is coordinate < -1
  //   clipflag &  2 if transformed x is coordinate >  1
  //   clipflag &  4 if transformed y is coordinate < -1
  //   clipflag &  8 if transformed y is coordinate >  1
  //   clipflag & 16 if transformed z is coordinate < -1
  //   clipflag & 32 if transformed z is coordinate >  1
  // illegal input returns 1|2|4|8|16|32 = 63
  int ClipFlag3d(
    const double* // 3d point
    ) const;
  int ClipFlag3d(
    int,          // number of points
    int,          // stride ( >= 3 )
    const double* // points
    ) const;
  int ClipFlag3dBox(
    const double*, // 3d boxmin
    const double*  // 3d boxmax
    ) const;
  int ClipFlag4d(
    const double* // 4d point
    ) const;
  int ClipFlag4d(
    int,          // number of points
    int,          // stride ( >= 4 )
    const double* // points
    ) const;

  double m_xform[4][4];

  BOOL Test() const;
};

BOOL RhinoGetRotationAngles( 
       // Determine angles so that
       // R3*R2*R1 maps world X, Y, Z to the input frame, where
       // R1 = rotation of angle 1 abount world Z
       // R2 = rotation of angle 2 abount world X
       // R3 = rotation of angle 3 abount world Z
       // (Useful in some old-fashioned rendering calculations.)
       const double*, const double*, const double*, // right handed 3d ortho normal frame
       double*, // angle1
       double*, // angle2
       double*  // angle3
       );

#endif
