#if !defined(RHMEM_INC_)
#define RHMEM_INC_

#if defined (cplusplus) || defined(_cplusplus) || defined(__cplusplus)
extern "C" {
#endif

/*
// These functions are used for all memory management in Rhino and
// any code used by Rhino.  They mimic the general behavior of
// the standard C library non-rh versions.
//
// Details:
//
// void* rhmalloc( site_t sz );
//
//   If sz is zero, NULL is returned.
//   If sz is positive and there is not enough memory to satify
//   the allocation request, the rhmem_error_handler(0) is called
//   and NULL is returned.
//        
//       
// void* rhcalloc( size_t num, size_t sz );
//
//   If sz or num is zero, NULL is returned.
//   If sz and num are positive and there is not enough memory to satify
//   the allocation request, the rhmem_error_handler(0) is called
//   and NULL is returned.
//               
//               
// void* rhrealloc( void* p, size_t sz );
//
//   If p is NULL, then rhmalloc(sz) is called.
//   If sz is zero, then rhfree(p) is called and
//   NULL is returned.
//   If p is not NULL, sz is positive, and there is not
//   enough memory to to satify the allocation request,
//   the rhmem_error_handler(0) is called and NULL is returned.
//   If p is not NULL and is known to be invalid, then
//   rhmem_error_handler(1) is called.
//               
//
// void rhfree( void* p );
//
//   NULL p is tolerated but considered poor style.
//   If p is not NULL and is known to be invalid, then
//   rhmem_error_handler(2) is called.
//               
//
// void* rhmemdup( const void* src, size_t sz );
//
//   If src is not NULL and sz is positive, then rhmalloc( sz )
//   is called to get memory, sz bytes of src are copied into this
//   memory, and the pointer to this memory is returned.
//   If rhmalloc() returns NULL, then NULL is returned.
//   If src is NULL or sz is zero, then NULL is returned.
//               
//
// char* rhstrdup( const char* src );
//
//   If src is not NULL, then rhmemdup( sc, (strlen(src)+1)*sizeof(src[0]) )
//   is called.  If rhmemdup() returns NULL, then NULL is returned.
//   If src is NULL, then NULL is returned.
//               
//
// rhmem_error_handler rhmem_register_error_handler( rhmem_error_handler my_handler);
//
//    Use this function to specify the error handler function that will be called
//    if an error occures.  See the description of rhmem_error_handler() for the
//    error handler function specification.
//
// int rhmem_error_handler( int error_number );
//
//   This function is called when an error occurs.  The values of error_number
//   are:
//
//     0: There is not enough memory to satisfy an rhmalloc() or rhrealloc()
//        request.
//
//        If rhmem_error_handler() returns 0, then rhmalloc() or rhrealloc()
//        returns NULL.  If rhmem_error_handler() returns 1, the allocation
//        is attempted again.
//
//     1: An invalid pointer was passed to rhrealloc().
//
//        If rhmem_error_handler() returns 0, then rhrealloc() returns NULL.
//        If rhmem_error_handler() returns 1, the deallocation is attempted
//        again;
//
//     2: An invalid pointer was passed to rhfree().
//
//        If rhmem_error_handler() returns 0, then free() returns.
//        If rhmem_error_handler() returns 1, the deallocation is
//        attempted again;
//
*/

/* memory manager */
void*  rhmalloc( size_t );
void*  rhcalloc( size_t, size_t );
void   rhfree( void* );
void*  rhrealloc( void*, size_t );
void*  rhmemdup( const void*, size_t );

char*  rhstrdup( const char* );
unsigned char* rhmbsdup( const unsigned char* );
#if defined(_WCHAR_T_DEFINED)
wchar_t* rhwcsdup( const wchar_t* );
#endif

/* define to handle _TCHAR* rhtcsdup( const _TCHAR* ) */
#if defined(_UNICODE)
#define rhtcsdup rhwcsdup
#elif defined(_MBCS)
#define rhtcsdup rhmbsdup
#else
#define rhtcsdup rhstrdup
#endif

/* error handling */
typedef int (*rhmem_error_handler)(int);
rhmem_error_handler rhmem_register_error_handler( rhmem_error_handler );

#if defined (cplusplus) || defined(_cplusplus) || defined(__cplusplus)
}
#endif

#endif
